/*!
 *  \file  IrisCConnection.h
 *  \brief IrisConnectionInterface implementation based on IrisC
 *  \date  Copyright ARM Limited 2017. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisCConnection_h
#define ARM_INCLUDE_IrisCConnection_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisInterface.h"
#include "iris/detail/IrisC.h"
#include "iris/detail/IrisErrorException.h"
#include "iris/detail/IrisDSO.h"

#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief Provide an IrisConnectionInterface which loads an IrisC library.
 * @see IrisClientConnection
 * @see IrisCoreConnection
 */
class IrisCConnection : public IrisConnectionInterface
{
private:
    /// IrisC API functions
    IrisC_HandleMessageFunction handleMessage_function;
    IrisC_ObtainInterfaceFunction obtainInterface_function;

    IrisC_RegisterChannelFunction registerChannel_function;
    IrisC_UnregisterChannelFunction unregisterChannel_function;

    IrisC_ProcessAsyncMessagesFunction processAsyncMessages_function;

    class RemoteInterface : public IrisInterface
    {
    private:
        IrisCConnection *irisc;

    public:
        RemoteInterface(IrisCConnection *irisc_)
            : irisc(irisc_)
        {
        }

    public: // IrisInterface
        virtual void irisHandleMessage(const uint64_t *message) IRIS_OVERRIDE
        {
            // Forward to the IrisC library
            int64_t status = irisc->IrisC_handleMessage(message);

            if (status != E_ok)
            {
                throw IrisErrorException(IrisErrorCode(status));
            }
        }

        virtual void *irisObtainInterface(IrisInstId instId, const char *requested_interface) IRIS_OVERRIDE
        {
            void *interface = NULL;

            int64_t status = irisc->IrisC_obtainInterface(instId, requested_interface, &interface);

            if (status != E_ok)
            {
                throw IrisErrorException(IrisErrorCode(status));
            }

            return interface;
        }
    } remote_interface;

    // Helper function to bridge IrisC_HandleMessageFunction to IrisInterface::irisHandleMessage
    static int64_t handleMessageToIrisInterface(void *context, const uint64_t *message)
    {
        if (context == NULL)
        {
            return E_invalid_context;
        }
        static_cast<IrisInterface*>(context)->irisHandleMessage(message);

        return E_ok;
    }

    // Helper function to bridge IrisC_ObtainInterfaceFunction to IrisInterface::obtainInterface
    static int64_t obtainIntefaceToIrisInterface(void *context, uint64_t instId,
                                                 const char *requested_interface,
                                                 void **interface_ptr_out)
    {
        if (context == NULL)
        {
            return E_invalid_context;
        }
        IrisInterface *iris_interface  = static_cast<IrisInterface*>(context);
        *interface_ptr_out = iris_interface->irisObtainInterface(instId, requested_interface);

        return E_ok;
    }

protected:
    /// Context pointer to use when calling IrisC_* functions. This is needed by subclasses as well.
    void *iris_c_context;

    /// Construct and empty object. Used by subclasses that needs to load a DSO and call init()
    IrisCConnection()
        : handleMessage_function(NULL)
        , obtainInterface_function(NULL)
        , registerChannel_function(NULL)
        , unregisterChannel_function(NULL)
        , processAsyncMessages_function(NULL)
        , remote_interface(this)
        , iris_c_context(NULL)
    {
    }

    /// Initialize by reading symbols from a DSO
    void init(IrisDSO &irisC_dso, const std::string &sym_prefix)
    {
#define INTERNAL_LOOKUP(name) irisC_dso.lookupSymbol(sym_prefix + #name, name ## _function)
        INTERNAL_LOOKUP(handleMessage);
        INTERNAL_LOOKUP(obtainInterface);
        INTERNAL_LOOKUP(registerChannel);
        INTERNAL_LOOKUP(unregisterChannel);
        INTERNAL_LOOKUP(processAsyncMessages);
#undef INTERNAL_LOOKUP
    }

    /// Wrapper functions to call the underlying IrisC functions

    int64_t IrisC_handleMessage(const uint64_t *message)
    {
        return (*handleMessage_function)(iris_c_context, message);
    }

    int64_t IrisC_obtainInterface(uint64_t instId, const char *requested_interface, void **interface_ptr_out)
    {
        return (*obtainInterface_function)(iris_c_context, instId, requested_interface, interface_ptr_out);
    }

    int64_t IrisC_registerChannel(IrisC_CommunicationChannel *channel, uint64_t *channel_id_out)
    {
        return (*registerChannel_function)(iris_c_context, channel, channel_id_out);
    }

    int64_t IrisC_unregisterChannel(uint64_t channel_id)
    {
        return (*unregisterChannel_function)(iris_c_context, channel_id);
    }

    int64_t IrisC_processAsyncMessages(uint64_t flags)
    {
        return (*processAsyncMessages_function)(iris_c_context, flags);
    }

public:
    IrisCConnection(IrisC_Functions *functions)
        : handleMessage_function(functions->handleMessage_function)
        , obtainInterface_function(functions->obtainInterface_function)
        , registerChannel_function(functions->registerChannel_function)
        , unregisterChannel_function(functions->unregisterChannel_function)
        , processAsyncMessages_function(functions->processAsyncMessages_function)
        , remote_interface(this)
        , iris_c_context(functions->iris_c_context)
    {
    }

    /// Populate an IrisC_Functions struct with the functions used by this instance
    void getIrisCFunctions(IrisC_Functions &functions) const
    {
        functions.Functions_version = 0;
        functions.iris_c_context = iris_c_context;
        functions.registerChannel_function = registerChannel_function;
        functions.unregisterChannel_function = unregisterChannel_function;
        functions.handleMessage_function = handleMessage_function;
        functions.obtainInterface_function = obtainInterface_function;
        functions.processAsyncMessages_function = processAsyncMessages_function;
    }

public: // IrisConnectionInterface
    /*!
     * @brief Register a communication channel.
     * @see IrisConnectionInterface::registerIrisInterfaceChannel
     */
    virtual uint64_t registerIrisInterfaceChannel(IrisInterface *iris_interface) IRIS_OVERRIDE
    {
        IrisC_CommunicationChannel channel;

        channel.CommunicationChannel_version = 0;
        channel.handleMessage_function = &IrisCConnection::handleMessageToIrisInterface;
        channel.handleMessage_context = static_cast<void *>(iris_interface);
        channel.obtainInterface_function = &IrisCConnection::obtainIntefaceToIrisInterface;
        channel.obtainInterface_context = static_cast<void *>(iris_interface);

        uint64_t channelId = IRIS_UINT64_MAX;

        IrisErrorCode status = static_cast<IrisErrorCode>(IrisC_registerChannel(&channel, &channelId));

        if (status != E_ok)
        {
            throw IrisErrorException(status);
        }

        return channelId;
    }

    /*!
     * @brief Unregister a communication channel.
     * @see IrisConnectionInterface::unregisterIrisInterfaceChannel
     */
    virtual void unregisterIrisInterfaceChannel(uint64_t channelId) IRIS_OVERRIDE
    {
        IrisErrorCode status = static_cast<IrisErrorCode>(IrisC_unregisterChannel(channelId));

        if (status != E_ok)
        {
            throw IrisErrorException(status);
        }
    }

    /*!
     * @brief Process asynchronous messages for the calling thread.
     * @see IrisConnectionInterface::processAsyncMessages
     */
    virtual IrisErrorCode processAsyncMessages(uint64_t flags) IRIS_OVERRIDE
    {
        return static_cast<IrisErrorCode>(IrisC_processAsyncMessages(flags));
    }

    /*!
     * @brief Get the IrisInterface for this connection.
     * @see IrisConnectionInterface::getIrisInterface
     */
    virtual IrisInterface *getIrisInterface() IRIS_OVERRIDE
    {
        return &remote_interface;
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisCConnection_h
