/*!
 *  \file  IrisClientConnection.h
 *  \brief Load, manage, and interface with the IrisClient library.
 *  \date  Copyright ARM Limited 2017. All rights reserved.
 */

#include "iris/detail/IrisCommon.h"
#include "iris/IrisCConnection.h"

#ifndef ARM_INCLUDE_IrisClientConnection_h
#define ARM_INCLUDE_IrisClientConnection_h

NAMESPACE_IRIS_START

/*!
 * @brief Provide an IrisConnectionInterface which loads the IrisClient library.
 * @see IrisCoreConnection
 */
class IrisClientConnection : public IrisCConnection
{
private:
    IrisDSO irisClient_dso;

    // IrisCore lifecycle functions

    IrisC_IrisClientConnectFunction connect_function;
    IrisC_IrisClientDisconnectFunction disconnect_function;


    int64_t IrisClient_connect(const char *hostname, uint16_t port)
    {
        return (*connect_function)(&iris_c_context, hostname, port);
    }

    int64_t IrisClient_disconnect()
    {
        return (*disconnect_function)(iris_c_context);
    }


public:
    /*!
     * @brief Get the default IrisClient library name for the host platform.
     * @return      Returns a static string containing the default name of the IrisClient library
     *              for the host platform.
     */
    static const std::string &getDefaultLibName()
    {
#if defined(_WIN32)
        #define INTERNAL_LIB_NAME   "IrisClient.dll"
#elif defined(__APPLE__)
        #define INTERNAL_LIB_NAME   "libIrisClient.dylib"
#elif defined(__unix__)
        #define INTERNAL_LIB_NAME   "libIrisClient.so"
#else
        #error "Unsupported platform"
#endif
        static std::string lib_name(INTERNAL_LIB_NAME);
#undef INTERNAL_LIB_NAME
        return lib_name;
    }

    /*!
     * @brief Load IrisClient library and use it to connect to an Iris server.
     *
     * @param   hostname        Hostname of the Iris server. This can be an IP address. Examples:
     *                                        |
     *                          ------------- | ------------------------------------------------------------
     *                          "192.168.0.5" | IP address of a different host.
     *                          "127.0.0.1"   | Loopback IP address to connect to a server on the same machine.
     *                          "localhost"   | Hostname of the loopback interface.
     *                          "foo.bar.com" | Hostname of a remote machine.
     * @param   port            Server port number to connect to on the host.
     * @param   library_name    Name of the IrisClient library to load. This defaults to the default name for
     *                          the library on the host platform.
     */
    IrisClientConnection(const std::string &hostname, uint16_t port,
                         const std::string &library_name = IrisClientConnection::getDefaultLibName())
        : IrisCConnection()
        , irisClient_dso(library_name)
        , connect_function(NULL)
        , disconnect_function(NULL)
    {
        IrisCConnection::init(irisClient_dso, "IrisClient_");
#define LOOKUP(name) irisClient_dso.lookupSymbol("IrisClient_" #name, name ## _function)
        LOOKUP(connect);
        LOOKUP(disconnect);
#undef LOOKUP

        int64_t code = IrisClient_connect(hostname.c_str(), port);
        if (code != iris::E_ok)
        {
            throw IrisErrorExceptionString(IrisErrorCode(code), "Failed to initialise IrisClient library");
        }
    }

    /*!
     * @brief Disconnect from the server and unload the IrisClient library.
     */
    ~IrisClientConnection()
    {
        int64_t status = IrisClient_disconnect();
        // Not a lot we can do if disconnect fails but asserting success means
        // we will at least discover the problem in debug builds.
        assert(status == iris::E_ok);
        (void)status;
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisClientConnection_h
