/*!
 *  \file  IrisCoreConnection.h
 *  \brief Load, manage, and interface with the IrisCore library.
 *  \date  Copyright ARM Limited 2017. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisCoreConnection_h
#define ARM_INCLUDE_IrisCoreConnection_h

#include "iris/detail/IrisCommon.h"
#include "iris/IrisCConnection.h"

NAMESPACE_IRIS_START

/*!
 * @brief Provide an IrisConnectionInterface which loads the IrisCore library.
 * @see IrisClientConnection
 */
class IrisCoreConnection : public IrisCConnection
{
private:
    IrisDSO irisCore_dso;

    // IrisCore lifecycle functions

    IrisC_IrisCoreInitFunction init_function;
    IrisC_IrisCoreShutdownFunction shutdown_function;


    int64_t IrisCore_init()
    {
        return (*init_function)(&iris_c_context);
    }

    int64_t IrisCore_shutdown()
    {
        return (*shutdown_function)(iris_c_context);
    }

public:
    static const std::string &getDefaultLibName()
    {
#if defined(_WIN32)
        #define LIB_NAME   "IrisCore.dll"
#elif defined(__APPLE__)
        #define LIB_NAME   "libIrisCore.dylib"
#elif defined(__unix__)
        #define LIB_NAME   "libIrisCore.so"
#else
        #error "Unsupported platform"
#endif
        static std::string lib_name(LIB_NAME);
#undef LIB_NAME
        return lib_name;
    }

    /*!
     * @brief Load IrisCore library and use it to connect to an Iris server.
     *
     * @param   library_name    Name of the IrisCore library to load. This defaults to the default name for
     *                          the library on the host platform.
     */
    explicit IrisCoreConnection(const std::string &library_name=getDefaultLibName())
        : IrisCConnection()
        , irisCore_dso(library_name)
        , init_function(NULL)
        , shutdown_function(NULL)
    {
        IrisCConnection::init(irisCore_dso, "IrisCore_");
#define LOOKUP(name) irisCore_dso.lookupSymbol("IrisCore_" #name, name ## _function)
        LOOKUP(init);
        LOOKUP(shutdown);
#undef LOOKUP

        if (IrisCore_init() != iris::E_ok)
        {
            throw IrisInternalError("Failed to initialise IrisCore library");
        }
    }

    /*!
     * @brief Shutdown Iris and unload the IrisCore library.
     */
    ~IrisCoreConnection()
    {
        int64_t status = IrisCore_shutdown();
        // Not a lot we can do if shutdown fails but asserting success means
        // we will at least discover the problem in debug builds.
        assert(status == iris::E_ok);
        (void)status;
    }
};


NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisCoreConnection_h
