/*!
 *  \file   IrisEventEmitter.h
 *  \brief  A utility class for emitting Iris Events.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisEventEmitter_h
#define ARM_INCLUDE_IrisEventEmitter_h

#include "iris/detail/IrisEventEmitterBase.h"

NAMESPACE_IRIS_START

/*!
 * @brief A helper class for generating Iris events
 *
 * @tparam  ARGS    Argument types corresponding to the fields in this event.
 *
 * Use IrisEventEmitter with IrisInstanceBuilder to add events to you Iris instance.
 * @code
       // Declare an event emitter
       iris::IrisEventEmitter<uint64_t, bool> my_event;

       // Add it to an Iris instance
       iris::IrisInstance my_instance(...);
       my_instance->getBuilder()->addEventSource("MY_EVENT", my_event)
           .addField("FOO",  "uint", 8, "A value")
           .addField("FLAG", "bool", 1, "A flag");

       // Emit an event
       my_event(0x1234, true);
   @endcode
 */
template <typename... ARGS>
class IrisEventEmitter : public IrisEventEmitterBase
{
public:
    /*! @brief Construct an event emitter */
    IrisEventEmitter()
        : IrisEventEmitterBase(sizeof...(ARGS))
    {
    }

    /*!
     * @brief Emit an event.
     *
     * The arguments to this function are the fields of the event source in the
     * same order that they appear in the template arguments to the IrisEventEmitter
     * class.
     */
    void operator() (ARGS... args)
    {
        emitEvent(args...);
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisEventEmitter_h
