/*!
 *  \file   IrisInstance.h
 *  \brief  Boiler-plate code for an Iris instance (client and components, etc).
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  The IrisInstance class offers:
 *  - infrastructure which is necessary for all Iris instances
 *  - infrastructure which is useful for Iris components
 *  - infrastructure which is useful for Iris clients
 *
 *  It must be stressed that it is optional to use this class to implement a correct Iris interface.
 *  This class does not form an interface between instances. It just forms an interface between
 *  the code of an instance and this class.
 *
 *  This class is useful for and used by both, component and clients.
 */

#ifndef ARM_INCLUDE_IrisInstance_h
#define ARM_INCLUDE_IrisInstance_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisFunctionDecoder.h"
#include "iris/detail/IrisCppAdapter.h"
#include "iris/detail/IrisDelegate.h"

#include "iris/IrisInstanceBuilder.h"

#include <cassert>

NAMESPACE_IRIS_START

/*!
 * @brief Event Callback Delegate
 *
 *  Used to register a function that can receive event callbacks.
 *
 * @code
        iris::IrisErrorCode ec_FOO(uint64_t esId, const iris::IrisValueMap &fields, uint64_t time,
                                   uint64_t sInstId, bool syncEc, std::string &errorMessageOut)
 * @endcode
 *
 *  Example:
 * @code
    class MyEventCallback
    {
    public:
        iris::IrisErrorCode impl_ec_FOO(uint64_t esId, const iris::IrisValueMap &fields, uint64_t time,
                                        uint64_t sInstId, bool syncEc, std::string &errorMessageOut)
        {
            ...
            return E_ok;
        }
    };

   iris_instance->irisRegisterEventCallback(my_event_callback, MyEventCallback, ec_FOO, "Handle event FOO");
   @endcode
 */
typedef IrisDelegate<uint64_t, const IrisValueMap&, uint64_t, uint64_t, bool, std::string&> EventCallbackDelegate;

class IrisInstantiationContext;

class IrisInstance
{
public:
    // --- Construction and destruction. ---

    /*!
     * @brief Register an Iris function implementation.
     * The function may be implemented in this class or in any other class.
     * The helper macro is here to avoid repeating the function name.
     * The 'impl_' prefix limits namespace pollution somewhat.
     */
    #define irisRegisterFunction(instancePtr, instanceType, functionName, functionInfoJson) registerFunction(instancePtr, #functionName, &instanceType::impl_##functionName, functionInfoJson, #instanceType)

    /*! @brief Register an event callback function using an EventCallbackDelegate */
    #define irisRegisterEventCallback(instancePtr, instanceType, functionName, description) \
        registerEventCallback<instanceType, &instanceType::impl_##functionName>(instancePtr, #functionName, \
                                                                                description, #instanceType)
    /*!
     * @defgroup InstanceFlags Instance Flags
     * @brief Flags that can be set when registering an IrisInstance.
     * @{
     */

    /*! @brief Uniquify instance name when registering. */
    static const uint64_t UNIQUIFY         = (1 << 0);

    /*! @brief Marshal future incoming messages on to the current thread. */
    static const uint64_t MARSHAL_REQUESTS = (1 << 1);

    /*! @brief Throw an exception when an Iris call returns an error response. */
    static const uint64_t THROW_ON_ERROR   = (1 << 2);

    /*! @brief Default flags used if not otherwise specified */
    static const uint64_t DEFAULT_FLAGS    = THROW_ON_ERROR;

    /*!
     * @}
     */


    /*!
     * @brief Construct a new Iris instance.
     *
     * @param   connection_interface    The IrisConnectionInterface that this instance
     *                                  should use to connect to the simulation.
     * @param   instName                Name of the instance. This should be prefixed with
     *                                  one of:
     *                                      - "client.".
     *                                      - "component.".
     *                                      - "framework.".
     *                                  as appropriate.
     * @param flags                     A bitwise or of @ref InstanceFlags.
     *                                  Client instances should usually set the flag
     *                                  iris::IrisInstance::UNIQUIFY.
     *                                  Component instances should usually set the flag
     *                                  iris::IrisInstance::MARSHAL_REQUESTS.
     */
    IrisInstance(IrisConnectionInterface *connection_interface = 0,
                 const std::string& instName = std::string(),
                 uint64_t flags = DEFAULT_FLAGS,
                 IrisThreadAbstraction *threadAbstraction = 0);

    /*!
     * @brief Construct a new Iris instance using an IrisInstantiationContext.
     *
     * @param   context    A context object that provides the necessary information
     *                     to instantiate an instance.
     */
    IrisInstance(IrisInstantiationContext *context,
                 IrisThreadAbstraction *threadAbstraction = 0);


    // Destructor.
    ~IrisInstance()
    {
        if (registered)
        {
            unregisterInstance();
        }
        delete builder;
        delete allocatedThreadAbstraction;
        delete simulation_run_state_mutex;
        delete state_changed_event_registry;
    }


    /*!
     * @brief Set remote connection interface.
     *
     * Used to set the IrisConnectionInterface if it was not set in the constructor.
     *
     * @param   connection_interface    The interface used to connect to an Iris simulation.
     */
    void setConnectionInterface(IrisConnectionInterface *connection_interface);

    /*!
     * @brief Process async requests.
     * Use this to keep the Iris system running while a thread is blocked waiting for something.
     */
    void processAsyncRequests();

    /*!
     * @brief Get remote iris interface.
     *
     * @return      Returns the IrisInterface that this instance sends requests and responses to.
     */
    IrisInterface *getRemoteIrisInterface()
    {
        return remoteIrisInterface;
    }


    /*!
     * @brief Set default error behavior for irisCall().
     *
     * @param   throw_on_error      If true calls made using irisCall() that respond with an
     *                              error response will throw an exception (the same behaviour
     *                              as irisCallThrow()).
     *                              If false calls made using irisCall() that respond with an
     *                              error response will return the error code and not throw an
     *                              exception (the same behaviour as irisCallNoThrow()).
     */
    void setThrowOnError(bool throw_on_error)
    {
        default_cppAdapter = throw_on_error? &throw_cppAdapter : &nothrow_cppAdapter;
    }

    /*!
     * @brief Register this instance if it wasn't registered when constructed.
     *
     * @param   instName                Name of the instance. This should be prefixed with
     *                                  one of:
     *                                      - "client.".
     *                                      - "component.".
     *                                      - "framework.".
     *                                  as appropriate.
     * @param flags                     A bitwise or of @ref InstanceFlags.
     *                                  Client instances should usually set the flag
     *                                  iris::IrisInstance::UNIQUIFY.
     *                                  Component instances should usually set the flag
     *                                  iris::IrisInstance::MARSHAL_REQUESTS.
     */
    IrisErrorCode registerInstance(const std::string &instName, uint64_t flags = DEFAULT_FLAGS);

    /*!
     * @brief Unregister this instance.
     *
     * Iris calls must not be made after the instance is unregistered.
     */
    IrisErrorCode unregisterInstance();


    /*!
     * @brief Set/add instance property.
     *
     *  This creates a new property or overwrites an existing one.
     *  Properties (name and value) are defined by the instance that has the property.
     *  (Properties are not to be confused with parameters where values are defined by clients
     *  or parent components and some parameters may change at runtime.)
     *  Properties are exposed by the function instance_getProperties().
     *  This should only ever be called upon initialisation, before other components have a chance
     *  to call instance_getProperties().
     *  Properties are constant and should not be changed at runtime.
     *  T may be bool, uint64_t, int64_t or std::string.
     *
     *  Note: To add a verbatim string property use @code setProperty("myStrProp", std::string("foo")) @endcode
     *  because C strings get caught by the bool overload of setProperty().
     *
     *  @param  propertyName    Name of the property.
     *  @param  propertyValue   Value of the property.
     */
    template<class T>
    void setProperty(const std::string &propertyName, const T &propertyValue)
    {
        assert(!instance_getProperties_called);
        propertyMap[propertyName].set(propertyValue);
    }


    // --- Interface for components. Provide functionality to clients. ---

    /*!
     * @brief Register an Iris function implementation.
     *
     *  This macros can be used instead of calling this function to avoid specifying the function name twice:
     *  irisRegisterFunction(instancePtr, instanceType, functionName, implFunctionName, functionInfoJson)
     *
     * @param  instance            An instance of class T on which to call the member function.
     * @param  name                Name of the function as it will be published.
     * @param  memberFunctionPtr   Pointer to the C++ implementation of the function.
     * @param  functionInfoJson    A string containing JSON encoded FunctionInfo object for this function.
     * @param  instanceTypeStr     The name of class T. This is only used for logging purposes.
     */
    template<class T>
    void registerFunction(T *instance, const std::string &name, void (T::*memberFunctionPtr)(IrisReceivedRequest &), const std::string &functionInfoJson, const std::string &instanceTypeStr)
    {
        functionDecoder.registerFunction(instance, name, memberFunctionPtr, functionInfoJson, instanceTypeStr);
    }

    /*!
     * @brief Unregister a function that was previously registered with registerFunction/irisRegisterFunction.
     */
    void unregisterFunction(const std::string &name)
    {
        functionDecoder.unregisterFunction(name);
    }

    /*!
     * @brief Register a general event callback.
     *
     * Event callbacks have the same signature, and only the description is different.
     *
     * @param  instance            An instance of class T on which to call the member function.
     * @param  name                Name of the function as it will be published.
     * @param  description         Description of this event callback function.
     * @param  memberFunctionPtr   Pointer to the C++ implementation of the function.
     * @param  instanceTypeStr     The name of class T. This is only used for logging purposes.
     */
    template<class T>
    void registerEventCallback(T *instance, const std::string &name, const std::string &description,
                               void (T::*memberFunctionPtr)(IrisReceivedRequest &),
                               const std::string &instanceTypeStr)
    {
        std::string funcInfoJson =
                "{descr:'" + description + "',"
                "args:{"
                "  instId:{type:'NumberU64', descr:'Target instance id.'},"
                "  esId:{type:'NumberU64', descr:'Event stream id.'},"
                "  fields:{type:'Object', descr:'Object which contains the names and values of event source fields.'},"
                "  time:{type:'NumberU64', descr:'Simulation time timestamp of the event.'},"
                "  sInstId:{type:'NumberU64', descr:'Source instId: Instance which generated and sent this event.'},"
                "  syncEc:{type:'Boolean', descr:'Synchronous callback behaviour.', optional:true},"
                "},"
                "retval:{type:'Null'}}";
        functionDecoder.registerFunction(instance, name, memberFunctionPtr, funcInfoJson, instanceTypeStr);
    }

    /*!
     * @brief Register a general event callback using an EventCallbackDelegate
     *
     * @param   delegate            EventCallbackDelegate to call to handle the function.
     * @param   name                Name of the function as it will be published.
     * @param   description         Description of this event callback function.
     * @param   dlgInstanceTypeStr  The name of the delegate type. This is only used for logging purposes.
     */
    void registerEventCallback(EventCallbackDelegate delegate, const std::string &name,
                               const std::string &description, const std::string &dlgInstanceTypeStr)
    {
        eventCallbacks[name] = ECD(delegate);
        registerEventCallback(this, name, description, &IrisInstance::impl_eventCallback, dlgInstanceTypeStr);
    }

    /*!
     * @brief Register a general event callback using an EventCallbackDelegate
     *
     * @param   instance            An instance of class T on which to call the delegate T::METHOD().
     * @param   name                Name of the function as it will be published.
     * @param   description         Description of this event callback function.
     * @param   dlgInstanceTypeStr  The name of the delegate type. This is only used for logging purposes.
     */
    template <typename T, IrisErrorCode (T::*METHOD)(uint64_t, const AttributeValueMap&, uint64_t,
                                                     uint64_t, bool, std::string&)>
    void registerEventCallback(T *instance, const std::string &name, const std::string &description,
                               const std::string &dlgInstanceTypeStr)
    {
        registerEventCallback(EventCallbackDelegate::make<T, METHOD>(instance),
                              name, description, dlgInstanceTypeStr);
    }

    /*!
     * @brief Unregister the named event callback function.
     */
    void unregisterEventCallback(const std::string &name);

    /*!
     * @brief Send response to remote iris interface.
     *
     * Call this from the function implementations registered with registerFunction()/irisRegisterFunction()
     *
     * @param   response    The Iris response message to send.
     */
    void sendResponse(const uint64_t *response)
    {
        remoteIrisInterface->irisHandleMessage(response);
    }


    // --- Interface for clients. Access to other components. ---

    /*!
     * @brief Get an IrisCppAdapter to call an Iris function of any other instance.
     *
     * Usage: @code irisCall().resource_read(...); @endcode for the Iris function @c resource_read().
     */
    IrisCppAdapter &irisCall() { return *default_cppAdapter; }

    /*!
     * @brief Get an IrisCppAdapter to call an Iris function of any other instance.
     *
     * When an Iris function returns an error response this adapter will return the error code
     * and not throw an exception.
     * Usage: @code iris::IrisErrorCode code = irisCallNoThrow().resource_read(...); @endcode
     */
    IrisCppAdapter &irisCallNoThrow() { return nothrow_cppAdapter; }

    /*!
     * @brief Get an IrisCppAdapter to call an Iris function of any other instance.
     * When an Iris function returns an error response this adapter will always throw an exception.
     * Usage: @code
       try
       {
           irisCall().resource_read(...);
       }
       catch (iris::IrisErrorException &e)
       {
           ...
       }
       @endcode
     */
    IrisCppAdapter &irisCallThrow() { return throw_cppAdapter; }

    /*!
     * @brief Send Iris request/notification and potentially wait for a response.
     *
     * @param   req     Iris request to send.
     *
     * @return          Returns true iff a non-error response was received (and so result
     *                  values need to be decoded).
     *
     *  Use this to manually call functions implemented in the called target but not
     *  implemented in IrisCppAdapter.
    */
    bool sendRequest(IrisRequest &req)
    {
        return irisCall().callAndPerhapsWaitForResponse(req);
    }


    // --- Misc functionality. ---

    /*!
     * @brief Get local IrisInterface of this instance.
     * This is the interface other instances use to send their requests and responses to this instance.
     *
     * @return          IrisInterface to send messages to this instance.
     */
    IrisInterface *getLocalIrisInterface() { return functionDecoder.getIrisInterface(); }


    /*!
     * @brief Get thread abstraction used by this instance.
     * @return          IrisThreadAbstraction object used by this instance.
     */
    IrisThreadAbstraction *getIrisThreadAbstraction()
    {
        return threadAbstraction;
    }


    /*!
     * @brief Get instance id of this instance.
     * This is valid after registerInstance() returned.
     *
     * @return          The instId for this instance.
     */
    IrisInstId getInstId() const { return thisInstanceInfo.instId; }


    /*!
     * @brief Get instance name of this instance.
     * This is valid after registerInstance() returned.
     *
     * @return          The instance name of this instance. This will be the same as the name
     *                  parameter passed to the constructor or registerInstance() unless this
     *                  instance was registered with the UNIQUIRY flag set and the name was modified
     *                  to make it unique.
     */
    const std::string &getInstanceName() const { return thisInstanceInfo.instName; }

    /*!
     * @brief Get the IrisInstanceBuilder object for this instance.
     * This can be used to set up meta-data and callbacks for standard Iris functions.
     *
     * @return          The IrisInstanceBuilder object for this instance.
     */
    IrisInstanceBuilder *getBuilder();

    /*!
     * @brief Set the event handler and enable the @c IRIS_STATE_CHANGED event.
     */
    void setEventHandler(IrisInstanceEvent *handler);

    /*!
     * @brief Send an @c IRIS_STATE_CHANGED event if the simulation is not running.
     */
    void notifyStateChanged();

private:
    void init(IrisConnectionInterface *connection_interface_ = 0,
              const std::string& instName = std::string(),
              uint64_t flags = DEFAULT_FLAGS,
              IrisThreadAbstraction *threadAbstraction_ = 0);

    // --- Iris function implementations ---
    typedef std::map<std::string,IrisValue> PropertyMap;

    void impl_instance_getProperties(IrisReceivedRequest &request);


    void impl_instance_ping(IrisReceivedRequest &request);

    void impl_eventCallback(IrisReceivedRequest &request);

    /// EventStreamCreate delegate to create event streams for the IRIS_STATE_CHANGED event.
    IrisErrorCode createEventStream(EventStream *&event_stream_out, const EventSourceInfo &info,
                                    const std::vector<std::string> &fields);

    /// Event callback to keep track of simulation run state. This is used to limit IRIS_STATE_CHANGED
    /// notification to changes when the simulation is not running.
    IrisErrorCode ec_SIMULATION_TIME_EVENT(uint64_t esId, const IrisValueMap &fields, uint64_t time,
                                           uint64_t sInstId, bool syncEc, std::string &error_message_out);

    /// Event callback to notify that the Iris system has shut down.
    IrisErrorCode ec_IRIS_SHUTDOWN_LEAVE(uint64_t esId, const IrisValueMap &fields, uint64_t time,
                                         uint64_t sInstId, bool syncEc, std::string &error_message_out);

    // --- Iris specific data and state ---

    /// Function decoder to decode incoming Iris requests.
    IrisFunctionDecoder functionDecoder;

    /// Object that manages sending requests for the cppAdapters
    IrisCppAdapter::RequestManager cppAdapter_request_manager;

    /// C++ adapter that throws an exception if a request returns an error response.
    IrisCppAdapter throw_cppAdapter;

    /// C++ adapter that returns an IrisErrorCode if a request returns an error response.
    IrisCppAdapter nothrow_cppAdapter;

    /// Default cppAdapter to use for unqualified irisCall().
    IrisCppAdapter *default_cppAdapter;

    /// Used to register the channel for our local IrisInterface
    /// and obtain the remote IrisInterface where requests and
    /// responses are sent.
    IrisConnectionInterface *connection_interface;

    /// Remote iris interface.
    /// This is where we send all requests and responses.
    IrisInterface *remoteIrisInterface;

    /// Thread abstraction in use.
    IrisThreadAbstraction *threadAbstraction;

    /// Thread abstraction allocated by this instance (just to free it).
    IrisThreadAbstraction *allocatedThreadAbstraction;

protected:
    /// InstanceInfo of this instance.
    InstanceInfo thisInstanceInfo;

private:
    /// Remember whether instance_getProperties() was already called.
    bool instance_getProperties_called;

    bool registered;

    uint64_t channelId;

    /// Logger.
    IrisLogger log;

    // --- Instance specific data and state ---

    /// Properties.
    PropertyMap propertyMap;

    /// Registered eventCallback delegates
    struct ECD
    {
        // Work around symbol length limits in Visual Studio (warning C4503)
        EventCallbackDelegate dlg;
        ECD() {}
        ECD(EventCallbackDelegate dlg_)
            : dlg(dlg_)
        {
        }
    };
    typedef std::map<std::string, ECD> EventCallbackMap;
    EventCallbackMap eventCallbacks;

    /// Helper object to populate this instance
    IrisInstanceBuilder *builder;

    /// Event registry for IRIS_STATE_CHANGED event.
    IrisEventRegistry *state_changed_event_registry;

    /// Track the run state of the simulation so that we only emit the IRIS_STATE_CHANGED
    /// event when the simulation is stopped.
    enum RunState
    {
        NOT_REGISTERED, //!< Callback has not yet been registered for IRIS_SIMULATION_TIME_EVENT.
        UNKNOWN,        //!< Callback is registered, but we have not seen an event yet.
        RUNNING,        //!< Simulation is running (IRIS_STATE_CHANGED should not be emitted).
        STOPPED         //!< Simulation is not running (IRIS_STATE_CHANGED should be emitted).
    } simulation_run_state;

    IrisThreadAbstractionMutex *simulation_run_state_mutex;
};




NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstance_h

