/*!
 *  \file   IrisInstanceBreakpoint.h
 *  \brief  Boiler-plate handling breakpoint of an IrisInstance. Breakpoint add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceBreakpoint class offers:
 *  - implements all breakpoint related Iris functions.
 *  - maintains and provides breakpoint information (type, address, rscId, ...).
 *  - converts between Iris breakpoint functions (breakpoint*()) and various C++ access functions
 */

#ifndef ARM_INCLUDE_IrisInstanceBreakpoint_h
#define ARM_INCLUDE_IrisInstanceBreakpoint_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

#include <cstdio>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisInstanceEvent;
class IrisEventRegistry;
class IrisReceivedRequest;

class EventStream;
struct EventSourceInfo;

/*!
 * @brief Set a breakpoint corresponding to the given information.
 * @code
       IrisErrorCode setBpt(BreakpointInfo &bptInfo)
   @endcode

 * The members of breakpoint information are guaranteed to be valid.
 * The BreakpointInfo is non-const as we may need to modify the meta info.
 * E.g. Align the address and fix the size of a data breakpoint, which could be useful in
 * some cases.
 * It should never modify the bptId, which is uniquely set by this add-on.
 *
 * Error: Return E_* error code if failed to set the breakpoint
 */
typedef IrisDelegate<BreakpointInfo&> BreakpointSetDelegate;

/*!
 * @brief Delete the breakpoint corresponding to the given information.
 * @code
      IrisErrorCode deleteBpt(const BreakpointInfo &bptInfo)
   @endcode
 *
 * The breakpoint of given information is guaranteed to be valid and existing.
 *
 * Error: Return E_* error code if fail to delete the breakpoint
 */
typedef IrisDelegate<const BreakpointInfo&> BreakpointDeleteDelegate;

/*!
 * @brief Breakpoint add-on for IrisInstance.
 *
 * This is used by instances which want to support breakpoint functionality.
 *
 * This class implements all iris breakpoint*() functions.
 * This class maintains breakpoint information (set by breakpoint_set() and exposed by breakpoint_getList()).
 *
 * Usage (example):
 * irisInstanceBpt = new iris::IrisInstanceBreakpoint(irisInstance);
 * irisInstanceBpt->setBreakpointSetDelegate(bptSetDel);           // Use this delegate for breakpoint set.
 * irisInstanceBpt->setBreakpointDeleteDelegate(bptDeleteDel);     // Use this delegate for breakpoint delete.
 *
 * // When a breakpoint is hit, notify the instances which enable the IRIS_BREAKPOINT_HIT event.
 * irisInstanceBpt->setEventHandler(irisInstanceEvent);
 *
 * See DummyComponent.h for a working example.
 */
class IrisInstanceBreakpoint
{

public:
    // --- Construction and destruction. ---
    IrisInstanceBreakpoint(IrisInstance *irisInstance = NULL);
    ~IrisInstanceBreakpoint();

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * Only use this method if NULL was passed to the constructor.
     *
     * @param   irisInstance    The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set breakpoint set delegate for all breakpoints set by this instance.
     *
     * @param   delegate    A BreakpointSetDelegate to call when a breakpoint is set.
     */
    void setBreakpointSetDelegate(BreakpointSetDelegate delegate);

    /*!
     * @brief Set breakpoint delete delegate for all breakpoints deleted by this instance.
     *
     * @param   delegate    A BreakpointDeleteDelegate to call when a breakpoint is deleted.
     */
    void setBreakpointDeleteDelegate(BreakpointDeleteDelegate delegate);

    /*!
     * @brief Set the event handler used to notify the clients who enable the IRIS_BREAKPOINT_HIT event.
     *
     * All breakpoint events are normal events and are handled through the same mechanism as other events.
     */
    void setEventHandler(IrisInstanceEvent *handler);

    /*!
     * @brief Notify clients that a code breakpoint was hit.
     *
     * Notifies clients by emitting an @c IRIS_BREAKPOINT_HIT event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint hit.
     * @param   pc          Value of a relevant program counter when the event hit.
     * @param   pcSpaceId   Memory space Id for the memory space that the PC address
     *                      corresponds to.
     */
    void notifyBreakpointHit(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId);

    /*!
     * @brief Notify clients that a data breakpoint was hit.
     *
     * Notifies clients by emitting an @c IRIS_BREAKPOINT_HIT event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint hit.
     * @param   pc          Value of a relevant program counter when the event hit.
     * @param   pcSpaceId   Memory space Id for the memory space that the PC address
     *                      corresponds to.
     * @param   accessAddr  The address of the data access that triggered the breakpoint.
     * @param   accessSize  The size of the data access that triggered the breakpoint
     * @param   accessRw    Indicate the direction of the access.
     *                      Value | Meaning
     *                      ------|------
     *                      "r"   | Read access
     *                      "w"   | Write access
     * @param   data        The data that was written/read during the access that triggered the
     *                      breakpoint.
     */
    void notifyBreakpointHitData(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId,
                                 uint64_t accessAddr, uint64_t accessSize,
                                 const std::string &accessRw, const std::vector<uint64_t> &data);

    /*!
     * @brief Notify clients that a register breakpoint was hit.
     *
     * Notifies clients by emitting an @c IRIS_BREAKPOINT_HIT event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint hit.
     * @param   pc          Value of a relevant program counter when the event hit.
     * @param   pcSpaceId   Memory space Id for the memory space that the PC address
     *                      corresponds to.
     * @param   accessRw    Indicate the direction of the access.
     *                      Value | Meaning
     *                      ------|------
     *                      "r"   | Read access
     *                      "w"   | Write access
     * @param   data        The data that was written/read during the access that triggered the
     *                      breakpoint.
     */
    void notifyBreakpointHitRegister(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId,
                                     const std::string &accessRw, const std::vector<uint64_t> &data);

    /*!
     * @brief Get BreakpointInfo for a breakpoint id.
     *
     * @param   bptId   The breakpoint id for which the BreakpointInfo is requested.
     * @return          A pointer to the BreakpointInfo for the requested breakpoint or
     *                  NULL if @a bptId is not a valid breakpoint id.
     */
    const BreakpointInfo *getBreakpointInfo(uint64_t bptId) const;

    /*!
     * @brief Add an optional component-specific condition that can be configured by clients.
     *
     * @param   name            The name of the condition.
     * @param   type            The type of the value that clients will set to configure the condition.
     * @param   description     A description of the condition.
     * @param   bpt_types       A list of breakpoint types that this condition can be applied to.
     *                          An empty list indicates all types.
     */
    void addCondition(const std::string &name, const std::string &type, const std::string &description,
                      const std::vector<std::string> bpt_types=std::vector<std::string>());

private:

    void impl_breakpoint_set(IrisReceivedRequest &request);

    void impl_breakpoint_delete(IrisReceivedRequest &request);

    void impl_breakpoint_getList(IrisReceivedRequest &request);

    void impl_breakpoint_getAdditionalConditions(IrisReceivedRequest &request);

    void beginBreakpointHit(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId);

    /// Create an event stream for the IRIS_BREAKPOINT_HIT event
    IrisErrorCode createEventStream(EventStream *&, const EventSourceInfo&, const std::vector<std::string>&);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// EventRegistry to handle IRIS_BREAKPOINT_HIT callbacks
    IrisEventRegistry *breakpoint_hit_registry;

    /// breakpoint information list.
    /// Densely allocated, may contain deleted breakpoint informations, whose bptId is set to IRIS_UINT64_MAX.
    std::vector<BreakpointInfo> bptInfos;

    /// List of free breakpoint ids.
    /// This is just here to help us allocating breakpoint ids densely and efficiently.
    std::vector<uint64_t> freeBptIds;

    /// Breakpoint id to esId and event action (True to enable event, False to disable event).
    std::map<uint64_t, std::pair<uint64_t, bool> > bptActions;

    /// List of component-specific conditions that can be applied to breakpoints.
    std::vector<BreakpointConditionInfo> additional_conditions;

    /// Delegate for breakpoint set
    BreakpointSetDelegate bptSetDelegate;

    /// Delegate for breakpoint delete
    BreakpointDeleteDelegate bptDeleteDelegate;

    /// Logger.
    IrisLogger log;
};


NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceBreakpoint_h
