/*!
 *  \file   IrisInstanceBuilder.h
 *  \brief  A high level interface to build up functionality on an IrisInstance
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceBuilder_h
#define ARM_INCLUDE_IrisInstanceBuilder_h

#include "iris/detail/IrisCommon.h"
#include "iris/IrisInstanceResource.h"
#include "iris/IrisInstanceEvent.h"
#include "iris/IrisEventEmitter.h"
#include "iris/IrisInstanceBreakpoint.h"
#include "iris/IrisInstanceMemory.h"
#include "iris/IrisInstanceImage.h"
#include "iris/IrisInstanceStep.h"
#include "iris/IrisInstancePerInstanceExecution.h"
#include "iris/IrisInstanceTable.h"
#include "iris/IrisInstanceDisassembler.h"
#include "iris/IrisInstanceDebuggableState.h"
#include "iris/IrisInstanceSemihosting.h"
#include "iris/IrisInstanceSyncLevel.h"

#include <cassert>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisRegisterEventEmitterBase;

/*!
 * @brief Helper class to set up Iris instances
 *
 * See DummyComponent.h for a working example.
 */
class IrisInstanceBuilder
{
private:
    template <typename T, T*(IrisInstanceBuilder::*INIT_METHOD)()>
    class LazyAddOn
    {
    private:
        IrisInstanceBuilder *parent;
        T* add_on;

    public:
        LazyAddOn(IrisInstanceBuilder *parent_)
            : parent(parent_)
            , add_on(NULL)
        {
        }

        ~LazyAddOn()
        {
            delete add_on;
        }

        T* operator-> ()
        {
            if (add_on == NULL)
            {
                init();
            }

            return add_on;
        }

        operator T* ()
        {
            if (add_on == NULL)
            {
                init();
            }

            return add_on;
        }

        T* getPtr()
        {
            return add_on;
        }

        void init()
        {
            assert(add_on == NULL);
            add_on = (parent->*INIT_METHOD)();
        }
    };
    IrisInstance *iris_instance; /// The Iris instance we are building.
#define INTERNAL_LAZY(addon) \
        addon * init##addon(); \
        LazyAddOn<addon, &IrisInstanceBuilder::init##addon>
    INTERNAL_LAZY(IrisInstanceResource)             inst_resource;      //!< IrisInstance resource add-on
    INTERNAL_LAZY(IrisInstanceEvent)                inst_event;         //!< IrisInstance event add-on
    INTERNAL_LAZY(IrisInstanceBreakpoint)           inst_breakpoint;    //!< IrisInstance breakpoint add-on
    INTERNAL_LAZY(IrisInstanceMemory)               inst_memory;        //!< IrisInstance memory add-on
    INTERNAL_LAZY(IrisInstanceImage)                inst_image;         //!< IrisInstance image loading add-on
    INTERNAL_LAZY(IrisInstanceImage_Callback)       inst_image_cb;      //!< IrisInstance image_loadDataRead() add-on
    INTERNAL_LAZY(IrisInstanceStep)                 inst_step;          //!< IrisInstance execution stepping add-on
    INTERNAL_LAZY(IrisInstancePerInstanceExecution) inst_per_inst_exec; //!< IrisInstance per-instance execution add-on
    INTERNAL_LAZY(IrisInstanceTable)                inst_table;         //!< IrisInstance table add-on
    INTERNAL_LAZY(IrisInstanceDisassembler)         inst_disass;        //!< IrisInstance disassembler add-on
    INTERNAL_LAZY(IrisInstanceDebuggableState)      inst_dbg_state;     //!< IrisInstance debuggable state add-on
    INTERNAL_LAZY(IrisInstanceSemihosting)          inst_semihost;      //!< IrisInstance semihosting add-on
    INTERNAL_LAZY(IrisInstanceSyncLevel)            inst_sync_level;    //!< IrisInstance sync-level add-on
#undef INTERNAL_LAZY

    /// Resource context
    //@{
    /// Default resource access delegates.
    /// These are used for all resources which do not have resource specific delegates assigned.
    ResourceReadDelegate default_reg_read_delegate;
    ResourceWriteDelegate default_reg_write_delegate;
    //@}

    /// Event context
    //@{
    struct RegisterEventInfo
    {
        IrisInstanceEvent::EventSourceInfoAndDelegate event_info;

        typedef std::vector<uint64_t> RscIdList;
        RscIdList rscId_list;
        IrisRegisterEventEmitterBase *event_emitter;

        RegisterEventInfo()
            : event_emitter(NULL)
        {
        }
    };

    std::vector<RegisterEventInfo*> register_read_event_info_list;
    std::vector<RegisterEventInfo*> register_update_event_info_list;

    RegisterEventInfo* active_register_read_event_info;
    RegisterEventInfo* active_register_update_event_info;

    RegisterEventInfo* find_register_event(const std::vector<RegisterEventInfo*>& register_event_info_list,
                                           const std::string &name);

    RegisterEventInfo* initRegisterReadEventInfo(const std::string &name);
    RegisterEventInfo* initRegisterUpdateEventInfo(const std::string &name);

    /// Fills in the REGISTER field enums, type, and size information based on the associated rscIds.
    /// Will throw IrisInternalError if any registers have an inconsistent type.
    uint64_t finalizeRegisterEvent(RegisterEventInfo *event_info, bool is_read);
    /// Add a rscId to the rscId_lists for the current register_read/update_event_info (if there are any)
    void associateRegisterWithTraceEvents(uint64_t rscId);
    //@}

    /// Breakpoint management context
    //@{
    IrisErrorCode setBreakpoint(BreakpointInfo &info);
    IrisErrorCode deleteBreakpoint(const BreakpointInfo &info);

    struct RegisterEventEmitterPair
    {
        IrisRegisterEventEmitterBase *read;
        IrisRegisterEventEmitterBase *update;

        RegisterEventEmitterPair()
            : read(NULL)
            , update(NULL)
        {
        }
    };
    typedef std::map<uint64_t, RegisterEventEmitterPair> RscIdEventEmitterMap;
    RscIdEventEmitterMap register_event_emitter_map;

    BreakpointSetDelegate user_setBreakpoint;
    BreakpointDeleteDelegate user_deleteBreakpoint;
    //@}

public:
    /*!
     * @brief Construct an IrisInstanceBuilder for an Iris instance.
     *
     * @param   iris_instance   The instance to build.
     */
    IrisInstanceBuilder(IrisInstance *iris_instance);

    /*!
     * @defgroup resource_apis IrisInstanceBuilder resource APIs
     * @brief Set up resource and register meta-data and delegates.
     * @{
     */

#define INTERNAL_RESOURCE_BUILDER_MIXIN(TYPE) \
        /*! @brief Set the @c name field. */ \
        /*! @param  name    The name field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setName(const std::string &name)       { info->resourceInfo.name = name; return *this; } \
        /*! @brief Set the @c cname field. */ \
        /*! @param  cname   The cname field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setCname(const std::string &cname)     { info->resourceInfo.cname = cname; return *this; } \
        /*! @brief Set the @c descr field. */ \
        /*! @param  descr   The descr field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setDescr(const std::string &descr)     { info->resourceInfo.descr = descr; return *this; } \
        /*! @brief Set the @c format field. */ \
        /*! @param  format  The format field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setFormat(const std::string &format)   { info->resourceInfo.format = format; return *this; } \
        /*! @brief Set the @c bitWidth field. */ \
        /*! @param  bitWidth  The bitWidth field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setBitWidth(uint64_t bitWidth)         { info->resourceInfo.bitWidth = bitWidth; return *this; } \
        /*! @brief Set the @c type field. */ \
        /*! @param  type  The type field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setType(const std::string &type)       { info->resourceInfo.type = type; return *this; } \
        /*! @brief Set the @c rwMode field. */ \
        /*! @param  rwMode  The rwMode field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setRwMode(const std::string &rwMode)   { info->resourceInfo.rwMode = rwMode; return *this; } \
        /*! @brief Set the @c subRscId field. */ \
        /*! @param  subRscId  The subRscId field of the ResourceInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setSubRscId(uint64_t subRscId)         { info->resourceInfo.subRscId = subRscId; return *this; } \
        /*! @brief Add a symbol to the enums field for numeric resources. */ \
        /*! This should be called multiple times to add multiple symbols. */ \
        /*! @param  symbol  The symbol string to be associated with the specified value. */ \
        /*! @param  value   The value of this symbol. */ \
        /*! @param  descr   A description of this symbol. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &addEnum(const std::string &symbol, const IrisValue &value, const std::string &descr = std::string()) \
        { \
            info->resourceInfo.enums.push_back(EnumElementInfo(value, symbol, descr)); \
            return *this; \
        } \
        /*! @brief Add a symbol to the enums field for string resources. */ \
        /*! This should be called multiple times to add multiple symbols. */ \
        /*! @param  value   The string value of this symbol. This is also used as the symbols string. */ \
        /*! @param  descr   A description of this symbol. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &addStringEnum(const std::string &stringValue, const std::string &descr = std::string()) \
        { \
            info->resourceInfo.enums.push_back(EnumElementInfo(IrisValue(stringValue), std::string(), descr)); \
            return *this; \
        } \
        /*! @brief Set the named boolean tag to true (e.g. isPc) */ \
        /*! @param  tag     The name of the tag to set. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setTag(const std::string &tag) \
        { \
            info->resourceInfo.tags[tag] = IrisValue(true); \
            return *this; \
        } \
        /*! @brief Set a tag to the specified value. */ \
        /*! @param  tag     The name of the tag to set. */ \
        /*! @param  value   The value to set the tag to. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setTag(const std::string &tag, const IrisValue &value) \
        { \
            info->resourceInfo.tags[tag] = value; \
            return *this; \
        } \
        /*! @brief Set the paramterInfo field. */ \
        /*! This will make the resource a paramter. */ \
        /*! @param  paramInfo     ParameterInfo object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setParameterInfo(const ParameterInfo &paramInfo) \
        { \
            info->resourceInfo.isParameter = true; \
            info->resourceInfo.parameterInfo = paramInfo; \
            return *this; \
        } \
        /*! @brief Set the delegate to read the resource */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceReadDelegate */ \
        /*! @param  readDelegate     ResourceReadDelegate object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setReadDelegate(ResourceReadDelegate readDelegate) \
        { \
            info->readDelegate = readDelegate; \
            return *this; \
        } \
        /*! @brief Set the delegate to write the resource */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceWriteDelegate */ \
        /*! @param  writeDelegate     ResourceWriteDelegate object. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setWriteDelegate(ResourceWriteDelegate writeDelegate) \
        { \
            info->writeDelegate = writeDelegate; \
            return *this; \
        } \
        /*! @brief Set the delegate to read the resource */ \
        /*! Set a delegate which calls METHOD() on an instance of class T. */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceReadDelegate */ \
        /*! @tparam T        A class that defines a method with the right signature to be a resource read delegate. */ \
        /*! @tparam METHOD   A resource read delegate method in class T. */ \
        /*! @param  instance The instance of class T on which to call METHOD. */ \
        /*! @return          A reference to this TYPE object allowing calls to be chained together. */ \
        template <typename T, IrisErrorCode (T::*METHOD)(const ResourceInfo &, ResourceReadResult &)> \
        TYPE &setReadDelegate(T* instance) \
        { \
            return setReadDelegate(ResourceReadDelegate::make<T, METHOD>(instance)); \
        } \
        /*! @brief Set the delegate to read the resource */ \
        /*! Set a delegate which calls function FUNC(). */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceReadDelegate */ \
        /*! @tparam FUNC      A resource read delegate function. */ \
        /*! @return           A reference to this TYPE object allowing calls to be chained together. */ \
        template <IrisErrorCode (*FUNC)(const ResourceInfo &, ResourceReadResult &)> \
        TYPE &setReadDelegate() \
        { \
            return setReadDelegate(ResourceReadDelegate::make<FUNC>()); \
        } \
        /*! @brief Set the delegate to write the resource */ \
        /*! Set a delegate which calls METHOD() on an instance of class T. */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceWriteDelegate */ \
        /*! @tparam T        A class that defines a method with the right signature to be a resource write delegate. */ \
        /*! @tparam METHOD   A resource write delegate method in class T. */ \
        /*! @param  instance The instance of class T on which to call METHOD. */ \
        /*! @return          A reference to this TYPE object allowing calls to be chained together. */ \
        template <typename T, IrisErrorCode (T::*METHOD)(const ResourceInfo &, const ResourceWriteValue &)> \
        TYPE &setWriteDelegate(T* instance) \
        { \
            return setWriteDelegate(ResourceWriteDelegate::make<T, METHOD>(instance)); \
        } \
        /*! @brief Set the delegate to write the resource */ \
        /*! Set a delegate which calls function FUNC(). */ \
        /*! If this is not set the default delegate will be used. */ \
        /*! @see IrisInstanceBuilder::setDefaultResourceWriteDelegate */ \
        /*! @tparam FUNC    A resource write delegate function. */ \
        /*! @return         A reference to this TYPE object allowing calls to be chained together. */ \
        template <IrisErrorCode (*FUNC)(const ResourceInfo &, const ResourceWriteValue &)> \
        TYPE &setWriteDelegate() \
        { \
            return setWriteDelegate(ResourceWriteDelegate::make<FUNC>()); \
        } \
        /*! @brief Return the rscId which was allocated for this resource.*/ \
        /*! @return         The rscId which was allocated for this resource.*/ \
        uint64_t getRscId() const \
        { \
            return info->resourceInfo.rscId; \
        }

#define INTERNAL_REGISTER_BUILDER_MIXIN(TYPE) \
        /*! @brief Set the @c lsbOffset field. */ \
        /*! @param   lsbOffset     The lsbOffset field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setLsbOffset(uint64_t lsbOffset) \
        { \
            regInfo->lsbOffset = lsbOffset; \
            return *this; \
        } \
        /*! @brief Set the @c canonicalRn field. */ \
        /*! @param   canonicalRn   The canonicalRn field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setCanonicalRn(uint64_t canonicalRn_) \
        { \
            regInfo->canonicalRn = canonicalRn_; \
            regInfo->hasCanonicalRn = true; \
            return *this; \
        } \
        /*! @brief Set the @c writeMask field. */ \
        /*! @param   writeMask     The writeMask field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setWriteMask(uint64_t writeMask) /* Call multiple times for resources > 64 bit. */ \
        { \
            regInfo->writeMask.push_back(writeMask); \
            return *this; \
        } \
        /*! @brief Set the @c resetData field. */ \
        /*! Set the reset value for numeric registers. */ \
        /*! @param   resetData     The resetData field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setResetData(uint64_t resetData) /* Call multiple times for resources > 64 bit. */ \
        { \
            regInfo->resetData.push_back(resetData); \
            return *this; \
        } \
        /*! @brief Set the @c resetString field. */ \
        /*! Set the reset value for string registers. */ \
        /*! @param   resetString   The resetString field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setResetString(const std::string &resetString) \
        { \
            regInfo->resetString = resetString; \
            return *this; \
        } \
        /*! @brief Set the @c addressOffset field. */ \
        /*! @param   addressOffset The addressOffset field of the RegisterInfo object. */ \
        /*! @return                A reference to this TYPE object allowing calls to be chained together. */ \
        TYPE &setAddressOffset(uint64_t addressOffset) \
        { \
            regInfo->addressOffset = addressOffset; \
            regInfo->hasAddressOffset = true; \
            return *this; \
        }

    /*!
     * @brief Used to set meta-data on a resource.
     */
    class ResourceBuilder
    {
    private:
        IrisInstanceResource::ResourceInfoAndAccess *info;

    public:
        ResourceBuilder(IrisInstanceResource::ResourceInfoAndAccess &info_)
            : info(&info_)
        {
        }

        ResourceBuilder()
            : info(NULL)
        {
        }

        INTERNAL_RESOURCE_BUILDER_MIXIN(ResourceBuilder)
    };

    class FieldBuilder;

    /*!
     * @brief Used to set meta-data on a register resource.
     */
    class RegisterBuilder
    {
    private:
        IrisInstanceResource::ResourceInfoAndAccess *info;
        RegisterInfo *regInfo;
        IrisInstanceResource *inst_resource;

    public:
        RegisterBuilder(IrisInstanceResource::ResourceInfoAndAccess &info_, IrisInstanceResource *inst_resource_)
            : info(&info_)
            , regInfo(&info_.resourceInfo.registerInfo)
            , inst_resource(inst_resource_)
        {
            info->resourceInfo.isRegister = true;
        }

        RegisterBuilder()
            : info(NULL)
            , regInfo(NULL)
            , inst_resource(NULL)
        {
        }

        INTERNAL_RESOURCE_BUILDER_MIXIN(RegisterBuilder)
        INTERNAL_REGISTER_BUILDER_MIXIN(RegisterBuilder)

        /*!
         * @brief Add a subregister field to this register.
         * By default the field will copy attributes from its parent register but any field can be overridden.
         * @param   name         Name of the register field.
         * @param   lsbOffset    The bit offset of this field inside its parent register.
         * @param   bitWidth     The size of the field.
         * @param   descr        Description of this field.
         * @return               A FieldBuilder object that allows to caller to set attributes for this field.
         */
        FieldBuilder addField(const std::string &name, uint64_t lsbOffset, uint64_t bitWidth, const std::string &descr);
    };

    /*!
     * @brief Used to set meta-data on a register field resource.
     */
    class FieldBuilder
    {
    protected:
        IrisInstanceResource::ResourceInfoAndAccess *info;
        RegisterInfo *regInfo;
        RegisterBuilder *parent_reg;

    public:
        FieldBuilder(IrisInstanceResource::ResourceInfoAndAccess &info_, RegisterBuilder *parent_reg_)
            : info(&info_)
            , regInfo(&info_.resourceInfo.registerInfo)
            , parent_reg(parent_reg_)
        {
        }

        FieldBuilder()
            : info(NULL)
            , regInfo(NULL)
            , parent_reg(NULL)
        {
        }

        INTERNAL_RESOURCE_BUILDER_MIXIN(FieldBuilder)
        INTERNAL_REGISTER_BUILDER_MIXIN(FieldBuilder)

        /*!
         * @brief Get the RegisterBuilder for the parent register
         * @return      The RegisterBuilder object for the parent register.
         */
        RegisterBuilder &parent()
        {
            return *parent_reg;
        }

        /*!
         * @brief Add another subregister field to the parent register.
         * @see RegisterBuilder::addField
         */
        FieldBuilder addField(const std::string &name, uint64_t lsbOffset, uint64_t bitWidth, const std::string &descr)
        {
            return parent().addField(name, lsbOffset, bitWidth, descr);
        }
    };

#undef INTERNAL_RESOURCE_BUILDER_MIXIN
#undef INTERNAL_REGISTER_BUILDER_MIXIN



    /*!
     * @brief Set default read access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all resources.
     *
     * Usage:
     * Pass an instance of ResourceReadDelegate into this to delegate reading to any class T:
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode myReadFunction(const iris::ResourceInfo &resourceInfo,
                                               iris::ResourceReadResult &result);
        };

        MyClass myInstanceOfMyClass;
        ResourceReadDelegate delegate =
            ResourceReadDelegate::make<MyClass, &MyClass::myReadFunction>(&myInstanceOfMyClass);

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setDefaultReadDelegate(delegate);
        builder->addResource(...); // Uses myReadFunction
        @endcode
     * @param   delegate    Delegate object which will be called to read resources.
     */
    void setDefaultResourceReadDelegate(ResourceReadDelegate delegate = ResourceReadDelegate())
    {
        default_reg_read_delegate = delegate;
    }

    /*!
     * @brief Set default read access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a resource read method.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode myReadFunction(const iris::ResourceInfo &resourceInfo,
                                               iris::ResourceReadResult &result);
        };

        MyClass myInstanceOfMyClass;

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setDefaultReadDelegate<MyClass, &MyClass::myReadFunction>(myInstanceOfMyClass);
        builder->addResource(...); // Uses myReadFunction
        @endcode
     * @tparam  T           Class which defines a resource read delegate method.
     * @tparam  METHOD      A method of class T which is a resource read delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode (T::*METHOD)(const ResourceInfo &, ResourceReadResult &)>
    void setDefaultResourceReadDelegate(T* instance)
    {
        setDefaultResourceReadDelegate(ResourceReadDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default read access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass in a global function to delegate resource reading to that function.
     * @code
        iris::IrisErrorCode myReadFunction(const iris::ResourceInfo &resourceInfo,
                                           iris::ResourceReadResult &result);

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setDefaultReadDelegate<myReadFunction>();
        builder->addResource(...); // Uses myReadFunction
        @endcode
     * @tparam   FUNC    A function which is a resource read delegate.
     */
    template <IrisErrorCode (*FUNC)(const ResourceInfo &, ResourceReadResult &)>
    void setDefaultResourceReadDelegate()
    {
        setDefaultResourceReadDelegate(ResourceReadDelegate::make<FUNC>());
    }

    /*!
     * @brief Set default write access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all resources.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a resource write method.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode myWriteFunction(const iris::ResourceInfo &resourceInfo, const uint64_t *data);
        };

        MyClass myInstanceOfMyClass;

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        iris::ResourceWriteDelegate delegate =
            iris::ResourceWriteDelegate::make<MyClass, &MyClass::myWriteFunction>(myInstanceOfMyClass);
        builder->setDefaultWriteDelegate(delegate);
        builder->addResource(...); // Uses myWriteFunction
        @endcode
     * @param   delegate    Delegate object which will be called to write resources.
     */
    void setDefaultResourceWriteDelegate(ResourceWriteDelegate delegate = ResourceWriteDelegate())
    {
        default_reg_write_delegate = delegate;
    }

    /*!
     * @brief Set default write access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a resource write method.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode myWriteFunction(const iris::ResourceInfo &resourceInfo, const uint64_t *data);
        };

        MyClass myInstanceOfMyClass;

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setDefaultWriteDelegate<MyClass, &MyClass::myWriteFunction>(myInstanceOfMyClass);
        builder->addResource(...); // Uses myWriteFunction
        @endcode
     * @tparam  T           Class which defines a resource write delegate method.
     * @tparam  METHOD      A method of class T which is a resource write delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode (T::*METHOD)(const ResourceInfo &, const ResourceWriteValue &)>
    void setDefaultResourceWriteDelegate(T* instance)
    {
        setDefaultResourceWriteDelegate(ResourceWriteDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default write access function for all subsequently added resources.
     *
     * Resources that do not explicitly override the access function via
     * @code addResource(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass in a global function to delegate resource writing to that function.
     * @code
        iris::IrisErrorCode myWriteFunction(const iris::ResourceInfo &resourceInfo, const uint64_t *data);

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setDefaultWriteDelegate<myWriteFunction>();
        builder->addResource(...); // Uses myWriteFunction
        @endcode
     * @tparam   FUNC    A function which is a resource write delegate.
     */
    template <IrisErrorCode (*FUNC)(const ResourceInfo &, const ResourceWriteValue &)>
    void setDefaultResourceWriteDelegate()
    {
        setDefaultResourceWriteDelegate(ResourceWriteDelegate::make<*FUNC>());
    }

    /*!
     * @brief Set both read and write resource delegates if they are defined in the same class.
     * @see setDefaultResourceReadDelegate
     * @see setDefaultResourceWriteDelegate
     * @tparam  T           Class which defines resource read and write delegate methods.
     * @tparam  READER      A method of class T which is a resource read delegate.
     * @tparam  WRITER      A method of class T which is a resource write delegate.
     * @param   instance    An instance of class T on which READER and WRITER should be called.
     */
    template <typename T, IrisErrorCode (T::*READER)(const ResourceInfo &, ResourceReadResult &),
                          IrisErrorCode (T::*WRITER)(const ResourceInfo &, const ResourceWriteValue &)>
    void setDefaultResourceDelegates(T* instance)
    {
        setDefaultResourceReadDelegate(ResourceReadDelegate::make<T, READER>(instance));
        setDefaultResourceWriteDelegate(ResourceWriteDelegate::make<T, WRITER>(instance));
    }

    /*!
     * @brief Begin new resource group.
     *
     * This has the following effects:
     * - Add a resource group if it does not yet exist.
     * - Assign all resources which are added through addResource() calls to this group,
     *   starting with rscId @a rscIdStart if possible. If @a rscIdStart is already used the next
     *   free id after @a rscIdStart will be used.
     *
     * @param   name        Name of the resource group.
     * @param   descr       Description of the resource group.
     * @param   rscIdStart  The number from which to start numbering new resources when they are added
     *                      to this group.
     * @param   cname       C-identifier style name to use for this group if it is different from @a name.
     *
     * @see addResource
     * @see addStringResource
     * @see addNoValueResource
     * @see addRegister
     * @see addStringRegister
     * @see addNoValueRegister
     */
    void beginResourceGroup(const std::string& name, const std::string& descr, uint64_t rscIdStart = 0,
                            const std::string& cname = std::string());

    /*!
     * @brief Add meta-information for one numeric resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'numeric'.
     * Call setType("numericFp") on the returned ResourceBuilder to add a 'numericFp' (pure floating point)
     * resource.
     * Use addStringResource() to add a resource of type 'string'.
     * Use addRegister() to add a numeric register resource.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name        Name of the resource. This is the same as the 'name' field of ResourceInfo.
     * @param  bitWidth    Width of the resource in bits.
     *                     This is the same as the 'bitWidth' field of ResourceInfo.
     * @param  descr       Human readable description of the resource.
     *                     This is the same as the 'descr' field of ResourceInfo.
     * @return             A ResourceBuilder object that can be used to set additional meta-data
     *                     for this resource.
     */
    ResourceBuilder addResource(const std::string &name, uint64_t bitWidth, const std::string &descr);

    /*!
     * @brief Add meta-information for one string resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'string'.
     * Use addResource() to add a resource of a type 'numeric' or 'numericFp'.
     * Use addStringRegister() to add a string register resource.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name        Name of the resource. This is the same as the 'name' field of ResourceInfo.
     * @param  descr       Human readable description of the resource.
     *                     This is the same as the 'descr' field of ResourceInfo.
     * @return             A ResourceBuilder object that can be used to set additional meta-data
     *                     for this resource.
     */
    ResourceBuilder addStringResource(const std::string &name, const std::string &descr);

    /*!
     * @brief Add meta-information for one noValue resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'noValue'.
     * Use addResource() to add a resource of a type 'numeric' or 'numericFp'.
     * Use addStringResource() to add a resource of a type 'string'.
     * Use addNoValueRegister() to add a noValue register resource.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name        Name of the resource. This is the same as the 'name' field of ResourceInfo.
     * @param  descr       Human readable description of the resource.
     *                     This is the same as the 'descr' field of ResourceInfo.
     * @param  format      The format used to display this resource.
     * @return             A ResourceBuilder object that can be used to set additional meta-data
     *                     for this resource.
     */
    ResourceBuilder addNoValueResource(const std::string &name, const std::string &descr, const std::string &format);

    /*!
     * @brief Add meta-information for one numeric register resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'numeric'.
     * Call setType("numericFp") on the returned RegisterBuilder to add a 'numericFp' (pure floating point)
     * register.
     * Use addStringRegister() to add a register of type 'string'.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name          Name of the register. This is the same as the 'name' field of ResourceInfo.
     * @param  bitWidth      Width of the resource in bits.
     *                       This is the same as the 'bitWidth' field of ResourceInfo.
     * @param  descr         Human readable description of the resource.
     *                       This is the same as the 'descr' field of ResourceInfo.
     * @param  addressOffset The address offset of this register inside the parent device.
     *                       This is the same as the 'addressOffset' field of RegisterInfo.
     * @param  canonicalRn   Canonical Register Number. This is the same as the 'canonicalRn' field
     *                       of RegisterInfo.
     * @return               A RegisterBuilder object that can be used to set additional meta-data
     *                       for this register resource.
     *
     * @remark A value of 2**64-1 (0xFFFFFFFFFFFFFFFF) for the arguments @a addressOffset and @a canonicalRn
     *         (the default value) is used to indicate that the field is not set.
     *         To set an addressOffset of 2**64-1 use
     *         @code addRegister(...).setAddressOffset(0xFFFFFFFFFFFFFFFF); @endcode
     *         To set a caconicalRn of 2**64-1 use
     *         @code addRegister(...).setCanonicalRn(0xFFFFFFFFFFFFFFFF); @endcode
     */
    RegisterBuilder addRegister(const std::string &name, uint64_t bitWidth, const std::string &descr,
                                 uint64_t addressOffset = IRIS_UINT64_MAX, uint64_t canonicalRn = IRIS_UINT64_MAX);

    /*!
     * @brief Add meta-information for one string register resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'string'.
     * Use addRegister() to add a register of type 'numeric'.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name          Name of the register. This is the same as the 'name' field of ResourceInfo.
     * @param  descr         Human readable description of the resource.
     *                       This is the same as the 'descr' field of ResourceInfo.
     * @return               A RegisterBuilder object that can be used to set additional meta-data
     *                       for this register resource.
     */
    RegisterBuilder addStringRegister(const std::string &name, const std::string &descr);

    /*!
     * @brief Add meta-information for one noValue resource.
     *
     * Resource group: beginResourceGroup() must have been called before calling this.
     * The added resource is automatically added to the last group added through beginResourceGroup().
     *
     * Type: The added resource is of type 'noValue'.
     * Use addRegister() to add a register of a type 'numeric' or 'numericFp'.
     * Use addStringRegister() to add a register of a type 'string'.
     * Use addNoValueResource() to add a noValue generic resource.
     *
     * The returned builder object is only valid until another resource is added. It is only intended to
     * modify the resource which was added last.
     *
     * @param  name        Name of the resource. This is the same as the 'name' field of ResourceInfo.
     * @param  descr       Human readable description of the resource.
     *                     This is the same as the 'descr' field of ResourceInfo.
     * @param  format      The format used to display this resource.
     * @return             A RegisterBuilder object that can be used to set additional meta-data
     *                     for this resource.
     */
    RegisterBuilder addNoValueRegister(const std::string &name, const std::string &descr, const std::string &format);

    /*!
     * @brief Set the register.canonicalRnScheme instance property
     *
     * This property is visible in the list of properties returned by instance_getProperties().
     *
     * This property defines the scheme used by the 'canonicalRn' member of the RegisterInfo object.
     * The default is 'arm.com/registers' which is used by all Arm components.
     * This default can be overridden by calling this function.
     * This should be called upon initialisation, before other instances have a chance to call
     * instance_getProperties().
     *
     * @param   canonicalRnScheme   Name of the canonical register number scheme used by this instance.
     */
    void setPropertyCanonicalRnScheme(const std::string &canonicalRnScheme);

    /*!
     * @brief Set the rscId which will be used for the next resource to be added.
     *
     * Resources which are added following this call get assigned rscIds starting at nextRscId (if possible).
     * It is usually not necessary to call this function. It may however be useful to achieve a certain rscId
     * numbering scheme which simplifies the evaluation of the rscId in the read/write function.
     *
     * @param   nextRscId   The rscId which will be used for the next resource to be added.
     */
    void setNextRscId(uint64_t nextRscId)
    {
        inst_resource->setNextRscId(nextRscId);
    }

    /*!
     * @brief Set a tag for a specific resource.
     *
     * @param   rscId   Resource Id for the resource which will have this tag set.
     * @param   tag     Name of the boolean tag which will be set to true.
     *
     * @see ResourceBuilder::setTag
     * @see RegisterBuilder::setTag
     */
    void setTag(uint64_t rscId, const std::string &tag);

    /*!
     * @}
     */

    /*!
     * @defgroup event_apis IrisInstanceBuilder event APIs
     * @{
     */

    /*!
     * @brief Used to set meta-data on an EventSource
     */
    class EventSourceBuilder
    {
    private:
        IrisInstanceEvent::EventSourceInfoAndDelegate &info;

    public:
        EventSourceBuilder(IrisInstanceEvent::EventSourceInfoAndDelegate &info_)
            : info(info_)
        {
        }

        /*!
         * @brief Set the @c name field.
         * @param  name     The name field of the EventSourceInfo object.
         * @return          A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setName(const std::string &name)
        {
            info.info.name = name;
            return *this;
        }

        /*!
         * @brief Set the @c descr field.
         * @param  descr    The descr field of the EventSourceInfo object.
         * @return          A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setDescription(const std::string &descr)
        {
            info.info.description = descr;
            return *this;
        }

        /*!
         * @brief Set the @c format field.
         * @param  format   The format field of the EventSourceInfo object.
         * @return          A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setFormat(const std::string &format)
        {
            info.info.format = format;
            return *this;
        }

        /*!
         * @brief Set the @c counter field.
         * @param  counter  The counter field of the EventSourceInfo object.
         * @return          A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setCounter(bool counter=true)
        {
            info.info.counter = counter;
            return *this;
        }

        /*!
         * @brief Set the @c ringBuffer field.
         * @param  ringBuffer  The ringBuffer field of the EventSourceInfo object.
         * @return             A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setRingBuffer(bool ringBuffer=true)
        {
            info.info.ringBuffer = ringBuffer;
            return *this;
        }

        /*!
         * @brief Hide/unhide this event source.
         * @param  hidden  If true this event source will not be listed in event_getEventSources() calls
         *                 but can still be accessed by event_getEventSource() for clients which know the
         *                 event source's name.
         * @return         A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setHidden(bool hidden=true)
        {
            info.info.isHidden = hidden;
            return *this;
        }

        /*!
         * @brief Add a field to this event source.
         *
         * This method constructs an EventSourceFieldInfo object and adds it to the EventSource.
         * It should be called multiple times to add multiple fields.
         *
         * @param  name         The name of the field.
         * @param  type         The type of the field.
         * @param  size         The size of the field in bytes.
         * @param  description  A human readable description of the field.
         * @return              A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &addField(const std::string &name, const std::string &type, uint64_t size,
                                     const std::string &description)
        {
            info.info.addField(name, type, size, description);
            return *this;
        }

        /*!
         * @brief Add an enum element for the last field added.
         *
         * This must be called after addField().
         *
         * @param  value        The value of the enum element.
         * @param  symbol       The symbol string that will be displayed instead of the value.
         * @param  description  A human readable description of this enum.
         * @return              A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &addEnumElement(uint64_t value, const std::string &symbol, const std::string &description = "")
        {
            if (info.info.fields.size() > 0)
            {
                info.info.fields.back().addEnumElement(value, symbol, description);
                return *this;
            }
            else
            {
                throw IrisInternalError("EventSourceInfo has no fields to add an enum element to.");
            }
        }

        /*!
         * @brief Set the delegate to create an event stream.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultEsCreateDelegate
         *
         * @param  delegate  EventStreamCreateDelegate object.
         * @return           A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        EventSourceBuilder &setEventStreamCreateDelegate(EventStreamCreateDelegate delegate)
        {
            info.createEventStream = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to create an event stream.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultEsCreateDelegate
         *
         * @tparam T         A class defines a method with the right signature to be an event
         *                   stream creation method.
         * @tparam METHOD    An event stream creation delegate method in class T.
         * @param  instance  The instance of class T on which to call METHOD.
         * @return           A reference to this EventSourceBuilder object allowing calls to be chained together.
         */
        template <typename T,
                  IrisErrorCode (T::*METHOD)(EventStream*&, const EventSourceInfo &, const std::vector<std::string>&)>
        EventSourceBuilder &setEventStreamCreateDelegate(T* instance)
        {
            return setEventStreamCreateDelegate(EventStreamCreateDelegate::make<T, METHOD>(instance));
        }
    };

    /*!
     * @brief Add meta-information for an event source.
     *
     * @param name          The name of the new event source.
     * @param isHidden      If true the event source will be hidden. @see EventSourceBuilder::setHidden
     * @return              An EventSourceBuilder object that can be used to set additional
     *                      attributes for this event source.
     *                      The returned EventSourceBuilder is only valid until the next call to
     *                      addEventSource().
     */
    EventSourceBuilder addEventSource(const std::string &name, bool isHidden = false)
    {
        return EventSourceBuilder(inst_event->addEventSource(name, isHidden));
    }

    /*!
     * @brief Add meta-information for an event source that uses an IrisEventEmitter.
     *
     * @param name          The name of the new event source.
     * @param event_emitter The IrisEventEmitter for this event source.
     * @param isHidden      If true the event source will be hidden. @see EventSourceBuilder::setHidden
     * @return              An EventSourceBuilder object that can be used to set additional
     *                      attributes for this event source.
     *                      The returned EventSourceBuilder is only valid until the next call to
     *                      addEventSource(), setRegisterReadEvent(), or setRegisterWriteEvent().
     */
    EventSourceBuilder addEventSource(const std::string &name, IrisEventEmitterBase &event_emitter, bool isHidden = false)
    {
        IrisInstanceEvent::EventSourceInfoAndDelegate &info = inst_event->addEventSource(name, isHidden);

        event_emitter.setIrisInstance(iris_instance);
        event_emitter.setEvSrcId(info.info.evSrcId);
        info.createEventStream = EventStreamCreateDelegate::make<IrisEventEmitterBase,
                                                                 &IrisEventEmitterBase::createEventStream>(&event_emitter);

        return EventSourceBuilder(info);
    }

    /*!
     * @brief Add a new register read event source.
     *
     * Any registers added after calling setRegisterReadEvent() and before the next call to
     * setRegisterReadEvent() or finalizeRegisterReadEvent() will be associated with this event.
     *
     * A call to setRegisterReadEvent() will implicitly call finalizeRegisterReadEvent() on the event
     * source with name @c name iff an event emitter object (type IrisRegisterEventEmitterBase) is
     * provided as an argument.
     *
     * If the register read event source already exists (identified by name), the active register read
     * event source will simply switch to it.
     *
     * Register read events have three standard fields:
     *  Field       | Description
     *  ----------- | ------------------------------------------------
     *  @c REGISTER | The Iris rscId of the register accessed.
     *  @c DEBUG    | True if the read originated from a debug access.
     *  @c VALUE    | The value that was read.
     *
     * @param   name            Name of the event source.
     * @param   description     Description of the event source.
     * @return                  An EventSourceBuilder for the event allowing extra custom fields
     *                          to be added.
     */
    EventSourceBuilder setRegisterReadEvent(const std::string &name, const std::string &description=std::string());

    /*!
     * @brief Add a new register read event source.
     *
     * Any registers added after calling setRegisterReadEvent() and before the next call to
     * setRegisterReadEvent() or finalizeRegisterReadEvent() will be associated with this event.
     *
     * A call to setRegisterReadEvent() will implicitly call finalizeRegisterReadEvent() on the event
     * source with name @c name iff an event emitter object (type IrisRegisterEventEmitterBase) is
     * provided as an argument.
     *
     * If the register read event source already exists (identified by name), the active register read
     * event source will simply switch to it.
     *
     * Register read events have three standard fields:
     *  Field       | Description
     *  ----------- | ------------------------------------------------
     *  @c REGISTER | The Iris rscId of the register accessed.
     *  @c DEBUG    | True if the read originated from a debug access.
     *  @c VALUE    | The value that was read.
     *
     * @param   name            Name of the event source.
     * @param   event_emitter   The event_emitter to associate with this this event source.
     * @return                  An EventSourceBuilder for the event allowing extra custom fields
     *                          to be added.
     */
    EventSourceBuilder setRegisterReadEvent(const std::string &name, IrisRegisterEventEmitterBase &event_emitter);

    /*!
     * Finalize set up of an IrisEventEmitter.
     *
     * Once all the registers associated with all the read events have been added call
     * finalizeRegisterReadEvent() to add the event sources to the IrisInstance.
     */
    void finalizeRegisterReadEvent();

    /*!
     * @brief Add a new register update event source.
     *
     * Any registers added after calling setRegisterUpdateEvent() and before the next call to
     * setRegisterUpdateEvent() or finalizeRegisterUpdateEvent() will be associated with this event.
     *
     * A call to setRegisterUpdateEvent will implicitly call finalizeRegisterUpdateEvent() on the
     * event source with name @c name iff an event emitter object (type IrisRegisterEventEmitterBase)
     * is provided as an argument.
     *
     * If the register update event source (identified by name) already exists, the active register
     * update event source will simply switch to it.
     *
     * Register update events have four standard fields:
     * Field        | Description
     * ------------ | -------------------------------------------------------------
     * @c REGISTER  | The Iris rscId of the register accessed.
     * @c DEBUG     | True if the update originated from a debug access.
     * @c OLD_VALUE | The value that would have been read before the access was made.
     * @c NEW_VALUE | The value that would be read after the access was made.
     *
     * @param   name            Name of the event source.
     * @param   description     Description of the event source.
     * @return                  An EventSourceBuilder for the event allowing extra custom fields
     *                          to be added.
     */
    EventSourceBuilder setRegisterUpdateEvent(const std::string &name, const std::string &description=std::string());

    /*!
     * @brief Add a new register update event source.
     *
     * Any registers added after calling setRegisterUpdateEvent() and before the next call to
     * setRegisterUpdateEvent() or finalizeRegisterUpdateEvent() will be associated with this event.
     *
     * A call to setRegisterUpdateEvent will implicitly call finalizeRegisterUpdateEvent() on the
     * event source with name @c name iff an event emitter object (type IrisRegisterEventEmitterBase)
     * is provided as an argument.
     *
     * If the register update event source (identified by name) already exists, the active register
     * update event source will simply switch to it.
     *
     * Register update events have four standard fields:
     * Field        | Description
     * ------------ | -------------------------------------------------------------
     * @c REGISTER  | The Iris rscId of the register accessed.
     * @c DEBUG     | True if the update originated from a debug access.
     * @c OLD_VALUE | The value that would have been read before the access was made.
     * @c NEW_VALUE | The value that would be read after the access was made.
     *
     * @param   name            Name of the event source.
     * @param   event_emitter   The event_emitter to associate with this this event source.
     * @return                  An EventSourceBuilder for the event allowing extra custom fields
     *                          to be added.
     */
    EventSourceBuilder setRegisterUpdateEvent(const std::string &name, IrisRegisterEventEmitterBase &event_emitter);

    /*!
     * @brief Finalize set up of an IrisEventEmitter.
     *
     * Once all the registers associated with all the write events have been added call
     * finalizeRegisterReadEvent() to add the event sources to the IrisInstance.
     */
    void finalizeRegisterUpdateEvent();


    /*!
     * @brief Set the delegate which helps to create a new event stream for the simulation specific event.
     *
     * Event sources that do not explicitly override the access function via
     * @code addEventSource(...).setEventStreamCreateDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is an event stream create method.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode createEventStream(iris::EventStream*&, const iris::EventSourceInfo&,
                                                 const std::vector<std::string>&)>
       };

       MyClass myInstanceOfMyClass;

       EventStreamCreateDelegate delegate = EventStreamCreateDelegate::make<MyClass, &MyClass::createEventStream>(myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultEsCreateDelegateC(delegate);
       builder->addEventSource(...); // Uses createEventStream
       @endcode
     * @param   delegate    Delegate object which will be called to create an event stream.
     */
    void setDefaultEsCreateDelegate(EventStreamCreateDelegate delegate)
    {
        inst_event->setDefaultEsCreateDelegate(delegate);
    }

    /*!
     * @brief Set the delegate which helps to create a new event stream for the simulation specific event.
     *
     * Event sources that do not explicitly override the access function via
     * @code addEventSource(...).setEventStreamCreateDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is an event stream creation method.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode createEventStream(iris::EventStream*&, const iris::EventSourceInfo&,
                                                  const std::vector<std::string>&)>
        };

        MyClass myInstanceOfMyClass;

        builder->setDefaultEsCreateDelegate<MyClass, &MyClass::createEventStream>(myInstanceOfMyClass);
        builder->addEventSource(...); // Uses createEventStream
        @endcode
     * @tparam  T           Class which defines an event stream creation method.
     * @tparam  METHOD      A method of class T which is an event stream creation method.
     * @param   instance    The instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(EventStream*&, const EventSourceInfo&, const std::vector<std::string>&)>
    void setDefaultEsCreateDelegate(T *instance)
    {
        setDefaultEsCreateDelegate(EventStreamCreateDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate which helps to create a new event stream for the simulation specific event.
     *
     * Event sources that do not explicitly override the access function via
     * @code addEventSource(...).setEventStreamCreateDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * Pass in a global function to delegate event stream creation to.
     * @code
        iris::IrisErrorCode createEventStream(iris::EventStream*&, const iris::EventSourceInfo&,
                                              const std::vector<std::string>&)>

        builder->setDefaultEsCreateDelegate<&MyClass::createEventStream>();
        builder->addEventSource(...); // Uses createEventStream
        @endcode
     * @tparam  FUNC        Global function to delegate event stream creation to.
     */
    template <IrisErrorCode(*FUNC)(EventStream*&, const EventSourceInfo&, const std::vector<std::string>&)>
    void setDefaultEsCreateDelegate()
    {
        setDefaultEsCreateDelegate(EventStreamCreateDelegate::make<FUNC>());
    }

    /*!
     * @}
     */

    /*!
     * @defgroup bpt_apis IrisInstanceBuilder breakpoint APIs
     * @{
     */

    /*!
     * @brief Set the delegate which is called when a breakpoint is set.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a breakpoint set delegate.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setBreakpoint(iris::BreakpointInfo&);
       };

       MyClass myInstanceOfMyClass;

       BreakpointSetDelegate delegate = BreakpointSetDelegate::make<MyClass, &MyClass::setBreakpoint>(myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setBreakpointSetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object which will be called to set a breakpoint.
     */
    void setBreakpointSetDelegate(BreakpointSetDelegate delegate)
    {
        if (inst_breakpoint.getPtr() == NULL)
        {
            // Ensure the underlying IrisInstanceBreakpoint object is initialised too.
            inst_breakpoint.init();
        }
        user_setBreakpoint = delegate;
    }

    /*!
     * @brief Set the delegate which is called when a breakpoint is set.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a breakpoint set delegate.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode setBreakpoint(iris::BreakpointInfo&);
        };

        MyClass myInstanceOfMyClass;

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setBreakpointSetDelegate<MyClass, &MyClass::setBreakpoint>(myInstanceOfMyClass);
        @endcode
     * @tparam  T           Class which defines a breakpoint set method.
     * @tparam  METHOD      A method of class T which is a breakpoint set delegate method.
     * @param   instance    The instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(BreakpointInfo&)>
    void setBreakpointSetDelegate(T* instance)
    {
        setBreakpointSetDelegate(BreakpointSetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate which is called when a breakpoint is set.
     *
     * Usage:
     * Pass in a global function to call when a breakpoint is set.
     * @code
       iris::IrisErrorCode setBreakpoint(iris::BreakpointInfo&);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setBreakpointSetDelegate<&MyClass::setBreakpoint>();
       @endcode
     * @tparam  FUNC        Global function to call when a breakpoint is set.
     */
    template <IrisErrorCode(*FUNC)(BreakpointInfo&)>
    void setBreakpointSetDelegate()
    {
        setBreakpointSetDelegate(BreakpointSetDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate which is called when a breakpoint is deleted.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a breakpoint delete delegate.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode deleteBreakpoint(iris::BreakpointInfo&);
       };

       MyClass myInstanceOfMyClass;

       BreakpointSetDelegate delegate = BreakpointSetDelegate::make<MyClass, &MyClass::deleteBreakpoint>(myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setBreakpointDeleteDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object which will be called to delete a breakpoint.
     */
    void setBreakpointDeleteDelegate(BreakpointDeleteDelegate delegate)
    {
        if (inst_breakpoint.getPtr() == NULL)
        {
            // Ensure the underlying IrisInstanceBreakpoint object is initialised too.
            inst_breakpoint.init();
        }
        user_deleteBreakpoint = delegate;
    }

    /*!
     * @brief Set the delegate which is called when a breakpoint is deleted.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a breakpoint delete delegate.
     * @code
        class MyClass
        {
            ...
            iris::IrisErrorCode deleteBreakpoint(iris::BreakpointInfo&);
        };

        MyClass myInstanceOfMyClass;

        iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
        builder->setBreakpointDeleteDelegate<MyClass, &MyClass::deleteBreakpoint>(myInstanceOfMyClass);
        @endcode
     * @tparam  T           Class which defines a breakpoint delete method.
     * @tparam  METHOD      A method of class T which is a breakpoint delete delegate method.
     * @param   instance    The instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const BreakpointInfo&)>
    void setBreakpointDeleteDelegate(T* instance)
    {
        setBreakpointDeleteDelegate(BreakpointDeleteDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate which is called when a breakpoint is deleted.
     *
     * Usage:
     * Pass in a global function to call when a breakpoint is deleted.
     * @code
       iris::IrisErrorCode deleteBreakpoint(iris::BreakpointInfo&);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setBreakpointDeleteDelegate<&MyClass::deleteBreakpoint>();
       @endcode
     * @tparam  FUNC        Global function to call when a breakpoint is deleted.
     */
    template <IrisErrorCode(*FUNC)(const BreakpointInfo&)>
    void setBreakpointDeleteDelegate()
    {
        setBreakpointDeleteDelegate(BreakpointDeleteDelegate::make<FUNC>());
    }

    /*!
     * @brief Notify clients that a code breakpoint was hit.
     *
     * This will emit an (@c IRIS_BREAKPOINT_HIT) event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint was hit.
     * @param   pc          Value of the program counter when the breakpoint was hit.
     * @param   pcSpaceId   Memory space id for the PC when the breakpoint was hit.
     */
    void notifyBreakpointHit(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId)
    {
        inst_breakpoint->notifyBreakpointHit(bptId, time, pc, pcSpaceId);
    }

    /*!
     * @brief Notify clients that a data breakpoint was hit (@c IRIS_BREAKPOINT_HIT)
     *
     * This will emit an (@c IRIS_BREAKPOINT_HIT) event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint was hit.
     * @param   pc          Value of the program counter when the breakpoint was hit.
     * @param   pcSpaceId   Memory space id for the PC when the breakpoint was hit.
     * @param   accessAddr  Address of the access that hit.
     * @param   accessSize  Size in bytes of the access that hit.
     * @param   accessRw    Access direction. Should be "r" for a read access or "w"
     *                      for a write access.
     * @param   data        The data transferred by the access that hit.
     */
    void notifyBreakpointHitData(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId,
                                 uint64_t accessAddr, uint64_t accessSize,
                                 const std::string &accessRw, const std::vector<uint64_t> &data)
    {
        inst_breakpoint->notifyBreakpointHitData(bptId, time, pc, pcSpaceId, accessAddr, accessSize, accessRw, data);
    }

    /*!
     * @brief Notify clients that a register breakpoint was hit (@c IRIS_BREAKPOINT_HIT)
     *
     * This will emit an (@c IRIS_BREAKPOINT_HIT) event.
     *
     * @param   bptId       Breakpoint id for the breakpoint that was hit.
     * @param   time        Simulation time that the breakpoint was hit.
     * @param   pc          Value of the program counter when the breakpoint was hit.
     * @param   pcSpaceId   Memory space id for the PC when the breakpoint was hit.
     * @param   accessRw    Access direction. Should be "r" for a read access or "w"
     *                      for a write access.
     * @param   data        The data transferred by the access that hit.
     */
    void notifyBreakpointHitRegister(uint64_t bptId, uint64_t time, uint64_t pc, uint64_t pcSpaceId,
                                     const std::string &accessRw, const std::vector<uint64_t> &data)
    {
        inst_breakpoint->notifyBreakpointHitRegister(bptId, time, pc, pcSpaceId, accessRw, data);
    }

    /*!
     * @brief Get the breakpoint info for a given breakpoint.
     *
     * @param   bptId   The breakpoint id of the breakpoint for which info is being requested.
     * @return          The breakpoint info for the requested breakpoint.
     *                  This will return NULL if @a bptId is invalid.
     */
    const BreakpointInfo *getBreakpointInfo(uint64_t bptId)
    {
        return inst_breakpoint->getBreakpointInfo(bptId);
    }

    /// Add an optional conponent-specific condition
    void addBreakpointCondition(const std::string &name, const std::string &type, const std::string &description,
                                const std::vector<std::string> bpt_types=std::vector<std::string>())
    {
        inst_breakpoint->addCondition(name, type, description, bpt_types);
    }

    /*!
     * @}
     */

    /*!
     * @defgroup memory_apis IrisInstanceBuilder memory APIs
     * @{
     */

    /*!
     * @brief Used to set meta-data for a memory space.
     */
    class MemorySpaceBuilder
    {
    private:
        IrisInstanceMemory::SpaceInfoAndAccess &info;

    public:

        MemorySpaceBuilder(IrisInstanceMemory::SpaceInfoAndAccess &info_)
            : info(info_)
        {
        }

        /*!
         * @brief Set the @c name field.
         * @param   name        The name field of the MemorySpaceInfo object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setName(const std::string &name)
        {
            info.spaceInfo.name = name;
            return *this;
        }

        /*!
         * @brief Set the @c description field.
         * @param   description The description field of the MemorySpaceInfo object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setDescription(const std::string &description)
        {
            info.spaceInfo.description = description;
            return *this;
        }

        /*!
         * @brief Set the @c minsAddr field.
         * @param   minAddr     The minAddr field of the MemorySpaceInfo object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setMinAddr(uint64_t minAddr)
        {
            info.spaceInfo.minAddr = minAddr;
            return *this;
        }

        /*!
         * @brief Set the @c maxAddr field.
         * @param   maxAddr     The maxAddr field of the MemorySpaceInfo object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setMaxAddr(uint64_t maxAddr)
        {
            info.spaceInfo.maxAddr = maxAddr;
            return *this;
        }

        /*!
         * @brief Set the @c description field.
         * @param   canonicalMsn The canonicalMsn field of the MemorySpaceInfo object.
         * @return               A reference to this MemorySpaceBuilder object allowing
         *                       calls to be chained together.
         */
        MemorySpaceBuilder &setCanonicalMsn(uint64_t canonicalMsn)
        {
            info.spaceInfo.canonicalMsn = canonicalMsn;
            return *this;
        }

        /*!
         * @brief Set the @c endianness field.
         * @param   endianness  The endianness field of the MemorySpaceInfo object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setEndianness(const std::string &endianness)
        {
            info.spaceInfo.endianness = endianness;
            return *this;
        }

        /*!
         * @brief Add an attribute to the @c attrib field.
         * @param   name        The name of this attribute.
         * @param   attrib      AttributeInfo for this attribute.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &addAttribute(const std::string &name, AttributeInfo attrib)
        {
            info.spaceInfo.attrib[name] = attrib;
            return *this;
        }

        /*!
         * @brief Set the default value for an attribute in the @c attrib field.
         * @param   name        The name of this attribute.
         * @param   value       Default value of the named attribute.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setAttributeDefault(const std::string &name, IrisValue value)
        {
            info.spaceInfo.attribDefaults[name] = value;
            return *this;
        }

        /*!
         * @brief Set the delegate to read this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryReadDelegate
         *
         * @param  delegate     MemoryReadDelegate object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setReadDelegate(MemoryReadDelegate delegate)
        {
            info.readDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to write to this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryWriteDelegate
         *
         * @param  delegate     MemoryWriteDelegate object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setWriteDelegate(MemoryWriteDelegate delegate)
        {
            info.writeDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to read sideband information.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultGetMemorySidebandInfoDelegate
         *
         * @param  delegate     MemoryGetSidebandInfoDelegate object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        MemorySpaceBuilder &setSidebandDelegate(MemoryGetSidebandInfoDelegate delegate)
        {
            info.sidebandDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to read this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryReadDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a memory
         *                      read delegate.
         * @tparam METHOD       A memory read delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                                        const AttributeValueMap &, MemoryReadResult &)>
        MemorySpaceBuilder &setReadDelegate(T* instance)
        {
            return setReadDelegate(MemoryReadDelegate::make<T, METHOD>(instance));
        }

        /*!
         * @brief Set the delegate to write to this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryWriteDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a memory
         *                      write delegate.
         * @tparam METHOD       A memory write delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                                        const AttributeValueMap &, const uint64_t*, MemoryWriteResult &)>
        MemorySpaceBuilder &setWriteDelegate(T* instance)
        {
            return setWriteDelegate(MemoryWriteDelegate::make<T, METHOD>(instance));
        }

        /*!
         * @brief Set the delegate to read sideband information.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultGetMemorySidebandInfoDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a memory
         *                      sideband info delegate.
         * @tparam METHOD       A memory sideband info delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo &, uint64_t, const IrisValueMap &,
                                                        const std::vector<std::string> &, IrisValueMap &)>
        MemorySpaceBuilder &setSidebandDelegate(T* instance)
        {
            return setSidebandDelegate(MemoryGetSidebandInfoDelegate::make<T, METHOD>(instance));
        }


        /*!
         * @brief Set the delegate to read this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryReadDelegate
         *
         * @tparam FUNC         A memory read delegate function.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                       const AttributeValueMap &, MemoryReadResult &)>
        MemorySpaceBuilder &setReadDelegate()
        {
            return setReadDelegate(MemoryReadDelegate::make<FUNC>());
        }

        /*!
         * @brief Set the delegate to write to this memory space.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultMemoryWriteDelegate
         *
         * @tparam FUNC         A memory write delegate function.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                       const AttributeValueMap &, const uint64_t*, MemoryWriteResult &)>
        MemorySpaceBuilder &setWriteDelegate()
        {
            return setWriteDelegate(MemoryWriteDelegate::make<FUNC>());
        }

        /*!
         * @brief Set the delegate to read sideband information.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultGetMemorySidebandInfoDelegate
         *
         * @tparam FUNC         A memory sideband info delegate function.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(const MemorySpaceInfo &, uint64_t, const IrisValueMap &,
                                       const std::vector<std::string> &, IrisValueMap &)>
        MemorySpaceBuilder &setSidebandDelegate()
        {
            return setSidebandDelegate(MemoryGetSidebandInfoDelegate::make<FUNC>());
        }

        /*!
         * @brief Get the memory space id for this memory space.
         *
         * This can be useful for to setting up address translations and to map access
         * requests to the correct memory space in memory access delegates.
         *
         * @return              The memory space id for this memory space.
         */
        uint64_t getSpaceId() const
        {
            return info.spaceInfo.spaceId;
        }
    };

    /*!
     * @brief Used to set meta-data for an address translation.
     */
    class AddressTranslationBuilder
    {
    private:
        IrisInstanceMemory::AddressTranslationInfoAndAccess &info;

    public:
        AddressTranslationBuilder(IrisInstanceMemory::AddressTranslationInfoAndAccess &info_)
            : info(info_)
        {
        }

        /*!
         * @brief Set the delegate to perform an address translation.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultAddressTranslationDelegate
         *
         * @param  delegate     MemoryAddressTranslateDelegate object.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        AddressTranslationBuilder &setTranslateDelegate(MemoryAddressTranslateDelegate delegate)
        {
            info.translateDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to perform an address translation.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultAddressTranslationDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a memory
         *                      address translation delegate.
         * @tparam METHOD       A memory address translation delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(uint64_t, uint64_t, uint64_t, MemoryAddressTranslationResult &)>
        AddressTranslationBuilder &setTranslateDelegate(T* instance)
        {
            return setTranslateDelegate(MemoryAddressTranslateDelegate::make<T, METHOD>(instance));
        }

        /*!
         * @brief Set the delegate to perform an address translation.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultAddressTranslationDelegate
         *
         * @tparam FUNC         An address translation delegate function.
         * @return              A reference to this MemorySpaceBuilder object allowing
         *                      calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(uint64_t, uint64_t, uint64_t, MemoryAddressTranslationResult &)>
        AddressTranslationBuilder &setTranslateDelegate()
        {
            return setTranslateDelegate(MemoryAddressTranslateDelegate::make<FUNC>());
        }

    };

    /*!
     * @brief Set memory.canonicalMsnScheme instance property.
     *
     * This property is visible in the list of properties returned by instance_getProperties().
     *
     * This property defines the scheme used by the 'canonicalMsn' member of the MemorySpaceInfo object.
     * The default is 'arm.com/memoryspaces' which is used by all ARM components.
     * This default can be overridden by calling this function.
     * This should be called upon initialisation, before other instances have a chance to call instance_getProperties().
     *
     * @param   canonicalMsnScheme  Name of the canonical memory space number scheme used by this instance.
     */
    void setPropertyCanonicalMsnScheme(const std::string &canonicalMsnScheme);


    /*!
     * @brief Set default read function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode readMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                          uint64_t byteWidth, uint64_t count,
                                          const iris::IrisValueMap &attrib,
                                          iris::MemoryReadResult &result);
       };

       MyClass myInstanceOfMyClass;
       iris::MemoryReadDelegate delegate =
           iris::MemoryReadDelegate::make<MyClass, &MyClass::readMemory>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryReadDelegate(delegate);
       builder->addMemorySpace(...); // Uses readMemory
       @endcode
     * @param   delegate    Delegate object which will be called to read memory.
     */
    void setDefaultMemoryReadDelegate(MemoryReadDelegate delegate = MemoryReadDelegate())
    {
        inst_memory->setDefaultReadDelegate(delegate);
    }

    /*!
     * @brief Set default read function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode readMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                          uint64_t byteWidth, uint64_t count,
                                          const iris::IrisValueMap &attrib,
                                          iris::MemoryReadResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryReadDelegate<MyClass, &MyClass::readMemory>(myInstanceOfMyClass);
       builder->addMemorySpace(...); // Uses readMemory
       @endcode
     * @tparam  T           Class which defines a memory read delegate method.
     * @tparam  METHOD      A method of class T which is a memory read delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                                    const AttributeValueMap &, MemoryReadResult &)>
    void setDefaultMemoryReadDelegate(T* instance)
    {
        setDefaultMemoryReadDelegate(MemoryReadDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default read function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       iris::IrisErrorCode readMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                      uint64_t byteWidth, uint64_t count,
                                      const iris::IrisValueMap &attrib,
                                      iris::MemoryReadResult &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryReadDelegate<readMemory>();

       builder->addMemorySpace(...); // Uses readMemory
       @endcode
     * @tparam FUNC         A memory read delegate function.
     */
    template <IrisErrorCode(*FUNC)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                   const AttributeValueMap &, MemoryReadResult &)>
    void setDefaultMemoryReadDelegate()
    {
        setDefaultMemoryReadDelegate(MemoryReadDelegate::make<FUNC>());
    }

    /*!
     * @brief Set default write function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode writeMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                          uint64_t byteWidth, uint64_t count,
                                          const iris::IrisValueMap &attrib,
                                          const uint64_t *data,
                                          iris::MemoryWriteResult &result);
       };

       MyClass myInstanceOfMyClass;
       iris::MemoryReadDelegate delegate =
           iris::MemoryWriteDelegate::make<MyClass, &MyClass::writeMemory>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryWriteDelegate(delegate);
       builder->addMemorySpace(...); // Uses writeMemory
       @endcode
     * @param   delegate    Delegate object which will be called to write memory.
     */
    void setDefaultMemoryWriteDelegate(MemoryWriteDelegate delegate = MemoryWriteDelegate())
    {
        inst_memory->setDefaultWriteDelegate(delegate);
    }

    /*!
     * @brief Set default write function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode writeMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                          uint64_t byteWidth, uint64_t count,
                                          const iris::IrisValueMap &attrib,
                                          const uint64_t *data,
                                          iris::MemoryWriteResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryWriteDelegate<MyClass, &MyClass::writeMemory>(&myInstanceOfMyClass);
       builder->addMemorySpace(...); // Uses writeMemory
       @endcode
     * @tparam  T           Class which defines a memory read delegate method.
     * @tparam  METHOD      A method of class T which is a memory read delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                                    const AttributeValueMap &, const uint64_t*, MemoryWriteResult &)>
    void setDefaultMemoryWriteDelegate(T* instance)
    {
        setDefaultMemoryWriteDelegate(MemoryWriteDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default write function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * Pass an instance of class T where T::METHOD() is a memory read method.
     * @code
       iris::IrisErrorCode writeMemory(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                      uint64_t byteWidth, uint64_t count,
                                      const iris::IrisValueMap &attrib,
                                      const uint64_t *data,
                                      iris::MemoryWriteResult &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultMemoryWriteDelegate<&writeMemory>();
       builder->addMemorySpace(...); // Uses writeMemory
       @endcode
     * @tparam  FUNC        Global function to call to write memory.
     */
    template <IrisErrorCode(*FUNC)(const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                                   const AttributeValueMap &, const uint64_t*, MemoryWriteResult &)>
    void setDefaultMemoryWriteDelegate()
    {
        setDefaultMemoryWriteDelegate(MemoryWriteDelegate::make<FUNC>());
    }

    /*!
     * @brief Add meta-information for one memory space.
     *
     * Typical use pattern:
     * @code
       addMemorySpace("name")
           .setDescription("description")
           .setMinAddr(...)
           .setMaxAddr(...)
           .setEndianness(...)
           .addAttribute(...)
           .addAttributeDefault(...);
       @endcode

       @param   name    Name of the memory space to add.
       @return          A MemorySpaceBuilder object which can be used to configure meta-data
                        for the memory space.
     */
    MemorySpaceBuilder addMemorySpace(const std::string &name)
    {
        return MemorySpaceBuilder(inst_memory->addMemorySpace(name));
    }

    /*!
     * @brief Set default address translation function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setTranslationDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode translateAddress(uint64_t inSpaceId, uint64_t address,
                                                uint64_t outSpaceId,
                                                iris::MemoryAddressTranslationResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::MemoryAddressTranslateDelegate delegate =
           iris::MemoryAddressTranslateDelegate::make<MyClass, &MyClass::translateAddress>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultAddressTranslateDelegate(delegate);
       builder->addMemorySpace(...); // Uses translateAddress
       @endcode
     * @param   delegate    Delegate object which will be called to translate addresses.
     */
    void setDefaultAddressTranslateDelegate(MemoryAddressTranslateDelegate delegate = MemoryAddressTranslateDelegate())
    {
        inst_memory->setDefaultTranslateDelegate(delegate);
    }

    /*!
     * @brief Set default address translation function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setTranslationDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode translateAddress(uint64_t inSpaceId, uint64_t address,
                                                uint64_t outSpaceId,
                                                iris::MemoryAddressTranslationResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultAddressTranslateDelegate<MyClass, &MyClass::translateAddress>(&myInstanceOfMyClass);
       builder->addMemorySpace(...); // Uses translateAddress
       @endcode
     * @tparam  T           Class which defines an address translation delegate method.
     * @tparam  METHOD      A method of class T which is an address translation delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t, uint64_t, uint64_t, MemoryAddressTranslationResult &)>
    void setDefaultAddressTranslateDelegate(T* instance)
    {
        setDefaultAddressTranslateDelegate(MemoryAddressTranslateDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default address translation function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the access function via
     * @code addMemorySpace(...).setTranslationDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       iris::IrisErrorCode translateAddress(uint64_t inSpaceId, uint64_t address,
                                            uint64_t outSpaceId,
                                            iris::MemoryAddressTranslationResult &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultAddressTranslateDelegate<&translateAddress>();
       builder->addMemorySpace(...); // Uses translateAddress
       @endcode
     * @tparam  FUNC        Global function to call to translate addresses.
     */
    template <IrisErrorCode(*FUNC)(uint64_t, uint64_t, uint64_t, MemoryAddressTranslationResult &)>
    void setDefaultAddressTranslateDelegate()
    {
        setDefaultAddressTranslateDelegate(MemoryAddressTranslateDelegate::make<FUNC>());
    }

    /*!
     * @brief Add an address translation.
     *
     * Add meta-data for the address translation from the memory space indicated by @a inSpaceId
     * to the memory space indicated by @a outSpaceId.
     *
     * By adding an address translation explicitly using this function the Iris instance can
     * tell clients which address translations are supported and a component can provide a
     * specific delegate function to perform that translation.
     *
     * @param   inSpaceId   Memory space id for the input memory space of this translation.
     * @param   outSpaceId  Memory space id for the output memory space of this translation.
     * @param   description A human readable description of this translation.
     * return               An AddressTranslationBuilder object which allows additional
     *                      configuration of this translation.
     */
    AddressTranslationBuilder addAddressTranslation(uint64_t inSpaceId, uint64_t outSpaceId,
                                                    const std::string &description)
    {
        return AddressTranslationBuilder(inst_memory->addAddressTranslation(inSpaceId, outSpaceId, description));
    }

    /*!
     * @brief Set default sideband info function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the sideband function via
     * @code addMemorySpace(...).setSidebandDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getSidebandInfo(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                               const iris::IrisValueMap &attrib,
                                               const std::vector<std::string> &request,
                                               iris::IrisValueMap &result);
       };

       MyClass myInstanceOfMyClass;

       iris::MemoryAddressTranslateDelegate delegate =
           iris::MemoryAddressTranslateDelegate::make<MyClass, &MyClass::getSidebandInfo>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultGetMemorySidebandInfoDelegate(delegate);
       builder->addMemorySpace(...); // Uses getSidebandInfo
       @endcode
     * @param   delegate    Delegate object which will be called to get sideband info.
     */
    void setDefaultGetMemorySidebandInfoDelegate(MemoryGetSidebandInfoDelegate delegate)
    {
        inst_memory->setDefaultGetSidebandInfoDelegate(delegate);
    }

    /*!
     * @brief Set default sideband info function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the sideband function via
     * @code addMemorySpace(...).setSidebandDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getSidebandInfo(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                               const iris::IrisValueMap &attrib,
                                               const std::vector<std::string> &request,
                                               iris::IrisValueMap &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultGetMemorySidebandInfoDelegate<MyClass, &MyClass::getSidebandInfo>(&myInstanceOfMyClass);
       builder->addMemorySpace(...); // Uses getSidebandInfo
       @endcode
     * @tparam  T           Class which defines a sideband info delegate method.
     * @tparam  METHOD      A method of class T which is a sideband info delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const MemorySpaceInfo &, uint64_t, const IrisValueMap &,
                                                    const std::vector<std::string> &, IrisValueMap &)>
    void setDefaultGetMemorySidebandInfoDelegate(T *instance)
    {
        setDefaultGetMemorySidebandInfoDelegate(MemoryGetSidebandInfoDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default sideband info function for all subsequently added memory spaces.
     *
     * Memory spaces that do not explicitly override the sideband function via
     * @code addMemorySpace(...).setSidebandDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       iris::IrisErrorCode getSidebandInfo(const iris::MemorySpaceInfo &spaceInfo, uint64_t address,
                                           const iris::IrisValueMap &attrib,
                                           const std::vector<std::string> &request,
                                           iris::IrisValueMap &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultGetMemorySidebandInfoDelegate<&getSidebandInfo>();
       builder->addMemorySpace(...); // Uses getSidebandInfo
       @endcode
     * @tparam  FUNC        Global function to call to get sideband info.
     */
    template <IrisErrorCode(*FUNC)(const MemorySpaceInfo &, uint64_t, const IrisValueMap &,
                                   const std::vector<std::string> &, IrisValueMap &)>
    void setDefaultGetMemorySidebandInfoDelegate()
    {
        setDefaultGetMemorySidebandInfoDelegate(MemoryGetSidebandInfoDelegate::make<FUNC>());
    }

    /*!
     * @}
     */

    /*!
     * @defgroup image_load_apis IrisInstanceBuilder image loading APIs
     * @{
     */

    /*!
    * @brief Set the delegate to load an image from a file
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode loadImageFile(const std::string &path);
       };

       MyClass myInstanceOfMyClass;

       iris::MemoryAddressTranslateDelegate delegate =
           iris::MemoryAddressTranslateDelegate::make<MyClass, &MyClass::loadImageFile>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageFileDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call for image loading.
     */
    void setLoadImageFileDelegate(ImageLoadFileDelegate delegate = ImageLoadFileDelegate())
    {
        inst_image->setLoadImageFileDelegate(delegate);
    }

    /*!
    * @brief Set the delegate to load an image from a file
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode loadImageFile(const std::string &path);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageFileDelegate<MyClass, &MyClass::loadImageFile>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines an image loading delegate method.
     * @tparam  METHOD      A method of class T which is an image loading delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const std::string &)>
    void setLoadImageFileDelegate(T *instance)
    {
        setLoadImageFileDelegate(ImageLoadFileDelegate::make<T, METHOD>(instance));
    }

    /*!
    * @brief Set the delegate to load an image from a file
     *
     * Usage:
     * @code
       iris::IrisErrorCode loadImageFile(const std::string &path);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageFileDelegate<&loadImageFile>();
       @endcode
     * @tparam  FUNC        Global function to call for image loading.
     */
    template <IrisErrorCode(*FUNC)(const std::string &)>
    void setLoadImageFileDelegate()
    {
        setLoadImageFileDelegate(ImageLoadFileDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate to load an image from provided data.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode loadImageData(const std::vector<uint64_t> &data, uint64_t dataSizeInBytes);
       };

       MyClass myInstanceOfMyClass;

       iris::MemoryAddressTranslateDelegate delegate =
           iris::MemoryAddressTranslateDelegate::make<MyClass, &MyClass::loadImageData>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageDataDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call for image loading.
     */
    void setLoadImageDataDelegate(ImageLoadDataDelegate delegate = ImageLoadDataDelegate())
    {
        inst_image->setLoadImageDataDelegate(delegate);
    }

    /*!
    * @brief Set the delegate to load an image from provided data.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode loadImageData(const std::vector<uint64_t> &data, uint64_t dataSizeInBytes);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageDataDelegate<MyClass, &MyClass::loadImageData>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines an image loading delegate method.
     * @tparam  METHOD      A method of class T which is an image loading delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const std::vector<uint64_t> &, uint64_t)>
    void setLoadImageDataDelegate(T *instance)
    {
        setLoadImageDataDelegate(ImageLoadDataDelegate::make<T, METHOD>(instance));
    }

    /*!
    * @brief Set the delegate to load an image from provided data.
     *
     * Usage:
     * @code
           iris::IrisErrorCode loadImageData(const std::vector<uint64_t> &data, uint64_t dataSizeInBytes);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setLoadImageDataDelegate<&loadImageData>();
       @endcode
     * @tparam  FUNC        Global function to call for image loading.
     */
    template <IrisErrorCode(*FUNC)(const std::vector<uint64_t> &, uint64_t)>
    void setLoadImageDataDelegate()
    {
        setLoadImageDataDelegate(ImageLoadDataDelegate::make<FUNC>());
    }

    /*!
     * @}
     */

    /*!
     * @defgroup image_readdata_apis IrisInstanceBuilder image readData callback APIs
     *
     * @{
     */

    /*!
     * @brief Open an image to be read using image_loadDataPull()/image_loadDataRead().
     * @param      filename  The name of the file to be read.
     * @return               The tag number to use when calling image_loadDataPull().
     */
    uint64_t openImage(const std::string &filename)
    {
        return inst_image_cb->openImage(filename);
    }

    /*!
     * @}
     */

    /*!
     * @defgroup step_apis IrisInstanceBuilder execution stepping APIs
     * @{
     */

    /*!
    * @brief Set the delegate to set remaining steps for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setRemainingSteps(uint64_t steps, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::RemainingStepSetDelegate delegate =
           iris::RemainingStepSetDelegate::make<MyClass, &MyClass::setRemainingSteps>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepSetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to set remaining steps.
     */
    void setRemainingStepSetDelegate(RemainingStepSetDelegate delegate = RemainingStepSetDelegate())
    {
        inst_step->setRemainingStepSetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to get remaining steps for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getRemainingSteps(uint64_t &steps, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::RemainingStepGetDelegate delegate =
           iris::RemainingStepGetDelegate::make<MyClass, &MyClass::getRemainingSteps>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepGetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to get remaining steps.
     */
    void setRemainingStepGetDelegate(RemainingStepGetDelegate delegate)
    {
        inst_step->setRemainingStepGetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to set remaining steps for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setRemainingSteps(uint64_t steps, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepSetDelegate<MyClass, &MyClass::setRemainingSteps>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a set remaining steps delegate method.
     * @tparam  METHOD      A method of class T which is a set remaining steps delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t, const std::string &)>
    void setRemainingStepSetDelegate(T* instance)
    {
        setRemainingStepSetDelegate(RemainingStepSetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to get remaining steps for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getRemainingSteps(uint64_t &steps, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepGetDelegate<MyClass, &MyClass::getRemainingSteps>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a get remaining steps delegate method.
     * @tparam  METHOD      A method of class T which is a get remaining steps delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t &, const std::string &)>
    void setRemainingStepGetDelegate(T* instance)
    {
        setRemainingStepGetDelegate(RemainingStepGetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to set remaining steps for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode setRemainingSteps(uint64_t steps, const std::string &unit);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepSetDelegate<&setRemainingSteps>();
       @endcode
     * @tparam  FUNC        Global function to call to set remaining steps.
     */
    template <IrisErrorCode(*FUNC)(uint64_t, const std::string &)>
    void setRemainingStepSetDelegate()
    {
        setRemainingStepSetDelegate(RemainingStepSetDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate to get remaining steps for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode getRemainingSteps(uint64_t &steps, const std::string &unit);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setRemainingStepGetDelegate<&getRemainingSteps>();
       @endcode
     * @tparam  FUNC        Global function to call to get remaining steps.
     */
    template <IrisErrorCode(*FUNC)(uint64_t &, const std::string &)>
    void setRemainingStepGetDelegate()
    {
        setRemainingStepGetDelegate(RemainingStepGetDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate to get step count for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getStepCount(uint64_t &count, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::StepCountGetDelegate delegate =
           iris::StepCountGetDelegate::make<MyClass, &MyClass::getStepCount>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setStepCountGetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to get step count.
     */
    //
    void setStepCountGetDelegate(StepCountGetDelegate delegate = StepCountGetDelegate())
    {
        inst_step->setStepCountGetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to get step count for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getStepCount(uint64_t &count, const std::string &unit);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setStepCountGetDelegate<MyClass, &MyClass::getStepCount>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a get step count delegate method.
     * @tparam  METHOD      A method of class T which is a get step count delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t &, const std::string &)>
    void setStepCountGetDelegate(T* instance)
    {
        setStepCountGetDelegate(RemainingStepGetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to get step count for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode getStepCount(uint64_t &count, const std::string &unit);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setStepCountGetDelegate<&getStepCount>();
       @endcode
     * @tparam  FUNC        Global function to call to get step count.
     */
    template <IrisErrorCode(*FUNC)(uint64_t &, const std::string &)>
    void setStepCountGetDelegate()
    {
        setStepCountGetDelegate(RemainingStepGetDelegate::make<FUNC>());
    }

    /*!
     * @}
     */

    /*
     * @brief exec_apis IrisInstanceBuilder per-instance execution APIs
     * @{
     */

    /*!
     * @brief Set the delegate to set execution state for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setState(bool enable_execution);
       };

       MyClass myInstanceOfMyClass;

       iris::PerInstanceExecutionStateSetDelegate delegate =
           iris::PerInstanceExecutionStateSetDelegate::make<MyClass, &MyClass::setState>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateSetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to set execution state.
     */
    void setExecutionStateSetDelegate(PerInstanceExecutionStateSetDelegate delegate = PerInstanceExecutionStateSetDelegate())
    {
        inst_per_inst_exec->setExecutionStateSetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to set execution state for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setState(bool enable_execution);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateSetDelegate<MyClass, &MyClass::setState>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a set execution state delegate method.
     * @tparam  METHOD      A method of class T which is a set execution state delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(bool)>
    void setExecutionStateSetDelegate(T *instance)
    {
        setExecutionStateSetDelegate(PerInstanceExecutionStateSetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to set execution state for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode setState(bool enable_execution);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateSetDelegate<&setState>();
       @endcode
     * @tparam  FUNC        Global function to call to set execution state.
     */
    template <IrisErrorCode(*FUNC)(bool)>
    void setExecutionStateSetDelegate()
    {
        setExecutionStateSetDelegate(PerInstanceExecutionStateSetDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate to get execution state for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getState(bool &execution_enabled);
       };

       MyClass myInstanceOfMyClass;

       iris::PerInstanceExecutionStateGetDelegate delegate =
           iris::PerInstanceExecutionStateGetDelegate::make<MyClass, &MyClass::getState>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateGetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to get execution state.
     */
    void setExecutionStateGetDelegate(PerInstanceExecutionStateGetDelegate delegate)
    {
        inst_per_inst_exec->setExecutionStateGetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to get execution state for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getState(bool &execution_enabled);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateGetDelegate<MyClass, &MyClass::getState>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a get execution state delegate method.
     * @tparam  METHOD      A method of class T which is a get execution state delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(bool&)>
    void setExecutionStateGetDelegate(T *instance)
    {
        setExecutionStateGetDelegate(PerInstanceExecutionStateGetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to get execution state for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode getState(bool &execution_enabled);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setExecutionStateGetDelegate<&getState>();
       @endcode
     * @tparam  FUNC        Global function to call to get execution state.
     */
    template <IrisErrorCode(*FUNC)(bool&)>
    void setExecutionStateGetDelegate()
    {
        setExecutionStateGetDelegate(PerInstanceExecutionStateGetDelegate::make<FUNC>());
    }

    /*!
     * @}
     */


    /*
     * @brief table_apis IrisInstanceBuilder table APIs
     * @{
     */

    class TableColumnBuilder;

    /*!
     * @brief Used to set meta-data for a table.
     */
    class TableBuilder
    {
    private:
        IrisInstanceTable::TableInfoAndAccess &info;

    public:
        TableBuilder(IrisInstanceTable::TableInfoAndAccess &info_)
            : info(info_)
        {
        }

        /*!
         * @brief Set the @c name field.
         * @param   name            The name field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setName(const std::string &name)
        {
            info.tableInfo.name = name;
            return *this;
        }

        /*!
         * @brief Set the @c description field.
         * @param   description     The description field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setDescription(const std::string &description)
        {
            info.tableInfo.description = description;
            return *this;
        }

        /*!
         * @brief Set the @c minIndex field.
         * @param   minIndex        The minIndex field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setMinIndex(uint64_t minIndex)
        {
            info.tableInfo.minIndex = minIndex;
            return *this;
        }

        /*!
         * @brief Set the @c maxIndex field.
         * @param   maxIndex        The maxIndex field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setMaxIndex(uint64_t maxIndex)
        {
            info.tableInfo.maxIndex = maxIndex;
            return *this;
        }

        /*!
         * @brief Set the @c indexFormatHint field.
         * @param   hint            The indexFormatHint field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setIndexFormatHint(const std::string &hint)
        {
            info.tableInfo.indexFormatHint = hint;
            return *this;
        }

        /*!
         * @brief Set the @c formatShort field.
         * @param   format          The formatShort field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setFormatShort(const std::string &format)
        {
            info.tableInfo.formatShort = format;
            return *this;
        }

        /*!
         * @brief Set the @c formatLong field.
         * @param   format          The formatLong field of the TableInfo object.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &setFormatLong(const std::string &format)
        {
            info.tableInfo.formatLong = format;
            return *this;
        }

        /*!
         * @brief Set the delegate to read the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableReadDelegate
         *
         * @param  delegate     TableReadDelegate object.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        TableBuilder &setReadDelegate(TableReadDelegate delegate)
        {
            info.readDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to write to the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableWriteDelegate
         *
         * @param  delegate     TableWriteDelegate object.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        TableBuilder &setWriteDelegate(TableWriteDelegate delegate)
        {
            info.writeDelegate = delegate;
            return *this;
        }

        /*!
         * @brief Set the delegate to read the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableReadDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a table read delegate.
         * @tparam METHOD       A table read delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(const TableInfo&, uint64_t, uint64_t, TableReadResult &)>
        TableBuilder &setReadDelegate(T *instance)
        {
            return setReadDelegate(TableReadDelegate::make<T, METHOD>(instance));
        }

        /*!
         * @brief Set the delegate to write to the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableWriteDelegate
         *
         * @tparam T            A class that defines a method with the right signature to be a table write delegate.
         * @tparam METHOD       A table write delegate method in class T.
         * @param  instance     The instance of class T on which to call METHOD.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        template <typename T, IrisErrorCode(T::*METHOD)(const TableInfo&, const TableRecords &, TableWriteResult &)>
        TableBuilder &setWriteDelegate(T *instance)
        {
            return setWriteDelegate(TableWriteDelegate::make<T, METHOD>(instance));
        }

        /*!
         * @brief Set the delegate to read the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableReadDelegate
         *
         * @tparam FUNC         A table read delegate function.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(const TableInfo&, uint64_t, uint64_t, TableReadResult &)>
        TableBuilder &setReadDelegate()
        {
            return setReadDelegate(TableReadDelegate::make<FUNC>());
        }

        /*!
         * @brief Set the delegate to write to the table.
         *
         * If this is not set the default delegate will be used.
         * @see IrisInstanceBuilder::setDefaultTableWriteDelegate
         *
         * @tparam FUNC         A table write delegate function.
         * @return              A reference to this TableBuilder object allowing calls to be chained together.
         */
        template <IrisErrorCode(*FUNC)(const TableInfo&, const TableRecords &, TableWriteResult &)>
        TableBuilder &setWriteDelegate()
        {
            return setWriteDelegate(TableWriteDelegate::make<FUNC>());
        }

        /*!
         * @brief Add a column with preconstructed TableColumnInfo.
         *
         * Call this multiple times for multiple columns.
         *
         * @see addColumn
         *
         * @param   columnInfo      A preconstructed TableColumnInfo object for new column.
         * @return                  A reference to this TableBuilder allowing calls to be chained together.
         */
        TableBuilder &addColumnInfo(const TableColumnInfo &columnInfo)
        {
            info.tableInfo.columns.push_back(columnInfo);
            return *this;
        }

        /*!
         * @brief Add a new column.
         *
         * Call this multiple times for multiple columns
         *
         * @see AddColumnInfo
         *
         * @param   name            The name of the new column.
         * @return                  A TableColumnBuilder object that can be used
         *                          to add meta-data for the new column.
         */
        TableColumnBuilder addColumn(const std::string &name);
    };

    /*!
     * @brief Used to set meta-data for a table column.
     */
    class TableColumnBuilder
    {
    private:
        TableBuilder &parent;
        TableColumnInfo &info;

    public:
        TableColumnBuilder(TableBuilder &parent_, TableColumnInfo &info_)
            : parent(parent_)
            , info(info_)
        {
        }

        /*!
         * @brief Add another column with preconstructed TableColumnInfo.
         *
         * @see TableBuilder::addColumnInfo
         * @see addColumn
         *
         * @param   columnInfo      A preconstructed TableColumnInfo object for new column.
         * @return                  A reference to the parent TableBuilder for this table.
         */
        TableBuilder &addColumnInfo(const TableColumnInfo &columnInfo)
        {
            return parent.addColumnInfo(columnInfo);
        }

        /*!
         * @brief Add another new column.
         *
         * Call this multiple times for multiple columns
         *
         * @see TableBuilder::addColumn
         *
         * @param   name            The name of the new column.
         * @return                  A TableColumnBuilder object that can be used
         *                          to add meta-data for the new column.
         */
        TableColumnBuilder addColumn(const std::string &name)          { return parent.addColumn(name); }

        /*!
         * @brief Stop building this column and go back to the parent table.
         *
         * @see addColumn
         * @see addColumnInfo
         *
         * @return                  The parent TableBuilder for this table.
         */
        TableBuilder &endColumn()
        {
            return parent;
        }

        /*!
         * @brief Set the @c name field.
         *
         * @param   name        The name field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setName(const std::string &name)
        {
            info.name = name;
            return *this;
        }

        /*!
         * @brief Set the @c description field.
         *
         * @param   description The description field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setDescription(const std::string &description)
        {
            info.description = description;
            return *this;
        }

        /*!
         * @brief Set the @c format field.
         *
         * @param   format      The format field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setFormat(const std::string &format)
        {
            info.format = format;
            return *this;
        }

        /*!
         * @brief Set the @c type field.
         *
         * @param   type        The type field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setType(const std::string &type)
        {
            info.type = type;
            return *this;
        }

        /*!
         * @brief Set the @c bitWidth field.
         *
         * @param   bitWidth    The bitWidth field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setBitWidth(uint64_t bitWidth)
        {
            info.bitWidth = bitWidth;
            return *this;
        }

        /*!
         * @brief Set the @c formatShort field.
         *
         * @param   format      The formatShort field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setFormatShort(const std::string &format)
        {
            info.formatShort = format;
            return *this;
        }

        /*!
         * @brief Set the @c formatLong field.
         *
         * @param   format      The formatLong field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setFormatLong(const std::string &format)
        {
            info.formatLong = format;
            return *this;
        }

        /*!
         * @brief Set the @c rwMode field.
         *
         * @param   rwMode      The rwMode field of the TableColumnInfo object.
         * @return              A TableColumnBuilder object that can be used to add meta-data for the new column.
         */
        TableColumnBuilder &setRwMode(const std::string &rwMode)
        {
            info.rwMode = rwMode;
            return *this;
        }
    };

    /*!
     * @brief Add meta-information for one table.
     *
     * Typical use pattern:
     * @code
       addTableInfo("name")
           .setDescription("description")
           .setMinIndex(...)
           .setMaxIndex(...)
           .setIndexFormatHint(...)
           .setFormatShort(...)
           .setFormatLong(...)
           .setReadDelegate(...)
           .setWriteDelegate(...)
           .addColumnInfo(...)
           .addColumnInfo(...)
           ...
       @endcode

       @param   name    Name of the new table.
       @return          A TableBuilder object than can be used to set meta-data for the new table.
     */
    TableBuilder addTable(const std::string &name)
    {
        return TableBuilder(inst_table->addTableInfo(name));
    }

    /*!
     * @brief Set default table read function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode readTable(const iris::TableInfo &tableInfo, uint64_t index,
                                         uint64_t count, iris::TableReadResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::TableReadDelegate delegate =
           iris::TableReadDelegate::make<MyClass, &MyClass::readTable>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableReadDelegate(delegate);
       builder->addTable(...); // Uses readTable
       @endcode
     * @param   delegate    Delegate object to call to read a table.
     */
    void setDefaultTableReadDelegate(TableReadDelegate delegate = TableReadDelegate())
    {
        inst_table->setDefaultReadDelegate(delegate);
    }

    /*!
     * @brief Set default table write function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode writeTable(const iris::TableInfo &tableInfo,
                                          const iris::TableRecords &records,
                                          iris::TableWriteResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::TableWriteDelegate delegate =
           iris::TableWriteDelegate::make<MyClass, &MyClass::writeTable>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableWriteDelegate(delegate);
       builder->addTable(...); // Uses writeTable
       @endcode
     * @param   delegate    Delegate object to call to write a table.
     */
    void setDefaultTableWriteDelegate(TableWriteDelegate delegate = TableWriteDelegate())
    {
        inst_table->setDefaultWriteDelegate(delegate);
    }

    /*!
     * @brief Set default table read function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode readTable(const iris::TableInfo &tableInfo, uint64_t index,
                                         uint64_t count, iris::TableReadResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableReadDelegate<MyClass, &MyClass::readTable>(&myInstanceOfMyClass);
       builder->addTable(...); // Uses readTable
       @endcode
     * @tparam  T           Class which defines a table read delegate method.
     * @tparam  METHOD      A method of class T which is a table read delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const TableInfo&, uint64_t, uint64_t, TableReadResult &)>
    void setDefaultTableReadDelegate(T *instance)
    {
        setDefaultTableReadDelegate(TableReadDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set default table write function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode writeTable(const iris::TableInfo &tableInfo,
                                          const iris::TableRecords &records,
                                          iris::TableWriteResult &result);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableWriteDelegate<MyClass, &MyClass::writeTable>(&myInstanceOfMyClass);
       builder->addTable(...); // Uses writeTable
       @endcode
     * @tparam  T           Class which defines a table write delegate method.
     * @tparam  METHOD      A method of class T which is a table write delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const TableInfo&, const TableRecords &, TableWriteResult &)>
    void setDefaultTableWriteDelegate(T *instance)
    {
        setDefaultTableWriteDelegate(TableWriteDelegate::make<T, METHOD>(instance));
    }

     /*!
     * @brief Set default table read function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setReadDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       iris::IrisErrorCode readTable(const iris::TableInfo &tableInfo, uint64_t index,
                                     uint64_t count, iris::TableReadResult &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableReadDelegate<&readTable>();
       builder->addTable(...); // Uses readTable
       @endcode
     * @tparam  FUNC        Global function to call to read a table.
     */
    template <IrisErrorCode(*FUNC)(const TableInfo&, uint64_t, uint64_t, TableReadResult &)>
    void setDefaultTableReadDelegate()
    {
        setDefaultTableReadDelegate(TableReadDelegate::make<FUNC>());
    }

    /*!
     * @brief Set default table write function for all subsequently added tables.
     *
     * Tables that do not explicitly override the access function via
     * @code addTable(...).setWriteDelegate(...) @endcode
     * will use this delegate.
     *
     * Usage:
     * @code
       iris::IrisErrorCode writeTable(const iris::TableInfo &tableInfo,
                                      const iris::TableRecords &records,
                                      iris::TableWriteResult &result);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDefaultTableWriteDelegate<&writeTable>();
       builder->addTable(...); // Uses writeTable
       @endcode
     * @tparam  FUNC        Global function to call to write a table.
     */
    template <IrisErrorCode(*FUNC)(const TableInfo&, const TableRecords &, TableWriteResult &)>
    void setDefaultTableWriteDelegate()
    {
        setDefaultTableWriteDelegate(TableWriteDelegate::make<FUNC>());
    }

    /*!
     * @}
     */

    /*!
     * @brief disass_apis IrisInstanceBuilder disassembler APIs
     * @{
     */

    /// Set the delegates to get the current disassembly mode
    void setGetCurrentDisassemblyModeDelegate(GetCurrentDisassemblyModeDelegate delegate)
    {
        inst_disass->setGetCurrentModeDelegate(delegate);
    }

    template <typename T, IrisErrorCode(T::*METHOD)(std::string &)>
    void setGetCurrentDisassemblyModeDelegate(T *instance)
    {
        setGetCurrentDisassemblyModeDelegate(GetCurrentDisassemblyModeDelegate::make<T, METHOD>(instance));
    }

    /// Set the delegate to get the disassembly of a chunk of memory
    void setGetDisassemblyDelegate(GetDisassemblyDelegate delegate)
    {
        inst_disass->setGetDisassemblyDelegate(delegate);
    }

    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t, const std::string &, MemoryReadResult &,
                                                    uint64_t, uint64_t, std::vector<DisassemblyLine> &)>
    void setGetDisassemblyDelegate(T *instance)
    {
        setGetDisassemblyDelegate(GetDisassemblyDelegate::make<T, METHOD>(instance));
    }

    template <IrisErrorCode(*FUNC)(uint64_t, const std::string &, MemoryReadResult &,
                                   uint64_t, uint64_t, std::vector<DisassemblyLine> &)>
    void setGetDisassemblyDelegate()
    {
        setGetDisassemblyDelegate(GetDisassemblyDelegate::make<FUNC>());
    }

    /// Set the delegate to get the disassembly of Opcode
    void setDisassembleOpcodeDelegate(DisassembleOpcodeDelegate delegate)
    {
        inst_disass->setDisassembleOpcodeDelegate(delegate);
    }

    template <typename T, IrisErrorCode(T::*METHOD)(const std::vector<uint64_t> &, uint64_t, const std::string &,
                                                    DisassembleContext &, DisassemblyLine &)>
    void setDisassembleOpcodeDelegate(T *instance)
    {
        setDisassembleOpcodeDelegate(DisassembleOpcodeDelegate::make<T, METHOD>(instance));
    }

    template <IrisErrorCode(*FUNC)(const std::vector<uint64_t> &, uint64_t, const std::string &,
                                   DisassembleContext &, DisassemblyLine &)>
    void setDisassembleOpcodeDelegate()
    {
        setDisassembleOpcodeDelegate(DisassembleOpcodeDelegate::make<FUNC>());
    }

    ///  Add a disassembly mode
    void addDisassemblyMode(const std::string &name, const std::string &description)
    {
        inst_disass->addDisassemblyMode(name, description);
    }

    /*!
     * @}
     */

    /*!
     * @brief debuggable_state_apis IrisInstanceBuilder debuggable state APIs
     * @{
     */

    /*!
     * @brief Set the delegate to set the debuggable state request flag for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setRequestFlag(bool request_debuggable_state);
       };

       MyClass myInstanceOfMyClass;

       iris::DebuggableStateSetRequestDelegate delegate =
           iris::DebuggableStateSetRequestDelegate::make<MyClass, &MyClass::setRequestFlag>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateSetRequestDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to set the debuggable state request flag.
     */
    void setDbgStateSetRequestDelegate(DebuggableStateSetRequestDelegate delegate = DebuggableStateSetRequestDelegate())
    {
        inst_dbg_state->setSetRequestDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to set the debuggable state request flag for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setRequestFlag(bool request_debuggable_state);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateSetRequestDelegate<MyClass, &MyClass::setRequestFlag>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a debuggable state request set delegate method.
     * @tparam  METHOD      A method of class T which is a debuggable state request set delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(bool)>
    void setDbgStateSetRequestDelegate(T* instance)
    {
        setDbgStateSetRequestDelegate(DebuggableStateSetRequestDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to set the debuggable state request flag for this instance.
     *
     * Usage:
     * @code
       iris::IrisErrorCode setRequestFlag(bool request_debuggable_state);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateSetRequestDelegate<&setRequestFlag>();
       @endcode
     * @tparam  FUNC        Global function to call to set the debuggable state request flag.
     */
    template <IrisErrorCode(*FUNC)(bool)>
    void setDbgStateSetRequestDelegate()
    {
        setDbgStateSetRequestDelegate(DebuggableStateSetRequestDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the delegate to get the debuggable state acknowledge flag for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getAcknowledgeFlag(bool &debuggable_state_acknowledge);
       };

       MyClass myInstanceOfMyClass;

       iris::DebuggableStateGetAcknowledgeDelegate delegate =
           iris::DebuggableStateGetAcknowledgeDelegate::make<MyClass, &MyClass::getAcknowledgeFlag>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateGetAcknowledgeDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to get the debuggable state acknowledge flag.
     */
    void setDbgStateGetAcknowledgeDelegate(DebuggableStateGetAcknowledgeDelegate delegate = DebuggableStateGetAcknowledgeDelegate())
    {
        inst_dbg_state->setGetAcknowledgeDelegate(delegate);
    }

    /*!
     * @brief Set the delegate to get the debuggable state acknowledge flag for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode getAcknowledgeFlag(bool &debuggable_state_acknowledge);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateGetAcknowledgeDelegate<MyClass, &MyClass::getAcknowledgeFlag>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a debuggable state get acknowledge delegate method.
     * @tparam  METHOD      A method of class T which is a debuggable state get acknowledge delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(bool &)>
    void setDbgStateGetAcknowledgeDelegate(T* instance)
    {
        setDbgStateGetAcknowledgeDelegate(DebuggableStateGetAcknowledgeDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate to get the debuggable state acknowledge flag for this instance.
     *
     * Usage:
     * @code
           iris::IrisErrorCode getAcknowledgeFlag(bool &debuggable_state_acknowledge);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateGetAcknowledgeDelegate<&getAcknowledgeFlag>();
       @endcode
     * @tparam  FUNC        Global function to call to get the debuggable state acknowledge flag.
     */
    template <IrisErrorCode(*FUNC)(bool &)>
    void setDbgStateGetAcknowledgeDelegate()
    {
        setDbgStateGetAcknowledgeDelegate(DebuggableStateGetAcknowledgeDelegate::make<FUNC>());
    }

    /*!
     * @brief Set both the debuggable state delegates.
     *
     * Usage:
       @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setRequestFlag(bool request_debuggable_state);
           iris::IrisErrorCode getAcknowledgeFlag(bool &debuggable_state_acknowledge);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setDbgStateDelegates<MyClass,
                                    &MyClass::setRequest,
                                    &MyClass::getAcknowledgeFlag>(&myInstanceOfMyClass);
       @endcode
     *
     * @tparam  T                   Class which defines both a debuggable state request set and a get
     *                              acknowledge delegate method.
     * @tparam  SET_REQUEST         A method of class T which is a debuggable state request set delegate.
     * @tparam  GET_ACKNOWLEDGE     A method of class T which is a debuggable state get acknowledge delegate.
     * @param   instance            An instance of class T on which @c SET_REQUEST and @c GET_ACKNOWLEDGE
     *                              should be called.
     */
    template <typename T, IrisErrorCode(T::*SET_REQUEST)(bool), IrisErrorCode(T::*GET_ACKNOWLEDGE)(bool &)>
    void setDbgStateDelegates(T* instance)
    {
        setDbgStateSetRequestDelegate<T, SET_REQUEST>(instance);
        setDbgStateGetAcknowledgeDelegate<T, GET_ACKNOWLEDGE>(instance);
    }

    /*!
     * @}
     */

    /*!
     * @brief semihosting_apis IrisInstanceBuilder semihosting APIs
     * @{
     */

    /*!
     * @brief Manage semihosting functionality
     */
    class SemihostingManager
    {
    private:
        IrisInstanceSemihosting *inst_semihost;

    public:
        SemihostingManager(IrisInstanceSemihosting *inst_semihost_)
            : inst_semihost(inst_semihost_)
        {
        }

        /*!
         * @brief  Instances which support semihosting extensions should call this to enable the
         *         @c IRIS_SEMIHOSTING_CALL_EXTENSION event.
         */
        void enableExtensions()
        {
            inst_semihost->enableExtensions();
        }

        /*!
         * @brief Read data for a given file descriptor.
         *
         * The exact behaviour of this method depends on the value of the size_hint and flags parameters.
         * If the NONBLOCK flag is set size_hint is informational only and the method will return immediately
         * with whatever data is already buffered (if any). If NONBLOCK is not set, the method will block
         * until at least size_hint bytes of data are available. Iris messages will continue to be processed
         * while this methods blocks.
         *
         * @param   fDes        File descriptor to read from.
         * @param   size_hint   A hint to how much data is expected to be read.
         * @param   flags       A bitwise or of @ref semihost_flags
         * @return              A vector of data that was read.
         */
        std::vector<uint8_t> readData(uint64_t fDes, size_t size_hint=0, uint64_t flags=semihost::DEFAULT)
        {
            return inst_semihost->readData(fDes, size_hint, flags);
        }

        /*
         * @brief Write data for a given file descriptor
         *
         * @param   fDes        File descriptor to read from.
         * @param   data        Buffer containing the data to write.
         * @param   size        Size of the data buffer in bytes.
         */
        void writeData(uint64_t fDes, const uint8_t *data, size_t size)
        {
            inst_semihost->writeData(fDes, data, size);
        }

        /*
         * @brief Write data for a given file descriptor
         *
         * @param   fDes        File descriptor to read from.
         * @param   data        Buffer containing the data to write.
         */
        void writeData(uint64_t fDes, const std::vector<uint8_t> &data)
        {
            writeData(fDes, &data.front(), data.size());
        }

        /*!
         * @brief Allow a client to perform a semihosting extension defined by @a operation and @a parameter.
         *
         * This might implement a user defined operation or override the default implementation for
         * a predefined operation.
         *
         * @param   operation   A number indicating the operation to perform. This is defined by the semihosting
         *                      standard for standard operations or by the client for user defined operations.
         * @param   parameter   A parameter to the operation. This meaning of this parameter is defined by the
         *                      operation.
         * @return              A pair of (bool success, uint64_t result).
         *                      If status is true, a client performed the function and returned the value in result.
         *                      If status is false, no client performed the function and result is 0.
         */
        std::pair<bool, uint64_t> semihostedCall(uint64_t operation, uint64_t parameter)
        {
            return inst_semihost->semihostedCall(operation, parameter);
        }
    };


    /*!
     * @brief Enable semihosting functionality for this instance and get a manager object to
     *        make use of it.
     *
     * @return          A SemihostingManager object to manage semihosting functionality for
     *                  this instance.
     */
    SemihostingManager enableSemihostingAndGetManager()
    {
        inst_semihost.init();
        return SemihostingManager(inst_semihost);
    }


    /*!
     * @}
     */

    /*!
     * @brief sync_level_apis IrisInstanceBuilder sync-level APIs
     * @{
     */

    /*!
     * @brief Set the delegate that will be called when the sync level is changed for this instance.
     *
     * Passing an empty delegate (the default argument) restores the default implementation which
     * always returns E_not_implemented for all requests.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setSynclevel(uint64_t new_level);
       };

       MyClass myInstanceOfMyClass;

       iris::SyncLevelSetDelegate delegate =
           iris::SyncLevelSetDelegate::make<MyClass, &MyClass::setSynclevel>(&myInstanceOfMyClass);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setSyncLevelSetDelegate(delegate);
       @endcode
     * @param   delegate    Delegate object to call to set sync level.
     */
    void setSyncLevelSetDelegate(SyncLevelSetDelegate delegate = SyncLevelSetDelegate())
    {
        inst_sync_level->setSyncLevelSetDelegate(delegate);
    }

    /*!
     * @brief Set the delegate that will be called when the sync level is changed for this instance.
     *
     * Usage:
     * @code
       class MyClass
       {
           ...
           iris::IrisErrorCode setSynclevel(uint64_t new_level);
       };

       MyClass myInstanceOfMyClass;

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setSyncLevelSetDelegate<MyClass, &MyClass::setSynclevel>(&myInstanceOfMyClass);
       @endcode
     * @tparam  T           Class which defines a set sync level delegate method.
     * @tparam  METHOD      A method of class T which is a set sync level delegate.
     * @param   instance    An instance of class T on which METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t)>
    void setSyncLevelSetDelegate(T* instance)
    {
        setSyncLevelSetDelegate(SyncLevelSetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the delegate that will be called when the sync level is changed for this instance.
     *
     * Usage:
     * @code
           iris::IrisErrorCode setSynclevel(uint64_t new_level);

       iris::IrisInstanceBuilder *builder = myIrisInstance.getBuilder();
       builder->setSyncLevelSetDelegate<&setSynclevel>();
       @endcode
     * @tparam  FUNC        Global function to call to set sync level.
     */
    template <IrisErrorCode(*FUNC)(uint64_t)>
    void setSyncLevelSetDelegate()
    {
        setSyncLevelSetDelegate(SyncLevelSetDelegate::make<FUNC>());
    }
    /*!
     * @}
     */


};

inline IrisInstanceBuilder::TableColumnBuilder IrisInstanceBuilder::TableBuilder::addColumn(const std::string &name)
{
    // Add a new column with default info
    info.tableInfo.columns.resize(info.tableInfo.columns.size() + 1);
    TableColumnInfo &col = info.tableInfo.columns.back();

    col.name = name;

    return TableColumnBuilder(*this, col);
}

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceBuilder_h
