/*!
 *  \file   IrisInstanceDebuggableState.h
 *  \brief  IrisInstance add-on to implement debuggableState functions.
 *  \date   Copyright ARM Limited 2017 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceDebuggableState_h
#define ARM_INCLUDE_IrisInstanceDebuggableState_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisDelegate.h"

NAMESPACE_IRIS_START

/*!
 * @brief Delegate to set the debuggable-state-request flag
 * @code
       IrisErrorCode setRequest(bool request);
   @endcode
 */
typedef IrisDelegate<bool> DebuggableStateSetRequestDelegate;


/*!
 * @brief Interface to stop the simulation time progress.
 * @code
       IrisErrorCode getAcknowledge(bool &acknowledge_out);
   @endcode
 */
typedef IrisDelegate<bool &> DebuggableStateGetAcknowledgeDelegate;

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @brief Debuggable-state add-on for IrisInstance
 */
class IrisInstanceDebuggableState
{
private:
    /// Associated IrisInstance
    IrisInstance *iris_instance;

    /// setRequest/getAcknowledge delegates
    DebuggableStateSetRequestDelegate setRequest;
    DebuggableStateGetAcknowledgeDelegate getAcknowledge;

public:
    IrisInstanceDebuggableState(IrisInstance *iris_instance = NULL);

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set the set request flag delegate.
     *
     *
     * @param   delegate    Delegate that will be called to set or clear the
     *                      debuggable-state request flag.
     */
    void setSetRequestDelegate(DebuggableStateSetRequestDelegate delegate)
    {
        setRequest = delegate;
    }

    /*!
     * @brief Set the get acknowledge flag delegate.
     *
     *
     * @param   delegate    Delegate that will be called to get the debuggable-state
     *                      acknowledge flag.
     */
    void setGetAcknowledgeDelegate(DebuggableStateGetAcknowledgeDelegate delegate)
    {
        getAcknowledge = delegate;
    }

private:
    /// Implementation of debuggableState_setRequest()
    void impl_debuggableState_setRequest(IrisReceivedRequest &request);

    /// Implementation of debuggableState_getAcknowledge()
    void impl_debuggableState_getAcknowledge(IrisReceivedRequest &request);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceSimulationTime_h
