/*!
 *  \file   IrisInstanceDisassembler.h
 *  \brief  Boiler-plate handling disassembly of an IrisInstance. disassembler add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceDisassembler class implements all disassembly related Iris functions.
 */

#ifndef ARM_INCLUDE_IrisInstanceDisassembler_h
#define ARM_INCLUDE_IrisInstanceDisassembler_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

#include <cstdio>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @defgroup disass_delegates Disassembler delegate functions
 * @{
 */

/*!
 * @brief Get the current disassembly mode
 *
 * @code
       IrisErrorCode getCurrentMode(std::string &currentMode)
   @endcode

 * Error: Return E_* error code if failed to get current mode
 */
typedef IrisDelegate<std::string &> GetCurrentDisassemblyModeDelegate;


/*!
 * @brief Get disassembly of a chunk of memory
 *
 * @code
       IrisErrorCode getDisassembly(uint64_t address, const std::string &mode, MemoryReadResult &memoryReadData,
                                    uint64_t count, uint64_t maxAddr, std::vector<DisassemblyLine> &disassemblyLineOut)
   @endcode
 *
 * Error: Return E_* error code if failed to disassemble
 */
typedef IrisDelegate<uint64_t, const std::string &, MemoryReadResult &,
                     uint64_t, uint64_t, std::vector<DisassemblyLine> &> GetDisassemblyDelegate;

/*!
 * @brief Get the disassembly for an individual opcode
 *
 * @code
       IrisErrorCode disassembleOpcode(const std::vector<uint64_t> &opcode, uint64_t address, const std::string &mode,
                                       DisassembleContext &context, DisassemblyLine &disassemblyLineOut)
   @endcode
 *
 * Error: Return E_* error code if failed to disassemble
 */
typedef IrisDelegate<const std::vector<uint64_t> &, uint64_t, const std::string &,
                    DisassembleContext &, DisassemblyLine &> DisassembleOpcodeDelegate;

/*
 * @}
 */

/*!
 * @brief Disassembler add-on for IrisInstance.
 *
 * This is used by instances which want to support disassembly functionality.
 *
 * This class implements all iris disassembler*() functions.
 *
 * Usage (example):
 * @code
    irisInstanceDisassembler = new iris::IrisInstanceDisassembler(irisInstance);
    irisInstanceDisassembler->setDasmCurrentModeGetDelegate(dasmCurrentModeGetDel);  // Get the current disassembly mode
    irisInstanceDisassembler->setDasmDisassemblyGetDelegate(dasmDisassemblyGetDel);  // Get the disassembly of a chunk of memory
    irisInstanceDisassembler->setDasmOpcodeDasmGetDelegate(dasmOpcodeDasmGetDel);    // Disassemble specific opcode
   @endcode
 *
 * See DummyComponent.h for a working example.
 */
class IrisInstanceDisassembler
{
public:

    /*!
     * @brief Contruct an IrisInstanceDisassember.
     *
     * @param   irisInstance    IrisInstance to attach this add-on to.
     */
    IrisInstanceDisassembler(IrisInstance *irisInstance = NULL);

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set the delegate to get the current disassembly mode.
     *
     * @param   delegate    Delegate object that will be called to get the
     *                      current disassembly mode.
     *
     */
    void setGetCurrentModeDelegate(GetCurrentDisassemblyModeDelegate delegate)
    {
        getCurrentMode = delegate;
    }

    /*!
     * @brief Set the delegate to get the disassembly of a chunk of memory.
     *
     * @param   delegate    Delegate object that will be called to get the
     *                      disassembly of a chunk of memory.
     *
     */
    void setGetDisassemblyDelegate(GetDisassemblyDelegate delegate)
    {
        getDisassembly = delegate;
    }

    /*!
     * @brief Set the delegate to get the disassembly of Opcode.
     *
     * @param   delegate    Delegate object that will be called to get the
     *                      disassembly of an opcode.
     *
     */
    void setDisassembleOpcodeDelegate(DisassembleOpcodeDelegate delegate)
    {
        disassembleOpcode = delegate;
    }

    /*!
     * @brief Add a disassembly mode.
     *
     * This should only be called during the initial set up of the instance after
     * which the list of disassembly modes should be static.
     *
     * @param   name            Name of the mode being added.
     * @param   description     Description of the mode being added.
     */
    void addDisassemblyMode(const std::string &name, const std::string &description);

private:

    void impl_disassembler_getModes(IrisReceivedRequest &request);

    void impl_disassembler_getCurrentMode(IrisReceivedRequest &request);

    void impl_disassembler_getDisassembly(IrisReceivedRequest &request);

    void impl_disassembler_disassembleOpcode(IrisReceivedRequest &request);

    /// Check the mode is supported or not, mode may be changed for the current mode
    /// If the mode is supported, isValidMode is true.
    void checkDisassemblyMode(std::string &mode, bool &isValidMode);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Delegate to get the disassembly CurrentMode
    GetCurrentDisassemblyModeDelegate getCurrentMode;

    /// Delegate to get the disassembly of a chunk of memory
    GetDisassemblyDelegate getDisassembly;

    /// Delegate to get the disassembly of Opcode
    DisassembleOpcodeDelegate disassembleOpcode;

    std::vector<DisassemblyMode> disassemblyModes;
    /// Logger.
    IrisLogger log;

};

NAMESPACE_IRIS_END

#endif// #ifndef ARM_INCLUDE_IrisInstanceDisassembler_h


