/*!
 *  \file   IrisInstanceEvent.h
 *  \brief  Boiler-plate handling event of an IrisInstance. Event add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceEvent class offers:
 *  - implements all event related Iris functions.
 *  - maintains and provides event source meta information.
 *  - converts between Iris event functions (event*()) and various C++ access functions
 */

#ifndef ARM_INCLUDE_IrisInstanceEvent_h
#define ARM_INCLUDE_IrisInstanceEvent_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"
#include "iris/detail/IrisRequest.h"

#include <cstdio>
#include <set>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

class EventStream;
class IrisEventRegistry;

/*!
 * @brief Delegate to create an EventStream.
 * @code
       IrisErrorCode create(EventStream *&evStream, const EventSourceInfo &srcInfo, const std::vector<std::string> &fields)
   @endcode
 *
 * Create a new event stream with the specified fields for an event source.
 *
 * The new event stream will be maintained and destroyed in the event add-on.
 *
 * Error: If the event stream could not be created: Return E_* error code
 * E.g. E_unknown_event_field.
 */
typedef IrisDelegate<EventStream *&, const EventSourceInfo&, const std::vector<std::string>&> EventStreamCreateDelegate;

/*!
 * @brief Event add-on for IrisInstance.
 *
 * This is used by instances which want to support event functionality.
 * Generally, there are two kinds of event sources:
 *   (1) Iris specific event sources. These are defined in Iris spec.
 *       E.g. @c IRIS_BREAKPONIT_HIT, @c IRIS_SIMULATION_TIME_EVENT.
 *   (2) Simulation specific event sources. These are not defined in Iris spec. They could be quite
 *       different for different simulations/instances. E.g. @c INST (Every instruction executed),
 *
 * This class implements all iris event*() functions.
 * This class maintains event source information (added by addEventSource() and exposed by
 * event_getEventSources()/event_getEventSource()).
 * This class maintains all event streams. Iris specific event streams are created by this add-on.
 * Simulation specific event streams are created by a delegate, which could be different for
 * different simulations/instances.
 */
class IrisInstanceEvent
{
public:
    /*!
     * @brief Contains the meta-data and delegates for a single EventSource.
     */
    struct EventSourceInfoAndDelegate
    {
        EventSourceInfo info;
        EventStreamCreateDelegate createEventStream;
    };

    /*!
     * @brief Construct an IrisInstanceEvent add-on.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceEvent(IrisInstance *irisInstance = NULL);
    ~IrisInstanceEvent();

    /*!
     * @brief Attach this IrisInstanceEvent add-on to a specific IrisInstance.
     *
     * This should only be used if no instance was attached when this object was constructed.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set the default delegate for creating EventStreams for the attached instance.
     *
     * @param   delegate    A delegate that will be called to create an event stream
     *                      for event sources in the attached instance that do no set
     *                      an event source specific delegate.
     */
    void setDefaultEsCreateDelegate(EventStreamCreateDelegate delegate);

    /*!
     * @brief Add meta-information for an event source.
     *
     * @param   name        The name of the event source.
     * @param   isHidden    If true this event source will be hidden. The EventSourceInfo will not be
     *                      included in the list of events sources returned by event_getEventSources()
     *                      but can still be accessed by event_getEventSource() if the client knows
     *                      the name of the hidden event.
     * @return              A reference to an object which keeps the meta-data and event specific delegates
     *                      (if applicable) for this event. The reference is valid until the next call to
     *                      addEventSource().
     */
    EventSourceInfoAndDelegate& addEventSource(const std::string &name, bool isHidden = false);

    /*!
     * @brief Add meta-information for an event source.
     *
     * @param   info    The meta-data and event specific delegates (if applicable) for a new
     *                  event to add.
     * @return          The evSrcId of the newly added event source.
     */
    uint64_t addEventSource(const EventSourceInfoAndDelegate &info);

private:

    // --- Iris function implementations ---

    void impl_event_getEventSources(IrisReceivedRequest &request);

    void impl_event_getEventSource(IrisReceivedRequest &request);

    void impl_eventStream_create(IrisReceivedRequest &request);

    void impl_eventStream_destroy(IrisReceivedRequest &request);

    void impl_eventStream_enable(IrisReceivedRequest &request);

    void impl_eventStream_disable(IrisReceivedRequest &request);

    void impl_eventStream_getCounter(IrisReceivedRequest &request);

    void impl_eventStream_setTraceRanges(IrisReceivedRequest &request);

    void impl_eventStream_getState(IrisReceivedRequest &request);

    void register_ec_IRIS_INSTANCE_REGISTRY_CHANGED();
    void ec_IRIS_INSTANCE_REGISTRY_CHANGED(IrisReceivedRequest &request);

    /// --- help functions ---

    EventStream *getEventStream(uint64_t esId);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Event source meta-information list.
    std::vector<EventSourceInfoAndDelegate> eventSources;
    std::map<std::string, uint64_t> srcNameToId;

    /// Event stream list.
    /// Contains NULL pointers for invalid event stream ids.
    std::vector<EventStream *> eventStreams;

    /// List of free event stream ids.
    /// This is just here to help us to reuse event stream ids.
    std::vector<uint64_t> freeEsIds;

    /// Delegate for event stream create
    EventStreamCreateDelegate defaultEsCreateDelegate;

    /// Logger.
    IrisLogger log;

    /// In order to automatically unregister an event stream when an instance disconnects
    /// IrisInstanceEvent listens to the global IRIS_INSTANCE_REGISTRY_CHANGED event stream.
    /// The callback for this is registered lazily and when it is registered
    /// instance_registry_changed_registered is set to true.
    bool instance_registry_changed_registered;
};


/*!
 * @brief Base class for event stream.
 *
 * This class is abstract as we don't know how to enable/disable an event for a simulation.
 */
class EventStream
{
public:

    /*!
     * @brief Construct a new event stream.
     */
    EventStream()
        : enabled(false)
        , req(NULL)
        , internal_req(NULL)
        , counter(false)
        , isWaiting(false)
        , selfReleaseAfterWaiting(false)
    {
    }

    virtual ~EventStream() {}

    /*!
     * @brief Trigger the event stream to be released.
     *
     * If this event stream is not waiting for any response, release it immediately.
     * Otherwise, release it when it has finished waiting.
     * Notice: (1) Do not touch anything related to this object after calling this function.
     *         (2) Do not call this function if this object is not created by 'new'.
     */
    void selfRelease()
    {
        // The request to destroy this event stream is nested and processed in the delegate to
        // wait for the response, so it is not multi-threaded and no need to protect the variables.
        if (!isWaiting)
        {
            delete this;
            return;
        }

        // It is waiting for the response of the current request.
        // Cancel the wait and release this object later (after the end of the wait).
        req->cancel();
        selfReleaseAfterWaiting = true;
    }

    /*!
     * @brief Enable this event stream
     *
     * @return      An error code indicating whether the event stream was successfully enabled.
     *              This should be E_ok if it was enabled or E_error_enabling_event_stream if
     *              it could not be enabled.
     */
    virtual IrisErrorCode enable() = 0;

    /*!
     * @brief Disable this event stream
     *
     * @return      An error code indicating whether the event stream was successfully disabled.
     *              This should be E_ok if it was disabled or E_error_enabling_event_stream if
     *              it could not be disabled.
     */
    virtual IrisErrorCode disable() = 0;

    /*!
     * @brief Query the current state of the event.
     *
     * Supported in the derived classes for specific event sources.
     *
     * @param   fields      A map which will be populated with the current values for this
     *                      event's fields.
     * @return              An error code indicating whether the operation was successful.
     */
    virtual IrisErrorCode getState(std::map<std::string, IrisValue> & fields)
    {
        (void)fields;
        return E_not_supported_for_event_source;
    }

    // --- Functions for basic properties ---

    /*!
     * @brief Initialise this event stream.
     *
     * @param   irisInstance    The IrisInstance that is producing this stream. This will be
     *                          used to send event callback requests.
     * @param   srcInfo         The meta-data for the event source generating this stream.
     * @param   ecInstId        The event callback instId: the instance that this stream calls
     *                          when an event fires.
     * @param   ecFunc          The event callback function: the function that is called
     *                          when an event fires.
     * @param   esId            The event stream id for this event stream.
     * @param   syncEc          True if this event stream is synchronous and should send event
     *                          callbacks as requests. If false event callbacks are sent as
     *                          notifications and do not wait for a response.
     */
    void setProperties(IrisInstance *irisInstance, const EventSourceInfo *srcInfo,
                       uint64_t ecInstId, const std::string &ecFunc, uint64_t esId,
                       bool syncEc);

    /*!
     * @brief Is this event stream currently enabled.
     *
     * @return                  @c true if this event stream is enabled or @c false if it disabled.
     */
    bool isEnabled() const
    {
        return enabled;
    }

    /*!
     * @brief Get the event callback instance id for this event stream.
     *
     * @return                  The instId for the instance that this event stream calls when an
     *                          event fires.
     */
    uint64_t getEcInstId() const
    {
        return ecInstId;
    }

    // --- Functions for the counter mode ---

    /*!
     * @brief Set the counter mode and starting value for this event stream.
     *
     * @param   startVal        The starting value of the counter.
     * @param   counterMode     The mode in which this counter operates.
     */
    void setCounter(uint64_t startVal, const EventCounterMode &counterMode);

    /*!
     * @brief Is this event stream a counter.
     *
     * @return                  @c true if this event stream is a counter, otherwise @c false.
     */
    bool isCounter() const
    {
        return counter;
    }

    /*!
     * @brief Get the current value of the counter.
     *
     * @return                  The current value of the event counter.
     */
    uint64_t getCountVal() const
    {
        return curVal;
    }

    // --- Functions for event stream with ranges

    /*!
     * @brief Set trace ranges for this event stream.
     *
     * @param   aspect      The field to check the range of.
     * @param   ranges      A list where each 3 elements form a 3-tuple of (mask, start, end)
     *                      values to configure ranges.
     * @return              An error code indicating whether the ranges could be set successfully.
     */
    IrisErrorCode setRanges(const std::string &aspect, const std::vector<uint64_t> &ranges);

    /*!
     * @brief Check the range for PC.
     *
     * This may be optionally called before generating the callback request (before calling emitEventBegin()).
     *
     * @param   pc      The program counter value to check.
     * @return          @c true if the pc value is in range or no range configured. @c false otherwise.
     */
    bool checkRangePc(uint64_t pc) const;

    // --- Functions to emit the event callback ---
    // Usage (example):
    //     emitEventBegin(time, pc);    // Start to emit the callback.
    //     addField(...);          // Add field value.
    //     addField(...);          // Add field value.
    //     ...
    //     emitEventEnd();             // Emit the callback.

    /*!
     * @brief Start to emit an event callback.
     *
     * @param   req     A request object to use to construct the event callback.
     * @param   time    The time in simulation ticks that the event occurred.
     * @param   pc      The program counter value when the event occurred.
     */
    void emitEventBegin(IrisRequest &req, uint64_t time, uint64_t pc = IRIS_UINT64_MAX);

    /*!
     * @brief Start to emit an event callback.
     *
     * @param   time    The time in simulation ticks that the event occurred.
     * @param   pc      The program counter value when the event occurred.
     */
    void emitEventBegin(uint64_t time, uint64_t pc = IRIS_UINT64_MAX);

    /*!
     * @brief Add uint field value.
     *
     * Fast variant for argument names up to 23 chars. Use this if you can.
     * This will also record the aspect value if the aspect of range check is set.
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addField(const IrisU64StringConstant &field, uint64_t value)
    {
        addFieldRangeHelper(field, value);
    }

    /*!
     * @brief Add sint field value.
     *
     * Fast variant for argument names up to 23 chars. Use this if you can.
     * This will also record the aspect value if the aspect of range check is set.
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addField(const IrisU64StringConstant &field, int64_t value)
    {
        addFieldRangeHelper(field, value);
    }

    /*!
     * @brief Add boolean field value.
     *
     * Fast variant for argument names up to 23 chars. Use this if you can.
     * This will also record the aspect value if the aspect of range check is set.
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addField(const IrisU64StringConstant &field, bool value)
    {
        addFieldRangeHelper(field, value);
    }

    /*!
     * @brief Add field value.
     *
     * This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.
     * Fast variant for argument names up to 23 chars. Use this if you can.
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    template<class T>
    void addField(const IrisU64StringConstant &field, const T &value)
    {
        fieldObj.member(field, value);
    }

    /*!
     * @brief Add uint field value.
     *
     * Slow variant for argument names with more than 23 chars. Do not use unless you have to.
     * This will also record the aspect value if the aspect of range check is set)
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addFieldSlow(const std::string &field, uint64_t value)
    {
        addFieldSlowRangeHelper(field, value);
    }

    /*!
     * @brief Add sint field value.
     *
     * Slow variant for argument names with more than 23 chars. Do not use unless you have to.
     * This will also record the aspect value if the aspect of range check is set)
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addFieldSlow(const std::string &field, int64_t value)
    {
        addFieldSlowRangeHelper(field, value);
    }

    /*!
     * @brief Add boolean field value.
     *
     * Slow variant for argument names with more than 23 chars. Do not use unless you have to.
     * This will also record the aspect value if the aspect of range check is set)
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    void addFieldSlow(const std::string &field, bool value)
    {
        addFieldSlowRangeHelper(field, value);
    }

    /*!
     * @brief Add field value.
     *
     * This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.
     * Slow variant for argument names with more than 23 chars. Do not use unless you have to.
     *
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    template<class T>
    void addFieldSlow(const std::string &field, const T &value)
    {
        fieldObj.memberSlow(field, value);
    }

    /*!
     * @brief Emit the callback.
     *
     * This will also check the ranges and maintain the counter.
     *
     * @param   send        If @c true event callbacks will be sent to the callee immediately.
     *                      If @c false the callback will not be sent immediately allowing the caller
     *                      to delay sending.
     */
    void emitEventEnd(bool send=true);

private:
    /// --- Help functions ---

    /// Increment the counter by one and check the counter mode.
    ///
    /// Return whether the callback should be triggered.
    bool counterTrigger();

    /// Check the ranges and return true if the current aspect value is in ranges.
    bool checkRanges() const;

    /// Helper function to implement addField() functions with range check.
    template<typename T>
    void addFieldRangeHelper(const IrisU64StringConstant &field, T value)
    {
        if (!aspect.empty() && aspect == toString(field))
        {
            aspectFound = true;
            curAspectValue = static_cast<uint64_t>(value);
        }

        fieldObj.member(field, value);
    }

    /// Helper function to implement addFieldSlow() functions with range check.
    template<typename T>
    void addFieldSlowRangeHelper(const std::string &field, T value)
    {
        if (aspect == field)
        {
            aspectFound = true;
            curAspectValue = static_cast<uint64_t>(value);
        }

        fieldObj.memberSlow(field, value);
    }

protected:

    /// --- basic members ---

    /// The iris instance which created this event.
    IrisInstance *irisInstance;

    /// The event source info
    const EventSourceInfo *srcInfo;

    /// Specify target instance which this event gets sent to.
    uint64_t ecInstId;

    /// The event callback function name specified by eventEnable().
    std::string ecFunc;

    /// The event stream id.
    uint64_t esId;

    /// Synchronous callback behavior.
    bool syncEc;

    /// Event is only generated when the event stream is enabled.
    bool enabled;

    /// Generate callback requests.
    IrisRequest *req;
    IrisRequest *internal_req;
    IrisU64JsonWriter::Object fieldObj;

    /// --- members for a counter ---

    /// Is a counter?
    bool counter;

    /// Start value and current value for a counter.
    uint64_t startVal;
    uint64_t curVal;

    /// Specified counter mode.
    EventCounterMode counterMode;

    /// --- members for range ---

    std::string aspect;
    std::vector<uint64_t> ranges;

    /// Found aspect in one of the fields.
    bool aspectFound;

    /// The current aspect value.
    uint64_t curAspectValue;

private:

    /// Is it waiting for the response of a synchronous event callback (Only a synchronous
    /// callback needs to wait for the response)
    bool isWaiting;

    /// Release this object after the end of waiting for response.
    bool selfReleaseAfterWaiting;
};

/*!
 * @brief Event stream class for Iris specific events.
 */
class IrisEventStream : public EventStream
{
public:
    IrisEventStream(IrisEventRegistry *registry_);

    virtual IrisErrorCode enable() IRIS_OVERRIDE;

    virtual IrisErrorCode disable() IRIS_OVERRIDE;

private:
    IrisEventRegistry *registry;
};

/*!
 * @brief Class to register iris event streams for an event.
 */
class IrisEventRegistry
{
public:
    /*!
     * @brief Return true if no event streams are registered.
     *
     * @return      @c true if no event streams are registered.
     */
    bool empty() const
    {
        return esSet.empty();
    }

    /*!
     * @brief Register an event stream.
     *
     * @param   evStream        The stream to be registered.
     * @return                  @c true if the stream was registered successfully.
     */
    bool registerEventStream(EventStream *evStream);

    /*!
     * @brief Unregister an event stream.
     *
     * @param   evStream        The stream to be registered.
     * @return                  @c true if the stream was unregistered successfully.
     */
    bool unregisterEventStream(EventStream *evStream);

    // --- Functions to emit the callback of all registered event streams ---
    // Usage (example):
    //     emitEventBegin(time, pc);    // Start to emit the callback.
    //     addField(...);          // Add field value.
    //     addField(...);          // Add field value.
    //     ...
    //     emitEventEnd();             // Emit the callback.

    void emitEventBegin(uint64_t time, uint64_t pc = IRIS_UINT64_MAX) const;

    /*!
     * @brief Add field value.
     *
     * This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.
     * Fast variant for argument names up to 23 chars. Use this if you can.
     *
     * @tparam  T           The type of @c value.
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    template<class T>
    void addField(const IrisU64StringConstant &field, const T &value) const
    {
        for (std::set<EventStream *>::const_iterator i = esSet.begin(), e = esSet.end(); i != e; i++)
            (*i)->addField(field, value);
    }

    /*!
     * @brief Add field value.
     *
     * This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.
     * Slow variant for argument names with more than 23 chars. Do not use unless you have to.
     *
     * @tparam  T           The type of @c value.
     * @param   field       The name of the field to set the value for.
     * @param   value       The value of the field.
     */
    template<class T>
    void addFieldSlow(const std::string &field, const T &value) const
    {
        for (std::set<EventStream *>::const_iterator i = esSet.begin(), e = esSet.end(); i != e; i++)
            (*i)->addFieldSlow(field, value);
    }

    /*!
     * @brief Emit the callback.
     *
     * This will also check the ranges and maintain the counter.
     */
    void emitEventEnd() const;

    typedef std::set<EventStream *>::const_iterator iterator;

    /*!
     * @brief Get an iterator to the beginning of the event stream set.
     *
     * @see end
     *
     * @return      An iterator to the beginning of the event stream set.
     */
    iterator begin() const
    {
        return esSet.begin();
    }

    /*!
     * @brief Get an iterator to the end of the event stream set.
     *
     * @see begin
     *
     * @return      An iterator to the end of the event stream set.
     */
    iterator end() const
    {
        return esSet.end();
    }

    ~IrisEventRegistry()
    {
        // Disable any remaining event streams.
        // Calling disable() on an EventStream will cause esSet to be modified so we need to loop without
        // using iterators which become invalidated.
        while (!esSet.empty())
        {
            (*esSet.begin())->disable();
        }
    }

private:
    // All registered event streams
    std::set<EventStream *> esSet;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceBreakpoint_h
