/*!
 * \file    IrisInstanceFactoryBuilder.h
 * \brief   A helper class to build instantiation parameter meta-data.
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisInstanceFactoryBuilder_h
#define ARM_INCLUDE_IrisInstanceFactoryBuilder_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/IrisParameterBuilder.h"

#include <vector>
#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief A builder class to construct instantiation parameter meta-data.
 */
class IrisInstanceFactoryBuilder
{
private:
    /// Vector of ResourceInfo for the parameters
    std::vector<ResourceInfo> parameters;

    /// Vector of ResourceInfo for hidden parameters
    std::vector<ResourceInfo> hidden_parameters;

    /// Prefix to add to all parameters
    std::string parameter_prefix;

    ResourceInfo &addParameterInternal(const std::string &name, uint64_t bitWidth, const std::string &descr,
                                       const std::string &type, bool hidden)
    {
        std::vector<ResourceInfo> &param_list = hidden? hidden_parameters : parameters;
        param_list.resize(parameters.size() + 1);
        ResourceInfo &info = param_list.back();

        info.name = name;
        info.bitWidth = bitWidth;
        info.descr = descr;
        info.type = type;

        return info;
    }

public:

    /*!
     * @brief   Construct an IrisInstanceFactoryBuilder
     *
     * @param   prefix      All parameters added to this builder will be prefixed with this string.
     */
    IrisInstanceFactoryBuilder(const std::string &prefix)
        : parameter_prefix(prefix)
    {
    }

    /*!
     * @brief Add a new numeric parameter
     * @param   name        Name of the parameter.
     * @param   bitWidth    Width of the parameter in bits.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addParameter(const std::string &name, uint64_t bitWidth, const std::string &descr)
    {
        return IrisParameterBuilder(addParameterInternal(parameter_prefix + name, bitWidth, descr, "" /*numeric*/, false));
    }

    /*!
     * @brief Add a new hidden numeric parameter
     * @param   name        Name of the parameter.
     * @param   bitWidth    Width of the parameter in bits.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addHidenParameter(const std::string &name, uint64_t bitWidth, const std::string &descr)
    {
        return IrisParameterBuilder(addParameterInternal(parameter_prefix + name, bitWidth, descr, "" /*numeric*/, true));
    }

    /*!
     * @brief Add a new string parameter
     * @param   name        Name of the parameter.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addStringParameter(const std::string &name, const std::string &descr)
    {
        return IrisParameterBuilder(addParameterInternal(parameter_prefix + name, 0, descr, "string", false));
    }

    /*!
     * @brief Add a new hidden string parameter
     * @param   name        Name of the parameter.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addHiddenStringParameter(const std::string &name, const std::string &descr)
    {
        return IrisParameterBuilder(addParameterInternal(parameter_prefix + name, 0, descr, "string", true));
    }

    /*!
     * @brief Add a new boolean parameter
     *
     * Boolean parameters are numeric parameters with a bitWidth of 1 and "true" and "false" enum symbols.
     *
     * @param   name        Name of the parameter.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addBooleanParameter(const std::string &name, const std::string &descr)
    {
        ResourceInfo &info = addParameterInternal(parameter_prefix + name, 1, descr, "numeric", false);

        // Be explicit about the range even though there are only two possible values anyway.
        info.parameterInfo.min.push_back(0);
        info.parameterInfo.max.push_back(1);

        // Add enum strings for the values
        info.enums.push_back(EnumElementInfo(IrisValue(0), "false", ""));
        info.enums.push_back(EnumElementInfo(IrisValue(1), "true", ""));

        return IrisParameterBuilder(info);
    }

    /*!
     * @brief Add a new hidden boolean parameter
     *
     * Boolean parameters are numeric parameters with a bitWidth of 1 and "true" and "false" enum symbols.
     *
     * @param   name        Name of the parameter.
     * @param   descr       Description of the parameter.
     * @return              An IrisParameterBuilder object which can be used to set further meta-data for
     *                      this parameter. The object is valid until another parameter is added.
     */
    IrisParameterBuilder addHiddenBooleanParameter(const std::string &name, const std::string &descr)
    {
        ResourceInfo &info = addParameterInternal(parameter_prefix + name, 1, descr, "numeric", true);

        // Be explicit about the range even though there are only two possible values anyway.
        info.parameterInfo.min.push_back(0);
        info.parameterInfo.max.push_back(1);

        // Add enum strings for the values
        info.enums.push_back(EnumElementInfo(IrisValue(0), "false", ""));
        info.enums.push_back(EnumElementInfo(IrisValue(1), "true", ""));

        return IrisParameterBuilder(info);
    }

    /*!
     * @brief Get all ResourceInfo for non-hidden parameters.
     *
     * @return  A vector of ResourceInfo. Iterators for this vector are invalidated if
     *          a new non-hidden parameter is added.
     */
    const std::vector<ResourceInfo> &getParameterInfo() const
    {
        return parameters;
    }

    /*!
     * @brief Get all ResourceInfo for hidden parameters.
     *
     * @return  A vector of ResourceInfo. Iterators for this vector are invalidated if
     *          a new hidden parameter is added.
     */
    const std::vector<ResourceInfo> &getHiddenParameterInfo() const
    {
        return hidden_parameters;
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceFactoryBuilder_h
