/*!
 *  \file   IrisInstanceImage.h
 *  \brief  Boiler-plate handling image loading of an IrisInstance:
 *              Image loading add-on to IrisInstance.
 *              Image loading callback add-on to the caller.
 *
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceImage class offers:
 *  - implements all image loading related Iris functions.
 *  - maintains and provides image meta-information (path, instanceSideFile, loadedRaw).
 *  - converts between Iris image loading functions (imageLoad*()) and various C++ access functions
 */

#ifndef ARM_INCLUDE_IrisInstanceImage_h
#define ARM_INCLUDE_IrisInstanceImage_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

#include <cstdio>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @brief Delegate function to load an image from the given file.
 *
 * The path may be absolute or relative to the current working directory.
 * @code
       IrisErrorCode loadImage(const std::string &path)
   @endcode
 *
 * Typical implementations try to load the file with supported formats.
 *
 * Errors: If the file specified by path could not be opened, E_error_opening_file is returned
 *         If the file could be opened but could not be read, E_io_error is returned
 *         If the image format is unknown, E_unknown_image_format is returned.
 *         If the image format is known but the image could not be loaded, E_image_format_error is returned
 */
typedef IrisDelegate<const std::string &> ImageLoadFileDelegate;

/*!
 * @brief Delegate to load an image from the given data.
 *
 * Bytes stored in little endian format.
 *
 * @code
       IrisErrorCode loadImage(const std::vector<uint64_t> &data, uint64_t dataSize)
   @endcode
 *
 * Typical implementations try to load the data with supported formats.
 *
 * Errors: If the image format is unknown, E_unknown_image_format is returned.
 *         If the image format is known but the image could not be loaded, E_image_format_error is returned
 */
typedef IrisDelegate<const std::vector<uint64_t> &, uint64_t> ImageLoadDataDelegate;

/*!
 * @brief Image loading add-on for IrisInstance.
 *
 * This is used by instances which support image loading.
 * This is also used by instances which want to make use of image_loadDataPull() to implement the
 * image_loadDataRead() callback.
 *
 * This class implements iris image*() functions.
 * This class maintains/implements two main things:
 * - Functions to load image.
 *   Could load image from file (by image_loadFile()) or data (by image_loadData()/image_loadDataPull()).
 *   Could load image according to the format or load raw data into memory (by specifying rawAddr and rawSpaceId).
 * - Image meta-information (exposed by image_getMetaInfoList() or cleared by image_clearMetaInfoList()).
 *
 * See DummyComponent.h for a working example.
 */
class IrisInstanceImage
{

public:
    /*!
     * @brief Construct a new IrisInstanceImage
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceImage(IrisInstance *irisInstance = 0);

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set image loading from file delegate.
     *
     * @param   delegate    The delegate that will be called to load an image from a file.
     */
    void setLoadImageFileDelegate(ImageLoadFileDelegate delegate);

    /*!
     * @brief Set image loading from (pushed/pulled) data delegate.
     *
     * @param   delegate    The delegate that will be called to load an image from a data buffer.
     */
    void setLoadImageDataDelegate(ImageLoadDataDelegate delegate);

    /*!
     * @brief Read file data into an uint64_t array and record the number of bytes read.
     *
     * @param   fileName    Name of the file to read.
     * @param   data        A reference to a vector which will be populated with the file contents.
     * @param   count       A reference to a variable which is be set to the number of bytes that were read.
     * @return              An error code indicating success of failure.
     */
    static IrisErrorCode readFileData(const std::string &fileName, std::vector<uint64_t> &data, uint64_t &count);

private:

    /// Load image from data, which is either from image_loadData() or image_loadDataPull()
    void loadImageFromData(IrisReceivedRequest &request, const ImageReadResult &imageData);

    /// --- Iris function implementations ---

    /// Image loading from file
    void impl_image_loadFile(IrisReceivedRequest &request);

    /// Image loading from data
    void impl_image_loadData(IrisReceivedRequest &request);

    /// Image loading by pulling data from the caller
    void impl_image_loadDataPull(IrisReceivedRequest &request);

    void impl_image_getMetaInfoList(IrisReceivedRequest &request);

    void impl_image_clearMetaInfoList(IrisReceivedRequest &request);

    /// Write raw data to a specific memory id and address.
    void writeRawDataToMemory(IrisReceivedRequest &request, const std::vector<uint64_t> &data, uint64_t count, uint64_t rawAddr, uint64_t rawSpaceId);

    /// Implementation of pull data by image_loadDataRead() from the caller.
    IrisErrorCode pullData(IrisInstId callerId, uint64_t tag, ImageReadResult &result);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// All image filenames and other meta-info as loaded by the client or loaded by the target
    /// since the list of meta-informations was last reset using image_clearMetaInfoList().
    /// Note: This only record files that have been loaded successfully.
    typedef std::vector<ImageMetaInfo> ImageMetaInfoList;
    ImageMetaInfoList metaInfos;

    /// Logger.
    IrisLogger log;

    ImageLoadFileDelegate loadFileDelegate;
    ImageLoadDataDelegate loadDataDelegate;
};


/*!
 * @brief Image loading add-on for IrisInstance clients implementing image_loadDataRead().
 *
 * This is used by instances which call the instances supporting image_loadDataPull().
 *
 * This class maintains/implements:
 * - Iris image_loadDataRead() function.
 * - Image opening, data reading.
 * - Tags of images.
 */
class IrisInstanceImage_Callback
{
public:

    /*!
     * @brief Construct an IrisInstanceImage_Callback add-on.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceImage_Callback(IrisInstance *irisInstance = 0);

    ~IrisInstanceImage_Callback();

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Open an image for read.
     *
     * @param   fileName    File name of the image file to read.
     * @return              An opaque tag number that is passed to image_loadDataRead()
     *                      to identify the file to read from.
     *                      This will return @c iris::IRIS_UINT64_MAX on failure to
     *                      open the image.
     */
    uint64_t openImage(const std::string &fileName);

protected:

    /// Implementation of Iris function image_loadDataRead()
    void impl_image_loadDataRead(IrisReceivedRequest &request);

private:

    /// Read data from an image identified by a specific tag.
    IrisErrorCode readImageData(uint64_t tag, uint64_t position, uint64_t size, bool end, ImageReadResult &result);

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Logger.
    IrisLogger log;

    /// The list of all opened images. An element is set to NULL, if that image is closed.
    typedef std::vector<FILE *> ImageList;
    ImageList images;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceImage_h
