/*!
 *  \file   IrisInstanceMemory.h
 *  \brief  Boiler-plate handling memory of an IrisInstance. Memory add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  The IrisInstanceMemory class offers:
 *  - implements all memory related Iris functions
 *  - feeds memory.* properties (memory related properties) to instance_getProperties() of the associated IrisInstance
 *  - infrastructure which is useful for Iris clients
 *  - maintains and provides memory meta-information (memory spaces, address translations, sideband information)
 *  - converts between Iris memory access functions (memory_read()) and various C++ access functions
 */

#ifndef ARM_INCLUDE_IrisInstanceMemory_h
#define ARM_INCLUDE_IrisInstanceMemory_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @brief Delegate to read memory data.
 *
 * @code
       IrisErrorCode read(const MemorySpaceInfo &spaceInfo, uint64_t address, uint64_t byteWidth,
                          uint64_t count, const AttributeValueMap &attrib, MemoryReadResult &result)
   @endcode
 *
 * spaceInfo, address, byteWidth, count are guaranteed to be valid.
 *
 * Typical implementations inspect the spaceId, address, byteWidth and count to determine
 * which memory elements should be read and then append the read elements to result.
 *
 * Return data is appended to result.data (which is a vector<uint64_t>):
 * Data elements read from ascending addresses, packed into uint64_ts such that the lowest
 * address is in the lowest bits.
 * Elements of byteWidth >= 2 are read with the endianness of the memory space inside each
 * element, but elements are stored with the lowest bits inside each uint64_t (for byteWidth < 8)
 * and with the lowest bits first in sequences of uint64_t (for byteWidth > 8).
 *
 * Error: Return E_* error code for read error. Append the address could not be read to result.error.
 */
typedef IrisDelegate<const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                     const AttributeValueMap &, MemoryReadResult &> MemoryReadDelegate;

/*!
 * @brief Delegate to write memory data.
 * @code
       IrisErrorCode write(const MemorySpaceInfo &spaceInfo, uint64_t address, uint64_t byteWidth,
                           uint64_t count, const AttributeValueMap &attrib, const uint64_t *data, MemoryWriteResult &result)
   @endcode
 *
 * @see MemoryReadDelegate
 * data contains the data elements to be written in the same format as MemoryReadResult.data for reads.
 */
typedef IrisDelegate<const MemorySpaceInfo&, uint64_t, uint64_t, uint64_t,
                     const AttributeValueMap &, const uint64_t *, MemoryWriteResult &> MemoryWriteDelegate;

/*!
 * @brief Delegate to translate an address.
 *
 * @code
       IrisErrorCode translate(uint64_t inSpaceId, uint64_t address,
                               uint64_t outSpaceId, MemoryAddressTranslationResult &result)
   @endcode
 * inSpaceId, address and outSpaceId are guaranteed to be valid.
 *
 * Typical implementations inspect the inSpaceId, outSpaceId to determine how to translate
 * the address.
 *
 * Return addresses are appended to result.address (which is a vector<uint64_t>):
 * If this array is empty then 'address' is not mapped in 'outSpaceId'.
 * If the array contains exactly one element then the mapping is unique.
 * If contains multiple addresses then 'address' is accessible in the same way under all of
 * these addresses in 'outSpaceId'.
 *
 * Error: Return E_* error code for translate error.
 */
typedef IrisDelegate<uint64_t, uint64_t, uint64_t, MemoryAddressTranslationResult &> MemoryAddressTranslateDelegate;


/*!
 * @ Delegate to get memory sideband info.
 *
 * @code
       IrisErrorCode getSidebandInfo(const MemorySpaceInfo &spaceInfo, uint64_t address,
                                     const IrisValueMap &attrib,
                                     const std::vector<std::string> &request,
                                     IrisValueMap &result)
   @endcode
 *
 * Returns sideband information for a range of addresses in a given memory space.
 */
typedef IrisDelegate<const MemorySpaceInfo &, uint64_t, const IrisValueMap &,
                     const std::vector<std::string> &, IrisValueMap &> MemoryGetSidebandInfoDelegate;



/*!
 * @brief Memory add-on for IrisInstance.
 *
 * This is used by instances which want to expose their own memory.
 *
 * This class implements all iris memory*() functions.
 * This class maintains/implements two main things:
 * - Memory space meta-information (exposed by memory_getMemorySpaces()).
 * - Forwarding memory read/write and address translate accesses to functions with a
 *   simple prototype which is easy to implement by components,
 *   hiding a lot of the complexity of memory_read(), memory_write() and memory_translateAddress()
 *
 * Usage (example):
 * @code
   irisInstance = new iris::IrisInstance(irisInterface, instanceName);
    irisInstanceMemory = new iris::IrisInstanceMemory(irisInstance);
   // Use these delegates for read/write for all following memory spaces.
   irisInstanceMemory->setDefaultReadDelegate<DummyComponent, &DummyComponent::readMemory>(this);
   irisInstanceMemory->setDefaultWriteDelegate<DummyComponent, &DummyComponent::writeMemory>(this);
   irisInstanceMemory->addMemorySpace("Memory"); // Add a memory address space.
   @endcode
 * See setDefaultReadDelegate() for an example of read/write delegates.
 *
 * See DummyComponent.h for a working example.
 * @see memory_apis
 */
class IrisInstanceMemory
{
public:

    /*!
     * @brief Entry in 'spaceInfos'.
     *
     * Contains static memory space information and information on how to access the space.
     */
    struct SpaceInfoAndAccess
    {
        MemorySpaceInfo spaceInfo;
        MemoryReadDelegate readDelegate;    // May be empty. In this case defaultReadDelegate is used.
        MemoryWriteDelegate writeDelegate;  // May be empty. In this case defaultWriteDelegate is used.
        MemoryGetSidebandInfoDelegate sidebandDelegate;  // May be empty. In this case sidebandDelegate is used.
    };

    /*!
     * @brief Contains static address translation information
     */
    struct AddressTranslationInfoAndAccess
    {
        AddressTranslationInfoAndAccess(uint64_t inSpaceId, uint64_t outSpaceId, const std::string &description)
            : translationInfo(inSpaceId, outSpaceId, description)
        {
        }

        MemorySupportedAddressTranslationResult translationInfo;
        MemoryAddressTranslateDelegate translateDelegate;
    };

    /*!
     * @brief Construct an IrisInstanceMemory.
     *
     * Optionally attaches to an IrisInstance.
     * @param   irisInstance    The IrisInstance to attach to.
     */
    IrisInstanceMemory(IrisInstance *irisInstance = 0);

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set default read function for all subsequently added memory spaces.
     *
     * @param   delegate    Delegate object which will be called to read memory.
     */
    void setDefaultReadDelegate(MemoryReadDelegate delegate = MemoryReadDelegate())
    {
        memReadDelegate = delegate;
    }

    /*!
     * @brief Set default write function for all subsequently added memory spaces.
     *
     * @param   delegate    Delegate object which will be called to write memory.
     */
    void setDefaultWriteDelegate(MemoryWriteDelegate delegate = MemoryWriteDelegate())
    {
        memWriteDelegate = delegate;
    }

    /*!
     * @brief Add meta-information for one memory space.
     *
     * @param   name    Name of the memory space.
     * @return          A reference to a SpaceInfoAndAccess object for this new memory
     *                  space. This reference is valid until the next time addMemorySpace() is called.
     */
    SpaceInfoAndAccess &addMemorySpace(const std::string &name);

    /*!
     * @brief Add one memory address translation as well as the translate interface.
     *
     * @param   inSpaceId       Memory space id for the input memory space of this translation.
     * @param   outSpaceId      Memory space id for the output memory space of this translation.
     * @param   description     A human readable description of this translation.
     * @return                  A reference to an AddressTranslationInfoAndAccess object for the
     *                          new translation. This reference is valid until the next time
     *                          addAddressTranslation() is called.
     */
    AddressTranslationInfoAndAccess &addAddressTranslation(uint64_t inSpaceId, uint64_t outSpaceId,
                                                           const std::string &description);

    /*!
     * @brief Set the default memory translation delegate.
     *
     * @param   delegate    Delegate object which will be called to translate addresses.
     */
    void setDefaultTranslateDelegate(MemoryAddressTranslateDelegate delegate = MemoryAddressTranslateDelegate())
    {
        translateDelegate = delegate;
    }

    /*!
     * @brief Set the default delegate to retrieve sideband information.
     *
     * @param   delegate    Delegate object which will be called to get sideband information for a memory space.
     */
    void setDefaultGetSidebandInfoDelegate(MemoryGetSidebandInfoDelegate delegate = MemoryGetSidebandInfoDelegate())
    {
        if (delegate.empty())
        {
            delegate = MemoryGetSidebandInfoDelegate::make<IrisInstanceMemory, &IrisInstanceMemory::getDefaultSidebandInfo>(this);
        }

        sidebandDelegate = delegate;
    }

private:
    /// --- Iris function implementations ---
    //@{
    void impl_memory_getMemorySpaces(IrisReceivedRequest &request);

    void impl_memory_read(IrisReceivedRequest &request);

    void impl_memory_write(IrisReceivedRequest &request);

    void impl_memory_translateAddress(IrisReceivedRequest &request);

    void impl_memory_getUsefulAddressTranslations(IrisReceivedRequest &request);

    void impl_memory_getSidebandInfo(IrisReceivedRequest &request);
    //@}

    /// Default sideband info get function always does nothing and returns an empty object.
    IrisErrorCode getDefaultSidebandInfo(const MemorySpaceInfo &spaceInfo, uint64_t address,
                                         const IrisValueMap &attrib,
                                         const std::vector<std::string> &request,
                                         IrisValueMap &result);
    // --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Memory space meta info.
    typedef std::vector<SpaceInfoAndAccess> SpaceInfoList;
    SpaceInfoList spaceInfos;

    /// Supported address translations
    typedef std::vector<AddressTranslationInfoAndAccess> SupportedTranslations;
    SupportedTranslations supportedTranslations;

    /// Default Memory read/write/translate delegates.
    MemoryReadDelegate memReadDelegate;
    MemoryWriteDelegate memWriteDelegate;
    MemoryAddressTranslateDelegate translateDelegate;

    /// Default memory sideband info get delegate.
    /// It is used for all memory spaces which do not have specific sideband delegates assigned
    MemoryGetSidebandInfoDelegate sidebandDelegate;

    /// Logger.
    IrisLogger log;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceMemory_h
