/*!
 *  \file   IrisInstancePerInstanceExecution.h
 *  \brief  Boiler-plate handling per-instance execution control of an IrisInstance.
 *          Per-instance execution control add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  Implement all per-instance execution control related Iris functions.
 */

#ifndef ARM_INCLUDE_IrisInstancePerInstanceExecution_h
#define ARM_INCLUDE_IrisInstancePerInstanceExecution_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

#include <cstdio>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @brief Delegate to set the execution state.
 *
 * Enable or disable the execution of instructions (or processing of work items).
 *
 * @code IrisErrorCode setState(bool enable) @endcode
 *
 * Return E_ok on success, otherwise return the error code.
 */
typedef IrisDelegate<bool> PerInstanceExecutionStateSetDelegate;

/*!
 * @brief Get the execution state.
 *
 * @a enabled should be set to true if the execution is enabled and false otherwise.
 *
 * @code IrisErrorCode getState(bool &enabled) @endcode
 *
 * Return E_ok on success, otherwise return the error code.
 */
typedef IrisDelegate<bool&> PerInstanceExecutionStateGetDelegate;


/*!
 * @brief Per-instance execution control add-on for IrisInstance.
 *
 * This is used by instances which want to support per-instance execution control functionality.
 *
 * This class implements all iris perInstanceExecution*() functions.
 */
class IrisInstancePerInstanceExecution
{
public:
    /*!
     * @brief Construct an IrisInstancePerInstanceExecution add-on.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstancePerInstanceExecution(IrisInstance *irisInstance = NULL);

    /*!
     * @brief Attach this IrisInstancePerInstanceExecution add-on to a specific IrisInstance.
     *
     * This should only be used if no instance was attached when this object was constructed.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set the delegate for setting execution state.
     *
     * @param   delegate    A delegate object which will be called to set execution state
     *                      for the attached instance.
     */
    void setExecutionStateSetDelegate(PerInstanceExecutionStateSetDelegate delegate);

    /*!
     * @brief Set the delegate for getting execution state.
     *
     * @param   delegate    A delegate object which will be called to get current
     *                      execution state for the attached instance.
     */
    void setExecutionStateGetDelegate(PerInstanceExecutionStateGetDelegate delegate);

private:

    void impl_perInstanceExecution_setState(IrisReceivedRequest &request);

    void impl_perInstanceExecution_getState(IrisReceivedRequest &request);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Delegate to set/get execution state.
    PerInstanceExecutionStateSetDelegate execStateSet;
    PerInstanceExecutionStateGetDelegate execStateGet;

    /// Logger.
    IrisLogger log;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstancePerInstanceExecution_h
