/*!
 *  \file   IrisInstanceResource.h
 *  \brief  Boiler-plate handling resources of an IrisInstance. Resource add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  The IrisInstanceResource class offers:
 *  - implements all resource related Iris functions:
 *  - feeds resource.* properties (resource related properties) to instance_getProperties() of the associated IrisInstance
 *  - infrastructure which is useful for Iris clients
 *  - maintains and provides resource meta-information (name, bitwidth)
 *  - converts between Iris resource access functions (resource_read()) and various C++ access functions
 */

#ifndef ARM_INCLUDE_IrisInstanceResource_h
#define ARM_INCLUDE_IrisInstanceResource_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"
#include <cassert>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

struct ResourceWriteValue
{
    const uint64_t *data;
    const std::string *str;
    ResourceWriteValue(): data(NULL), str(NULL) {}
};

/*!
 * @brief Delegate to read resources.
 * @code
       IrisErrorCode read(const ResourceInfo &resourceInfo, ResourceReadResult &result)
   @endcode
 *
 * resourceInfo.rscId is guaranteed to be valid.
 *
 * Typical implementations inspect the rscId, canonicalRn, addressOffset or even name or cname value to
 * determine which resource should be read and then append the read data to result:
 *
 *     Return data (no undefined bits):
 *     Append data to result.data (which is a vector<uint64_t>. Append one uint64_t if resource is <= 64 bits.
 *     Append multiple uint64_t for wider resources, least significant uint64_t first.
 *
 *     Return data with undefined bits:
 *     Same as above, except: In addition append a mask which contains 1 bits for all undefined bits to
 *     result.undefinedBits (same format and length as result.data) and set all undefined bits to 0 in
 *     result.data.
 *
 * Error: If the resource could not be read: Return E_* error code (e.g.
 * E_error_reading_write_only_resource, E_error_reading_resource or E_not_implemented) and leave result untouched.
 */
typedef IrisDelegate<const ResourceInfo &, ResourceReadResult &> ResourceReadDelegate;


/*!
 * @brief Delegate to read resources.
 * @code
       IrisErrorCode write(const ResourceInfo &resourceInfo, const ResourceWriteValue &value)
   @endcode
 *
 * resourceInfo.rscId is guaranteed to be valid.
 *
 * Typical implementations inspect the rscId, canonicalRn, addressOffset or even name or cname value to
 * determine which resource should be written.
 *
 * @c data contains the data for all resources to be written in the same format as ResourceReadResult.data for reads.
 * The number of elements in the data array is resourceInfo.getDataSizeInU64Chunks().
 * (data is only evaluated for string resources.)
 */
typedef IrisDelegate<const ResourceInfo &, const ResourceWriteValue &> ResourceWriteDelegate;


/*!
 * @brief Resource add-on for IrisInstance.
 *
 * This class implements all iris resource*() functions.
 * This class maintains/implements two main things:
 * - Resource meta-information (exposed by resource_getList() and resource_getListOfResourceGroups()).
 * - Forwarding Resource read/write access to functions with a
 *   simple prototype which is easy to implement by components,
 *   hiding a lot of the complexity of resource_read() and resource_write().
 *
 * In most cases an instance should not use IrisInstanceResource directly but should use IrisInstanceBuilder instead.
 */
class IrisInstanceResource
{
public:
    /*!
     * @brief Entry in 'resourceInfos'.
     *
     * Contains static resource information and information on how to access the resource.
     */
    struct ResourceInfoAndAccess
    {
        ResourceInfo resourceInfo;
        ResourceReadDelegate readDelegate;   // May be invalid. In this case defaultReadDelegate is used.
        ResourceWriteDelegate writeDelegate; // May be invalid. In this case defaultWriteDelegate is used.
    };

    /*!
     * @brief Construct an IrisInstanceResource.
     *
     * Optionally attaches to an IrisInstance.
     * @param   irisInstance    The IrisInstance to attach to.
     */
    IrisInstanceResource(IrisInstance *irisInstance = 0);

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Add a new resource.
     *
     * @param   type    The type of the resource. This should be one of:
     *                   - "numeric"
     *                   - "numericFp"
     *                   - "String"
     *                   - "noValue"
     * @param   name    The name of the resource.
     * @param   descr   A human readable description of the resource.
     * @return          A reference to a ResourceInfoAndAccess object for this new resource.
     *                  This reference is valid until the next time addResource() is called.
     */
    ResourceInfoAndAccess &addResource(const std::string &type,
                                       const std::string &name,
                                       const std::string &descr);

    /*!
     * @brief Begin new resource group.
     *
     * This has these effects:
     * - Add a resource group (only if it does not yet exist).
     * - Assign all resources which are added through addResource() calls to this group, starting with
     *   rscId rscIdStart (if possible).
     *
     * @param   name        The name of the resource group.
     * @param   descr       A description of this resource group.
     * @param   rscIdStart  Start counting from this rscId when new resources are added.
     * @param   cname       A C-identifier version of the resource name if different from @a name.
     */
    void beginResourceGroup(const std::string& name, const std::string& descr, uint64_t rscIdStart = 0,
                            const std::string& cname = std::string());

    /*!
     * @brief Set next rscId.
     *
     * Resources which are added following this call get assigned rscIds starting at nextRscId (if possible).
     * It is usually not necessary to call this function. It may however be useful to achieve a certain rscId
     * naming scheme which simplifies the evaluation of the rscId in the read/write function.
     *
     * @param   nextRscId_  Next rscId
     */
    void setNextRscId(uint64_t nextRscId_)
    {
        nextRscId = nextRscId_;
    }

    /*!
     * @brief Set a tag for a specific resource.
     *
     * @param   rscId   Resource Id for the resource which will have this tag set.
     * @param   tag     Name of the boolean tag which will be set to true.
     *
     * @see IrisInstanceBuilder::setTag
     */
    void setTag(uint64_t rscId, const std::string &tag);

    /*!
     * @brief Get the resource info for a resource that was already added.
     *
     * @param   rscId   A resource id for a resource that was already added.
     * @return          A pointer to the ResourceInfoAndAccess object for the requested resource.
     *                  This pointer is valid until the next call to addResource().
     *                  If @a rscId is not a valid id this function will return NULL.
     */
    ResourceInfoAndAccess *getResourceInfo(uint64_t rscId);

protected:
    // --- Iris function implementations ---
    
    void impl_resource_getList(IrisReceivedRequest &request);


    void impl_resource_getListOfResourceGroups(IrisReceivedRequest &request);


    void impl_resource_read(IrisReceivedRequest &request);

    
    void impl_resource_write(IrisReceivedRequest &request);
    
private:

    /// Generate free resource id.
    uint64_t generateFreeRegId()
    {
        // Skip already occupied rscIds.
        while (rscIdToIndex.find(nextRscId) != rscIdToIndex.end())
            nextRscId++;
        
        return nextRscId++;
    }

    // --- State ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;
    
    /// Logger.
    IrisLogger log;
    
    /// Resource meta info.
    typedef std::vector<ResourceInfoAndAccess> ResourceInfoList;
    ResourceInfoList resourceInfos;
        
    /// rscId to index into resourceInfos.
    typedef std::map<uint64_t,size_t> RegIdToIndex;
    RegIdToIndex rscIdToIndex;
        
    /// Group meta info.
    typedef std::vector<ResourceGroupInfo> GroupInfoList;
    GroupInfoList groupInfos;
    
    /// Group name to index into groupInfos.
    typedef std::map<std::string,size_t> GroupNameToIndex;
    GroupNameToIndex groupNameToIndex;
    
    /// AddResource() adds resources automatically to this group.
    ResourceGroupInfo *currentAddGroup;
    
    /// rscId to assigne to the next resource whichg is added through addResource() etc.
    uint64_t nextRscId;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceResource_source
