/*!
 *  \file   IrisInstanceSemihosting.h
 *  \brief  IrisInstance add-on to implement semihosting functionality.
 *  \date   Copyright ARM Limited 2017 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceSemihosting_h
#define ARM_INCLUDE_IrisInstanceSemihosting_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisLogger.h"

#include "iris/IrisInstanceEvent.h"

NAMESPACE_IRIS_START

class IrisInstance;
class IrisInstanceEvent;
class IrisReceivedRequest;
class IrisThreadAbstractionMutex;

namespace semihost
{

/*!
 * @defgroup semihost_flags Semihosting data request flag constants
 * @brief Flags used to define the behaviour of the readData() method.
 * @{
 */

/*!
 * @brief Request that the client send data when the user presses the return key.
 */
static const uint64_t COOKED = (0 << 0);

/*!
 * @brief Request that the client send data to the model as soon as it is available.
 */
static const uint64_t RAW = (1 << 0);

/*!
 * @brief Block until enough data has been provided based on the specified size hint.
 *
 * @remark While the function blocks Iris messages to all instances will continue to be processed.
 */
static const uint64_t BLOCK = (0 << 1);

/*!
 * @brief Do not block if sufficient data is not available.
 */
static const uint64_t NONBLOCK = (1 << 1);

/*!
 * @brief If the buffer is empty emit an event to request input.
 */
static const uint64_t EMIT_EVENT = (0 << 2);

/*!
 * @brief Do not request input. Return only the data currently buffered.
 */
static const uint64_t NO_EVENT = (1 << 2);

/*!
 * @brief The default flags used if none are specified.
 */
static const uint64_t DEFAULT = COOKED | BLOCK | EMIT_EVENT;

/*!
 * @}
 */

} // namespace semihost

class IrisInstanceSemihosting
{
private:
    /// The IrisInstance this helper is attached to.
    IrisInstance *iris_instance;

    /// Corresponding Event helper for our IrisInstance. This is required to manage
    /// the IRIS_SEMIHOSTING_INPUT* events.
    IrisInstanceEvent *inst_event;

    /// Map of evSrcId to event
    std::map<uint64_t, unsigned> evSrcId_map;

    /// Map if Event enum to evSrcId
    std::vector<IrisEventRegistry> event_registries;

    /// Per-file descriptor buffered input.
    std::map<uint64_t, std::vector<uint8_t> > buffered_input_data;

    /// Mutex to guard access to the input buffer.
    IrisThreadAbstractionMutex *buffer_mutex;

    /// Mutex to guard access to extension related state.
    IrisThreadAbstractionMutex *extension_mutex;

    uint64_t extension_retval;

    IrisLogger log;

    enum ExtensionState
    {
        XS_DISABLED,          // Semihosting extensions are not supported
        XS_DORMANT,           // No ongoing semihosting extension call in progress
        XS_WAITING_FOR_REPLY, // Event has been emitted, waiting for a reply for a client
        XS_RETURNED,          // A client instance has called semihosting_return()
        XS_NOT_IMPLEMENTED    // A client instance has called semihosting_notImplemented()
    } extension_state;

public:
    IrisInstanceSemihosting(IrisInstance *iris_instance = NULL, IrisInstanceEvent *inst_event = NULL);

    ~IrisInstanceSemihosting();

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   iris_instance   The instance to attach to.
     */
    void attachTo(IrisInstance *iris_instance);

    /*!
     * @brief Set the corresponding IrisInstanceEvent object to use to manage semihosting events.
     *
     * This must not be called more than once and must be called with a Event add-on that is
     * attached to the same IrisInstance as this semihosting add-on.
     *
     * @param   handler     The event add-on for this Iris instance.
     */
    void setEventHandler(IrisInstanceEvent *handler);

    /*!
     * @brief Read data for a given file descriptor.
     *
     * The exact behaviour of this method depends on the value of the size_hint and flags parameters.
     * If the NONBLOCK flag is set size_hint is informational only and the method will return immediately
     * with whatever data is already buffered (if any). If NONBLOCK is not set, the method will block
     * until at least size_hint bytes of data are available. Iris messages will continue to be processed
     * while this methods blocks.
     *
     * @param   fDes        File descriptor to read from.
     * @param   size_hint   A hint to how much data is expected to be read.
     * @param   flags       A bitwise or of @ref semihost_flags
     * @return              A vector of data that was read.
     */
    std::vector<uint8_t> readData(uint64_t fDes, size_t size_hint=0, uint64_t flags=semihost::DEFAULT);

    /*
     * @brief Write data for a given file descriptor
     *
     * @param   fDes        File descriptor to read from.
     * @param   data        Buffer containing the data to write.
     * @param   size        Size of the data buffer in bytes.
     */
    void writeData(uint64_t fDes, const uint8_t *data, size_t size);

    /*!
     * @brief  Instances which support semihosting extensions should call this to enable the
     *         @c IRIS_SEMIHOSTING_CALL_EXTENSION event.
     */
    void enableExtensions();

    /*!
     * @brief Allow a client to perform a semihosting extension defined by @a operation and @a parameter.
     *
     * This might implement a user defined operation or override the default implementation for
     * a predefined operation.
     *
     * @param   operation   A number indicating the operation to perform. This is defined by the semihosting
     *                      standard for standard operations or by the client for user defined operations.
     * @param   parameter   A parameter to the operation. This meaning of this parameter is defined by the
     *                      operation.
     * @return              A pair of (bool success, uint64_t result).
     *                      If status is true, a client performed the function and returned the value in result.
     *                      If status is false, no client performed the function and result is 0.
     */
    std::pair<bool, uint64_t> semihostedCall(uint64_t operation, uint64_t parameter);

private:
    /// Internal semihosting_provideInputData() implementation.
    void impl_semihosting_provideInputData(IrisReceivedRequest &request);

    /// Internal semihosting_return() implementation.
    void impl_semihosting_return(IrisReceivedRequest &request);

    /// Internal semihosting_notImplemented() implementation.
    void impl_semihosting_notImplemented(IrisReceivedRequest &request);

    /// Delegate for creating event streams
    IrisErrorCode createEventStream(EventStream *&stream_out, const EventSourceInfo &info,
                                    const std::vector<std::string> &requested_fields);

    /// Issue an IRIS_SEMIHOSTING_CALL event
    void notifyCall(uint64_t operation, uint64_t parameter);

    class SemihostingEventStream;

    IrisErrorCode enableEventStream(EventStream *stream, unsigned event_type);
    IrisErrorCode disableEventStream(EventStream *stream, unsigned event_type);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceSemihosting_h
