/*!
 *  \file   IrisInstanceSimulation.h
 *  \brief  IrisInstance add-on to implement simulation_* functions.
 *  \date   Copyright ARM Limited 2017 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceSimulation_h
#define ARM_INCLUDE_IrisInstanceSimulation_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisObjects.h"

#include "iris/IrisInstantiationContext.h"

#include <vector>
#include <string>
#include <map>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;
class IrisThreadAbstractionMutex;
class IrisInstanceEvent;
class IrisEventRegistry;

class EventStream;

/*!
 * @brief Delegate to instantiate the simulation
 *
 * @code
       IrisErrorCode instantiate(InstantiationResult &result_out)
   @endcode
 */
typedef IrisDelegate<InstantiationResult &> SimulationInstantiateDelegate;

/*!
 * @brief Provides context to a reset delegate call.
 */
class IrisSimulationResetContext
{
private:
    /// Set if allowPartialReset=true.
    static const uint64_t ALLOW_PARTIAL = (1 << 0);

    /// Representation for boolean options to simulation_reset().
    uint64_t flags;

    /// Extract an individual flag as a bool.
    bool getFlag(uint64_t mask) const
    {
        return (flags & mask) != 0;
    }

    /// Set an individual flag from a bool.
    void setFlag(uint64_t mask, bool value)
    {
        flags &= ~mask;
        flags |= (value? mask : 0);
    }

public:
    IrisSimulationResetContext()
        : flags(0)
    {
    }

    /*!
     *  @brief Get the allowPartialReset flag.
     *
     *  @return     Returns true if simulation_reset() was called with allowPartialReset=true.
     */
    bool getAllowPartialReset() const
    {
        return getFlag(ALLOW_PARTIAL);
    }

    // Set/clear the allowPartialReset flag.
    void setAllowPartialReset(bool value=true)
    {
        setFlag(ALLOW_PARTIAL, value);
    }
};

/*!
 * @brief Delegate to reset the simulation
 *
 * @code IrisErrorCode reset(const IrisSimulationResetContext &) @endcode
 */
typedef IrisDelegate<const IrisSimulationResetContext &> SimulationResetDelegate;

/*!
 * @brief Delegate to request that the simulation be shut down.
 *
 * @code IrisErrorCode requestShutdown() @endcode
 */
typedef IrisDelegate<> SimulationRequestShutdownDelegate;

/*!
 * @brief Delegate to get list of parameter information
 * @code IrisErrorCode getInstantiationParameterInfo(std::vector<ResourceInfo> &parameters_out) @endcode
 */
typedef IrisDelegate<std::vector<ResourceInfo> &> SimulationGetParameterInfoDelegate;

/*!
 * @brief Delegate to set the value of an instantiation parameter
 * @code IrisErrorCode setInstantiationParameterValue(const InstantiationParameterValue &value) @endcode
 */
typedef IrisDelegate<const InstantiationParameterValue &> SimulationSetParameterValueDelegate;

/*!
 * @brief List of IRIS_SIMULATION_PHASE events
 */
enum IrisSimulationPhase
{
    IRIS_SIM_PHASE_INITIAL_PLUGIN_LOADING_COMPLETE,
    IRIS_SIM_PHASE_INSTANTIATE_ENTER,
    IRIS_SIM_PHASE_INSTANTIATE,
    IRIS_SIM_PHASE_INSTANTIATE_LEAVE,
    IRIS_SIM_PHASE_INIT_ENTER,
    IRIS_SIM_PHASE_INIT,
    IRIS_SIM_PHASE_INIT_LEAVE,
    IRIS_SIM_PHASE_BEFORE_END_OF_ELABORATION,
    IRIS_SIM_PHASE_END_OF_ELABORATION,
    IRIS_SIM_PHASE_INITIAL_RESET_ENTER,
    IRIS_SIM_PHASE_INITIAL_RESET,
    IRIS_SIM_PHASE_INITIAL_RESET_LEAVE,
    IRIS_SIM_PHASE_START_OF_SIMULATION,
    IRIS_SIM_PHASE_RESET_ENTER,
    IRIS_SIM_PHASE_RESET,
    IRIS_SIM_PHASE_RESET_LEAVE,
    IRIS_SIM_PHASE_END_OF_SIMULATION,
    IRIS_SIM_PHASE_TERMINATE_ENTER,
    IRIS_SIM_PHASE_TERMINATE,
    IRIS_SIM_PHASE_TERMINATE_LEAVE
};
static const size_t IrisSimulationPhase_total = 20;

/*!
 * @brief An IrisInstance add-on that adds simulation functions for the SimulationEngine instance.
 */
class IrisInstanceSimulation
{
private:
    /// The IrisInstance this helper is attached to.
    IrisInstance *iris_instance;

    /// The IrisConnectionInterface to use for the IrisInstantiationContext passed to the
    /// instantiate delegate.
    IrisConnectionInterface *connection_interface;

    /// Delegate to instantiate the simulation.
    SimulationInstantiateDelegate instantiate;

    /// Delegate to reset the simulation
    SimulationResetDelegate reset;

    /// Delegate to request that the simulation be shut down.
    SimulationRequestShutdownDelegate requestShutdown;

    /// Delegate to get instantiation parameter info.
    SimulationGetParameterInfoDelegate getParameterInfo;

    /// Delegate to set an instantiation parameter value
    SimulationSetParameterValueDelegate setParameterValue;

    /// Unless parameter_info_cache_state is set to CACHE_DISABLED getParameterInfo() will be called
    /// at most once and its result stored in cached_parameter_info.
    enum
    {
        CACHE_DISABLED, //! Do not cache parameter info. Call getParameterInfo() each time.
        CACHE_EMPTY,    //! Caching is enabled but we haven't called getParameterInfo() yet.
        CACHE_SET       //! Caching is enabled and we have called getParameterInfo().
    } parameter_info_cache_state;

    /// Used to store cached parameter information when caching is enabled.
    std::vector<ResourceInfo> cached_parameter_info;

    /// Guard access to internal data structures
    IrisThreadAbstractionMutex *mutex;

    /// Event registries for the IRIS_SIM_PHASE_* events
    std::vector<IrisEventRegistry *> simulation_phase_event_registries;

    /// Maps event source Id to the simulation phase for the event.
    std::map<uint64_t, IrisSimulationPhase> evSrcId_to_phase;

    IrisLogger log;

    /// Tracks whether we have been notified that the simulation has been sufficiently
    /// initialised to return from waitForInstantiation().
    bool simulation_has_been_initialised;

    /// A list of request ids that are waiting for a response to waitForInstantiation.
    std::vector<uint64_t> requests_waiting_for_instantiation;

public:

    /*!
     * @brief Construct an IrisInstanceSimulation add-on.
     *
     * @param   iris_instance           The IrisInstance to attach this add-on to.
     * @param   connection_interface    The connection interface that will be used when
     *                                  the simulation is instantiated.
     */
    IrisInstanceSimulation(IrisInstance *iris_instance = NULL,
                           IrisConnectionInterface *connection_interface = NULL);
    ~IrisInstanceSimulation();

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   iris_instance   The IrisInstance to attach to.
     */
    void attachTo(IrisInstance *iris_instance);

    /*!
     * @brief Set the IrisConnectionInterface to use instantiation.
     *
     * This will be passed to the instantiate() delegate when the simulation is instantiated.
     */
    void setConnectionInterface(IrisConnectionInterface *connection_interface_)
    {
        connection_interface = connection_interface_;
    }

    /*!
     * @brief Set the instantiate() delegate.
     *
     * @param   delegate    A delegate object which will be called to instantiate the simulation.
     */
    void setInstantiateDelegate(SimulationInstantiateDelegate delegate)
    {
        instantiate = delegate;
    }

    /*!
     * @brief Set the instantiate() delegate.
     *
     * Set the delegate to call a method in class T.
     *
     * @tparam  T           Class which defines an instantiate delegate method.
     * @tparam  METHOD      A method of class @a T which is an instantiate delegate.
     * @param   instance    An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(InstantiationResult &)>
    void setInstantiateDelegate(T *instance)
    {
        setInstantiateDelegate(SimulationInstantiateDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the instantiate() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam   FUNC       A function which is an instantiate delegate.
     */
    template <IrisErrorCode(*FUNC)(InstantiationResult &)>
    void setInstantiateDelegate()
    {
        setInstantiateDelegate(SimulationInstantiateDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the reset() delegate.
     *
     * @param   delegate    A delegate object which will be called to reset the simulation.
     */
    void setResetDelegate(SimulationResetDelegate delegate)
    {
        reset = delegate;
    }

    /*!
     * @brief Set the reset() delegate.
     *
     * Set the delegate to call a method in class T.
     *
     * @tparam  T           Class which defines a reset delegate method.
     * @tparam  METHOD      A method of class @a T which is a reset delegate.
     * @param   instance    An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const IrisSimulationResetContext &)>
    void setResetDelegate(T *instance)
    {
        setResetDelegate(SimulationResetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the reset() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam   FUNC       A function which is a reset delegate.
     */
    template <IrisErrorCode(*FUNC)(const IrisSimulationResetContext &)>
    void setResetDelegate()
    {
        setResetDelegate(SimulationResetDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the requestShutdown() delegate.
     *
     * @param   delegate    A delegate object which will be called to request that the
     *                      simulation be shut down.
     */
    void setRequestShutdownDelegate(SimulationRequestShutdownDelegate delegate)
    {
        requestShutdown = delegate;
    }

    /*!
     * @brief Set the requestShutdown() delegate.
     *
     * Set the delegate to call a method in class T.
     *
     * @tparam  T           Class which defines a requestShutdown delegate method.
     * @tparam  METHOD      A method of class @a T which is a requestShutdown delegate.
     * @param   instance    An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)()>
    void setRequestShutdownDelegate(T *instance)
    {
        setRequestShutdownDelegate(SimulationRequestShutdownDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the requestShutdown() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam   FUNC       A function which is a requestShutdown delegate.
     */
    template <IrisErrorCode(*FUNC)()>
    void setRequestShutdownDelegate()
    {
        setRequestShutdownDelegate(SimulationRequestShutdownDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the getParameterInfo() delegate.
     *
     * @param   delegate        A delegate object which will be called to get instantiation
     *                          parameter info for the simulation.
     * @param   cache_result    If true the delegate will only be called once and the result
     *                          cached and used to subsequent calls to
     *                          @c simulation_getInstantiationParameterInfo().
     *                          If false the result is not cached and the delegate will be called every time.
     */
    void setGetParameterInfoDelegate(SimulationGetParameterInfoDelegate delegate, bool cache_result=true)
    {
        getParameterInfo = delegate;
        parameter_info_cache_state = cache_result? CACHE_EMPTY : CACHE_DISABLED;
        cached_parameter_info.clear();
    }

    /*!
     * @brief Set the getParameterInfo() delegate.
     *
     * Set the delegate to call a method in class T.
     *
     * @tparam  T               Class which defines a getParameterInfo delegate method.
     * @tparam  METHOD          A method of class @a T which is a getParameterInfo delegate.
     * @param   instance        An instance of class @a T on which @a METHOD should be called.
     * @param   cache_result    If true the delegate will only be called once and the result
     *                          cached and used to subsequent calls to
     *                          @c simulation_getInstantiationParameterInfo().
     *                          If false the result is not cached and the delegate will be called every time.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(std::vector<ResourceInfo> &)>
    void setGetParameterInfoDelegate(T *instance, bool cache_result=true)
    {
        typedef SimulationGetParameterInfoDelegate D;
        setGetParameterInfoDelegate(D::make<T, METHOD>(instance), cache_result);
    }

    /*!
     * @brief Set the getParameterInfo() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam  FUNC            A function which is a getParameterInfo delegate.
     * @param   cache_result    If true the delegate will only be called once and the result
     *                          cached and used to subsequent calls to
     *                          @c simulation_getInstantiationParameterInfo().
     *                          If false the result is not cached and the delegate will be called every time.
     */
    template <IrisErrorCode(*FUNC)(std::vector<ResourceInfo> &)>
    void setGetParameterInfoDelegate(bool cache_result=true)
    {
        typedef SimulationGetParameterInfoDelegate D;
        setGetParameterInfoDelegate(D::make<FUNC>(), cache_result);
    }

    /*!
     * @brief Set the setParameterValue() delegate.
     *
     * @param   delegate        A delegate object which will be called to set instantiation
     *                          parameter values before instantiation.
     */
    void setSetParameterValueDelegate(SimulationSetParameterValueDelegate delegate)
    {
        setParameterValue = delegate;
    }

    /*!
     * @brief Set the setParameterValue() delegate.
     *
     * Set the delegate to call a method in class T.
     *
     * @tparam  T               Class which defines a setParameterValue delegate method.
     * @tparam  METHOD          A method of class @a T which is a setParameterValue delegate.
     * @param   instance        An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(const InstantiationParameterValue &)>
    void setSetParameterValueDelegate(T *instance)
    {
        setSetParameterValueDelegate(SimulationSetParameterValueDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the setParameterValue() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam  FUNC            A function which is a setParameterValue delegate.
     */
    template <IrisErrorCode(*FUNC)(const InstantiationParameterValue &)>
    void setSetParameterValueDelegate()
    {
        setSetParameterValueDelegate(SimulationSetParameterValueDelegate::make<FUNC>());
    }

    /*!
     * @brief Move from pre-instantiation to post-instantiation phase.
     *
     * This effects which functions are published.
     * Only call this function if the simulation is instantiated though some means outside of Iris.
     * This object will automatically enter post-instantiation phase when the simulation is successfully
     * instantiated by an Iris call to simulation_instantiate().
     */
    void enterPostInstantiationPhase();

    /*!
     * @brief Set up IRIS_SIM_PHASE* events.
     *
     * @param   handler     An IrisInstanceEvent add-on which is attached to the same instance as this add-on.
     */
    void setEventHandler(IrisInstanceEvent *handler);

    /*!
     * @brief Emit an IRIS_SIM_PHASE* event for the supplied phase.
     *
     * @param   time    The simulation time at which the event occurred.
     * @param   phase   The simulation phase that was reached.
     */
    void notifySimPhase(uint64_t time, IrisSimulationPhase phase);

private:
    /// Internal simulation_getInstantiationParameterInfo() implementation.
    void impl_simulation_getInstantiationParameterInfo(IrisReceivedRequest &request);

    /// Internal simulation_setInstantiationParameterValues() implementation.
    void impl_simulation_setInstantiationParameterValues(IrisReceivedRequest &request);

    /// Internal simulation_instantiate() implementation.
    void impl_simulation_instantiate(IrisReceivedRequest &request);

    /// Internal simulation_reset() implementation.
    void impl_simulation_reset(IrisReceivedRequest &request);

    /// Internal simulation_requestShutdown() implementation.
    void impl_simulation_requestShutdown(IrisReceivedRequest &request);

    /// Internal simulation_waitForInstantiation() implementation.
    void impl_simulation_waitForInstantiation(IrisReceivedRequest &request);

    /// EventStreamCreate delegate to create event streams for the IRIS_SIM_PHASE* events.
    IrisErrorCode createEventStream(EventStream *&event_stream_out, const EventSourceInfo &info,
                                    const std::vector<std::string> &fields);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceSimulation_h

