/*!
 *  \file   IrisInstanceSimulationTime.h
 *  \brief  IrisInstance add-on to implement simulationTime functions.
 *  \date   Copyright ARM Limited 2017 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceSimulationTime_h
#define ARM_INCLUDE_IrisInstanceSimulationTime_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisDelegate.h"

#include <vector>
#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief Delegate to resume the simulation time progress.
 * @code IrisErrorCode run(); @endcode
 */
typedef IrisDelegate<> SimulationTimeRunDelegate;


/*!
 * @brief Delegate to stop the simulation time progress.
 * @code IrisErrorCode stop(); @endcode
 */
typedef IrisDelegate<> SimulationTimeStopDelegate;


/*!
 * @brief Delegate to get the simulation time.
 * @code IrisErrorCode getTime(uint64_t &ticks, uint64_t &tickHz, bool &running); @endcode
 */
typedef IrisDelegate<uint64_t &, uint64_t &, bool &> SimulationTimeGetDelegate;

/*!
 * @brief The reasons why the simulation time is stopped.
 */
enum TIME_EVENT_REASON
{
    TIME_EVENT_UNKNOWN,                //!< Simulation stopped for a different reason.
    TIME_EVENT_STOP,                   //!< simulationTime_stop() was called.
    TIME_EVENT_BREAKPOINT,             //!< Breakpoint was hit.
    TIME_EVENT_TRACE_COUNTER_OVERFLOW  //!< CounterMode.overflowStopSim.
};

class IrisInstance;
class IrisInstanceEvent;
class IrisEventRegistry;
class IrisReceivedRequest;

class EventStream;
struct EventSourceInfo;

/*!
 * @brief Simulation time add-on for IrisInstance
 */
class IrisInstanceSimulationTime
{
private:
    /// Associated IrisInstance
    IrisInstance *iris_instance;

    /// EventRegistry to handle IRIS_SIMULATION_TIME_EVENT callbacks
    IrisEventRegistry *simulation_time_event_registry;

    /// Simulation time run/stop/get delegates.
    SimulationTimeRunDelegate run_delegate;
    SimulationTimeStopDelegate stop_delegate;
    SimulationTimeGetDelegate get_time_delegate;

public:
    /*!
     * @brief Construct an IrisInstanceSimulationTime add-on.
     *
     * @param   iris_instance   An IrisInstance to attach this add-on to.
     * @param   inst_event      An IrisInstanceEvent add-on that is already attached to IrisInstance.
     *                          This is used to set up simulation time events.
     */
    IrisInstanceSimulationTime(IrisInstance *iris_instance = NULL, IrisInstanceEvent *inst_event = NULL);
    ~IrisInstanceSimulationTime();

    /*!
     * @brief Attach this IrisInstance add-on to a specific IrisInstance.
     *
     * @param   irisInstance    An IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set event handler to use to send simulation time related events
     *
     * @param   handler         An IrisInstanceEvent add-on that is already attached to IrisInstance.
     *                          This is used to set up simulation time events.
     */
    void setEventHandler(IrisInstanceEvent *handler);

    /*!
     * @brief Set the run() delegate.
     *
     * @param   delegate    A delegate that will be called to start/resume progress of simulation time.
     */
    void setSimTimeRunDelegate(SimulationTimeRunDelegate delegate)
    {
        run_delegate = delegate;
    }

    /*!
     * @brief Set the run() delegate.
     *
     * @tparam  T               Class which defines a run delegate method.
     * @tparam  METHOD          A method of class @a T which is a run delegate.
     * @param   instance        An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)()>
    void setSimTimeRunDelegate(T* instance)
    {
        setSimTimeRunDelegate(SimulationTimeRunDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the run() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam  FUNC            A function which is a run delegate.
     */
    template <IrisErrorCode(*FUNC)()>
    void setSimTimeRunDelegate()
    {
        setSimTimeRunDelegate(SimulationTimeRunDelegate::make<FUNC>());
    }

    /*!
     * @brief Set the stop() delegate.
     *
     * @param   delegate    A delegate that will be called to stop progress of simulation time.
     */
    void setSimTimeStopDelegate(SimulationTimeStopDelegate delegate)
    {
        stop_delegate = delegate;
    }

    /*!
     * @brief Set the stop() delegate.
     *
     * @tparam  T               Class which defines a stop delegate method.
     * @tparam  METHOD          A method of class @a T which is a stop delegate.
     * @param   instance        An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)()>
    void setSimTimeStopDelegate(T* instance)
    {
        setSimTimeStopDelegate(SimulationTimeStopDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the stop() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam  FUNC            A function which is a stop delegate.
     */
    template <IrisErrorCode(*FUNC)()>
    void setSimTimeStopDelegate()
    {
        setSimTimeStopDelegate(SimulationTimeStopDelegate::make<FUNC>());
    }


    /*!
     * @brief Set the getTime() delegate.
     *
     * @param   delegate    A delegate that will be called to get the current simulation time.
     */
    void setSimTimeGetDelegate(SimulationTimeGetDelegate delegate)
    {
        get_time_delegate = delegate;
    }

    /*!
     * @brief Set the getTime() delegate.
     *
     * @tparam  T               Class which defines a getTime delegate method.
     * @tparam  METHOD          A method of class @a T which is a getTime delegate.
     * @param   instance        An instance of class @a T on which @a METHOD should be called.
     */
    template <typename T, IrisErrorCode(T::*METHOD)(uint64_t &, uint64_t&, bool&)>
    void setSimTimeGetDelegate(T* instance)
    {
        setSimTimeGetDelegate(SimulationTimeGetDelegate::make<T, METHOD>(instance));
    }

    /*!
     * @brief Set the getTime() delegate.
     *
     * Set the delegate to a global function.
     *
     * @tparam  FUNC            A function which is a getTime delegate.
     */
    template <IrisErrorCode(*FUNC)(uint64_t &, uint64_t&, bool&)>
    void setSimTimeGetDelegate()
    {
        setSimTimeGetDelegate(SimulationTimeGetDelegate::make<FUNC>());
    }

    /// Generate the IRIS_SIMULATION_TIME_EVENT event callback.
    void notifySimulationTimeEvent(TIME_EVENT_REASON reason = TIME_EVENT_UNKNOWN);

private:
    /// Run/Stop/Get simulation time.
    void impl_simulationTime_run(IrisReceivedRequest &request);
    void impl_simulationTime_stop(IrisReceivedRequest &request);
    void impl_simulationTime_get(IrisReceivedRequest &request);

    IrisErrorCode createEventStream(EventStream *&, const EventSourceInfo&, const std::vector<std::string>&);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceSimulationTime_h
