/*!
 *  \file   IrisInstanceStep.h
 *  \brief  Boiler-plate handling stepping of an IrisInstance. Step add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceStep class implements all stepping related Iris functions.
 */

#ifndef ARM_INCLUDE_IrisInstanceStep_h
#define ARM_INCLUDE_IrisInstanceStep_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"
#include "iris/detail/IrisLogger.h"

#include <cstdio>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;

/*!
 * @brief Delegate to set the remaining steps measured in the specified unit.
 *
 * @code IrisErrorCode setRemainingSteps(uint64_t steps, const std::string &unit) @endcode
 *
 * Error: Return E_* error code if failed to set the steps
 */
typedef IrisDelegate<uint64_t, const std::string &> RemainingStepSetDelegate;

/*!
 * @brief Delegate to get the value of the current remaining steps.
 *
 *     IrisErrorCode getRemainingSteps(uint64_t &steps, const std::string &unit)
 *
 * Error: Return E_* error code if failed to get the remaining steps
 * Interface for remaining steps get
 */
typedef IrisDelegate<uint64_t &, const std::string &> RemainingStepGetDelegate;

/*!
 * @brief Delegate to get the value of the step count.
 *
 * @code IrisErrorCode getStepCount(uint64_t &count, const std::string &unit) @endcode
 *
 * Error: Return E_* error code if failed to get the step count
 */
typedef IrisDelegate<uint64_t &, const std::string &> StepCountGetDelegate;

/*!
 * @brief Step add-on for IrisInstance.
 *
 * This is used by instances which want to support stepping functionality.
 *
 * This class implements all iris step*() functions.
 */
class IrisInstanceStep
{
public:
    /*!
     * @brief Construct an IrisInstanceStep add-on.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceStep(IrisInstance *irisInstance = NULL);

    /*!
     * @brief Attach this IrisInstanceStep add-on to a specific IrisInstance.
     *
     * This should only be used if no instance was attached when this object was constructed.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Set the delegate for setting remaining steps.
     *
     * @param   delegate    A delegate object which will be called to set the remaining steps
     *                      for the attached instance.
     */
    void setRemainingStepSetDelegate(RemainingStepSetDelegate delegate);

    /*!
     * @brief Set the delegate for getting remaining steps.
     *
     * @param   delegate    A delegate object which will be called to get the remaining steps
     *                      for the attached instance.
     */
    void setRemainingStepGetDelegate(RemainingStepGetDelegate delegate);


    /*!
     * @brief Set the delegate for getting the step count.
     *
     * @param   delegate    A delegate object which will be called to get the step count
     *                      for the attached instance.
     */
    void setStepCountGetDelegate(StepCountGetDelegate delegate);

private:

    void impl_step_setup(IrisReceivedRequest &request);

    void impl_step_getRemainingSteps(IrisReceivedRequest &request);

    void impl_step_getStepCounterValue(IrisReceivedRequest &request);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Delegate to set/get the remaining steps
    RemainingStepSetDelegate stepSetDel;
    RemainingStepGetDelegate stepGetDel;

    /// Delegate to get the step count
    StepCountGetDelegate stepCountGetDel;

    /// Logger.
    IrisLogger log;
};


NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceStep_h
