/*!
 *  \file   IrisInstanceSyncLevel.h
 *  \brief  IrisInstance add-on to implement sync-level functionality.
 *  \date   Copyright ARM Limited 2017 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisInstanceSyncLevel_h
#define ARM_INCLUDE_IrisInstanceSyncLevel_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisLogger.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"

#include "iris/IrisInstanceEvent.h"

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;
class IrisThreadAbstractionMutex;

/*!
 * @brief Delegate to inform the simulation that the sync-level has changed.
 *
 * @code IrisErrorCode setSyncLevel(uint64_t level) @endcode
 */
typedef IrisDelegate<uint64_t> SyncLevelSetDelegate;

/*!
 * @brief Synchronisation level add-on for IrisInstance.
 *
 * This is used by instances which want to support syncLevel functionality.
 */
class IrisInstanceSyncLevel
{
private:
    /// The IrisInstance this helper is attached to.
    IrisInstance *iris_instance;

    /// Current effective sync-level and request counts.
    SyncLevelState sync_level_state;

    /// Guard access to the request_counts table.
    IrisThreadAbstractionMutex *mutex;

    IrisLogger log;

    /// Delegate to tell the simulation that there is a new effective sync-level
    SyncLevelSetDelegate setSyncLevel;

public:
    /*!
     * @brief Construct an IrisInstanceSyncLevel add-on.
     *
     * @param   iris_instance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceSyncLevel(IrisInstance *iris_instance = NULL);

    ~IrisInstanceSyncLevel();

    /*!
     * @brief Attach this IrisInstanceSyncLevel add-on to a specific IrisInstance.
     *
     * This should only be used if no instance was attached when this object was constructed.
     *
     * @param   iris_instance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *iris_instance);

    /*!
     * @brief Set the delegate for setting the syncronisation level.
     *
     * @param   delegate    A delegate object which will be called to set the sync level
     *                      for the attached instance when the effective sync level changes.
     */
    void setSyncLevelSetDelegate(SyncLevelSetDelegate delegate)
    {
        setSyncLevel = delegate;
    }

private:
    /// Internal implementation of both:
    ///     syncLevel_request()
    ///     syncLevel_release()
    void syncLevel_update(IrisReceivedRequest &request, bool is_request);

    /// Internal syncLevel_request() implementation.
    /// Forwards to impl_syncLevel_request_release(..., true).
    void impl_syncLevel_request(IrisReceivedRequest &request);

    /// Internal syncLevel_release() implementation.
    /// Forwards to impl_syncLevel_request_release(..., false).
    void impl_syncLevel_release(IrisReceivedRequest &request);

    /// Internal syncLevel_get() implementation.
    void impl_syncLevel_get(IrisReceivedRequest &request);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstanceSyncLevel_h

