/*!
 *  \file   IrisInstanceTable.h
 *  \brief  Boiler-plate handling table interface of an IrisInstance. Table add-on to IrisInstance.
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 *  The IrisInstanceTable class implements all table related Iris functions.
 */

#ifndef ARM_INCLUDE_IrisInstanceTable_h
#define ARM_INCLUDE_IrisInstanceTable_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisDelegate.h"

NAMESPACE_IRIS_START

class IrisInstance;
class IrisReceivedRequest;


/*!
 * @brief Delegate to read table data.
 *
 * @code
       IrisErrorCode read(const TableInfo &tableInfo, uint64_t index, uint64_t count, TableReadResult &result)
   @endcode
 * tableInfo, index, count are guaranteed to be valid. count is non-zero.
 *
 * Read results are in the TableReadResult.
 * Errors about reading table cell values are in TableReadResult.
 */
typedef IrisDelegate<const TableInfo&, uint64_t, uint64_t, TableReadResult &> TableReadDelegate;

/*!
 * @brief Delegate to write table data.
 *
 * @code
       IrisErrorCode write(const TableInfo &tableInfo, const TableRecords &records, TableWriteResult &result)
   @endcode
 * records is guaranteed not empty.
 *
 * Errors about writing table cell values are in TableWriteResult.
 */
typedef IrisDelegate<const TableInfo&, const TableRecords &, TableWriteResult &> TableWriteDelegate;

/*!
 * @brief Table add-on for IrisInstance.
 *
 * This is used by instances which want to support table functionality.
 */
class IrisInstanceTable
{
public:
    /*!
     * @brief Entry in 'tableInfos'.
     *
     * Contains static table information and information on how to access the table.
     */
    struct TableInfoAndAccess
    {
        TableInfo tableInfo;
        TableReadDelegate readDelegate;    //!< May be empty. In this case defaultReadDelegate is used.
        TableWriteDelegate writeDelegate;  //!< May be empty. In this case defaultWriteDelegate is used.
    };

    /*!
     * @brief Construct an IrisInstanceTable add-on.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    IrisInstanceTable(IrisInstance *irisInstance = NULL);

    /*!
     * @brief Attach this IrisInstanceTable add-on to a specific IrisInstance.
     *
     * This should only be used if no instance was attached when this object was constructed.
     *
     * @param   irisInstance    The IrisInstance to attach this add-on to.
     */
    void attachTo(IrisInstance *irisInstance);

    /*!
     * @brief Add meta-information for one table.
     *
     * @param   name    The name of this table.
     * @return          A reference to a TableInfoAndAccess object that can be used to set meta-data
     *                  and access delegates for this table.
     */
    TableInfoAndAccess &addTableInfo(const std::string &name);

    /*!
     * @brief Set the default delegate for reading table data.
     *
     * @param   delegate    A delegate object which will be called to read table data for
     *                      tables in the attached instance that do not set a table specific delegate.
     */
    void setDefaultReadDelegate(TableReadDelegate delegate = TableReadDelegate())
    {
        defaultReadDelegate = delegate;
    }

    /*!
     * @brief Set the default delegate for writing table data.
     *
     * @param   delegate    A delegate object which will be called to write table data for
     *                      tables in the attached instance that do not set a table specific delegate.
     */
    void setDefaultWriteDelegate(TableWriteDelegate delegate = TableWriteDelegate())
    {
        defaultWriteDelegate = delegate;
    }

private:

    void impl_table_getList(IrisReceivedRequest &request);

    void impl_table_read(IrisReceivedRequest &request);

    void impl_table_write(IrisReceivedRequest &request);

    /// --- state ---

    /// This object is attached to one specific IrisInstance. This is an add-on to an IrisInstance.
    IrisInstance *irisInstance;

    /// Table meta info.
    typedef std::vector<TableInfoAndAccess> TableInfoAndAccessList;
    TableInfoAndAccessList tableInfos;

    /// The default table read/write delegate always return E_not_implemented
    TableReadDelegate defaultReadDelegate;
    TableWriteDelegate defaultWriteDelegate;
};


NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInstanceTable_h
