/*!
 * \file    IrisInstantiationContext.h
 * \brief   Helper class used to instantiate Iris instances from generic factories.
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisInstantiationContext_h
#define ARM_INCLUDE_IrisInstantiationContext_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisUtils.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisUtils.h"

#include <vector>
#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief Provides context when instantiating an Iris instance from a factory.
 */
class IrisInstantiationContext
{
private:
    IrisConnectionInterface *connection_interface;

    /// Result object returned by the instantiate function.
    /// This is used to collect warnings/errors and indicate the overall status.
    InstantiationResult &result;

    /// Map of parameter name to parameter value.
    /// This should contain all parameters
    IrisValueMap params;

    /// Instance prefix. One of:
    ///     "component"
    ///     "client"
    ///     "framework"
    std::string prefix;

    /// The portion of the instance name after the prefix.
    std::string component_name;

    uint64_t instance_flags;

    /// InstantiationContexts for sub-components of this context which are owned by this context.
    std::vector<IrisInstantiationContext *> children;

    void errorInternal(const std::string &severity,
                       const std::string &code,
                       const std::string &parameterName,
                       const char *format,
                       va_list args);

    /// Validate parameter values and populate the params map, setting warnings and errors
    /// as appropriate as it goes.
    void processParameters(const std::vector<ResourceInfo> &param_info_,
                           const std::vector<InstantiationParameterValue> &param_values_);


    /// Private constructor used to create child contexts
    IrisInstantiationContext(const IrisInstantiationContext *parent, const std::string &instance_name);

public:
    IrisInstantiationContext(IrisConnectionInterface *connection_interface_,
                             InstantiationResult &result_,
                             const std::vector<ResourceInfo> &param_info_,
                             const std::vector<InstantiationParameterValue> &param_values_,
                             const std::string &prefix_,
                             const std::string &component_name_,
                             uint64_t instance_flags_);

    ~IrisInstantiationContext();

    /*!
     * @brief Get an IrisInstanceContext pointer for a subcomponent instance.
     *
     * For example the you might call getSubcomponentContext("cpu0") on the context
     * for instance "component.cluster0" to get the context to instantiate "component.cluster0.cpu0".
     * The object pointed to by the return value is owned by its parent context and has the same
     * lifetime as the parent context.
     *
     * @param   child_name      The name of a child instance.
     * @return                  A pointer to an IrisInstantiationContext object for the named child.
     */
    IrisInstantiationContext *getSubcomponentContext(const std::string &child_name);

    /*!
     * @brief Get the value of an instantiation parameter.
     *
     * @tparam  T       The type of the @a value. This must be a type which is appropriate
     *                  to receive the value for this parameter.
     * @param   name    The name of the parameter.
     * @param   value   A reference to a value to type @a T which will receive the value
     *                  of the named parameter.
     */
    template <typename T>
    void getParameter(const std::string &name, T &value)
    {
        IrisValueMap::const_iterator it = params.find(name);
        if (it != params.end())
        {
            it->second.get(value);
        }
        else
        {
            throw IrisInternalError("Instance tried to read invalid parameter");
        }
    }

    /*!
     * @brief Get the value of a large numeric instantiation parameter.
     *
     * This is used for numeric parameters which are outside the range for uint64_t/int64_t.
     *
     * @param   name    The name of the parameter.
     * @param   value   A reference to a value to type @a T which will receive the value
     *                  of the named parameter.
     */
    void getParameter(const std::string &name, std::vector<uint64_t> &value);

    /*!
     * @brief Get flags to use when registering the instance being instantiated.
     *
     * @return          A value to use for the flags argument of IrisInstance::IrisInstance()
     *                  or IrisInstance::registerInstance().
     */
    uint64_t getRecommendedInstanceFlags() const
    {
        return instance_flags;
    }

    /*!
     * @brief Get instance name to use when registering the instance being instantiated.
     *
     * @return          A value to use for the @c instName argument of IrisInstance::IrisInstance()
     *                  or IrisInstance::registerInstance().
     */
    std::string getInstanceName() const
    {
        return prefix + "." + component_name;
    }

    /*!
     * @brief Get connection interface to use to register the instance being instantiated.
     *
     * @return          A value to use for the @c connection_interface argument of IrisInstance::IrisInstance().
     */
    IrisConnectionInterface *getConnectionInterface() const
    {
        return connection_interface;
    }

    /*!
     * @brief Add a warning to the InstantiationResult.
     *
     * @see parameterWarning
     *
     * @param   code            An error code symbol. This should be one of the codes specified
     *                          for the InstantiationError object.
     * @param   format          A printf style format string.
     * @param   ...             Printf substitution arguments.
     */
    void warning(const std::string &code, const char *format, ...) INTERNAL_IRIS_PRINTF(3, 4);

    /*!
     * @brief Add a warning to the InstantiationResult.
     *
     * @see warning
     *
     * @param   code            An error code symbol. This should be one of the codes specified
     *                          for the InstantiationError object.
     * @param   parameterName   The name of the parameter this warning relates to.
     * @param   format          A printf style format string.
     * @param   ...             Printf substitution arguments.
     */
    void parameterWarning(const std::string &code, const std::string &parameterName, const char *format, ...) INTERNAL_IRIS_PRINTF(4, 5);
    /*!
     * @brief Add a error to the InstantiationResult.
     *
     * @see parameterError
     *
     * @param   code            An error code symbol. This should be one of the codes specified
     *                          for the InstantiationError object.
     * @param   format          A printf style format string.
     * @param   ...             Printf substitution arguments.
     */
    void error(const std::string &code, const char *format, ...) INTERNAL_IRIS_PRINTF(3, 4);

    /*!
     * @brief Add a error to the InstantiationResult.
     *
     * @see error
     *
     * @param   code            An error code symbol. This should be one of the codes specified
     *                          for the InstantiationError object.
     * @param   parameterName   The name of the parameter this error relates to.
     * @param   format          A printf style format string.
     * @param   ...             Printf substitution arguments.
     */
    void parameterError(const std::string &code, const std::string &parameterName, const char *format, ...) INTERNAL_IRIS_PRINTF(4, 5);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisInstantiationContext_h
