/*!
 * \file    IrisParameterBuilder.h
 * \brief   Helper class to construct instantiation parameters.
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisParameterBuilder_h
#define ARM_INCLUDE_IrisParameterBuilder_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"

#include <vector>
#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief Helper class to construct instantiation parameters.
 */
class IrisParameterBuilder
{
private:
    ResourceInfo &info;

    IrisParameterBuilder &setValueExtend(std::vector<uint64_t> &arr, uint64_t value, uint64_t extension)
    {
        arr.resize(info.getDataSizeInU64Chunks(), extension);
        arr[0] = value;

        return *this;
    }

    IrisParameterBuilder &setValueExtend(std::vector<uint64_t> &arr, const std::vector<uint64_t> &value, uint64_t extension)
    {
        size_t param_size = info.getDataSizeInU64Chunks();
        if (param_size < value.size())
        {
            throw IrisInternalError("Invalid parameter configuration");
        }
        arr = value;
        arr.resize(info.getDataSizeInU64Chunks(), extension);

        return *this;
    }

    IrisParameterBuilder &setValueSignExtend(std::vector<uint64_t> &arr, int64_t value)
    {
        return setValueExtend(arr, static_cast<uint64_t>(value), (value < 0)? IRIS_UINT64_MAX : 0);
    }

    IrisParameterBuilder &setValueZeroExtend(std::vector<uint64_t> &arr, uint64_t value)
    {
        return setValueExtend(arr, value, 0);
    }

    IrisParameterBuilder &setValueSignExtend(std::vector<uint64_t> &arr, const std::vector<uint64_t> &value)
    {
        return setValueExtend(arr, value, (static_cast<int64_t>(value.back()) < 0)? IRIS_UINT64_MAX : 0);
    }

    IrisParameterBuilder &setValueZeroExtend(std::vector<uint64_t> &arr, const std::vector<uint64_t> &value)
    {
        return setValueExtend(arr, value, 0);
    }

    IrisParameterBuilder &setValueDouble(std::vector<uint64_t> &arr, double value)
    {
        arr.resize(1);
        *static_cast<double *>((void*)(&arr[0])) = value;

        return *this;
    }

public:

    /*!
     * @brief Construct a parameter builder for a given parameter resource.
     * @param   info_   The resource info object for the parameter being built.
     */
    IrisParameterBuilder(ResourceInfo &info_)
        : info(info_)
    {
        info.isParameter = true;
    }

    /*!
     * @brief Set the @c name field.
     * @param   name        The @c name field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setName(const std::string &name)
    {
        info.name = name;
        return *this;
    }

    /*!
     * @brief Set the @c descr field.
     * @param   descr       The @c descr field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDescr(const std::string &descr)
    {
        info.descr = descr;
        return *this;
    }

    /*!
     * @brief Set the @c format field.
     * @param   format      The @c format field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setFormat(const std::string &format)
    {
        info.format = format;
        return *this;
    }

    /*!
     * @brief Set the @c bitWidth field.
     * @param   bitWidth    The @c bitWidth field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setBitWidth(uint64_t bitWidth)
    {
        info.bitWidth = bitWidth;
        return *this;
    }

    /*!
     * @brief Set the @c rwMode field.
     * @param   rwMode      The @c rwMode field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRwMode(const std::string &rwMode)
    {
        info.rwMode = rwMode;
        return *this;
    }

    /*!
     * @brief Set the @c subRscId field.
     * @param   subRscId    The @c subRscId field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setSubRscId(uint64_t subRscId)
    {
        info.subRscId = subRscId;
        return *this;
    }

    /*!
     * @brief Set the @c topology field.
     * @param   value       The @c topology field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setTopology(bool value = true)
    {
        info.parameterInfo.topology = value;
        return *this;
    }

    /*!
     * @brief Set the @c initOnly field.
     * @param   value       The @c initOnly field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setInitOnly(bool value = true)
    {
        info.parameterInfo.initOnly = value;
        return *this;
    }

    /*!
     * @brief Set the @c min field.
     * @param   min         The @c min field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMin(uint64_t min)
    {
        return setValueZeroExtend(info.parameterInfo.min, min);
    }

    /*!
     * @brief Set the @c max field.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMax(uint64_t max)
    {
        return setValueZeroExtend(info.parameterInfo.max, max);
    }

    /*!
     * @brief Set both the @c min field and the @c max field.
     * @param   min         The @c min field of the ParameterInfo object.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRange(uint64_t min, uint64_t max)
    {
        return setMin(min).setMax(max);
    }

    /*!
     * @brief Set the @c min field.
     *
     * Use this variant to set values that are >= 2**64.
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMin(const std::vector<uint64_t> &min)
    {
        return setValueZeroExtend(info.parameterInfo.min, min);
    }

    /*!
     * @brief Set the @c max field.
     *
     * Use this variant to set values that are >= 2**64.
     *
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMax(const std::vector<uint64_t> &max)
    {
        return setValueZeroExtend(info.parameterInfo.max, max);
    }

    /*!
     * @brief Set both the @c min field and the @c max field.
     *
     * Use this variant to set values that are >= 2**64.
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRange(const std::vector<uint64_t> &min, const std::vector<uint64_t> &max)
    {
        return setMin(min).setMax(max);
    }

    /*!
     * @brief Set the @c min field.
     *
     * This implies that the parameter type is "numericSigned".
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMinSigned(int64_t min)
    {
        return setValueSignExtend(info.parameterInfo.min, min)
              .setType("numericSigned");
    }

    /*!
     * @brief Set the @c max field.
     *
     * This implies that the parameter type is "numericSigned".
     *
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMaxSigned(int64_t max)
    {
        return setValueSignExtend(info.parameterInfo.max, max)
              .setType("numericSigned");
    }

    /*!
     * @brief Set both the @c min field and the @c max field.
     *
     * This implies that the parameter type is "numericSigned".
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRangeSigned(int64_t min, int64_t max)
    {
        return setValueSignExtend(info.parameterInfo.min, min)
              .setValueSignExtend(info.parameterInfo.max, max)
              .setType("numericSigned");
    }

    /*!
     * @brief Set the @c min field.
     *
     * This implies that the parameter type is "numericSigned".
     * Use this variant for signed values that are out of range for int64_t.
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMinSigned(const std::vector<uint64_t> &min)
    {
        return setValueSignExtend(info.parameterInfo.min, min)
              .setType("numericSigned");
    }

    /*!
     * @brief Set the @c max field.
     *
     * This implies that the parameter type is "numericSigned".
     * Use this variant for signed values that are out of range for int64_t.
     *
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMaxSigned(const std::vector<uint64_t> &max)
    {
        return setValueSignExtend(info.parameterInfo.max, max)
              .setType("numericSigned");
    }

    /*!
     * @brief Set both the @c min field and the @c max field.
     *
     * This implies that the parameter type is "numericSigned".
     * Use this variant for signed values that are out of range for int64_t.
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRangeSigned(const std::vector<uint64_t> &min, const std::vector<uint64_t> &max)
    {
        return setValueSignExtend(info.parameterInfo.min, min)
              .setValueSignExtend(info.parameterInfo.max, max)
              .setType("numericSigned");
    }

    /*!
     * @brief Set the @c min field for floating point parameters.
     *
     * This implies that the parameter type is "numericFp".
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMinFloat(double min)
    {
        return setValueDouble(info.parameterInfo.min, min)
              .setType("numericFp");
    }

    /*!
     * @brief Set the @c max field for floating point parameters.
     *
     * This implies that the parameter type is "numericFp".
     *
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setMaxFloat(double max)
    {
        return setValueDouble(info.parameterInfo.max, max)
              .setType("numericFp");
    }

    /*!
     * @brief Set both the @c min field and the @c max field.
     *
     * This implies that the parameter type is "numericFp".
     *
     * @param   min         The @c min field of the ParameterInfo object.
     * @param   max         The @c max field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setRangeFloat(double min, double max)
    {
        return setValueDouble(info.parameterInfo.min, min)
              .setValueDouble(info.parameterInfo.max, max)
              .setType("numericFp");
    }

    /*!
     * @brief Add an enum symbol for this parameter.
     *
     * @param   symbol      The enum symbol that is being added.
     * @param   value       The value associated with that symbol.
     * @param   descr       A description explaining the meaning of the symbol.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &addEnum(const std::string &symbol, const IrisValue &value, const std::string &descr = std::string())
    {
        info.enums.push_back(EnumElementInfo(value, symbol, descr));
        return *this;
    }

    /*!
     * @brief Add a string enum symbol for this parameter.
     *
     * For string enums both the symbol and value are the same.
     *
     * @param   value       The value associated with that symbol.
     * @param   descr       A description explaining the meaning of the symbol.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &addStringEnum(const std::string &value, const std::string &descr = std::string())
    {
        info.enums.push_back(EnumElementInfo(IrisValue(value), std::string(), descr));
        return *this;
    }

    /*!
     * @brief Set a boolean tag for this parameter resource.
     *
     * @param   tag         The name of the tag to set.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setTag(const std::string &tag)
    {
        info.tags[tag] = IrisValue(true);
        return *this;
    }

    /*!
     * @brief Set a tag for this parameter resource.
     *
     * @param   tag         The name of the tag to set.
     * @param   value       The value to set for this tag.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setTag(const std::string &tag, const IrisValue &value)
    {
        info.tags[tag] = value;
        return *this;
    }

    /*!
     * @brief Set the default value for a string parameter.
     *
     * @param   value       The @c defaultString field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefault(const std::string &value)
    {
        info.parameterInfo.defaultString = value;
        return *this;
    }

    /*!
     * @brief Set the default value for a numeric parameter.
     *
     * @param   value       The @c defaultData field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefault(uint64_t value)
    {
        return setValueZeroExtend(info.parameterInfo.defaultData, value);
    }

    /*!
     * @brief Set the default value for a numeric parameter.
     *
     * Use this variant for values that are >= 2**64.
     *
     * @param   value       The @c defaultData field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefault(const std::vector<uint64_t> &value)
    {
        return setValueZeroExtend(info.parameterInfo.defaultData, value);
    }

    /*!
     * @brief Set the default value for a numericSigned parameter.
     *
     * @param   value       The @c defaultData field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefaultSigned(int64_t value)
    {
        return setValueSignExtend(info.parameterInfo.defaultData, value);
    }

    /*!
     * @brief Set the default value for a numericSigned parameter.
     *
     * Use this variant for values that are out of range for int64_t.
     *
     * @param   value       The @c defaultData field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefaultSigned(const std::vector<uint64_t> &value)
    {
        return setValueSignExtend(info.parameterInfo.defaultData, value);
    }

    /*!
     * @brief Set the default value for a numericFp parameter.
     *
     * @param   value       The @c defaultData field of the ParameterInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setDefaultFloat(double value)
    {
        return setValueDouble(info.parameterInfo.defaultData, value);
    }

    /*!
     * @brief Set the type of this parameter.
     *
     * The bitWidth field must be set before setting the type.
     *
     * @param   type        The @c type field of the ResourceInfo object.
     * @return              A reference to this IrisParameterBuilder object allowing calls to be chained together.
     */
    IrisParameterBuilder &setType(const std::string &type)
    {
        if ((info.bitWidth != 32) && (info.bitWidth != 64) && (type == "numericFp"))
        {
            throw IrisInternalError("Invalid parameter configuration."
                                    " NumericFp parameters must have a bitWidth of 32 or 64");
        }

        info.type = type;
        return *this;
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisParameterBuilder_h
