/*!
 * \file    IrisPluginFactory.h
 * \brief   A generic plugin factory for instantiating plugin instances
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 */

#ifndef ARM_INCLUDE_IrisPluginFactory_h
#define ARM_INCLUDE_IrisPluginFactory_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisThreadAbstractionDefault.h"
#include "iris/detail/IrisU64JsonReader.h"
#include "iris/detail/IrisU64JsonWriter.h"
#include "iris/detail/IrisFunctionInfo.h"
#include "iris/IrisCConnection.h"
#include "iris/IrisInstance.h"
#include "iris/IrisInstantiationContext.h"
#include "iris/IrisInstanceFactoryBuilder.h"

#include <vector>
#include <string>

NAMESPACE_IRIS_START

/*!
 * @brief Set meta-data for instantiating a plugin instance.
 */
class IrisPluginFactoryBuilder : public IrisInstanceFactoryBuilder
{
private:
    /// Name of the plugin. Usually this will be the same name as the plugin library.
    /// It is used to construct the default isntName for plugin instances and the plugin factory.
    std::string plugin_name;

    /// Prefix used for instances of this plugin and the plugin factory.
    std::string instance_name_prefix;

    /// Default name for plugin instances. Default is the same as plugin_name
    std::string default_instance_name;

public:
    /*!
     * @param   name    The name of the plugin to build.
     */
    IrisPluginFactoryBuilder(const std::string &name)
        : IrisInstanceFactoryBuilder(/*parameter_prefix=*/ "")
        , plugin_name(name)
        , instance_name_prefix("client.plugin")
    {
    }

    /*!
     * @brief Override the plugin name.
     *
     * The factory provides a sensible default for this so it should only be overridden if there is
     * a good reason to do so.
     *
     * @param   name    The name of the plugin.
     */
    void setPluginName(const std::string &name)
    {
        plugin_name = name;
    }

    /*!
     * @brief Get the plugin name.
     * @return          The name of the plugin.
     */
    const std::string &getPluginName() const
    {
        return plugin_name;
    }

    /*!
     * @brief Override the instance name prefix (default is "client.plugin").
     *
     * The factory provides a sensible default for this so it should only be overridden if there is
     * a good reason to do so.
     *
     * @param   prefix  The prefix that will be used for instances of this plugin.
     */
    void setInstanceNamePrefix(const std::string &prefix)
    {
        instance_name_prefix = prefix;
    }

    /*!
     * @brief Get the prefix that should be used for instances of this plugin.
     * @return          The prefix that will be used for instances of this plugin.
     */
    const std::string &getInstanceNamePrefix() const
    {
        return instance_name_prefix;
    }

    /*!
     * @brief Override the default instance name for plugin instances.
     *
     * The factory provides a sensible default for this so it should only be overridden if there is
     * a good reason to do so.
     *
     * @param   name        The default name for plugin instances.
     */
    void setDefaultInstanceName(const std::string &name)
    {
        default_instance_name = name;
    }

    /*!
     * @brief Get the default name that should be used for plugin instances.
     *
     * @return              The default name for plugin instances.
     */
    const std::string &getDefaultInstanceName() const
    {
        if (default_instance_name.empty())
        {
            return getPluginName();
        }
        else
        {
            return default_instance_name;
        }
    }
};


template <typename PLUGIN_INSTANCE>
class IrisPluginFactory
{
private:
    /// Remote iris interface to handle the requests from this plugin.
    IrisCConnection connection_interface;

    /// Iris instance of this plugin factory.
    IrisInstance factory_instance;

    /// List of loaded plugins
    std::vector<PLUGIN_INSTANCE*> plugin_instances;

    IrisThreadAbstractionMutex *plugin_instances_mutex;

    /// Contains parameter and configuration information and has methods to set up the same.
    IrisPluginFactoryBuilder builder;

    /// Implementation of plugin_getInstantiationParameterInfo() function
    void impl_plugin_getInstantiationParameterInfo(IrisReceivedRequest &req)
    {
        factory_instance.sendResponse(req.generateOkResponse(builder.getParameterInfo()));
    }

    /// Implementation of plugin_instantiate() function
    void impl_plugin_instantiate(IrisReceivedRequest &req)
    {
        InstantiationResult result;
        result.success = true; // Assume we will succeed until proven otherwise

        uint64_t instance_flags = IrisInstance::DEFAULT_FLAGS;

        std::string instName;

        if (!req.getOptionalArg(ISTR("instName"), instName))
        {
            instName = builder.getDefaultInstanceName();
            instance_flags |= IrisInstance::UNIQUIFY;
        }

        std::vector<InstantiationParameterValue> param_values;
        req.getOptionalArg(ISTR("paramValues"), param_values);

        // Build the full parameter info list
        const std::vector<ResourceInfo> &param_info = builder.getParameterInfo();
        const std::vector<ResourceInfo> &hidden_param_info = builder.getHiddenParameterInfo();

        std::vector<ResourceInfo> all_param_info;
        all_param_info.insert(all_param_info.end(), param_info.begin(), param_info.end());
        all_param_info.insert(all_param_info.end(), hidden_param_info.begin(), hidden_param_info.end());

        IrisInstantiationContext init_context(&connection_interface, result,
                                              all_param_info, param_values,
                                              builder.getInstanceNamePrefix(),
                                              instName, instance_flags);


        // Parameters have been validated. If they all passed we can instantiate the plugin.

        if (result.success)
        {
            try
            {
                IrisThreadAbstractionLock lock(*plugin_instances_mutex);

                plugin_instances.push_back(new PLUGIN_INSTANCE(init_context));

                if (!result.success)
                {
                    // The plugin instance set an error in its constructor so destroy it.
                    delete plugin_instances.back();
                    plugin_instances.pop_back();
                }
            }
            catch (IrisErrorException &e)
            {
                result.success = false;
                result.errors.resize(result.errors.size() + 1);

                InstantiationError &error = result.errors.back();
                error.severity = "error";
                error.code = "error_general_error";
                error.message = e.getMessage();
            }
            catch (...)
            {
                result.success = false;
                result.errors.resize(result.errors.size() + 1);

                InstantiationError &error = result.errors.back();
                error.severity = "error";
                error.code = "error_general_error";
                error.message = "Internal error while instantiating plugin";
            }
        }

        factory_instance.sendResponse(req.generateOkResponse(result));
    }

public:
    IrisPluginFactory(IrisC_Functions *iris_c_functions, const std::string& plugin_name)
        : connection_interface(iris_c_functions)
        , factory_instance(&connection_interface)
        , builder(plugin_name)
    {
        PLUGIN_INSTANCE::buildPluginFactory(builder);

        typedef IrisPluginFactory<PLUGIN_INSTANCE> Self;

        factory_instance.irisRegisterFunction(this, Self, plugin_getInstantiationParameterInfo,
                                              function_info::plugin_getInstantiationParameterInfo);

        factory_instance.irisRegisterFunction(this, Self, plugin_instantiate,
            "{descr:'Instantiate an instance of the " + builder.getPluginName() + " plugin',"
            "args:{"
            "  instName:{type:'String', descr:'Used to construct the instance name for the new instance."
                " Instance name will be \"" + builder.getInstanceNamePrefix() +  "<instName>\"',"
                "defval:'" + builder.getDefaultInstanceName() + "', optional:true},"
            "  paramValues:{type:'Array', descr:'Instantiation parameter values'}"
            "},"
            "retval:{type:'InstantiationResult', descr:'Indicates success of and errors/warnings"
                                                       " that occurred during plugin instantiation.'}}");

        // Register factory instance
        uint64_t flags = IrisInstance::DEFAULT_FLAGS
                       | IrisInstance::UNIQUIFY;

        std::string factory_instName = "framework.plugin." + builder.getPluginName() + "Factory";
        factory_instance.registerInstance(factory_instName, flags);
        factory_instance.setProperty("componentType", std::string("PluginFactory"));

        IrisLogger log("IrisPluginFactory");
        IrisThreadAbstractionDefault thread_abstraction(log);
        plugin_instances_mutex = thread_abstraction.createMutex();
    }

    ~IrisPluginFactory()
    {
        {
            IrisThreadAbstractionLock lock(*plugin_instances_mutex);

            // Clean up plugin instances
            typename std::vector<PLUGIN_INSTANCE *>::iterator it;
            for (it = plugin_instances.begin(); it != plugin_instances.end(); ++it)
            {
                delete *it;
            }
        }

        delete plugin_instances_mutex;
    }

    // Unregister factory instance. Call this when unloading a plugin before simulation termination.
    IrisErrorCode unregisterInstance()
    {
        return factory_instance.unregisterInstance();
    }

    // Implementation of the plugin entry point.
    // This will initialize an IrisPluginFactory the first time it is called.
    static int64_t initPlugin(IrisC_Functions *functions, const std::string &plugin_name)
    {
        static IrisPluginFactory<PLUGIN_INSTANCE> *factory = NULL;

        if (factory == NULL)
        {
            factory = new IrisPluginFactory<PLUGIN_INSTANCE>(functions, plugin_name);
            return E_ok;
        }
        else
        {
            return E_plugin_already_loaded;
        }
    }
};

#if defined(_WIN32)
    #if defined(_MSC_VER)
        #define IRIS_PLUGIN_EXPORT __declspec(dllexport)
    #else
        #error "Unsupported Compiler"
    #endif
#elif defined(__unix__) || defined(__unix) || defined(unix)
    #if __GNUC__ >= 4
        #define IRIS_PLUGIN_EXPORT __attribute__((visibility ("default")))
    #else
        #define IRIS_PLUGIN_EXPORT
    #endif
#else
    #error "Unsupported OS"
#endif

/*!
 * @brief Use this macro to create an Iris plugin entry point.
 *
 * @param   plugin_instance     Objects of this type will be instantiated for each plugin instance created.
 */
#define IRIS_PLUGIN_FACTORY(plugin_instance) \
    extern "C" IRIS_PLUGIN_EXPORT int64_t irisInitPlugin(IrisC_Functions *functions) \
    { \
        return ::iris::IrisPluginFactory<plugin_instance>::initPlugin(functions, #plugin_instance); \
    }

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisPluginFactory_h
