/*!
 *  \file   IrisEventEmitter.h
 *  \brief  A utility class for emitting Iris Events
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisRegisterEventEmitter_h
#define ARM_INCLUDE_IrisRegisterEventEmitter_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisRegisterEventEmitterBase.h"

NAMESPACE_IRIS_START

/*!
 * @brief An EventEmitter class for register read events
 *
 * @tparam  REG_T   The type of the register being read.
 * @tparam  ARGS    Types of any custom fields that this event source defines
 *                  in addition to the standard fields defined for register read events.
 *
 * Use IrisRegisterReadEventEmitter with IrisInstanceBuilder to add register read events
 * to you Iris instance.
 * @code
       // Declare an event emitter
       iris::IrisRegisterReadEventEmitter<uint64_t> reg_read_event;

       // Add it to an Iris instance
       iris::IrisInstance my_instance(...);
       iris::IrisInstanceBuilder *builder = my_instance->getBuilder();
       builder->setRegisterReadEvent("READ_REG", reg_read_event);

       // Add some registers that will be traced by this event
       builder->setNextRscId(0x1000);
       builder->addRegister("X0", 64, "Register X0");
       builder->addRegister("X1", 64, "Register X1");
       builder->addRegister("X2", 64, "Register X2");
       builder->addRegister("X3", 64, "Register X3");


       // Now that the Instance builder has the meta-data for the registers we need
       // to finalize the register read event to populate the event meta-data.
       builder->finalizeRegisterReadEvent();

       uint64_t readRegister(unsigned reg_index, bool is_debug)
       {
            uint64_t value = readRegValue(reg_index);

            // Emit an event
            reg_read_event(0x1000 | reg_index, is_debug, value);

            return value;
        }
   @endcode
 */
template <typename REG_T, typename... ARGS>
class IrisRegisterReadEventEmitter : public IrisRegisterEventEmitterBase
{
public:
    IrisRegisterReadEventEmitter()
        : IrisRegisterEventEmitterBase(sizeof...(ARGS) + 3)
    {
    }

    /*!
     * @brief Emit an event
     *
     * @param   rscId       Resource id for the register that was accessed.
     * @param   debug       True if this access originated from a debug access.
     * @param   value       The register value that was read during this event.
     * @param   args        Any additional custom fields for this event.
     */
    void operator() (uint64_t rscId, bool debug, REG_T value, ARGS... args)
    {
        // Emit event
        emitEvent(rscId, debug, value, args...);

        // Check if this event indicates a breakpoint was hit
        if (!debug)
        {
            checkBreakpointHit(rscId, value, /*is_read=*/ true);
        }
    }

};

/*!
 * @brief An EventEmitter class for register update events
 *
 * @tparam  REG_T   The type of the register being read.
 * @tparam  ARGS    Types of any custom fields that this event source defines
 *                  in addition to the standard fields defined for register update events.
 *
 * Use IrisRegisterUpdateEventEmitter with IrisInstanceBuilder to add register update events
 * to you Iris instance.
 * @code
       // Declare an event emitter
       iris::IrisRegisterUpdateEventEmitter<uint64_t> reg_update_event;

       // Add it to an Iris instance
       iris::IrisInstance my_instance(...);
       iris::IrisInstanceBuilder *builder = my_instance->getBuilder();
       builder->setRegisterUpdateEvent("WRITE_REG", reg_update_event);

       // Add some registers that will be traced by this event
       builder->setNextRscId(0x1000);
       builder->addRegister("X0", 64, "Register X0");
       builder->addRegister("X1", 64, "Register X1");
       builder->addRegister("X2", 64, "Register X2");
       builder->addRegister("X3", 64, "Register X3");

       // Now that the Instance builder has the meta-data for the registers we need
       // to finalize the register update event to populate the event meta-data.
       builder->finalizeRegisterUpdateEvent();

       void writeRegister(unsigned reg_index, bool is_debug, uint64_t new_value)
       {
            uint64_t old_value = readRegValue(reg_index);
            writeRegValue(reg_index, new_value);

            // Emit an event
            reg_update_event(0x1000 | reg_index, is_debug, old_value, new_value);
        }
   @endcode
 */
template <typename REG_T, typename... ARGS>
class IrisRegisterUpdateEventEmitter : public IrisRegisterEventEmitterBase
{
public:
    IrisRegisterUpdateEventEmitter()
        : IrisRegisterEventEmitterBase(sizeof...(ARGS) + 4)
    {
    }

    /*!
     * @brief Emit an event
     *
     * @param   rscId       Resource id for the register that was accessed.
     * @param   debug       True if this access originated from a debug access.
     * @param   old_value   The register value before the event.
     * @param   new_value   The register value after the event.
     * @param   args        Any additional custom fields for this event.
     */
    void operator() (uint64_t rscId, bool debug, REG_T old_value, REG_T new_value, ARGS... args)
    {
        // Emit event
        emitEvent(rscId, debug, old_value, new_value, args...);

        // Check if this event indicates a breakpoint was hit
        if (!debug)
        {
            checkBreakpointHit(rscId, new_value, /*is_read=*/ false);
        }
    }

};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisRegisterEventEmitter_h
