/*!
 *  \file  IrisC.h
 *  \brief C Interface between IrisCore library, plugin libraries and in-process instances.
 *  \date  Copyright ARM Limited 2017. All rights reserved.
 *
 *  Generally Iris instances will not need to use the C interface or IrisCore directly and should
 *  use the C++ interfaces provided by IrisSupportLib.
 *
 *  All IrisCore functions are internally thread safe and may be called from any thread.
 *
 *  All memory passed through IrisC is owned by the caller and must not be accessed by the callee
 *  after the call has returned. The callee must always make a copy of the memory if it needs to
 *  access it later.
 *
 *  Function pointers and interface pointers passed through IrisC are owned by the instance that
 *  originated them and must be valid for the lifetime of the instance.
 */

#ifndef ARM_INCLUDE_Iris_C_h
#define ARM_INCLUDE_Iris_C_h

#define __STDC_LIMIT_MACROS
#include <stdint.h>
#include <stddef.h>

#ifdef __cplusplus
extern "C"
{
#endif // __cplusplus


/*!
 * \defgroup IrisCoreLifecycle
 * @{
 */

/*!
 *   Initialise the IrisGlobalInstance
 *
 *      int64_t IrisCore_init(void **iris_core_context_out)
 *
 *   This function must be called before calling any other IrisCore functions to obtain
 *   a context pointer.
 *
 *   \param[in]  iris_core_context_out  Context pointer to use for all calls to IrisCore functions.
 *                                      This should be passed as the context argument when calling
 *                                      all other IrisCore functions.
 *   \return                            An IrisErrorCode indicating whether the call was successful.
 */
typedef int64_t (*IrisC_IrisCoreInitFunction) (void **iris_core_context_out);

/*!
 *   Destroy the IrisGlobalInstance
 *
 *      int64_t IrisCore_shutdown(void *iris_core_context)
 *
 *   Once IrisCore_shutdown() returns no future calls should be made using this context pointer.
 *
 *   \param[in]  iris_core_context  Context pointer for IrisGlobalInstance
 *   \return                        An IrisErrorCode indicating whether the call was successful.
 */
typedef int64_t (*IrisC_IrisCoreShutdownFunction) (void *iris_core_context);

/* @} */

/*!
 * \defgroup IrisClientLifecycle
 * @{
 */

/*!
 *   Initialise and connect an Iris client.
 *
 *      int64_t IrisClient_connect(void **iris_client_context_out, const char *hostname, uint16_t port);
 *
 *   This function must be called before calling any other IrisCore functions to obtain
 *   a context pointer.
 *
 *   \param[out]  iris_client_context_out  Context pointer to use for all calls to IrisClient functions.
 *                                         This should be passed as the context argument when calling
 *                                         all other IrisClient functions.
 *   \param[in]   hostname                 Hostname or IP address of Iris server to connect to.
 *   \param[in]   port                     Port number of Iris server to connect to.
 *   \return                               An IrisErrorCode indicating whether the connection was successful.
 */
typedef int64_t (*IrisC_IrisClientConnectFunction) (void **iris_client_context_out, const char *hostname, uint16_t port);

/*!
 *   Disconnect and destroy an Iris client.
 *
 *      int64_t IrisClient_disconnect(void *iris_core_context)
 *
 *   Once IrisClient_disconnect() returns no future calls should be made using this context pointer.
 *
 *   \param[in]  iris_client_context  Context pointer for Iris client.
 *   \return                        An IrisErrorCode indicating whether the call was successful.
 */
typedef int64_t (*IrisC_IrisClientDisconnectFunction) (void *iris_client_context);

/* @} */

/*!
 * \defgroup IrisFunctionCall Iris function call functions
 * @{
 */

/*!
 *   Handle an Iris request or response
 *
 *      int64_t handleMessage(void *handle_message_context, const uint64_t *message)
 *
 *      int64_t IrisCore_handleMessage(void *iris_core_context, const uint64_t *message)
 *
 *      int64_t IrisClient_handleMessage(void *iris_core_context, const uint64_t *message)
 *
 *   \param[in]  handle_message_context  Context pointer for the response function.
 *   \param[in]  message                 U64JSON encoded request/response message.
 *   \return                             An IrisErrorCode indicating whether the message could be
 *                                       handled. Note that when making a request a return value of
 *                                       E_Ok does not indicate that the request was carried out
 *                                       successfully. Iris calls are asynchronous and might not have
 *                                       been executed at the point when handleMessage() returns.
 */
typedef int64_t (*IrisC_HandleMessageFunction) (void *handle_message_context, const uint64_t *message);

/*!
 *  Obtain a native interface to an instance
 *  All Iris instances must implement this function.
 *
 *      int64_t obtainInterface(void *obtain_interface_context,
 *                              uint64_t instId,
 *                              const char *requested_interface,
 *                              void **interface_ptr_out)
 *
 *      int64_t IrisCore_obtainInterface(void *iris_core_context,
 *                                       uint64_t instId,
 *                                       const char *requested_interface,
 *                                       void **interface_ptr_out)
 *
 *      NOTE: obtainInterface() is not supported for instances accessed remotely over TCP connection
 *            with IrisClient.
 *      int64_t IrisClient_obtainInterface(void *iris_core_context,
 *                                         uint64_t instId,
 *                                         const char *requested_interface,
 *                                         void **interface_ptr_out)
 *
 *  \param[in]  obtain_interface_context  Context pointer for the obtain interface function.
 *  \param[in]  instId                    Destination instId
 *  \param[in]  requested_interface       Name of the interface the caller wants to obtain.
 *  \param[out] interface_ptr_out         Returns a pointer to the interface that was requested
 *                                        or NULL if the requested interface could not be obtained.
 *  \return                               An IrisErrorCode indicating whether the call was successful.
 */
typedef int64_t (*IrisC_ObtainInterfaceFunction) (void *obtain_interface_context,
                                                  uint64_t instId,
                                                  const char *requested_interface,
                                                  void **interface_ptr_out);
/**@}*/

/*!
 * \defgroup IrisCoreC Iris C API functions
 *  These functions are used by implemened by IrisCore and used by all
 *  Iris instances to communicate with IrisCore.
 *
 * @{
 */

struct IrisC_CommunicationChannel
{
    /*! IrisC_CommunicationChannel struct version.
     *  Must be set to 0.
     */
    uint64_t CommunicationChannel_version;

    /*! The IrisC_HandleMessageFunction to call send messages (requests
     *  and responses) to the instance.
     *  Must be set by the caller.
     */
    IrisC_HandleMessageFunction handleMessage_function;

    /*! Context pointer passed to handle_message_function.
     *  Must be set by the caller.
     */
    void *handleMessage_context;

    /*! The IrisC_ObtainInterfaceFunction to call to obtain native
     *  interfaces for this instance.
     *  Must be set by the caller.
     *  May be NULL if native interfaces are not supported.
     */
    IrisC_ObtainInterfaceFunction obtainInterface_function;

    /*! Context pointer passed to obtain_interface_function.
     *  Must be set by the caller.
     */
    void *obtainInterface_context;
};

/*!
 * Register a communication channel.
 * Communication channels are used by IrisC libraries to route calls to instances.
 * When a channel is registered with an IrisC library (IrisCore or IrisClient) it is identified
 * by a unique channel Id returned by the library. This Id should be passed as the channelId parameter
 * when calling instanceRegistry_registerInstance() to ensure requests and responses are sent to the
 * right place.
 *
 *     int64_t IrisCore_registerChannel(void *iris_core_context,
 *                                      IrisC_CommunicationChannel *channel,
 *                                      uint64_t *channel_id);
 *
 *     int64_t IrisClient_registerChannel(void *iris_core_context,
 *                                        IrisC_CommunicationChannel *channel,
 *                                        uint64_t *channel_id);
 *
 * \param[in]  iris_core_context  IrisCore context pointer.
 * \param[in]  channel            Contains function pointers for the interfaces used to make
 *                                calls to an instance.
 * \param[out] channel_id         channel_id will be set to an id number used by IrisCore to identify
 *                                the channel. This id is used when registering instances using
 *                                the instanceRegistry_registerInstance() function.
 * \return                        An error code indicating whether the channel could be
 *                                registered.
 */
typedef int64_t (*IrisC_RegisterChannelFunction) (void *iris_core_context,
                                                  IrisC_CommunicationChannel *channel,
                                                  uint64_t *channel_id);

/*!
 * Unregister a channel.
 * This will implicitly unregister any instance using this channel if they are not already
 * been unregistered. Once a channel is unregistered no further calls will be made to this
 * channel.
 *
 *     int64_t IrisCore_unregisterChannel(void *iris_core_context, uint64_t channel_id);
 *
 *     int64_t IrisClient_unregisterChannel(void *iris_core_context, uint64_t channel_id);
 *
 * \param[in]  iris_core_context  IrisCore context pointer.
 * \param[in]  channel_id         The id for the channel being unregistered. Any instances that
 *                                have been registered using this channel will be automatically
 *                                unregistered if they have not already done so themselves.
 * \return                        An error code indicating whether the channel could be
 *                                unregistered.
 */
typedef int64_t (*IrisC_UnregisterChannelFunction) (void *iris_core_context, uint64_t channel_id);

/*!
 * Process messages that have been marshalled to the calling thread.
 *
 *     int64_t IrisCore_processAsyncMessages(void *iris_core_context, uint64_t flags);
 *
 *     int64_t IrisClient_processAsyncMessages(void *iris_client_context, uint64_t flags);
 *
 * \param[in]  iris_c_context  Context pointer for an IrisC library (IrisCore or IrisClient).
 * \param[in]  flags           Bitwise or of IrisC_AsyncMessage_* flags.
 * \return  An IrisErrorCode.
 */
typedef int64_t (*IrisC_ProcessAsyncMessagesFunction) (void *iris_c_context, uint64_t flags);

/*! IrisC_processAsyncMessages flag.
 *
 *  Default non-blocking behaviour. Returns immediately if there are no outstanding messages
 *  to process.
 */
static const uint64_t IrisC_AsyncMessage_Default = 0;

/*! IrisC_processAsyncMessages flag.
 *
 *  If there are no outstanding messages to process, block until there is one.
 *  This is useful when waiting for a response to a request. By calling processAsyncMessages()
 *  like this:
 *
 *      while (no_response_received)
 *          processAsyncMessages(context, IrisC_AsyncMessage_Blocking);
 *
 *  you wait for the response and also avoid deadlock situations where the recipient
 *  of you request makes a request itself that needs to be marshalled to the thread
 *  being blocked.
 */
static const uint64_t IrisC_AsyncMessage_Blocking = (1 << 0);

/**@}*/

/*!
 *  Data structure containing all the IrisC_* function
 *  (NOTE: this excludes IrisCore_* and IrisClient_* functions)
 *
 */
struct IrisC_Functions
{
    /*! IrisC_Functions struct version.
     *  Must be set to 0.
     */
    uint64_t Functions_version;

    /*! Context pointer to use when calling all IrisC functions */
    void *iris_c_context;

    /*! Pointer to an IrisC library's registerChannel function */
    IrisC_RegisterChannelFunction      registerChannel_function;

    /*! Pointer to an IrisC library's unregisterChannel function */
    IrisC_UnregisterChannelFunction    unregisterChannel_function;

    /*! Pointer to an IrisC library's handleMessage function */
    IrisC_HandleMessageFunction        handleMessage_function;

    /*! Pointer to an IrisC library's obtainInterface function */
    IrisC_ObtainInterfaceFunction      obtainInterface_function;

    /*! Pointer to an IrisC library's processAsyncMessages function */
    IrisC_ProcessAsyncMessagesFunction processAsyncMessages_function;
};

/*!
 * Initialize a plugin library.
 * This function should be implemented by plugin libraries as the entry point.
 *
 *     void irisInitPlugin(IrisC_Functions *functions);
 *
 * \param[in]  functions   a pointer to an IrisC_Functions struct which contains
 *                         pointers to the IrisCore library's IrisC functions so
 *                         that it can register instance(s) and interact with the
 *                         Iris world.
 *
 * \return                 An error code indicating whether the plugin was
 *                         initialized successfully.
 *
 */
typedef int64_t (*IrisC_InitPluginFunction) (IrisC_Functions *functions);

#ifdef __cplusplus
} // extern "C"
#endif // __cplusplus

#endif // ARM_INCLUDE_Iris_C_h
