/*!
 *  \file   IrisCppAdapter.h
 *  \brief  C++ Iris adapter: Call Iris functions from C++ (for clients, components and framework).
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  C++ Iris adapter: Call Iris functions from C++ (for clients, components and framework).
 */

#ifndef ARM_INCLUDE_IrisCppAdapter_h
#define ARM_INCLUDE_IrisCppAdapter_h

#include "iris/detail/IrisCommon.h"

#include "iris/detail/IrisObjects.h"
#include "iris/detail/IrisErrorCode.h"
#include "iris/detail/IrisRequest.h"
#include "iris/detail/IrisPendingRequestSet.h"
#include "iris/detail/IrisThreadAbstraction.h"
#include "iris/detail/IrisU64Json.h"
#include "iris/detail/IrisLogger.h"
#include <string>
#include <vector>
#include <map>


NAMESPACE_IRIS_START

/// Iris C++ adapter: Convert C++ function calls into Iris requests.
class IrisCppAdapter
{
private:
    friend class IrisInstance;
    class RequestManager
    {
    public:
        /// Constructor.
        RequestManager(IrisLogger &log)
            : connection_interface(0)
            , remoteIrisInterface(0)
            , pendingRequests(log)
            , nextRequestId(0)
            , instIdInUpperHalf(0)
        {
        }

        /// Set remote iris interface.
        /// This is the interface we send all requests and responses to.
        void setConnectionInterface(IrisConnectionInterface *connection_interface_)
        {
            assert(connection_interface_);
            connection_interface = connection_interface_;
            remoteIrisInterface = connection_interface->getIrisInterface();
        }


        /// Get remote iris interface.
        /// This is the interface we send all requests and responses to.
        IrisInterface *getRemoteIrisInterface() const { return remoteIrisInterface; }

        /// Get local iris interface which should receive all responses.
        IrisInterface *getLocalIrisInterface() { return pendingRequests.getIrisInterface(); }

        /// Set thread abstraction.
        void setThreadAbstraction(IrisThreadAbstraction *threadAbstraction)
        {
            assert(threadAbstraction);
            pendingRequests.init(threadAbstraction);
        }

        /// Set instId of the instance sending requests using this object.
        /// This is put into the upper 32 bits of all request ids.
        void setInstId(uint64_t instId)
        {
            instIdInUpperHalf = instId << 32;
        }

        /// Call Iris function and wait for response.
        /// Returns true iff we got a non-error result.
        bool callAndPerhapsWaitForResponse(IrisRequest &req, bool throw_on_error)
        {
            if (req.isRequest())
            {
                // Check whether this is the second call of a split request/response call:
                // Check whether we already got a result.
                if (!req.hasResponse())
                {
                    // Assign request id.
                    if (req.isRequest())
                        req.setRequestId(generateRequestId());

                    // Send request.
                    pendingRequests.addRequest(&req);
                    remoteIrisInterface->irisHandleMessage(req.generateRequestMessage());
                    if (req.isSplit())
                        return false; // For split request/response we are done after sending the request.

                    // Wait for response for this request.
                    while (!req.hasResponse())
                    {
                        IrisErrorCode code = connection_interface->processAsyncMessages(AsyncMessage_Blocking);
                        if (code == E_no_marshalling_for_current_thread)
                        {
                            // There is no marshalling to this thread so no need to process async
                            // events. We can just block the thread to wait for a response.
                            pendingRequests.waitForResponse(req);
                            assert(req.hasResponse());
                        }
                        else if (code != E_ok)
                        {
                            // An unexpected error occurred, propagate it.
                            throw IrisErrorException(code);
                        }
                    }
                }

                // Process response.
                req.processResponse();

                if (req.getErrorCode() != E_ok)
                {
                    if (throw_on_error)
                    {
                        // Iris errors throw an exception.
                        throw IrisErrorExceptionString(req.getErrorCode(), req.getErrorMessageAndData());
                    }
                    else
                    {
                        lastErrorMessage = req.getErrorMessageAndData();
                        return false; // E_<someerror>
                    }
                }
                return true; // E_ok
            }
            else
            {
                // This is a notification. We will not get a response.
                remoteIrisInterface->irisHandleMessage(req.generateRequestMessage());
                return false;
            }
        }

        /// Cancel a pending request. No need to wait for the response.
        void cancelPendingRequest(IrisRequest &req)
        {
            pendingRequests.erasePendingRequest(&req);
        }

        /// Generate a unique request id.
        /// This contains the source instId (our instId) in the upper 32 bits.
        uint64_t generateRequestId()
        {
            return instIdInUpperHalf | (nextRequestId++);
        }

        std::string getLastErrorMessage() const
        {
            return lastErrorMessage;
        }

    private:
        /// Iris Connection Interface
        IrisConnectionInterface *connection_interface;

        /// Remote iris interface.
        /// This is the interface we send all requests and responses to.
        IrisInterface *remoteIrisInterface;

        /// Response queue for reponses from outgoing requests.
        /// All responses are sent to this queue.
        IrisPendingRequestSet pendingRequests;

        /// Error message of last error which was returned.
        /// This is only meaningful if this object is used single threaded (which is not mandatory).
        /// This is not used when throwOnError==true which is preferred when using this instance from multiple threads.
        std::string lastErrorMessage;

        /// Next request id.
        uint32_t nextRequestId;

        /// instId of this instance in the upper 32 bits.
        uint64_t instIdInUpperHalf;
    };

public:
    IrisCppAdapter(RequestManager *request_manager_, bool throwOnError_)
        : request_manager(request_manager_)
        , throwOnError(throwOnError_)
    {
    }

    // --- Iris functions ---
    //
    // Translate C++ to outgoing Iris call.
    //
    // General signature guideline:
    // There are always at least two C++ functions per Iris function defined here:
    // 1. One which supports split request/response, which therefore accepts an IrisRequest& as the first parameter.
    // 2. One which does not support split request/response and so always blocks until the function returns.
    // (There may be variants of 1. and 2. for certain use-cases.)
    //
    // IrisErrorCode IRIS_FUNCTION_NAME(IrisRequest &req, [IrisInstId instId,] RETURN_VALUE, MANDATORY_ARGS, OPTIONAL_ARGS);
    // IrisErrorCode IRIS_FUNCTION_NAME([IrisInstId instId,] RETURN_VALUE, MANDATORY_ARGS, OPTIONAL_ARGS);

    IrisErrorCode instanceRegistry_registerInstance(IrisRequest &req, InstanceInfo &instanceInfoOut, const std::string &instName, uint64_t channelId, bool uniquify = false, bool marshalRequests = false)
    {
        if (!req.hasRequest())
        {
            req.initRequest("instanceRegistry_registerInstance", IrisInstIdGlobalInstance);
            req.addArg(ISTR("instName"), instName);
            req.addArg(ISTR("channelId"), channelId);
            if (uniquify)
                req.addArg(ISTR("uniquify"), uniquify);
            if (marshalRequests)
                req.addArg(ISTR("marshalRequests"), marshalRequests);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), instanceInfoOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instanceRegistry_registerInstance(InstanceInfo &instanceInfoOut, const std::string &instName, uint64_t channelId, bool uniquify = false, bool marshalRequests = false)
    {
        IrisRequest req;
        return instanceRegistry_registerInstance(req, instanceInfoOut, instName, channelId, uniquify, marshalRequests);
    }


    IrisErrorCode instanceRegistry_unregisterInstance(IrisRequest &req, IrisInstId aInstId)
    {
        if (!req.hasRequest())
        {
            req.initRequest("instanceRegistry_unregisterInstance", IrisInstIdGlobalInstance);
            req.addArg(ISTR("aInstId"), aInstId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // req.getResultReader().persistNull(); // Function does not return anything (returns null).
        }
        return req.getErrorCode();
    }
    IrisErrorCode instanceRegistry_unregisterInstance(IrisInstId aInstId)
    {
        IrisRequest req;
        return instanceRegistry_unregisterInstance(req, aInstId);
    }


    IrisErrorCode instanceRegistry_getList(IrisRequest &req, std::vector<InstanceInfo> &instanceListOut, const std::string &prefix = std::string())
    {
        if (!req.hasRequest())
        {
            req.initRequest("instanceRegistry_getList", IrisInstIdGlobalInstance);
            if (!prefix.empty())
                req.addArg(ISTR("prefix"), prefix);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), instanceListOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instanceRegistry_getList(std::vector<InstanceInfo> &instanceListOut, const std::string &prefix = std::string())
    {
        IrisRequest req;
        return instanceRegistry_getList(req, instanceListOut, prefix);
    }


    IrisErrorCode instanceRegistry_getInstanceInfoByInstId(IrisRequest &req, InstanceInfo &instanceInfoOut, IrisInstId aInstId)
    {
        if (!req.hasRequest())
        {
            req.initRequest("instanceRegistry_getInstanceInfoByInstId", IrisInstIdGlobalInstance);
            req.addArg(ISTR("aInstId"), aInstId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), instanceInfoOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instanceRegistry_getInstanceInfoByInstId(InstanceInfo &instanceInfoOut, IrisInstId aInstId)
    {
        IrisRequest req;
        return instanceRegistry_getInstanceInfoByInstId(req, instanceInfoOut, aInstId);
    }


    IrisErrorCode instanceRegistry_getInstanceInfoByName(IrisRequest &req, InstanceInfo &instanceInfoOut, const std::string &instName)
    {
        if (!req.hasRequest())
        {
            req.initRequest("instanceRegistry_getInstanceInfoByName", IrisInstIdGlobalInstance);
            req.addArg(ISTR("instName"), instName);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), instanceInfoOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instanceRegistry_getInstanceInfoByName(InstanceInfo &instanceInfoOut, const std::string &instName)
    {
        IrisRequest req;
        return instanceRegistry_getInstanceInfoByName(req, instanceInfoOut, instName);
    }


    IrisErrorCode instance_getProperties(IrisRequest &req, IrisInstId instId, std::map<std::string,IrisValue> &propertyMapOut)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("instance_getProperties"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), propertyMapOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instance_getProperties(IrisInstId instId, std::map<std::string,IrisValue> &propertyMapOut)
    {
        IrisRequest req;
        return instance_getProperties(req, instId, propertyMapOut);
    }


    IrisErrorCode instance_checkFunctionSupport(IrisRequest &req, IrisInstId instId, bool &out, const std::vector<FunctionSupportRequest> &functions)
    {
        if (!req.hasRequest())
        {
            req.initRequest("instance_checkFunctionSupport", instId);
            req.addArg(ISTR("functions"), functions);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instance_checkFunctionSupport(IrisInstId instId, bool &out, const std::vector<FunctionSupportRequest> &functions)
    {
        IrisRequest req;
        return instance_checkFunctionSupport(req, instId, out, functions);
    }


    IrisErrorCode instance_getFunctionInfo(IrisRequest &req, IrisInstId instId, std::map<std::string,FunctionInfo> &functionInfoMapOut, const std::string &prefix = std::string())
    {
        if (!req.hasRequest())
        {
            req.initRequest("instance_getFunctionInfo", instId);
            if (!prefix.empty())
                req.addArg(ISTR("prefix"), prefix);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), functionInfoMapOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode instance_getFunctionInfo(IrisInstId instId, std::map<std::string,FunctionInfo> &functionInfoMapOut, const std::string &prefix = std::string())
    {
        IrisRequest req;
        return instance_getFunctionInfo(req, instId, functionInfoMapOut, prefix);
    }


    // T must be supported by a global PersistPSH specialization, see IrisObjects.h, example InstanceInfo.
    template<class T>
    IrisErrorCode instance_ping(IrisRequest &req, IrisInstId instId, T &payloadOut, T payload)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("instance_ping"), instId);
            req.addArg(ISTR("payload"), payload);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), payloadOut);
        }
        return req.getErrorCode();
    }

    // T must be supported by a global PersistPSH specialization, see IrisObjects.h, example InstanceInfo.
    template<class T>
    IrisErrorCode instance_ping(IrisInstId instId, T &payloadOut, T payload)
    {
        IrisRequest req;
        return instance_ping(req, instId, payloadOut, payload);
    }

    IrisErrorCode instance_ping(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("instance_ping"), instId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode instance_ping(IrisInstId instId)
    {
        IrisRequest req;
        return instance_ping(req, instId);
    }


    // Resource interface.
    IrisErrorCode resource_getList(IrisRequest &req, IrisInstId instId, std::vector<ResourceInfo> &resourceInfosOut, uint64_t rscId = IRIS_UINT64_MAX, const std::string &group = std::string())
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("resource_getList"), instId);
            if (rscId != IRIS_UINT64_MAX)
                req.addArg(ISTR("rscId"), rscId);
            if (!group.empty())
                req.addArg(ISTR("group"), group);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), resourceInfosOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode resource_getList(IrisInstId instId, std::vector<ResourceInfo> &resourceInfosOut, uint64_t rscId = IRIS_UINT64_MAX, const std::string &group = std::string())
    {
        IrisRequest req;
        return resource_getList(req, instId, resourceInfosOut, rscId, group);
    }


    IrisErrorCode resource_getListOfResourceGroups(IrisRequest &req, IrisInstId instId, std::vector<ResourceGroupInfo> &resourceGroupInfosOut)
    {
        if (!req.hasRequest())
        {
            req.initRequest("resource_getListOfResourceGroups", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), resourceGroupInfosOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode resource_getListOfResourceGroups(IrisInstId instId, std::vector<ResourceGroupInfo> &resourceGroupInfosOut)
    {
        IrisRequest req;
        return resource_getListOfResourceGroups(req, instId, resourceGroupInfosOut);
    }


    IrisErrorCode resource_read(IrisRequest &req, IrisInstId instId, ResourceReadResult &resourceReadResultOut, const std::vector<uint64_t> &rscIds)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("resource_read"), instId);
            req.addArg(ISTR("rscIds"), rscIds);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), resourceReadResultOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode resource_read(IrisRequest &req, IrisInstId instId, ResourceReadResult &resourceReadResultOut, uint64_t rscId)
    {
        std::vector<uint64_t> rscIds;
        rscIds.push_back(rscId);
        return resource_read(req, instId, resourceReadResultOut, rscIds);
    }
    
    IrisErrorCode resource_read(IrisInstId instId, ResourceReadResult &resourceReadResultOut, const std::vector<uint64_t> &rscIds)
    {
        IrisRequest req;
        return resource_read(req, instId, resourceReadResultOut, rscIds);
    }
    IrisErrorCode resource_read(IrisInstId instId, ResourceReadResult &resourceReadResultOut, uint64_t rscId)
    {
        IrisRequest req;
        return resource_read(req, instId, resourceReadResultOut, rscId);
    }


    IrisErrorCode resource_write(IrisRequest &req, IrisInstId instId, ResourceWriteResult &resourceWriteResultOut,
                                 const std::vector<uint64_t> &rscIds, const std::vector<uint64_t> &data,
                                 const std::vector<std::string> *strings = NULL)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("resource_write"), instId);
            req.addArg(ISTR("rscIds"), rscIds);
            req.addArg(ISTR("data"), data);
            if (strings != NULL)
                req.addArg(ISTR("strings"), *strings);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), resourceWriteResultOut);
        }
        return req.getErrorCode();
    }
    // Only for resources <= 64 bit.
    IrisErrorCode resource_write(IrisRequest &req, IrisInstId instId, ResourceWriteResult &resourceWriteResultOut, uint64_t rscId, uint64_t value)
    {
        std::vector<uint64_t> rscIds;
        rscIds.push_back(rscId);
        std::vector<uint64_t> data;
        data.push_back(value);
        return resource_write(req, instId, resourceWriteResultOut, rscIds, data);
    }
    // Only for string resources.
    IrisErrorCode resource_write(IrisRequest &req, IrisInstId instId, ResourceWriteResult &resourceWriteResultOut, uint64_t rscId, const std::string &value)
    {
        std::vector<uint64_t> rscIds;
        rscIds.push_back(rscId);
        std::vector<uint64_t> data;
        std::vector<std::string> strings;
        strings.push_back(value);
        return resource_write(req, instId, resourceWriteResultOut, rscIds, data, &strings);
    }
    
    IrisErrorCode resource_write(IrisInstId instId, ResourceWriteResult &resourceWriteResultOut,
                                 const std::vector<uint64_t> &rscIds, const std::vector<uint64_t> &data,
                                 const std::vector<std::string> &strings)
    {
        IrisRequest req;
        return resource_write(req, instId, resourceWriteResultOut, rscIds, data, &strings);
    }
    IrisErrorCode resource_write(IrisInstId instId, ResourceWriteResult &resourceWriteResultOut, const std::vector<uint64_t> &rscIds, const std::vector<uint64_t> &data)
    {
        IrisRequest req;
        return resource_write(req, instId, resourceWriteResultOut, rscIds, data);
    }
    // Only for resources <= 64 bit.
    IrisErrorCode resource_write(IrisInstId instId, ResourceWriteResult &resourceWriteResultOut, uint64_t rscId, uint64_t value)
    {
        IrisRequest req;
        return resource_write(req, instId, resourceWriteResultOut, rscId, value);
    }
    // Only for string resources.
    IrisErrorCode resource_write(IrisInstId instId, ResourceWriteResult &resourceWriteResultOut, uint64_t rscId, const std::string &value)
    {
        IrisRequest req;
        return resource_write(req, instId, resourceWriteResultOut, rscId, value);
    }

    //--- Memory interface ---

    IrisErrorCode memory_getMemorySpaces(IrisRequest &req,
                                         IrisInstId instId,
                                         std::vector<MemorySpaceInfo> &spaceInfosOut)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("memory_getMemorySpaces"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), spaceInfosOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_getMemorySpaces(IrisInstId instId,
                                         std::vector<MemorySpaceInfo> &spaceInfosOut)
    {
        IrisRequest req;
        return memory_getMemorySpaces(req, instId, spaceInfosOut);
    }


    IrisErrorCode memory_read(IrisRequest &req,
                              IrisInstId instId,
                              MemoryReadResult &memoryReadResultOut,
                              uint64_t spaceId,
                              uint64_t address,
                              uint64_t byteWidth,
                              uint64_t count,
                              IrisValueMap *attrib = NULL)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("memory_read"), instId);
            req.addArg(ISTR("spaceId"), spaceId);
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("byteWidth"), byteWidth);
            req.addArg(ISTR("count"), count);
            if (attrib)
                req.addArg(ISTR("attrib"), *attrib);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), memoryReadResultOut);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_read(IrisInstId instId,
                              MemoryReadResult &memoryReadResultOut,
                              uint64_t spaceId,
                              uint64_t address,
                              uint64_t byteWidth,
                              uint64_t count,
                              IrisValueMap *attrib = NULL)
    {
        IrisRequest req;
        return memory_read(req, instId, memoryReadResultOut, spaceId, address, byteWidth, count, attrib);
    }


    IrisErrorCode memory_write(IrisRequest &req,
                               IrisInstId instId,
                               MemoryWriteResult &memoryWriteResult,
                               uint64_t spaceId,
                               uint64_t address,
                               uint64_t byteWidth,
                               uint64_t count,
                               const std::vector<uint64_t> &data,
                               IrisValueMap *attrib = NULL)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("memory_write"), instId);
            req.addArg(ISTR("spaceId"), spaceId);
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("byteWidth"), byteWidth);
            req.addArg(ISTR("count"), count);
            req.addArg(ISTR("data"), data);
            if (attrib)
                req.addArg(ISTR("attrib"), *attrib);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), memoryWriteResult);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_write(IrisInstId instId,
                               MemoryWriteResult &memoryWriteResult,
                               uint64_t spaceId,
                               uint64_t address,
                               uint64_t byteWidth,
                               uint64_t count,
                               const std::vector<uint64_t> &data,
                               IrisValueMap *attrib = NULL)
    {
        IrisRequest req;
        return memory_write(req, instId, memoryWriteResult, spaceId, address, byteWidth, count, data, attrib);
    }


    IrisErrorCode memory_getUsefulAddressTranslations(IrisRequest &req,
                                                      IrisInstId instId,
                                                      std::vector<MemorySupportedAddressTranslationResult> &result)
    {
        if (!req.hasRequest())
        {
            req.initRequest("memory_getUsefulAddressTranslations", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_getUsefulAddressTranslations(IrisInstId instId,
                                                      std::vector<MemorySupportedAddressTranslationResult> &result)
    {
        IrisRequest req;
        return memory_getUsefulAddressTranslations(req, instId, result);
    }


    IrisErrorCode memory_translateAddress(IrisRequest &req,
                                          IrisInstId instId,
                                          MemoryAddressTranslationResult &result,
                                          uint64_t spaceId,
                                          uint64_t address,
                                          uint64_t outSpaceId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("memory_translateAddress"), instId);
            req.addArg(ISTR("spaceId"), spaceId);
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("outSpaceId"), outSpaceId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_translateAddress(IrisInstId instId,
                                          MemoryAddressTranslationResult &result,
                                          uint64_t spaceId,
                                          uint64_t address,
                                          uint64_t outSpaceId)
    {
        IrisRequest req;
        return memory_translateAddress(req, instId, result, spaceId, address, outSpaceId);
    }


    IrisErrorCode memory_getSidebandInfo(IrisRequest &req, IrisInstId instId,
                                         IrisValueMap &result, uint64_t spaceId,
                                         uint64_t address, const IrisValueMap *attrib = NULL,
                                         const std::vector<std::string> *request = NULL)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("memory_getSidebandInfo"), instId);
            req.addArg(ISTR("spaceId"), spaceId);
            req.addArg(ISTR("address"), address);
            if (attrib)
                req.addArg(ISTR("attrib"), *attrib);
            if (request)
                req.addArg(ISTR("request"), *request);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result);
        }
        return req.getErrorCode();
    }
    IrisErrorCode memory_getSidebandInfo(IrisInstId instId, IrisValueMap &result,
                                         uint64_t spaceId, uint64_t address,
                                         const IrisValueMap *attrib = NULL,
                                         const std::vector<std::string> *request = NULL)
    {
        IrisRequest req;
        return memory_getSidebandInfo(req, instId, result, spaceId, address, attrib, request);
    }

    //--- Image Loading interface ---

    IrisErrorCode image_loadFile(IrisRequest &req,
                                 IrisInstId instId,
                                 const std::string &path,
                                 uint64_t rawAddr,
                                 uint64_t rawSpaceId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_loadFile"), instId);
            req.addArg(ISTR("path"), path);
            if (rawAddr != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawAddr"), rawAddr);
            if (rawSpaceId != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawSpaceId"), rawSpaceId);
        }

        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode image_loadFile(IrisInstId instId,
                                 const std::string &path,
                                 uint64_t rawAddr = IRIS_UINT64_MAX,
                                 uint64_t rawSpaceId = IRIS_UINT64_MAX)
    {
        IrisRequest req;
        return image_loadFile(req, instId, path, rawAddr, rawSpaceId);
    }


    IrisErrorCode image_loadData(IrisRequest &req,
                                 IrisInstId instId,
                                 const std::vector<uint64_t> &data,
                                 uint64_t size,
                                 const std::string &path,
                                 uint64_t rawAddr,
                                 uint64_t rawSpaceId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_loadData"), instId);
            req.addArg(ISTR("data"), data);
            req.addArg(ISTR("size"), size);
            if (path != "")
                req.addArg(ISTR("path"), path);
            if (rawAddr != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawAddr"), rawAddr);
            if (rawSpaceId != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawSpaceId"), rawSpaceId);
        }

        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode image_loadData(IrisInstId instId,
                                 const std::vector<uint64_t> &data,
                                 uint64_t size,
                                 const std::string &path = "",
                                 uint64_t rawAddr = IRIS_UINT64_MAX,
                                 uint64_t rawSpaceId = IRIS_UINT64_MAX)
    {
        IrisRequest req;
        return image_loadData(req, instId, data, size, path, rawAddr, rawSpaceId);
    }


    IrisErrorCode image_loadDataPull(IrisRequest &req,
                                     IrisInstId instId,
                                     uint64_t tag,
                                     const std::string &path,
                                     uint64_t rawAddr,
                                     uint64_t rawSpaceId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_loadDataPull"), instId);
            req.addArg(ISTR("tag"), tag);
            if (path != "")
                req.addArg(ISTR("path"), path);
            if (rawAddr != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawAddr"), rawAddr);
            if (rawSpaceId != IRIS_UINT64_MAX)
                req.addArg(ISTR("rawSpaceId"), rawSpaceId);
        }

        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode image_loadDataPull(IrisInstId instId,
                                     uint64_t tag,
                                     const std::string &path = "",
                                     uint64_t rawAddr = IRIS_UINT64_MAX,
                                     uint64_t rawSpaceId = IRIS_UINT64_MAX)
    {
        IrisRequest req;
        return image_loadDataPull(req, instId, tag, path, rawAddr, rawSpaceId);
    }


    IrisErrorCode image_loadDataRead(IrisRequest &req,
                                     IrisInstId instId,
                                     ImageReadResult &result,
                                     uint64_t tag,
                                     uint64_t position,
                                     uint64_t size,
                                     bool end)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_loadDataRead"), instId);
            req.addArg(ISTR("tag"), tag);
            req.addArg(ISTR("position"), position);
            req.addArg(ISTR("size"), size);
            if (end)
                req.addArg(ISTR("end"), end);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result);
        }
        return req.getErrorCode();
    }
    IrisErrorCode image_loadDataRead(IrisInstId instId,
                                     ImageReadResult &result,
                                     uint64_t tag,
                                     uint64_t position,
                                     uint64_t size,
                                     bool end = false)
    {
        IrisRequest req;
        return image_loadDataRead(req, instId, result, tag, position, size, end);
    }

    IrisErrorCode image_getMetaInfoList(IrisRequest &req,
                                        IrisInstId instId,
                                        std::vector<ImageMetaInfo> &result)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_getMetaInfoList"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result);
        }
        return req.getErrorCode();
    }
    IrisErrorCode image_getMetaInfoList(IrisInstId instId, std::vector<ImageMetaInfo> &result)
    {
        IrisRequest req;
        return image_getMetaInfoList(req, instId, result);
    }

    IrisErrorCode image_clearMetaInfoList(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("image_clearMetaInfoList"), instId);
        }

        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode image_clearMetaInfoList(IrisInstId instId)
    {
        IrisRequest req;
        return image_clearMetaInfoList(req, instId);
    }


    //--- Breakpoint interface ---

    IrisErrorCode breakpoint_set_code(IrisRequest &req,
                                      IrisInstId instId,
                                      uint64_t &bptId,
                                      uint64_t address,
                                      uint64_t spaceId,
                                      uint64_t size = 0,
                                      bool dontStop = false,
                                      BreakpointAction *action = NULL,
                                      const IrisValueMap &conditions = IrisValueMap())
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("breakpoint_set"), instId);
            req.addArg(ISTR("type"), "code");
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("spaceId"), spaceId);
            if (size != 0)
                req.addArg(ISTR("size"), size);
            if (dontStop)
                req.addArg(ISTR("dontStop"), true);
            if (action)
                req.addArg(ISTR("action"), *action);
            if (!conditions.empty())
                req.addArg(ISTR("conditions"), conditions);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), bptId);
        }
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_set_code(IrisInstId instId,
                                      uint64_t &bptId,
                                      uint64_t address,
                                      uint64_t spaceId,
                                      uint64_t size = 0,
                                      bool dontStop = false,
                                      BreakpointAction *action = NULL,
                                      const IrisValueMap &conditions = IrisValueMap())
    {
        IrisRequest req;
        return breakpoint_set_code(req, instId, bptId, address, spaceId, size, dontStop, action, conditions);
    }


    IrisErrorCode breakpoint_set_data(IrisRequest &req,
                                      IrisInstId instId,
                                      uint64_t &bptId,
                                      uint64_t address,
                                      uint64_t spaceId,
                                      uint64_t size = 0,
                                      const std::string &rwMode = "",
                                      bool dontStop = false,
                                      BreakpointAction *action = NULL,
                                      const IrisValueMap &conditions = IrisValueMap())
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("breakpoint_set"), instId);
            req.addArg(ISTR("type"), "data");
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("spaceId"), spaceId);
            if (size != 0)
                req.addArg(ISTR("size"), size);
            if (!rwMode.empty())
                req.addArg(ISTR("rwMode"), rwMode);
            if (dontStop)
                req.addArg(ISTR("dontStop"), true);
            if (action)
                req.addArg(ISTR("action"), *action);
            if (!conditions.empty())
                req.addArg(ISTR("conditions"), conditions);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), bptId);
        }
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_set_data(IrisInstId instId,
                                      uint64_t &bptId,
                                      uint64_t address,
                                      uint64_t spaceId,
                                      uint64_t size = 0,
                                      const std::string &rwMode = "",
                                      bool dontStop = false,
                                      BreakpointAction *action = NULL,
                                      const IrisValueMap &conditions = IrisValueMap())
    {
        IrisRequest req;
        return breakpoint_set_data(req, instId, bptId, address, spaceId, size, rwMode, dontStop, action, conditions);
    }


    IrisErrorCode breakpoint_set_register(IrisRequest &req,
                                          IrisInstId instId,
                                          uint64_t &bptId,
                                          uint64_t rscId,
                                          const std::string &rwMode = "",
                                          bool dontStop = false,
                                          BreakpointAction *action = NULL,
                                          const IrisValueMap &conditions = IrisValueMap())
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("breakpoint_set"), instId);
            req.addArg(ISTR("type"), "register");
            req.addArg(ISTR("rscId"), rscId);
            if (!rwMode.empty())
                req.addArg(ISTR("rwMode"), rwMode);
            if (dontStop)
                req.addArg(ISTR("dontStop"), true);
            if (action)
                req.addArg(ISTR("action"), *action);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), bptId);
            if (!conditions.empty())
                req.addArg(ISTR("conditions"), conditions);
        }
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_set_register(IrisInstId instId,
                                          uint64_t &bptId,
                                          uint64_t rscId,
                                          const std::string &rwMode = "",
                                          bool dontStop = false,
                                          BreakpointAction *action = NULL,
                                          const IrisValueMap &conditions = IrisValueMap())
    {
        IrisRequest req;
        return breakpoint_set_register(req, instId, bptId, rscId, rwMode, dontStop, action, conditions);
    }


    IrisErrorCode breakpoint_delete(IrisRequest &req, IrisInstId instId, uint64_t bptId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("breakpoint_delete"), instId);
            req.addArg(ISTR("bptId"), bptId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_delete(IrisInstId instId, uint64_t bptId)
    {
        IrisRequest req;
        return breakpoint_delete(req, instId, bptId);
    }


    IrisErrorCode breakpoint_getList(IrisRequest &req,
                                     IrisInstId instId,
                                     std::vector<iris::BreakpointInfo> &bptInfos,
                                     uint64_t bptId = iris::IRIS_UINT64_MAX)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("breakpoint_getList"), instId);
            if (bptId != iris::IRIS_UINT64_MAX)
                req.addArg(ISTR("bptId"), bptId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), bptInfos);
        }
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_getList(IrisInstId instId,
                                     std::vector<iris::BreakpointInfo> &bptInfos,
                                     uint64_t bptId = iris::IRIS_UINT64_MAX)
    {
        IrisRequest req;
        return breakpoint_getList(req, instId, bptInfos, bptId);
    }

    IrisErrorCode breakpoint_getAdditionalConditions(IrisRequest &req, IrisInstId instId,
                                                     std::vector<BreakpointConditionInfo> &conditions,
                                                     const std::string &type="")
    {
        if (!req.hasRequest())
        {
            req.initRequest("breakpoint_getAdditionalConditions", instId);
            if (type != "")
            {
                req.addArg(ISTR("type"), type);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), conditions);
        }
        return req.getErrorCode();
    }
    IrisErrorCode breakpoint_getAdditionalConditions(IrisInstId instId, std::vector<BreakpointConditionInfo> &conditions,
                                                     const std::string &type="")
    {
        IrisRequest req;
        return breakpoint_getAdditionalConditions(req, instId, conditions, type);
    }


    //--- Simulation interface ---

    IrisErrorCode simulation_getInstantiationParameterInfo(IrisRequest &req, IrisInstId instId,
                                                          std::vector<ResourceInfo> &info_out,
                                                          const std::string &prefix="")
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_getInstantiationParameterInfo", instId);
            if (!prefix.empty())
            {
                req.addArg(ISTR("prefix"), prefix);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), info_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_getInstantiationParameterInfo(IrisInstId instId, std::vector<ResourceInfo> &info_out,
                                                           const std::string &prefix="")
    {
        IrisRequest req;
        return simulation_getInstantiationParameterInfo(req, instId, info_out, prefix);
    }
    IrisErrorCode simulation_getInstantiationParameterInfo(std::vector<ResourceInfo> &info_out, const std::string &prefix="")
    {
        return simulation_getInstantiationParameterInfo(IrisInstIdSimulationEngine, info_out, prefix);
    }

    IrisErrorCode simulation_setInstantiationParameterValues(IrisRequest &req, IrisInstId instId,
                                                            const std::vector<InstantiationParameterValue> &values)
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_setInstantiationParameterValues", instId);
            req.addArg(ISTR("data"), values);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_setInstantiationParameterValues(IrisInstId instId,
                                                            const std::vector<InstantiationParameterValue> &values)
    {
        IrisRequest req;
        return simulation_setInstantiationParameterValues(req, instId, values);
    }
    IrisErrorCode simulation_setInstantiationParameterValues(const std::vector<InstantiationParameterValue> &values)
    {
        return simulation_setInstantiationParameterValues(IrisInstIdSimulationEngine, values);
    }

    IrisErrorCode simulation_instantiate(IrisRequest &req, IrisInstId instId, InstantiationResult &result_out)
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_instantiate", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_instantiate(IrisInstId instId, InstantiationResult &result_out)
    {
        IrisRequest req;
        return simulation_instantiate(req, instId, result_out);
    }
    IrisErrorCode simulation_instantiate(InstantiationResult &result_out)
    {
        return simulation_instantiate(IrisInstIdSimulationEngine, result_out);
    }

    IrisErrorCode simulation_reset(IrisRequest &req, IrisInstId instId, bool allowPartialReset=false)
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_reset", instId);
            if (allowPartialReset)
            {
                req.addArg(ISTR("allowPartialReset"), allowPartialReset);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_reset(IrisInstId instId, bool allowPartialReset=false)
    {
        IrisRequest req;
        return simulation_reset(req, instId, allowPartialReset);
    }
    IrisErrorCode simulation_reset(bool allowPartialReset=false)
    {
        return simulation_reset(IrisInstIdSimulationEngine, allowPartialReset);
    }

    IrisErrorCode simulation_requestShutdown(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_requestShutdown", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_requestShutdown(IrisInstId instId)
    {
        IrisRequest req;
        return simulation_requestShutdown(req, instId);
    }
    IrisErrorCode simulation_requestShutdown()
    {
        return simulation_requestShutdown(IrisInstIdSimulationEngine);
    }

    IrisErrorCode simulation_waitForInstantiation(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest("simulation_waitForInstantiation", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulation_waitForInstantiation(IrisInstId instId)
    {
        IrisRequest req;
        return simulation_waitForInstantiation(req, instId);
    }
    IrisErrorCode simulation_waitForInstantiation()
    {
        return simulation_waitForInstantiation(IrisInstIdSimulationEngine);
    }

    IrisErrorCode simulationTime_run(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("simulationTime_run"), instId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode simulationTime_run(IrisInstId instId)
    {
        IrisRequest req;
        return simulationTime_run(req, instId);
    }
    IrisErrorCode simulationTime_run()
    {
        return simulationTime_run(IrisInstIdSimulationEngine);
    }


    IrisErrorCode simulationTime_stop(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("simulationTime_stop"), instId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode simulationTime_stop(IrisInstId instId)
    {
        IrisRequest req;
        return simulationTime_stop(req, instId);
    }
    IrisErrorCode simulationTime_stop()
    {
        return simulationTime_stop(IrisInstIdSimulationEngine);
    }


    IrisErrorCode simulationTime_get(IrisRequest &req, IrisInstId instId, SimulationTimeObject &time)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("simulationTime_get"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), time);
        }
        return req.getErrorCode();
    }
    IrisErrorCode simulationTime_get(IrisInstId instId, SimulationTimeObject &time)
    {
        IrisRequest req;
        return simulationTime_get(req, instId, time);
    }
    IrisErrorCode simulationTime_get(SimulationTimeObject &time)
    {
        return simulationTime_get(IrisInstIdSimulationEngine, time);
    }


    //--- Per-instance execution control ---

    IrisErrorCode perInstanceExecution_setState(IrisRequest &req, IrisInstId instId, bool enable)
    {
        if (!req.hasRequest())
        {
            req.initRequest("perInstanceExecution_setState", instId);
            req.addArg(ISTR("enable"), enable);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode perInstanceExecution_setState(IrisInstId instId, bool enable)
    {
        IrisRequest req;
        return perInstanceExecution_setState(req, instId, enable);
    }


    IrisErrorCode perInstanceExecution_getState(IrisRequest &req, IrisInstId instId, bool &enable)
    {
        if (!req.hasRequest())
        {
            req.initRequest("perInstanceExecution_getState", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            std::map<std::string, bool> result;
            persist(req.getResultReader(), result);

            assert(result.size() == 1 && result.count("enable") && "Invalid result for per-instance execution state get");
            enable = result["enable"];
        }
        return req.getErrorCode();
    }
    IrisErrorCode perInstanceExecution_getState(IrisInstId instId, bool &enable)
    {
        IrisRequest req;
        return perInstanceExecution_getState(req, instId, enable);
    }


    IrisErrorCode perInstanceExecution_setStateAll(IrisRequest &req, std::vector<uint64_t> &instanceSet,
                                                   bool enable)
    {
        if (!req.hasRequest())
        {
            req.initRequest("perInstanceExecution_setStateAll", /*dstInstId*/ 0);
            req.addArg(ISTR("instanceSet"), instanceSet);
            req.addArg(ISTR("enable"), enable);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode perInstanceExecution_setStateAll(std::vector<uint64_t> &instanceSet, bool enable)
    {
        IrisRequest req;
        return perInstanceExecution_setStateAll(req, instanceSet, enable);
    }


    IrisErrorCode perInstanceExecution_getStateAll(IrisRequest &req, std::vector<uint64_t> &enabledSet,
                                                   std::vector<uint64_t> &disabledSet)
    {
        if (!req.hasRequest())
        {
            req.initRequest("perInstanceExecution_getStateAll", /*dstInstId*/ 0);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            IrisU64JsonReader::Object o = req.getResultReader().openObject();
            o.member(ISTR("enabledSet"), enabledSet);
            o.member(ISTR("disabledSet"), disabledSet);
        }
        return req.getErrorCode();
    }
    IrisErrorCode perInstanceExecution_getStateAll(std::vector<uint64_t> &enabledSet, std::vector<uint64_t> &disabledSet)
    {
        IrisRequest req;
        return perInstanceExecution_getStateAll(req, enabledSet, disabledSet);
    }


    //--- Event interface ---

    IrisErrorCode event_getEventSources(IrisRequest &req,
                                        IrisInstId instId,
                                        std::vector<iris::EventSourceInfo> &evSrcInfos)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("event_getEventSources"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), evSrcInfos);
        }
        return req.getErrorCode();
    }
    IrisErrorCode event_getEventSources(IrisInstId instId,
                                        std::vector<iris::EventSourceInfo> &evSrcInfos)
    {
        IrisRequest req;
        return event_getEventSources(req, instId, evSrcInfos);
    }


    IrisErrorCode event_getEventSource(IrisRequest &req, IrisInstId instId,
                                       EventSourceInfo &evSrcInfo, const std::string &name)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("event_getEventSource"), instId);
            req.addArg(ISTR("name"), name);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), evSrcInfo);
        }
        return req.getErrorCode();
    }
    IrisErrorCode event_getEventSource(IrisInstId instId, EventSourceInfo &evSrcInfo,
                                       const std::string &name)
    {
        IrisRequest req;
        return event_getEventSource(req, instId, evSrcInfo, name);
    }


    IrisErrorCode eventStream_create(IrisRequest &req,
                                     IrisInstId instId,
                                     uint64_t &esId,
                                     uint64_t evSrcId,
                                     uint64_t ecInstId,
                                     const std::vector<std::string> *fields = NULL,
                                     const std::string ecFunc = "",
                                     bool counter = false,
                                     uint64_t startVal = 0,
                                     const iris::EventCounterMode *counterMode = NULL,
                                     bool disable = false,
                                     bool ringBuffer = false,
                                     bool syncEc = false)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("eventStream_create"), instId);
            req.addArg(ISTR("evSrcId"), evSrcId);
            req.addArg(ISTR("ecInstId"), ecInstId);
            if (fields)
                req.addArg(ISTR("fields"), *fields);
            if (!ecFunc.empty())
                req.addArg(ISTR("ecFunc"), ecFunc);
            if (counter)
                req.addArg(ISTR("counter"), counter);
            if (startVal)
                req.addArg(ISTR("startVal"), startVal);
            if (counterMode)
                req.addArg(ISTR("counterMode"), *counterMode);
            if (disable)
                req.addArg(ISTR("disable"), disable);
            if (ringBuffer)
                req.addArg(ISTR("ringBuffer"), ringBuffer);
            if (syncEc)
                req.addArg(ISTR("syncEc"), syncEc);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), esId);
        }
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_create(IrisInstId instId,
                                     uint64_t &esId,
                                     uint64_t evSrcId,
                                     uint64_t ecInstId,
                                     const std::vector<std::string> *fields = NULL,
                                     const std::string ecFunc = "",
                                     bool counter = false,
                                     uint64_t startVal = 0,
                                     const iris::EventCounterMode *counterMode = NULL,
                                     bool disable = false,
                                     bool ringBuffer = false,
                                     bool syncEc = false)
    {
        IrisRequest req;
        return eventStream_create(req, instId, esId, evSrcId, ecInstId, fields, ecFunc, counter, startVal,
                                 counterMode, disable, ringBuffer, syncEc);
    }


    IrisErrorCode eventStream_destroy(IrisRequest &req, IrisInstId instId, uint64_t esId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("eventStream_destroy"), instId);
            req.addArg(ISTR("esId"), esId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_destroy(IrisInstId instId, uint64_t esId)
    {
        IrisRequest req;
        return eventStream_destroy(req, instId, esId);
    }


    IrisErrorCode eventStream_enable(IrisRequest &req, IrisInstId instId, uint64_t esId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("eventStream_enable"), instId);
            req.addArg(ISTR("esId"), esId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_enable(IrisInstId instId, uint64_t esId)
    {
        IrisRequest req;
        return eventStream_enable(req, instId, esId);
    }


    IrisErrorCode eventStream_disable(IrisRequest &req, IrisInstId instId, uint64_t esId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("eventStream_disable"), instId);
            req.addArg(ISTR("esId"), esId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_disable(IrisInstId instId, uint64_t esId)
    {
        IrisRequest req;
        return eventStream_disable(req, instId, esId);
    }


    IrisErrorCode eventStream_getCounter(IrisRequest &req, IrisInstId instId, uint64_t &counterVal, uint64_t esId)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("eventStream_getCounter"), instId);
            req.addArg(ISTR("esId"), esId);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), counterVal);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_getCounter(IrisInstId instId, uint64_t &counterVal, uint64_t esId)
    {
        IrisRequest req;
        return eventStream_getCounter(req, instId, counterVal, esId);
    }


    IrisErrorCode eventStream_setTraceRanges(IrisRequest &req, IrisInstId instId, uint64_t esId,
                                             const std::vector<uint64_t> &ranges,
                                             const std::string &aspect)
    {
        if (!req.hasRequest())
        {
            req.initRequest("eventStream_setTraceRanges", instId);
            req.addArg(ISTR("esId"), esId);
            req.addArg(ISTR("ranges"), ranges);
            req.addArg(ISTR("aspect"), aspect);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_setTraceRanges(IrisInstId instId, uint64_t esId,
                                             const std::vector<uint64_t> &ranges,
                                             const std::string &aspect)
    {
        IrisRequest req;
        return eventStream_setTraceRanges(req, instId, esId, ranges, aspect);
    }


    IrisErrorCode eventStream_getState(IrisRequest &req, IrisInstId instId, uint64_t esId,
                                       std::map<std::string, IrisValue> fields)
    {
        if (!req.hasRequest())
        {
            req.initRequest("eventStream_getState", instId);
            req.addArg(ISTR("esId"), esId);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), fields);
        return req.getErrorCode();
    }
    IrisErrorCode eventStream_getState(IrisInstId instId, uint64_t esId, std::map<std::string, IrisValue> fields)
    {
        IrisRequest req;
        return eventStream_getState(req, instId, esId, fields);
    }


    /// The event callback function.
    /// 'ecFOO' is just used as a placeholder for the real function name, which is specified by 'ecFunc'
    IrisErrorCode ec_FOO(IrisRequest &req, const std::string &ecFunc, IrisInstId ecInstId, uint64_t esId,
                         const std::map<std::string, IrisValue> &fields,
                         uint64_t sInstId, uint64_t time, bool syncEc = false)
    {
        if (!req.hasRequest())
        {
            if (syncEc)
            {
                req.initRequest(ecFunc, ecInstId);
                req.addArg(ISTR("syncEc"), syncEc);
            }
            else
            {
                // Notification needs no response
                req.initNotification(ecFunc, ecInstId);
            }

            req.addArg(ISTR("esId"), esId);
            req.addArg(ISTR("fields"), fields);
            req.addArg(ISTR("time"), time);
            req.addArg(ISTR("sInstId"), sInstId);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode ec_FOO(const std::string &ecFunc, IrisInstId ecInstId, uint64_t esId,
                         const std::map<std::string, IrisValue> &fields,
                         uint64_t sInstId, uint64_t time, bool syncEc = false)
    {
        IrisRequest req;
        return ec_FOO(req, ecFunc, ecInstId, esId, fields, sInstId, time, syncEc);
    }


    ///--- Step interface ---

    IrisErrorCode step_setup(IrisRequest &req, IrisInstId instId, uint64_t steps,
                             const std::string &unit)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("step_setup"), instId);
            req.addArg(ISTR("steps"), steps);
            req.addArg(ISTR("unit"), unit);
        }
        callAndPerhapsWaitForResponse(req);
        return req.getErrorCode();
    }
    IrisErrorCode step_setup(IrisInstId instId, uint64_t steps, const std::string &unit)
    {
        IrisRequest req;
        return step_setup(req, instId, steps, unit);
    }


    IrisErrorCode step_getRemainingSteps(IrisRequest &req, IrisInstId instId, uint64_t &steps,
                                         const std::string &unit)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("step_getRemainingSteps"), instId);
            req.addArg(ISTR("unit"), unit);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), steps);
        return req.getErrorCode();
    }
    IrisErrorCode step_getRemainingSteps(IrisInstId instId, uint64_t &steps, const std::string &unit)
    {
        IrisRequest req;
        return step_getRemainingSteps(req, instId, steps, unit);
    }


    IrisErrorCode step_getStepCounterValue(IrisRequest &req, IrisInstId instId, uint64_t &count,
                                           const std::string &unit)
    {
        if (!req.hasRequest())
        {
            req.initRequest("step_getStepCounterValue", instId);
            req.addArg(ISTR("unit"), unit);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), count);
        return req.getErrorCode();
    }
    IrisErrorCode step_getStepCounterValue(IrisInstId instId, uint64_t &count, const std::string &unit)
    {
        IrisRequest req;
        return step_getStepCounterValue(req, instId, count, unit);
    }

    ///--- Disassembler interface ---

    IrisErrorCode disassembler_getModes(IrisRequest &req, IrisInstId instId, std::vector<DisassemblyMode> &disassemblyModes)
    {
        if (!req.hasRequest())
        {
            req.initRequest("disassembler_getModes", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), disassemblyModes);
        return req.getErrorCode();
    }
    IrisErrorCode disassembler_getModes(IrisInstId instId, std::vector<DisassemblyMode> &disassemblyModes)
    {
        IrisRequest req;
        return disassembler_getModes(req, instId, disassemblyModes);
    }

    IrisErrorCode disassembler_getCurrentMode(IrisRequest &req, IrisInstId instId, std::string &currentMode)
    {
        if (!req.hasRequest())
        {
            req.initRequest("disassembler_getCurrentMode", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), currentMode);
        return req.getErrorCode();
    }
    IrisErrorCode disassembler_getCurrentMode(IrisInstId instId, std::string &currentMode)
    {
         IrisRequest req;
         return disassembler_getCurrentMode(req, instId, currentMode);
    }

    IrisErrorCode disassembler_getDisassembly(IrisRequest &req, IrisInstId instId,
                                              std::vector<DisassemblyLine> &disassemblyLineOut,
                                              uint64_t spaceId, uint64_t address,
                                              std::string mode, uint64_t count, uint64_t maxAddr,
                                              AttributeValueMap &attrib)
    {
        if (!req.hasRequest())
        {
            req.initRequest("disassembler_getDisassembly", instId);
            req.addArg(ISTR("spaceId"), spaceId);
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("mode"), mode);
            req.addArg(ISTR("count"), count);
            if (maxAddr != IRIS_UINT64_MAX)
            {
                req.addArg(ISTR("maxAddr"), maxAddr);
            }
            if (!attrib.empty())
            {
                req.addArg(ISTR("attrib"), attrib);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), disassemblyLineOut);
        return req.getErrorCode();
    }
    IrisErrorCode disassembler_getDisassembly(IrisInstId instId, std::vector<DisassemblyLine> &disassemblyLineOut,
                                              uint64_t spaceId, uint64_t address, std::string mode,
                                              uint64_t count, uint64_t maxAddr, AttributeValueMap &attrib)
    {
        IrisRequest req;
        return disassembler_getDisassembly(req, instId, disassemblyLineOut, spaceId, address, mode, count, maxAddr, attrib);
    }
    IrisErrorCode disassembler_getDisassembly(IrisInstId instId, std::vector<DisassemblyLine> &disassemblyLineOut,
                                              uint64_t spaceId, uint64_t address, std::string mode,
                                              uint64_t count, uint64_t maxAddr)
    {
        IrisRequest req;
        AttributeValueMap attrib;
        return disassembler_getDisassembly(req, instId, disassemblyLineOut, spaceId, address, mode, count, maxAddr, attrib);
    }
    IrisErrorCode disassembler_getDisassembly(IrisInstId instId, std::vector<DisassemblyLine> &disassemblyLineOut,
                                              uint64_t spaceId, uint64_t address, std::string mode, uint64_t count)
    {
        IrisRequest req;
        AttributeValueMap attrib;
        uint64_t maxAddr = IRIS_UINT64_MAX;
        return disassembler_getDisassembly(req, instId, disassemblyLineOut, spaceId, address, mode, count, maxAddr, attrib);
    }


    IrisErrorCode disassembler_disassembleOpcode(IrisRequest &req, IrisInstId instId, DisassemblyLine &disassemblyLineOut,
                                                 std::vector<uint64_t> opcode, uint64_t address, std::string mode,
                                                 DisassembleContext &context)
    {
        if (!req.hasRequest())
        {
            req.initRequest("disassembler_disassembleOpcode", instId);
            req.addArg(ISTR("opcode"), opcode);
            req.addArg(ISTR("address"), address);
            req.addArg(ISTR("mode"), mode);
            if (!context.empty())
            {
                req.addArg(ISTR("context"), context);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), disassemblyLineOut);
        return req.getErrorCode();
    }
    IrisErrorCode disassembler_disassembleOpcode(IrisInstId instId, DisassemblyLine &disassemblyLineOut,
                                                std::vector<uint64_t> opcode, uint64_t address,
                                                std::string mode, DisassembleContext &context)
    {
        IrisRequest req;
        return disassembler_disassembleOpcode(req, instId, disassemblyLineOut, opcode, address, mode, context);
    }
    IrisErrorCode disassembler_disassembleOpcode(IrisInstId instId, DisassemblyLine &disassemblyLineOut,
                                                std::vector<uint64_t> opcode, uint64_t address, std::string mode)
    {
        IrisRequest req;
        DisassembleContext context;
        return disassembler_disassembleOpcode(req, instId, disassemblyLineOut, opcode, address, mode, context);
    }

    ///--- table interface ---

    IrisErrorCode table_getList(IrisRequest &req, IrisInstId instId, TableInfoList &tableList)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("table_getList"), instId);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), tableList);
        return req.getErrorCode();
    }
    IrisErrorCode table_getList(IrisInstId instId, TableInfoList &tableList)
    {
        IrisRequest req;
        return table_getList(req, instId, tableList);
    }


    IrisErrorCode table_read(IrisRequest &req, IrisInstId instId, TableReadResult &result,
                             uint64_t tableId, uint64_t index, uint64_t count = 1)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("table_read"), instId);
            req.addArg(ISTR("tableId"), tableId);
            req.addArg(ISTR("index"), index);
            if (count != 1)
                req.addArg(ISTR("count"), count);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), result);
        return req.getErrorCode();
    }
    IrisErrorCode table_read(IrisInstId instId, TableReadResult &result, uint64_t tableId,
                             uint64_t index, uint64_t count = 1)
    {
        IrisRequest req;
        return table_read(req, instId, result, tableId, index, count);
    }


    IrisErrorCode table_write(IrisRequest &req, IrisInstId instId, TableWriteResult &result,
                              uint64_t tableId, const TableRecords &records)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("table_write"), instId);
            req.addArg(ISTR("tableId"), tableId);
            req.addArg(ISTR("records"), records);
        }
        if (callAndPerhapsWaitForResponse(req))
            persist(req.getResultReader(), result);
        return req.getErrorCode();
    }
    IrisErrorCode table_write(IrisInstId instId, TableWriteResult &result, uint64_t tableId,
                              const TableRecords &records)
    {
        IrisRequest req;
        return table_write(req, instId, result, tableId, records);
    }

    IrisErrorCode tcpServer_start(IrisRequest &req, uint16_t &actualPort, uint16_t minPort, uint16_t maxPort)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("tcpServer_start"), IrisInstIdGlobalInstance);
            req.addArg(ISTR("minPort"), uint64_t(minPort));
            req.addArg(ISTR("maxPort"), uint64_t(maxPort));
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            uint64_t port64 = 0;
            persist(req.getResultReader(), port64);
            actualPort = uint16_t(port64);
        }
        return req.getErrorCode();
    }
    IrisErrorCode tcpServer_start(uint16_t &actualPort, uint16_t minPort, uint16_t maxPort)
    {
        IrisRequest req;
        return tcpServer_start(req, actualPort, minPort, maxPort);
    }

    IrisErrorCode tcpServer_stop(IrisRequest &req)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("tcpServer_stop"), IrisInstIdGlobalInstance);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No return value
        }
        return req.getErrorCode();
    }
    IrisErrorCode tcpServer_stop()
    {
        IrisRequest req;
        return tcpServer_stop(req);
    }

    IrisErrorCode tcpServer_getPort(IrisRequest &req, uint16_t &port_out)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("tcpServer_getPort"), IrisInstIdGlobalInstance);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            uint64_t port64 = 0;
            persist(req.getResultReader(), port64);
            port_out = uint16_t(port64);
        }
        return req.getErrorCode();
    }
    IrisErrorCode tcpServer_getPort(uint16_t &port_out)
    {
        IrisRequest req;
        return tcpServer_getPort(req, port_out);
    }

    IrisErrorCode plugin_load(IrisRequest &req, std::vector<InstanceInfo> &instance_info_out, const std::string &path)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("plugin_load"), IrisInstIdGlobalInstance);
            req.addArg(ISTR("path"), path);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), instance_info_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode plugin_load(std::vector<InstanceInfo> &instance_info_out, const std::string &path)
    {
        IrisRequest req;
        return plugin_load(req, instance_info_out, path);
    }

    IrisErrorCode service_connect(IrisRequest &req, const std::string &hostname, uint16_t port)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("service_connect"), IrisInstIdGlobalInstance);
            req.addArg(ISTR("hostname"), hostname);
            req.addArg(ISTR("port"), uint64_t(port));
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // no return value
        }
        return req.getErrorCode();
    }

    IrisErrorCode service_connect(const std::string &hostname, uint16_t port)
    {
        IrisRequest req;
        return service_connect(req, hostname, port);
    }

    IrisErrorCode service_disconnect(IrisRequest &req, const std::string &hostname, uint16_t port)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("service_disconnect"), IrisInstIdGlobalInstance);
            req.addArg(ISTR("hostname"), hostname);
            req.addArg(ISTR("port"), uint64_t(port));
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // no return value
        }
        return req.getErrorCode();
    }

    IrisErrorCode service_disconnect(const std::string &hostname, uint16_t port)
    {
        IrisRequest req;
        return service_disconnect(req, hostname, port);
    }

    IrisErrorCode plugin_getInstantiationParameterInfo(IrisRequest &req, IrisInstId instId,
                                                       std::vector<ResourceInfo> &parameter_info_out)
    {
        if (!req.hasRequest())
        {
            req.initRequest("plugin_getInstantiationParameterInfo", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), parameter_info_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode plugin_getInstantiationParameterInfo(IrisInstId instId,
                                                       std::vector<ResourceInfo> &parameter_info_out)
    {
        IrisRequest req;
        return plugin_getInstantiationParameterInfo(req, instId, parameter_info_out);
    }

    IrisErrorCode plugin_instantiate(IrisRequest &req, IrisInstId instId,
                                     InstantiationResult &result_out,
                                     const std::vector<InstantiationParameterValue> &paramValues,
                                     const std::string &instName)
    {
        if (!req.hasRequest())
        {
            req.initRequest(ISTR("plugin_instantiate"), instId);
            req.addArg(ISTR("instName"), instName);
            req.addArg(ISTR("paramValues"), paramValues);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), result_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode plugin_instantiate(IrisInstId instId,
                                     InstantiationResult &result_out,
                                     const std::vector<InstantiationParameterValue> &paramValues)
    {
        IrisRequest req;
        req.initRequest(ISTR("plugin_instantiate"), instId);
        req.addArg(ISTR("paramValues"), paramValues);

        return plugin_instantiate(req, instId, result_out, paramValues, "");
    }
    IrisErrorCode plugin_instantiate(IrisInstId instId,
                                     InstantiationResult &result_out,
                                     const std::vector<InstantiationParameterValue> &paramValues,
                                     const std::string &instName)
    {
        IrisRequest req;

        return plugin_instantiate(req, instId, result_out, paramValues, instName);
    }

    IrisErrorCode debuggableState_setRequest(IrisRequest &req, IrisInstId instId, bool debug_state_request=true)
    {
        if (!req.hasRequest())
        {
            req.initRequest("debuggableState_setRequest", instId);
            if (!debug_state_request)
            {
                req.addArg(ISTR("request"), debug_state_request);
            }
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode debuggableState_setRequest(IrisInstId instId, bool debug_state_request=true)
    {
        IrisRequest req;
        return debuggableState_setRequest(req, instId, debug_state_request);
    }

    IrisErrorCode debuggableState_getAcknowledge(IrisRequest &req, IrisInstId instId, bool &debuggable_state_acknowledged)
    {
        if (!req.hasRequest())
        {
            req.initRequest("debuggableState_getAcknowledge", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), debuggable_state_acknowledged);
        }
        return req.getErrorCode();
    }
    IrisErrorCode debuggableState_getAcknowledge(IrisInstId instId, bool &debuggable_state_acknowledged)
    {
        IrisRequest req;
        return debuggableState_getAcknowledge(req, instId, debuggable_state_acknowledged);
    }

    IrisErrorCode semihosting_provideInputData(IrisRequest &req, IrisInstId instId,
                                               uint64_t fDes, const std::vector<uint64_t> &data, uint64_t size)
    {
        if (!req.hasRequest())
        {
            req.initRequest("semihosting_provideInputData", instId);
            req.addArg(ISTR("fDes"), fDes);
            req.addArg(ISTR("data"), data);
            req.addArg(ISTR("size"), size);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }

        return req.getErrorCode();
    }

    IrisErrorCode semihosting_provideInputData(IrisInstId instId, uint64_t fDes, const std::vector<uint64_t> &data,
                                               uint64_t size)
    {
        IrisRequest req;
        return semihosting_provideInputData(req, instId, fDes, data, size);
    }

    IrisErrorCode semihosting_return(IrisRequest &req, IrisInstId instId, uint64_t retval)
    {
        if (!req.hasRequest())
        {
            req.initRequest("semihosting_return", instId);
            req.addArg(ISTR("retval"), retval);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }

        return req.getErrorCode();
    }
    IrisErrorCode semihosting_return(IrisInstId instId, uint64_t retval)
    {
        IrisRequest req;
        return semihosting_return(req, instId, retval);
    }

    IrisErrorCode semihosting_notImplemented(IrisRequest &req, IrisInstId instId)
    {
        if (!req.hasRequest())
        {
            req.initRequest("semihosting_notImplemented", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }

        return req.getErrorCode();
    }
    IrisErrorCode semihosting_notImplemented(IrisInstId instId)
    {
        IrisRequest req;
        return semihosting_notImplemented(req, instId);
    }

    IrisErrorCode syncLevel_request(IrisRequest &req, IrisInstId instId, uint64_t syncLevel)
    {
        if (!req.hasRequest())
        {
            req.initRequest("syncLevel_request", instId);
            req.addArg(ISTR("syncLevel"), syncLevel);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode syncLevel_request(IrisInstId instId, uint64_t syncLevel)
    {
        IrisRequest req;
        return syncLevel_request(req, instId, syncLevel);
    }

    IrisErrorCode syncLevel_release(IrisRequest &req, IrisInstId instId, uint64_t syncLevel)
    {
        if (!req.hasRequest())
        {
            req.initRequest("syncLevel_release", instId);
            req.addArg(ISTR("syncLevel"), syncLevel);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            // No result
        }
        return req.getErrorCode();
    }
    IrisErrorCode syncLevel_release(IrisInstId instId, uint64_t syncLevel)
    {
        IrisRequest req;
        return syncLevel_release(req, instId, syncLevel);
    }

    IrisErrorCode syncLevel_get(IrisRequest &req, IrisInstId instId, SyncLevelState &state_out)
    {
        if (!req.hasRequest())
        {
            req.initRequest("syncLevel_get", instId);
        }
        if (callAndPerhapsWaitForResponse(req))
        {
            persist(req.getResultReader(), state_out);
        }
        return req.getErrorCode();
    }
    IrisErrorCode syncLevel_get(IrisInstId instId, SyncLevelState &state_out)
    {
        IrisRequest req;
        return syncLevel_get(req, instId, state_out);
    }

    /// Call Iris function and wait for response.
    /// Returns true iff we got a non-error result.
    bool callAndPerhapsWaitForResponse(IrisRequest &req)
    {
        return request_manager->callAndPerhapsWaitForResponse(req, throwOnError);
    }

    /// Return the error message contained in the most recently received error response.
    std::string getLastErrorMessage() const
    {
        return request_manager->getLastErrorMessage();
    }

private:

    IrisCppAdapter(const IrisCppAdapter& other);


    RequestManager *request_manager;

    /// Error behavior: Throw exception or return error code.
    bool throwOnError;
}; // class IrisCppAdapter

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisCppAdapter_h

