/*!
 *  \file   IrisDelegate.h
 *  \brief  Generic C++ delegates
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 * A delegate class that lets you call a member function on an instance of an arbitrary class
 * or a global function.
 * IrisDelegate objects are small, containing only two pointers, so they should be
 * passed by value.
 *
 * To use IrisDelegate write a typedef to create a new type for your delegate
 * defining the signature of the delegate:
 *
 *      class MyClass
 *      {
 *      private:
 *          unsigned N;
 *
 *      public:
 *          MyClass(unsigned N_)
 *              : N(N_)
 *          {
 *          }
 *
 *          IrisErrorCode addN(unsigned &out, unsigned n)
 *          {
 *              out = N + n;
 *              return E_ok;
 *          }
 *      };
 *
 *      // Iris delegates always return IrisErrorCode but the argument types are given as
 *      // template arguments.
 *      typedef IrisDelegate<unsigned &, unsigned> MyDelegate;
 *
 * IrisDelegate instances can be default constructed. This will create an empty delegate
 * which always returns E_not_implemented.
 *
 *      MyDelegate d;
 *      MyClass instance(5);
 *
 *      // To create a delegate object call the static make method on the delegate class
 *      // passing it the instance to call the method on. The type of the instance and a
 *      // method pointer must be given as template parameters:
 *      d = MyDelegate::make<MyClass, &MyClass::addN>(&instance);
 *      unsigned out;
 *      d(out, 4); // --> E_ok
 *                 //     out == 9
 */

#ifndef ARM_INCLUDE_IrisDelegate_h
#define ARM_INCLUDE_IrisDelegate_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisErrorCode.h"

#include <cstddef>

NAMESPACE_IRIS_START

/// A generic specialisable delegate class
template <typename... ARGS> class IrisDelegate
{
private:
    /// Context pointer to give the method_caller function context.
    /// This will be set to an instance pointer for calling member
    /// functions, and NULL otherwise.
    /// Stored as void* so we don't have to template on its type.
    void *context;

    /// The delegate call is performed by calling the function pointer method_caller
    /// which is always set to one of the three functions below:
    ///     * callMemberFunction()
    ///     * callGlobalFunction()
    ///     * callEmpty()
    typedef IrisErrorCode(MethodCaller)(void *, ARGS...);
    MethodCaller *method_caller;

    /// This is the implementation used for method_caller in non-empty delegates
    /// that call a member function on an instance.
    /// It is templated on the real type of instance and the method
    /// pointer to call with it. When we create a delegate we instantiate
    /// a version of this static method with the particular class/method
    /// we want to delegate to.
    template <typename T, IrisErrorCode(T::*FUNC)(ARGS...)>
    IrisErrorCode static callMemberFunction(void *instance, ARGS... args)
    {
        T *t_instance = static_cast<T*>(instance);
        return (t_instance->*FUNC)(args...);
    }

    /// This is the implementation used for method_caller in non-empty delegates
    /// that call a global funcion.
    template <IrisErrorCode(*FUNC)(ARGS...)>
    IrisErrorCode static callGlobalFunction(void * /*context*/, ARGS... args)
    {
        return (*FUNC)(args...);
    }

    /// This is the implementation used for method_caller in empty delegates.
    /// It works like callMemberFunction() except it ignores its arguments and just
    /// returns E_not_implemented.
    IrisErrorCode static callEmpty(void * /*context*/, ARGS... /*args*/)
    {
        return E_not_implemented;
    }

    // This constructor shouldn't be used directly.
    // Use the static function make() instead
    IrisDelegate(void *context_, MethodCaller *method_caller_)
        : context(context_)
        , method_caller(method_caller_)
    {
        // Nothing else to initialise
    }

public:
    /// Construct an empty delegate.
    IrisDelegate()
        : context(NULL)
        , method_caller(&callEmpty)
    {
        // Nothing else to initialise
    }

    /// Create a delegate object that calls a member function
    template <typename T, IrisErrorCode(T::*FUNC)(ARGS...)>
    static IrisDelegate<ARGS...> make(T* instance)
    {
        return IrisDelegate<ARGS...>(instance, &callMemberFunction<T, FUNC>);
    }

    /// Create a delegate object that calls a non-member function
    template <IrisErrorCode(*FUNC)(ARGS...)>
    static IrisDelegate<ARGS...> make()
    {
        return IrisDelegate<ARGS...>(NULL, &callGlobalFunction<FUNC>);
    }

    /// Call the delegate
    IrisErrorCode operator() (ARGS... args)
    {
        return (*method_caller)(context, args...);
    }

    bool operator== (const IrisDelegate<ARGS...> &other) const
    {
        return (context == other.context) && (method_caller == other.method_caller);
    }

    bool operator!= (const IrisDelegate<ARGS...> &other) const
    {
        return !(*this == other);
    }

    /// Returns true if this delegate is empty.
    /// Emtpy delegates throw an exception when called.
    bool empty() const
    {
        return method_caller == &callEmpty;
    }
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisDelegate_h
