/*!
 *  \file   IrisErrorException.h
 *  \brief  Error exceptions used in the Iris utilities.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Error exceptions used in the Iris utilities.
 *  These exceptions are just used inside the utilities and internal infrastructure.
 *  Exceptions must never cross the IrisInterface.
 */

#ifndef ARM_INCLUDE_IrisErrorException_h
#define ARM_INCLUDE_IrisErrorException_h

#include "iris/detail/IrisCommon.h"
#include "IrisErrorCode.h"
#include "IrisUtils.h"
#include "IrisInterface.h" // uint64_t
#include "IrisU64StringConstant.h"
#include <exception>
#include <string.h>
#include <cassert>
#include <iostream>


NAMESPACE_IRIS_START


    /// Base class for all Iris error exceptions.
    /// This variant just takes an error code.
    class IrisErrorException: public std::exception
    {
    public:
        /// Constructor.
        IrisErrorException(IrisErrorCode errorCode_):errorCode(errorCode_), whatBuffer(0) {}

        /// Destructor.
        virtual ~IrisErrorException() throw() { delete[] whatBuffer; }

        /// Get Iris error code.
        IrisErrorCode getErrorCode() const { return errorCode; }
        
        /// Set error code.
        /// This is used by code which can refine a generic error (e.g. E_type_mismatch)
        /// into a more specific one (E_argument_type_mismatch).
        void setErrorCode(IrisErrorCode errorCode_) { errorCode = errorCode_; }

        /// Get error code string.
        std::string getErrorCodeString() const { return irisErrorCodeString(errorCode); }

        /// Get message associated with this error.
        /// This usually does not contain the error code.
        virtual std::string getMessage() const { return std::string(); }

        /// Set message associated with this error.
        virtual void setMessage(const std::string & /*message_*/) { assert(!"setMessage() not supported by this error class"); }
        
        /// Get context data associated with this error.
        /// This function either returns an empty string (if no context data 
        /// is present) or returns a JSON object (in textual JSON).
        /// Examples: 
        /// For E_argument_type_mismatch: {name: "<ARGUMENT_NAME>"}
        virtual std::string getData() const { return std::string(); }
        
        /// Get string which contains errorCode, message and data in a useful form.
        /// This function allocates memory which is owned and freed by this object.
        virtual const char* what() const throw() IRIS_OVERRIDE
        {
            if (!whatBuffer)
            {
                std::string str = std::string(type()) + "(errorCode=" + getErrorCodeString();
                std::string message = getMessage();
                if (!message.empty())
                    str += ", message=" + message;
                std::string data = getData();
                if (!data.empty())
                    str += ", data=" + data;
                str += ")";
                char *buffer = new char[str.length() + 1];
                memcpy(buffer, str.c_str(), str.length() + 1);
                const_cast<IrisErrorException*>(this)->whatBuffer = buffer;
            }
            return whatBuffer;
        }

        /// Classname of this exception.
        virtual const char *type() const { return "IrisErrorException"; }

    protected:
        IrisErrorCode errorCode;
        char *whatBuffer;
    };


    /// This variant takes a constant and static C string.
    class IrisErrorExceptionCString: public IrisErrorException
    {
    public:
        /// Constructor.
        /// message: Does not need to contain details of the errorCode.
        /// The message pointer must be a constant string and must be valid indefinitely. This object does not take a copy of it.
        IrisErrorExceptionCString(IrisErrorCode errorCode_, const char *message_):IrisErrorException(errorCode_), message(message_) {}

        /// Destructor.
        virtual ~IrisErrorExceptionCString() throw() {}

        /// Get message associated with this error.
        /// This usually does not contain the error code.
        virtual std::string getMessage() const IRIS_OVERRIDE { return message; }

        /// Classname of this exception.
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorExceptionCString"; }

    protected:
        const char *message;
    };


    /// This variant takes a std::string.
    class IrisErrorExceptionString: public IrisErrorException
    {
    public:
        /// Constructor.
        /// message: Does not need to contain details of the errorCode.
        IrisErrorExceptionString(IrisErrorCode errorCode_, const std::string &message_, const std::string &data_ = std::string()):IrisErrorException(errorCode_), message(message_), data(data_) {}

        /// Destructor.
        virtual ~IrisErrorExceptionString() throw() {}

        /// Get message associated with this error.
        virtual std::string getMessage() const IRIS_OVERRIDE { return message; }

        /// Set message associated with this error.
        virtual void setMessage(const std::string &message_) IRIS_OVERRIDE { message = message_; }

        /// Get context data associated with this error.
        virtual std::string getData() const IRIS_OVERRIDE { return data; }

        /// Classname of this exception.
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorExceptionString"; }

        /// Set context data.
        void setData(const std::string &data_) { data = data_; }
        
    protected:  
        std::string message;
        std::string data; // Optional: Some exceptions have context data (e.g, argument name, parameter name).
    };


    /// U64JSON encoding error.
    class IrisErrorU64JsonEncodingError: public IrisErrorExceptionString
    {
    public:
        IrisErrorU64JsonEncodingError(const std::string &message_ = std::string()): IrisErrorExceptionString(E_u64json_encoding_error, message_) {}
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorU64JsonEncodingError"; }
    };


    /// JSON parse error.
    class IrisErrorJsonParseError: public IrisErrorExceptionString
    {
    public:
        IrisErrorJsonParseError(const std::string &message_ = std::string()): IrisErrorExceptionString(E_json_parse_error, message_) {}
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorJsonParseError"; }
    };


    /// Type mismatch when getting data from an object.
    /// For E_type_mismatch, E_argument_type_mismatch.
    class IrisErrorTypeMismatch: public IrisErrorExceptionString
    {
    public:
        IrisErrorTypeMismatch(const std::string &message_, const std::string &name_ = std::string(), IrisErrorCode errorCode_ = E_type_mismatch): IrisErrorExceptionString(errorCode_, message_) { setName(name_); }
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorTypeMismatch"; }
        void setName(const std::string &name) { setData("{name:" + quoteString(name) + "}"); }
        void setName(const IrisU64StringConstant &name) { setData("{name:" + quoteString(toString(name)) + "}"); }
    };
        
    
    /// Unknown member when looking up a mandatory member in an Object.
    class IrisErrorUnknownMember: public IrisErrorExceptionString
    {
    public:
        IrisErrorUnknownMember(const std::string &member): IrisErrorExceptionString(E_unknown_member, "Unknown object member.") { setName(member); }
        IrisErrorUnknownMember(const IrisU64StringConstant &member): IrisErrorExceptionString(E_unknown_member, "Unknown object member.") { setName(member); }
        virtual const char *type() const IRIS_OVERRIDE { return "IrisErrorUnknownMember"; }
        void setName(const std::string &name) { setData("{name:" + quoteString(name) + "}"); }
        void setName(const IrisU64StringConstant &name) { setData("{name:" + quoteString(toString(name)) + "}"); }
    };

    /// Internal error.
    /// For errors not caused by calling Iris functions.
    class IrisInternalError: public IrisErrorExceptionString
    {
    public:
        IrisInternalError(const std::string &message_ = std::string()): IrisErrorExceptionString(E_internal_error, message_) {}
        virtual const char *type() const IRIS_OVERRIDE { return "IrisInternalError"; }
    };

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisErrorException_h

