/*!
 *  \file   IrisEventEmitterBase.h
 *  \brief  A utility class for emitting Iris Events
 *  \date   Copyright ARM Limited 2016 All Rights Reserved.
 *
 */

#ifndef ARM_INCLUDE_IrisEventEmitterBase_h
#define ARM_INCLUDE_IrisEventEmitterBase_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisValue.h"
#include "iris/IrisInstanceEvent.h"

#include <vector>
#include <cassert>

NAMESPACE_IRIS_START

class IrisInstance;
class IrisInstanceEvent;

class IrisEventEmitterBase
{
public:
    struct FieldInfo
    {
        std::string name;
        unsigned stream_count; /// Number of streams that require this field.
                               /// if it is 0 the field will not be emitted.

        FieldInfo(const std::string &name_)
            : name(name_)
            , stream_count(0)
        {
        }
    };

private:
    IrisInstance *iris_instance;
    IrisEventRegistry event_registry;

    std::vector<FieldInfo> fields;
    size_t num_fields;

    uint64_t evSrcId;

    uint64_t getSimulationTime();

    void emitField(std::vector<FieldInfo>::iterator &field)
    {
        assert(field == fields.end());
        (void)field;
    }

    template <typename ARG0, typename... REST>
    void emitField(std::vector<FieldInfo>::iterator &field, ARG0 arg0, REST... rest)
    {
        if (field->stream_count > 0)
        {
            event_registry.addFieldSlow(field->name, arg0);
        }

        ++field;
        emitField(field, rest...);
    }

protected:
    template <typename... ARGS>
    void emitEvent(ARGS... args)
    {
        if (isEnabled())
        {
            assert(fields.size() == num_fields);

            event_registry.emitEventBegin(getSimulationTime(), IRIS_UINT64_MAX);

            auto field_iterator = fields.begin();
            emitField(field_iterator, args...);

            event_registry.emitEventEnd();
        }
    }

    IrisEventEmitterBase(size_t num_fields_)
        : iris_instance(NULL)
        , num_fields(num_fields_)
        , evSrcId(IRIS_UINT64_MAX)
    {
        (void)num_fields; // This is only used for an assert
    }

    ~IrisEventEmitterBase()
    {
        // The EventEmitterEventStream objects will call back in to this object when they
        // are disabled so we need to disable them all up front to make sure they are disabled
        // before this object begins being destroyed.
        while (!event_registry.empty())
        {
            (*event_registry.begin())->disable();
        }
    }
public:
    void setIrisInstance(IrisInstance *iris_instance_)
    {
        assert(iris_instance == NULL);
        iris_instance = iris_instance_;
    }

    void setEvSrcId(uint64_t evSrcId_)
    {
        evSrcId = evSrcId_;
    }

    /// Get the Event Source Id for this event.
    uint64_t getEvSrcId() const
    {
        return evSrcId;
    }

    bool isEnabled() const
    {
        return !event_registry.empty();
    }

    IrisErrorCode createEventStream(EventStream *&stream_out, const EventSourceInfo& info,
                                    const std::vector<std::string>& requested_fields);
};

NAMESPACE_IRIS_END

#endif // ARM_INCLUDE_IrisEventEmitterBase_h
