/*!
 *  \file   IrisFunctionDecoder.h
 *  \brief  Function registry, call functions and provide function meta info.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  This implements a function registry which handles calling functions and retrieving fucntion meta-info (interface discovery).
 */

#ifndef ARM_INCLUDE_IrisFunctionDecoderInternal_h
#define ARM_INCLUDE_IrisFunctionDecoderInternal_h

#include "iris/detail/IrisCommon.h"
#include "IrisInterface.h"
#include "IrisLogger.h"
#include "IrisReceivedRequest.h"
#include "IrisJsonParser.h"
#include "iris/detail/IrisFunctionInfo.h"
#include <cassert>
#include <map>
#include <string>

NAMESPACE_IRIS_START


class IrisFunctionDecoder: public IrisInterface
{
public:
    /// Constructor.
    IrisFunctionDecoder(IrisLogger &log_): localIrisInterfaceHandlingResponses(0), remoteIrisInterface(0), log(log_), enableUnsupportedArgumentDetection(false)
    {

#define REGISTER_FUNCTION(functionName) \
            registerFunction(this, #functionName, &IrisFunctionDecoder::impl_##functionName, \
                            function_info::functionName, "IrisFunctionDecoder")

        // Register interface discovery functions. These are directly handled by this class.
        REGISTER_FUNCTION(instance_checkFunctionSupport);
        REGISTER_FUNCTION(instance_getFunctionInfo);

#undef REGISTER_FUNCTION

        enableUnsupportedArgumentDetection = getEnvU64("IRIS_CHECK_UNSUPPORTED_ARGUMENTS") > 0;
    }


    /// Destructor.
    virtual ~IrisFunctionDecoder()
    {
        // Free FunctionMapEntryFunctionoids.
        for (FunctionMap::iterator i = functionMap.begin(); i != functionMap.end(); i++)
            delete i->second;
    }


    /// Get IrisInterface of function decoder.
    /// All function calls are fed into this interface.
    IrisInterface *getIrisInterface() { return this; }


    /// Optionally splice responses to another object.
    /// If this is not called (or called with 0) then response messages must never be
    /// send to this object (else internal error).
    void forwardResponsesTo(IrisInterface *irisInterface_)
    {
        localIrisInterfaceHandlingResponses = irisInterface_;
    }


    /// Send decoding errors and responses to functions implemented in this object to the remote interface.
    /// This must be called before the function decoder is usable.
    void setRemoteIrisInterface(IrisInterface *remoteIrisInterface_)
    {
        assert(remoteIrisInterface_ != 0);
        remoteIrisInterface = remoteIrisInterface_;
    }


    /// Register member function of instance type T.
    template<class T>
    void registerFunction(T *instance, const std::string &name, void (T::*memberFunctionPtr)(IrisReceivedRequest &), const std::string &functionInfoJson, const std::string &instanceTypeStr)
    {
        if (functionMap.find(name) != functionMap.end())
            throw IrisInternalError("IrisFunctionDecoderInternal::registerFunction(" + name + "): function is already registered");
        FunctionMapEntry *functionMapEntry = new FunctionMapEntryFunctionoid<T>(instance, memberFunctionPtr, name, functionInfoJson, instanceTypeStr);
        functionMap[name] = functionMapEntry;
    }

    void unregisterFunction(const std::string &name)
    {
        FunctionMap::iterator it = functionMap.find(name);
        if (it == functionMap.end())
        {
            throw IrisInternalError("IrisFunctionDecoderInternal::unregisterFunction(" + name = "): "
                                    "No function by that name registered");
        }
        delete it->second;
        functionMap.erase(it);
    }

private:

    /// Check unsupported arguments.
    /// Extract each parameter name in the request message, and check if it is in the function info.
    /// Throw error E_unsupported_argument_name if an argument is not supported.
    void checkUnsupportedArguments(const uint64_t *message, const IrisValue &functionInfoU64Data)
    {
        IrisU64JsonReader functionInfoReader = IrisU64JsonReader(functionInfoU64Data);
        IrisU64JsonReader::Object functionInfoObject = functionInfoReader.openObject();

        // Function information for arguments
        IrisU64JsonReader::Object functionInfoArgs;
        if (functionInfoObject.memberOptional(ISTR("args")))
            functionInfoArgs = functionInfoObject.openObject();

        IrisU64JsonReader requestReader(message);
        IrisU64JsonReader::Request req = requestReader.openRequest();

        // All input parameters
        std::string paramName;
        while(req.readNextParam(paramName))
        {
            // Skip the value, as we only care about the parameter name.
            requestReader.skip();

            // Skip instId, which is special and it is not a real argument.
            if (paramName == "instId")
                continue;

            // Check such name in function information
            // Note: The performance of this can be drastically improved later:
            //   (1) We can keep paramName a U64JSON string without converting to a std::string.
            //       This is not supported in the U64JsonReader yet.
            //   (2) We can encourage generators of function calls to sort their arguments
            //       alphabetically. This way we compare the list of actual arguments against the
            //       list of formal arguments by comparing two list sequentially element by element,
            //       since we can rely on both lists to be sorted.
            if (!functionInfoArgs.memberOptionalSlow(paramName))
            {
                IrisErrorUnknownMember e(paramName);
                e.setErrorCode(E_unsupported_argument_name);
                e.setMessage("Unsupported argument");
                throw e;
            }
        }
    }

    // --- IrisInterface implementation ---

    virtual void irisHandleMessage(const uint64_t *message) IRIS_OVERRIDE
    {
        if (IrisU64JsonReader::isRequest(message))
        {
            irisHandleRequest(message);
        }
        else
        {
            irisHandleResponse(message);
        }
    }

    void irisHandleRequest(const uint64_t *message)
    {
        assert(IrisU64JsonReader::isRequest(message));

        IrisReceivedRequest request(message);
        FunctionMap::const_iterator i = functionMap.find(request.getFunctionName());
        if (i == functionMap.end())
        {
            // Function not found: Return E_function_not_supported_by_instance if this is a request.
            if (request.isRequest())
            {
                if (!remoteIrisInterface)
                    throw IrisInternalError("IrisFunctionDecoderInternal::irisHandleRequest(): function not supported"
                                            " by instance but no remoteIrisInterface set to report the"
                                            " E_function_not_supported_by_instance error to");
                remoteIrisInterface->irisHandleMessage(request.generateErrorResponse(E_function_not_supported_by_instance,
                                                                                "instance does not support this function"));
                return;
            }
            else
            {
                // Notification: All we can do is log this.
                log.error("IrisFunctionDecoderInternal::irisHandleRequest(): notification '%s' not supported by"
                          " instance instId=%lld", request.getFunctionName().c_str(),
                          (long long)(IrisU64JsonReader::getInstId(message)));
            }
        }

        try
        {
            if (enableUnsupportedArgumentDetection)
                checkUnsupportedArguments(message, i->second->functionInfoU64Data);

            // Call registered function.
            i->second->call(request);
        }
        catch(IrisErrorException &e)
        {
            if (!remoteIrisInterface)
                throw IrisInternalError("IrisFunctionDecoderInternal::irisHandleRequest(): function caused an error"
                                        " but no remoteIrisInterface set to report the error to");
            remoteIrisInterface->irisHandleMessage(request.generateErrorResponse(e.getErrorCode(), e.getMessage(), e.getData()));
        }
    }

    void irisHandleResponse(const uint64_t *message)
    {
        assert(!IrisU64JsonReader::isRequest(message));

        // Forward to response handler (if any).
        if (localIrisInterfaceHandlingResponses)
            localIrisInterfaceHandlingResponses->irisHandleMessage(message);
        else
            throw IrisInternalError("IrisFunctionDecoderInternal::irisHandleResponse() called but no local response"
                                    " handler was set (forwardResponsesTo() was not called)");
    }

    virtual void *irisObtainInterface(IrisInstId /*instId*/, const char * /*requested_interface*/) IRIS_OVERRIDE
    {
        throw IrisErrorExceptionString(E_function_not_supported_by_instance, "obtainInterface() is reserved for future use.");
    }


    // --- Iris function implementations ---


    void impl_instance_checkFunctionSupport(IrisReceivedRequest &request)
    {
        // Read arguments and check for functions and args.
        IrisU64JsonReader::Array functionsArray = request.getArgArray(ISTR("functions"));
        bool result = true;
        while (!functionsArray.reachedEnd())
        {
            // Get function name.
            IrisU64JsonReader::Object o = functionsArray.openObject();
            std::string functionName;
            o.member(ISTR("name"), functionName);

            // Find function info.
            FunctionMap::const_iterator i = functionMap.find(functionName);
            if (i == functionMap.end())
            {
                // Function not found.
                result = false;
                break;
            }

            // Check args (if specified).
            if (o.memberOptional(ISTR("args")))
            {
                // Get FunctionInfo::args meta information.
                IrisU64JsonReader functionInfoReader = IrisU64JsonReader(i->second->functionInfoU64Data);
                IrisU64JsonReader::Object functionInfoObject = functionInfoReader.openObject();
                functionInfoObject.member(ISTR("args"));
                IrisU64JsonReader::Object functionInfoArgs = functionInfoObject.openObject();

                // Check all args.
                IrisU64JsonReader::Array argsArray = o.openArray();
                while(!argsArray.reachedEnd())
                {
                    // Get arg name.
                    std::string argName;
                    argsArray.element(argName);

                    // Check arg name.
                    if (!functionInfoArgs.memberOptionalSlow(argName))
                    {
                        // Argument not found.
                        result = false;
                        break;
                    }
                }
                if (!result)
                    break;
            }
        }

        // Return result.
        IrisU64JsonWriter &resultWriter = request.startOkResponse();
        resultWriter.persist(result);
        remoteIrisInterface->irisHandleMessage(request.endOkResponse());
    }


    void impl_instance_getFunctionInfo(IrisReceivedRequest &request)
    {
        // Read arguments.
        std::string prefix = request.getOptionalArgString(ISTR("prefix"), std::string());

        // Generate result object.

        // For all registered functions.
        IrisU64JsonWriter &resultWriter = request.startOkResponse();
        IrisU64JsonWriter::Object o = resultWriter.openObject();
        for (FunctionMap::const_iterator i = functionMap.begin(); i != functionMap.end(); i++)
        {
            if (!stringStartsWith(i->first, prefix))
                continue;
            o.memberSlow(i->first, i->second->functionInfoU64Data);
        }
        o.close();

        // Return result.
        remoteIrisInterface->irisHandleMessage(request.endOkResponse());
    }


    /// Function map entry.
    /// Contains:
    /// - functionoid to call function
    /// - function meta info
    class FunctionMapEntry
    {
    public:
        /// Constructor.
        /// instanceTypeStr is only used for error messages.
        FunctionMapEntry(const std::string &functionName, const std::string &functionInfoJson, const std::string &instanceTypeStr)
        {
            IrisU64JsonWriter writer;
            try
            {
                convertJsonToU64Json(functionInfoJson, writer);
            }
            catch(const IrisErrorException &e)
            {
                // We use a temprary logger with the type name, not the instanc name, since propagating the instance name into
                // this object is quite tedious and generally not necessary: This error message is really a compile-time error
                // message attached to a type which always applies to all instances of a type.
                IrisLogger logger(instanceTypeStr);
                logger.error("Error while parsing FunctionInfo JSON for function '%s':\n%s\n", functionName.c_str(), e.what());
            }
            writer.copyToIrisValue(functionInfoU64Data);
        }

        /// Virtual destructor.
        virtual ~FunctionMapEntry() {}

        /// Call registered function.
        virtual void call(IrisReceivedRequest &) = 0;

        /// Function meta information (FunctionInfo etc).
        IrisValue functionInfoU64Data;
    };


    /// Add functionoid to FunctionMapEntry.
    /// The functionoid allows the FunctionMapEntry to call member functions of arbitrary instances.
    /// This is a simplistic delegate mechanism which works since all functions have the same
    /// prototype.
    template<class T>
    class FunctionMapEntryFunctionoid: public FunctionMapEntry
    {
    public:
        FunctionMapEntryFunctionoid(T *instance_, void (T::*memberFunctionPtr_)(IrisReceivedRequest &), const std::string &functionName, const std::string &functionInfoJson, const std::string &instanceTypeStr): FunctionMapEntry(functionName, functionInfoJson, instanceTypeStr), instance(instance_), memberFunctionPtr(memberFunctionPtr_)
        {
        }

        virtual void call(IrisReceivedRequest &req) IRIS_OVERRIDE
        {
            (instance->*memberFunctionPtr)(req);
        }

    private:
        T *instance;
        void (T::*memberFunctionPtr)(IrisReceivedRequest &);
    };


    /// Function map type.
    typedef std::map<std::string,FunctionMapEntry*> FunctionMap;


    // --- private data ---

    /// Map function names onto member function pointers.
    FunctionMap functionMap;

    /// Splice incoming reponses to another local IrisInterface.
    IrisInterface *localIrisInterfaceHandlingResponses;

    /// Remote iris interface (counter of the local irtis interface).
    /// We send error responses and responses to functions implemented by this object (interface discovery) to this interface.
    IrisInterface *remoteIrisInterface;

    /// Logger.
    IrisLogger &log;

    /// Detecting unsupported arguments is optional, as it causes extra overhead.
    bool enableUnsupportedArgumentDetection;
}; // class IrisFunctionDecoderInternal

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisFunctionDecoder_h

