/*!
 *  \file   IrisHistEndianness.h
 *  \brief  Host endianness abstraction.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  Host endianness abstraction.
 *
 *  Note: Only very few places in Iris require dealing explicitly with the host endianness.
 *  The U64JSON fomat is always stored in host endianness and never requires conversion.
 *  Explicit conversions happen for example in the IrisTcpClient and IrisTcpServer on
 *  big endian machines (which are getting very rare these days).
 *
 *  Explicit conversions also happen when accessing string values of string registers.
 *
 *  Iris is designed to be fast on little-endian machines.
 *
 *  Note: The big-endian code is not yet tested.
 */

#ifndef ARM_INCLUDE_IrisHostEndianness_h
#define ARM_INCLUDE_IrisHostEndianness_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisInterface.h" // uint64_t etc

// Configuration: Specify whether we are on a little endian host or big endian host.
// Virtually all useful hosts are little endian today, so we guess little endian here.
#ifndef IRIS_HOST_LITTLE_ENDIAN
#define IRIS_HOST_LITTLE_ENDIAN 1
#endif


NAMESPACE_IRIS_START


/// Read a uint32_t from memory bytes in little endian.
inline uint32_t readU32LE(const void *p)
{
#if IRIS_HOST_LITTLE_ENDIAN
    return *reinterpret_cast<const uint32_t*>(p);
#else
    uint8_t bytes = reinterpret_cast<const uint8_t*>(p)
    return bytes[0] | (bytes[1] << 8) | (bytes[2] << 16) | (bytes[3] << 24);
#endif
}


/// Pack a uint32_t into a string as binary in little endian.
inline std::string packU32IntoStringLE(uint32_t u32)
{
#if IRIS_HOST_LITTLE_ENDIAN
    return std::string((const char *)&u32, 4);
#else
    std::string r = "abcd";
    r[0] = (char)(uint8_t)(u32);
    r[1] = (char)(uint8_t)(u32 >> 8);
    r[2] = (char)(uint8_t)(u32 >> 16);
    r[3] = (char)(uint8_t)(u32 >> 24);
    return r;
#endif
}


/// Pack a uint64_t into a string as binary in little endian.
inline std::string packU64IntoStringLE(uint64_t u64)
{
#if IRIS_HOST_LITTLE_ENDIAN
    return std::string((const char *)&u64, 4);
#else
    std::string r = "abcdefgh";
    r[0] = (char)(uint8_t)(u64);
    r[1] = (char)(uint8_t)(u64 >> 8);
    r[2] = (char)(uint8_t)(u64 >> 16);
    r[3] = (char)(uint8_t)(u64 >> 24);
    r[4] = (char)(uint8_t)(u64 >> 32);
    r[5] = (char)(uint8_t)(u64 >> 40);
    r[6] = (char)(uint8_t)(u64 >> 48);
    r[7] = (char)(uint8_t)(u64 >> 56);
    return r;
#endif
}


/// Pack a uint64_t array into a string as binary in little endian.
inline std::string packU64ArrayIntoStringLE(const uint64_t *data, size_t numElements)
{
#if IRIS_HOST_LITTLE_ENDIAN
    return std::string((const char *)data, numElements * 8);
#else
    std::string r;
    for (size_t = 0; i < numElements; i++)
        r.append(packU64IntoStringLE(data[i]));
    return r;
#endif
}


/// Reverse byte order.
inline uint32_t reverseByteOrder(uint32_t x)
{
    return (x << 24) | ((x << 8) & 0xff0000) | ((x >> 8) & 0xff00) | (x >> 24);
}


/// Pack a uint64_t into a uint8_t  as binary in little endian.
inline void convertLEU64ArrayIntoHostEndianness(uint64_t * data, uint64_t length)
{
#if IRIS_HOST_LITTLE_ENDIAN
    //To avoid unused parameter errors.
    (void) data;
    (void) length;
#else
    for (uint64_t i = 0; i < length; i++)
    {
        reverseByteOrder(data[i]);
    }
#endif
}


/// Reverse byte order.
inline uint64_t reverseByteOrder(uint64_t x)
{
    return (uint64_t(reverseByteOrder(uint32_t(x))) << 32) | reverseByteOrder(uint32_t(x >> 32));
}


/// Convert a sequence of uint64_t values to a string, assuming that the uint64_t data is
/// formatted accoding to the register_read()/register_write() specification:_
/// First uint64_t elemnt is the length of the string in bytes, all following elements
/// contain the string characters, first char in the LSB (little-endian regardless of host endianness).
inline std::string convertU64StringToString(const uint64_t *data)
{
#if IRIS_HOST_LITTLE_ENDIAN
    return std::string((const char *)data + 1, data[0]);
#else
    std::string r;
    uint64_t len = *(data++);
    while (len >= 8)
    {
        r.append(packU64IntoStringLE(*(data++)));
        len -= 8;
    }
    if (len > 0)
    {
        uint64_t last = *data;
        while (len > 0)
        {
            r.append(1, (char)last);
            last >>= 8;
            len--;
        }
    }
    return r;
#endif
}


/// Convert a string to a sequence of uint64_T values, assuming that the uint64_t data is
/// formatted accoding to the register_read()/register_write() specification:_
/// First uint64_t elemnt is the length of the string in bytes, all following elements
/// contain the string characters, first char in the LSB (little-endian regardless of host endianness).
///
/// dataOut: String is written into this buffer. This buffer must be at least N=(str.size() + 15) / 8 elements long.
inline void convertStringToU64String(const std::string &str, uint64_t *dataOut)
{
    uint64_t len = uint64_t(str.size());
    dataOut[(len + 7) / 8] = 0; // Clear bytes in last uint64_t element.
    *dataOut = len;
    memcpy(dataOut + 1, str.data(), str.size());
#if IRIS_HOST_LITTLE_ENDIAN
#else
    for (size_t i = 1; i < (len + 15) / 8; i++)
        dataOut[i] = reverseBytesU64(dataOut[i]);
#endif
}

/// Convert a sequence of bytes to an array of NumberU64, assuming that the NumberU64 is formatted
/// according to the image_loadFile() specification:
/// The bytes are stored in little endian in the NumberU64 values. The last
/// element in the array may contain 1..8 valid bytes. Unused bytes are set to 0.
///
/// u64Array: Bytes are written into this buffer. This buffer must be at least N=(str.size() + 7) / 8 elements long.
inline void convertBytesToU64Array(const char* data, uint64_t numBytes, uint64_t *u64Array)
{
    uint64_t numU64 = (numBytes + 7) / 8;
    u64Array[numU64 - 1] = 0; // Clear bytes in last uint64_t element.
    memcpy(u64Array, data, numBytes);
#if IRIS_HOST_LITTLE_ENDIAN
#else
    for (size_t i = 0; i < numU64; i++)
        u64Array[i] = reverseBytesU64(u64Array[i]);
#endif
}

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisHostEndianness_h
