/*!
 *  \file   IrisInterface.h
 *  \brief  Declaration of the main Iris function call mechanism.
 *  \date   Copyright ARM Limited 2014-2017 All Rights Reserved.
 *
 *  All Iris function calls, requests, responses and notifications go through IrisInterfaces.
 */

#ifndef ARM_INCLUDE_IrisInterface_h
#define ARM_INCLUDE_IrisInterface_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisC.h"
#include "iris/detail/IrisErrorCode.h"
#include <sys/types.h> // size_t


NAMESPACE_IRIS_START
    
/// Terminology:
///
/// This interface and all Iris helper classes use the following terminology:
///
/// - request: Same as in JSON RPC 2.0: This is a function call which delivers 
///            a response. This is often used synonymously for 'request or 
///            notification' if the difference between request and 
///            notification is not important.
///
/// - notification: Same as in JSON RPC 2.0: This is a function call which 
///            does not deliver a response (also no error indication and no 
///            indication of completion).
///
/// - response: Same as in JSON RPC 2.0: This is the response sent in the 
///            other direction in response to a request.
///
/// - message: Use as a common term for request, notification or response. 
///           Message transported between instances.
///
/// - function call: Used synonymously to request.
///
/// - caller: Instance which originally issued a request or notification and 
///           which receives responses to these requests. Instance which calls 
///           irisHandleMessage() on the callee.
///
/// - callee: Instance which receives a request or notification and which 
///           potentially send a response. Instance which calls 
///           irisHandleMessage() on the caller.
///
/// - thread: Host thread. Not a co-routine/fiber.
///
/// - instance: Entity which can send and/or receive messages.
///
/// - instId: Densely allocated integer id uniquely identifying each instance.
    

/*!
 * @brief Instance ids are densely allocated unsigned integers.
 */
typedef uint64_t IrisInstId;

/*!
 * @brief Instance id of framework.GlobalInstance.
 */
static const IrisInstId IrisInstIdGlobalInstance = 0;

/*!
 * @brief Instance id of framework.SimulationEngine
 */
static const IrisInstId IrisInstIdSimulationEngine = 1;

// Artificial limit to the number of instances which can be handled by Iris (total number of instances in the whole system).
// This mainly exists to catch programming errors like instId = -1.
static const IrisInstId IrisMaxTotalInstances = 2*1024*1024;


/*!
 * @brief Iris interface.
 *
 * This interface receives incoming requests, notifications and incoming
 * responses. This is the main (and only) entry point of Iris communication
 * between C++ instances. Every instance implements an IrisInterface to
 * receive requests, notifications and responses and every instance has a
 * pointer to a sending IrisInterface to be able to send requests,
 * notifications and responses.
 * All other Iris classes (belonging to IrisSupportLib) are just helper
 * classes and generally only maintain source code compatibility (as far as
 * possible). It is not mandatory to use the helper classes for a correct
 * Iris implementation.
 */
class IrisInterface
{
public:
    /*! Send an Iris message.
     *
     * Send request (call function with return value), send notification
     * (function call without return value and without error notification),
     * or response (containing a return value or an error).
     *
     * When handling a request
     * irisHandleMessage() processes the given message. The response may be
     * generated either synchronously from within irisHandleMessage() or
     * asynchronously. If irisHandleMessage() does not generate a response
     * itself then it must make arrangements for the response to be generated
     * by some other means.
     *
     * This function may (in fact will most likely) return before the
     * semantically called function returns a response. This function may
     * return before the request or notification reaches the callee. The
     * return value (or even just the information that the function completed)
     * is transmitted by the callee calling irisHandleMessage() on the
     * original caller. The IrisGlobalInstance takes care of routing function
     * calls and responses to the correct instances.
     *
     * caller::irisHandleMessage() may be called from within
     * callee::irisHandleMessage() or from any other thread after
     * irisHandleMessage() was entered. irisHandleMessage() may be called from
     * any host thread.
     *
     * This function returns nothing. irisHandleMessage() will be called on
     * the caller asynchronously to deliver the response to a request.
     *
     * @param   message     This is a JSON RPC 2.0 encoded request,
     *                      notification, or response encoded in U64JSON
     *                      instead of textual JSON.
     */
    virtual void irisHandleMessage(const uint64_t *message) = 0;


    /*!
     * @brief Obtain a native interface provided by an in-process instance.
     *
     * @param   instId                  Identifies the instance to obtain the interface from.
     *                                  This is used to route the call to the correct instance
     *                                  the same way that the instId parameter is used in
     *                                  Iris function calls. End points (instances) that
     *                                  implement obtainInterface should ignore the instId argument
     *                                  other than to confirm that it matches their own instId.
     * @param   requested_interface     The name of the interface requested
     * @return                          Pointer to the requested interface.
     */
    virtual void *irisObtainInterface(IrisInstId instId, const char *requested_interface) = 0;

protected:
    /// Forbid destruction of instances through this interface.
    virtual ~IrisInterface() {}
};

// Duplicate the AsyncMessage flags in the iris namespace
/*!
 * @defgroup async_message_flags Asynchronous message processing flags.
 *
 * @brief Constants that can be passed to the flags parameter of
 *        IrisConnectionInterface::processAsyncMessages.
 * @{
 */

/*!
 * @brief Default non-blocking behaviour.
 *
 * Returns immediately if there are no outstanding messages to process.
 */
static const uint64_t AsyncMessage_Default      = IrisC_AsyncMessage_Default;

/*!
 * @brief Default non-blocking behaviour.
 *
 * Returns immediately if there are no outstanding messages to process.
 */
static const uint64_t AsyncMessage_NonBlocking  = AsyncMessage_Default;

/*!
 * @brief Block until there are messages to process.
 *
 *  If there are no outstanding messages to process, block until there is one.
 *  This is useful when waiting for a response to a request. By calling processAsyncMessages()
 *  like this:
 *  @code
        while (no_response_received)
            processAsyncMessages(AsyncMessage_Blocking);
    @endcode
 *  you wait for the response and also avoid deadlock situations where the recipient
 *  of you request makes a request itself that needs to be marshalled to the thread
 *  being blocked.
 */
static const uint64_t AsyncMessage_Blocking     = IrisC_AsyncMessage_Blocking;

/*!
 * @}
 */

/*!
 * @brief Interface used by an IrisInstance to join an Iris system
 */
class IrisConnectionInterface
{
public:

    /*!
     * @brief Register a local IrisInterface with the system.
     *
     * This allows it to receive messages (requests and responses).
     *
     * @param   iris_interface  The IrisInterface to be called to communicate with the
     *                          new channel.
     * @return                  The unique channelId used to identify this channel when
     *                          registering instances.
     */
    virtual uint64_t registerIrisInterfaceChannel(IrisInterface *iris_interface) = 0;

    /*!
     * @brief Unregister a previously registered channel.
     *
     * This will automatically unregister all instances associated with that channel.
     *
     * @param   channelId   The id of a channel that was previously registered through
     *                      this interface.
     */
    virtual void unregisterIrisInterfaceChannel(uint64_t channelId) = 0;

    /*!
     * @brief Forward any pending messages that are waiting to be processed in the executing thread.
     *
     * @param   flags   One of the @ref async_message_flags.
     */
    virtual IrisErrorCode processAsyncMessages(uint64_t flags) = 0;

    /*!
     * @brief Get the remote IrisInterface for this connection.
     *
     * @return      The remote IrisInterface for this connection.
     */
    virtual IrisInterface *getIrisInterface() = 0;

protected:
    virtual ~IrisConnectionInterface() {}
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisInterface_h

