/*!
 *  \file   IrisObjects.h
 *  \brief  Iris Objects and functions to convert them to/from U64JSON.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Iris Objects and functions to convert them to/from U64JSON.
 */

#ifndef ARM_INCLUDE_IrisObjects_h
#define ARM_INCLUDE_IrisObjects_h

#include "iris/detail/IrisCommon.h"
#include "IrisPersistPSH.h"
#include "IrisU64JsonReader.h"
#include "IrisU64JsonWriter.h"
#include "IrisInterface.h" // uint64_t
#include "IrisHostEndianness.h"
#include <string>
#include <vector>
#include <map>


NAMESPACE_IRIS_START


// See IrisPersistPSH.h for deatils why we wrap persist() functions in PersistPSH structs.


// --- Iris objects and their persist() functions. ---------------------------------------------------------------

/// Instance info for instance registry.
struct InstanceInfo
{
    InstanceInfo(IrisInstId instId_ = 0, const std::string& instName_ = std::string()): instId(instId_), instName(instName_) {}
    IrisInstId instId;
    std::string instName;
};

/// Read/write InstanceInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,InstanceInfo> { static void persist(PERSISTER &p, InstanceInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("instId"), inOrOut.instId);
    o.member(ISTR("instName"), inOrOut.instName);
}};


/// Argument for instance_checkFunctionSupport().
struct FunctionSupportRequest
{
    FunctionSupportRequest() {}
    /// Convenience constructor for up to three arguemnts.
    /// Note that typically caller will only check zero arguments or one critical argument.
    FunctionSupportRequest(const std::string &_name, const std::string &arg0 = std::string(), const std::string &arg1 = std::string(), const std::string &arg2 = std::string()): name(_name)
    {
        if (!arg0.empty())
            args.push_back(arg0);
        if (!arg1.empty())
            args.push_back(arg1);
        if (!arg2.empty())
            args.push_back(arg2);
    }
    std::string name; // Function name.
    std::vector<std::string> args; // Function argument names to check for.
};

/// Read/write FunctionSupportRequest.
template<class PERSISTER>
struct PersistPSH<PERSISTER,FunctionSupportRequest> { static void persist(PERSISTER &p, FunctionSupportRequest &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("name"), inOrOut.name);
    bool hasArgs = !inOrOut.args.empty();
    o.memberOptional(ISTR("args"), inOrOut.args, hasArgs);
}};
    
    
/// EnumElementInfo, returned by resource_getList() in ResourceInfo.enums and other contexts (instanceGetFunctionInfo()).
struct EnumElementInfo
{       
    IrisValue value;
    std::string symbol; // Optional: Empty means missing.
    std::string descr;  // Optional: Empty means missing.
    
    EnumElementInfo() {}
    EnumElementInfo(const IrisValue &value_, const std::string &symbol_, const std::string &descr_): value(value_), symbol(symbol_), descr(descr_) {}
};

/// Read/write EnumElementInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,EnumElementInfo> { static void persist(PERSISTER &p, EnumElementInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("value"), inOrOut.value);
    bool hasSymbol = !inOrOut.symbol.empty();
    o.memberOptional(ISTR("symbol"), inOrOut.symbol, hasSymbol);
    bool hasDescr = !inOrOut.descr.empty();
    o.memberOptional(ISTR("descr"), inOrOut.descr, hasDescr);
}};

    
/// AttributeInfo, returned by instanceGetFunctionInfo(), also used in other contexts.
struct AttributeInfo
{
    std::string type; // Mandatory: Argument type.
    std::string descr; // Mandatory: Argument description.
    std::vector<EnumElementInfo> enums; // Optional: Enum values.
    bool optional; // Optional: true means: Optional argument.
    IrisValue defval; // Optional: Default value for optional arguments.
    
    AttributeInfo(): optional(false) {}
};

/// Read/write AttributeInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,AttributeInfo> { static void persist(PERSISTER &p, AttributeInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("type"), inOrOut.type);
    bool hasDescr = !inOrOut.descr.empty();
    o.memberOptional(ISTR("descr"), inOrOut.descr, hasDescr);
    bool hasEnums = !inOrOut.enums.empty();
    o.memberOptional(ISTR("enums"), inOrOut.enums, hasEnums);
    bool hasOptional = inOrOut.optional;
    o.memberOptional(ISTR("optional"), inOrOut.optional, hasOptional);
    bool hasDefval = !inOrOut.defval.isEmptyOrNull();
    o.memberOptional(ISTR("defval"), inOrOut.defval, hasDefval);
}};


/// FunctionInfo, returned by instanceGetFunctionInfo().
struct FunctionInfo
{
    std::string descr; // Function description.
    std::map<std::string,AttributeInfo> args; // Function argument information.
    AttributeInfo retval; // Function return value information.
};

/// Read/write FunctionInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,FunctionInfo> { static void persist(PERSISTER &p, FunctionInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("descr"), inOrOut.descr);
    o.member(ISTR("args"), inOrOut.args);
    o.member(ISTR("retval"), inOrOut.retval);
}};


/// --- Resource ---

struct ParameterInfo
{
    std::vector<uint64_t> defaultData;  // Optional: Default value of the parameter (for numeric parameters).
    std::string defaultString;          // Optional: Default string of the parameter (for string parameters).
    std::vector<uint64_t> min;          // Optional: minimum numeric value accepted (inclusive).
    std::vector<uint64_t> max;          // Optional: maximum numeric value accepted (inclusive).
    bool initOnly;                      // Optional: True for Init-time parameter, False for Run-time parameter. Default is false.
    bool topology;                      // Optional: Iff present and True: This is a topology parameter which affects the topology of the system.

    ParameterInfo(): initOnly(false), topology(false) {}
};

/// Read/write ParameterInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ParameterInfo> { static void persist(PERSISTER &p, ParameterInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    bool hasDefaultData = !inOrOut.defaultData.empty();
    o.memberOptional(ISTR("defaultData"), inOrOut.defaultData, hasDefaultData);
    bool hasDefaultString = !inOrOut.defaultString.empty();
    o.memberOptional(ISTR("defaultString"), inOrOut.defaultString, hasDefaultString);
    bool hasMin = !inOrOut.min.empty();
    o.memberOptional(ISTR("min"), inOrOut.min, hasMin);
    bool hasMax = !inOrOut.max.empty();
    o.memberOptional(ISTR("max"), inOrOut.max, hasMax);
    bool isInitOnly = inOrOut.initOnly;
    o.memberOptional(ISTR("initOnly"), inOrOut.initOnly, isInitOnly);
    bool isTopology = inOrOut.topology;
    o.memberOptional(ISTR("topology"), inOrOut.topology, isTopology);
}};

struct RegisterInfo
{
    uint64_t canonicalRn; // Optional: Missing iff hasCanonicalRn == false.
    bool hasCanonicalRn;
    uint64_t lsbOffset; // Optional: 2**64-1 means missing.
    std::vector<uint64_t> writeMask; // Optional: Empty array means missing.
    std::vector<uint64_t> resetData; // Optional: Empty array means missing.
    std::string resetString;         // Optional: Empty string means missing (For type 'string' only).
    uint64_t addressOffset; // Optional: Missing iff hasAddressOffset == false.
    bool hasAddressOffset;

    RegisterInfo()
        : canonicalRn(0)
        , hasCanonicalRn(false)
        , lsbOffset(IRIS_UINT64_MAX)
        , addressOffset(0)
        , hasAddressOffset(false)
    {
    }
};

/// Read/write RegisterInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,RegisterInfo> { static void persist(PERSISTER &p, RegisterInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.memberOptional(ISTR("canonicalRn"), inOrOut.canonicalRn, inOrOut.hasCanonicalRn);
    bool hasLsbOffset = inOrOut.lsbOffset != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("lsbOffset"), inOrOut.lsbOffset, hasLsbOffset);
    bool hasWriteMask = !inOrOut.writeMask.empty();
    o.memberOptional(ISTR("writeMask"), inOrOut.writeMask, hasWriteMask);
    bool hasResetData = !inOrOut.resetData.empty();
    o.memberOptional(ISTR("resetData"), inOrOut.resetData, hasResetData);
    bool hasResetString = !inOrOut.resetString.empty();
    o.memberOptional(ISTR("resetString"), inOrOut.resetString, hasResetString);
    o.memberOptional(ISTR("addressOffset"), inOrOut.addressOffset, inOrOut.hasAddressOffset);
}};


/// ResourceInfo, returned by resource_getList().
struct ResourceInfo
{
    uint64_t rscId;     // Mandatory.
    std::string name;   // Mandatory.
    std::string cname;  // Optional: Empty means missing.
    std::string descr;  // Optional: Empty means missing.
    std::string format; // Optional: Empty means missing.
    uint64_t bitWidth;  // Mandatory.
    std::string type;   // Optional: Empty means 'numeric'. Must be either empty, 'numeric', 'numericFp', 'string' or 'noValue'.
    std::vector<EnumElementInfo> enums; // Optional: Empty array means missing.
    std::string rwMode; // Optional: Empty means 'rw'. Must be either empty, 'rw', 'r' or 'w'.
    uint64_t subRscId;  // Optional: 2**64-1 means missing.
    std::map<std::string,IrisValue> tags; // Optional: Empty means missing.

    bool isParameter;
    ParameterInfo parameterInfo;

    bool isRegister;
    RegisterInfo registerInfo;

    ResourceInfo()
        : rscId(IRIS_UINT64_MAX)
        , bitWidth(0)
        , subRscId(IRIS_UINT64_MAX)
        , isParameter(false)
        , isRegister(false)
    {
    }

    /// Return number of uint64_t units which the data of this resource occupies in the
    /// ResourceReadResult.data/.undefinedBits fields and in the data argument of resource
    /// write delegates.
    /// data: Pointer to the first uint64_t element of the data for this resource.
    /// Only used for string resources (length of string in bytes).
    size_t getDataSizeInU64Chunks() const
    {
        return (bitWidth + 63) / 64;
    }

    /// Convert a string to a sequence of uint64_t values.
    /// This is used by resource read delegates to put a string into the data buffer.
    static void stringToData(const std::string &str, std::vector<uint64_t> &data)
    {
        size_t start = data.size();
        data.resize(start + (str.size() + 15) / 8);
        convertStringToU64String(str, &data[start]);
    }

    /// Convert a sequence of uint64_t values to a string.
    /// This is used by resource write delegates to retrieve the actual string value from the data buffer.
    static std::string dataToString(const uint64_t *data)
    {
        return convertU64StringToString(data);
    }
};

/// Read/write ResourceInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ResourceInfo> { static void persist(PERSISTER &p, ResourceInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    bool isInstantiationParameter = inOrOut.isParameter && (inOrOut.rscId == IRIS_UINT64_MAX);
    bool hasRscId = !isInstantiationParameter;
    o.memberOptional(ISTR("rscId"), inOrOut.rscId, hasRscId);

    o.member(ISTR("name"), inOrOut.name);

    bool hasCname = !inOrOut.cname.empty();
    o.memberOptional(ISTR("cname"), inOrOut.cname, hasCname);

    bool hasDescr = !inOrOut.descr.empty();
    o.memberOptional(ISTR("descr"), inOrOut.descr, hasDescr);

    bool hasFormat = !inOrOut.format.empty();
    o.memberOptional(ISTR("format"), inOrOut.format, hasFormat);
    o.member(ISTR("bitWidth"), inOrOut.bitWidth);

    bool hasType = !inOrOut.type.empty();
    o.memberOptional(ISTR("type"), inOrOut.type, hasType);

    bool hasEnums = !inOrOut.enums.empty();
    o.memberOptional(ISTR("enums"), inOrOut.enums, hasEnums);

    bool hasRwMode = !inOrOut.rwMode.empty();
    o.memberOptional(ISTR("rwMode"), inOrOut.rwMode, hasRwMode);

    bool hasSubRscId = inOrOut.subRscId != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("subRscId"), inOrOut.subRscId, hasSubRscId);

    bool hasTags = !inOrOut.tags.empty();
    o.memberOptional(ISTR("tags"), inOrOut.tags, hasTags);
    o.memberOptional(ISTR("parameterInfo"), inOrOut.parameterInfo, inOrOut.isParameter);
    o.memberOptional(ISTR("registerInfo"), inOrOut.registerInfo, inOrOut.isRegister);
}};


/// ResourceGroupInfo, returned by resource_getListOfResourceGroups().
struct ResourceGroupInfo
{
    std::string name;   // Mandatory.
    std::string cname;  // Optional: Empty means missing.
    std::string descr;  // Optional: Empty means missing.
    typedef std::vector<uint64_t> ResourceList;
    ResourceList resourceList; // Mandatory.
};

/// Read/write ResourceInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ResourceGroupInfo> { static void persist(PERSISTER &p, ResourceGroupInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("name"), inOrOut.name);
    bool hasCname = !inOrOut.cname.empty();
    o.memberOptional(ISTR("cname"), inOrOut.cname, hasCname);
    bool hasDescr = !inOrOut.descr.empty();
    o.memberOptional(ISTR("descr"), inOrOut.descr, hasDescr);
    o.member(ISTR("resourceList"), inOrOut.resourceList);
}};


/// ResourceReadResult, returned by resource_read()
struct ResourceReadResult
{
    typedef std::vector<uint64_t> U64Array;
    U64Array data;                    // Mandatory. List of numeric resource values read.
    std::vector<std::string> strings; // Optional: List of string resource values read. Empty means missing.
    U64Array undefinedBits;           // Optional: Empty if missing (which means there are no undefined bits).
    U64Array error;                   // Optional: Empty means missing.
};

/// Read/write ResourceReadResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ResourceReadResult> { static void persist(PERSISTER &p, ResourceReadResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("data"), inOrOut.data);
    bool hasStringValues = !inOrOut.strings.empty();
    o.memberOptional(ISTR("strings"), inOrOut.strings, hasStringValues);
    bool hasUndefinedBits = !inOrOut.undefinedBits.empty();
    o.memberOptional(ISTR("undefinedBits"), inOrOut.undefinedBits, hasUndefinedBits);
    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};


/// ResourceWriteResult, returned by resource_write()
struct ResourceWriteResult
{
    typedef std::vector<uint64_t> U64Array;
    U64Array error;          // Optional: Empty means missing.
};

/// Read/write ResourceWriteResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ResourceWriteResult> { static void persist(PERSISTER &p, ResourceWriteResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};


// --- Memory objects

/// The map for attribute informations
typedef std::map<std::string, AttributeInfo> AttributeInfoMap;

/// TODO: Delete this later. It is only used in fastsim.
/// The map for attribute value
typedef std::map<std::string, IrisValue> AttributeValueMap;

/// The map for IrisValue
typedef std::map<std::string, IrisValue> IrisValueMap;

struct MemorySpaceInfo
{
    uint64_t spaceId;            // Mandatory.
    std::string name;            // Mandatory.
    std::string description;     // Optional: Empty means missing.
    uint64_t minAddr;            // Optional: Default is 0. Minimum address in this address space.
    uint64_t maxAddr;            // Optional: Default is 2**64-1. Maximum address in this address space (inclusive).
    uint64_t canonicalMsn;       // Optional: 2**64-1 means missing.
    std::string endianness;      // Optional: Empty means 'little'. Must be either empty, 'little', 'big', 'be32', 'variable' or 'none'.
    AttributeInfoMap attrib;     // Optional: Attributes of this memory space.
    IrisValueMap attribDefaults; // Optional: Default values for attributes of this memory space.

    MemorySpaceInfo() : spaceId(0), minAddr(0), maxAddr(IRIS_UINT64_MAX), canonicalMsn(IRIS_UINT64_MAX) {}
};

/// Read/write MemorySpaceInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,MemorySpaceInfo> { static void persist(PERSISTER &p, MemorySpaceInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("spaceId"), inOrOut.spaceId);
    o.member(ISTR("name"), inOrOut.name);
    bool hasDescription = !inOrOut.description.empty();
    o.memberOptional(ISTR("description"), inOrOut.description, hasDescription);
    bool hasMinAddr = inOrOut.minAddr == 0;
    o.memberOptional(ISTR("minAddr"), inOrOut.minAddr, hasMinAddr);
    bool hasMaxAddr = inOrOut.maxAddr != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("maxAddr"), inOrOut.maxAddr, hasMaxAddr);
    bool hasCanonicalMsn = inOrOut.canonicalMsn != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("canonicalMsn"), inOrOut.canonicalMsn, hasCanonicalMsn);
    bool hasEndianness = !inOrOut.endianness.empty();
    o.memberOptional(ISTR("endianness"), inOrOut.endianness, hasEndianness);
    bool hasAttrib = !inOrOut.attrib.empty();
    o.memberOptional(ISTR("attrib"), inOrOut.attrib, hasAttrib);
    bool hasAttribDefaults = !inOrOut.attribDefaults.empty();
    o.memberOptional(ISTR("attribDefaults"), inOrOut.attribDefaults, hasAttribDefaults);
}};

struct MemoryReadResult
{
    typedef std::vector<uint64_t> U64Array;
    U64Array data;           // Mandatory.
    U64Array error;          // Optional: List of addresses which failed to return a value. Empty means missing.
};

/// Read/write MemoryReadResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,MemoryReadResult> { static void persist(PERSISTER &p, MemoryReadResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("data"), inOrOut.data);
    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};

struct MemoryWriteResult
{
    typedef std::vector<uint64_t> U64Array;
    U64Array error;          // Optional: List of addresses which failed to return a value. Empty means missing.
};

/// Read/write MemoryWriteResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,MemoryWriteResult> { static void persist(PERSISTER &p, MemoryWriteResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};

struct MemoryAddressTranslationResult
{
    typedef std::vector<uint64_t> U64Array;
    U64Array address;
};

/// Read/write MemoryAddressTranslationResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,MemoryAddressTranslationResult> { static void persist(PERSISTER &p, MemoryAddressTranslationResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("address"), inOrOut.address);
}};

struct MemorySupportedAddressTranslationResult
{
    uint64_t inSpaceId;
    uint64_t outSpaceId;
    std::string description;

    MemorySupportedAddressTranslationResult() : inSpaceId(0), outSpaceId(0) {}
    MemorySupportedAddressTranslationResult(uint64_t inSpaceId_, uint64_t outSpaceId_, std::string description_) :
        inSpaceId(inSpaceId_), outSpaceId(outSpaceId_), description(description_) {}
};

/// Read/write MemorySupportedAddressTranslationResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,MemorySupportedAddressTranslationResult> { static void persist(PERSISTER &p, MemorySupportedAddressTranslationResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("inSpaceId"), inOrOut.inSpaceId);
    o.member(ISTR("outSpaceId"), inOrOut.outSpaceId);
    o.member(ISTR("description"), inOrOut.description);
}};


// --- Image Loading objects

struct ImageReadResult
{
    std::vector<uint64_t> data;     // Mandatory: Data bytes returned by the read. The bytes are stored in little endian in the NumberU64 values.
    uint64_t size;                  // Mandatory: Size of the data read. (The array length is insufficient as it has a granularity of 8 bytes.)

    ImageReadResult() : size(0) {}
};

/// Read/write ImageReadResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ImageReadResult> { static void persist(PERSISTER &p, ImageReadResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("data"), inOrOut.data);
    o.member(ISTR("size"), inOrOut.size);
}};

struct ImageMetaInfo
{
    std::string path;       // Mandatory.
    bool instanceSideFile;  // Mandatory.
    uint64_t instId;        // Mandatory.
    uint64_t rawAddr;       // Optional: 2**64-1 means missing.
    uint64_t rawSpaceId;    // Optional: 2**64-1 means missing.

    ImageMetaInfo() : rawAddr(IRIS_UINT64_MAX), rawSpaceId(IRIS_UINT64_MAX) {}
    ImageMetaInfo(const std::string &path_, bool instanceSide_, uint64_t instId_) : path(path_), instanceSideFile(instanceSide_), instId(instId_), rawAddr(IRIS_UINT64_MAX), rawSpaceId(IRIS_UINT64_MAX) {}
    ImageMetaInfo(const std::string &path_, bool instanceSide_, uint64_t instId_, uint64_t rawAddr_, uint64_t rawSpaceId_) : path(path_), instanceSideFile(instanceSide_), instId(instId_), rawAddr(rawAddr_), rawSpaceId(rawSpaceId_) {}
};

/// Read/write ImageMetaInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,ImageMetaInfo> { static void persist(PERSISTER &p, ImageMetaInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("path"), inOrOut.path);
    o.member(ISTR("instanceSideFile"), inOrOut.instanceSideFile);
    o.member(ISTR("instId"), inOrOut.instId);
    bool hasRawAddr = inOrOut.rawAddr != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("rawAddr"), inOrOut.rawAddr, hasRawAddr);
    bool hasRawSpaceId = inOrOut.rawSpaceId != IRIS_UINT64_MAX;
    o.memberOptional(ISTR("rawSpaceId"), inOrOut.rawSpaceId, hasRawSpaceId);
}};


// --- Breakpoint ---

/// Breakpoint info
struct BreakpointInfo
{
    uint64_t bptId;              // Mandatory. Breakpoint id.
    std::string type;            // Mandatory. Breakpoint type. Is one of: code, data or resource.
    uint64_t address;            // Optional: 2**64-1 means missing. Address or start address of range. Mandatory for code and data breakpoints
    uint64_t size;               // Optional: Default is 0. Size of address range in bytes for code and data breakpoints.
    uint64_t spaceId;            // Optional: 2**64-1 means missing. Memory space id. Mandatory for code and data breakpoints.
    uint64_t rscId;              // Optional: 2**64-1 means missing. Resource id. Mandatory for register breakpoints
    std::string rwMode;          // Optional: Is one of "r", "w" or "rw" for read-only, write-only or read-write (default). Only relevant for data and register breakpoints.
    bool dontStop;               // Optional: Iff present and True: Do not stop simulation on breakpoint hit, but still generate an IRIS_BREAKPOINT_HIT event with field DONTSTOP=True.
    uint64_t instId;             // Mandatory: Instance id of the instance which created the breakpoint.
    IrisValueMap conditions;     // Optional: The values of any additional conditions set on this breakpoint. If not present/empty
                                 //           no additional conditions are set.


    BreakpointInfo() :
        bptId(IRIS_UINT64_MAX), address(IRIS_UINT64_MAX), size(0), spaceId(IRIS_UINT64_MAX),
        rscId(IRIS_UINT64_MAX), rwMode(""), dontStop(false)
    {
    }
};

/// Read/write BreakpointInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,BreakpointInfo> { static void persist(PERSISTER &p, BreakpointInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    // Mandatory
    o.member(ISTR("bptId"), inOrOut.bptId);
    o.member(ISTR("instId"), inOrOut.instId);
    o.member(ISTR("type"), inOrOut.type);

    // Optional for different types
    bool isCodeOrData = (inOrOut.type == "code" || inOrOut.type == "data");
    o.memberOptional(ISTR("address"), inOrOut.address, isCodeOrData);
    o.memberOptional(ISTR("spaceId"), inOrOut.spaceId, isCodeOrData);
    o.memberOptional(ISTR("size"), inOrOut.size, isCodeOrData);
    bool hasRwMode = ((inOrOut.type == "data" || inOrOut.type == "register") && !inOrOut.rwMode.empty());
    o.memberOptional(ISTR("rwMode"), inOrOut.rwMode, hasRwMode);
    bool isRegister = (inOrOut.type == "register");
    o.memberOptional(ISTR("rscId"), inOrOut.rscId, isRegister);

    // Optional
    bool continueExecution = inOrOut.dontStop;
    o.memberOptional(ISTR("dontStop"), inOrOut.dontStop, continueExecution);

    bool has_conditions = inOrOut.conditions.empty();
    o.memberOptional(ISTR("conditions"), inOrOut.conditions, has_conditions);
}};

struct BreakpointAction
{
    uint64_t esId;       // Mandatory. Event stream id for eventStream* actions.
    std::string action;  // Mandatory. Must be one of: eventStream_enable or eventStream_disable.

    BreakpointAction(uint64_t esId_, bool enableEvStream) :
        esId(esId_),
        action(enableEvStream ? "eventStream_enable" : "eventStream_disable")
    {}

    BreakpointAction() : esId(IRIS_UINT64_MAX) {}
};

/// Read/write BreakpointAction.
template<class PERSISTER>
struct PersistPSH<PERSISTER,BreakpointAction> { static void persist(PERSISTER &p, BreakpointAction &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("esId"), inOrOut.esId);
    o.member(ISTR("action"), inOrOut.action);
}};

struct BreakpointConditionInfo
{
    std::string name;                  // Mandatory: Name of the condition
    std::string type;                  // Mandatory: Type of the condition value (string, numeric, etc)
    std::string description;           // Mandatory: A description of the condition. This should indicate the
                                       //            circumstances that will cause the breakpoint to trigger.
    std::vector<std::string> bptTypes; // Optional: List of breakpoint types ('code', 'data', etc) that this
                                       //           condition is applicable to.
};

/// Read/write BreakpointConditionInfo
template <typename PERSISTER>
struct PersistPSH<PERSISTER, BreakpointConditionInfo> { static void persist(PERSISTER &p, BreakpointConditionInfo &inOrOut)
{
     typename PERSISTER::Object o = p.openObject();

     o.member(ISTR("name"), inOrOut.name);
     o.member(ISTR("type"), inOrOut.type);
     o.member(ISTR("description"), inOrOut.description);
     bool has_bpt_types = !inOrOut.bptTypes.empty();
     o.memberOptional(ISTR("bptTypes"), inOrOut.bptTypes, has_bpt_types);
}};

// --- event ---

struct EventSourceFieldInfo
{
    std::string name;                   // Mandatory. Uniquely identify the event source field within an event source.
    std::string type;                   // Mandatory. May be one of "uint", "int", "bool", "float", "string" and "bytes".
    uint64_t size;                      // Mandatory. Size of the field value in bytes. 0 means variable size (for type="string" only).
    std::string description;            // Optional. Description of the event source field.
    std::vector<EnumElementInfo> enums; // Optional. Array of EnumElementInfo objects which describe symbols for numeric event field values.

    EventSourceFieldInfo() {}
    EventSourceFieldInfo(const std::string &name_, const std::string &type_, uint64_t size_,
                         const std::string descr = "") :
        name(name_), type(type_), size(size_), description(descr) {}

    void addEnumElement(uint64_t value, const std::string symble, const std::string descr = "")
    {
        enums.push_back(EnumElementInfo(IrisValue(value), symble, descr));
    }
};

/// Read/write EventSourceFieldInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,EventSourceFieldInfo> { static void persist(PERSISTER &p, EventSourceFieldInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("type"), inOrOut.type);
    o.member(ISTR("size"), inOrOut.size);

    bool hasDescription = !inOrOut.description.empty();
    o.memberOptional(ISTR("description"), inOrOut.description, hasDescription);
    bool hasEnums = !inOrOut.enums.empty();
    o.memberOptional(ISTR("enums"), inOrOut.enums, hasEnums);
}};


struct EventSourceInfo
{
    std::string name;                         // Mandatory. Uniquely identify the event source within a target instance.
    std::string description;                  // Optional. Description of the event source field.
    std::string format;                       // Optional. Iris-text-format.
    std::vector<EventSourceFieldInfo> fields; // Mandatory. Event sources which do not have any fields (which is a rare case) must return an empty array.
    bool counter;                             // Optional. If present and True this event source supports the counter features.
    bool ringBuffer;                          // Optional. If present and True this event source supports being buffered in a ring buffer.
    uint64_t evSrcId;                         // Mandatory. Opaque event source id.
    bool isHidden;                            // Is hidden event source. Do not persist this member.

    EventSourceInfo() : description(""), format(""), counter(false), ringBuffer(false), isHidden(false) {}

    EventSourceFieldInfo &addField(const std::string &name_, const std::string &type, uint64_t size, const std::string &descr)
    {
        fields.push_back(EventSourceFieldInfo(name_, type, size, descr));
        return fields.back();
    }

    const EventSourceFieldInfo *getField(const std::string &name_) const
    {
        for (std::vector<EventSourceFieldInfo>::const_iterator i = fields.begin(), e = fields.end(); i != e; i++)
            if (i->name == name_)
                return &(*i);
        return NULL;
    }
};

/// Read/write EventSourceInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,EventSourceInfo> { static void persist(PERSISTER &p, EventSourceInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("fields"), inOrOut.fields);
    o.member(ISTR("evSrcId"), inOrOut.evSrcId);

    bool hasDescription = !inOrOut.description.empty();
    o.memberOptional(ISTR("description"), inOrOut.description, hasDescription);
    bool hasFormat = !inOrOut.format.empty();
    o.memberOptional(ISTR("format"), inOrOut.format, hasFormat);
    bool isCounter = inOrOut.counter;
    o.memberOptional(ISTR("counter"), inOrOut.counter, isCounter);
    bool hasRingBuffer = inOrOut.ringBuffer;
    o.memberOptional(ISTR("ringBuffer"), inOrOut.ringBuffer, hasRingBuffer);
}};


struct EventCounterMode
{
    bool overflowReload;        // Optional. Set internal counter value to startVal on overflow.
    bool overflowDisableTrace;  // Optional. Disable event on overflow. The client must call eventDisable() to destroy the event stream.
    bool overflowStopSim;       // Optional. Stop simulation on overflow.
    bool overflowTrace;         // Optional. Call ecFOO() callback on overflow.
    bool nonOverflowTrace;      // Optional. Call ecFOO() callback for all non-overflow events.

    EventCounterMode() : overflowReload(false), overflowDisableTrace(false), overflowStopSim(false),
            overflowTrace(false), nonOverflowTrace(false) {}
};

/// Read/write EventCounterMode.
template<class PERSISTER>
struct PersistPSH<PERSISTER,EventCounterMode> { static void persist(PERSISTER &p, EventCounterMode &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    bool isOverflowReload = inOrOut.overflowReload;
    o.memberOptional(ISTR("overflowReload"), inOrOut.overflowReload, isOverflowReload);
    bool isOverflowDisableTrace = inOrOut.overflowDisableTrace;
    o.memberOptional(ISTR("overflowDisableTrace"), inOrOut.overflowDisableTrace, isOverflowDisableTrace);
    bool isOverflowStopSim = inOrOut.overflowStopSim;
    o.memberOptional(ISTR("overflowStopSim"), inOrOut.overflowStopSim, isOverflowStopSim);
    bool isOverflowTrace = inOrOut.overflowTrace;
    o.memberOptional(ISTR("overflowTrace"), inOrOut.overflowTrace, isOverflowTrace);
    bool isNonOverflowTrace = inOrOut.nonOverflowTrace;
    o.memberOptional(ISTR("nonOverflowTrace"), inOrOut.nonOverflowTrace, isNonOverflowTrace);
}};


// --- Simulation time execution control ---

struct SimulationTimeObject
{
    uint64_t ticks;     // Mandatory. Current simulation time in ticks. One tick is 1/tickHz seconds long.
    uint64_t tickHz;    // Mandatory. Time resolution of the ticks value in Hz. Example: 1000 means 1 tick = 1 ms.
    bool running;       // Iff True the simulation time is running, else it is stopped.

    SimulationTimeObject()
        : ticks(IRIS_UINT64_MAX)
        , tickHz(IRIS_UINT64_MAX)
        , running(false)
    {
    }
};

/// Read/write SimulationTimeObject.
template<class PERSISTER>
struct PersistPSH<PERSISTER,SimulationTimeObject> { static void persist(PERSISTER &p, SimulationTimeObject &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("ticks"), inOrOut.ticks);
    o.member(ISTR("tickHz"), inOrOut.tickHz);
    o.member(ISTR("running"), inOrOut.running);
}};

// --- disassembly ---
typedef std::map<std::string, IrisValue> DisassembleContext;

struct DisassemblyMode
{
    std::string name;
    std::string descr;
    DisassemblyMode() {}
    DisassemblyMode(std::string name_, std::string descr_) : name(name_), descr(descr_) {}
};

/// Disassemble ModeInfo
template<class PERSISTER>
struct PersistPSH<PERSISTER,DisassemblyMode> { static void persist(PERSISTER &p, DisassemblyMode &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("descr"), inOrOut.descr);
}};

struct DisassemblyLine
{
    uint64_t addr;
    std::string opcode;
    std::string disass;
    DisassemblyLine() : addr(0), opcode(""), disass(""){}
    DisassemblyLine(uint64_t addr_, std::string opcode_, std::string disass_) :
                    addr(addr_), opcode(opcode_), disass(disass_) {}
};

/// DisassembleInfo
template<class PERSISTER>
struct PersistPSH<PERSISTER,DisassemblyLine> { static void persist(PERSISTER &p, DisassemblyLine &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();
    o.member(ISTR("addr"), inOrOut.addr);
    o.member(ISTR("opcode"), inOrOut.opcode);
    o.member(ISTR("disass"), inOrOut.disass);
}};

// --- table ---

struct TableColumnInfo
{
    std::string name;         // Column title and record field name.
    std::string description;  // Description of the record field, describing the semantics of the record fields in free form.
    std::string format;       // Optional: Iris-text-format.
    std::string type;         // Value type of this record field. Must be one of the following: "NumberU64", "NumberS64", "Array of NumberU64", "Boolean", "String".
    uint64_t bitWidth;            // The interpretation of 'bitWidth' depends on the value of 'type'.
    std::string formatShort;  // Optional: Iris-Text-Format specification when displaying the value in a textual form. Must only be a single line and should be short.
    std::string formatLong;   // Optional: Iris-Text-Format specification when displaying the value in a textual form. May contain multiple lines and should be self descriptive.
    std::string rwMode;       // Optional: Either "r" or "rw" for read-only or read-write (default).

    TableColumnInfo()
        : bitWidth(0)
        , rwMode("rw")
    {
    }

    TableColumnInfo(std::string name_,std::string type_, uint64_t bitWidth_)
        : name(name_)
        , type(type_)
        , bitWidth(bitWidth_)
        , rwMode("rw")
    {
    }
};
/// Read/write TableColumnInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableColumnInfo> { static void persist(PERSISTER &p, TableColumnInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("description"), inOrOut.description);
    o.member(ISTR("type"), inOrOut.type);
    o.member(ISTR("bitWidth"), inOrOut.bitWidth);
    o.member(ISTR("rwMode"), inOrOut.rwMode);

    bool hasFormat = !inOrOut.format.empty();
    o.memberOptional(ISTR("format"), inOrOut.format, hasFormat);
    bool hasFormatShort = !inOrOut.formatShort.empty();
    o.memberOptional(ISTR("formatShort"), inOrOut.formatShort, hasFormatShort);
    bool hasFormatLong = !inOrOut.formatLong.empty();
    o.memberOptional(ISTR("formatLong"), inOrOut.formatLong, hasFormatLong);
}};

typedef std::vector<TableColumnInfo> TableColumnInfos;

struct TableInfo
{
    std::string name;                      // Name of the table.
    std::string description;               // Description of the table, describing the semantics of the table in free form.
    uint64_t tableId;                      // Opaque table id used to read and write table contents.
    TableColumnInfos columns;              // List of meta information for the columns in the table.
    uint64_t minIndex;                     // Optional: Minimum row index (inclusive). Default is 0.
    uint64_t maxIndex;                     // Optional: Maximum row index (inclusive). Default is 2**64-1.
    std::string indexFormatHint;           // Optional: Hint for clients on how to display the index of a record. Must be "hide", "dec" or "hex". Default is "hex".
    std::string formatShort;               // Optional: If present then clients should format the data of each row according to this format and should not (by default) show the individual columns.
    std::string formatLong;                // Optional: Long version of 'formatShort' intended to be displayed in a bubble help when hovering over a table row.

    TableInfo()
        : tableId(0)
        , minIndex(0)
        , maxIndex(IRIS_UINT64_MAX)
        , indexFormatHint("hex")
    {
    }
};

typedef std::vector<TableInfo> TableInfoList;

/// Read/write TableInfo.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableInfo> { static void persist(PERSISTER &p, TableInfo &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("description"), inOrOut.description);
    o.member(ISTR("tableId"), inOrOut.tableId);
    o.member(ISTR("columns"), inOrOut.columns);
    o.member(ISTR("minIndex"), inOrOut.minIndex);
    o.member(ISTR("maxIndex"), inOrOut.maxIndex);
    o.member(ISTR("indexFormatHint"), inOrOut.indexFormatHint);

    bool hasFormatShort = !inOrOut.formatShort.empty();
    o.memberOptional(ISTR("formatShort"), inOrOut.formatShort, hasFormatShort);
    bool hasFormatLong = !inOrOut.formatLong.empty();
    o.memberOptional(ISTR("formatLong"), inOrOut.formatLong, hasFormatLong);
}};

struct TableRecord
{
    uint64_t index;        // Index of this record.
    IrisValueMap row;      // Name, type and semantics as described in TableColumnInfo in TableInfo.

    TableRecord() : index(IRIS_UINT64_MAX) {}
    TableRecord(uint64_t index_) : index(index_) {}

    void addCell(const std::string &name, const IrisValue &value) { row[name] = value; }
};
/// Read/write TableRecord.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableRecord> { static void persist(PERSISTER &p, TableRecord &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("index"), inOrOut.index);
    o.member(ISTR("row"), inOrOut.row);
}};

struct TableCellError
{
    uint64_t index;      // Index in table (row number).
    std::string name;    // Column name (field name) as in TableColumnInfo.
    int64_t errorCode;  // E_* error code.

    TableCellError() : index(IRIS_UINT64_MAX) {}
    TableCellError(uint64_t index_, std::string name_, uint64_t errorCode_) :
        index(index_), name(name_), errorCode(errorCode_) {}
};
/// Read/write TableCellError.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableCellError> { static void persist(PERSISTER &p, TableCellError &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("index"), inOrOut.index);
    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("errorCode"), inOrOut.errorCode);
}};

typedef std::vector<TableRecord> TableRecords;
typedef std::vector<TableCellError> TableCellErrors;

struct TableReadResult
{
    TableRecords data;      // The actual data read from the table.
    TableCellErrors error;  // Optional: List of errors which occurred while reading one or more table cells.
};
/// Read/write TableReadResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableReadResult> { static void persist(PERSISTER &p, TableReadResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("data"), inOrOut.data);
    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};


struct TableWriteResult
{
    TableCellErrors error;  // Optional: List of errors which occurred while writing one or more table cells.
};
/// Read/write TableWriteResult.
template<class PERSISTER>
struct PersistPSH<PERSISTER,TableWriteResult> { static void persist(PERSISTER &p, TableWriteResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    bool hasError = !inOrOut.error.empty();
    o.memberOptional(ISTR("error"), inOrOut.error, hasError);
}};

// --- Instantiation ---

struct InstantiationError
{
    std::string severity;      // Severity of the error or warning. Always one of "error" or "warning".
    std::string code;          // Error or warning code as a string symbol.
    std::string message;       // Free format error or warning message, potentially multiple lines long.
    std::string parameterName; // Optional: Name of the offending parameter for parameter related errors.
                               //           Mandatory for parameter related errors. Must not be present
                               //           for other errors/warnings.
};

/// Read/write InstantiationError.
template<typename PERSISTER>
struct PersistPSH<PERSISTER,InstantiationError> { static void persist(PERSISTER &p, InstantiationError &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("code"), inOrOut.code);
    o.member(ISTR("message"), inOrOut.message);
    bool has_parameter_name = !inOrOut.parameterName.empty();
    o.memberOptional(ISTR("parameterName"), inOrOut.parameterName, has_parameter_name);
    o.member(ISTR("severity"), inOrOut.severity);
}};

struct InstantiationResult
{
    bool success;                           // If True the simulation was instantiated successfully.
                                            // In this case 'errors' may only contain warnings or may be empty.
                                            // If False the simulation was not instantiated.
                                            // In this case 'errors' contains at least one error.
    std::vector<InstantiationError> errors; // List of errors and/or warnings which occurred during instantiation.

    InstantiationResult()
        : success(false)
    {
    }
};

/// Read/write InstantiationResult.
template<typename PERSISTER>
struct PersistPSH<PERSISTER,InstantiationResult> { static void persist(PERSISTER &p, InstantiationResult &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("success"), inOrOut.success);
    o.member(ISTR("errors"), inOrOut.errors);
}};

struct InstantiationParameterValue
{
    /// Name of the parameter to set. This is the same as the ResourceInfo.name field for this parameter.
    std::string name;

    /// Value of this parameter. Can be a string or array of NumberU64.
    IrisValue value;

    /// Default constructor
    InstantiationParameterValue() {}

    /// Construct a numeric parameter value
    InstantiationParameterValue(const std::string name_, uint64_t uint_value)
        : name(name_)
    {
        std::vector<uint64_t> data(1, uint_value);
        IrisU64JsonWriter::writeU64ArrayToIrisValue(data, value);
    }

    /// Construct a numericSigned parameter value
    InstantiationParameterValue(const std::string name_, int64_t int_value)
        : name(name_)
    {
        std::vector<uint64_t> data(1, static_cast<uint64_t>(int_value));
        IrisU64JsonWriter::writeU64ArrayToIrisValue(data, value);
    }

    /// Construct a numericFp parameter value
    InstantiationParameterValue(const std::string name_, double dbl_value)
        : name(name_)
    {
        const void *vp = &dbl_value;
        std::vector<uint64_t> data(1, *reinterpret_cast<const uint64_t*>(vp));
        IrisU64JsonWriter::writeU64ArrayToIrisValue(data, value);
    }

    /// Construct a string parameter value
    InstantiationParameterValue(const std::string name_, const std::string &value_)
        : name(name_)
        , value(value_)
    {
    }

    /// Construct a large (> 64-bit) numeric parameter value
    InstantiationParameterValue(const std::string name_, const std::vector<uint64_t> &value_)
        : name(name_)
    {
        IrisU64JsonWriter::writeU64ArrayToIrisValue(value_, value);
    }

    /// Set the value by parsing a string representation of the value using information from the
    /// ResourceInfo object for this parameter.
    /// This function does not check the validity of the value, it just uses the ResourceInfo
    /// to determine the type.
    IrisErrorCode setFromStringAndInfo(const::std::string &str_value, const ResourceInfo &info);

    /// Return true if this parameter value is a string.
    bool isString() const
    {
        return value.getType() == IrisValue::String;
    }

    /// Return true if this parameter value is numeric.
    bool isNumeric() const
    {
        return value.getType() == IrisValue::Array;
    }

    IrisErrorCode getValue(std::string &value_out) const
    {
        if (isString())
        {
            value_out = value.getString();
            return E_ok;
        }
        else
        {
            return E_type_mismatch;

        }
    }

    IrisErrorCode getValue(std::vector<uint64_t> &array_out) const
    {
        if (isNumeric())
        {
            try
            {
                iris::IrisU64JsonReader::readU64ArrayFromIrisValue(array_out, value);
                return E_ok;
            }
            catch (IrisErrorException &)
            {
                return E_type_mismatch;
            }
        }
        else
        {
            return E_type_mismatch;
        }
    }

    IrisErrorCode getValue(uint64_t &value_out) const
    {
        std::vector<uint64_t> u64arr_value;
        IrisErrorCode code = getValue(u64arr_value);
        if (code == E_ok)
        {
            if (u64arr_value.size() > 1)
            {
                code = E_data_size_error;
            }
            else
            {
                value_out = u64arr_value[0];
            }
        }
        return code;
    }

    IrisErrorCode getValue(int64_t &value_out) const
    {
        uint64_t u64_value = 0;
        IrisErrorCode code = getValue(u64_value);
        if (code == E_ok)
        {
            value_out = static_cast<int64_t>(u64_value);
        }
        return code;
    }
};

/// Read/write InstantiationParameterValue
template<typename PERSISTER>
struct PersistPSH<PERSISTER,InstantiationParameterValue> { static void persist(PERSISTER &p, InstantiationParameterValue &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("name"), inOrOut.name);
    o.member(ISTR("value"), inOrOut.value);
}};

// --- SyncLevel ---

struct SyncLevelState
{
    uint64_t syncLevel; //! Current effective sync-level. This is always the highest requested sync-level.
    std::vector<uint64_t> counts; //! Count of requests for each sync-level

    SyncLevelState()
        : syncLevel(0)
        , counts(4, 0)
    {
    }
};

template<typename PERSISTER>
struct PersistPSH<PERSISTER,SyncLevelState> { static void persist(PERSISTER &p, SyncLevelState &inOrOut)
{
    typename PERSISTER::Object o = p.openObject();

    o.member(ISTR("syncLevel"), inOrOut.syncLevel);
    o.member(ISTR("counts"), inOrOut.counts);
}};

// --- Generic objects and their persist functions. ---------------------------------------------------------------

/// Read vector<uint64_t>. Specialization for Array of NumberU64.
template<>
struct PersistPSH<IrisU64JsonReader, std::vector<uint64_t> > { static void persist(IrisU64JsonReader &reader, std::vector<uint64_t> &out)
{
    reader.persist(out);
}};

/// Write vector<uint64_t>. Specialization for Array of NumberU64.
template<>
struct PersistPSH<IrisU64JsonWriter, std::vector<uint64_t> > { static void persist(IrisU64JsonWriter &writer, std::vector<uint64_t> &in)
{
    writer.persist(in);
}};

/// Read vector<T>.
/// out is overwritten.
template<class T>
struct PersistPSH<IrisU64JsonReader,std::vector<T> > { static void persist(IrisU64JsonReader &reader, std::vector<T> &out)
{
    IrisU64JsonReader::Array a = reader.openArray();
    out.resize(a.getLength());
    for (uint64_t i = 0; i < a.getLength(); i++)
        a.element(out[i]);
}};

/// Write vector<T>.
template<class T>
struct PersistPSH<IrisU64JsonWriter,std::vector<T> > { static void persist(IrisU64JsonWriter &writer, const std::vector<T> &in)
{
    IrisU64JsonWriter::Array a = writer.openArray();
    for (size_t  i = 0; i < in.size(); i++)
        a.element(in[i]);
}};
    
    
/// Read map<string,T>.
/// out is overwritten.
template<class T>
struct PersistPSH<IrisU64JsonReader,std::map<std::string,T> > { static void persist(IrisU64JsonReader &reader, std::map<std::string,T> &out)
{
    IrisU64JsonReader::Object o = reader.openObject();
    out.clear();
    std::string name;
    T value;
    while (o.readNextMember(name, value))
        out[name] = value;
}};
    
/// Write map<string,T>.
template<class T>
struct PersistPSH<IrisU64JsonWriter,std::map<std::string,T> > { static void persist(IrisU64JsonWriter &writer, const std::map<std::string,T> &in)
{
    IrisU64JsonWriter::Object o = writer.openObject();
    for (typename std::map<std::string,T>::const_iterator i = in.begin(); i != in.end(); i++)
        o.memberSlow(i->first, i->second);
}};
    




NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisObjects_h

