/*!
 *  \file   IrisPendingRequestSet.h
 *  \brief  Set of pending request which wait for a response.
 *  \date   Copyright ARM Limited 2014-2015 All Rights Reserved.
 *
 *  Set of pending request which wait for a response.
 */

#ifndef ARM_INCLUDE_IrisPendingRequestSet_h
#define ARM_INCLUDE_IrisPendingRequestSet_h

#include "iris/detail/IrisCommon.h"

#include "IrisInterface.h"
#include "IrisRequest.h"
#include "IrisThreadAbstraction.h"
#include "IrisLogger.h"
#include <cassert>
#include <set>

NAMESPACE_IRIS_START


class IrisPendingRequestSet: public IrisInterface
{
public:
    /// Constructor.
    IrisPendingRequestSet(IrisLogger &log_): threadAbstraction(0), responseArrivedCondVar(0), pendingRequestSetMutex(0), log(log_) {}


    /// Destructor.
    virtual ~IrisPendingRequestSet()
    {
        delete responseArrivedCondVar;
        delete pendingRequestSetMutex;
    }


    /// Set thread abstraction.
    /// It is mandatory to set a thread abstraction before using this object.
    /// This function must only be called once.
    void init(IrisThreadAbstraction *threadAbstraction_)
    {
        assert(threadAbstraction == 0);
        assert(threadAbstraction_ != 0);
        threadAbstraction = threadAbstraction_;
        responseArrivedCondVar = threadAbstraction->createConditionVariable();
        pendingRequestSetMutex = threadAbstraction->createMutex();
    }


    /// Add request which waits for a response.
    /// The caller continues to own the memory of the request. The request must not be freed until
    /// a response arrived for it (until request.hasResponse() became true).
    /// Requests autamtically get removed from this set once a response arrived for them.
    void addRequest(IrisRequest *req)
    {
        IrisThreadAbstractionLock lock(*pendingRequestSetMutex);
        assert(pendingRequestSet.find(req) == pendingRequestSet.end());
        pendingRequestSet.insert(req);
    }

    /// Erase a pending request. No need to wait for the response.
    void erasePendingRequest(IrisRequest *req)
    {
        IrisThreadAbstractionLock lock(*pendingRequestSetMutex);
        pendingRequestSet.erase(req);
    }


    /// Wait for any response to come in.
    /// This function blocks until a response comes in.
    /// There is no guarantee that any of the pending requests received a response when this function returns.
    /// Once a response comes in for request r r.hasResponse() becomes true.
    void waitForResponse(const IrisRequest &req)
    {
        responseArrivedCondVar->lock();
        while (!req.hasResponse())
            responseArrivedCondVar->wait();
        responseArrivedCondVar->unlock();
    }


    /// Get Iris interface.
    /// Responses get fed into this.
    /// irisHandleMessage() must only be called with response messages on this interface.
    IrisInterface *getIrisInterface() { return this; }


    // --- IrisInterface implementation (only responses are handled)

    /// Handle incoming response.
    /// This is called from any thread.
    /// This class cannot handle incoming requests. Pass incoming requests to an IrisFunctionDecoder
    /// and splice incoming responses from the function decoder to this object using
    /// IrisFunctionDecoder::forwardResponsesTo().
    virtual void irisHandleMessage(const uint64_t *message) IRIS_OVERRIDE
    {
        if (IrisU64JsonReader::isRequest(message))
            throw IrisInternalError("IrisPendingRequestSet::irisHandleMessage() must only be called with responses.");

        // Get request id, fast.
        uint64_t requestId = IrisU64JsonReader::getRequestId(message);

        // Find request for this response.
        IrisRequest *req = NULL;
        {
            IrisThreadAbstractionLock lock(*pendingRequestSetMutex);
            for (PendingRequestSet::iterator i = pendingRequestSet.begin(); i != pendingRequestSet.end(); i++)
            {
                if ((*i)->getRequestId() == requestId)
                {
                    req = *i;
                    pendingRequestSet.erase(i);
                    break;
                }
            }
        }

        if (req == NULL)
        {
            // No request is waiting for this response. This is not fatal for us, we just ignore it and log the warning.
            log.warning("IrisPendingRequestSet: Received response with request id 0x%016llx but no such"
                        " request is waiting for a response (numPendingRequests=%lld).\n",
                        (long long)requestId, (long long)pendingRequestSet.size());
        }
        else
        {
            // Found request.
            responseArrivedCondVar->lock();
            req->setResponse(message);
            responseArrivedCondVar->notifyAll();
            responseArrivedCondVar->unlock();
        }
    }

    virtual void *irisObtainInterface(IrisInstId /*instId*/, const char * /*requested_interface*/) IRIS_OVERRIDE
    {
        throw IrisErrorExceptionString(E_function_not_supported_by_instance, "obtainInterface() is reserved for future use.");
    }

private:
    /// Thread abstraction.
    IrisThreadAbstraction *threadAbstraction;

    /// Event which gets set when a new response arrives.
    IrisThreadAbstractionConditionVariable *responseArrivedCondVar;

    /// Mutex which protects pendingRequestSet.
    IrisThreadAbstractionMutex *pendingRequestSetMutex;

    /// Set of pending requests.
    /// This is accessed by multiple threads and is protected by mutex pendingRequestSetMutex.
    typedef std::set<IrisRequest*> PendingRequestSet;
    PendingRequestSet pendingRequestSet;

    /// Logger.
    IrisLogger &log;
}; // class IrisPendingRequestSet

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisPendingRequestSet_h

