/*!
 *  \file   IrisPersistPSH.h
 *  \brief  Iris persist partial specialisation helper.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Iris persist partial specialisation helper. This file provides two completely independent things:
 *  1. Meachanism to provide partial function template specialisation which is not directly supported by C++ (PersistPSH).
 *  2. Base template forwarding persist(PERSISTER,T) to PERSISTER.persist(T).
 * 
 *  This file is used by IrisObjects.h, IrisU64JsonReader.h and IrisU64JsonWriter.h.
 */

#ifndef ARM_INCLUDE_IrisPersistPSH_h
#define ARM_INCLUDE_IrisPersistPSH_h

#include "iris/detail/IrisCommon.h"
NAMESPACE_IRIS_START


/// Note on PersistPSH:
/// PersistPSH means PersistPartialSpecialisationHelper.
/// In C++ function templates cannot be partially specialised while class templates can be
/// partially specialised. PersistPSH is a class template which always just contains one static
/// member function persist().
///
/// What we want:
/// A function template "template<class PERSISTER, class T> void persist(PERSISTER &p, T &inOrOut);"
/// which can be partially specialised for all relevant types in Iris and/or for the two persisters
/// IrisU64JsonReader and IrisU64JsonWriter.
/// This is not possible in C++ since function templates cannot be partially specialised. Always
/// fully specialising them would be possible but then we would need to repeat exactly the same code
/// for reading and writing which defeats the whole point of using a persister template.
///
/// What we do instead:
/// We follow the example 4 illustrated in Herb Sutters article "Why not specialising function templates?"
/// (http://www.gotw.ca/publications/mill17.htm).
/// 1. We define a single function template "template<class PERSISTER, class T> void persist(PERSISTER &p, T &inOrOut)"
/// which redirects to a static member function 'persist()' of a class template
/// "template<class PERSISTER, class T> struct PersistPSH;".
/// We do not specialise this function template at all.
/// 2. We declare the unspecialised base template class "template<class PERSISTER, class T> struct PersistPSH;" and define
/// as many specialisations of it as we want. The specialisation rules for _class_ templates apply, not the ones
/// for function templates.
/// 3. Just by chance we have a good use for the unspecialised base template class: We forward all calls to overloaded
/// persist() member functions of the persisters (IrisU64JsonReader and IrisU64JsonWriter). This is not necessary to make
/// 1.+2. work and it is unrelated to the function template vs. class template problem.


/// Declare the PersistPSH (PersistPartialSpecialisationHelper) class template. This is specialised for all Iris objects like
/// InstanceInfo etc. (See item 2. in the explanation of PersistPSH above.)
/// Also: Define the base class template to forward all persist() calls to the persisters. (See item 3. in the explanation
/// of PersistPSH above.)
/// Note: If you get an error like "no matching member function for call to 'persist'" in this function it means that you
/// should implement a specialized struct PersistPSH, see below.
/// PERSITER is either IrisU64JsonReader or IrisU64JsonWriter.
/// T is any type.
template<class PERSISTER, class T>
struct PersistPSH { static void persist(PERSISTER &p, T &inOrOut)
{
    p.persist(inOrOut);
}};


/// Define the persist() function template.
/// It always forwards to the static member function persist() in the PersistPSH class template.
/// This function template must never be specialised. Specialise PersistPSH instead (which is done in IrisOjects.h for example).
/// (See item 1. in the explanation of PersistPSH above.)
template<class PERSISTER, class T>
void persist(PERSISTER &p, T &inOrOut) { PersistPSH<PERSISTER, T>::persist(p, inOrOut); }
    
/// Do the same in contexts where we just have a 'const T&', which usually happens in writing contexts.
/// This discards the const qualifier, but with the symmetric persist() approach, sharing one function for
/// read and write we lose this compiler checking anyway.
/// We trust that writers never modify their input data.
template<class PERSISTER, class T>
void persist(PERSISTER &p, const T &inOrOut) { PersistPSH<PERSISTER, T>::persist(p, const_cast<T&>(inOrOut)); }

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisPersistPSH_h

