/*!
 *  \file   IrisReceivedRequest.h
 *  \brief  Iris request received by a callee.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  This object holds all relevant data of an Iris request and notification in a callee.
 *  It provides utility functions to conveniently access function arguments and generate return values.
 */

#ifndef ARM_INCLUDE_IrisReceivedRequest_h
#define ARM_INCLUDE_IrisReceivedRequest_h

#include "iris/detail/IrisCommon.h"
#include "IrisInterface.h"
#include "IrisErrorCode.h"
#include "IrisU64JsonReader.h"
#include "IrisU64JsonWriter.h"
#include "IrisObjects.h"
#include "IrisJsonParser.h"
#include <string>

NAMESPACE_IRIS_START

class IrisReceivedRequest
{
public:
    /// Constructor.
    IrisReceivedRequest(const uint64_t *message_)
        : requestReader(message_)
        , message(message_)
    {
        // Set functionName, request
        decode();
    }

    const uint64_t* getMessage()
    {
        return message;
    }

    // --- Reading request/notification (reading args) interface ---

    /// Return true iff this is a request, returns false if this is a notification.
    bool isRequest() const
    {
        return !request.isNotification();
    }

    /// Get function name.
    std::string getFunctionName() const
    {
        return functionName;
    }

    /// Get request id of request.
    uint64_t getRequestId() const
    {
        return isRequest()? request.getRequestId() : 0;
    }

    /// Get instId of caller.
    IrisInstId getCallerInstId() const
    {
        return getRequestId() >> 32;
    }

    /// Get mandatory string argument.
    std::string getArgString(const IrisU64StringConstant &argName) const
    {
        std::string r;
        getArg(argName, r);
        return r;
    }

    /// Get optional string argument.
    std::string getOptionalArgString(const IrisU64StringConstant &argName, const std::string &defaultValue) const
    {
        std::string r = defaultValue;
        getOptionalArg(argName, r);
        return r;
    }

    /// Get mandatory bool argument.
    bool getArgBool(const IrisU64StringConstant &argName) const
    {
        bool r;
        getArg(argName, r);
        return r;
    }

    /// Get optional bool argument.
    bool getOptionalArgBool(const IrisU64StringConstant &argName, bool defaultValue) const
    {
        bool r = defaultValue;
        getOptionalArg(argName, r);
        return r;
    }

    /// Get mandatory uint64_t argument.
    uint64_t getArgU64(const IrisU64StringConstant &argName) const
    {
        uint64_t r;
        getArg(argName, r);
        return r;
    }

    /// Get optional uint64_t argument.
    uint64_t getOptionalArgU64(const IrisU64StringConstant &argName, uint64_t defaultValue) const
    {
        uint64_t r = defaultValue;
        getOptionalArg(argName, r);
        return r;
    }

    /// Get mandatory int64_t argument.
    int64_t getArgS64(const IrisU64StringConstant &argName) const
    {
        int64_t r;
        getArg(argName, r);
        return r;
    }

    /// Get optional int64_t argument.
    int64_t getOptionalArgS64(const IrisU64StringConstant &argName, int64_t defaultValue) const
    {
        int64_t r = defaultValue;
        getOptionalArg(argName, r);
        return r;
    }

    /// Get mandatory arg of any type.
    /// Use this also for IrisValue and any of the types/structs defined in IrisObjects.h.
    template<class T>
    void getArg(const IrisU64StringConstant &argName, T &out) const
    {
        try
        {
            request.param(argName, out);
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            throw e;
        }
        catch(IrisErrorUnknownMember &e)
        {
            e.setErrorCode(E_missing_mandatory_argument);
            e.setMessage("Missing mandatory argument.");
            throw e;
        }
    }

    /// Get mandatory arg of any type.
    /// Use this also for IrisValue and any of the types/structs defined in IrisObjects.h.
    template<class T>
    void getArgSlow(const std::string &argName, T &out) const
    {
        try
        {
            request.param(argName, out);
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            throw e;
        }
        catch(IrisErrorUnknownMember &e)
        {
            e.setErrorCode(E_missing_mandatory_argument);
            e.setMessage("Missing mandatory argument.");
            throw e;
        }
    }

    /// Get optional arg of any type.
    /// Returns true if the argument was present, else false.
    template<class T>
    bool getOptionalArg(const IrisU64StringConstant &argName, T &out) const
    {
        try
        {
            return request.paramOptional(argName, out);
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            throw e;
        }
    }

    /// Get optional arg of any type.
    /// Returns true if the argument was present, else false.
    template<class T>
    bool getOptionalArgSlow(const std::string &argName, T &out) const
    {
        try
        {
            return request.paramOptional(argName, out);
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            throw e;
        }
    }
    
    
    /// Read mandatory array argument.
    IrisU64JsonReader::Array getArgArray(const IrisU64StringConstant &argName)
    {
        try
        {
            request.param(argName);
            return requestReader.openArray();
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            e.setName(argName);
            throw e;
        }
        catch(IrisErrorUnknownMember &e)
        {
            e.setErrorCode(E_missing_mandatory_argument);
            e.setMessage("Missing mandatory argument.");
            throw e;
        }
    }


    /// Read mandatory object argument.
    IrisU64JsonReader::Object getArgObject(const IrisU64StringConstant &argName)
    {
        try
        {
            request.param(argName);
            return requestReader.openObject();
        }
        catch(IrisErrorTypeMismatch &e)
        {
            e.setErrorCode(E_argument_type_mismatch);
            e.setName(argName);
            throw e;
        }
        catch(IrisErrorUnknownMember &e)
        {
            e.setErrorCode(E_missing_mandatory_argument);
            e.setMessage("Missing mandatory argument.");
            throw e;
        }
    }


    // --- Generating response interface ---

    /// Generate error response and return it.
    /// This may be called also after startOkResponse and overwrites a partial generated OK response.
    const uint64_t *generateErrorResponse(IrisErrorCode errorCode, const std::string &errorMessage, const std::string &data = std::string())
    {
        responseWriter.reset();
        responseWriterResponse = responseWriter.openResponse(getCallerInstId());
        responseWriterResponse.setErrorSlow(errorCode, errorMessage);
        if (!data.empty())
        {
            // Add data value to the end of the response
            convertJsonToU64Json(data, responseWriter);
        }
        responseWriterResponse.setRequestId(getRequestId());
        responseWriterResponse.close();
        return responseWriter.getMessage();
    }

    /// Generate ok null response (= function returning void in C++).
    /// This must not be called after another response has been started or generated yet (e.g. with generateErrorResponse()).
    const uint64_t *generateOkNullResponse()
    {
        responseWriterResponse = responseWriter.openResponse(getCallerInstId());
        responseWriterResponse.setRequestId(getRequestId());
        // By default the request is initialised to an Ok null response so nothing left to set
        responseWriterResponse.close();
        return responseWriter.getMessage();
    }

    /// Start ok response and return result writer.
    /// The caller must use the returned IrisU64JsonWriter to write the value of the
    /// 'result' member of the response object. At least a 'null' value must be written.
    /// This must not be called after another response has been started or generated yet (e.g. with generateErrorResponse()).
    IrisU64JsonWriter &startOkResponse()
    {
        assert(responseWriter.isEmpty());
        responseWriterResponse = responseWriter.openResponse(getCallerInstId());
        responseWriterResponse.setRequestId(getRequestId());
        responseWriterResponse.setResult(); // Caller will write result value after startOkResponse().
        return responseWriter;
    }

    /// Finalize ok response and return it.
    /// This must be called after startOkResponse() and after the result value was written.
    /// The return value can be passed to IrisInterface::irisHandleResponse().
    const uint64_t *endOkResponse()
    {
        responseWriterResponse.close();
        return responseWriter.getMessage();
    }

    /// Generate ok response returning a non-null value.
    /// This must not be called after another response has been started or generated yet (e.g. with generateErrorResponse()).
    template <typename RESULT_T>
    const uint64_t *generateOkResponse(const RESULT_T &result)
    {
        iris::persist(startOkResponse(), result);
        return endOkResponse();
    }


private:
    /// Decode incoming request. requestReader must be initialized with the incoming message.
    void decode()
    {
        try
        {
            request = requestReader.openRequest();
            functionName = request.getMethod();
        }
        catch(IrisErrorException &)
        {
            throw IrisErrorExceptionCString(E_malformatted_request, "Received malformatted request. Expecting optional id (NumberU64), mandatory method (String), optional params (Object).");
        }
    }
    
    /// Read request function name and arguments.
    IrisU64JsonReader requestReader;

    /// Function name.
    std::string functionName;

    /// Helper for reading function arguments.
    IrisU64JsonReader::Request request;

    /// Write response.
    IrisU64JsonWriter responseWriter;

    /// Helper for writing the response object (remembering the state between startOkResposnse() and endOkResponse()).
    IrisU64JsonWriter::Response responseWriterResponse;

    /// Iris message.
    const uint64_t *message;
};



NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisReceivedRequest_h

