/*!
 *  \file   IrisRequest.h
 *  \brief  Iris request object. Allow efficient creation aof and acces to function arguments and return values.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  This object holds all relevant data of an Iris request (and notification) and its response.
 *  It provides utility functions to conveniently access function arguents and generate
 *  and access return values.
 */

#ifndef ARM_INCLUDE_IrisRequest_h
#define ARM_INCLUDE_IrisRequest_h

#include "iris/detail/IrisCommon.h"
#include "IrisUtils.h"
#include "IrisErrorCode.h"
#include "IrisInterface.h"
#include "IrisU64JsonReader.h"
#include "IrisU64JsonWriter.h"
#include "IrisJsonProducer.h"
#include "IrisU64StringConstant.h"
#include <string>
#include <sstream>

NAMESPACE_IRIS_START

/// Request on the caller side.
/// This may represent a request or a notification.
/// This allows to add function name, arguments and allows to retrieve the response (error or result).
/// (See IrisReceivedRequest for a representation of a request on the callee side.)
class IrisRequest
{
public:
    /// Constructor.
    IrisRequest(bool split_ = false)
        : responseMessage(NULL)
        , hasResponseState(false)
        , errorCode(E_no_response_yet)
        , isRequestState(true)
        , isSplit_(split_)
        , is_notification(true) // assume we are a notification until a request id is set
        , requestId(0)
        , canceled(false)
    {
    }

    /// Destructor.
    virtual ~IrisRequest()
    {
        delete[] responseMessage;
    }

    // --- Reading aspects of the request. ---

    /// Return true iff we have initialized a request or notification.
    bool hasRequest() const { return !requestWriter.isEmpty(); }

    /// Return true iff this is a request, returns false if this is a notification.
    bool isRequest() const { return isRequestState; }

    /// Set this request to be split or not
    void setSplit(bool isSplit__) { isSplit_ = isSplit__; }

    /// Return true iff this request should be dealt with as a split request/response.
    /// This does not affect the behavior of this object but it will affect the behavior of for example IrisCppAdapter.
    bool isSplit() const { return isSplit_; }

    /// Get request id. (Only valid if this is not a notification.)
    uint64_t getRequestId() const { return requestId; }

    // --- Generating request/notification interface ---

    /// Initialize request (set function name, destination instance and request id).
    /// This function must be called at the start when generating a request message. After this
    /// function the addArg() functions are usually called to add parameters.
    template<class STRING> // IrisU64StringConstant or std::string
    void initRequest(const STRING &functionName, IrisInstId dstInstId)
    {
        assert(isRequestState == true);
        requestWriterRequestObject = requestWriter.openRequest(functionName, dstInstId);
        // addArg() will be called in the following to add parameters to the params object.
    }

    /// Initialize request (set function name and destination instance).
    template<class STRING> // IrisU64StringConstant or std::string
    void initNotification(const STRING &functionName, IrisInstId dstInstId)
    {
        isRequestState = false;
        requestWriterRequestObject = requestWriter.openRequest(functionName, dstInstId);
        // addArg() will be called in the following to add parameters to the params object.
    }

    /// Add argument to request.
    /// (This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.)
    /// Fast variant for argument names up to 23 chars. Use this if you can.
    /// Usage: bla.addArg(ISTR("foo"), bar);
    template<class T>
    void addArg(const IrisU64StringConstant &argName, const T& arg)
    {
        requestWriterRequestObject.param(argName, arg);
    }

    /// Add argument to request.
    /// (This is supported for all types supported by IrisU64JsonWriter and IrisObjects.h.)
    /// Slow variant for argument names with more than 23 chars. Do not use unless you have to.
    template<class T>
    void addArgSlow(const std::string &argName, const T& arg)
    {
        requestWriterRequestObject.param(argName, arg);
    }

    /// Add an argument of object to request.
    /// The caller must persist the object contents immediately afterwards.
    IrisU64JsonWriter::Object addObjectArg(const IrisU64StringConstant &argName)
    {
        requestWriterRequestObject.param(argName);
        return requestWriter.openObject();
    }

    /// Set request id.
    /// This must be called only for requests (not for notifications) and it must be called before generateRequestMessage().
    void setRequestId(uint64_t requestId_)
    {
        is_notification = false;
        requestId = requestId_;
    }

    /// Generate request message.
    /// This function must be called after initRequest(), addArg() and potentially setRequestId() (only for requests) were called to finalize and retrieve the request message.
    const uint64_t *generateRequestMessage()
    {
        if (!is_notification)
            requestWriterRequestObject.setRequestId(requestId);
        requestWriterRequestObject.close();
        return requestWriter.getMessage();
    }


    // --- Reading response and setting response interface ---

    /// Set response for this request. Take a copy of the memory pointed to by message.data.
    /// This is called by irisHandleResponse() on any host thread.
    /// Warning: This is called from any thread.
    void setResponse(const uint64_t *responseMessage_)
    {
        // Store response and do no further processing on it, we want to block the calling thread as short as possible.
        assert(responseMessage_);
        assert(!hasResponseState);
        assert(responseMessage == 0);

        responseMessage = IrisU64JsonReader::copyValue(responseMessage_);

        // Make sure all modifications ended up in memory.
        irisCompilerAndMemoryBarrier();

        // Tell other threads we have the data. This is possible lock free since this only ever goes from false to true.
        hasResponseState = true;
    }

    /// Process response.
    /// This must be called once (and only once) after hasResponse() became true.
    /// getErrorCode() and the other result inspection function may only be called after this function.
    void processResponse()
    {
        assert(hasResponseState);
        assert(errorCode == E_no_response_yet);

        // Parse response object.
        IrisU64JsonReader r(responseMessage);
        IrisU64JsonReader::Response response = r.openResponse();

        if (response.isError())
        {
            if (response.getError(errorCode, errorMessage))
            {
                // Remember error data for caller.
                errorDataReader = response.getErrorDataReader();

                // Remember error data for our selves.
                internalErrorDataReader = errorDataReader;
            }
        }
        else
        {
            assert(response.getErrorCode() == E_ok);
            errorCode = E_ok;
            // We do not know how to parse the result here but the caller knows and will call getResultReader() to read the result.
            resultReader = response.getResultReader();
        }
    }

    /// Get result reader.
    IrisU64JsonReader &getResultReader() { return resultReader; }

    /// Get result reader.
    IrisU64JsonReader &getErrorDataReader() { return errorDataReader; }

    /// Returns true iff a response was received for this request.
    bool hasResponse() const { return hasResponseState; }

    /// Get error code of response. This is only valid if
    /// gotResponse() is true. Else this returns E_no_response_yet.
    IrisErrorCode getErrorCode() const { return errorCode; }

    /// Get error message of response. This is only valid if
    /// gotResponse() is true. Else this returns E_no_response_yet.
    std::string getErrorMessage() const { return errorMessage; }

    /// Get error message and data (containing error.message and error.data in compact readable JSON format).
    /// This does not contain the error code.
    std::string getErrorMessageAndData() const
    {
        std::stringstream s;
        s << errorMessage;
        if (internalErrorDataReader.isInitialized())
        {
            s << ", data=";
            convertU64JsonToJson(IncorrectJsonSingleLineFormat, internalErrorDataReader.getData(), internalErrorDataReader.getSize(), s);
        }
        return s.str();
    }

    /// Disable this request. It is not necessary to wait for the response anymore.
    /// This could interrupt the waiting for a response.
    void cancel() { canceled = true; }
    bool isCanceled() { return canceled; }

private:
    // Disallow assignment
    IrisRequest &operator= (const IrisRequest &);

private:
    /// Response.
    uint64_t *responseMessage;

    /// True iff we have a response.
    /// Volatile as it is checked by a debug thread (e.g. while(req.hasResponse());),
    /// and can be set by another thread such as TcpClient/TcpServer thread.
    volatile bool hasResponseState;

    /// Read result within response.
    IrisU64JsonReader resultReader;

    /// Read error.data within response (if any).
    IrisU64JsonReader errorDataReader;

    /// Read error.data within response (if any) from getErrorMessageAndData().
    IrisU64JsonReader internalErrorDataReader;

    /// Write request.
    IrisU64JsonWriter requestWriter;

    /// Helper for writing the request object (remembering the state between initRequest() and generateRequestMessage()).
    IrisU64JsonWriter::Request requestWriterRequestObject;

    /// Error code of the received response (if any).
    IrisErrorCode errorCode;

    /// Error message of the received response (if any).
    std::string errorMessage;

    /// True if this is a request, false if this is a notification.
    bool isRequestState;

    /// IrisCppAdapter will deal with split request differently.
    /// Calling function f(req) the first time will send the request and will not return a result.
    /// Once req.hasResponse() becomes true calling f(req) a second time will retrieve the result from the response.
    /// This allows to use IrisCppAdapter with non-blocking function call semantics.
    bool isSplit_;

    /// true if this is a notification
    bool is_notification;

    /// Request id (if this is a request and not a notification).
    uint64_t requestId;

    /// True if the wait for the response is canceled.
    /// Volatile as it may be set and checked by multiple threads.
    volatile bool canceled;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisRequest_h

