/*!
 *  \file   IrisThreadAbstraction.h
 *  \brief  Thread abstraction interface. Keeps the Iris infrastructure independent of a concrete thread implementation.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Thread abstraction interface. Keeps the Iris infrastructure independent of a concrete thread implementation.
 *  See IrisThreadAbstractionDefault for a conrete implelentation.
 */

#ifndef ARM_INCLUDE_IrisThreadAbstraction_h
#define ARM_INCLUDE_IrisThreadAbstraction_h

#include "iris/detail/IrisCommon.h"
#include <string>
#include <vector>

NAMESPACE_IRIS_START



class IrisThreadAbstractionThread;
class IrisThreadAbstractionRunnable;
class IrisThreadAbstractionConditionVariable;
class IrisThreadAbstractionMutex;


/// Main thread abstraction interface.
/// This allows to create things like threads, mutexes, events etc.
class IrisThreadAbstraction
{
public:
    // Destructor.
    virtual ~IrisThreadAbstraction() {}
    
    /// Create a thread.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    virtual IrisThreadAbstractionThread *createThread(IrisThreadAbstractionRunnable *runnable) = 0;

    /// Create condition variable.
    /// Condition variables can be waited upon and can be notified.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    /// If no mutex is provided an internal mutex is threated.
    virtual IrisThreadAbstractionConditionVariable *createConditionVariable(IrisThreadAbstractionMutex *mutex = 0) = 0;

    /// Create mutex.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    virtual IrisThreadAbstractionMutex *createMutex() = 0;
}; // class IrisThreadAbstraction


/// Thread.
class IrisThreadAbstractionThread
{
public:    
    /// Destructor.
    virtual ~IrisThreadAbstractionThread() {}

    /// Start thread.
    virtual void start() = 0;

    /// Join thread.
    /// Thread must end cooperatively by other means before or after calling this.
    virtual void join() = 0;

    /// Kill thread.
    /// Thread is killed non-cooperatively.
    /// Never use this if it can be avoided.
    /// Killing a thread is never clean and will most likely leak resources.
    virtual void kill() = 0;

    /// Is thread running?
    /// This is false before start() is called and after the thread exited.
    virtual bool isRunning() const = 0;

    /// Get the id for this thread.
    /// This must only be called after start().
    virtual unsigned long getThreadId() const = 0;
};

    
/// Thread runnable.
/// This is always implemented by the user of the abstraction.
/// Instances of this object must never be freed before the thread joined().
class IrisThreadAbstractionRunnable
{
public:
    /// Destructor.
    virtual ~IrisThreadAbstractionRunnable() {}

    /// Actual thread functionality.
    /// When this returns the thread is terminated.
    virtual void run() = 0;
};

    
/// Condition variable.
/// Condition variables help to wait for arbitrary conditions in a race-free way,
/// for example between producers and consumers.
/// Condition variables are always associated with a mutex.
/// If no mutex is provided an internal lock is created and used.
///
/// Typical use pattern:
///   Producer:
///     acquire();
///     produceItem();
///     notify(); (or notifyAll();)
///     release();
///   Consumer:
///     acquire();
///     while (!isItemAvailable())
///       wait();
///     consumeItem();
///     release();
class IrisThreadAbstractionConditionVariable
{
public:
    /// Destructor.
    virtual ~IrisThreadAbstractionConditionVariable() {}

    /// Lock access to shared variables.
    virtual void lock() = 0;
    
    /// Unlock access to shared variables.
    virtual void unlock() = 0;
    
    /// Wait (block) until this thread gets notified through notify() or notifyAll().
    /// The mutex must be lock()ed before calling this function (for example to produce an element).
    ///
    /// This atomically unlocks the mutex and then waits until notified.
    /// Once notified this locks the mutex again and returns.
    ///
    /// The mutex must be unlocked after this returns (for example after consuming the produced element).
    virtual void wait() = 0;
    
    /// Notify one thread of all threads currently waiting in wait().
    /// The mutex must be locked when calling this function.
    /// More than one thread may run when calling this function.
    virtual void notify() = 0;
    
    /// Notify all threads currently waiting in wait().
    /// The mutex must be locked when calling this function.
    virtual void notifyAll() = 0;
};


/// Mutex.
class IrisThreadAbstractionMutex
{
public:
    /// Destructor.
    virtual ~IrisThreadAbstractionMutex() {}

    /// Lock mutex.
    virtual void lock() = 0;

    /// Unlock mutex.
    virtual void unlock() = 0;
};

    
/// Lock. Critical section.
class IrisThreadAbstractionLock
{
public:
    /// Constructor locks mutex.
    explicit IrisThreadAbstractionLock(IrisThreadAbstractionMutex &mutex_): mutex(mutex_)
    {
        mutex.lock();
    }

    /// Destructor unlocks mutex.
    ~IrisThreadAbstractionLock()
    {
        mutex.unlock();
    }

private:
    /// Mutex locked by this lock.
    IrisThreadAbstractionMutex &mutex;
};

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisThreadAbstraction_h

