/*!
 *  \file   IrisThreadAbstractionDefault.h
 *  \brief  Default thread abstraction implementationinter using pthreads on Linux and Windows threads on Windows.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Default thread abstraction implementationinter using pthreads on Linux and Windows threads on Windows.
 */

#ifndef ARM_INCLUDE_IrisThreadAbstractionDefault_h
#define ARM_INCLUDE_IrisThreadAbstractionDefault_h

#include "iris/detail/IrisCommon.h"

#ifndef _WIN32
#include <pthread.h>
#include <signal.h>
#endif
#include <stdexcept>
#include <cassert>
#include "IrisErrorException.h"
#include "IrisThreadAbstraction.h"
#include "IrisLogger.h"
#include "IrisSystemInclude.h"

NAMESPACE_IRIS_START


class IrisThreadAbstractionDefault;


/// Thread.
class IrisThreadAbstractionThreadDefault: public IrisThreadAbstractionThread
{
public:
    /// Constructor.
    IrisThreadAbstractionThreadDefault(IrisThreadAbstractionRunnable *runnable_, IrisLogger &log_): runnable(runnable_), isRunning_(false), log(log_) {}

    /// Destructor.
    virtual ~IrisThreadAbstractionThreadDefault()
    {
        if (isRunning())
        {
            log.warning("deleting IrisThreadAbstractionThread instance before thread ended, killing thread\n");
            kill();
        }
    }

    /// Start thread.
    virtual void start() IRIS_OVERRIDE
    {
#ifdef _WIN32
        threadId = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)(IrisThreadAbstractionThreadDefault::run), (void *)this, 0, NULL);
        if (!threadId)
            throw IrisErrorExceptionCString(E_thread_error, "CreateThread() failed");
#else
        if (pthread_create(&threadId, 0, IrisThreadAbstractionThreadDefault::run, this) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_create() failed");
#endif
    }

    /// Join thread.
    /// Thread must end cooperatively by other means before or after calling this.
    virtual void join() IRIS_OVERRIDE
    {
#ifdef _WIN32
        //wait for thread to die.
        if (WaitForSingleObject(threadId, INFINITE) != WAIT_OBJECT_0)
            throw IrisErrorExceptionCString(E_thread_error, "WaitForSingleObject() failed");
#else
        int status = pthread_join(threadId, 0);
        if (status != 0)
        {
            std::string code;
            switch (status)
            {
            case EDEADLK: code = "EDEADLK"; break;
            case EINVAL:  code = "EINVAL";  break;
            case ESRCH:   code = "ESRCH";   break;
            default:      code = int64ToString(status); break;

            }
            throw IrisErrorExceptionString(E_thread_error, "pthread_join() returned " + code);
        }
#endif
    }

    /// Kill thread.
    /// Thread is killed non-cooperatively.
    /// Never use this if it can be avoided.
    /// Killing a thread is never clean and will most likely leak resources.
    virtual void kill() IRIS_OVERRIDE
    {
#ifdef _WIN32
        if (CloseHandle(threadId) == FALSE)
            throw IrisErrorExceptionCString(E_thread_error, "CloseHandle() failed");
#else
        if (pthread_kill((pthread_t)threadId, 9) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_kill() failed");
#endif
    }

    /// Is thread running?
    /// This is false before start() is called and after the thread exited.
    virtual bool isRunning() const IRIS_OVERRIDE
    {
        return isRunning_;
    }

    virtual unsigned long getThreadId() const IRIS_OVERRIDE
    {
#ifdef _WIN32
        return static_cast<unsigned long>(GetThreadId(threadId));
#else
        return static_cast<unsigned long>(threadId);
#endif
    }

private:
    /// Thread function thunk.
    static void *run(void *voidthis)
    {
        IrisThreadAbstractionThreadDefault *this_ = static_cast<IrisThreadAbstractionThreadDefault*>(voidthis);

        this_->isRunning_ = true;
        this_->runnable->run();
        this_->isRunning_ = false;
        return 0;
    }

    /// Runnable.
    IrisThreadAbstractionRunnable *runnable;

    /// Is thread running?
    /// This is false before start() is called and after the thread exited.
    bool isRunning_;

#ifdef _WIN32
    HANDLE threadId;
#else
    /// Thread id.
    pthread_t threadId;
#endif

    /// Logger.
    IrisLogger &log;
};


/// Mutex.
class IrisThreadAbstractionMutexDefault: public IrisThreadAbstractionMutex
{
public:
    // Constructor
    IrisThreadAbstractionMutexDefault()
    {
#ifdef _WIN32
        InitializeCriticalSection(&mutex);
#else
        if (pthread_mutex_init(&mutex, 0) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_mutex_init failed");
#endif
    }

    // Destructor.
    virtual ~IrisThreadAbstractionMutexDefault()
    {
#ifdef _WIN32
        DeleteCriticalSection(&mutex);
#else
        if (pthread_mutex_destroy(&mutex) != 0)
        {
            // We failed to destroy the mutex but there is not a lot that we can do about it
            assert(!"pthread_mutex_destroy failed");
        }
#endif
    }

    /// Lock mutex.
    virtual void lock() IRIS_OVERRIDE
    {
#ifdef _WIN32
        EnterCriticalSection(&mutex);
#else
        if (pthread_mutex_lock(&mutex) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_mutex_lock failed");
#endif
    }

    /// Unlock mutex.
    virtual void unlock() IRIS_OVERRIDE
    {
#ifdef _WIN32
        LeaveCriticalSection(&mutex);
#else
        if (pthread_mutex_unlock(&mutex) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_mutex_unlock failed");
#endif
    }

private:
    /// Get pointer to internal mutext representation (for ConditionVariable implementation).
#ifdef _WIN32
    CRITICAL_SECTION *
#else
    pthread_mutex_t *
#endif
    getInternalMutexPointer() { return &mutex; }
    friend class IrisThreadAbstractionConditionVariableDefault;

#ifdef _WIN32
    CRITICAL_SECTION mutex;
#else
    pthread_mutex_t mutex;
#endif
};


/// Condition variable.
/// Condition variables help to wait for arbitrary conditions in a race-free way,
/// for example between producers and consumers.
/// Condition variables are always associated with a mutex.
/// If no mutex is provided an internal lock is created and used.
///
/// Typical use pattern:
///   Producer:
///     acquire();
///     produceItem();
///     notify(); (or notifyAll();)
///     release();
///   Consumer:
///     acquire();
///     while (!isItemAvailable())
///       wait();
///     consumeItem();
///     release();
class IrisThreadAbstractionConditionVariableDefault: public IrisThreadAbstractionConditionVariable
{
public:
    /// Constructor.
    /// If no mutex is provided an internal one is created and used.
    IrisThreadAbstractionConditionVariableDefault(IrisThreadAbstractionDefault *threadAbstraction, IrisLogger &log, IrisThreadAbstractionMutex *mutex_ = 0);

    /// Destructor.
    virtual ~IrisThreadAbstractionConditionVariableDefault()
    {
#ifdef _WIN32
        // Noo need (nor possible) to delete condvar.
#else
        if (pthread_cond_destroy(&condvar) != 0)
        {
            // We failed to destroy the condvar but there is not a lot that we can do about it
            assert(!"pthread_cond_destroy failed");
        }
#endif
        delete ownMutex;
    }

    /// Lock access to shared variables.
    virtual void lock() IRIS_OVERRIDE
    {
        mutex->lock();
    }

    /// Unlock access to shared variables.
    virtual void unlock() IRIS_OVERRIDE
    {
        mutex->unlock();
    }

    /// Wait (block) until this thread gets notified through notify() or notifyAll().
    /// The mutex must be lock()ed before calling this function (for example to produce an element).
    ///
    /// This atomically unlocks the mutex and then waits until notified.
    /// Once notified this locks the mutex again and returns.
    ///
    /// The mutex must be unlocked after this returns (for example after consuming the produced element).
    virtual void wait() IRIS_OVERRIDE
    {
#ifdef _WIN32
        SleepConditionVariableCS(&condvar, mutex->getInternalMutexPointer(), INFINITE);
#else
        if (pthread_cond_wait(&condvar, mutex->getInternalMutexPointer()) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_cond_wait failed");
#endif
    }

    /// Notify one thread of all threads currently waiting in wait().
    /// The mutex must be locked when calling this function.
    /// More than one thread may run when calling this function.
    virtual void notify() IRIS_OVERRIDE
    {
#ifdef _WIN32
        WakeConditionVariable(&condvar);
#else
        if (pthread_cond_signal(&condvar) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_cond_signal");
#endif

    }

    /// Notify all threads currently waiting in wait().
    /// The mutex must be locked when calling this function.
    virtual void notifyAll() IRIS_OVERRIDE
    {
#ifdef _WIN32
        WakeAllConditionVariable(&condvar);
#else
        if (pthread_cond_broadcast(&condvar) != 0)
            throw IrisErrorExceptionCString(E_thread_error, "pthread_cond_broadcast");
#endif
    }

private:
#ifdef _WIN32
    CONDITION_VARIABLE condvar;
#else
    pthread_cond_t condvar;
#endif
    IrisThreadAbstractionMutexDefault *mutex;
    IrisThreadAbstractionMutexDefault *ownMutex;
};


class IrisThreadAbstractionDefault: public IrisThreadAbstraction
{
public:
    /// Constructor.
    IrisThreadAbstractionDefault(IrisLogger &log_): log(log_)
    {
    }

    /// Create a thread.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    virtual IrisThreadAbstractionThread *createThread(IrisThreadAbstractionRunnable *runnable) IRIS_OVERRIDE
    {
        return new IrisThreadAbstractionThreadDefault(runnable, log);
    }

    /// Create condition variable.
    /// Condition variables can be waited upon and can be notified.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    /// If no mutex is provided an internal mutex is threated.
    virtual IrisThreadAbstractionConditionVariable *createConditionVariable(IrisThreadAbstractionMutex *mutex = 0) IRIS_OVERRIDE
    {
        return new IrisThreadAbstractionConditionVariableDefault(this, log, mutex);
    }

    /// Create mutex.
    /// The returned object must be freed with delete and its lifetime is not attached to this object.
    virtual IrisThreadAbstractionMutex *createMutex() IRIS_OVERRIDE
    {
        return new IrisThreadAbstractionMutexDefault();
    }

private:
    /// Logger.
    IrisLogger &log;
}; // class IrisThreadAbstractionDefault


/// Get the current thread id.
inline unsigned long getCurrentThreadId()
{
#ifdef _WIN32
    return GetCurrentThreadId();
#else
    return (unsigned long)pthread_self();
#endif
}

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisThreadAbstractionDefault_h

