/*!
 *  \file   IrisU64JsonReader.h
 *  \brief  Read U64JSON data.
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Read U64JSON data.
 */

#ifndef ARM_INCLUDE_IrisU64JsonReader_h
#define ARM_INCLUDE_IrisU64JsonReader_h

#include "iris/detail/IrisCommon.h"

#include "IrisUtils.h"
#include "IrisValue.h"
#include "IrisErrorCode.h"
#include "IrisU64Json.h"
#include "IrisInterface.h"
#include "IrisU64JsonWriter.h"
#include "IrisPersistPSH.h"
#include "IrisErrorException.h"
#include <string>
#include <vector>
#include <iterator>
#include <cassert>

NAMESPACE_IRIS_START


/// Read U64JSON messages, e.g. read function name, argument names and argument values etc.
///
/// The access functions which read data may throw exceptions. It depends on the type of exception
/// in what state this object is after an exception:
///
/// IrisErrorTypeMismatch: The state is the same as it was before issuing the particular read operation.
///
/// IrisErrorUnknownMember: The state is the same as it was before issuing the particular read operation.
///
/// IrisErrorU64JsonEncodingError: The state is undefined. Further reads produce undefined results. This object needs to be discarded or re-initialized.
///
class IrisU64JsonReader
{
public:
    class Array;
    
    class Object
    {
    public:
        /// Constructor: Leave object uninitialized. Object will always be overwritten by copy.
        Object(): reader(0), n(0), dataEnd(0), dataStart(0) {}

        /// Internal constructor. Only use IrisU64JsonReader::openObject() to create instances of Object.
        Object(IrisU64JsonReader *reader_, uint64_t n_, const uint64_t *dataEnd_): reader(reader_), n(n_), dataEnd(dataEnd_), dataStart(reader_->readPos) {}

        /// Assignment has move semantics.
        /// Move semantics are necessary in order not to skip the entire Object when the temporary copy in openObject() gets destructed.
        const Object &operator =(const Object &o)
        {
            reader = o.reader;
            n = o.n;
            dataEnd = o.dataEnd;
            dataStart = o.dataStart;
            const_cast<Object*>(&o)->reader = 0;
            return *this;
        }

        /// Destructor, implicitly advances read cursor of enclosing reader to continue reading after this object.
        ~Object() { close(); }

        /// Close object (no longer read from this object) and advance read cursor of enclosing reader to continue reading after this object.
        void close() { if (reader) reader->readPos = dataEnd; }

        /// Get number of members.
        uint64_t getNumMembers() const { return n; }

        /// Initialize with an empty object.
        void initWithEmptyObject()
        {
            dataStart = dataEnd = reinterpret_cast<const uint64_t *>(1); // We will not read from this location anyway. 1 not to trigger asserts.
            n = 0;
        }

        /// Get value of mandatory member.
        template<class T>
        void member(const IrisU64StringConstant &name, T &out) const
        {
            if (!find(name))
                throw IrisErrorUnknownMember(name);
            try
            {
                iris::persist(*reader, out);
            }
            catch(IrisErrorTypeMismatch &e)
            {
                e.setName(name);
                throw e;
            }
        }

        /// Get value of mandatory member.
        template<class T>
        void memberSlow(const std::string &name, T &out) const
        {
            if (!find(name))
                throw IrisErrorUnknownMember(name);
            try
            {
                iris::persist(*reader, out);
            }
            catch(IrisErrorTypeMismatch &e)
            {
                e.setName(name);
                throw e;
            }
        }

        /// Get value of optional member.
        /// If the member is present return true and return its value in out.
        /// If the member is not present return false and out is not touched (and for example can keep its default value).
        template<class T>
        bool memberOptional(const IrisU64StringConstant &name, T &out) const
        {
            if (!find(name))
                return false;
            try
            {
                iris::persist(*reader, out);
            }
            catch(IrisErrorTypeMismatch &e)
            {
                e.setName(name);
                throw e;
            }
            return true;
        }

        /// Get value of optional member.
        /// If the member is present return true and return its value in out.
        /// If the member is not present return false and out is not touched (and for example can keep its default value).
        template<class T>
        bool memberOptionalSlow(const std::string &name, T &out) const
        {
            if (!find(name))
                return false;
            try
            {
                iris::persist(*reader, out);
            }
            catch(IrisErrorTypeMismatch &e)
            {
                e.setName(name);
                throw e;
            }
            return true;
        }

        /// Get value of optional member (variant for persist() functions which are identical for read and write).
        /// If the member is present return isPresent=true and return its value in out.
        /// If the member is not present return isPresent=false and out is not touched (and for example can keep its default value).
        template<class T>
        void memberOptional(const IrisU64StringConstant &name, T &out, bool &isPresent) const
        {
            isPresent = memberOptional(name, out);
        }

        /// Get value of optional member (variant for persist() functions which are identical for read and write).
        /// If the member is present return isPresent=true and return its value in out.
        /// If the member is not present return isPresent=false and out is not touched (and for example can keep its default value).
        template<class T>
        void memberOptionalSlow(const std::string &name, T &out, bool &isPresent) const
        {
            isPresent = memberOptionalSlow(name, out);
        }

        /// Position read cursor of enclosing reader at value of member (for example to read a nested object value for a member).       
        void member(const IrisU64StringConstant &name) const
        {
            if (!find(name))
                throw IrisErrorUnknownMember(name);
        }

        /// Position read cursor of enclosing reader at value of member (for example to read a nested object value for a member).       
        void memberSlow(const std::string &name) const
        {
            if (!find(name))
                throw IrisErrorUnknownMember(name);
        }

        /// Position read cursor of enclosing reader at value of member if it exists.
        /// If it exists return true.
        /// If it does not exist return false and the read cursor of the enclosing reader is undefined.
        bool memberOptional(const IrisU64StringConstant &name) const
        {
            return find(name);
        }

        /// Position read cursor of enclosing reader at value of member if it exists.
        /// If it exists return true.
        /// If it does not exist return false and the read cursor of the enclosing reader is undefined.
        bool memberOptionalSlow(const std::string &name) const
        {
            return find(name);
        }

        // Read next member pair from current read position.
        // This is used to enumerate all members of an object.
        // T is usually IrisValue as one usually does not know the type in advance.
        // This function conflicts with member() and memberOptional() and should not be used together with these.
        // Returns true if a member was read, else false (= tried to read past end of container).
        template<class T>
        bool readNextMember(std::string &name, T &out)
        {
            if (reader->readPos >= dataEnd)
                return false;
            reader->persist(name, /*persistMemberName=*/true);
            try
            {
                iris::persist(*reader, out);
            }
            catch(IrisErrorTypeMismatch &e)
            {
                e.setName(name);
                throw e;
            }
            return true;
        }
        
        // Read next member name from current read position.
        // This is used to enumerate all members of an object.
        // This function conflicts with member() and memberOptional() and should not be used together with these.
        // Returns true if a member was read, else false (= tried to read past end of container).
        bool readNextMember(std::string &name)
        {
            if (reader->readPos >= dataEnd)
                return false;
            reader->persist(name, /*persistMemberName=*/true);
            return true;
        }
        
        /// Open nested object.
        /// Usage:
        /// outerObject.member("member_name_of_nested_object")
        /// nestedObject = outerObject.openObject()
        /// <read from nested object>
        /// Note: Reading from the outer object invalidates the read position of the inner object.
        /// The intended use is to complete reading from the inner object before continuing to read from the outer object.
        Object openObject()
        {
            return reader->openObject();
        }

        /// Open nested array.
        /// Usage:
        /// outerObject.member("member_name_of_nested_array")
        /// nestedArray = outerObject.openArray()
        /// <read from nested array>
        /// Note: Reading from the outer object invalidates the read position of the inner array.
        /// The intended use is to complete reading from the inner array before continuing to read from the outer object.
        Array openArray()
        {
            return reader->openArray();
        }

        /// Reset: Start reading members again which may have already been read.
        /// Normally members must be read in alphabetical order for performance reasons.
        /// After a call to reset() members may be read from the beginning of the sorted array again
        /// without a performance impact. This is rarely needed and exists mostly for testing purposes.
        void reset()
        {
            reader->readPos = dataStart;
        }

    private:
        /// Find member and return pointer to value.
        /// Return true iff member was found in which case reader->readPos point to the value of that member.
        bool find(const std::string &name) const
        {
            reader->readPos = dataStart;
            while (reader->readPos < dataEnd)
            {
                std::string aname;
                reader->persist(aname, /*persistMemberName=*/true);
                if (aname == name)
                    return true;
                reader->skip();
            }
            return false;
        }

        /// Find member and return pointer to value.
        /// Return true iff member was found in which case reader->readPos point to the value of that member.
        bool find(const IrisU64StringConstant &name) const
        {
            reader->readPos = dataStart;
            while (reader->readPos < dataEnd)
            {
                if (name.data0 == *(reader->readPos))
                {
                    reader->readPos++;
                    if (name.containerLength == 1)
                        return true;
                    else if (name.data1 == *(reader->readPos))
                    {
                        reader->readPos++;
                        if (name.containerLength == 2)
                            return true;
                        else if (name.data2 == *(reader->readPos))
                        {
                            // We blindly assume name.containerLength == 3 here because it must be either 1, 2 or 3.
                            assert(name.containerLength == 3);
                            reader->readPos++;
                            return true;
                        }
                    }
                }
                
                reader->skip(); // Skip member name.
                reader->skip(); // Skip value.
            }
            return false;
        }

        IrisU64JsonReader *reader;
        uint64_t n;
        const uint64_t *dataEnd;
        const uint64_t *dataStart;
    };

    class Array
    {
    public:
        /// Constructor: Leave array uninitialized. Array will always be overwritten by copy.
        Array(): reader(0), n(0), dataEnd(0), dataStart(0) {}
        
        /// Internal constructor. Only use IrisU64JsonReader::openArray() to create instances of Array.
        Array(IrisU64JsonReader *reader_, uint64_t n_, const uint64_t *dataEnd_): reader(reader_), n(n_), dataEnd(dataEnd_), dataStart(reader_->readPos) {}

        /// Destructor, implicitly advances read cursor of enclosing reader to continue reading after this array.
        ~Array() { close(); }
        
        /// Copy constructor has move semantics.
        /// Move semantics are necessary in order not to skip the entire Array when the temporary copy in openArray() gets destructed.
        Array(const Array &a): reader(a.reader), n(a.n), dataEnd(a.dataEnd), dataStart(a.dataStart) { const_cast<Array*>(&a)->reader = 0; }
        
        /// Assignment has move semantics.
        /// Move semantics are necessary in order not to skip the entire Array when the temporary copy in openArray() gets destructed.
        const Array &operator =(const Array &a)
        {
            reader = a.reader;
            n = a.n;
            dataEnd = a.dataEnd;
            dataStart = a.dataStart;
            const_cast<Array*>(&a)->reader = 0;
            return *this;
        }

        /// Close array (no longer read from this array) and advance read cursor of enclosing reader to continue reading after this array.
        void close() { if (reader) reader->readPos = dataEnd; }

        /// Get number of elements in array.
        uint64_t getLength() const { return n; }
        
        /// Return true iff there are no more elements to read from this array.
        bool reachedEnd() const { return reader->readPos == dataEnd; }

        /// Read element from array.
        template<class T>
        void element(T &out)
        {
            iris::persist(*reader, out);
        }

        /// Open nested object.
        /// Usage:
        /// nestedObject = outerArray.openObject()
        /// <read from nested object>
        /// Note: Reading from the outer array invalidates the read position of the inner object and vice versa.
        /// The intended use is to complete reading from the inner object before continuing to read from the outer array.
        Object openObject()
        {
            return reader->openObject();
        }

        /// Open nested array.
        /// Usage:
        /// nestedArray = outerArray.openArray()
        /// <read from nested array>
        /// Note: Reading from the outer array invalidates the read position of the inner array and vice versa.
        /// The intended use is to complete reading from the inner array before continuing to read the outer array.
        Array openArray()
        {
            return reader->openArray();
        }
        
    private:
        IrisU64JsonReader *reader;
        uint64_t n;
        const uint64_t *dataEnd;
        const uint64_t *dataStart;
    };

    class Request
    {
    public:
        /// Constructor: Leave request uninitialized. Request will always be overwritten by copy.
        Request()
            : reader(NULL)
            , method("<invalid>")
            , dataEnd(NULL)
            , dataStart(NULL)
        {
        }

        /// Internal constructor. Only use IrisU64JsonReader::openRequest() to create instances of Request.
        Request(IrisU64JsonReader *reader_, const uint64_t *dataEnd_)
            : reader(reader_)
            , dataEnd(dataEnd_)
            , dataStart(reader_->readPos)
        {
            // reader should be pointing at the element after the header
            // but we want to look at the header to see if we are a request or a notification
            uint64_t type = *(reader->readPos - 1) & u64json::c_msb8mask;
            is_notification = type == u64json::c_notification;

            assert(is_notification || type == u64json::c_request);

            // Read request method
            reader->readPos = &dataStart[method_offset];
            iris::persist(*reader, method);

            // Start reading the params
            params = reader->openObject();
        }

        /// Destructor, implicitly advances read cursor of enclosing reader to continue reading after this request.
        ~Request()
        {
            close();
        }

        /// Close Request (no longer read from this request) and advance read cursor of enclosing reader to continue reading after this request.
        void close()
        {
            if (reader)
            {
                reader->readPos = dataEnd;
            }
        }

        uint64_t getRequestId() const
        {
            if (!isNotification())
            {
                return dataStart[request_id_offset];
            }
            else
            {
                throw IrisInternalError("Tried to read request id from a notification");
            }
        }

        bool isNotification() const
        {
            return is_notification;
        }

        /// Get destination instId
        uint64_t getInstId() const
        {
            return dataStart[instId_offset];
        }

        const std::string &getMethod() const
        {
            return method;
        }

        /// Get value of mandatory param.
        template<class T>
        void param(const IrisU64StringConstant &name, T &out) const
        {
            params.member(name, out);
        }

        /// Get value of mandatory param.
        template<class T>
        void paramSlow(const std::string &name, T &out) const
        {
            params.memberSlow(name, out);
        }

        /// Get value of optional param.
        /// If the param is present return true and return its value in out.
        /// If the param is not present return false and out is not touched (and for example can keep its default value).
        template<class T>
        bool paramOptional(const IrisU64StringConstant &name, T &out) const
        {
            return params.memberOptional(name, out);
        }

        /// Get value of optional param.
        /// If the param is present return true and return its value in out.
        /// If the param is not present return false and out is not touched (and for example can keep its default value).
        template<class T>
        bool paramOptionalSlow(const std::string &name, T &out) const
        {
            return params.memberOptionalSlow(name, out);
        }

        /// Get value of optional param (variant for persist() functions which are identical for read and write).
        /// If the param is present return isPresent=true and return its value in out.
        /// If the param is not present return isPresent=false and out is not touched (and for example can keep its default value).
        template<class T>
        void paramOptional(const IrisU64StringConstant &name, T &out, bool &isPresent) const
        {
            params.memberOptional(name, out, isPresent);
        }

        /// Get value of optional param (variant for persist() functions which are identical for read and write).
        /// If the param is present return isPresent=true and return its value in out.
        /// If the param is not present return isPresent=false and out is not touched (and for example can keep its default value).
        template<class T>
        void paramOptionalSlow(const std::string &name, T &out, bool &isPresent) const
        {
            params.memberOptionalSlow(name, out, isPresent);
        }

        /// Position read cursor of enclosing reader at value of param (for example to read a nested object value for a param).
        void param(const IrisU64StringConstant &name) const
        {
            params.member(name);
        }

        /// Position read cursor of enclosing reader at value of param (for example to read a nested object value for a param).
        void paramSlow(const std::string &name) const
        {
            params.memberSlow(name);
        }

        /// Position read cursor of enclosing reader at value of param if it exists.
        /// If it exists return true.
        /// If it does not exist return false and the read cursor of the enclosing reader is undefined.
        bool paramOptional(const IrisU64StringConstant &name) const
        {
            return params.memberOptional(name);
        }

        /// Position read cursor of enclosing reader at value of param if it exists.
        /// If it exists return true.
        /// If it does not exist return false and the read cursor of the enclosing reader is undefined.
        bool paramOptionalSlow(const std::string &name) const
        {
            return params.memberOptionalSlow(name);
        }

        // Read next param name from current read position.
        // This is used to enumerate all params of an request.
        // This function conflicts with param() and paramOptional() and should not be used together with these.
        // Returns true if a param was read, else false (= tried to read past end of container).
        bool readNextParam(std::string &name)
        {
            return params.readNextMember(name);
        }

        /// Open nested object.
        /// Usage:
        /// request.param("param_name_of_nested_object")
        /// nestedObject = request.openObject()
        /// <read from nested object>
        /// Note: Reading from the request invalidates the read position of the inner object.
        /// The intended use is to complete reading from the inner object before continuing to read from the outer request.
        Object openObject()
        {
            return params.openObject();
        }

        /// Open nested array.
        /// Usage:
        /// outerObject.param("param_name_of_nested_array")
        /// nestedArray = outerObject.openArray()
        /// <read from nested array>
        /// Note: Reading from the outer request invalidates the read position of the inner array.
        /// The intended use is to complete reading from the inner array before continuing to read from the outer request.
        Array openArray()
        {
            return params.openArray();
        }

        /// Reset: Start reading params again which may have already been read.
        /// Normally params must be read in alphabetical order for performance reasons.
        /// After a call to reset() params may be read from the beginning of the sorted array again
        /// without a performance impact. This is rarely needed and exists mostly for testing purposes.
        void reset()
        {
            params.reset();
        }

    private:
        IrisU64JsonReader *reader;

        /// Request fields
        std::string method; //! Name of the function being called

        /// Request 'params' object
        Object params;

        /// Message data start and end pointers
        const uint64_t *dataEnd;
        const uint64_t *dataStart;

        // True if this is a notification
        bool is_notification;

        /// Field offset constants
        static const size_t request_id_offset = 0;
        static const size_t instId_offset     = 1;
        static const size_t method_offset     = 2;
    };

    class Response
    {
    public:
        /// Constructor: Leave request uninitialized. Response will always be overwritten by copy.
        Response()
            : reader(NULL)
            , dataEnd(NULL)
            , dataStart(NULL)
        {
        }

        /// Internal constructor. Only use IrisU64JsonReader::openRequest() to create instances of Response.
        Response(IrisU64JsonReader *reader_, const uint64_t *dataEnd_)
            : reader(reader_)
            , dataEnd(dataEnd_)
            , dataStart(reader_->readPos)
        {
        }

        /// Destructor, implicitly advances read cursor of enclosing reader to continue reading after this request.
        ~Response()
        {
            close();
        }

        /// Close Response (no longer read from this request) and advance read cursor of enclosing reader to continue reading after this request.
        void close()
        {
            if (reader)
            {
                reader->readPos = dataEnd;
            }
        }

        uint64_t getRequestId() const
        {
            return dataStart[request_id_offset];
        }

        /// Get destination instId
        uint64_t getInstId() const
        {
            return dataStart[instId_offset];
        }

        IrisErrorCode getErrorCode() const
        {
            return static_cast<IrisErrorCode>(dataStart[error_code_offset]);
        }

        /// Returns true if this response represents an error
        bool isError() const
        {
            return getErrorCode() != E_ok;
        }

        /// Get result value
        template <typename RESULT_T>
        void getResult(RESULT_T &result) const
        {
            if (isError())
            {
                throw IrisInternalError("Tried to read result from an error response");
            }

            // Set read position to the beginning of the result
            reader->readPos = &dataStart[result_offset];
            iris::persist(*reader, result);
        }

        /// Returns a reader object primed to read the result field of the response
        IrisU64JsonReader getResultReader() const
        {
            if (isError())
            {
                throw IrisInternalError("Tried to read result from an error response");
            }

            // Set read position to the beginning of the result
            reader->readPos = &dataStart[result_offset];

            IrisU64JsonReader result_reader;
            result_reader.initToSingleValue(*reader);
            return result_reader;
        }

        /// Get error information
        /// Returns true if this response includes the optional 'data' field.
        /// The caller should read the 'data' field using the regular reader functions
        /// after calling getError().
        bool getError(IrisErrorCode &code_out, std::string &message_out) const
        {
            if (!isError())
            {
                throw IrisInternalError("Tried to read error from a non-error response");
            }

            code_out = getErrorCode();

            // Set read position to the beginning of the error information.
            reader->readPos = &dataStart[error_obj_offset];

            // Read the mandatory error 'message' field.
            iris::persist(*reader, message_out);

            return reader->readPos < dataEnd;
        }

        /// Get error information
        /// Returns true if the optional 'data' field was read. If this method returns false
        /// the response did not include the 'data' field and data_out was not touched.
        template <typename DATA_T>
        bool getError(IrisErrorCode &code_out, std::string &message_out, DATA_T &data_out) const
        {
            bool has_data = getError(code_out, message_out);

            if (has_data)
            {
                // Read the optional error 'data' field.
                iris::persist(*reader, data_out);
            }

            return has_data;
        }

        /// Returns a reader object primed to read the result field of the response
        IrisU64JsonReader getErrorDataReader() const
        {
            if (!isError())
            {
                throw IrisInternalError("Tried to read error data from a non-error response");
            }

            // Set read position to the beginning of the error information.
            reader->readPos = &dataStart[error_obj_offset];

            // Skip the 'message' field.
            reader->skip();

            IrisU64JsonReader data_reader;
            data_reader.initToSingleValue(*reader);
            return data_reader;
        }

    private:
        IrisU64JsonReader *reader;

        /// Message data start and end pointers
        const uint64_t *dataEnd;
        const uint64_t *dataStart;

        /// Field offset constants
        static const size_t request_id_offset = 0;
        static const size_t instId_offset     = 1;
        static const size_t error_code_offset = 2;
        static const size_t error_obj_offset  = 3; // } The error object and result have the same offset.
        static const size_t result_offset     = 3; // } One of the other is used depending on the value of error code.
    };



    /// Construct uninitialized.
    /// init() must be called before using the reader.
    IrisU64JsonReader(): readPos(0), dataEnd(0) {}


    /// Construct from raw buffer.
    /// data must be != 0.
    IrisU64JsonReader(const uint64_t *data_, const uint64_t *dataEnd_) 
    { 
        init(data_, dataEnd_); 
    }


    /// Construct from Iris message.
    /// message_ must be != 0 and message_->data must be != 0.
    IrisU64JsonReader(const uint64_t *message_) 
    { 
        init(message_); 
    }
    
    
    /// Construct from writer.
    /// This reader is only valid as long as the writer is unchanged.
    IrisU64JsonReader(const IrisU64JsonWriter &writer)
    {
        init(writer);
    }


    /// Construct from IrisValue.
    /// This must only ever be called for an array or an object. 
    /// This does not work for any other type. Calling this for any other type causes an internal error exception.
    /// This reader is only valid as long as the IrisValue is unchanged.    
    IrisU64JsonReader(const IrisValue &arrayOrObject)
    {
        init(arrayOrObject);
    }


    /// Re-initialize to a raw buffer.
    void init(const uint64_t *data_, const uint64_t *dataEnd_)
    {
        readPos = data_;
        dataEnd = dataEnd_;
        assert(readPos);
        assert(dataEnd);
    }


    /// Re-initialize to a message.
    void init(const uint64_t *message_)
    {
        assert(message_);
        init(message_, message_ + getSizeOfValue(message_));
    }


    /// Re-initialize to the current read position of another reader (for example to read a value later or in another context).
    void init(const IrisU64JsonReader &otherReader)
    {
        assert(this != &otherReader);
        readPos = otherReader.readPos;
        dataEnd = otherReader.dataEnd;
    }


    /// Re-initialize to the current read position of another reader (for example to read a value later or in another context).
    /// This variant limits the reader to the value pointed to by readPos.
    void initToSingleValue(const IrisU64JsonReader &otherReader)
    {
        init(otherReader);
        size_t size = getSizeOfCurrentValue();
        if (readPos + size > dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        dataEnd = readPos + size;
    }


    /// Initialize to read from an IrisU64JsonWriter. This is intended just for testing.
    void init(const IrisU64JsonWriter &writer)
    {
        readPos = writer.getMessage();
        dataEnd = readPos + writer.getSize();
    }


    /// Construct from IrisValue.
    /// This must only ever be called for an array or an object. 
    /// This does not work for any other type. Calling this for any other type causes an internal error exception.
    /// This reader is only valid as long as the IrisValue is unchanged.    
    void init(const IrisValue &arrayOrObject)
    {
        if ((arrayOrObject.getType() != IrisValue::Array) && (arrayOrObject.getType() != IrisValue::Object))
            throw IrisInternalError("IrisU64JsonReader::init(IrisValue&) called for type other than array or object.");
        size_t size = 0;
        readPos = arrayOrObject.getU64JsonDataForObjectOrArray(size);
        dataEnd = readPos + size;
    }


    /// Clear reader.
    /// Nothing can be read from this reader after this function was called.
    /// See initWithEmptyObject() if you want to have an empty object container.
    void clear()
    {
        readPos = 0;
        dataEnd = 0;
    }


    /// Initialize reader with an empty object.
    void initWithEmptyObject()
    {
        readPos = u64json::empty_object;
        dataEnd = readPos + (sizeof(u64json::empty_object) / sizeof(u64json::empty_object[0]));
    }


    /// Return true iff this reader is initialized with data and can be used for reading.
    bool isInitialized() const { return readPos != 0; }


    /// Get acces to raw data. Returns current read position.
    const uint64_t *getData() const { return readPos; }


    /// Get size.
    size_t getSize() const { return size_t(dataEnd - readPos); }


    /// Return true iff no more data can be read.
    bool isEmpty() const { return readPos == dataEnd; }
    
    
    /// Return true iff no more data can be read.
    bool reachedEnd() const { return readPos == dataEnd; }

    
    /// Open Object.
    Object openObject()
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        if ((readPos[0] >> 60) != 0xb)
            throw IrisErrorTypeMismatch("Expected object.");
        if (readPos + 2 > dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t containerLength = readPos[0] & ~u64json::c_msb4mask;
        if (readPos + containerLength > dataEnd)
            throw IrisErrorU64JsonEncodingError("Object length exceeds data length.");
        if (containerLength < 2)
            throw IrisErrorU64JsonEncodingError("Malformatted object.");
        uint64_t n = readPos[1];
        readPos += 2;
        return Object(this, n, readPos + containerLength - 2);
    }


    /// Open Array.
    Array openArray()
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        if ((readPos[0] >> 60) != 0xa)
            throw IrisErrorTypeMismatch("Expected array.");
        if (readPos + 2 > dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t containerLength = readPos[0] & ~u64json::c_msb4mask;
        if (readPos + containerLength > dataEnd)
            throw IrisErrorU64JsonEncodingError("Array length exceeds data length.");
        if (containerLength < 2)
            throw IrisErrorU64JsonEncodingError("Malformatted array.");
        uint64_t n = readPos[1];
        readPos += 2;
        return Array(this, n, readPos + containerLength - 2);
    }

    /// Open Request.
    Request openRequest()
    {
        if (readPos >= dataEnd - 3)
            throw IrisErrorU64JsonEncodingError("Out of data.");

        uint64_t value_type = readPos[0] & u64json::c_msb8mask;
        if ((value_type != u64json::c_request) && (value_type != u64json::c_notification))
            throw IrisErrorTypeMismatch("Expected request.");

        // 0xe0vv_xxxx_xxxx_xxxx
        //   vv = jsonrpc version
        // xx.. = container length
        uint64_t jsonrpc_version  = readPos[0] & 0x00ff000000000000ull;
        uint64_t container_length = readPos[0] & 0x0000ffffffffffffull;

        if (jsonrpc_version != u64json::c_jsonrpc2)
            throw IrisErrorU64JsonEncodingError("Unrecognised JSON-RPC version");

        if (readPos + container_length > dataEnd)
            throw IrisErrorU64JsonEncodingError("Request length exceeds data length.");

        if (container_length <= 3)
            throw IrisErrorU64JsonEncodingError("Malformatted request.");

        readPos += 1;
        return Request(this, readPos + container_length - 1);
    }

    /// Open Response.
    Response openResponse()
    {
        if (readPos >= dataEnd - 4)
            throw IrisErrorU64JsonEncodingError("Out of data.");

        if ((readPos[0] & u64json::c_msb8mask) != u64json::c_response)
            throw IrisErrorTypeMismatch("Expected response.");

        // 0xe2vv_xxxx_xxxx_xxxx
        //   vv = jsonrpc version
        // xx.. = container length
        uint64_t jsonrpc_version  = readPos[0] & 0x00ff000000000000ull;
        uint64_t container_length = readPos[0] & 0x0000ffffffffffffull;

        if (jsonrpc_version != u64json::c_jsonrpc2)
            throw IrisErrorU64JsonEncodingError("Unrecognised JSON-RPC version");

        if (readPos + container_length > dataEnd)
            throw IrisErrorU64JsonEncodingError("Response length exceeds data length.");

        if (container_length <= 4)
            throw IrisErrorU64JsonEncodingError("Malformatted response.");

        readPos += 1;
        return Response(this, readPos + container_length - 1);
    }


    /// Read uint64_t.
    void persist(uint64_t &out)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *(readPos++);
        if ((n < 0x1000000000000000ull) || (n >= 0xf000000000000000ull))
        {
            out = n; // Most numbers will be in this case.
            return;
        }
        if (n < 0x2000000000000000ull) // Negative NumberS64
        {
            readPos--;
            throw IrisErrorTypeMismatch("Cannot convert negative number to NumberU64.");
        }


        // Handle other cases.
        switch(n >> 56)
        {
        // case 0x0x: // NumberU64: handled in the if() above
        // case 0x1x: // Negative NumberS64: handled in the if() above
        case 0xc0: // NumberU64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            out = *(readPos++);
            break;
        case 0xc1: // NumberS64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            n = *(readPos++);
            if (int64_t(n) < 0)
            {
                readPos -= 2;
                throw IrisErrorTypeMismatch("Cannot convert negative number to NumberU64.");
            }
            out = n;
            break;
        case 0xca: // Number as double
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            {
                const void *vp = readPos++;
                double d = *reinterpret_cast<const double*>(vp);
                if ((d < 0.0) || (d > double(IRIS_UINT64_MAX)))
                {
                    readPos -= 2;
                    throw IrisErrorTypeMismatch("Double out of range when converting to NumberU64.");
                }
                out = uint64_t(d);
            }
            break;
        case 0xce: // False
            out = 0;
            break;
        case 0xcf: // True
            out = 1;
            break;
        // case 0xfx: // NumberU64: handled in the if() above
        default:
            {
                readPos--;
                throw IrisErrorTypeMismatch("Cannot convert non-numeric type to Number.");
            }
        }
    }


    /// Persist NumberS64.
    void persist(int64_t &out)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *(readPos++);
        if (n < 0x1000000000000000ull)
        {
            out = int64_t(n);
            return;
        }
        if (n < 0x2000000000000000ull)
        {
            out = int64_t(0xf000000000000000ull | n);
            return;
        }

        // Handle other cases.
        switch(n >> 56)
        {
        // case 0x0x: // NumberU64: handled in the if() above
        // case 0x1x: // Negative NumberS64: handled in the if() above
        case 0xc0: // NumberU64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            n = *(readPos++);
            if (int64_t(n) < 0)
            {
                readPos -= 2;
                throw IrisErrorTypeMismatch("Cannot convert high positive number to NumberS64.");
            }
            out = n;
            break;
        case 0xc1: // NumberS64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            out = int64_t(*(readPos++));
            break;
        case 0xca: // Number as double
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            {
                const void *vp = readPos++;
                double d = *reinterpret_cast<const double*>(vp);
                if ((d < double(IRIS_INT64_MIN)) || (d >= double(IRIS_INT64_MAX)))
                {
                    readPos -= 2;
                    throw IrisErrorTypeMismatch("Double out of range when converting to NumberS64.");
                }
                out = int64_t(d);
            }
            break;
        case 0xce: // False
            out = 0;
            break;
        case 0xcf: // True
            out = 1;
            break;
        // case 0xfx: // NumberU64: handled by default (is invalid)
        default:
            {
                readPos--;
                throw IrisErrorTypeMismatch("Cannot convert non-numeric type or high positive number to NumberS64.");
            }
        }
    }


    /// Persist double.
    void persist(double &out)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *(readPos++);
        if ((n < 0x1000000000000000ull) || (n >= 0xf000000000000000ull))
        {
            out = double(n);
            return;
        }
        if (n < 0x2000000000000000ull)
        {
            out = double(int64_t(0xf000000000000000ull | n));
            return;
        }

        // Handle other cases.
        switch(n >> 56)
        {
        // case 0x0x: // NumberU64: handled in the if() above
        // case 0x1x: // Negative NumberS64: handled in the if() above
        case 0xc0: // NumberU64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            out = double(*(readPos++));
            break;
        case 0xc1: // NumberS64
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            out = double(int64_t(*(readPos++)));
            break;
        case 0xca: // Number as double
            if (readPos >= dataEnd)
                throw IrisErrorU64JsonEncodingError("Out of data.");
            {
                const void *vp = readPos++;
                out = *reinterpret_cast<const double*>(vp);
            }
            break;
        case 0xce: // False
            out = 0.0;
            break;
        case 0xcf: // True
            out = 1.0;
            break;
        // case 0xfx: // NumberU64: handled by if() above.
        default:
            {
                readPos--;
                throw IrisErrorTypeMismatch("Cannot convert non-numeric type to double.");
            }
        }
    }


    /// Persist Boolean.
    void persist(bool &out)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *(readPos++);
        switch (n)
        {
        case 0xce00000000000000ull:
        case 0:
            out = false;
            break;
        case 0xcf00000000000000ull:
        case 1:
            out = true;
            break;
        default:
            {
                readPos--;
                throw IrisErrorTypeMismatch("Cannot convert value to bool.");
            }
        }
    }


    /// Persist String.
    void persist(std::string &out, bool persistMemberName = false)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = readPos[0];
        if ((n >= 0x2000000000000000ull) && (n < 0x8000000000000000ull))
        {
            // Short string (0..255 chars) which has char str[6] in 0x20..0x7f (or is shorter).
            uint64_t len = n & 255;
            uint64_t containerLength = (len >> 3) + 1;
            if (readPos + containerLength > dataEnd) // >= is OK
                throw IrisErrorU64JsonEncodingError("Out of data.");
            const char *p = reinterpret_cast<const char *>(readPos) + 1;
            // Note: This works only on little-endian machines.
            out = std::string(p, len);
            readPos += containerLength;
            return;
        }
        else if ((n >> 56) == 0xcc)
        {
            persistLongString(out);
        }
        else 
        {
            // Note: When we read an object member name and fail we must not throw IrisErrorTypeMismatch()
            // as this would indicate to the user that he may try to read the member again with a different
            // type.
            // Instead this is an encoding error which is a lot more fatal and leaves this object in an
            // undefined state.
            if (persistMemberName)
                throw IrisErrorU64JsonEncodingError("Member name is not a string.");
            else
                throw IrisErrorTypeMismatch("Value is not a string."); 
        }
    }


    /// Persist Array of NumberU64.
    void persist(std::vector<uint64_t> &out)
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");

        uint64_t n = *(readPos++);
        uint64_t len = n & ~u64json::c_msb4mask;
        out.clear();
        switch(n >> 60)
        {
        case 0x8:
            if (readPos + len > dataEnd)
            {
                readPos--;
                throw IrisErrorU64JsonEncodingError("Out of data.");
            }

            // Copy all uint64_t elements in the array
            std::copy(readPos, readPos + len, std::back_inserter(out));
            readPos += len;
            break;
        case 0xa:
            if (readPos + len - 1 > dataEnd)
            {
                readPos--;
                throw IrisErrorU64JsonEncodingError("Out of data.");
            }

            {
                const uint64_t* startPos = readPos - 1;
                uint64_t numElements = *(readPos++);
                try
                {
                    while (numElements--)
                    {
                        uint64_t element;
                        persist(element);
                        out.push_back(element);
                    }
                }
                catch (IrisErrorTypeMismatch &e)
                {
                    // Reset to start position if any element is not uint64_t
                    readPos = startPos;
                    throw e;
                }
            }
            break;
        default:
            readPos--;
            throw IrisErrorTypeMismatch("Expected an array of uint64_t (msb = 0x8 or msb = 0xa)");
        }
    }


    /// Persist Null.
    void persistNull()
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *(readPos++);
        if (n != u64json::c_null)
        {
            readPos--;
            throw IrisErrorTypeMismatch("Got non-Null value when expecting Null.");
        }
    }
    
    
    /// Persist IrisValue.
    void persist(IrisValue &out)
    {
        IrisValue::Type type = getType();
        switch (type)
        {
        case IrisValue::Null: 
            out.setNull(); 
            persistNull();
            break;
            
        case IrisValue::Boolean:
            {
                bool b;
                persist(b);
                out.setBool(b);
            }
            break;

        case IrisValue::NumberU64:
            {
                uint64_t u64;
                persist(u64);
                out.setU64(u64);
            }
            break;
            
        case IrisValue::NumberS64:
            {
                int64_t s64;
                persist(s64);
                out.setS64(s64);
            }
            break;

        case IrisValue::NumberDouble:
            {
                double d;
                persist(d);
                out.setDouble(d);
            }
            break;

        case IrisValue::String:
            {
                std::string str;
                persist(str);
                out.setString(str);
            }
            break;

        case IrisValue::Message:
        case IrisValue::Object:
        case IrisValue::Array:
            {
                const uint64_t *data = readPos;
                skip();
                size_t size = readPos - data;
                out.setU64JsonDataForObjectOrArray(type, data, size);
            }
            break;
        }
    }

    
    /// Get type of current value (at current read position).
    /// This does not modify the read position.
    IrisValue::Type getType()
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        uint64_t n = *readPos;
        switch (n >> 60)
        {
        case 0: return IrisValue::NumberU64;
        case 1: return IrisValue::NumberS64;
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7: return IrisValue::String;
        case 8:
        case 0xa: return IrisValue::Array;
        case 0xb: return IrisValue::Object;
        case 0xc:
            switch (n >> 56)
            {
            case 0xc0: return IrisValue::NumberU64;
            case 0xc1: return IrisValue::NumberS64;
            case 0xca: return IrisValue::NumberDouble;
            case 0xcc: return IrisValue::String;
            case 0xcd: return IrisValue::Null;
            case 0xce: 
            case 0xcf: return IrisValue::Boolean;
            default:
                throw IrisErrorU64JsonEncodingError("Invalid MSB8 in uint64_t " + int64ToString(n, "0x%llX."));
            }
        case 0xe: return IrisValue::Message;
        case 0xf: return IrisValue::NumberU64;
        default:
            throw IrisErrorU64JsonEncodingError("Invalid MSB4 in uint64_t " + int64ToString(n, "0x%llX."));
        }        
    }
    

    /// Get size of current value (at current read position)
    /// This does not modify the read position.
    size_t getSizeOfCurrentValue() const
    {
        if (readPos >= dataEnd)
            throw IrisErrorU64JsonEncodingError("Out of data.");
        return getSizeOfValue(readPos);
    }

    /// Get size of an IrisU64Json value
    static size_t getSizeOfValue(const uint64_t *value)
    {
        uint64_t n = *value;
        switch (n >> 60)
        {
        case 0:
        case 1:
            return 1;
        case 2:
        case 3:
        case 4:
        case 5:
        case 6:
        case 7: // 2-7: Short string.
            return ((n & 255) >> 3) + 1;
        case 8:
            return (n & ~u64json::c_msb4mask) + 1;
        case 0xa:
        case 0xb: // Arrays and Objects have explicit container sizes.
            return n & ~u64json::c_msb4mask;
        case 0xc:
            switch (n >> 56)
            {
            case 0xc0:
            case 0xc1:
            case 0xca:
                return 2;
            case 0xcc:
                return (((n & ~u64json::c_msb8mask) + 7) >> 3) + 1;
            case 0xcd:
            case 0xce:
            case 0xcf:
                return 1;
            default:
                throw IrisErrorU64JsonEncodingError("Invalid MSB8 in uint64_t " + int64ToString(n, "0x%llX."));
            }
        case 0xe: // 0xervv_xxxx_xxxx_xxxx where xxx_... is container length
            return (n & ~u64json::c_msb16mask);
        case 0xf:
            return 1;
        default:
            throw IrisErrorU64JsonEncodingError("Invalid MSB4 in uint64_t " + int64ToString(n, "0x%llX."));
        }
    }


    /// Skip current value.
    void skip()
    {
        readPos += getSizeOfCurrentValue();
        if (readPos > dataEnd)
        {
            readPos = dataEnd;
            throw IrisErrorU64JsonEncodingError("Out of data.");
        }
    }


    /// Get requestId without decoding anything else.
    /// This throws IrisErrorException if message does not contain a U64JSON message
    static uint64_t getRequestId(const uint64_t *message)
    {
        if ((message[0] & u64json::c_msb4mask) == u64json::c_message)
        {
            return message[1];
        }
        else
        {
            throw IrisErrorException(E_u64json_encoding_error);
        }
    }

    /// Get isntId without decoding anything else.
    /// This throws IrisErrorException if message does not contain a U64JSON message
    static uint64_t getInstId(const uint64_t *message)
    {
        if ((message[0] & u64json::c_msb4mask) == u64json::c_message)
        {
            return message[2];
        }
        else
        {
            throw IrisErrorException(E_u64json_encoding_error);
        }
    }

    /// Return true if this U64Json encoded value is a message
    static bool isMessage(const uint64_t *message)
    {
        return (message[0] & u64json::c_msb4mask) == u64json::c_message;
    }

    /// Return true iff this message is a request or notification.
    /// Return false if this is a response or is malformatted.
    static bool isRequest(const uint64_t *message)
    {
        uint64_t value_type = message[0] & u64json::c_msb8mask;
        return (value_type == u64json::c_request) || (value_type == u64json::c_notification);
    }

    /// Return true iff this message is a response.
    /// Return false if this is a request, notification, or is malformatted.
    static bool isResponse(const uint64_t *message)
    {
        return (message[0] & u64json::c_msb8mask) == u64json::c_response;
    }

    /// Return true iff this message is an instanceRegistry_registerInstance() request.
    static bool is_instanceRegistry_registerInstance_request(const uint64_t *message)
    {
        if (!isRequest(message))
        {
            return false;
        }

        static const uint64_t u64_encoded_name[] = {
            0x636e6174736e6921, // instanc
            0x7274736967655265, // eRegistr
            0x7473696765725f79, // y_regist
            0x6e6174736e497265, // erInstan
            0x0000000000006563  // ce
        };

        for (size_t i=0; i < sizeof(u64_encoded_name)/sizeof(u64_encoded_name[0]); i++)
        {
            static const size_t method_offset = 3;
            if (u64_encoded_name[i] != message[i + method_offset])
            {
                return false;
            }
        }

        // If we reach here the method name matched
        return true;
    }


    static void readU64ArrayFromIrisValue(std::vector<uint64_t> &u64Array, const IrisValue &value)
    {
        IrisU64JsonReader reader(value);
        reader.persist(u64Array);
    }

    /// Copy a U64JSON value to a new buffer.
    /// Returns a newly allocated uint64_t array. This is owned by the caller
    /// and must be deleted by the caller.
    static uint64_t *copyValue(const uint64_t *value_to_copy)
    {
        size_t val_size = getSizeOfValue(value_to_copy);
        uint64_t *copy = new uint64_t[val_size];

        memcpy(copy, value_to_copy, val_size * sizeof(uint64_t));

        return copy;
    }

private:
    // Persist MSB 0xcc String.
    // This must only be called when readPos[0] is 0xcc.
    void persistLongString(std::string &out)
    {
        uint64_t len = readPos[0] & ~u64json::c_msb8mask;
        uint64_t containerLength = (len + 15) >> 3;
        if (readPos + containerLength > dataEnd) // >= is OK
            throw IrisErrorU64JsonEncodingError("Out of data.");
        // Note: This works only on little-endian machines.
        out = std::string(reinterpret_cast<const char *>(readPos) + 8, len);
        readPos += containerLength;
    }

    /// Current read position.
    const uint64_t *readPos;

    /// End of data.
    const uint64_t *dataEnd;
}; // class IrisU64JsonReader


NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisU64JsonReader_h

