/*!
 *  \file   IrisU64JsonWriter.h
 *  \brief  Generate U64JSON data.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  Generate U64JSON data.
 */

#ifndef ARM_INCLUDE_IrisU64JsonWriter_h
#define ARM_INCLUDE_IrisU64JsonWriter_h

#include "iris/detail/IrisCommon.h"
#include "iris/detail/IrisU64StringConstant.h"
#include "IrisInterface.h"
#include "IrisU64Json.h"
#include "IrisPersistPSH.h"
#include "IrisValue.h"
#include "IrisU64StringConstant.h"
#include <cassert>
#include <string>
#include <string.h>


NAMESPACE_IRIS_START

class IrisU64JsonReader;


/// Generate a U64JSON message.
/// This object contains and owns the generated Iris message (retreived through getMessage()).
/// The lifetime of the Irisc message returned by getMessage() is the same as the lifetime of this object.
/// This object allocates, owns and destroys the memory returned by getMessage().
//
/// The intended use pattern is:
/// void doFoo()
/// {
///   IrisU64JsonWriter writer;
///   writer.persist(...); /// also openObject() etc.
///   /// ...
///   irisInterface->irisHandleMessage(writer.getMessage());
/// }
class IrisU64JsonWriter
{
public:
    class Object
    {
    public:
        /// Copy and assignment allowed.

        /// Constructor: Leave object uninitialized.
        Object(): writer(0), startIndex(0) {}

        /// Constructor: Start object in writer.
        Object(IrisU64JsonWriter *writer_): writer(writer_)
        {
            assert(writer);
            writer->ensureSpace(2);
            startIndex = writer->getWriteIndex();
            // Mark container invalid so we fall over quickly when we do not close it.
            // close() overwrites this with the proper value.
            *((writer->writePos)++) = u64json::c_invalid;
            *((writer->writePos)++) = 0; // numMembers
        }

        /// Destructor: Implicitly close object.
        ~Object() { close(); }

        /// Close object: No futher members may be added after this.
        void close()
        {
            if (!writer)
                return;
            makeConsistent();
            writer = 0;
        }
        
        /// Make all fields of the object consistent.
        /// More members may be added after calling this. This is idempotent.
        /// This must be called before reading the object (for example for error messages).
        /// This is implicitly called by the destructor and close().
        /// This is rarely called directly. This allows you to print incomplete objects in error messages.
        void makeConsistent()
        {
            // Update container length and MSB.
            uint64_t containerLen = writer->getWriteIndex() - startIndex;
            writer->message_data[startIndex] = u64json::c_object | containerLen;
        }
                
        /// Persist object member and its value.
        template<class T>
        void member(const IrisU64StringConstant &name, const T &in)
        {
            writer->persist(name);
            iris::persist(*writer, in);
            writer->message_data[startIndex + 1]++;
        }

        /// Persist object member and its value.
        template<class T>
        void memberSlow(const std::string &name, const T &in)
        {
            writer->persist(name);
            iris::persist(*writer, in);
            writer->message_data[startIndex + 1]++;
        }

        /// Persist object member name. The caller must persist the value using writer functions immediately afterwards.
        void member(const IrisU64StringConstant &name)
        {
            writer->persist(name);
            writer->message_data[startIndex + 1]++;
        }

        /// Persist object member name. The caller must persist the value using writer functions immediately afterwards.
        void memberSlow(const std::string &name)
        {
            writer->persist(name);
            writer->message_data[startIndex + 1]++;
        }

        /// Persist null.
        void memberNull(const IrisU64StringConstant &name)
        {
            writer->persist(name);
            writer->persistNull();
            writer->message_data[startIndex + 1]++;
        }

        /// Persist null.
        void memberNullSlow(const std::string &name)
        {
            writer->persist(name);
            writer->persistNull();
            writer->message_data[startIndex + 1]++;
        }

        /// Persist object member name and a placeholder with a certain size.
        /// patchU60() can be used to fill the placeholder with a value.
        /// This function returns the index in message_data of the placeholder.
        size_t memberPlaceholder(const IrisU64StringConstant &name, size_t containerLength = 1)
        {
            writer->persist(name);
            size_t index = writer->writePos - writer->message_data;
            for (size_t i = 0; i < containerLength; i++)
                *((writer->writePos)++) = u64json::c_invalid;
            writer->message_data[startIndex + 1]++;
            return index;
        }

        /// Persist object member name and a placeholder with a certain size.
        /// patchU60() can be used to fill the placeholder with a value.
        /// This function returns the index in message_data of the placeholder.
        size_t memberPlaceholderSlow(const std::string &name, size_t containerLength = 1)
        {
            writer->persist(name);
            size_t index = writer->writePos - writer->message_data;
            for (size_t i = 0; i < containerLength; i++)
                *((writer->writePos)++) = u64json::c_invalid;
            writer->message_data[startIndex + 1]++;
            return index;
        }

        /// Persist object member and its value if isPresent==true. Donot persist anything if isPresent==false.
        /// (for persist() functions which are identical for read and write.)
        template<class T>
        void memberOptional(const IrisU64StringConstant &name, const T &in, bool isPresent)
        {
            if (isPresent)
                member(name, in);
        }

        /// Persist object member and its value if isPresent==true. Donot persist anything if isPresent==false.
        /// (for persist() functions which are identical for read and write.)
        template<class T>
        void memberOptionalSlow(const std::string &name, const T &in, bool isPresent)
        {
            if (isPresent)
                memberSlow(name, in);
        }
        
    private:
        /// IrisU64JsonWriter this object writes into.
        IrisU64JsonWriter *writer;

        /// Start index of this object in the writer->message_data array.
        /// This is stored as an index and not a pointer since the data array may be reallocated to increase its size.
        uint64_t startIndex;
    };
    friend class Object;

    class Array
    {
    public:
        /// Copy and assignment allowed.

        Array(IrisU64JsonWriter *writer_): writer(writer_)
        {
            writer->ensureSpace(2);
            startIndex = writer->getWriteIndex();
            // Mark container invalid so we fall over quickly when we do not close it.
            // close() overwrites this with the proper value.
            *((writer->writePos)++) = u64json::c_invalid;
            *((writer->writePos)++) = 0; // numMembers
        }

        ~Array() { close(); }

        void close()
        {
            if (!writer)
                return;
            // Update container length.
            uint64_t containerLen = writer->getWriteIndex() - startIndex;
            writer->message_data[startIndex] = u64json::c_array | containerLen;
            writer = 0;
        }

        /// Add one element to the array.
        template<class T>
        void element(const T &in)
        {
            iris::persist(*writer, in);
            writer->message_data[startIndex + 1]++;
        }

        /// Notify array container that n elements were just added to the array container.
        /// The elements are addes by calling functions on the writer directly (bypassing this object).
        /// This method has to be used to add arrays, objects and null (persistNull()).
        void elementsAdded(uint64_t n)
        {
            writer->message_data[startIndex + 1] += n;
        }

        /// Notify array container that one element was just added to the array container.
        /// The elements are addes by calling functions on the writer directly (bypassing this object).
        /// This method has to be used to add arrays, objects and null (persistNull()).
        void elementAdded()
        {
            writer->message_data[startIndex + 1]++;
        }

    private:
        /// IrisU64JsonWriter this array writes into.
        IrisU64JsonWriter *writer;

        /// Start index of this array in the writer->message_data array.
        /// This is stored as an index and not a pointer since the data array may be reallocated to increase its size.
        uint64_t startIndex;
    };
    friend class Array;

    class Request
    {
    public:
        /// Copy and assignment allowed.

        /// Constructor: Leave object uninitialized.
        Request()
            : writer(NULL)
            , startIndex(0)
        {
        }

        /// Constructor: Start object in writer.
        template <typename STRING>
        Request(IrisU64JsonWriter *writer_, const STRING &method, uint64_t dest_instId)
            : writer(writer_)
            , is_notification(true) // assume until a request id is set
        {
            assert(writer);
            writer->ensureSpace(3);
            startIndex = writer->getWriteIndex();
            // Mark container invalid so we fall over quickly when we do not close it.
            // close() overwrites this with the proper value.
            *((writer->writePos)++) = u64json::c_invalid;
            *((writer->writePos)++) = 0; // request id placeholder
            *((writer->writePos)++) = dest_instId;
            iris::persist(*writer, method);

            params = writer->openObject();
            // Also add instId as a parameter
            param(ISTR("instId"), dest_instId);
        }

        /// Destructor: Implicitly close object.
        ~Request() { close(); }

        /// Close object: Nothing further may be added after this.
        void close()
        {
            if (!writer)
                return;

            makeConsistent();
            writer = 0;
        }

        /// Make all fields of the object consistent.
        /// More params may be added after calling this. This is idempotent.
        /// This must be called before reading the request (for example for error messages).
        /// This is implicitly called by the destructor and close().
        /// This is rarely called directly. This allows you to print incomplete requests in error messages.
        void makeConsistent()
        {
            params.makeConsistent();

            // Update container length and MSB.
            uint64_t containerLen = writer->getWriteIndex() - startIndex;
            if (containerLen > u64json::c_msg_max)
            {
                throw IrisInternalError("U64JSON encoding error: message too large");
            }

            // message[0] = 0xetvv_xxxx_xxxx_xxxx
            //    t    = type (request=0, notification=1, response=2)
            //    vv   = jsonrpc version
            //    xx.. = container length
            writer->message_data[startIndex] = isNotification()? u64json::c_notification
                                                               : u64json::c_request;
            writer->message_data[startIndex] |= (u64json::c_jsonrpc2 | containerLen);
        }

        /// Persist param and its value.
        template<class T>
        void param(const IrisU64StringConstant &name, const T &in)
        {
            params.member(name, in);
        }

        /// Persist param and its value.
        template<class T>
        void paramSlow(const std::string &name, const T &in)
        {
            params.memberSlow(name, in);
        }

        /// Persist param name. The caller must persist the value using writer functions immediately afterwards.
        void param(const IrisU64StringConstant &name)
        {
            params.member(name);
        }

        /// Persist param name. The caller must persist the value using writer functions immediately afterwards.
        void paramSlow(const std::string &name)
        {
            params.memberSlow(name);
        }

        /// Persist null.
        void paramNull(const IrisU64StringConstant &name)
        {
            params.memberNull(name);
        }

        /// Persist null.
        void paramNullSlow(const std::string &name)
        {
            params.memberNullSlow(name);
        }

        /// Persist param name and a placeholder with a certain size.
        /// patchU60() can be used to fill the placeholder with a value.
        /// This function returns the index in message_data of the placeholder.
        size_t paramPlaceholder(const IrisU64StringConstant &name, size_t containerLength = 1)
        {
            return params.memberPlaceholder(name, containerLength);
        }

        /// Persist param name and a placeholder with a certain size.
        /// patchU60() can be used to fill the placeholder with a value.
        /// This function returns the index in message_data of the placeholder.
        size_t paramPlaceholderSlow(const std::string &name, size_t containerLength = 1)
        {
            return params.memberPlaceholderSlow(name, containerLength);
        }

        /// Persist param and its value if isPresent==true. Do not persist anything if isPresent==false.
        /// (for persist() functions which are identical for read and write.)
        template<class T>
        void paramOptional(const IrisU64StringConstant &name, const T &in, bool isPresent)
        {
            params.memberOptional(name, in, isPresent);
        }

        /// Persist param and its value if isPresent==true. Do not persist anything if isPresent==false.
        /// (for persist() functions which are identical for read and write.)
        template<class T>
        void paramOptionalSlow(const std::string &name, const T &in, bool isPresent)
        {
            params.memberOptionalSlow(name, in, isPresent);
        }

        /// Set the request id member
        void setRequestId(uint64_t request_id)
        {
            writer->message_data[request_id_offset] = request_id;
            is_notification = false;
        }

        /// Returns true if this request is a notification (has no request id set)
        bool isNotification() const
        {
            return is_notification;
        }

    private:
        /// IrisU64JsonWriter this request writes into.
        IrisU64JsonWriter *writer;

        /// Start index of this request in the writer->message_data array.
        /// This is stored as an index and not a pointer since the data array may be reallocated to increase its size.
        uint64_t startIndex;

        /// The request's 'params' member
        Object params;

        /// True if this is a notification (a request id has been set)
        bool is_notification;

        /// Field offset constants
        static const size_t request_id_offset = 1;
        static const size_t instId_offset     = 2;
    };
    friend class Request;

    class Response
    {
    public:
        /// Copy and assignment allowed.

        /// Constructor: Leave object uninitialized.
        Response()
            : writer(NULL)
            , startIndex(0)
        {
        }

        /// Constructor: Start object in writer.
        Response(IrisU64JsonWriter *writer_, uint64_t dest_instId)
            : writer(writer_)
        {
            assert(writer);
            writer->ensureSpace(3);
            startIndex = writer->getWriteIndex();
            // Mark container invalid so we fall over quickly when we do not close it.
            // close() overwrites this with the proper value.
            *((writer->writePos)++) = u64json::c_invalid;
            *((writer->writePos)++) = IRIS_UINT64_MAX; // request id placeholder
            *((writer->writePos)++) = dest_instId;

            // By default set this to a non-error response returning null.
            *((writer->writePos)++) = static_cast<uint64_t>(E_ok);
            writer->persistNull();

        }

        /// Destructor: Implicitly close object.
        ~Response() { close(); }

        /// Close object: Nothing further may be added after this.
        void close()
        {
            if (!writer)
                return;
            makeConsistent();
            writer = 0;
        }

        /// Make all fields of the object consistent.
        /// More params may be added after calling this. This is idempotent.
        /// This must be called before reading the response (for example for error messages).
        /// This is implicitly called by the destructor and close().
        /// This is rarely called directly. This allows you to print incomplete requests in error messages.
        void makeConsistent()
        {
            // Update container length and MSB.
            uint64_t containerLen = writer->getWriteIndex() - startIndex;
            if (containerLen > u64json::c_msg_max)
            {
                throw IrisInternalError("U64JSON encoding error: message too large");
            }

            // message[0] = 0xe1vv_xxxx_xxxx_xxxx
            //    vv   = jsonrpc version
            //    xx.. = container length
            writer->message_data[startIndex] = u64json::c_response | u64json::c_jsonrpc2 | containerLen;
        }

        /// Make this response an error response and set the error code and message.
        /// This version of setError() omits the optional data field but this can be added by
        /// persisting the data using writer functions immediately afterwards.
        void setError(IrisErrorCode code, const IrisU64StringConstant &message)
        {
            writer->message_data[startIndex + error_code_offset] = static_cast<uint64_t>(code);

            // Make sure we are writing to the correct offset.
            // Allows an error that has already been set to be replaced or allows a non-error
            // response to be changed to an error response.
            writer->writePos = &writer->message_data[startIndex + error_obj_offset];

            iris::persist(*writer, message);
        }

        /// Make this response an error response and set the error code and message.
        /// This version of setErrorSlow() omits the optional data field but this can be added by
        /// persisting the data using writer functions immediately afterwards.
        void setErrorSlow(IrisErrorCode code, const std::string &message)
        {
            writer->message_data[startIndex + error_code_offset] = static_cast<uint64_t>(code);

            // Make sure we are writing to the correct offset.
            // Allows an error that has already been set to be replaced or allows a non-error
            // response to be changed to an error response.
            writer->writePos = &writer->message_data[startIndex + error_obj_offset];

            iris::persist(*writer, message);
        }

        /// Make this response an error response and set the error code and message.
        /// This version of setError() includes the optional data field.
        template <typename DATA_T>
        void setError(IrisErrorCode code, const IrisU64StringConstant &message, const DATA_T &data)
        {
            // This part is the same whether we include the data field or not.
            setError(code, message);

            // Now add the data
            iris::persist(*writer, data);
        }

        /// Make this response an error response and set the error code and message.
        /// This version of setError() includes the optional data field.
        template <typename DATA_T>
        void setErrorSlow(IrisErrorCode code, const std::string &message, const DATA_T &data)
        {
            // This part is the same whether we include the data field or not.
            setErrorSlow(code, message);

            // Now add the data
            iris::persist(*writer, data);
        }

        /// Make this response a non-error response.
        /// The caller must persist the value using writer functions immediately afterwards.
        void setResult()
        {
            writer->message_data[startIndex + error_code_offset] = static_cast<uint64_t>(E_ok);

            // Make sure we are writing to the correct offset.
            // Allows a result that has already been set to be replaced or allows an error
            // response to be changed to a non-error response.
            writer->writePos = &writer->message_data[startIndex + result_offset];
        }

        /// Make this response a non-error response and set the result.
        template <typename RESULT_T>
        void setResult(const RESULT_T &data)
        {
            setResult();

            iris::persist(*writer, data);
        }

        /// Set the request id member
        void setRequestId(uint64_t request_id)
        {
            writer->message_data[startIndex + request_id_offset] = request_id;
        }

    private:
        /// IrisU64JsonWriter this response writes into.
        IrisU64JsonWriter *writer;

        /// Start index of this response in the writer->message_data array.
        /// This is stored as an index and not a pointer since the data array may be reallocated to increase its size.
        uint64_t startIndex;

        /// Field offset constants
        static const size_t request_id_offset = 1;
        static const size_t instId_offset     = 2;
        static const size_t error_code_offset = 3;
        static const size_t error_obj_offset  = 4; // } The error object and result have the same object.
        static const size_t result_offset     = 4; // } One of the other is used depending on the value of error code.
    };
    friend class Response;


    /// Constructor.
    IrisU64JsonWriter()
        : message_data(NULL)
        , writePos(0)
        , dataEnd(0)
    {
        // (No need to allocate memory upfront. We might be unused. ensureSpace() will automatically allocate memory.)
    }


    /// Destructor.
    ~IrisU64JsonWriter() { delete[] message_data; }


    /// Reset: Start writing at beginning of message again.
    void reset()
    {
        writePos = message_data;
    }


    /// Finalize and get message.
    /// May be called multiple times.
    /// The returned Iris message and also the data contained in it is owned by this object.
    /// The lifetime of the returned message and its data is the same as the lifetime of this object.
    const uint64_t *getMessage() const
    {
        return message_data;
    }

    /// Raw access to data and size.
    size_t getSize() const { return writePos - message_data; }


    /// Return true iff message is empty.
    bool isEmpty() const { return message_data == writePos; }
    
    
    /// Copy data from this writer into a IrisValue container.
    /// The writer is unaffected.
    void copyToIrisValue(IrisValue &value);

    /// Static function to copy an array of NumberU64 to IrisValue.
    static void writeU64ArrayToIrisValue(const std::vector<uint64_t> &u64Array, IrisValue &value);

    /// Static function to copy an array of uint8_t to IrisValue. The IrisValue will be encoded
    /// as a U64 array where each element of the array contains 8 uint8_t values.
    static void writePackedU8ArrayToIrisValue(const uint8_t* u8Array, uint64_t len, IrisValue &value);

    /// Open Object.
    Object openObject()
    {
        return Object(this);
    }


    /// Open Array.
    Array openArray()
    {
        return Array(this);
    }


    /// Open a request.
    template <typename STRING>
    Request openRequest(const STRING &func, uint64_t dest_instId)
    {
        return Request(this, func, dest_instId);
    }

    /// Open a response.
    Response openResponse(uint64_t dest_instId)
    {
        return Response(this, dest_instId);
    }


    /// Persist uint32_t.
    void persist(uint32_t in)
    {
        ensureSpace(1);
        // Small positive number (60 bits).
        *(writePos++) = uint64_t(in); // sets msb to 0
    }
    
    
    /// Persist uint64_t.
    void persist(uint64_t in)
    {
        ensureSpace(2); // need 1 or 2
        if ((in < 0x1000000000000000) || (in >= 0xf000000000000000))
        {
            // Small or high positive number (60 bits).
            *(writePos++) = in; // sets msb to 0 or 0xf
        }
        else
        {
            // All other positive numbers.
            *(writePos++) = u64json::c_u64;
            *(writePos++) = uint64_t(in);
        }
    }


    /// Persist int32_t.
    void persist(int32_t in)
    {
        ensureSpace(1);
        // Small positive or negative number (60 bits).        
        *(writePos++) = in >= 0 ? int64_t(in) : int64_t(in) & 0x1fffFFFFffffFFFFull; // sets msb to 0 or 1
    }
    
    
    /// Persist int64_t.
    void persist(int64_t in)
    {
        ensureSpace(2); // need 1 or 2
        if (uint64_t(in) < 0x1000000000000000)
        {
            // Small positive number (60 bits).
            *(writePos++) = uint64_t(in); // sets msb to 0
        }
        else if (uint64_t(in) >= 0xf000000000000000)
        {
            // Small negative number (60 bits).
            *(writePos++) = uint64_t(in) & 0x1fffffffffffffff; // sets msb to 1
        }
        else
        {
            // Big positive or negative number (all other numbers).
            *(writePos++) = u64json::c_s64;
            *(writePos++) = uint64_t(in);
        }
    }


    /// Persist double.
    void persist(double in)
    {
        ensureSpace(2); // need 2
        // Number as a double.
        *(writePos++) = u64json::c_double;
        const void *vp = &in;
        *(writePos++) = *reinterpret_cast<const uint64_t*>(vp);
    }

    /// Persist Boolean.
    void persist(bool in)
    {
        ensureSpace(1);
        *(writePos++) = in ? u64json::c_true : u64json::c_false;
    }


    /// Persist String.
    void persist(const char *in)
    {
        persist(std::string(in));
    }

    /// Persist String.
    void persist(const std::string &in)
    {
        // Note: This implementation only works on little endian machines.

        if ((in.length() <= 6) || ((in.length() <= 255) && (int8_t(in[6]) >= 0x20)))
        {
            // Short string < 256 chars with in[6] in 0x20-0x7f
            uint64_t len = in.length();
            uint64_t containerLen = (len >> 3) + 1;
            ensureSpace(containerLen);
            writePos[containerLen - 1] = 0; // Clear last element so that unused bytes are 0. This may write writePos[0].
            *writePos = 0x2000000000000000 | len;
            memcpy(reinterpret_cast<char*>(writePos) + 1, in.c_str(), len);
            writePos += containerLen;
        }
        else
        {            
            // Long string >= 256 chars or string >= 7 chars with in[6] not in 0x20..0x7f.
            persistLongString(in, u64json::c_string);
        }
    }
    
    /// Persist Array of uint8_t.
    void persist(const uint8_t* in, uint64_t u8len)
    {
        uint64_t u64len = (u8len + 7) / 8;

        ensureSpace(u64len + 1);
        *(writePos++) = u64json::c_arrayu64 + u64len;

        uint64_t pos = 0;
        for (uint64_t i = 0; i < u8len / 8; i++)
        {
            uint64_t u64Value = 0;
            for (uint64_t j = 0; j < 8; j++, pos++)
            {
                u64Value |= (uint64_t) in[pos] << (j * 8);
            }
            *(writePos++) = u64Value;
        }
        if (uint64_t bytesLeft = u8len % 8)
        {
            uint64_t u64Value = 0;
            for (uint64_t j = 0; j < bytesLeft; j++, pos++)
                u64Value |= (uint64_t) in[pos] << (j * 8);
            *(writePos++) = u64Value;
        }
    }

    /// Persist Array of NumberU64.
    void persist(const std::vector<uint64_t> &in)
    {
        uint64_t len = in.size();
        ensureSpace(len + 1);
        *(writePos++) = u64json::c_arrayu64 + len;

        if (len > 0)
        {
            memcpy(writePos, in.data(), len*sizeof(uint64_t));
        }
        writePos += len;
    }

    /// Persist string constant in U64JSON format.
    void persist(const IrisU64StringConstant &in)
    {
        ensureSpace(in.containerLength);
        *(writePos++) = in.data0;
        if (in.containerLength >= 2)
        {
            *(writePos++) = in.data1;
            if (in.containerLength >= 3)
            {
                assert(in.containerLength < 4);
                *(writePos++) = in.data2;
            }
        }
    }

    
    /// Persist true.
    void persistTrue()
    {
        ensureSpace(1);
        *(writePos++) = u64json::c_true;
    }


    /// Persist false.
    void persistFalse()
    {
        ensureSpace(1);
        *(writePos++) = u64json::c_false;
    }


    /// Persist Null.
    void persistNull()
    {
        ensureSpace(1);
        *(writePos++) = u64json::c_null;
    }


    /// Persist IrisValue.
    void persist(const IrisValue &in)
    {
        switch (in.getType())
        {
        case IrisValue::Null: 
            persistNull();
            break;
            
        case IrisValue::Boolean:
            persist(in.getBool());
            break;

        case IrisValue::NumberU64:
            persist(in.getU64());
            break;
            
        case IrisValue::NumberS64:
            persist(in.getS64());
            break;

        case IrisValue::NumberDouble:
            persist(in.getDouble());
            break;

        case IrisValue::String:
            persist(in.getString());
            break;

        case IrisValue::Message:
        case IrisValue::Object:
        case IrisValue::Array:
            {
                // Objects and array stored in IrisValues are already un U64JSON format.
                // We just need to append the data to ours.
                size_t size;
                const uint64_t *data = in.getU64JsonDataForObjectOrArray(size);
                ensureSpace(size);
                memcpy(writePos, data, size * sizeof(uint64_t));
                writePos += size;
            }
            break;
        }
    }

    
    /// Patch 60 bit wide low or high uint64_t value into existing message at index.
    void patchU60(size_t index, uint64_t in)
    {
        if (message_data + index >= dataEnd)
            throw IrisInternalError("IrisU64JsonWriter::patchU60(): index out of range.");
        if ((in < 0x1000000000000000) || (in >= 0xf000000000000000))
        {
            // Small or high positive number (60 bits).
            message_data[index] = in; // sets msb to 0 or 0xf
        }
        else
        {
            throw IrisInternalError("IrisU64JsonWriter::patchU60(): value out of range. Must be 60 bits with the 4 MSB either 0x0 or 0xf.");
        }
    }
    
    
    /// Add or overwrite an object member value potentially nested deep inside an object hierarchy.
    /// *this must be an empty writer when calling this function. This writer will contain a modified copy of originalMessage when this function returns.
    /// memberPath is the path to the member which is to be added or modified, using the dot as hierarchy separator. It may have the following formats:
    /// "": Replace the outermost object. originalMessage will not be used at all (will be completely replaced).
    /// "foo": Replace value of/add member foo in the outer most object.
    /// "foo.bar": Replace value of/add member bar in the outer object foo. And so on for deeper paths.
    /// Note: Do not use this to construct messages. Use the mechanisms of IrisaU64JsonWriter directly to construct messages. 
    ///       This functiuon is very inefficient and only exists to replace request ids in instanceRegistry_registerInstance()
    ///       calls which are an absolute corner case.
    void setObjectMember(IrisU64JsonReader &originalMessage, const std::string &memberPath, const IrisU64JsonWriter &value);


    /// Copy and encoded request and but modify its "id" field
    /// This function only exists to replace request ids in instanceRegistry_registerInstance().
    void copyMessageAndModifyId(const uint64_t *message, uint64_t new_id)
    {
        if ((message[0] & u64json::c_msb4mask) != u64json::c_message)
        {
            throw IrisErrorTypeMismatch("Expected message.");
        }

        size_t message_size = message[0] & ~u64json::c_msb16mask;

        ensureSpace(message_size);
        memcpy(writePos, message, message_size*sizeof(uint64_t));

        writePos[1] = new_id;

        // Set writePos to the end of the request
        writePos += message_size;
    }

    /// Make space for at least n elements from writePos.
    /// After this writePos[0..n-1] may be written.
    /// Internal menthod. Users never have to call this. This is just public to allow benchmarks to get rid of the memory allocation overhead.
    void ensureSpace(uint64_t n)
    {
        if (writePos + n <= dataEnd)
            return;

        increaseSize(n); // Will usually increase by more than n.
    }

    /// Returns true if nothing has been written in to this writer.
    bool empty() const
    {
        return getWriteIndex() == 0;
    }

private:
    
    // Increase size of message->data at least by n elements and relocate writePos and dataEnd.
    void increaseSize(uint64_t n);


    /// Get current write index into message_data.
    uint64_t getWriteIndex() const
    {
        return uint64_t(writePos - message_data);
    }
    
    
    /// Persist long string (MSB 0xcc).
    void persistLongString(const std::string &in, uint64_t msb)
    {
        uint64_t len = in.length();
        uint64_t containerLen = (len + 15) >> 3;
        ensureSpace(containerLen);
        writePos[containerLen - 1] = 0; // Clear last element so that unused bytes are 0. This may write writePos[0].
        *writePos = msb | len;
        memcpy(writePos + 1, in.c_str(), len);
        writePos += containerLen;
    }


    /// Internal recursive implementation of setObjectMember().
    void setObjectMemberInternal(IrisU64JsonReader &originalMessage, const std::vector<std::string> &members, size_t memberIndex, const IrisU64JsonWriter &value, bool insertObject);
    
    
    /// Iris message container manipulated by this object.
    uint64_t *message_data;

    /// Current write position.
    uint64_t *writePos;

    /// End of allocated memory pointed to by message->data.
    uint64_t *dataEnd;
}; // class IrisU64JsonWriter




NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisU64JsonWriter_h

