/*!
 *  \file   IrisU64StringConstant.h.
 *  \brief  Define compile-time string constants in U64JSON format.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 * 
 * Define compile-time string constants in U64JSON format.
 * 
 * Usage:
 * ISTR("foo")
 * Defines a string constant in U64JSON format for the string "foo". The string must be <= 23 chars long.
 * This defines a constant IrisU64StringConstant() struct instance with:
 * - contentLength (length of the U64JSON scontainer for this string, this is 1..3)
 * - data0         (first U64 element of the string)
 * - data1         (second U64 element of the string, only for strings >= 8 chars)
 * - data2         (third U64 element of the string, only for strings >= 16 chars)
 * 
 * This is only useful (and intended) for short constant strings like parameter names, object member names etc.
 * For strings >= 24 chars just use 'const char *' or std::string as usual.
 * 
 * Using ISTR() is just a performance optimization and does not provide new functionality.
 * 
 * Example:
 *   params.addArg(ISTR("rscId"), 42);
 * This is functionally identical to, but faster as:
 *   params.addArg("rscId", 42);
 * 
 */

#ifndef ARM_INCLUDE_IrisU64StringConstant_h
#define ARM_INCLUDE_IrisU64StringConstant_h

#include "iris/detail/IrisCommon.h"
#include "IrisInterface.h" // uint64_t
#include <string>


NAMESPACE_IRIS_START


#define ISTR(s) DefineIrisU64StringConstant(s)

#define DefineIrisU64StringConstant(s) DefineIrisU64StringConstant_internal(s "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0")
#define DefineIrisU64StringConstant_internal(s) ((void)::iris::IrisU64StringConstantStaticAssert<(sizeof(s) < 25 + 22)>::string_too_long_error__must_be_max_23_chars,::iris::IrisU64StringConstant(sizeof(s) - 23, s[0], s[1], s[2], s[3], s[4], s[5], s[6], s[7], s[8], s[9], s[10], s[11], s[12], s[13], s[14], s[15], s[16], s[17], s[18], s[19], s[20], s[21], s[22]))

struct IrisU64StringConstant
{
    IrisU64StringConstant(unsigned size, uint8_t c0, uint8_t c1, uint8_t c2, uint8_t c3, uint8_t c4, uint8_t c5, uint8_t c6, uint8_t c7, uint8_t c8, uint8_t c9, uint8_t c10, uint8_t c11, uint8_t c12, uint8_t c13, uint8_t c14, uint8_t c15, uint8_t c16, uint8_t c17, uint8_t c18, uint8_t c19, uint8_t c20, uint8_t c21, uint8_t c22):
        containerLength((size >> 3) + 1),
        data0(size | (uint64_t(c6?c6:0x20) << 56) | (uint64_t(c5) << 48) | (uint64_t(c4) << 40) | (uint64_t(c3) << 32) | (c2 << 24) | (c1 << 16) | (c0 << 8)),
        data1((uint64_t(c14) << 56) | (uint64_t(c13) << 48) | (uint64_t(c12) << 40) | (uint64_t(c11) << 32) | (c10 << 24) | (c9 << 16) | (c8 << 8) | c7),
        data2((uint64_t(c22) << 56) | (uint64_t(c21) << 48) | (uint64_t(c20) << 40) | (uint64_t(c19) << 32) | (c18 << 24) | (c17 << 16) | (c16 << 8) | c15) {}
    
    /// Length of U64JSON container. 1, 2 or 3.
    uint64_t containerLength;
    
    /// U64JSON data of string.
    uint64_t data0, data1, data2;
    
    /// Max length.
    static const size_t MaxLength = 23;
};
    
// Compile-time assert that string is <= 23 chars.
template<bool B>
struct IrisU64StringConstantStaticAssert {};
template<>
struct IrisU64StringConstantStaticAssert<true> { static const int string_too_long_error__must_be_max_23_chars = 0; };

    
/// Convert back to string.
/// This should only ever be used for error messages and other non-perfornace critical situations as it defeates the sole purpose of IrisU64StringConstant.
inline std::string toString(const IrisU64StringConstant &str)
{
    return std::string(reinterpret_cast<const char *>(&str.data0) + 1, str.data0 & 255);
}
    
    
NAMESPACE_IRIS_END
    
#endif // #ifndef ARM_INCLUDE_IrisU64StringConstant_h
