/*!
 *  \file   IrisUtils.h
 *  \brief  Utilities used by the Iris framework.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  Utilities used by the Iris framework.
 */

#ifndef ARM_INCLUDE_IrisUtils_h
#define ARM_INCLUDE_IrisUtils_h

#include "iris/detail/IrisCommon.h"
#include "IrisInterface.h" // uint64_t
#include <string>
#include <vector>
#include <stdio.h>
#include <iostream>
#include <iomanip>
#include <sstream>

#ifdef _MSC_VER
// Note: _snprintf is deprecated; use _snprintf_s instead
# define iris_snprintf(x, y, format, ...) _snprintf_s(x, y, _TRUNCATE, format, __VA_ARGS__)
# define iris_strtoull _strtoui64
# define iris_setenv(x,y,z) _putenv_s(x,y)
#else
# define iris_snprintf snprintf
# define iris_strtoull strtoull
# define iris_setenv(x,y,z) setenv(x,y,z)
#endif

#ifdef __GNUC__
    #define INTERNAL_IRIS_PRINTF(fmt_index, args_index) __attribute__((format(printf, fmt_index, args_index)))
#else
    #define INTERNAL_IRIS_PRINTF(fmt_index, args_index)
#endif
NAMESPACE_IRIS_START

    /// Stay independent of the messy inttypes.h, stdint.h and cstdint headers which have
    /// conflicting semantics and availability on different compilers. We want Iris headers to
    /// always compile, independent of any other previous #include statements.
    static const uint64_t IRIS_UINT64_MAX = uint64_t(0xffffFFFFffffFFFFull);
    static const uint64_t IRIS_UINT64_MSB1 = uint64_t(0x8000000000000000ull);
    static const uint64_t IRIS_UINT64_MSB4 = uint64_t(0xf000000000000000ull);
    static const int64_t IRIS_INT64_MIN = int64_t(-0x8000000000000000ll);
    static const int64_t IRIS_INT64_MAX = int64_t(0x7fffFFFFffffFFFFll);


    /// Convert int64_t/uint64_t to string.
    /// format must expect one 'll' integer type as argument and must generate a string which is shorter than 80 chars.
    inline std::string int64ToString(int64_t i, const char *format = "%lld")
    {
        char buf[80];
        iris_snprintf(buf, sizeof(buf), format, (long long)i);
        buf[sizeof(buf) - 1] = 0;
        return buf;
    }


    /// Return true iff str starts with prefix.
    inline bool stringStartsWith(const std::string &str, const std::string& prefix)
    {
        return str.compare(0, prefix.size(), prefix) == 0;
    }

    /// Return true iff str starts with prefix (case insensitive compare).
    inline bool caseInsensitiveStringStartsWith(const std::string &str, const std::string& prefix)
    {
        if (prefix.length() > str.length())
            return false;

        for (unsigned i = 0; i < prefix.length(); i++)
        {
            if (tolower(str[i]) != tolower(prefix[i]))
                return false;
        }

        return true;
    }

    /// Compares two strings, ignoring case.
    inline bool caseInsensitiveEqual(const std::string &str1, const std::string& str2)
    {
        if (str1.length() != str2.length())
            return false;

        return caseInsensitiveStringStartsWith(str1, str2);
    }

    /// Strip spaces at the beginning and the end of a string.
    inline std::string stripSpaces(const std::string & s)
    {
        size_t start = s.find_first_not_of(" \t\n\r\v");
        if (start == std::string::npos)
            return std::string();
        size_t end = s.find_last_not_of(" \t\n\r\v");
        return s.substr(start, end + 1 - start);
    }


    /// Construct a valid C identifier from an arbitrary (non-C-identifier) string using the 'RegisterInfo cname' rules:
    /// Applied in this order:
    /// - All non-C identifier chars are replaced with underscores.
    /// - If the first char is a digit an 'N' is prepended.
    /// - If the first char is an underscore an 'U' is prepended.
    /// - If name is the empty string then "NONAME" is returned.
    /// Note that the last 3 rules never kick in for virtually all object identifying names. R9 stays R9 and 'FPU Registers' becomes FPU_Registers.
    /// Note that the focus is not on uniqueness but on readability of the C-identifier, since this is part of the user interface.
    /// name: Arbitrary input string.
    /// Return cname: Valid C identifier based on name, not starting with an underscore.
    inline std::string nameToCname(const std::string& name)
    {
        if (name.empty())
            return "NONAME";
        std::string r = name;
        for (size_t i = 0; i < r.length(); i++)
            if (!isalnum((unsigned)r[i]))
                r[i] = '_';
        if (isdigit(name[0]))
            return "N" + r;
        if (name[0] == '_')
            return "U" + r;
        return r;
    }


    /// Split string at separator char.
    /// Optionally strip spaces in the beginning and the end of each string after splitting.
    /// fieldsOut is overwritten and contains the fields after splitting.
    /// The separator char is not part of the fields after splitting.
    /// Splitting an empty string results in an empty list of strings.
    /// Splitting a non-empty string which does not contain any separator results in a list with exactly one string.
    /// Sequences of separators or separators at the beginning or the end of the string result in empty fields.
    inline void splitString(std::vector<std::string> &fieldsOut, const std::string &in, char separator, int maxSplit = 0, bool stripSpaces_ = false)
    {
        fieldsOut.clear();
        if (in.empty())
            return;

        size_t pos = 0;
        for (int split = 0; pos != std::string::npos; split++)
        {
            std::string field;
            size_t next = in.find(separator, pos);
            if ((next == std::string::npos) || ((maxSplit > 0) && (split >= maxSplit)))
            {
                field = in.substr(pos);
                pos = std::string::npos;
            }
            else
            {
                field = in.substr(pos, next - pos);
                pos = next + 1;
            }
            if (stripSpaces_)
                field = stripSpaces(field);
            fieldsOut.push_back(field);
        }
    }


    /// Join strings.
    /// An empty list results in an empty string.
    /// A list with one entry results in this entyr being returned.
    inline std::string joinString(const std::vector<std::string> &fields, const std::string &separator)
    {
        if (fields.empty())
            return std::string();
        std::string r = fields[0];
        for (size_t i = 1; i < fields.size(); i++)
        {
            r.append(separator);
            r.append(fields[i]);
        }
        return r;
    }


    /// Replace substrings string.
    /// Does nothing if from is empty.
    inline void replaceString(std::string &str, const std::string &from, const std::string &to)
    {
        if (from.empty())
            return;
        for (size_t pos = 0;;)
        {
            size_t start = str.find(from, pos);
            if (start == std::string::npos)
                break;
            str.replace(start, from.length(), to);
            pos = start + to.length();
        }
    }


    /// Expand unpritable characters unto C escape sequences.
    inline std::string expandUnprintable(const std::string &str)
    {
        std::string r;
        for (size_t i = 0; i < str.length(); i++)
        {
            switch (str[i])
            {
            case  '\r': r += "\\r"; break;
            case  '\n': r += "\\n"; break;
            case  '\v': r += "\\v"; break;
            case  '\t': r += "\\t"; break;
            case  '\b': r += "\\b"; break;
            case  '\a': r += "\\a"; break;
            case  '\0': r += "\\0"; break;
            default:
                if (isprint((unsigned char)str[i]))
                    r += str[i];
                else
                    r += int64ToString((unsigned char)str[i], "\\x02x");
            }
        }
        return r;
    }


    /// Emit correctly quoted string, escaping non-printable chars and quotes (according to JSON rules).
    inline std::ostream &quoteString(std::ostream &os, const char *str, size_t size, bool addQuotes)
    {
        if (addQuotes)
            os << '"';
        const char *end = str + size;
        for (const char *s = str; s < end; s++)
        {
            if ((*s == '\\') || (*s == '"') || (*s < '\x20'))
            {
                os.write(str, s - str); // Write all non-special characters up to this point
                switch (*s)
                {
                case '\b': os << "\\b"; break;
                case '\f': os << "\\f"; break;
                case '\n': os << "\\n"; break;
                case '\r': os << "\\r"; break;
                case '\t': os << "\\t"; break;
                case '\\': os << "\\\\"; break;
                case '"' : os << "\\\""; break;
                default:
                    os << "\\u"
                       << std::hex
                       << std::setw(4)
                       << std::setfill('0')
                       << reinterpret_cast<const unsigned char&>(*s) << std::dec << std::setw(0);
                    break;
                }

                str = s + 1; // Update str to the next char to write
            }
        }

        os.write(str, end - str); // Write all remaining characters to the end of the string

        if (addQuotes)
            os << '"';
        return os;
    }


    /// Emit correctly quoted string, escaping non-printable chars and quotes (according to JSON rules).
    inline std::string quoteString(const std::string &str, bool addQuotes = true)
    {
        std::stringstream ss;
        quoteString(ss, str.c_str(), str.length(), addQuotes);
        return ss.str();
    }


    /// Return true iff str should be parsed withg strtod() rather than with strtoull().
    /// This does not determine if s is avalid floatimng point number which is checked
    /// when really parsing the floating point number.
    /// This handles: Floating point, hex, dec.
    inline bool looksLikeFloatingPoint(const std::string &s)
    {
        if (s.find_first_of(".pP") != std::string::npos)
            return true; // Float-hex or float using a dot.
        if ((s.length() < 2) || (tolower(s[1]) == 'x'))
            return false; // Just one char or integer-hex.
        if ((s.length() >= 3) && (tolower(s[2]) == 'x'))
            return false; // Signed hex.
        if (s.find_first_of(".eE") != std::string::npos)
            return true;
        return false;
    }

    /// Return true iff path (hierarchical path with dots) starts with prefix where prefix must
    /// either be exactly the same as path or must be one or more leading elements of the path.
    ///
    /// Always returns true for an empty prefix.
    /// isPrefixOfHierarchicalPath("foo.bar", "foo.bar") returns true.
    /// isPrefixOfHierarchicalPath("foo.bar", "foo") returns true.
    /// isPrefixOfHierarchicalPath("foo.bar", "fo") returns false.
    /// isPrefixOfHierarchicalPath("foo.bar", "") returns true.
    inline bool isPrefixOfHierarchicalPath(const std::string &str, const std::string& prefix)
    {
        if (prefix.empty())
            return true;
        if (str.compare(0, prefix.size(), prefix) != 0)
            return false;
        if (str.length() == prefix.length())
            return true;
        return str[prefix.length()] == '.';
    }

    /// Compiler and memory barrier.
    /// This is useful for situations where explicit synchronisation through mutexes is not necessary,
    /// for example for flags set only once by an alien thread:
    /// void publishValue(int val)
    /// {
    ///     value = val;
    ///     irisCompilerAndMemoryBarrier(); // <== This is necessary so the compiler (or CPU) does not set hasValue to true before 'value' was updated.
    ///     hasValue = true;
    /// }
    ///
    /// Note: Calling this function is not necessary in a lot of places since calling any external non-inlined function of any library
    /// (for example all pthread_* functions, all libc functions etc) also acts as a compiler and memory barrier.
    /// Calling this is usually only necessary when playing tricks without mutexes.
    void irisCompilerAndMemoryBarrier();

    /// Generate a std::string using printf style formatting
    std::string stringvprintf(const char *format, va_list args);

    uint64_t getEnvU64(const std::string &var, uint64_t default_val=0);

NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisUtils_h

