/*!
 *  \file   IrisValue.h
 *  \brief  Representation of an arbitrary JSON value in C++.
 *  \date   Copyright ARM Limited 2015 All Rights Reserved.
 *
 *  Representation of an arbitrary JSON value in C++.
 * 
 *  Note: Most aspects of Iris are 100% statically typed. Use IrisValue only if you have to
 *  e.g. if an argument/member may have any value (i.e. is not statically typed). Using this
 *  is exotic.
 * 
 *  Design goals:
 *  - Lean memory footprint. Value semantics (copy).
 *  - Direct and easy access to Number, String and Boolean values from C++.
 *  - Can also contain arbitary JSON values (like nested arrays and objects).
 * 
 *  Use cases:
 *  - values in InstanceProperties object (mostly String or Number, but may be anything)
 *  - defval (default value) in AttributeInfo (mostly String or Number, but may be anything)
 */

#ifndef ARM_INCLUDE_IrisValue_h
#define ARM_INCLUDE_IrisValue_h

#include "iris/detail/IrisCommon.h"
#include "IrisInterface.h" // uint64_t
#include "IrisErrorException.h"
#include "IrisU64Json.h"
#include <string>
#include <stdio.h>

NAMESPACE_IRIS_START



class IrisValue
{
public:
    /// JSON Type.
    enum Type 
    {
        Null,      /// Access via getType(). Internally stored in type.
        Boolean,   /// Access via getBool(), getU64() or getS64(). Internally stored in number.
        NumberU64, /// Access via getU64(). Internally stored in number.
        NumberS64, /// Access via getS64(). Internally stored in number.
        NumberDouble, /// Access via getDouble(). Internally stored in number.
        String,    /// Access via getString(). Internally store in str.
        Array,     /// Access via getU64JsonData() and getU64DataSize(). Internally stored in str.
        Object,    /// Access via getU64JsonData() and getU64DataSize(). Internally stored in str.
        Message,   /// Access via getU64JsonData() and getU64DataSize(). Internally stored in str.
    };
    
    // Copy is allowed (implicit operator= and copy constructor).

    /// Construct Null.
    IrisValue(): type(Null), number(0) {}

    /// Get type.
    Type getType() const { return type; }
    
    
    // -- get/set access for each type ---
    
    /// null    
    void getNull() const
    {
        // Just check type.
        if (type !=Null)
            throw IrisErrorTypeMismatch("Not implicitly converting non-Null type to Null.");
    }
    void setNull() { type = Null; }
    // Default constructor constructs null.
    // No assignment operator for null. Assign IrisValue() to get a null.
  
    /// bool
    bool getBool() const
    { 
        switch (type)
        {
        case Boolean: return number != 0;
        case NumberU64:
        case NumberS64:
            if (number == 0) return false;
            else if (number == 1) return true;
            else throw IrisErrorTypeMismatch("Not implicitly converting non-0/1 integer to bool.");
        default:
            throw IrisErrorTypeMismatch("Not implicitly converting Null/String/Array/Object/NumberDouble to bool.");
        }
    }
    void get(bool &b) const { b = getBool(); }
    void setBool(bool b) { type = Boolean; number = b; }
    void set(bool b) { setBool(b); }
    explicit IrisValue(bool b) { set(b); } // Allow only explicit conversion because there are way too many implicit conversions (for example from all pointers) to bool in C++.
    IrisValue &operator=(bool b) { set(b); return *this; }
    
    /// U64
    uint64_t getU64() const
    { 
        switch (type)
        {
        case Boolean:
        case NumberU64:
            return number;
        case NumberS64:
            if (int64_t(number) >= 0) return number;
            else throw IrisErrorTypeMismatch("Not implicitly converting negative integer to NumberU64.");
        default:
            throw IrisErrorTypeMismatch("Not implicitly converting Null/String/Array/Object/NumberDouble to NumberU64.");
        }
    }
    void get(uint64_t &u64) const { u64 = getU64(); }
    void setU64(uint64_t u64) { type = NumberU64; number = u64; }
    void set(uint64_t u64) { setU64(u64); }
    IrisValue(uint64_t u64) { set(u64); }
    IrisValue &operator=(uint64_t u64) { set(u64); return *this; }

    // Initialize a U64 from smaller uint types
    IrisValue(uint32_t u32) { set(uint64_t(u32)); }
    IrisValue &operator=(uint32_t u32) { set(uint64_t(u32)); return *this; }

    IrisValue(uint16_t u16) { set(uint64_t(u16)); }
    IrisValue &operator=(uint16_t u16) { set(uint64_t(u16)); return *this; }

    IrisValue(uint8_t u8) { set(uint64_t(u8)); }
    IrisValue &operator=(uint8_t u8) { set(uint64_t(u8)); return *this; }

    /// S64
    int64_t getS64() const
    { 
        switch (type)
        {
        case Boolean:
        case NumberS64:
            return int64_t(number);
        case NumberU64:
            if (int64_t(number) >= 0) return number;
            else throw IrisErrorTypeMismatch("Not implicitly converting high positive integer to NumberS64.");
        default:
            throw IrisErrorTypeMismatch("Not implicitly converting Null/String/Array/Object/NumberDouble to NumberS64.");
        }
    }
    void get(int64_t &s64) const { s64 = getS64(); }
    void setS64(int64_t s64) { type = NumberS64; number = uint64_t(s64); }
    void set(int64_t s64) { setS64(s64); }
    IrisValue(int64_t s64) { set(s64); }
    IrisValue &operator=(int64_t s64) { set(s64); return *this; }

    // Initialize an S64 from smaller int types
    IrisValue(int32_t u32) { set(int64_t(u32)); }
    IrisValue &operator=(int32_t u32) { set(int64_t(u32)); return *this; }

    IrisValue(int16_t u16) { set(int64_t(u16)); }
    IrisValue &operator=(int16_t u16) { set(int64_t(u16)); return *this; }

    IrisValue(int8_t u8) { set(int64_t(u8)); }
    IrisValue &operator=(int8_t u8) { set(int64_t(u8)); return *this; }

    /// double
    double getDouble() const
    { 
        switch (type)
        {
        case Boolean:
        case NumberS64:
            return double(int64_t(number));
        case NumberU64:
            return double(number);
        case NumberDouble:
            {
                const void *vp = &number;
                return *reinterpret_cast<const double*>(vp);
            }
        default:
            throw IrisErrorTypeMismatch("Not implicitly converting Null/String/Array/Object to NumberDouble.");
        }
    }
    void get(double &d) const { d = getDouble(); }
    void setDouble(double d) { type = NumberDouble; const void *vp = &d; number = *reinterpret_cast<const uint64_t*>(vp); }
    void set(double d) { setDouble(d); }
    IrisValue(double d) { set(d); }
    IrisValue &operator=(double d) { set(d); return *this; }

    /// string
    std::string getString() const
    { 
        switch (type)
        {
        case String: return str;
        default:
            throw IrisErrorTypeMismatch("Not implicitly converting non-String to String.");
        }
    }
    void get(std::string &str_) const { str_ = getString(); }
    void setString(const std::string &str_) { type = String; str = str_; }
    void set(const std::string &str_) { setString(str_); }
    IrisValue(const char * str_) { set(std::string(str_)); }
    IrisValue(const std::string &str_) { set(str_); }
    IrisValue &operator=(const std::string &str_) { set(str_); return *this; }

    /// Access for objects and arrays.
    /// Use an IrisU64JsonReader(this) to read objects and arrays.
    const uint64_t *getU64JsonDataForObjectOrArray(size_t &sizeOut) const
    {
        switch (type)
        {
        case Object:
        case Array:
        case Message:
            sizeOut = size_t(str.length() / 8);
            return (const uint64_t*)(str.data());
        default:
            throw IrisErrorTypeMismatch("getU64JsonDataForObjectOrArray() may only be called for Object or Array.");
        }
    }
    void setU64JsonDataForObjectOrArray(Type type_, const uint64_t *data, size_t size)
    {
        switch (type_)
        {
        case Object:
        case Array:
        case Message:
            type = type_;
            str = std::string((const char *)data, size_t(size * 8));
            break;
        default:
            throw IrisErrorTypeMismatch("setU64JsonDataForObjectOrArray() may only be called for Object or Array.");
        }        
    }

    IrisValue(Type type_, const uint64_t *data, size_t size)
    {
        setU64JsonDataForObjectOrArray(type_, data, size);
    }

    IrisValue(Type type_, const std::vector<uint64_t> &data)
    {
        setU64JsonDataForObjectOrArray(type_, &data[0], data.size());
    }

    /// Return true iff the value is Null, empty string, 0, empty array or empty object.
    bool isEmptyOrNull() const
    {
        switch(type)
        {
        case Null: return true;
        case Boolean:
        case NumberU64:
        case NumberS64:
            return number == 0;
        case NumberDouble:
            return (number == 0) || (number == IRIS_UINT64_MSB1); // 0.0 or -0.0
        case String:
            return str.empty();
        case Array:
        case Object:
        case Message:
            {
                if (str.length() < 8) 
                    return true;
                const uint64_t *data = (const uint64_t*)(str.data());
                switch ((*data) & IRIS_UINT64_MSB4)
                {
                case u64json::c_arrayu64: return (*data & ~IRIS_UINT64_MSB4) == 0;
                case u64json::c_array:
                case u64json::c_object: return (*data & ~IRIS_UINT64_MSB4) == 2;
                case u64json::c_message:
                    {
                        switch((*data) & u64json::c_msb8mask)
                        {
                        case u64json::c_request:
                        case u64json::c_notification: return (*data & ~u64json::c_msb16mask) == 3;
                        case u64json::c_response: return (*data & ~u64json::c_msb16mask) == 4;
                        }
                    }
                }
            }
            return false;
        }
        return false; // Can never come here.
    }
    
    
private:
    /// Type of this value.
    /// This also determines how data is stored in number or str.
    Type type; 
    
    /// Storage for NumberU64, NumberS64, Boolean, NumberDouble (0/1 for false/true).
    uint64_t number;
    
    /// Storage for String, Array and Object.
    /// String: Directly stores the string.
    /// Array and Object: Stores the U64JSON data for these.
    std::string str;
};
    
    
NAMESPACE_IRIS_END

#endif // #ifndef ARM_INCLUDE_IrisValue_h

