/**
 * \file    Framework/include/DebugCoffSymbols.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup debug_info_database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef COFFDEBUGSYMBOLS_H
#define COFFDEBUGSYMBOLS_H

#include <cstdint>

#include <map>
#include <string>
#include <vector>
#include <list>
#include <iostream>

#include "Coff.h"
#include "DebugDWARFInfo.h"
#include "Dwarf2Constants.h"


namespace DebugInfoDB
{
#define nameLength 1024
    struct filenamePos
    {
        std::string filename;
        uint32_t address;
    };
    class SymbolEntry
    {

    public:
        void Print();
        // container for CoffArchSyment
        struct COFFArchSyment coffArchSyment;

        // add alternative symbol table entries here
        // ...
        std::list <union COFFAuxent *> AuxEntries;
        uint32_t address;
    };

    class CoffDebugSymbol
    {
    };

    class CoffVariableInfo
    {
    public:
        // constructor
        CoffVariableInfo();
        ~CoffVariableInfo();
        CoffVariableInfo(const CoffVariableInfo &variableInfo);
        uint32_t type;
        std::string name;
        bool isStackVariable;
        bool isRegParam;
        int32_t indexOrOffset;

    };
    class CoffBlockInfo
    {

    public:
        // constructor
        CoffBlockInfo();
        ~CoffBlockInfo();
        CoffBlockInfo(const CoffBlockInfo &blockInfo);
        std::list<CoffVariableInfo> variableList;
        uint64_t startAddress;
        uint64_t endAddress;
        std::string functionName;
    };

    class CoffFunctionInfo
    {

    public:
        // constructor
        CoffFunctionInfo();
        ~CoffFunctionInfo();
        CoffFunctionInfo(const CoffFunctionInfo &functionInfo);
        std::list <CoffBlockInfo> basicBlockList;
        std::list<CoffVariableInfo> variableList;
        uint64_t startAddress;
        uint64_t endAddress;
        std::string functionName;
    };



    // The set of all Debug Symbols in one file is stored in the CoffDebugSymbols
    class CoffDebugSymbols
    {
    public:
        CoffDebugSymbols();
        std::string GetFilename(){return filename;}

        // get a list of all symbols for one storage class
        std::list<SymbolEntry *> &GetSymbols(uint8_t storageClass);
        void Print();

    protected:

        friend class CoffDebugSymbolsDB;
    private:

        // access functions
        uint32_t GetWordFromCharArray(const uint8_t data[], uint32_t pos, bool bigendian);
         uint16_t GetShortFromCharArray(const uint8_t data[], uint32_t pos, bool bigendian);
        void AddNewCoffDebugSymbols(const uint8_t data[], const uint8_t strings[],uint32_t &currentPos, uint32_t size,uint32_t stringSize,bool bigendian,std::string &sectionFileName, DWARFInfoDB *dwarfInfo,std::list<filenamePos> &coffFilesInSegment);
        void InsertSymbolTableEntryIncludingAuxEntries(const uint8_t data[], const uint8_t strings[], uint32_t &currentPos, uint32_t size,uint32_t stringSize,bool bigendian, bool &filenameDetected,std::string &sectionFileName,std::list<filenamePos> &coffFilesInSegment);
        void ProcessFunctionsAndVariables(DWARFInfoDB *dwarfInfo,RCPointer<DWARFInfoTreeNode> insertionPoint,std::map<uint32_t,RCPointer<DWARFInfoTreeNode> > &typeMap,uint64_t &start, uint64_t &end);
        void ProcessGlobalVariables(DWARFInfoDB *dwarfInfo,RCPointer<DWARFInfoTreeNode> insertionPoint,std::map<uint32_t,RCPointer<DWARFInfoTreeNode> > &typeMap);
        void AddVariable(DWARFInfoDB *dwarfInfo,std::list<CoffVariableInfo>::iterator &varIt,RCPointer<DWARFInfoTreeNode> &newFunction,uint64_t startAddress,uint64_t endAddress,std::map<uint32_t,RCPointer<DWARFInfoTreeNode> > &typeMap);
        void InitializeBasicTypeTable(std::map<uint32_t,RCPointer<DWARFInfoTreeNode> > &typeMap,DWARFInfoDB *dwarfInfo,RCPointer<DWARFInfoTreeNode> insertionPoint);

        void PrintArchelonHeader(COFF_ASCNHDR &SectionHeader);
        void PrintSectionHeader(COFF_SCNHDR &sectionHeader);
        std::string GetTypeAsString(uint32_t type);
        void SetFramePointerID(uint32_t id){framePointerID=id;}
        uint32_t GetFramePointerID(){return framePointerID;}

        std::list<CoffFunctionInfo> &GetFunctionInfo(){return functionInfo;}

        void SetVerbose(bool verbose_){verbose=verbose_;}

        // list to store function definitions
        // ----------------------------------------
        // final list only consisting of complete functions
        std::list<CoffFunctionInfo> functionInfo;
        // used for nested loops
        std::list<CoffFunctionInfo> functionInfoStack;

        // list to store basic block definitions
        std::list<CoffBlockInfo> blockInfo;
        // intermediate data structure to put blocks on the stack
        std::list<CoffBlockInfo> blockInfoStack;

        std::list<CoffVariableInfo> globalList;
        // iterators to handle lists
        std::list<CoffFunctionInfo>::iterator functionInfoPos;
        std::list<CoffBlockInfo>::iterator blockInfoPos;

        uint32_t nestingLevel;
        // private members
        std::string filename;
        std::string lastKnownSymbol;
        bool verbose;
        std::list <SymbolEntry *> symEntries;
        std::map<uint32_t,RCPointer<DWARFInfoTreeNode> > treeNodeMap;
        uint32_t framePointerID;

    };






    // For debug purposes:
    std::ostream &operator << (std::ostream &ostr, const CoffDebugSymbols &debugSymbols);

    class CoffDebugSymbolsDB
    {
    public:
        CoffDebugSymbolsDB();
        ~CoffDebugSymbolsDB();

        // copy constructor
        CoffDebugSymbolsDB(const CoffDebugSymbolsDB &a);

        // assignment
        const CoffDebugSymbolsDB &operator=(const CoffDebugSymbolsDB& a);

        void LoadCoffDebugStrings(const uint8_t strings[], uint32_t size);
        void LoadCoffDebugSymbols(const uint8_t data[], uint32_t size,std::string &sectionFileName);
        void ProcessDebugInfo(bool bigendian,std::string &sectionFileName,DWARFInfoDB *dwarfInfo);
        std::list<filenamePos> &GetCoffFilesInSegment(){return coffFilesInSegment;}

        void Clear();
        void PrintRawData(const uint8_t data[], uint32_t size);
        std::list <CoffDebugSymbols *> debugSymbols;
        void SetVerbose(bool _verbose){verbose=_verbose;}
        void SetFramePointerID(uint32_t id){framePointerID=id;}
        uint32_t GetFramePointerID(){return framePointerID;}
        bool containsInfo;



    private:
        // for every '.file' entry we have exactly one 'CoffDebugSymbols' entry
        uint8_t *symInfo;
        uint8_t *stringInfo;
        uint32_t stringSize;
        uint32_t symSize;
        bool verbose;
        std::list<filenamePos> coffFilesInSegment;
        uint32_t framePointerID;

    };
}

#endif
