/**
 * \file    Framework/include/DebugDWARFCallFrameInfo.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup Debug Info Database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef DEBUGDWARFCALLFRAMEINFO_H
#define DEBUGDWARFCALLFRAMEINFO_H

#include <cstdint>

#include <string>
#include <vector>
#include "Dwarf2Constants.h"

namespace DebugInfoDB
{
    // common information entry
    struct DWARFCallFrameInfoCIE
    {
        std::string augmentation;
        uint32_t code_alignment_factor;
        int32_t data_alignment_factor;
        std::vector<uint8_t> initial_instructions; // DW_CFA_ opcodes
        uint8_t version;
        uint8_t return_address_register;
    };

    // frame description entry
    struct DWARFCallFrameInfoFDE
    {
        uint64_t initial_location;
        uint64_t end_location;  // this is initial_location + address_range - 1
        uint32_t CIE_pointer; // index into cieList array
        std::vector<uint8_t> instructions; // DW_CFA_ opcodes
    };

    // return value of GetRegisterRulesForPreviousFrame(): a rule how to get the previous value for a register
    struct DWARFCallFrameInfoRegisterRule
    {
        DWARFCallFrameInfoRegisterRule(uint32_t rule_ = DW_CFA_RR_unspecified, int32_t n_ = 0): rule(rule_), n(n_) {}
        uint32_t rule; // see DW_CFA_RR_ constants (undefined/same value/offset/register)
        int32_t n; // either a signed offset or a register number
    };


    // main DebugInfo class
    class DWARFCallFrameInfoDB
    {
    public:
        // public methods
        DWARFCallFrameInfoDB(): isBigEndian(false), addressTypeSize(4), cfaOffsetIsFactoredOffset(false), restoreStateAlsoRestoresCFA(false), adjustCfaOffset(false) {}
        void Clear();
        void LoadDWARF(const uint8_t data[], uint32_t size, bool isBigEndian, uint32_t addressTypeSize);
        void DumpLikeReadELF();
        // this is the main API function: call this repeatedly and execute the register rules on a virtual register set to get the call stack
        bool GetRegisterRulesForPreviousFrame(uint64_t pc, uint32_t &cfaRegisterNumber_out, int32_t &cfaOffset_out, uint32_t &returnAddressRegisterNumber_out, std::vector<DWARFCallFrameInfoRegisterRule> &registerRules_out);
        // get endianness
        bool IsBigEndian() const { return isBigEndian; }
        uint32_t GetAddressTypeSize() const { return addressTypeSize; }
        void SetFeatures(bool cfaOffsetIsFactoredOffset_, bool restoreStateAlsoRestoresCFA_, bool adjustCfaOffset_ = false)
        {
            cfaOffsetIsFactoredOffset = cfaOffsetIsFactoredOffset_;
            restoreStateAlsoRestoresCFA = restoreStateAlsoRestoresCFA_;
            adjustCfaOffset = adjustCfaOffset_;
        }

    private:
        // private helpers:
        void ExecuteCallFrameInstructions(const std::vector<uint8_t>& instructions, uint64_t startPC, uint64_t pc, uint32_t code_alignment_factor,
                                          int32_t data_alignment_factor, const std::vector<DWARFCallFrameInfoRegisterRule>& initialRules, bool print,
                                          uint32_t &cfaRegisterNumber_out, int32_t &cfaOffset_out, std::vector<DWARFCallFrameInfoRegisterRule> &registerRules_out);

        // private data
        std::vector<DWARFCallFrameInfoCIE> cieList;
        std::vector<DWARFCallFrameInfoFDE> fdeList;
        bool isBigEndian;
        uint32_t addressTypeSize;
        bool cfaOffsetIsFactoredOffset;
        bool restoreStateAlsoRestoresCFA;
        bool adjustCfaOffset;
    };

}

#endif
