/**
 * \file    Framework/include/DebugDWARFInfo.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup debug_info_database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef DEBUGDWARFINFO_H
#define DEBUGDWARFINFO_H

#include <cstdint>

#include <string>
#include <vector>
#include <set>
#include <map>

#include "Dwarf2Constants.h"
#include "ReferenceCounting.h"

namespace DebugInfoDB
{
    class DWARFInfoDB;
    struct DWARFInfoAttributeSettingContext;

    // simple byte block
    struct DWARFInfoByteBlock
    {
        DWARFInfoByteBlock(const uint8_t *block_, uint32_t len_): block(block_), len(len_) {}
        DWARFInfoByteBlock(): block(0), len(0) {}
        const uint8_t *block;
        uint32_t len;
        std::vector<uint8_t> vect() const { return std::vector<uint8_t>(block, block + len); }
    };


    // location list entry
    struct DWARFInfoLocationListEntry
    {
        uint64_t start;
        uint64_t end;
        uint64_t base;     // base address which needs to be added to start and end if relative == true
        bool relative;  // hint whether this are absolute or relative addresses (the documentation says clearly that these addresses are always relative to base, but many compilers just put absolute addresses in here)
        DWARFInfoByteBlock location; // location expression
    };


    // find location list entry for a specific PC
    // use heuristics to determine whether addresses are relative or absolute (try both)
    // returns 0 if no matching entry is found
    const DWARFInfoLocationListEntry *FindLocationListEntry(uint64_t pc, const std::vector<DWARFInfoLocationListEntry>& location);


    // ranges entry
    struct DWARFInfoRangesEntry
    {
        DWARFInfoRangesEntry(uint64_t start_ = 0, uint64_t end_ = 0, uint64_t base_ = 0, bool relative_ = false): start(start_), end(end_), base(base_), relative(relative_) {}
        uint64_t start;
        uint64_t end;
        uint64_t base;     // base address which needs to be added to start and end if relative == true
        bool relative;  // hint whether this are absolute or relative addresses (the documentation says clearly that these addresses are always relative to base, but many compilers just put absolute addresses in here)
    };

    // base class for all items in the debug info tree structure
    class DWARFInfoTreeNode: public RCObject
    {
    public:
        enum
        {
            DEFAULT_DW_AT_decl_file = 0xffffffff,
            DEFAULT_DW_AT_decl_line = 0xffffffff,
            DEFAULT_DW_AT_decl_column = 0xffff,
            DEFAULT_DW_AT_start_scope = 0xffffffff
        };

        // constructor and destructor
        DWARFInfoTreeNode():
            tag((DWARFTag)0),
            offsetInDebugInfoSection(0),
            locationListOffset(0),
            flags(0),
            rangesOffset(0),
            db(0),
            abbrevCode(0),
            name(""),
            byte_size(0),
            coffMode(false),
            addressTypeSize(4),
            visibility((DWARFVisibility)0),
            decl_file(DEFAULT_DW_AT_decl_file),
            decl_line(DEFAULT_DW_AT_decl_line),
            decl_column(DEFAULT_DW_AT_decl_column),
            accessibility((DWARFAccessibility)0),
            declaration(false),
            start_scope(DEFAULT_DW_AT_start_scope)
        {
            hasFlags[0] = hasFlags[1] = hasFlags[2] = 0;
        }

        virtual ~DWARFInfoTreeNode();

        // get children of this node
        const std::vector<RCPointer<DWARFInfoTreeNode> > &GetChildren() const; // logical constness
        // get children of this node, do not load compilation units
        // (never use this function, always use GetChildren())
        const std::vector<RCPointer<DWARFInfoTreeNode> > &GetChildrenButDoNotLoadCompilationUnits() const { return children; }
        // add child
        void AddChild(RCPointer<DWARFInfoTreeNode> child);
        // clear children
        void ClearChildren();

        // these functions need to be reimplemented in a derived class
        // (the functions in the derived class must not call the implementation of the base class)
        inline virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        enum { PRF_LIKEREADELF = 1 };
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        void PrintRecursive(uint32_t flags, int level, bool recursive = true) const;

        // check whether a given address is in 'ranges' (or low_pc/high_pc)
        // returns false if no address range specified in this object
        bool IsAddressInRanges(uint64_t address) const;
        // returns true if no address range specified in this object
        bool IsAddressInRangesOrNoRange(uint64_t address) const;
        // returns true is we have at least one address range
        bool HasAddressRange() const { return !GetRanges().empty(); }

        // set TAG attribute
        void SetTag(DWARFTag tag_){tag=tag_;}



        // check whether this node has a certain attribute: (only in this node)
        bool HasAttributeNonRecursive(DWARFAttribute attr) const
        {
            if (attr < 96)
            {
                return (hasFlags[attr >> 5] & (1 << (attr & 31))) != 0;
            }
            else if(attr == DW_AT_MIPS_linkage_name
                    || attr == DW_AT_linkage_name)
            {
                return (hasFlags[2] & (1 << 31)) != 0;
            }
            else
            {
                return false;
            }
        }

        // check whether this node has a certain attribute: (also in abtract_origin)
        // (abstract_origin is not followed recursively)
        bool HasAttribute(DWARFAttribute attr) const
        {
            if(HasAttributeNonRecursive(attr))
                return true;
            else if(abstract_origin)
                return abstract_origin->HasAttributeNonRecursive(attr);
            else
                return false;
        }

        // set a 'has' flag for a specific attribute
        void SetAttributeFlag(DWARFAttribute attr)
        {
            if(attr < 96)
                hasFlags[attr >> 5] |= (1 << (attr & 31));
            else if(attr == DW_AT_MIPS_linkage_name
                    || attr == DW_AT_linkage_name)
                hasFlags[2] |= (1 << 31);
        }


    private:
        // this private function is here so it can be inlined
        // return either this object or the abstract_origin if the given attribute is missing in this
        // (abstract_origin is not followed recursively)
        const DWARFInfoTreeNode *GetInstanceContainingAttribute(DWARFAttribute attr) const { if((!HasAttributeNonRecursive(attr)) && abstract_origin) return abstract_origin.GetPtr(); else return this; }
    public:
        // get minimum range start
        uint64_t GetMinimumRangeStart();

        // attribute read functions:
        // attributes present in this base class:
        virtual DWARFVisibility GetVisibility() const { return GetInstanceContainingAttribute(DW_AT_visibility)->visibility; }
        virtual const char *GetName() const { return GetInstanceContainingAttribute(DW_AT_name)->name; }
        virtual uint16_t GetDeclColumn() const { return GetInstanceContainingAttribute(DW_AT_decl_column)->decl_column; }
        virtual uint32_t GetDeclFile() const { return GetInstanceContainingAttribute(DW_AT_decl_file)->decl_file; }
        virtual uint32_t GetDeclLine() const { return GetInstanceContainingAttribute(DW_AT_decl_line)->decl_line; }
        virtual DWARFAccessibility GetAccessibility() const { return GetInstanceContainingAttribute(DW_AT_accessibility)->accessibility; }
        virtual bool GetDeclaration() const { return GetInstanceContainingAttribute(DW_AT_declaration)->declaration; }
        virtual const DWARFInfoTreeNode *GetAbstractOrigin() const { return GetInstanceContainingAttribute(DW_AT_abstract_origin)->abstract_origin.GetPtr(); }
        virtual const DWARFInfoTreeNode *GetType() const { return GetInstanceContainingAttribute(DW_AT_type)->type.GetPtr(); }
        virtual uint32_t GetByteSize() const { return GetInstanceContainingAttribute(DW_AT_byte_size)->byte_size; }
        virtual const std::vector<DWARFInfoRangesEntry>& GetRanges() const { if(ranges.empty() && abstract_origin) return abstract_origin->ranges; else return ranges; }
        virtual std::vector<DWARFInfoRangesEntry>& GetRanges() { if(ranges.empty() && abstract_origin) return abstract_origin->ranges; else return ranges; }
        // for GetLowPc() please use GetRanges
        // for GetHighPc() please use GetRanges
        virtual uint32_t GetStartScope() const { return GetInstanceContainingAttribute(DW_AT_start_scope)->start_scope; }

        // attributes present in derived classes (just return a default value here)
        virtual const std::vector<DWARFInfoLocationListEntry>& GetLocation() const { static const std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual std::vector<DWARFInfoLocationListEntry>& GetLocationList()    { static std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual std::vector<DWARFInfoLocationListEntry>& GetFrameBaseList()   { static std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual const std::vector<DWARFInfoLocationListEntry>& GetSegment() const { static const std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual uint64_t GetConstValue() const { return 0; }
        virtual const char *GetConstStringValue() const { return 0; }
        virtual bool GetExternal() const { return false; }
        virtual bool GetArtificial() const { return false; }
        virtual const char *GetLinkageName() const { return ""; }
        virtual int32_t GetBitOffset() const { return 0; }
        virtual uint32_t GetBitSize() const { return 0; }
        virtual int32_t GetStmtList() const { return 0; }
        virtual uint32_t GetLanguage() const { return 0; }
        virtual const char *GetCompDir() const { return ""; }
        virtual RCPointer<DWARFInfoTreeNode> GetContainingType() const { return 0; }
        virtual uint32_t GetUpperBound() const { return 0; }
        virtual DWARFAttrEncoding GetEncoding() const { return DWARFAttrEncoding(0); }
        virtual const std::vector<DWARFInfoLocationListEntry>& GetDataMemberLocation() const { static const std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual RCPointer<DWARFInfoTreeNode> GetSpecification() const { return 0; }
        virtual uint16_t GetInline() const { return 0; }
        virtual const char *GetProducer() const { return ""; }
        virtual bool GetPrototyped() const { return 0; }
        virtual int32_t GetMacroInfo() const { return 0; }
        virtual const std::vector<DWARFInfoLocationListEntry>& GetFrameBase() const { static const std::vector<DWARFInfoLocationListEntry> l; return l; }
        virtual DWARFVirtuality GetVirtuality() const { return DWARFVirtuality(0); }
        virtual uint32_t GetStrideSize() const { return 0; }
        virtual void SetRangesList( std::vector<DWARFInfoRangesEntry> ranges_){ranges=ranges_;}

#if 0
        virtual uint32_t GetOrdering() const { return 0; }
        virtual uint32_t GetDiscr() const { return 0; }
        virtual uint32_t GetDiscrValue() const { return 0; }
        virtual uint32_t GetImport() const { return 0; }
        virtual uint32_t GetStringLength() const { return 0; }
        virtual uint32_t GetCommonReference() const { return 0; }
        virtual uint32_t GetDefaultValue() const { return 0; }
        virtual uint32_t GetIsOptional() const { return 0; }
        virtual uint32_t GetLowerBound() const { return 0; }
        virtual uint32_t GetReturnAddr() const { return 0; }
        virtual uint32_t GetAddressClass() const { return 0; }
        virtual uint32_t GetBaseTypes() const { return 0; }
        virtual uint32_t GetCallingConvention() const { return 0; }
        virtual uint32_t GetCount() const { return 0; }
        virtual uint32_t GetDiscrList() const { return 0; }
        virtual uint32_t GetFriend() const { return 0; }
        virtual uint32_t GetIdentifierCase() const { return 0; }
        virtual uint32_t GetNamelistItem() const { return 0; }
        virtual uint32_t GetPriority() const { return 0; }
        virtual uint32_t GetStaticLink() const { return 0; }
        virtual uint32_t GetUseLocation() const { return 0; }
        virtual uint32_t GetVariableParameter() const { return 0; }
        virtual uint32_t GetVtableElemLocation() const { return 0; }
        virtual uint32_t GetAssociated() const { return 0; }
        virtual uint32_t GetDataLocation() const { return 0; }
        virtual uint32_t GetStride() const { return 0; }
        virtual uint32_t GetEntryPc() const { return 0; }
        virtual uint32_t GetUseUtf8() const { return 0; }
        virtual uint32_t GetExtension() const { return 0; }
        virtual uint32_t GetTrampoline() const { return 0; }
        virtual uint32_t GetCallColumn() const { return 0; }
        virtual uint32_t GetCallFile() const { return 0; }
        virtual uint32_t GetCallLine() const { return 0; }
        virtual uint32_t GetDescription() const { return 0; }
#endif

    protected:
        // helpers used by SetAttribute()
        inline const char * ReadStringValue   (DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        inline uint64_t        ReadIntValue      (DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        inline DWARFInfoByteBlock ReadBlockValue(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        inline void         ReadReferenceValue(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context, RCPointer<DWARFInfoTreeNode> *referenceToResolve);
        // read a location expression or a location list
        // a simple location expression (provided as form=block) will result in a location list with a single entry having the start and end addresses set to 0..MAXADDR
        inline std::vector<DWARFInfoLocationListEntry> ReadLocationValue(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        inline std::vector<DWARFInfoRangesEntry> ReadRangesValue(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        // printing helpers
        std::string PrintReference(const RCPointer<DWARFInfoTreeNode>& ref) const;
        void PrintLocation(const std::string& prefix, const std::string& attr, const std::vector<DWARFInfoLocationListEntry>& locationList) const;

    public:
        // public data
        DWARFTag tag;                                      // one of the DW_TAG_* constants
        RCPointer<DWARFInfoTreeNode> parent;            // pointer to parent node (NULL means: this is the root node)
        uint32_t offsetInDebugInfoSection;                 // offset of this node in debug_info section, for printout and references
        uint32_t locationListOffset;                       // offset in debug_loc section, only used for readelf printout
        uint32_t flags;                                    // misc flags:
        enum { REFERENCE_IS_ABSOLUTE = (1 << 0), CHILDREN_NOT_FULLY_LOADED = (1 << 1) };
        uint32_t rangesOffset;                             // offset in debug_ranges section, only used for readelf printout
        DWARFInfoDB *db;                                // pointer to main debug info class
        uint32_t abbrevCode;                               // only used for readelf printout
        const char *name;                               // value of DW_AT_name attribute (if present)
        RCPointer<DWARFInfoTreeNode> type;              // value of DW_AT_type attribute (if present)
        uint32_t byte_size;                                // value of DW_AT_byte_size attribute (if present)
        bool  coffMode;                                 // 'true', if DWARF structures were generated from COFF origin
        uint32_t addressTypeSize;                          // size of addresses, pointers

    private:
        // private data
        uint32_t hasFlags[3];                              // flags which indicate which attributes are present
        std::vector<RCPointer<DWARFInfoTreeNode> > children; // children


    protected:
        // some attributes are in the base class since they are common to a lot of debug entry types
        DWARFVisibility visibility;                        // value of DW_AT_visibility attribute (if present)
        uint32_t decl_file;                                // value of DW_AT_decl_file attribute (if present)
        uint32_t decl_line;                                // value of DW_AT_decl_line attribute (if present)
        uint16_t decl_column;                              // value of DW_AT_decl_column attribute (if present)
        DWARFAccessibility accessibility;                  // value of DW_AT_accessibility attribute (if present)
        bool declaration:1;                             // value of DW_AT_declaration attribute (if present)
        RCPointer<DWARFInfoTreeNode> abstract_origin;   // value of DW_AT_abstract_origin attribute (if present)
        std::vector<DWARFInfoRangesEntry> ranges;            // value of DW_AT_ranges attribute (if present), or of low_pc/high_pc attribute
        uint32_t start_scope;                              // value of DW_AT_start_scope attribute (if present)
    };


    // unresolved references during parsing of the .debug_info section
    struct DWARFInfoUnresolvedReference
    {
        RCPointer<DWARFInfoTreeNode> *referenceToResolve;
        uint32_t targetOffsetInDebugInfoSection;
    };


    // attribute setting context
    struct DWARFInfoAttributeSettingContext
    {
        std::vector<DWARFInfoUnresolvedReference> unresolvedReferences;
        const uint8_t *compilation_unit_end;
        uint32_t offsetOfCompilationUnitInDebugInfoSection;
        uint64_t baseAddressOfCompilationUnit; // base address of the current compilation unit
        uint32_t version; // DWARF version (2 or 3 or so)
    };

    // simplify the declaration of member functions which automatically access the abstract_origin if the attribute is missing in the current instance
#define GetAttributeAbstractOrigin(cattr,attr,class,rtype) \
        rtype Get##cattr() const { if((!HasAttributeNonRecursive(DW_AT_##attr)) && abstract_origin && dynamic_cast<const TreeNode_DW_TAG_##class*>(abstract_origin.GetPtr())) \
                return reinterpret_cast<const TreeNode_DW_TAG_##class*>(abstract_origin.GetPtr())->attr; else return attr; }

    // DW_TAG_compile_unit
    class TreeNode_DW_TAG_compile_unit: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_compile_unit()
            : comp_dir("")
            , producer("")
            , language(DWARFSourceLanguage(0))
            , macro_info(-1)
            , stmt_list(-1)
            , compilation_unit_start(0)
            , compilation_unit_end(0)
        {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        const char *GetCompDir() const { return comp_dir; }
        const char *GetProducer() const { return producer; }
        uint32_t GetLanguage() const { return language; }
        int32_t GetMacroInfo() const { return macro_info; }
        int32_t GetStmtList() const { return stmt_list; }

        // get/set the offset/size of the outer compilation unit (not compile_unit node)
        uint32_t GetCompilationUnitStart() const { return compilation_unit_start; }
        uint32_t GetCompilationUnitEnd  () const { return compilation_unit_end  ; }
        void  SetCompilationUnitStart(uint32_t start) { compilation_unit_start = start; }
        void  SetCompilationUnitEnd  (uint32_t end  ) { compilation_unit_end   = end  ; }

    private:
        // private data
        const char *comp_dir;
        const char *producer;
        DWARFSourceLanguage language;
        int32_t macro_info;
        int32_t stmt_list;

        uint32_t compilation_unit_start;
        uint32_t compilation_unit_end;
    };

    // DW_TAG_partial_unit
    class TreeNode_DW_TAG_partial_unit :
        public TreeNode_DW_TAG_compile_unit
    {
    public:
        TreeNode_DW_TAG_partial_unit() :
            TreeNode_DW_TAG_compile_unit()
        {
        }
    };

    // DW_TAG_base_type
    class TreeNode_DW_TAG_base_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_base_type()
            : bit_size(0)
            , bit_offset(-1)
            , encoding(DWARFAttrEncoding(0))
        {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        int32_t GetBitOffset() const { return bit_offset; }
        void  SetBitOffset(int32_t bit_offset_){bit_offset=bit_offset_; }
        uint32_t GetBitSize() const { return bit_size; }
//      void  SetBitSize() (uint32_t bit_size_){bit_size=bit_size_;}
        DWARFAttrEncoding GetEncoding() const { return encoding; }
        uint32_t bit_size;
        void  SetEncoding(DWARFAttrEncoding encoding_)  { encoding=encoding_;}

    private:
        // private data
        int32_t bit_offset;
        DWARFAttrEncoding encoding;
    };

    // DW_TAG_enumeration_type
    class TreeNode_DW_TAG_enumeration_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_enumeration_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_enumerator
    class TreeNode_DW_TAG_enumerator: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_enumerator(): const_value(0) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        uint64_t GetConstValue() const { return const_value; }

    private:
        // private data
        uint64_t const_value;
    };

    // DW_TAG_typedef
    class TreeNode_DW_TAG_typedef: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_typedef() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_array_type
    class TreeNode_DW_TAG_array_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_array_type(): stride_size(0) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(StrideSize,stride_size,array_type,uint32_t);

        virtual uint32_t GetByteSize() const;

    private:
        // private data
        uint32_t stride_size;
    };

    // DW_TAG_subrange_type
    class TreeNode_DW_TAG_subrange_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_subrange_type(): upper_bound(0) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(UpperBound,upper_bound,subrange_type,uint32_t);

    private:
        // private data
        uint32_t upper_bound;
    };

    // DW_TAG_structure_type
    class TreeNode_DW_TAG_structure_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_structure_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(ContainingType,containing_type,structure_type,RCPointer<DWARFInfoTreeNode>);
        GetAttributeAbstractOrigin(Specification,specification,structure_type,RCPointer<DWARFInfoTreeNode>);

    private:
        // private data
        RCPointer<DWARFInfoTreeNode> containing_type;
        RCPointer<DWARFInfoTreeNode> specification;
    };

    // DW_TAG_union_type
    class TreeNode_DW_TAG_union_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_union_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(ContainingType,containing_type,union_type,RCPointer<DWARFInfoTreeNode>);
        GetAttributeAbstractOrigin(Specification,specification,union_type,RCPointer<DWARFInfoTreeNode>);

    private:
        // private data
        RCPointer<DWARFInfoTreeNode> containing_type;
        RCPointer<DWARFInfoTreeNode> specification;
    };

    // DW_TAG_pointer_type
    class TreeNode_DW_TAG_pointer_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_pointer_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_const_type
    class TreeNode_DW_TAG_const_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_const_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_reference_type
    class TreeNode_DW_TAG_reference_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_reference_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_volatile_type
    class TreeNode_DW_TAG_volatile_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_volatile_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_unspecified_parameters
    class TreeNode_DW_TAG_unspecified_parameters: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_unspecified_parameters() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_label
    class TreeNode_DW_TAG_label: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_label() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };

    // DW_TAG_member
    class TreeNode_DW_TAG_member: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_member(): bit_offset(-1), bit_size(0), artificial(false) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        const std::vector<DWARFInfoLocationListEntry>& GetDataMemberLocation() const { return data_member_location; }
        int32_t GetBitOffset() const { return bit_offset; }
        uint32_t GetBitSize() const { return bit_size; }
        bool GetArtificial() const { return artificial; }

    private:
        // private data
        std::vector<DWARFInfoLocationListEntry> data_member_location;
        int32_t bit_offset;
        uint32_t bit_size;
        bool artificial;
    };

    // DW_TAG_inheritance
    class TreeNode_DW_TAG_inheritance: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_inheritance()
            : virtuality(DWARFVirtuality(0))
        {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        const std::vector<DWARFInfoLocationListEntry>& GetDataMemberLocation() const { return data_member_location; }
        DWARFVirtuality GetVirtuality() const { return virtuality; }

    private:
        // private data
        std::vector<DWARFInfoLocationListEntry> data_member_location;
        DWARFVirtuality virtuality;
    };

    // DW_TAG_variable
    class TreeNode_DW_TAG_variable: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_variable() :
            linkage_name(""),
            const_value(0),
            const_string_value(0),
            external(false),
            artificial(false)
        {}

        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(Location,location,variable,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(Segment,segment,variable,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(LinkageName,linkage_name,variable,const char *);
        GetAttributeAbstractOrigin(ConstValue,const_value,variable,uint64_t);
        GetAttributeAbstractOrigin(External,external,variable,bool);
        GetAttributeAbstractOrigin(Artificial,artificial,variable,bool);
        virtual std::vector<DWARFInfoLocationListEntry>& GetLocationList() {return location; }
        virtual void SetLocationList(std::vector<DWARFInfoLocationListEntry> list){location=list;}
        virtual const char *GetConstStringValue() const { return const_string_value; }

    private:
        // private data
        std::vector<DWARFInfoLocationListEntry> location;
        std::vector<DWARFInfoLocationListEntry> segment;
        const char *linkage_name;
        uint64_t const_value;
        const char *const_string_value;
        bool external:1;
        bool artificial:1;
    };

    // DW_TAG_formal_parameter
    class TreeNode_DW_TAG_formal_parameter: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_formal_parameter():artificial(false) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(Location,location,formal_parameter,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(Segment,segment,formal_parameter,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(Artificial,artificial,formal_parameter,bool);

    private:
        // private data
        std::vector<DWARFInfoLocationListEntry> location;
        std::vector<DWARFInfoLocationListEntry> segment;
        bool artificial:1;
    };

    // DW_TAG_subprogram
    class TreeNode_DW_TAG_subprogram: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_subprogram()
            : linkage_name("")
            , inlined(DWARFInline(0))
            , external(false)
            , prototyped(false)
            , artificial(false)
            , virtuality(DWARFVirtuality(0))
        {
        }
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(FrameBase,frame_base,subprogram,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(Artificial,artificial,subprogram,bool);
        GetAttributeAbstractOrigin(External,external,subprogram,bool);
        GetAttributeAbstractOrigin(Prototyped,prototyped,subprogram,bool);
        GetAttributeAbstractOrigin(LinkageName,linkage_name,subprogram,const char *);
        #define inline inlined
        GetAttributeAbstractOrigin(Inline,inline,subprogram,uint16_t);
        #undef inline
        GetAttributeAbstractOrigin(Virtuality,virtuality,subprogram,DWARFVirtuality);
        GetAttributeAbstractOrigin(VtableElemLocation,vtable_elem_location,subprogram,const std::vector<DWARFInfoLocationListEntry>&);
        GetAttributeAbstractOrigin(ContainingType,containing_type,subprogram,RCPointer<DWARFInfoTreeNode>);
        GetAttributeAbstractOrigin(Specification,specification,subprogram,RCPointer<DWARFInfoTreeNode>);

        std::vector<DWARFInfoLocationListEntry> &GetFrameBaseList(){return frame_base;}
        void SetFrameBaseList(std::vector<DWARFInfoLocationListEntry> list_){frame_base=list_;}
    private:
        // private data
        std::vector<DWARFInfoLocationListEntry> frame_base;
        const char *linkage_name;
        DWARFInline inlined;
        bool external:1;
        bool prototyped:1;
        bool artificial:1;
        DWARFVirtuality virtuality;
        std::vector<DWARFInfoLocationListEntry> vtable_elem_location;
        RCPointer<DWARFInfoTreeNode> containing_type;
        RCPointer<DWARFInfoTreeNode> specification;
    };

    // DW_TAG_subroutine_type
    class TreeNode_DW_TAG_subroutine_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_subroutine_type():prototyped(false) {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;

        GetAttributeAbstractOrigin(Prototyped,prototyped,subroutine_type,bool);

    private:
        // private data
        bool prototyped:1;
    };

    // DW_TAG_lexical_block
    class TreeNode_DW_TAG_lexical_block: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_lexical_block() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // DW_TAG_inlined_subroutine
    class TreeNode_DW_TAG_inlined_subroutine: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_inlined_subroutine() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // DW_TAG_namespace
    class TreeNode_DW_TAG_namespace: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_namespace() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // DW_TAG_class_type
    class TreeNode_DW_TAG_class_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_class_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // DW_TAG_template_type_param
    class TreeNode_DW_TAG_template_type_param: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_template_type_param() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // DW_TAG_ptr_to_member_type
    class TreeNode_DW_TAG_ptr_to_member_type: public DWARFInfoTreeNode
    {
    public:
        TreeNode_DW_TAG_ptr_to_member_type() {}
        virtual void SetAttribute(DWARFAttribute attribute, DWARFFormat form, const uint8_t *&debug_info, DWARFInfoAttributeSettingContext *context);
        virtual void PrintTAGSpecificAttributes(uint32_t flags, const std::string& prefix) const;
    };


    // main DebugInfo class
    class DWARFInfoDB
    {
    public:
        // public methods
        DWARFInfoDB();
        ~DWARFInfoDB();
        void Clear();
        void LoadDWARF(const uint8_t debug_info[], uint32_t debug_info_size,
                       const uint8_t debug_abbrev[], uint32_t debug_abbrev_size,
                       const uint8_t debug_loc[], uint32_t debug_loc_size,
                       const uint8_t debug_str[], uint32_t debug_str_size,
                       const uint8_t debug_ranges[], uint32_t debug_ranges_size,
                       bool isBigEndian_, uint32_t addressTypeSize_, bool onlyCompilationUnits = false);
        void PrintRecursive(uint32_t flags);
        void SetVerbose(bool verbose);
        void SetQuiet(bool quiet);
        // main API function
        DWARFInfoTreeNode *GetRoot() { return &root; }
        const DWARFInfoTreeNode *GetRoot() const { return &root; }
        // get endianness and address type size
        bool IsBigEndian() const { return isBigEndian; }
        uint32_t GetAddressTypeSize() const { return addressTypeSize; }
        void DumpDebugLoc();

        // convenience function which returns the 'producer' field of the first compilation unit found (or empty string if no compilation unit contains a non-empty producer entry)
        // (parameter should never be specified)
        const char *GetProducer(const DWARFInfoTreeNode *node = 0) const;
        // get a list of all procducers separated by linefeeds, format 'Producer: $producer\n'
        // (reset by LoadDWARF())
        const std::string& GetProducers() const { return producers; }
        // get human readable info about loaded info (including producers)
        // (reset by LoadDWARF())
        std::string GetHRInfo() const;
        // get number of compilation units currently being loaded (they are loaded on demand)
        uint32_t GetNumberOfCompilationUnitsLoaded() const { return numCompilationUnitsLoaded; }
        // get number of total compilation units
        uint32_t GetNumberOfCompilationUnitsTotal() const { return numCompilationUnitsTotal; }
        // force load all compilation units
        // (this may take very long for large amounts of debug info/1 minuet for 45MB of data, data will also be expanded by a factor of 15!)
        // (never use this function)
        void ForceLoadAllCompilationUnits();
        // get all (top level) compilation units
        const std::vector<RCPointer<DWARFInfoTreeNode> >& GetCompilationUnits();

        // get the value of an enum symbol regardless of scope
        // Note: We do not want to trigger loading of all compilation units if there are so many that this would take
        // a minute to load, so we only look up the symbols in the units we already have loaded anyway. This means that
        // this function behaves differently even when called with the same parameters, depending on what other functions
        // have already been called on this db. The effect is that this works well for requests which are to some degree
        // local enough to the modules which are currectly being debugged.
        bool GetEnumSymbolValue(const char *name, uint64_t &uval, int64_t &sval, const DWARFInfoTreeNode *node = 0) const;
        // get the outermost subprogram which contains pc
        const DWARFInfoTreeNode *FindOuterMostSubprogram(uint64_t pc, const DWARFInfoTreeNode *node = 0) const;
        // get the innermost subprogram which contains pc
        const DWARFInfoTreeNode *FindInnerMostSubprogram(uint64_t pc, const DWARFInfoTreeNode *node = 0) const;
        // create a new debug info node (also used by DebugCoffSymbol)
        RCPointer<DWARFInfoTreeNode> CreateTreeNodeForTAG(DWARFTag tag, uint32_t offsetInDebugInfo = 0xffffffff);


    private:
        // private data
        DWARFInfoTreeNode root;
        uint32_t addressTypeSize;
        bool isBigEndian:1;
        bool verbose:1;
        bool quiet:1;
        uint32_t version;
        std::string producers; // reset by LoadDWARF()
        std::string hrInfo; // reset by LoadDWARF()

        // global context
        const uint8_t *data_debug_info;
        const uint8_t *data_debug_info_end;
        uint32_t       data_debug_info_size;
        const uint8_t *data_debug_abbrev;
        const uint8_t *data_debug_abbrev_end;
        uint32_t       data_debug_abbrev_size;
        const uint8_t *data_debug_str;
        const uint8_t *data_debug_str_end;
        const uint8_t *data_debug_loc;
        const uint8_t *data_debug_loc_end;
        uint32_t       data_debug_loc_size;
        const uint8_t *data_debug_ranges;
        const uint8_t *data_debug_ranges_end;
        // current abbreviation table
        uint32_t abbrevCode2IdxLimit;
        uint32_t *abbrevCode2Idx;
        uint32_t abbrevTableLimit;
        uint32_t abbrevTableSize;
        uint32_t *abbrevTable;
        uint32_t abbrevTableLastBase;
        uint32_t numErroneousLocalVarsRelocated;
        std::set<std::string> producerSet;
        uint32_t numCompilationUnitsLoaded;
        uint32_t numCompilationUnitsTotal;
        uint32_t numTreeNodes;
        std::map<uint32_t,DWARFInfoTreeNode *> offset2compileUnitNode; // offset is the byte offset of the inner compile_unit node in the .debug_info section

        // private offset2node cache (used to resolve unresolved references
        class Offset2NodeCache
        {
        public:
            Offset2NodeCache(DWARFInfoDB *db_): compileUnit(0), compileUnitNextIndex(0), offset2node(0), db(db_) {}
            ~Offset2NodeCache() { delete[] offset2node; }
            DWARFInfoTreeNode *Lookup(uint32_t offsetInDebugInfoSection);

        private:
            void InitRecursive(DWARFInfoTreeNode *compileUnit, uint32_t compilationUnitStart);
            TreeNode_DW_TAG_compile_unit *compileUnit;
            uint32_t compileUnitNextIndex;
            DWARFInfoTreeNode **offset2node;
            DWARFInfoDB *db;
        };

        // private helpers:
        inline const uint8_t *copyMemAndAddZeroBytes(const uint8_t *debug_sect, uint32_t size, const char *name, const uint8_t *& endOut);
        void ClearDataPointers();
        void LoadCompilationUnit(uint32_t compilationUnitStart, bool onlyCompileUnits);

        // friend classes
        friend class DWARFInfoTreeNode;
    };

}

#endif
