/**
 * \file    Framework/include/DebugFunctionInfo.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup debug_info_database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef DEBUGFUNCTIONINFO_H
#define DEBUGFUNCTIONINFO_H

#include <cstdint>

#include <vector>
#include <string>

namespace DebugInfoDB
{
    class FunctionInfoDB;

    class FunctionInfo
    {
    public:
        FunctionInfo(const std::string &name_, uint64_t lowPC_, uint64_t highPC_, bool inlined_=false);

        const std::string GetName() const { return name; }
        uint64_t GetLowPC() const { return lowPC; }
        uint64_t GetHighPC() const { return highPC; }
        uint64_t GetSize() const { return highPC - lowPC; }
        bool InRange(uint64_t pc) const { return (lowPC <= pc) && (pc < highPC); }
        bool IsInlined() const { return inlined; }
        bool IsThumb () const {return isThumb;}

        // adjust ordering to make enclosed ranges 'greater', ensuring they appear after the enclosers
        inline bool operator < (const FunctionInfo &b) const
        {
            if (lowPC != b.lowPC)
            {
                return lowPC < b.lowPC;
            }
            else
            {
                // enclosing range < enclosed range
                return b.highPC < highPC;
            }
        }

    protected:
        std::string name;
        uint64_t lowPC, highPC;
        bool inlined;
        bool isThumb;

        // used to create the enclosing tree structure. unsafe after sortings.
        FunctionInfo * parent;

        friend class FunctionInfoDB;
        // for a given lowPC, with highPC == lowPC, this is the 'greatest' range
        // thus upper bound will always return ranges with lowPC > this.lowPC
        FunctionInfo(uint64_t lowPC_) : lowPC(lowPC_), highPC (lowPC_) {}
    };

    class FunctionInfoDB
    {
    public:
        enum { INVALID_ID = 0xffffffff };

        FunctionInfoDB();

        void Clear();

        unsigned int GetNumberOfFunctions() const;

        std::vector<std::string> GetAllFunctionNames() const;

        // get function information for an Id (ids go from 0 to N-1, where N == GetNumberOfFunctions())
        const FunctionInfo *GetFunctionInfoForId(size_t id) const;

        // get function id for a function name (yet slow)
        size_t GetFunctionInfoIdForName(const std::string& name) const;

        // Tries to find a function which contains a certain PC, else return NULL -- fast !
        inline const FunctionInfo *
        FindFunctionInfo(uint64_t pc) const
        {
            if (lastFunctionInfoFound &&
                lastFunctionInfoFound->InRange(pc))
            {
                return lastFunctionInfoFound;
            }
            else
            {
                return lastFunctionInfoFound = FindFunctionInfoInternal(pc);
            }
        }

        // same as above but return the id of the function info
        // may return INVALID_ID if pc is not in any function
        inline size_t FindFunctionInfoId(uint64_t pc) const
        {
            const FunctionInfo *fi = FindFunctionInfo(pc);
            if(fi)
                return fi - &(db[0]);
            else
                return INVALID_ID;
        }

        // add a function
        void AddFunctionInfo(const FunctionInfo &func);

        // add a section (sections are only used by GuessSizeForZeroSizeFunctions())
        void AddSectionInfo(const FunctionInfo &func);

                // set this if the core has Thumb mode
                // it controls the low bit of the start address, which is set to 1 for Thumb functions
                void AdjustThumbFunctions ();

                // set this if you want to allow overlapped functions, default false
                // it may happen when LOCAL symbols are loaded
                void AllowOverlappedFunctions (bool allow);

        // call this to guess sizes for all functions which have a zero size
        void GuessSizeForZeroSizeFunctions ();

                // local symbols may show up inside already existing function ranges (enclosed)
                // two strategies to deal with this problem:

                // a. speedup the search in presence of enclosed functions by using the holes search structure
                void PrepareHoles (int stepsBeforeSearchingHoles);
                // computes fully qualified name for enclosed functions for better display
                void ComputeFullEnclosedNames ();

                // b. completely eliminate functions appearing within other functions (enclosed)
                // this permits using the simple/fast search algorithm
                // can be used prior to guessing the size for zero size functions (only outer functions this time)
                void EliminateEnclosedFunctions ();

        private:
                // computes parent pointers according to enclosusre structure
                // assumes parents show up before childrent in dbT
                // parents pointers are invalidated by (during) sorting
                void ComputeParents () const;

    protected:
        typedef std::vector<FunctionInfo> dbT;
        dbT db; // this is the main database
        dbT sections; // this is only used by GuessSizeForZeroSizeFunctions()
        bool unclean; // db and sections may only be accessed if this is false, if its true Update() will set it to false
                // keeps track of holes between functions. speed up FindFunctionInfo when function is unknown
                // needed only when GuessSizeForZeroSizeFunctions is used.
                dbT holesdb;
                // if we GuessSizeForZeroSizeFunctions without EliminateEnclosedFunctions -> we need a moire complex search algorithm
                bool overlappingAllowed;
                // if >=1 and disjunct ranges then holes will not be searched. tunes up the hole searching process and offers a clean fallback to original range search algorithm
                // if -1, no holes machinery is used
                int stepsBeforeSearchingHoles;
                bool eliminatedEnclosedFunctions;

        mutable const FunctionInfo *lastFunctionInfoFound;

        const FunctionInfo *FindFunctionInfoInternal(uint64_t pc) const;

        // mark containers as 'unclean'
        void SetUnclean();

        // this must be called internally before 'db' is used
        // if unclean is true sort functions and then set unclean to false
        void Update() const;

#ifdef UNIT_TEST
public:
                void Prepare () const {Update ();}
                const vector <FunctionInfo> & holes () {return holesdb;}
#endif
    };
}

#endif
