/**
 * \file    Framework/include/DebugInfoDB.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup debug_info_database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef DEBUGINFODB_H
#define DEBUGINFODB_H

#include <cstdint>

#include "DebugLineInfo.h"
#include "DebugFunctionInfo.h"
#include "DebugCoffSymbols.h"
#include "DebugDWARFCallFrameInfo.h"
#include "DebugDWARFInfo.h"


#define DEBUGINFODB_VERSION 3

namespace DebugInfoDB
{
    class DebugInfoRoot
    {
    public:
        // NOTE: Never ever reorder the first two virtual functions in the declaration!!!
        virtual ~DebugInfoRoot();
        // NOTE: Never ever reorder the first two virtual functions in the declaration!!!
        virtual uint32_t GetVersion() const { return DEBUGINFODB_VERSION; }
        // NOTE: Never ever reorder the first two virtual functions in the declaration!!!

        // default constructor, costructs an object containing all kinds of debug info (none of the pointers will be NULL)
        DebugInfoRoot();

        // constructor, which constructs an object points to foreign parts of the debug info
        // (this object will not allocate/free anything passed to this constructor)
        DebugInfoRoot(LineInfoDB *lineInfo_, CoffDebugSymbolsDB *coffDebugSymbols = 0,FunctionInfoDB *functionInfo_ = 0, DWARFCallFrameInfoDB *dwarfCallFrameInfo_ = 0, DWARFInfoDB *dwarfInfo_ = 0);

        // public methods
        void Clear();

        // access functions:

        // get line info
        LineInfoDB *GetLineInfo() { return lineInfo; }
        const LineInfoDB *GetLineInfo() const { return lineInfo; }


        // get COFF Debug Symbols
        CoffDebugSymbolsDB *GetCoffDebugSymbols() { return coffDebugSymbols; }
        void SetCoffDebugSymbols(CoffDebugSymbolsDB *symbols){coffDebugSymbols = symbols;}
//      const CoffDebugSymbolsDB *GetCoffDebugSymbols() {return coffDebugSymbols ;}

        // get function info
        FunctionInfoDB *GetFunctionInfo() { return functionInfo; }
        const FunctionInfoDB *GetFunctionInfo() const { return functionInfo; }

        // get DWARF call frame info
        DWARFCallFrameInfoDB *GetDWARFCallFrameInfo() { return dwarfCallFrameInfo; }
        const DWARFCallFrameInfoDB *GetDWARFCallFrameInfo() const { return dwarfCallFrameInfo; }

        // get DWARF debug info
        DWARFInfoDB *GetDWARFInfo() { return dwarfInfo; }
        const DWARFInfoDB *GetDWARFInfo() const { return dwarfInfo; }

        void SetDWARFInfo(DWARFInfoDB *dwarfInfo_){dwarfInfo=dwarfInfo_;}
        void SetFramePointerID(uint32_t id){framePointerID=id;}
        uint32_t GetFramePointerID(){return framePointerID;}
        // convenience access functions:

        // get function name for address
        // (uses various sources of information to locate the function:
        //  1. FunctionInfoDB
        //  2. Symbol Table)
        // return empty string if no function found
        const std::string GetFunctionName(uint64_t address) const;

        // put functions in debug info into the FunctionInfoDB
        void PutDebugInfoFunctionsIntoFunctionInfoDB(const DWARFInfoTreeNode *node = 0);

        void SetQuiet(bool quiet);

        // get/set ELF characteristics
#define EM_MACHINE_TYPE_NOT_SET 0xffffffff /* this is an invalid ELF machine type since the ELF machine type is 16 bit */
        uint32_t GetELFMachineType() const { return elfMachineType; }
        void SetELFMachineType(uint16_t elfMachineType_) { elfMachineType = elfMachineType_; }
        bool HasELFEndianness() const { return hasELFEndianness; }
        bool IsELFEndiannessBigEndian() { return elfEndiannessBigEndian; }
        void SetELFEndiannessBigEndian(bool elfEndiannessBigEndian_) { elfEndiannessBigEndian = elfEndiannessBigEndian_; hasELFEndianness = true; }
        uint64_t GetApplicationEntryPoint() { return applicationEntryPoint; }
        void SetApplicationEntryPoint (uint64_t entryPoint) { applicationEntryPoint = entryPoint; }

    private:
        // data containers
        LineInfoDB *lineInfo;
        CoffDebugSymbolsDB *coffDebugSymbols;
        FunctionInfoDB *functionInfo;
        DWARFCallFrameInfoDB *dwarfCallFrameInfo;
        DWARFInfoDB *dwarfInfo;

        // these pointers are used to delete debug info allocated by this object
        LineInfoDB *lineInfoToDelete;
        CoffDebugSymbolsDB *coffDebugSymbolsToDelete;
        FunctionInfoDB *functionInfoToDelete;
        DWARFCallFrameInfoDB *dwarfCallFrameInfoToDelete;
        DWARFInfoDB *dwarfInfoToDelete;

        // other data
        uint32_t elfMachineType;
        bool hasELFEndianness;
        bool elfEndiannessBigEndian;
        uint32_t framePointerID;
        uint64_t applicationEntryPoint;
    };
}

#endif

