/**
 * \file    Framework/include/DebugLineInfo.h
 * \date    Copyright (c) 2004 by ARM, Herzogenrath, Germany
 * \ingroup debug_info_database
 *
 * This file contains proprietary, unpublished source code
 * solely owned by ARM, Herzogenrath, Germany.
 * All rights reserved.
 */

#ifndef DEBUGLINEINFO_H
#define DEBUGLINEINFO_H

#include <cstdint>

#include <map>
#include <string>
#include <vector>
#include <iostream>

#include "Coff.h"
#include "DebugCoffSymbols.h"

// offset of line number in Auxiliary entry in COFF symbol table entries (only Archelon generated files)
#define LINENO_OFFSET_AUX 46

namespace DebugInfoDB
{
    class LineInfo
    {
    public:
        LineInfo(uint64_t address,
                 uint16_t srcfileIdx,
                 uint32_t line,
                 uint16_t column,
                 bool is_stmt,
                 bool basic_block = false,
                 bool end_sequence = false);
        uint64_t GetAddress() const         { return address; }
        uint16_t GetSourceFileIndex() const { return srcFileIdx; }
        uint32_t GetLine() const            { return line; }
        uint16_t GetColumn() const          { return column; }
        bool  IsStatement() const        { return is_stmt; }
        bool  IsBasicBlock() const       { return basic_block; }
        bool  IsEndSequence() const      { return end_sequence; }

        inline bool operator < (const LineInfo &b) const
        { return (address < b.address); }

        bool operator == (const LineInfo &b) const
        { return (address == b.address) &&
                AlmostEqual(*this, b); }

        static inline bool LessBySourceReference(const LineInfo &a, const LineInfo &b)
        {
            if (a.srcFileIdx < b.srcFileIdx) return true;
            else if (a.srcFileIdx > b.srcFileIdx) return false;

            else if (a.line < b.line) return true;
            else if (a.line > b.line) return false;

            else return (a.column < b.column);
        }

        // Checks if all info is equal (doesn't compare address)
        static bool AlmostEqual(const LineInfo &a, const LineInfo &b)
        {
            return
                (a.srcFileIdx == b.srcFileIdx) &&
                (a.line == b.line) &&
                (a.column == b.column) &&
                (a.is_stmt == b.is_stmt) &&
                (a.basic_block == b.basic_block) &&
                (a.end_sequence == b.end_sequence);
        }

    protected:
        uint64_t address;        // 8 byte
        uint32_t line;            // 4 byte
        uint16_t srcFileIdx;    // 2 byte
        uint16_t column;        // 2 byte
        bool  is_stmt      : 1;    // 1 byte
        bool  basic_block  : 1;
        bool  end_sequence : 1;
        // total size of class = 20 bytes

        friend class LineInfoDB;
    };

    // For debug purposes:
    std::ostream &operator << (std::ostream &ostr, const LineInfo &lineInfo);

    class LineInfoDB
    {
    public:
        LineInfoDB();
        ~LineInfoDB();

        // copy constructor
        LineInfoDB(const LineInfoDB &a);

        // assignment
        const LineInfoDB &operator=(const LineInfoDB& a);

        void Clear();

        // This returns the index for an existing or newly added source file name
        unsigned int GetSourceFileIndex(const std::string &filename);

        // This returns the index for an existing source file. If the source file could not be found it will return -1.
        // If allowPatrialMatch is false the filename will only be found if it matches exactly one filename in the DB,
        // else it will also be found if in filename some parent directories have not been specified (do not use unless absolutely necessary).
        unsigned int LookupSourceFileIndex(const std::string &filename, bool allowPartialMatch = false) const;

        // This is the reverse operation, return the source file name for an index
        const std::string &GetSourceFile(unsigned int srcFileIndex) const;

        // This can be used to derive the highest valid source file index.
        unsigned int GetNumberOfSourceFiles() const { return (unsigned int)sourceFiles.size(); }

        // add a single line info element
        // jo: AddLineInfo is somewhat performance critical, keep it simple, this used to do insertion sort, now it does almost nothing
        void AddLineInfo(const LineInfo &newInfo) {    lineInfos.push_back(newInfo); lineInfosIsUnclean = true; }

        // get number of line infos
        uint32_t GetNumberOfLineInfos() const { UpdateLineInfos(); return (uint32_t)lineInfos.size(); }
        // get specific line info (id is not the address but an id in the range [0..GetNumberOfLineInfos()-1])
        const LineInfo *GetLineInfo(uint32_t id) const { UpdateLineInfos(); return (id < lineInfos.size()) ? &lineInfos[id] : 0; }

        /* Find the info which is valid for this address (address of
           info found is lower or equal). If no valid info could be
           found a NULL pointer is returned. */
        const LineInfo *Find(uint64_t addr) const;

        // return all address ranges covered by the statement surrounding addr
        // for each element of the list:
        //    first: start address of the range
        //    second: size of the range
        std::vector<std::pair<uint64_t,uint64_t> > GetAddressRangesForStatement(uint64_t addr) const;

        // return all address ranges not covered by line info
        std::vector<std::pair<uint64_t,uint64_t> > GetAddressRangesNotCoveredByLineInfo() const;

        // sort address ranges (sort to start address)
        static void SortAddressRanges(std::vector<std::pair<uint64_t,uint64_t> >& ranges);

        // invert address ranges
        // ranges needs to be sorted and ranges may not overlap
        static void InvertAddressRanges(std::vector<std::pair<uint64_t,uint64_t> >& ranges);

            /** Find the highest line number of a source file which
            produced debug line info. If source file did not produce
            any code 0 will be returned. */
        uint32_t GetLastSourceLine(uint16_t sourceFileIndex) const;

        /** Find the line infos of the first instruction of each code
            sequence generated from a certain source line. Due to
            inlining and code reordering because of optimization there
            might be multiple sequences of instructions generated by
            one source line. This is why the function might return
            multiple addresses. If no such instructions are found an
            empty vector is returned. (Typical application is to
            acquire all addresses to set a source level breakpoint on a
            certain source line.) Parameter 'sourceFileIndex' must be lower than
            GetNumberOfSourceFiles() which is checked by an
            assertion. If 'line' is too big (beyond the end of the
            source file) it will just result in an empty vector. The
            pointers to the LineInfos returned must not be freed in
            any case ! */
        const std::vector<const LineInfo *> FindFirstInstructionsOfSourceLine(uint16_t sourceFileIndex, uint32_t line) const;
        /** Returns true if the previous function returns non-empty vector.
            Implemented for performance reasons. */
        bool  SourceLineHasInstructions(uint16_t sourceFileIndex, uint32_t line) const;

        /** This will find all line-infos which belong to a certain *
            source line. This means you will get the addresses of all
            * instructions which were generated from a source
            line. Due * to inlining and optimizations there might be
            multiple * sequences of instructions from one source
            line. Parameter 'sourceFileIndex' must be lower than
            GetNumberOfSourceFiles() which is checked by an
            assertion. If 'line' is too big (beyond the end of the
            source file) it will just result in an empty vector. The
            pointers to the LineInfos returned must not be freed in
            any case ! */
        const std::vector<const LineInfo *> FindAllInstructionsOfSourceLine(uint16_t sourceFileIndex, uint32_t line) const;

        // add offset to all addresses - e.g. used by linker
        void Relocate(uint64_t offset);

        void AddForeignSourceFilename(const std::string &srcFileName, uint64_t foreignIdx);

        // old method
        void LoadCOFF(const uint8_t data[], uint32_t size, bool bigendian);

        void LoadCoffSymbolInfo(const uint8_t data[], uint32_t size, bool bigendian);

        void PrintSectionHeader(COFF_SCNHDR &sectionHeader);

        void PrintArchelonHeader(COFF_ASCNHDR &archelonHeader);

        void PrintRawData(const uint8_t data[], uint32_t size,uint32_t start);

        uint32_t GetWordFromCharArray(const uint8_t data[], uint32_t pos, bool bigendian);


        // new method which uses Archelon generated Debug info
        void LoadCOFFDebugLineInfo(const uint8_t data[], uint32_t size,const uint8_t symboldata[], uint32_t symbolsize, bool bigendian,std::string sectionFileName,std::list<filenamePos> &coffFilesInSegment);

        // load .debug_line section
        // additional sections are only used to analyze, where line info is present in the file
        void LoadDWARF(const uint8_t debug_line[], uint32_t debug_line_size,
                       const uint8_t debug_info[], uint32_t debug_info_size,
                       const uint8_t debug_abbrev[], uint32_t debug_abbrev_size,
                       const uint8_t debug_loc[], uint32_t debug_loc_size,
                       const uint8_t debug_str[], uint32_t debug_str_size,
                       const uint8_t debug_ranges[], uint32_t debug_ranges_size,
                       bool isBigEndian, uint32_t addressTypeSize);

        // load line info for a single compilation unit:
        // debug_line points to the start of the data somewhere inside the
        // .debug_line section and debug_line_size is the remaining available
        // data in the .debug_line section
        void LoadDWARFCompilationUnit(const uint8_t debug_line[], uint32_t debug_line_size,
                                      bool isBigEndian, uint32_t address_size);

        // get extended address shift
        // (always 0 for .debug_line and >= 0 for .debug_line_ext)
        // (this allows fine grain line info with a finer granularity than real addresses)
        uint32_t GetExtendedAddressShift() const { return extendedAddressShift; }

        // set extended address shift
        // (used by the loader to set the extended address shift which is characteristic to the data loaded by LoadDWARF())
        // (this value is stored in ELF files in the absolute symbol debug_line_ext_shift_value__)
        void SetExtendedAddressShift(uint32_t shift) { extendedAddressShift = shift; }

        void Merge(const LineInfoDB &other);

        // convert this database from extended addresses (as defined by extendedAddressShift) to
        // real addresses (comparable to the PC)
        void ConvertExtendedToRealAddresses();

        // If possible avoid using this method:
        const std::vector<LineInfo> &GetDB() const { UpdateLineInfos(); return lineInfos; }

        // For debugging
        void Dump(std::ostream &ostr) const;




    private:
        // private types
        class SourceFile
        {
        public:
            SourceFile(const std::string &filename = std::string(), uint64_t foreignIndex = uint64_t(-1));
            ~SourceFile();

            std::string filename;

            typedef std::vector<std::vector<const LineInfo *> > ReverseLookupT;
            mutable ReverseLookupT *srcLine2addresses; // (cache)table for faster reverse lookup (line => addresses)

            uint64_t  foreignIndex;    // used for COFF format line info
        };
        typedef std::vector<SourceFile> sourceFilesT;
        typedef std::vector<LineInfo> lineInfosT;


        // private methods

        // get source file index for symbol index (see AddCOFFSourceFilename)
        uint32_t GetSourceFileIndexForForeignIndex(uint64_t foreignIndex) const;

        // This must be called internally if some change of the data (e.g. AddLineInfo()) occurs
        void InvalidateCaches() const;

        // this must be called internally before 'lineInfos' is used
        void UpdateLineInfos() const;

        // This gives back a reverse lookup table for a source
        // file. It will create such a table if no such table has been
        // created so far.
        const SourceFile::ReverseLookupT *GetReverseLookupTable(uint16_t sourceFileIndex) const;


        // private data
        sourceFilesT sourceFiles;
        lineInfosT lineInfos;
        bool lineInfosIsUnclean; // jo: set by AddLineInfo and reset by UpdateLineInfos()
        const std::string emptyString; // used to return references to empty strings

        // extended address shift: to get real addresses (addresses comparable to the PC) the addresses in this
        // container need to be shifted right by this)
        // (this variable is not used in the algorithms in this class, it is just a container so that anyone who
        // has access to this LineInfo has access to this value)
        uint32_t extendedAddressShift;
    };
}

#endif
