/*!
 * \file    SystemGenComponent.h
 * \date    Copyright 2002-2013 ARM Limited. All rights reserved.
 */

#ifndef SYSTEMGENCOMPONENT_H
#define SYSTEMGENCOMPONENT_H

#include <map>
#include "SGComponentWithPorts.h"
#include "sg/SGControlProtocol.h"
#include "sg/Attributes.h"
#include "sg/SGSimulationContext.h"
#include "sg/SGSchedulerInterfaceForComponents.h"
#include "CADIProtocol.h"
#include "CADIBase.h"
#include "cp/SerializationInterface.h"
#include "iris/IrisInstance.h"

namespace sg
{
    class EventSourceBase;
    class CADIIrisAdapter;
    class MTWrapperConfig;
}

class SystemGenComponent : public sg::ComponentWithPorts, public cp::Serializable
{
public:

    struct property_value_pair_t {
        char const*   property;
        char const*   value;

        // Size of value, or if zero then value is a NUL terminated string
        size_t        size;
    };

    SystemGenComponent(
        const char*                     instName,
        sg::SimulationContext*          simulationContext,
        const sg::Params &              params,
        property_value_pair_t const*    property_list
        );
    ~SystemGenComponent();

    virtual void serializeState(const char *name, cp::SerializationInterface *stream) SG_OVERRIDE;

    // 'getInstanceName()' is a LISA keyword and a LISA langage feature, do not alter
    // (the implementation may be changed however)
    std::string getInstanceName() const { return getName(); }
    // get hierarchical instance name in CADI consistent way - without top component
    // (for top component itself - empty string)
    std::string getInstancePath() const;

    // default implementation of some behaviors
    void step() {}
    void simShutdown();
    void cadiRefresh(uint32_t refreshReason);
    // this should be called by LISA components whenever a breakpoint is hit
    // (must always be called before the simHalt())
    void simBreakpointHit(uint32_t cadiBptNum);

    // MaxSim-style message functions for all SG components
    //
    // For non-fatal message() calls prior to init(), then the messages are
    // buffered and drained just before init() is called.
    //
    // For fatal messages() then all prior messages are drained to std::cout and
    // then the fatal message is generated to std::cout and then abort() called.
    //
    // See also setShutdownAtReset(), which means that you can use non-fatal
    // error messages and then generate a fatal error message at reset().
    void message(const std::string &msg, sg::message::MessageType msgType);
    void message(sg::message::MessageType msgType, const char *fmt, ...) SG_FORMATSTR(printf, 3, 4);

    void addExtendedTargetFeature(const std::string &etf);

    void addTraceSource(sg::EventSourceBase *es);

    const std::string& getExtendedTargetFeatures() const { return extendedTargetFeatures; }

    // get global interface
    // (intended to be used from LISA to get a list of all components in the system from within any component)
    eslapi::CAInterface *getGlobalInterface();
    // get SCADI interface
    // (intended to be used from LISA to access the SCADI interface of the own component)
    eslapi::CADI *getSCADI();

    void buildDebugInfo__(const sg::CADIBase::RegInfo* regInfoTable, uint32_t regInfoTableSize,
                          const sg::CADIBase::RegInfoInstanceSpecific* regInfoInstanceSpecificTable,
                          uint32_t regInfoInstanceSpecificTableSize,
                          const sg::CADIBase::ParamInfo* paramInfoTable, uint32_t paramInfoTableSize);

    virtual std::string getHierarchicalName() const SG_OVERRIDE { return hierarchicalName; }

    // Return a hierarchical name of subcomponent: "hierarchicalName:subcompName"
    // This is used to pass the hierarchicalName into a subcomponent's constructor. Usually used
    // in instantiation stage.
    // E.g. If we use a factory to instantiate a subcomponent:
    //     add(factory->instantiate(subcompHierName("name")));
    //
    // E.g. If we new a subcomponent:
    //     Component* comp = new Subcomp(subcompHierName("name"));
    //     add(comp);    // Or: comp->setParent(this);
    //
    // Note: Some sub-components' name already contains dot '.' (such as "cpu0.l1cache"),
    //       so a colon is used as the separator instead. Otherwise we can not extract
    //       the subcomponent name.
    inline const std::string subcompHierName(const std::string& subcompName)
    {
        return hierarchicalName + ":" + subcompName;
    }

    // In the case this is the main component of an EVS of a complete system
    // the EVS name prefix might be dropped from parameters and CADI targets
    // to be called by the SCXExportedVirtualSubsystem if needed
    inline void dropCadiNamePrefix()
    {
        cadiNamePrefix = "";
    }

    // Access to the Iris instance for LISA components which want to access other component (or even clients) in the system.
    // 'getIrisInstance()' is a LISA keyword.
    iris::IrisInstance *getIrisInstance()
    {
        return irisInstance;
    }

protected:
    static int64_t             GetParameterValue(const sg::Params &params,
                                               const std::string &param_name,
                                               int64_t val_default,
                                               int64_t min_val, int64_t max_val);

    static bool GetParameterValue(const sg::Params &params, const std::string &param_name, bool val_default);

    static const std::string GetParameterValue(const sg::Params &params,
                                        const std::string &param_name,
                                        const std::string &val_default);

    static const std::string GetParameterValue(const sg::Params &params,
                                        const std::string &param_name,
                                        const char *val_default)
    {
        return GetParameterValue(params, param_name, std::string(val_default));
    }

protected:
    sg::SimulationContext* simulationContext__;
    sg::CADIBase* cadi__;
    // the following map allows to get an cadi interface of sub component by name
    std::map< std::string, eslapi::CAInterface* > componentName2CADI__;
    const sg::Params externParameters__;
    std::map<std::string,std::string> properties__; // contents of the properties section of this component

    // CADIPort (the CADI interface forwards some CADI calls to this port)
    sg::CADIProtocol cadiPort__;
    // applications loaded by the component and its command line params
    std::map<std::string, std::string> applications2params;

    iris::IrisInstance* irisInstance;
    sg::CADIIrisAdapter *cadi_iris_adapter__;

    // contains the target name prefix including '.'
    std::string cadiNamePrefix;

    sg::SchedulerInterfaceForComponents * sched__;
    void initScheduler();

    // It uses the properties__["mtdomains"] and externParameters__.
    //
    // out_.error will indicate if there was an error or not.  It will have
    // already issued an error message to message(MSG_ERROR) and
    // setShutdownAtReset().
    void fillOutMTWrapperConfig__(sg::MTWrapperConfig& out_);

    // list of all CADI parameters
    std::vector<eslapi::CADIParameterInfo_t> parameterInfos__;

private:
    void connectControlPorts(sg::ComponentBase *c) SG_OVERRIDE;
    void disconnectControlPorts(sg::ComponentBase *c) SG_OVERRIDE;
    void populateChildControlPort(sg::CADIMap &, const std::string&) SG_OVERRIDE;

    /// Copy a LISA property into to the corresponding Iris string property.
    void copyPropertyToIrisString(const std::string &lisa_name, const std::string &iris_name);

    /// Copy a LISA property into to the corresponding Iris NumberU64 property.
    void copyPropertyToIrisU64(const std::string &lisa_name, const std::string &iris_name);

private:
    std::string hierarchicalName;

    std::string extendedTargetFeatures;

public:
    sg::SGControlProtocol control_port;
    sg::SGControlProtocol child_control_port;

protected:
    // If 'emergency' is true then drain all to std::cout.
    //
    // Otherwise, if bufferMessages then will buffer messages.
    //
    // If ! bufferMessages and control port 'message' is bound then drain all
    // bufferedMessages.
    //
    // NOTE that the message port is always bound by the generated simgen
    // templates and so we need an explicit way of saying when it is safe to
    // drain.
    void drainBufferedMessages__(bool emergency_ = false);

    // Indicate we need no longer buffer messages.
    void dontBufferMessages__();
    
    /// If you call setShutdownAtReset() then when reset() is called then a
    /// MSG_FATAL_ERROR will terminate the simulation.
    void setShutdownAtReset__()      { shutdownAtReset__ = true; }
    
    bool isShutdownAtReset__() const { return shutdownAtReset__; }
    
private:
    // Buffer strings from message() when not yet connected.
    //
    // MSG_FATAL_ERROR drains all prior messages and then prints out the fatal
    // message and then abort()s.
    //
    // The buffered messages are retained in order and drained when
    // drainBufferedMessages() is called.
    std::vector< std::pair<std::string, sg::message::MessageType> > bufferedMessages__;

    bool bufferMessages__  { true  };
    bool shutdownAtReset__ { false };
};

#endif

