/*!
 * \file    BaseCheckpoint.h
 * \brief   Helper classes for adding checkpointing to components.
 * \date    Copyright 2014 ARM Limited. All rights reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 */

#ifndef BASE_CHECKPOINT_H
#define BASE_CHECKPOINT_H

#include "cp/CheckpointInterface.h"
#include "cp/SerializationInterface.h"
#include "sg/Attributes.h"
#include "sg/SGComponentTrace.h"

namespace cp
{

class BaseCheckpoint : public CheckpointInterface
                     , public SerializationInterface::Callback
{
protected:
    enum Operation
    {
        SAVE,
        RESTORE
    } current_operation;

    cp::StreamInterface *current_si;

private:
    eslapi::CAInterface *parent;
    typedef std::map<std::string, cp::SerializationInterface *> SerializationInterfaceMap;
    SerializationInterfaceMap open_streams;

    // trace sources
    sg::EventSource0 save_start_trace;
    sg::EventSource0 save_end_trace;
    sg::EventSource0 restore_start_trace;
    sg::EventSource0 restore_end_trace;
    sg::EventSource1<const char *> message_trace;

    void fire_save_start_trace();
    void fire_save_end_trace();
    void fire_restore_start_trace();
    void fire_restore_end_trace();
    void fire_message(const char *message);

    bool save_restore_helper(cp::StreamInterface *si, Operation op);

protected:

    // Open a named stream to save to/restore from.
    SerializationInterface *open_stream(const char *name);

    // Implemented by subclass to 'describe' the state of the checkpoint.
    // This is done by calling open_stream() and calling  the array(), field(),
    // and object() methods on the SerializationInferface that it returns.
    // This allow us to express the mapping between internal state and checkpoint
    // state once, ensuring that all operations stay in sync.
    virtual void state() = 0;

    // Optionally implemented by subclass to 'describe' the preconditions
    // for checkpointing, i.e. whether a checkpoint saved in a simulation instance
    // can be restored in a new simulation instance.
    // This is done by calling the array(), field(), and object() methods on the
    // SerializationInterface passed in as a parameter, without having to
    // distinguish the case where we declare the preconditions and where we check them.
    virtual void preconditions(const char * /*stream_name*/, cp::SerializationInterface * /*stream*/) {}

    virtual std::string get_component_name();

public: // cp::CheckpointInterface
    virtual eslapi::CAInterface *ObtainInterface(eslapi::if_name_t ifName, eslapi::if_rev_t minRev, eslapi::if_rev_t *actualRev) SG_OVERRIDE;

    virtual bool saveState(cp::StreamInterface *si)    SG_OVERRIDE;
    virtual bool restoreState(cp::StreamInterface *si) SG_OVERRIDE;

public: // cp::SerializationInterface::Callback
    virtual void report_message(const char *message) SG_OVERRIDE { fire_message(message); }

public:
    BaseCheckpoint(eslapi::CAInterface *parent);
    virtual ~BaseCheckpoint();
};

} // namespace cp

#endif // BASE_CHECKPOINT_H
