/*!
 * \file    SerializationInterface.h
 * \brief   Interface to serialize and deserialize data.
 * \date    Copyright 2014-2015 ARM Limited. All rights reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 */

#ifndef SERIALIZATION_H
#define SERIALIZATION_H

#include "sg/Attributes.h"
#include <sg/IncludeMeFirst.h>
#include <cstdint>
#include <algorithm>
#include <map>
#include <string>
#include <set>
#include <cassert>

namespace cp
{

class SerializationInterface;

class Serializable
{
public:
    virtual ~Serializable() {}

    virtual void serializeState(const char *name, SerializationInterface *serializer) = 0;
};

namespace SerializableHierarchy
{

// SerializationInterface includes utility methods to deal with the
// (de)serialization of containers that hold pointers to objects belonging to a
// class hierarchy.
//
// In order to use those utility methods, the root class in the hierarchy needs
// to implement the following interface and all the concrete classes must
// provide their unique implementation of getTypeID().
class Base : public Serializable
{

public:

    // Return a textual ID that uniquely identifies this type
    virtual std::string getTypeID() const = 0;

    virtual ~Base() {}

};

// Additionally, a way to create objects of the correct concrete type must be
// provided to SerializationInterface; this takes the form of an instance of the
// following Factory interface per concrete type that can be serialized to the
// stream. Each instance is only capable to handle one type, whose ID causes
// matches(std::string const&) to return true.
class Factory
{

public:

    // Return true if the type, whose ID has been passed as the only parameter,
    // can be handled; return false, otherwise.
    virtual bool matches(std::string const& type_id) const = 0;

    // Create a new instance of the concrete class and set its state according
    // to the data read from the stream.
    virtual Base* create(char const* name, SerializationInterface* si) = 0;

    virtual ~Factory() {}

};

} // namespace SerializableHierarchy

class SerializationInterface
{
private:
    // Implementation for getter()/setter() versions of array() and field()
    // This avoids duplicating the implementation for methods that take const
    // and non-const getters.
    // These methods are not safe to use directly because the types of the getter
    // and setter are not guaranteed to match up.
    // You should only use the public versions below.

    // Serialize/deserialize a named field using getter() and setter().
    template <typename T, typename CLS, typename GETTER, typename SETTER>
    void field_helper(const char *name, CLS *component, GETTER getter, SETTER setter)
    {
        T t;

        if (serializing())
        {
            t = (component->*getter)();
        }

        field(name, t);

        if (!serializing())
        {
            (component->*setter)(t);
        }
    }

    // Serialize/deserialize a named field using getter() and setter() which take a parameter.
    template <typename T, typename CLS, typename GETTER, typename SETTER, typename PARAM_TYPE>
    void field_helper(const char *name, CLS *component, GETTER getter, SETTER setter, PARAM_TYPE param)
    {
        T t;

        if (serializing())
        {
            t = (component->*getter)(param);
        }

        field(name, t);

        if(!serializing())
        {
            (component->*setter)(t, param);
        }
    }

    // Serialize/deserialize a named array using getter() and setter().
    template <typename T, typename CLS, typename GETTER, typename SETTER>
    void array_helper(const char *name, CLS *component, GETTER getter, SETTER setter, std::size_t size)
    {
        T *t_array = new T[size];

        if (serializing())
        {
            memcpy(t_array, (component->*getter)(), size * sizeof(T));
        }

        array(name, t_array, size);

        if (!serializing())
        {
            (component->*setter)(t_array);
        }

        delete[] t_array;
    }

    // Serialize/deserialize a named array using getter() and setter() which take a parameter.
    template <typename T, typename CLS, typename GETTER, typename SETTER, typename PARAM_TYPE>
    void array_helper(const char *name, CLS *component, GETTER getter, SETTER setter, std::size_t size, PARAM_TYPE param)
    {
        T *t_array = new T[size];

        if (serializing())
        {
            memcpy(t_array, (component->*getter)(param), size * sizeof(T));
        }

        array(name, t_array, size);

        if (!serializing())
        {
            (component->*setter)(t_array, param);
        }

        delete[] t_array;
    }

    // Utility class to use SerializableHierarchy::Factory with STL algorithms
    class SerializableHierarchyMatcher
    {
    public:
        SerializableHierarchyMatcher(std::string type_id_) : type_id(type_id_) {}
        bool operator()(SerializableHierarchy::Factory* factory)
        {
            return factory->matches(type_id);
        }
    private:
        std::string type_id;
    };

public:
    class Callback
    {
    public:
        // used for reporting errors during serialization operation.
        virtual void report_message(const char *message) = 0;
        virtual ~Callback() {}
    };

    // Reporting warnings/errors during serialization operation.

    virtual void report_message(const char *message) = 0;

    // Serialize/deserialize named fields and arrays.
    // These methods are implemented by derived classes and read or write the
    // field/array as appropriate for the operation being performed.

    virtual void field(const char *name, uint8_t  &value) = 0;
    virtual void field(const char *name, uint16_t &value) = 0;
    virtual void field(const char *name, uint32_t &value) = 0;
    virtual void field(const char *name, uint64_t &value) = 0;
    virtual void field(const char *name, int8_t   &value) = 0;
    virtual void field(const char *name, int16_t  &value) = 0;
    virtual void field(const char *name, int32_t  &value) = 0;
    virtual void field(const char *name, int64_t  &value) = 0;
    virtual void field(const char *name, double   &value) = 0;
    virtual void field(const char *name, bool     &value) = 0;
    virtual void field(const char *name, std::string &value) = 0;

    virtual void start_object(const char *name) = 0;
    virtual void end_object(const char *name) = 0;

    // Pass a pointer in when deserializing an array if its size needs to be known.
    virtual void start_array(const char *name, std::size_t *size = 0) = 0;
    virtual void end_array(const char *name) = 0;

    // Returns true  if a serialize operation is being performed,
    //         false if a deserialize operation is being performed.
    virtual bool serializing() const = 0;


    // Serialize/deserialize a named field.
    // The field is passed by reference to this method and read or written as appropriate
    // for the operation being performed.
    template <typename T>
    void field(const char *name, T &t);

    // Serialize anything that implements Serializable
    void field(const char* name, Serializable& t)
    {
        t.serializeState(name, this);
    }
    void field(const char* name, Serializable* t)
    {
        t->serializeState(name, this);
    }

    template <typename T>
    void array(const char *name, T *t_array, std::size_t size)
    {
        start_array(name);
        for (std::size_t i=0; i < size; i++)
        {
            field(t_array[i]);
        }
        end_array(name);
    }

    template<template<typename T, typename All = std::allocator<T> > class RESIZABLE_CONTAINER, class SERIALIZABLE>
    void array(char const* name, RESIZABLE_CONTAINER<SERIALIZABLE>& serializables, SERIALIZABLE val = SERIALIZABLE())
    {
        std::size_t number_of_serializables;
        start_array(name, &number_of_serializables);
        if (!serializing())
            serializables.resize(number_of_serializables, val);
        for (typename RESIZABLE_CONTAINER<SERIALIZABLE>::iterator it = serializables.begin();
            it != serializables.end(); it++)
            (*it).serializeState(NULL, this);
        end_array(name);
    }

    // Serialize to/from an array std::sets of objects that can be serialized
    template<typename T>
    void array(const char *name, std::set<T> &cb)
    {
        std::size_t size = 0;
        start_array(name, &size);
        if (serializing()) {
            for(typename std::set<T>::iterator i = cb.begin(); i != cb.end(); i++)
            {
                T value = *i;
                field(value);
            }
        } else {
            cb.clear();
            for(std::size_t i = 0; i < size; i++) {
                T value;
                field(value);
                cb.insert(value);
            }
        }
        end_array(name);
    }

    // This method is used to (de)serialize a SequenceContainer whose items are of a type
    // for which a suitable implementation of field exists.
    template< template< typename T, typename Alloc = std::allocator<T> > class SequenceContainer
            , typename Item >
    void field(char const* name, SequenceContainer<Item>& items)
    {
        std::size_t number_of_items = 0;
        start_array(name, &number_of_items);

        if (serializing())
        {
            for (typename SequenceContainer<Item>::iterator it = items.begin(); it != items.end(); it++)
            {
                Item item = *it;
                field(item);
            }
        }
        else
        {
            items.clear();
            for(std::size_t i = 0; i < number_of_items; i++)
            {
                Item item;
                field(item);

                items.insert(items.end(), item);
            }
        }

        end_array(name);
    }

    // This method is used to (de)serialize a std::map. The map's Key and T types are
    // required to be of a type for which a suitable implementation of field exists. Note
    // that this implies that this also covers the case where T is a std::map itself.
    template< typename Key, typename T >
    void field(char const* name, std::map<Key, T>& items)
    {
        std::size_t number_of_items = 0;
        start_array(name, &number_of_items);

        if (serializing())
        {
            for (typename std::map<Key, T>::iterator it = items.begin(); it != items.end(); it++)
            {
                Key key = it->first;
                T value = it->second;

                start_array();
                field(key);
                field(value);
                end_array();
            }
        }
        else
        {
            // Start with an empty std::map
            items.clear();

            for (std::size_t i = 0; i < number_of_items; i++)
            {
                Key key;
                T value;

                start_array();
                field(key);
                field(value);
                end_array();

                items[key] = value;
            }
        }
        end_array(name);
    }

    // This method is used to (de)serialize a SequenceContainer that holds pointers to
    // objects of a class hierarchy. The pointer type is a pointer to the base class,
    // while the actual pointed-to objects are of concrete types.
    //
    // The type of the base class is taken as the template parameter
    // SerializableHierarchyElement and the implementation assumes it implements
    // the SerializableHierarchy::Base interface.
    //
    // An additional third parameter is required to create objects of the correct concrete
    // type from the serialization stream. It takes the form of a Container of pointers to
    // implementations of the SerializableHierarchy::Factory interface, which has to be
    // provided by the caller and that is assumed to contain one instance per concrete
    // type (each SerializableHierarchy::Factory instance is only capable to handle one
    // type).
    template< template< typename T, typename Alloc = std::allocator<T> > class SequenceContainer
            , typename SerializableHierarchyElement
            , template< typename T, typename Alloc = std::allocator<T> > class Container >
    void field(char const* name,
               SequenceContainer<SerializableHierarchyElement*>& objects,
               Container<SerializableHierarchy::Factory*>& factories)
    {
        size_t number_of_objects;
        start_array(name, &number_of_objects);

        std::string type_id;
        if (serializing())
        {
            // Serialization case

            for (typename SequenceContainer<SerializableHierarchyElement*>::iterator it = objects.begin();
                it != objects.end(); it++)
            {
                // Assume SerializableHierarchyElement is a SerializableHierarchy::Base
                assert(dynamic_cast<SerializableHierarchy::Base*>(*it));
                SerializableHierarchy::Base* object = static_cast<SerializableHierarchy::Base*>(*it);

                // Output a type identifier to the serialization stream, so that we can
                // create the right type of object when we deserialize the stream
                type_id = object->getTypeID();
                field(NULL, type_id);

                // SerializableHierarchy::Base is a Serializable, so delegate serialization to the object itself
                field(static_cast<Serializable*>(object));
            }
        }
        else
        {
            // Deserialization case

            // Start with an empty SequenceContainer
            objects.clear();

            for (size_t i = 0; i < number_of_objects; i++)
            {
                // Read the type identifier from the serialization stream
                field(type_id);

                // Find the factory that can create the object of the right type: it is
                // assumed that such factory can always be found (an assertion will fail,
                // otherwise)
                typename Container<SerializableHierarchy::Factory*>::iterator it = std::find_if(factories.begin(),
                                                                                                factories.end(),
                                                                                                SerializableHierarchyMatcher(type_id));
                assert(it != factories.end());

                SerializableHierarchy::Base* object = (*it)->create(NULL, this);

                // Assume that what the factory creates is a SerializableHierarchyElement
                assert(dynamic_cast<SerializableHierarchyElement*>(object));

                objects.insert(objects.end(), static_cast<SerializableHierarchyElement*>(object));
            }
        }
        end_array(name);
    }

    // These versions with no name argument should be used to add elements to an array.
    void start_array(std::size_t *size = 0)   { start_array(NULL, size);    }
    void end_array()                          { end_array(NULL);            }
    void start_object()                       { start_object(NULL);         }
    void end_object()                         { end_object(NULL);           }
    template <typename T>
    void field(T &t)                          { field(NULL, t);             }
    template <typename T>
    void array(T *t_array, std::size_t size)  { array(NULL, t_array, size); }
    void field(Serializable *t)               { field(NULL, t);             }
    void array(Serializable **serializables, std::size_t size) { array(NULL, serializables, size); }

    // NOTE: Serializable version of array() has a different signature to other array methods()
    //       it takes Serializable **, not Serializable *
    void array(const char *name, Serializable **serializables, std::size_t size)
    {
        start_array(name);
        for (std::size_t i=0; i < size; i++)
        {
            start_object();
            serializables[i]->serializeState(NULL, this);
            end_object();
        }
        end_array(name);
    }

    // Serialize/deserialize a named field using getter() and setter().
    // i.e: si->field("my_field", my_class, &MyClass::get_my_field, &MyClass::set_my_field);
    // Must be used inside an object.
    template <typename T, typename COMP_CLS, typename CLS>
    void field(const char *name, COMP_CLS *component, T (CLS::*getter)(), void (CLS::*setter)(T))
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(), void (CLS::*)(T)>(name, cls_component, getter, setter);
    }

    template <typename T, typename COMP_CLS, typename CLS>
    void field(const char *name, COMP_CLS *component, T (CLS::*getter)() const, void (CLS::*setter)(T))
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)() const, void (CLS::*)(T)>(name, cls_component, getter, setter);
    }

    // Serialize/deserialize a named field using getter() and setter() which take a parameter.
    // i.e: si->field("my_field", my_class, &MyClass::get_my_field, &MyClass::set_my_field, my_param);
    // Must be used inside an object.
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void field(const char *name, COMP_CLS *component, T (CLS::*getter)(PARAM_TYPE), void (CLS::*setter)(T, PARAM_TYPE), PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(PARAM_TYPE), void (CLS::*)(T, PARAM_TYPE), PARAM_TYPE>(name, cls_component, getter, setter, param);
    }
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void field(const char *name, COMP_CLS *component, T (CLS::*getter)(PARAM_TYPE) const, void (CLS::*setter)(T, PARAM_TYPE), PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(PARAM_TYPE) const, void (CLS::*)(T, PARAM_TYPE), PARAM_TYPE>(name, cls_component, getter, setter, param);
    }

    // Serialize/deserialize a named array using getter() and setter().
    // i.e: si->array("my_field", my_class, &MyClass::get_my_array, &MyClass::set_my_array);
    // Must be used inside an object.
    template <typename T, typename COMP_CLS, typename CLS>
    void array(const char *name, COMP_CLS *component, const T *(CLS::*getter)(), void (CLS::*setter)( const T*), std::size_t size)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(), void (CLS::*)( const T*)>(name, cls_component, getter, setter, size);
    }
    template <typename T, typename COMP_CLS, typename CLS>
    void array(const char *name, COMP_CLS *component, const T *(CLS::*getter)() const, void (CLS::*setter)( const T*), std::size_t size)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)() const, void (CLS::*)( const T*)>(name, cls_component, getter, setter, size);
    }

    // Serialize/deserialize a named array using getter() and setter() which take a parameter.
    // i.e: si->array("my_field", my_class, &MyClass::get_my_array, &MyClass::set_my_array, my_param);
    // Must be inside an object.
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void array(const char *name, COMP_CLS *component, const T *(CLS::*getter)(PARAM_TYPE), void (CLS::*setter)(const T*, PARAM_TYPE), std::size_t size, PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(PARAM_TYPE), void (CLS::*)(const T*, PARAM_TYPE), PARAM_TYPE>(name, cls_component, getter, setter, size, param);
    }
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void array(const char *name, COMP_CLS *component, const T *(CLS::*getter)(PARAM_TYPE) const, void (CLS::*setter)(const T*, PARAM_TYPE), std::size_t size, PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(PARAM_TYPE) const, void (CLS::*)(const T*, PARAM_TYPE), PARAM_TYPE>(name, cls_component, getter, setter, size, param);
    }


    // Serialize/deserialize a named field using getter() and setter().
    // i.e: si->field(my_class, &MyClass::get_my_field, &MyClass::set_my_field);
    // Must be inside an array.
    template <typename T, typename COMP_CLS, typename CLS>
    void field(COMP_CLS *component, T (CLS::*getter)(), void (CLS::*setter)(T))
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(), void (CLS::*)(T)>(NULL, cls_component, getter, setter);
    }

    template <typename T, typename COMP_CLS, typename CLS>
    void field(COMP_CLS *component, T (CLS::*getter)() const, void (CLS::*setter)(T))
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)() const, void (CLS::*)(T)>(NULL, cls_component, getter, setter);
    }

    // Serialize/deserialize a unnamed field using getter() and setter() which take a parameter.
    // i.e: si->field("my_field", my_class, &MyClass::get_my_field, &MyClass::set_my_field, my_param);
    // Must be inside an array.
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void field(COMP_CLS *component, T (CLS::*getter)(PARAM_TYPE), void (CLS::*setter)(T, PARAM_TYPE), PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(PARAM_TYPE), void (CLS::*)(T, PARAM_TYPE), PARAM_TYPE>(cls_component, getter, setter, param);
    }
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void field(COMP_CLS *component, T (CLS::*getter)(PARAM_TYPE) const, void (CLS::*setter)(T, PARAM_TYPE), PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        field_helper<T, CLS, T (CLS::*)(PARAM_TYPE) const, void (CLS::*)(T, PARAM_TYPE), PARAM_TYPE>(NULL, cls_component, getter, setter, param);
    }

    // Serialize/deserialize a unnamed array using getter() and setter().
    // i.e: si->array(my_class, &MyClass::get_my_array, &MyClass::set_my_array);
    // Must be inside an array.
    template <typename T, typename COMP_CLS, typename CLS>
    void array(COMP_CLS *component, const T *(CLS::*getter)(), void (CLS::*setter)( const T*), std::size_t size)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(), void (CLS::*)( const T*)>(NULL, cls_component, getter, setter, size);
    }
    template <typename T, typename COMP_CLS, typename CLS>
    void array(COMP_CLS *component, const T *(CLS::*getter)() const, void (CLS::*setter)( const T*), std::size_t size)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)() const, void (CLS::*)( const T*)>(NULL, cls_component, getter, setter, size);
    }

    // Serialize/deserialize a unnamed array using getter() and setter() which take a parameter.
    // i.e: si->array(my_class, &MyClass::get_my_array, &MyClass::set_my_array, my_param);
    // Must be inside an array.
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void array(COMP_CLS *component, const T *(CLS::*getter)(PARAM_TYPE), void (CLS::*setter)(PARAM_TYPE, const T*), std::size_t size, PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(PARAM_TYPE), void (CLS::*)(PARAM_TYPE, const T*), PARAM_TYPE>(NULL, cls_component, getter, setter, size, param);
    }
    template <typename T, typename COMP_CLS, typename CLS, typename PARAM_TYPE>
    void array( COMP_CLS *component, const T *(CLS::*getter)(PARAM_TYPE) const, void (CLS::*setter)(PARAM_TYPE, const T*), std::size_t size, PARAM_TYPE param)
    {
        // COMP_CLS must be either the same type as CLS or inherit from it
        CLS *cls_component = static_cast<CLS*>(component);
        array_helper<T, CLS, const T *(CLS::*)(PARAM_TYPE) const, void (CLS::*)(PARAM_TYPE, const T*), PARAM_TYPE>(NULL, cls_component, getter, setter, size, param);
    }

    // convenience version which uses templates to detect the array size
    template<typename T, std::size_t n>
    inline void array(const char *name, T (&t_array)[n])
    {
        array(name, t_array, n);
    }

    // convenience version which uses templates to detect the array size
    template<typename T, std::size_t n>
    inline void array(T (&t_array)[n])
    {
        array(t_array, n);
    }

    virtual ~SerializationInterface() {}
};

} // namespace cp

#endif // SERIALIZATION_H
