/*!
 *  \file   StreamInterface.h
 *  \brief  Stream I/O interface for use by checkpoint API
 *  \date   Copyright ARM Limited 2013 All Rights Reserved.
 *
 */

#ifndef STREAM_INTERFACE_H
#define STREAM_INTERFACE_H

#include "eslapi/CAInterface.h"

#include <string>

namespace cp
{
    class StreamInterface : public eslapi::CAInterface
    {
    public:
        /**
         *  An I/O mechanism implemented by a CADI client to provide a way of
         *  getting checkpoint state in and out of the model. The client
         *  provides a StreamInterface pointer to the model, and the model
         *  opens named streams for reading/writing, and can perform reads from
         *  and writes to any open stream in any arbitrary order.
         *
         *  The API is similar to file I/O in POSIX, and the client might
         *  implement the streams as a set of files, one per stream, stored in
         *  directory.
         *
         *  The client may buffer reads and writes, by reading the underlying
         *  storage in larger blocks than requested and coalescing writes into
         *  a buffer.
         *
         *  Reads and writes may return less than the number of bytes requested,
         *  in which case the model will retry with the remainder of the buffer
         *  to be accessed.
         *
         *  Writes are not guaranteed to complete until the stream is closed.
         */

        enum StreamMode
        {
            ReadOnly,    // Read only stream
            WriteOnly,   // Write only stream
            WriteAppend, // Write only stream that's appended to
        };

        /*
         *  Open a stream for reading/writing. The streamName must be a
         *  string containing one or more [a-z0-9_.] characters.
         *  Specifically, upper case characters, quotes, punctuation and
         *  filesystem delimiters such as ':', '/' and '\' are forbidden
         *  and will cause openStream to fail.
         *
         *  If the client doesn't have a stream by this name, openStream
         *  with mode = ReadOnly fails; and openStream with mode = WriteOnly
         *  or mode = WriteAppend creates a new stream.
         *
         *  If the client does have a stream by this name, openStream with
         *  mode = WriteOnly deletes the stream's previous contents, and
         *  creates a new empty stream, but openStream with mode = WriteAppend
         *  will open the existing stream and put the write pointer at the end
         *  of the stream.
         *
         *  On error, the stream ID -1 is returned. On success, a stream
         *  ID of 0 or greater is returned. Multiple streams may be
         *  opened simultaneously. Stream IDs become invalid once a stream
         *  is closed, and a stream ID may be reused by subsequent
         *  streamOpen() calls.
         */
        virtual int32_t openStream(const char *streamName, StreamMode mode) = 0;

        /*
         *  Close an open stream and invalidate the stream ID.
         *  If the stream is writable, any buffered stream contents are
         *  guaranteed to be written.
         */
        virtual void closeStream(int32_t stream) = 0;

        /*
         * Reads up to count bytes from a readable stream into the buffer
         * buf.
         *
         * Returns -1 on error (if the stream ID is invalid, isn't a readable stream),
         * Returns 0 if the end of the stream has been reached.
         * Returns the number of bytes read (which may be 1...count) on success.
         */
        virtual int32_t readBuf(int32_t stream, void *buf, uint32_t count) = 0;

        /*
         * Writes up to count bytes from the buffer buf into a writeable
         * stream.
         *
         * Returns -1 on error (if the stream ID is invalid, isn't a writeable stream,
         * or some other fatal error occurs such as the client running out of storage)
         * Returns the number of bytes written (which may be 0...count) on success.
         */
        virtual int32_t writeBuf(int32_t stream, const void *buf, uint32_t count) = 0;

        static eslapi::if_name_t IFNAME()     { return "StreamInterface0"; }
        static eslapi::if_rev_t  IFREVISION() { return 0; }
    };
}

#endif

// end of file StreamInterface.h
