/*
 * SCX: scx_cadi_target.h - SystemC eXport support: Debug target implementation,
 *                           including registers, memory spaces and parameters.
 *
 * Copyright 2016 ARM Limited.
 * All rights reserved.
 *
 * WARNING! This file is not part of a stable public API and
 *          is subject to change without notice.
 */

#ifndef ARM_SCX_CADI_TARGET_H
#define ARM_SCX_CADI_TARGET_H

/*!
 * \file        scx_cadi_target.h
 * \brief       Debug target implementation, including registers, memory spaces
 *              and parameters.
 * \date        Copyright 2016 ARM Limited. All rights reserved.
 *
 * SystemC eXport support Debug target implementation, including registers,
 * memory spaces and parameters.
 */

/* Includes */
#include <eslapi/CADITypes.h>
#include <eslapi/CADI.h>
#include <eslapi/CAInterfaceRegistry.h>
#include <sg/Attributes.h>
#include <sg/UnorderedMap.h>
#include <PVVersion.h>
#include <TxLib/TxLib.h>

#include <string>
#include <list>
#include <utility>
#include <vector>

#include "scx/scx_simcontrol.h"
#include "scx/scx_debug_target_if.h"

/*!
 * \brief Specialized \c operator==() for \c eslapi::CADIAddr_t type.
 */
inline bool
operator==(const eslapi::CADIAddr_t & l, const eslapi::CADIAddr_t & r) {
    return ((l.space == r.space) && (l.addr == r.addr));
}

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Globals */

/*!
 * Max value for CADI register bits wide attribute.
 */
const uint32_t
SCX_CADI_REG_BITS_WIDE_MAX = 128;

/*!
 * \brief Specialized hash function for \c eslapi::CADIAddr_t type.
 */
struct hash {
    size_t
    operator()(const eslapi::CADIAddr_t & k) const {
        return (std::hash<eslapi::CADIMemSpace_t>()(k.space)
                ^ std::hash<eslapi::CADIAddrSimple_t>()(k.addr));
    }
};

/* Forwards */
class CADIDisassembler;

/* Datatypes */

/*!
 * \brief   SystemC eXport support: Debug target implementation, including
 *          registers, memory spaces and parameters.
 *
 * \note    This version of the class implements \c eslapi::CADI interface.
 */
class scx_cadi_target:
    public virtual scx_debug_target_if,
    protected virtual eslapi::CADI,
    protected virtual scx::scx_simcallback_if {

    /* Construction */
    public:
        scx_cadi_target(const std::string & target,
                        const std::string & instance,
                        const std::string & type,
                        const std::string & version = SYSGEN_PVMODELLIB_VER_STR "." SYSGEN_PVMODELLIB_BUILD_STR);
        virtual
        ~scx_cadi_target();

    /* Accessors - Debug target base */
    public:

        /*!
         * \brief   Return this debug target name.
         * \see     eslapi::CADITargetFeatures_t
         */
        virtual std::string
        get_target_name() const SG_OVERRIDE {
            return m_features.targetName;
        }

        /*!
         * \brief   Return this CADI instance name.
         * \see     eslapi::CADITargetInfo_t
         */
        virtual std::string
        get_instance_name() const SG_OVERRIDE {
            return m_info.instanceName;
        }

        /*!
         * \brief   Return this CADI component type.
         * \see     eslapi::CADITargetInfo_t
         */
        virtual std::string
        get_component_type() const SG_OVERRIDE {
            return m_info.componentType;
        }

        /*!
         * \brief   Return this CADI target version.
         * \see     eslapi::CADITargetFeatures_t
         */
        virtual std::string
        get_target_version() const SG_OVERRIDE {
            return m_features.targetVersion;
        }

        /*!
         * \brief   Return a pointer to underlying CADI instance.
         */
        virtual eslapi::CADI *
        get_cadi() const SG_OVERRIDE {
            return dynamic_cast<eslapi::CADI *>(const_cast<scx_cadi_target *>(this));
        }

        /*!
         * \brief   Register new eslapi::CAInterface with this CADI target.
         */
        virtual void
        register_interface(eslapi::if_name_t if_name,
                           eslapi::if_rev_t if_rev,
                           eslapi::CAInterface * if_ptr) SG_OVERRIDE;

    /* CADI target - Registers */
    public:

        /*!
         * \brief   Set new scx_debug_registers_if for use by this CADI target.
         */
        virtual void
        set_registers_if(scx_debug_registers_if * registers_if) SG_OVERRIDE;

        /*!
         * \brief   Add a register group to this CADI target.
         */
        virtual uint32_t
        add_group(const std::string & name, const std::string & description = "") SG_OVERRIDE;

        /*!
         * \brief   Return current register group ID, or 0 when this CADI target has no register.
         */
        virtual uint32_t
        get_curr_group_id() const SG_OVERRIDE {
            return ((m_curr_group == NULL)? 0: m_curr_group->groupID);
        }

        /*!
         * \brief   Add a register to this CADI target.
         */
        virtual uint32_t
        add_register(const std::string & name,
                     const std::string & description,
                     uint32_t bits_wide,
                     scx_debug_access_type attr = SCX_DEBUG_ACCESS_READ_WRITE) SG_OVERRIDE;

        /*!
         * \brief   Return current register number, or 0 when this CADI target has no register.
         */
        virtual uint32_t
        get_curr_register_number() const SG_OVERRIDE {
            return ((m_curr_reg == NULL)? 0: m_curr_reg->regNumber);
        }

        /*!
         * \brief   Specify given register is of compound type, that is, contains subregisters or bitfields.
         */
        virtual void
        register_is_compound(uint32_t num) SG_OVERRIDE;

        /*!
         * \brief   Add a bitfield to given \a parent register to this CADI target.
         */
        virtual uint32_t
        add_bitfield(uint32_t parent,
                     const std::string & name,
                     const std::string & description,
                     uint32_t offset,
                     uint32_t bits_wide) SG_OVERRIDE;
        /*!
         * \brief   Specify symbolic values for given \a register.
         */
        virtual void
        register_has_symbols(uint32_t num,
                             uint32_t num_symbols,
                             const char * * symbols) SG_OVERRIDE;

        /*!
         * \brief   Specify given register is of program counter, so can be used for disassembly display.
         */
        virtual void
        register_is_pc(uint32_t num) SG_OVERRIDE;

        /*!
         * \brief   Specify given register is stack pointer.
         */
        virtual void
        register_is_sp(uint32_t num) SG_OVERRIDE;

        /*!
         * \brief   Specify given register is link register.
         */
        virtual void
        register_is_lr(uint32_t num) SG_OVERRIDE;

        /*!
         * \brief   Specify given register is status register.
         */
        virtual void
        register_is_status(uint32_t num) SG_OVERRIDE;

        /*!
         * \brief   Add specified register to specified group.
         */
        virtual void
        register_is_in_group(uint32_t reg, uint32_t group) SG_OVERRIDE;

        /*!
         * \brief   Perform checks related to register breakpoint when given register is accessed.
         */
        virtual bool
        register_bpt_check(uint32_t num, bool read) SG_OVERRIDE;

        /*!
         * \brief   Return whether given register has breakpoint(s).
         */
        virtual bool
        has_register_bpt(uint32_t num) SG_OVERRIDE;

    /* CADI target - Memory spaces */
    public:

        /*!
         * \brief   Set new scx_debug_memory_if for use by this CADI target.
         */
        virtual void
        set_memory_if(scx_debug_memory_if * memory_if) SG_OVERRIDE;

        /*!
         * \brief   Add memory space to this CADI target.
         */
        virtual uint32_t
        add_space(const std::string & name,
                  const std::string & description = "",
                  uint32_t bits_per_mau = 8,
                  const uint64_t & max = uint64_t(~0),
                  const uint64_t & min = 0,
                  bool progmem = false,
                  bool virtmem = false,
                  unsigned char endianness = 0) SG_OVERRIDE;

        /*!
         * \brief   Add memory block to this CADI target.
         */
        virtual uint32_t
        add_block(const std::string & name,
                  const std::string & description = "",
                  const uint64_t & start = 0,
                  const uint64_t & end = 0,
                  scx_debug_access_type attr = SCX_DEBUG_ACCESS_READ_WRITE) SG_OVERRIDE;

        /*!
         * \brief   Perform checks related to memory breakpoint when given address is accessed.
         */
        virtual bool
        memory_bpt_check(uint32_t space, const uint64_t & addr, bool read) SG_OVERRIDE;

        /*!
         * \brief   Return whether given memory address has breakpoint(s).
         */
        virtual bool
        has_memory_bpt(uint32_t space, const uint64_t & addr) SG_OVERRIDE;

    /* CADI target - Parameters */
    public:

        /*!
         * \brief   Set new scx_debug_params_if for use by this CADI target.
         */
        virtual void
        set_params_if(scx_debug_params_if * params_if) SG_OVERRIDE;

        /*!
         * \brief   Add parameter to this CADI target.
         */
        virtual uint32_t
        add_param(const std::string & name,
                  const std::string & description,
                  scx_debug_param_value_type type = SCX_DEBUG_PARAM_INVALID,
                  bool runtime = false,
                  int64_t min = 0,
                  int64_t max = 0,
                  int64_t value = 0,
                  const char * str = "") SG_OVERRIDE;

    /* CADI target - Control */
    public:

        /*!
         * \brief   Set new scx_debug_control_if for use by this CADI target.
         */
        virtual void
        set_control_if(scx_debug_control_if * control_if) SG_OVERRIDE;

        /*!
         * \brief   Perform checks related to program breakpoint when given
         *          address is accessed.
         */
        virtual bool
        program_bpt_check(uint32_t space, const uint64_t & addr) SG_OVERRIDE;

        /*!
         * \brief   Return whether given program address has breakpoint(s).
         */
        virtual bool
        has_program_bpt(uint32_t space, const uint64_t & addr) SG_OVERRIDE;

        /*!
         * \brief   Called by a model when it has reached a stopped state.
         */
        virtual void
        stop_acknowledge(bool success) SG_OVERRIDE;

        /*!
         * \brief   Called by a model when it has executed specified number of steps.
         */
        virtual void
        step_acknowledge(bool success) SG_OVERRIDE;

        /*!
         * \brief   Called by a model when it finishes its reset sequence.
         */
        virtual void
        reset_done() SG_OVERRIDE;

    /* CADI target - Disassembly */
    public:

        /*!
         * \brief   Set new scx_debug_disass_if for use by this CADI target.
         */
        virtual void
        set_disass_if(scx_debug_disass_if * disass_if) SG_OVERRIDE;

        /*!
         * \brief   Add mode to this debug target.
         */
        virtual void
        add_mode(uint32_t mode,
                 const std::string & name) SG_OVERRIDE;

    /* Implementation - CAInterface */
    private:

        /*
         * Static helper method to return the interface name for the
         * eslapi::CAInterface.
         */
        static eslapi::if_name_t
        IFNAME() {
            return "eslapi.CADI2";
        }

        /*
         * Static helper method to return the current interface revision
         * for eslapi::CAInterface.
         */
        static eslapi::if_rev_t
        IFREVISION() {
            return 0;
        }

        /*
         * Ask the component for a handle to a different interface,
         * if the component implements that interface.
         * The return value is NULL if the component doesn't implement the specified
         * interface; otherwise it is a pointer to an instance of
         * eslapi::CAInterface that can be cast to the appropriate interface type.
         */
        virtual eslapi::CAInterface *
        ObtainInterface(eslapi::if_name_t if_name,
                        eslapi::if_rev_t min_rev,
                        eslapi::if_rev_t * actual_rev) SG_OVERRIDE;

    /* Implementation - CADI base */
    private:

        /* Call-back object */
        struct callback_obj {
            eslapi::CADICallbackObj * m_obj;
            char m_enable[eslapi::CADI_CB_Count];
        };

        /* Call-back object list type */
        typedef std::list<callback_obj> callback_obj_list_type;

        /* Member variables */
        eslapi::CADITargetInfo_t m_info;
        eslapi::CADITargetFeatures_t m_features;
        eslapi::CAInterfaceRegistry m_ifs_registry;
        callback_obj_list_type m_callbacks;
            /* List of registered CADI call-back objects */
        TxMutex m_callbacks_mutex;
            /* Mutex for list of registered CADI call-back object */
        typedef std::vector<std::string> reset_levels_type;
        reset_levels_type m_reset_levels;
        typedef std::vector<eslapi::CADIExecMode_t> exec_modes_type;
        exec_modes_type m_exec_modes;
        std::string m_error_message;
        scx_debug_registers_if * m_registers_if;
            /* Model debug resources: - registers */
        scx_debug_params_if * m_params_if;
            /*                        - parameters */
        scx_debug_memory_if * m_memory_if;
            /*                        - memory */
        scx_debug_control_if * m_control_if;
            /*                        - execution control */
        scx_debug_disass_if * m_disass_if;
            /*                        - disassembly */

        /* CADI */

        /*
         * When an error is detected,
         * this routine is called to get the error message.
         */
        virtual eslapi::CADIReturn_t
        CADIXfaceGetError(uint32_t maxMessageLength,
                          uint32_t * actualMessageLength,
                          char * errorMessage) SG_OVERRIDE;

        /*
         * A target's debugger shall call this function when it attaches to a
         * target.
         * This function is typically called once per target.
         * The debugger is free to call it more often, if needed.
         * This call is used to determine the features supported by the target.
         */
        virtual eslapi::CADIReturn_t
        CADIXfaceGetFeatures(eslapi::CADITargetFeatures_t * features) SG_OVERRIDE;

        /*
         * Return the targetInfo for this model. To be filled by the model.
         */
        virtual eslapi::CADIReturn_t
        CADIGetTargetInfo(eslapi::CADITargetInfo_t * targetInfo) SG_OVERRIDE;

        /*
         * A target's debugger should call this to establish with the target a
         * callback object that will handle asynchronous information from the
         * target.
         * The callback routines should not make calls to the target.
         * It is possible for a debugger to receive a callback while in the
         * middle of a call (for example, receiving a modeChange callback from
         * within a eslapi::CADIExecStop call).
         */
        virtual eslapi::CADIReturn_t
        CADIXfaceAddCallback(eslapi::CADICallbackObj *callbackObj,
                             char enable[eslapi::CADI_CB_Count]) SG_OVERRIDE;

        /*
         * A debugger should call this to remove any callback objects it might
         * have added.
         */
        virtual eslapi::CADIReturn_t
        CADIXfaceRemoveCallback(eslapi::CADICallbackObj *callbackObj) SG_OVERRIDE;

        /*
         * Helper to define CADI proxy-able methods.
         */
        #define SCX_PROXY_METHOD_DECL(fun, ...) \
            virtual eslapi::CADIReturn_t \
            fun(__VA_ARGS__) SG_OVERRIDE; \
            eslapi::CADIReturn_t \
            proxy_ ## fun(__VA_ARGS__)

    /* Implementation - Execution control */
    private:

        /*
         * Many processors have more than just a run state and a breakpoint
         * state.
         * This call allows the debugger to determine what these states are.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetModes(uint32_t startModeIndex,
                         uint32_t desiredNumOfModes,
                         uint32_t * actualNumOfModes,
                         eslapi::CADIExecMode_t * execModes) SG_OVERRIDE;

        /*
         * Many targets have more than just one reset level.
         * This call allows the debugger to determine what these levels are.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetResetLevels(uint32_t startResetLevelIndex,
                               uint32_t desiredNumOfResetLevels,
                               uint32_t * actualNumOfResetLevels,
                               eslapi::CADIResetLevel_t * resetLevels) SG_OVERRIDE;

        /*
         * This is a way to set to the target's execution modes.
         * This call will return immediately, possibly before the target
         * execution mode has been reached.
         */
        virtual eslapi::CADIReturn_t
        CADIExecSetMode(uint32_t execMode) SG_OVERRIDE;

        /*
         * This call allows the debugger to determine the execution state of
         * the target.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetMode(uint32_t *execMode) SG_OVERRIDE;

        /*
         * This function returns immediately and a separate notification
         * informs the debugger that the execution state has changed.
         * Typically this call would result in the 'modeChange' callback (if
         * enabled) for eslapi::CADI_EXECMODE_Run followed by
         * eslapi::CADI_EXECMODE_Stop.
         *
         * The same callback is possible with eslapi::CADI_EXECMODE_Bpt in case
         * of the target hitting a breakpoint, or eslapi::CADI_EXECMODE_Error
         * in the case of target errors during it's execution.
         */
        virtual eslapi::CADIReturn_t
        CADIExecSingleStep(uint32_t instructionCount,
                           int8_t stepCycle,
                           int8_t stepOver) SG_OVERRIDE;

        /*
         * Upon receipt of this call, the target will reset its execution
         * related internal state, as well as reset its registers and memories
         * to a predefined state, but it will not change breakpoints or
         * callbacks.
         */
        virtual eslapi::CADIReturn_t
        CADIExecReset(uint32_t resetLevel) SG_OVERRIDE;

        /*
         * This function returns immediately and a separate notification informs the debugger when
         * execution state has changed.
         */
        virtual eslapi::CADIReturn_t
        CADIExecContinue(void) SG_OVERRIDE;

        /*
         * This causes the target's execution to stop.
         */
        virtual eslapi::CADIReturn_t
        CADIExecStop(void) SG_OVERRIDE;

        /*
         * Get currrent instruction count of the specific target this interface is connected to.
         */
        SCX_PROXY_METHOD_DECL(CADIGetInstructionCount, uint64_t & instructionCount);

        /*
         * Get cycle count.
         */
        SCX_PROXY_METHOD_DECL(CADIGetCycleCount, uint64_t & cycleCount, bool systemCycles);

    /* Implementation - Load application */
    private:
        /* Typedefs */
        typedef std::vector<std::string> apps_list_type;

        /* Member variables */
        apps_list_type m_apps_to_load; /* list of apps pending for loading */
        apps_list_type m_apps_loaded; /* list of apps which were loaded successfully */
        bool m_app_load_pending;
        bool m_reset_done;

        virtual eslapi::CADIReturn_t
        CADIExecLoadApplication(const char *filename,
                                bool loadData,
                                bool verbose,
                                const char *parameters) SG_OVERRIDE;

        /*
         * Remove the symbol information of a specific image which has been loaded previously.
         */
        virtual eslapi::CADIReturn_t
        CADIExecUnloadApplication(const char *filename) SG_OVERRIDE;

        /*
         * Get list of image filenames currently loaded in target.
         * This list shall be preserved until calling eslapi::CADIExecReset(0)
         * or until a custom CADI reset is called with eslapi::CADIExecReset()
         * that clears the program memory.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetLoadedApplications(uint32_t startIndex,
                                      uint32_t desiredNumberOfApplications,
                                      uint32_t *actualNumberOfApplicationsReturnedOut,
                                      char *filenamesOut,
                                      uint32_t filenameLength,
                                      char *parametersOut,
                                      uint32_t parametersLength) SG_OVERRIDE;

    /* Implementation - Registers */
    private:

        /* Typedefs */
        typedef std::unordered_map<uint32_t, eslapi::CADIRegGroup_t *> group_map_type;
        typedef std::list<eslapi::CADIRegGroup_t *> group_list_type;
        typedef std::unordered_map<uint32_t, eslapi::CADIRegInfo_t *> reg_info_map_type;
        typedef std::list<eslapi::CADIRegInfo_t *> reg_info_list_type;
        typedef std::unordered_map<uint32_t, reg_info_list_type> group_reg_info_list_map_type;
        typedef std::list<uint32_t> reg_component_list_type;
        typedef std::unordered_map<uint32_t, reg_component_list_type> compound_reg_component_list_map_type;
        typedef std::unordered_map<uint32_t, uint32_t> reg_parent_map_type;

        /* Member variables */
        group_list_type m_groups;         /* List of groups */
        group_map_type m_group_map;       /* Group map */
        eslapi::CADIRegGroup_t * m_curr_group;
        reg_info_list_type m_reg_infos;         /* List of register information */
        reg_info_map_type m_reg_info_map;       /* Register information map */
        eslapi::CADIRegInfo_t * m_curr_reg;
        group_reg_info_list_map_type m_group_reg_infos_map;
        compound_reg_component_list_map_type m_reg_compounds_map;
        reg_parent_map_type m_reg_parent_map;
        bool m_pc_is_virtual;
        eslapi::CADIRegGroup_t * m_pseudo_group;
        eslapi::CADIRegInfo_t * m_extended_target_features_reg;
        eslapi::CADIRegInfo_t * m_pcmemspace_reg;
        eslapi::CADIRegInfo_t * m_haltcore_reg;
        int m_haltcore_value;
        bool m_haltcore_supported;
            /* Whether to expose HALT_CORE register or not. Disabled by default. */
        std::string m_extended_target_features;
        TxMutex m_reg_read_write_mutex;
            /* Mutex for reading/writing registers */

        /* CADI */

        /*
         * This call is used to retrieve register groups from the target.
         */
        virtual eslapi::CADIReturn_t
        CADIRegGetGroups(uint32_t groupIndex,
                         uint32_t desiredNumOfRegGroups,
                         uint32_t * actualNumOfRegGroups,
                         eslapi::CADIRegGroup_t * reg) SG_OVERRIDE;

        /*
         * A target's debugger should call this once after connecting to the target.
         * All registers should be reported even if they are a piece of a compound register.
         * All register numbers must be unique, even with respect to registers in other groups.
         */
        virtual eslapi::CADIReturn_t
        CADIRegGetMap(uint32_t groupID,
                      uint32_t startRegisterIndex,
                      uint32_t desiredNumOfRegisters,
                      uint32_t * actualNumOfRegisters,
                      eslapi::CADIRegInfo_t * reg) SG_OVERRIDE;

        /*
         * This call gets the information about a compound register (as reported by a CADIRegGetMap call).
         * A model _may_ expose a set of registers in such a way that they form a Directed Acyclic Graph
         * (DAG), it _must not_ form a cyclic graph.  A debugger _must_ be able to cope with an acyclic
         * graph, it _may_ detect a cyclic graph and report an error.
         */
        virtual eslapi::CADIReturn_t
        CADIRegGetCompound(uint32_t reg,
                           uint32_t componentIndex,
                           uint32_t desiredNumOfComponents,
                           uint32_t *actualNumOfcomponents,
                           uint32_t *components) SG_OVERRIDE;

        /*
         * This function writes a register.
         */
        SCX_PROXY_METHOD_DECL(CADIRegWrite,
                              uint32_t regCount,
                              eslapi::CADIReg_t * reg,
                              uint32_t * numOfRegsWritten,
                              uint8_t doSideEffects);

        /*
         * This function reads the values of the target's registers.
         */
        SCX_PROXY_METHOD_DECL(CADIRegRead,
                              uint32_t regCount,
                              eslapi::CADIReg_t * reg,
                              uint32_t * numRegsRead,
                              uint8_t doSideEffects);

        /*
         * This methood returns the PC of the instruction that is executed next
         * from an ISA perspective.
         */
        virtual uint64_t
        CADIGetPC() SG_OVERRIDE;
        virtual uint64_t
        CADIGetPC(bool* is_virtual) SG_OVERRIDE;

    /* Implementation - Memory */
    private:

        /* Typedefs */
        typedef std::unordered_map<uint32_t, eslapi::CADIMemSpaceInfo_t *> space_info_map_type;
        typedef std::list<eslapi::CADIMemSpaceInfo_t *> space_info_list_type;
        typedef std::list<eslapi::CADIMemBlockInfo_t *> block_info_list_type;
        typedef std::map<uint64_t, eslapi::CADIMemBlockInfo_t *> addr_block_info_map_type;
        typedef std::unordered_map<uint32_t, addr_block_info_map_type> space_addr_block_info_map_type;

        /* Member variables */
        space_info_list_type m_space_infos;     /* List of spaces */
        space_info_map_type m_space_info_map;   /* Space map */
        eslapi::CADIMemSpaceInfo_t * m_curr_space;
        block_info_list_type m_block_infos;     /* List of block information */
        eslapi::CADIMemBlockInfo_t * m_curr_block;
        space_addr_block_info_map_type m_space_addr_block_info_map;
        TxMutex m_memory_read_write_mutex;
            /* Mutex for reading/writting memory*/

        /* CADI */

        /*
         * A target's debugger should call this once after connecting to the target but before accessing
         * memory. This gets the number of independent address spaces available on the target. Different
         * memory spaces should be used to separate distinct memory areas with overlapping address values
         * (like program and data memory in a Harvard architecture).
         */
        virtual eslapi::CADIReturn_t
        CADIMemGetSpaces(uint32_t startMemSpaceIndex,
                         uint32_t desiredNumOfMemSpaces,
                         uint32_t *actualNumOfMemSpaces,
                         eslapi::CADIMemSpaceInfo_t *memSpaces) SG_OVERRIDE;

        /*
         * A target's debugger should call this once for each memory space specified by eslapi::CADIMemGetSpaces
         * before accessing memory in that space. This should return the layout of the memory in a single
         * address space. No two blocks with the same parent should overlap. This call returns existing
         * memory blocks only. The caller can assume that any memory not in a block is a hole or invalid
         * memory.
         */
        virtual eslapi::CADIReturn_t
        CADIMemGetBlocks(uint32_t memorySpace,
                         uint32_t memBlockIndex,
                         uint32_t desiredNumOfMemBlocks,
                         uint32_t *actualNumOfMemBlocks,
                         eslapi::CADIMemBlockInfo_t *memBlocks) SG_OVERRIDE;

        /*
         * This causes the target's memory to get new values. If the write spans a hole in memory,
         * the target should stop writing at the beginning of the hole and reflect the discrepancy
         * in the numUnitsWritten.
         * Writes to memory should never trigger breakpoints directly themselves. However, it is
         * target implementation-defined whether it recognises a modify-triggered breakpoint when
         * the simulation starts again.
         */
        SCX_PROXY_METHOD_DECL(CADIMemWrite,
                              eslapi::CADIAddrComplete_t startAddress,
                              uint32_t unitsToWrite,
                              uint32_t unitSizeInBytes,
                              const uint8_t *data,
                              uint32_t *actualNumOfUnitsWritten,
                              uint8_t doSideEffects);

        /*
         * This causes memory to be read from the target. If the read spans a hole in memory, the target should
         * stop reading at the beginning of the hole and reflect the discrepancy in the numUnitsRead.
         */
        SCX_PROXY_METHOD_DECL(CADIMemRead,
                              eslapi::CADIAddrComplete_t startAddress,
                              uint32_t unitsToRead,
                              uint32_t unitSizeInBytes,
                              uint8_t *data,
                              uint32_t *actualNumOfUnitsRead,
                              uint8_t doSideEffects);

    /* Implementation - Breakpoints */
    private:

        /* Typedefs */
        typedef std::list<eslapi::CADIBptDescription_t *> bpt_info_list_type;
        typedef std::unordered_map<eslapi::CADIBptNumber_t, eslapi::CADIBptDescription_t *> bpt_info_map_type;
        typedef std::unordered_map<uint32_t, bpt_info_list_type> reg_bpts_map_type;
        typedef std::unordered_map<eslapi::CADIAddr_t, bpt_info_list_type, hash> mem_bpts_map_type;
        typedef std::list<eslapi::CADIBptNumber_t> bpt_num_list_type;

        /* Member variables */
        eslapi::CADIBptNumber_t m_next_bpt_number;
        bpt_info_list_type m_bpt_infos;         /* List of breakpoint information */
        bpt_info_map_type m_bpt_info_map;       /* Breakpoint information map */
        reg_bpts_map_type m_reg_bpts_map;       /* Register-breakpoints map */
        mem_bpts_map_type m_mem_bpts_map;       /* Memory-breakpoints map */
        mem_bpts_map_type m_prog_bpts_map;      /* Program-breakpoints map */
        TxMutex m_bpt_mutex;
            /* Mutex for accessing breakpoints */

        /* CADI */

        /*
         * If the debugger attaches to a target that already has breakpoints set,
         * this is a way for the debugger to find out what they are.
         */
        virtual eslapi::CADIReturn_t
        CADIBptGetList(uint32_t startBptIndex,
                       uint32_t desiredNumOfBpts,
                       uint32_t * actualNumOfBpts,
                       eslapi::CADIBptDescription_t * breakpoints) SG_OVERRIDE;

        /*
         * Read a breakpoint for a specific breakpoint ID.
         * This can be used to retrieve the current ignoreCount of a specific breakpoint.
         */
        virtual eslapi::CADIReturn_t
        CADIBptRead(eslapi::CADIBptNumber_t breakpointId,
                    eslapi::CADIBptRequest_t * requestOut) SG_OVERRIDE;

        /*
         * This sets a, possibly complex, breakpoint in the target.
         * The breakpoint will be initially enabled or disabled depending on the value of 'request->enabled'.
         */
        SCX_PROXY_METHOD_DECL(CADIBptSet,
                              eslapi::CADIBptRequest_t * request,
                              eslapi::CADIBptNumber_t * breakpoint);

        /*
         * This function removes a breakpoint from the target.
         */
        SCX_PROXY_METHOD_DECL(CADIBptClear, eslapi::CADIBptNumber_t breakpointId);

        /*
         * This function enables or disables a breakpoint on the target.
         * This only makes sense when the target supports enabling and disabling of hardware breakpoints.
         * Otherwise, this sort of breakpoint management must be done in the host.
         */
        virtual eslapi::CADIReturn_t
        CADIBptConfigure(eslapi::CADIBptNumber_t breakpointId,
                         eslapi::CADIBptConfigure_t configuration) SG_OVERRIDE;

    /* Implementation - Parameters */
    private:

        /* Typedefs */
        typedef std::unordered_map<uint32_t, eslapi::CADIParameterInfo_t *> param_info_map_type;
        typedef std::list<eslapi::CADIParameterInfo_t *> param_info_list_type;
        typedef std::unordered_map<std::string, uint32_t> param_id_map_type;

        /* Member variables */
        param_info_list_type m_param_infos;         /* List of parameter information */
        param_info_map_type m_param_info_map;       /* Parameter information map */
        param_id_map_type m_param_id_map;           /* Parameter ID map */
        TxMutex m_param_get_set_mutex;
            /* Mutex for setting/getting parameters */

        /* CADI */

        /*
         * Get list of supported parameters and parameter information.
         */
        virtual eslapi::CADIReturn_t
        CADIGetParameters(uint32_t startIndex,
                          uint32_t desiredNumOfParams,
                          uint32_t * actualNumOfParams,
                          eslapi::CADIParameterInfo_t * params) SG_OVERRIDE;

        /*
         * Get parameter info for a specific parameter name.
         */
        virtual eslapi::CADIReturn_t
        CADIGetParameterInfo(const char * parameterName,
                             eslapi::CADIParameterInfo_t * param) SG_OVERRIDE;

        /*
         * Get current parameter values.
         */
        SCX_PROXY_METHOD_DECL(CADIGetParameterValues,
                              uint32_t parameterCount,
                              uint32_t * actualNumOfParamsRead,
                              eslapi::CADIParameterValue_t * paramValuesOut);

        /*
         * Set current parameter values.
         */
        SCX_PROXY_METHOD_DECL(CADISetParameters,
                              uint32_t parameterCount,
                              eslapi::CADIParameterValue_t * parameters,
                              eslapi::CADIFactoryErrorMessage_t * error);

    /* Implementation - Disassembly */

        /* Typedefs */
        typedef std::vector<std::pair<uint32_t, std::string> > mode_map_type;

        /* Member variables */
        mode_map_type m_modes;
        CADIDisassembler * m_disassembler;

        /* CADI */
        /*
         * This call is used to retrieve the disassembler for a target.
         */
        virtual eslapi::CADIDisassembler *
        CADIGetDisassembler(void) SG_OVERRIDE;

    /* Implementation - Unsupported */
    private:

        /*
         * Targets will sometimes have specialized commands.
         * This command is designed to allow the debugger to pass a string
         * containing one of these commands to a target.
         * The target is free to interpret this command in any way that it
         * wishes.
         */
        virtual eslapi::CADIReturn_t
        CADIXfaceBypass(uint32_t commandLength, const char *command,
                        uint32_t maxResponseLength, char *response) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This gets the list of the target's exception vectors.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetExceptions(uint32_t startExceptionIndex,
                              uint32_t desiredNumOfExceptions,
                              uint32_t *actualNumOfExceptions,
                              eslapi::CADIException_t *exceptions) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This method raises an exception.
         */
        virtual eslapi::CADIReturn_t
        CADIExecAssertException(uint32_t exception,
                                eslapi::CADIExceptionAction_t action) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This is used to expose the pipe simulated inside of a cycle-accurate simulation.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetPipeStages(uint32_t startPipeStageIndex,
                              uint32_t desiredNumOfPipeStages,
                              uint32_t *actualNumOfPipeStages,
                              eslapi::CADIPipeStage_t *pipeStages) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This is used to expose the fields of the pipe simulated inside of a cycle-accurate simula tion.
         */
        virtual eslapi::CADIReturn_t
        CADIExecGetPipeStageFields(uint32_t startPipeStageFieldIndex,
                                   uint32_t desiredNumOfPipeStageFields,
                                   uint32_t *actualNumOfPipeStageFields,
                                   eslapi::CADIPipeStageField_t *pipeStageFields) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * Get commmited PCs. Returns the list of all commited PCs in the current cycle (e.g., for multi-issue
         * processors)
         * Implementing this method is not required for single-issue processors and may result in simulation slowdowns
         * in certain situations.
         */
        virtual eslapi::CADIReturn_t
        CADIGetCommitedPCs(int startIndex,
                           int desiredCount,
                           int * actualCount,
                           uint64_t * pcs) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * The debugger calls this when it wants to get the list of active overlays. This would typically be done
         * when a breakpoint is hit.
         * <p><b>Language Domain Requirement.</b> When overlays are implemented, some overlay ID needs to be
         * stored in the symbol table and in the target software. The symbol table should store the starting
         * address, memory space, and byte count for each overlay. This will allow the ID to be sent to the host
         * when an overlay happens and allow the host to understand what it means.
         */
        virtual eslapi::CADIReturn_t
        CADIMemGetOverlays(uint32_t activeOverlayIndex,
                           uint32_t desiredNumOfActiveOverlays,
                           uint32_t *actualNumOfActiveOverlays,
                           eslapi::CADIOverlayId_t *overlays) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This call is used to get the cache information for a memory space.
         */
        virtual eslapi::CADIReturn_t
        CADIGetCacheInfo(uint32_t memSpaceID,
                         eslapi::CADICacheInfo_t* cacheInfo) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This call is used to perform a cache read.
         */
        virtual eslapi::CADIReturn_t
        CADICacheRead(eslapi::CADIAddr_t addr,
                      uint32_t linesToRead,
                      uint8_t* data,
                      uint8_t* tags,
                      bool* is_dirty,
                      bool* is_valid,
                      uint32_t* numLinesRead,
                      bool doSideEffects) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This call is used to perform a cache write.
         */
        virtual eslapi::CADIReturn_t
        CADICacheWrite(eslapi::CADIAddr_t addr,
                       uint32_t linesToWrite,
                       const uint8_t* data,
                       const uint8_t* tags,
                       const bool* is_dirty,
                       const bool* is_valid,
                       uint32_t* numLinesWritten,
                       bool doSideEffects) SG_OVERRIDE {
            return eslapi::CADI_STATUS_CmdNotSupported;
        }

        /*
         * This function translates a virtual address to a physical one.
         */
        virtual eslapi::CADIAddrComplete_t
        VirtualToPhysical(eslapi::CADIAddrComplete_t vaddr) SG_OVERRIDE {
            eslapi::CADIAddrComplete_t paddr;

            return paddr;
        }

        /*
         * This function translates a physical address to a virtual one.
         */
        virtual eslapi::CADIAddrComplete_t
        PhysicalToVirtual(eslapi::CADIAddrComplete_t paddr) SG_OVERRIDE {
            eslapi::CADIAddrComplete_t vaddr;

            return vaddr;
        }

    /* Implementation - Simulation control */
    private:

        /*
         * Notification that simulation is running.
         */
        virtual void
        notify_running() SG_OVERRIDE;

        /*
         * Notification that simulation has stopped.
         */
        virtual void
        notify_stopped() SG_OVERRIDE;

        /*
         * Notification that simulation is at debuggable point.
         */
        virtual void
        notify_debuggable() SG_OVERRIDE;

        /*
         * Notification that simulation is idle.
         */
        virtual void
        notify_idle() SG_OVERRIDE;

    /* Implementation - Helper functions */

        /*
         * Add given register to extended target features.
         */
        void
        add_to_extended_target_features(uint32_t num, const char * token);

        /*
         * Notify simulation mode (e.g. running, stopped...) has changed.
         */
        void
        notify_mode_change(eslapi::CADI_EXECMODE_t mode,
                           eslapi::CADIBptNumber_t bp = 0);

        /*
         * Notify debuggers to update their display after state of target changed
         * spontaneously, when simulation is stopped.
         */
        void
        notify_refresh(uint32_t reason);

        /*
         * Perform checks related to breakpoint when resource is accessed.
         */
        bool
        bpt_check(const bpt_info_list_type & bpts, bool read);

        /*
         * Return whether any given breakpoints is enabled.
         */
        bool
        has_bpt(const bpt_info_list_type & bpts);

        /*
         * Return memory block contaning given address, or NULL if not found.
         */
        eslapi::CADIMemBlockInfo_t *
        find_block(uint32_t space, const uint64_t & addr);

        /*
         * Report an error.
         */
        void
        report_error(const char * fun, const char * fmt, ...) SG_FORMATSTR(printf, 3,4);

        /*
         * Load any apps that are lined up, waiting for reset_done.
         */
        bool load_applications();
};

}   /* namespace scx */

#endif  /* defined(ARM_SCX_CADI_TARGET_H) */
