/*
 * SCX: scx_debug_control_if.h - SystemC eXport support: Debug excution control interface.
 *
 * Copyright 2016-2017 ARM Limited.
 * All rights reserved.
 *
 * WARNING! This file is not part of a stable public API and
 *          is subject to change without notice.
 */

#ifndef ARM_SCX_DEBUG_CONTROL_IF_H
#define ARM_SCX_DEBUG_CONTROL_IF_H

/*!
 * \file        scx_debug_control_if.h
 * \brief       SystemC eXport support: Debug excution control interface.
 * \date        Copyright 2016-2017 ARM Limited. All rights reserved.
 *
 * This file specifies interfaces provided by models to debug target
 * for controlling their execution.
 */

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Datatypes */

/*!
 * \brief       Debug excution control interface.
 *
 * This interface specifies operations provided by models to debug target
 * for controlling their execution.
 */
class scx_debug_control_if {

    /* Construction */
    protected:

        /*!
         * \brief   Destructor.
         *
         * This version of the function does not allow destroying instance
         * through this interface.
         */
        virtual
        ~scx_debug_control_if() {
        }

    /* Debug support */
    public:

        /*!
         * \brief   Notify a breakpoint has been set at given program address.
         *
         * The model excution control is expected to call back
         * scx_debug_target::program_pbt_check(), when it detects that program
         * execution has reached given address.
         *
         * \param   space program space identifier
         * \param   addr program address
         * \see     scx_debug_target::program_pbt_check()
         */
        virtual void
        debug_program_bpt_set(uint32_t space, const uint64_t & addr) = 0;

        /*!
         * \brief   Notify a breakpoint has been cleared at given program address.
         *
         * The model execution control can release any resource it might have
         * used for detecting program execution reached given address.
         *
         * \param   space program space identifier
         * \param   addr program address
         */
        virtual void
        debug_program_bpt_clear(uint32_t space, const uint64_t & addr) = 0;

        /*!
         * This causes the model's execution state to change to running state.
         * This function returns immediately without changing global simulation
         * state.
         *
         * \return  \c true when model's execution mode succesfully changed to
         *          running, \c false otherwise.
         */
        virtual bool
        debug_run() = 0;

        /*!
         * This causes the model's execution to step for specified number of
         * instructions.
         * This function returns immediately without changing global simulation
         * state.
         * The model is then expected to call scx_debug_target::step_acknowledge()
         * once specified number of instructions have been executed.
         *
         * \param   count number of instruction to be executed.
         * \return  \c true when model's execution mode succesfully changed to
         *          stepping, \c false otherwise.
         * \see     scx_debug_target::step_acknowledge()
         */
        virtual bool
        debug_step(uint32_t count) = 0;

        /*!
         * This causes the model's execution to stop at a debug-able point.
         * This function returns immediately without changing global simulation
         * state.
         * The model is then expected to call scx_debug_target::stop_acknowledge()
         * once it's execution mode changed to stopped,
         * unless it returned @c false.
         *
         * \return  \c true when model's execution mode succesfully changed to
         *          stopping, \c false when simulation can stop immediately.
         * \see     scx_debug_target::stop_acknowledge()
         */
        virtual bool
        debug_stop() {
            return false;
        }

        /*!
         * Upon receipt of this call, the model will reset its execution
         * related internal state, as well as reset its parameters to a
         * predefined state, but it will not change callbacks.
         *
         * \param   level unused.
         */
        virtual void
        debug_reset(uint32_t level) = 0;

        /*!
         * \brief   Get current instruction count of this model.
         *
         * \param   instruction_count returned model's instruction count.
         * \return  \c true when model's instruction count succesfully retrieved,
         *          \c false otherwise (e.g. the model does not execute software).
         */
        virtual bool
        debug_get_instruction_count(uint64_t & instruction_count) = 0;

        /*!
         * \brief   Get current cycle count of this model.
         *
         * \param   cycle_count returned model's cycle count
         * \param   system_cycles \c true to retrieve system cycle count,
         *          \c false to retrieve model specific cycle count.
         * \return  \c true when model's cycle count succesfully retrieved,
         *          \c false otherwise (e.g. the model does not execute software).
         *
         * \note    Not all models support cycle count or system cycles. When
         *          not supported, models return either:
         *          - approximation of cycle count, e.g. instruction count
         *          - \c false.
         */
        virtual bool
        debug_get_cycle_count(uint64_t & cycle_count, bool system_cycles) = 0;

        /*!
         * \brief   Get memory space that program counter points to.
         *
         * \see     scx_debug_target_if::add_space()
         */
        virtual uint32_t
        debug_get_pcmemspace() = 0;

        /*!
         * \brief   Set execution state of the target.
         * \param   stay_halted \c true for target to ignore any run/step commands,
         *          \c false to have target advancing normally.
         *
         * This function is called by the CADI target to support \c HALT_CORE register
         * or, in Iris, to implement \c setPersistentExecutionState().
         */
        virtual void
        set_execution_state(bool stay_halted) = 0;

        /*!
         * \brief   Check whether the target is at a debuggable point or not.
         *
         * \return  \c true when model is at a debuggable point,
         *          \c false otherwise.
         */
        virtual bool
        is_debuggable() = 0;

};

}   /* namespace scx */

#endif  /* defined(ARM_SCX_DEBUG_CONTROL_IF_H) */
