/*
 * SCX: scx_debug_memory_if.h - SystemC eXport support: Debug memory resource interface.
 *
 * Copyright 2016 ARM Limited.
 * All rights reserved.
 *
 * WARNING! This file is not part of a stable public API and
 *          is subject to change without notice.
 */

#ifndef ARM_SCX_DEBUG_MEMORY_IF_H
#define ARM_SCX_DEBUG_MEMORY_IF_H

/**
 * \file        scx_debug_memory_if.h
 * \brief       SystemC eXport support: Debug memory resource interface.
 * \date        Copyright 2016 ARM Limited. All rights reserved.
 *
 * This file specifies interfaces provided by models to debug target for accessing
 * their memory resource.
 */

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Datatypes */

/*!
 * \brief       Debug memory resource interface.
 *
 * This interface specifies operations provided by models to debug target for accessing
 * their memory resource.
 */
class scx_debug_memory_if {

    /* Construction */
    protected:

        /*!
         * \brief   Destructor.
         *
         * This version of the function does not allow destroying instance
         * through this interface.
         */
        virtual
        ~scx_debug_memory_if() {
        }

    /* Debug support */
    public:

        /*!
         * \brief   Debug memory read.
         *
         * This function performs a read from a model memory resource,
         * but without any of the delays, waits, event notifications,
         * or side effects associated with regular memory transaction.
         * In other words, this function shall be non-intrusive.
         *
         * The endianness of the data buffer \a datap will be according to the
         * endianness declared in the memory space used for the read.
         *
         * \param   space memory space identifier
         * \param   addr memory address
         * \param   size number of bytes to read
         * \param   datap pointer to data buffer (own by debug target).
         * \return  \c true when memory successfully read, \c false otherwise
         * \see     scx_debug_target_if::add_space(), scx_debug_target_if::add_block()
         */
        virtual bool
        debug_memory_read(uint32_t space,
                          const uint64_t & addr,
                          uint32_t size,
                          uint8_t * datap) = 0;

        /*!
         * \brief   Debug memory write.
         *
         * This function performs a write to a model memory resource,
         * but without any of the delays, waits, event notifications,
         * or side effects associated with regular memory transaction.
         * In other words, this function shall be non-intrusive.
         *
         * The endianness of the data buffer \a datap will be according to the
         * endianness declared in the memory space used for the write.
         *
         * \param   space memory space identifier
         * \param   addr memory address
         * \param   size number of bytes to write
         * \param   datap pointer to data buffer (own by debug target).
         * \return  \c true when memory successfully written, \c false otherwise
         * \see     scx_debug_target_if::add_space(), scx_debug_target_if::add_block()
         */
        virtual bool
        debug_memory_write(uint32_t space,
                           const uint64_t & addr,
                           uint32_t size,
                           const uint8_t * datap) = 0;

        /*!
         * \brief   Notify a breakpoint has been set at given memory address.
         *
         * The model memory resource is expected to invalidate all DMI pointers
         * previously established and containing given memory address.
         * The model memory resource is expected to call back
         * scx_debug_target::memory_pbt_check(), when it detects that memory
         * has been accessed at given address.
         *
         * \param   space memory space identifier
         * \param   addr memory address
         * \see     scx_debug_target::memory_pbt_check()
         */
        virtual void
        debug_memory_bpt_set(uint32_t space, const uint64_t & addr) = 0;

        /*!
         * \brief   Notify a breakpoint has been cleared at given memory address.
         *
         * The model memory resource can establish DMI pointers containing
         * given memory address, providing no other breakpoints are set at
         * given memory address.
         * The model memory resource can stop "watching" given address.
         *
         * \param   space memory space identifier
         * \param   addr memory address
         */
        virtual void
        debug_memory_bpt_clear(uint32_t space, const uint64_t & addr) = 0;
};

}   /* namespace scx */

#endif  /* defined(ARM_SCX_DEBUG_MEMORY_IF_H) */
