/*
 * SCX: scx_debug_registers_if.h - SystemC eXport support: Debug registers resource interface.
 *
 * Copyright 2016 ARM Limited.
 * All rights reserved.
 *
 * WARNING! This file is not part of a stable public API and
 *          is subject to change without notice.
 */

#ifndef ARM_SCX_DEBUG_REGISTERS_IF_H
#define ARM_SCX_DEBUG_REGISTERS_IF_H

/**
 * \file        scx_debug_registers_if.h
 * \brief       SystemC eXport support: Debug registers resource interface.
 * \date        Copyright 2016 ARM Limited. All rights reserved.
 *
 * This file specifies interfaces provided by models to debug target
 * for accessing their registers resource.
 */

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Datatypes */

/*!
 * \brief       Debug registers resource interface.
 *
 * This interface specifies operations provided by models to debug target
 * for accessing their registers resource.
 */
class scx_debug_registers_if {

    /* Construction */
    protected:

        /*!
         * \brief   Destructor.
         *
         * This version of the function does not allow destroying instance
         * through this interface.
         */
        virtual
        ~scx_debug_registers_if() {
        }

    /* Debug support */
    public:

        /*!
         * \brief   Debug register read.
         *
         * This function performs a read from a model register resource,
         * but without any of the delays, waits, event notifications,
         * or side effects associated with regular memory transaction.
         * In other words, this function shall be non-intrusive.
         *
         * The endianness of the data buffer \a datap will be little endian.
         *
         * \param   num register number
         * \param   datap pointer to data buffer (own by debug target).
         * \return  \c true when register successfully read, \c false otherwise
         * \see     scx_debug_target_if::add_register(),
         *          scx_debug_target_if::add_bitfield()
         */
        virtual bool
        debug_register_read(uint32_t num, uint8_t * datap) = 0;

        /*!
         * \brief   Debug register write.
         *
         * This function performs a write to a model register resource,
         * but without any of the delays, waits, event notifications,
         * or side effects associated with regular memory transaction.
         * In other words, this function shall be non-intrusive.
         *
         * The endianness of the data buffer \a datap will be little endian.
         *
         * \param   num register number
         * \param   datap pointer to data buffer (own by debug target).
         * \return  \c true when register successfully written, \c false otherwise
         * \see     scx_debug_target_if::add_register(),
         *          scx_debug_target_if::add_bitfield()
         */
        virtual bool
        debug_register_write(uint32_t num, const uint8_t * datap) = 0;

        /*!
         * \brief   Notify a breakpoint has been set on given register.
         *
         * The model register resource is expected to invalidate all DMI pointers
         * previously established and containing given register.
         * The model register resource is expected to call back
         * scx_debug_target::register_pbt_check(), when it detects that given
         * register has been accessed.
         *
         * \param   num register number
         * \see     scx_debug_target::register_pbt_check()
         */
        virtual void
        debug_register_bpt_set(uint32_t num) = 0;

        /*!
         * \brief   Notify a breakpoint has been cleared on given register.
         *
         * The model register resource can establish DMI pointers containing
         * given register, providing no other breakpoints are set on given
         * register.
         *
         * \param   num register number
         */
        virtual void
        debug_register_bpt_clear(uint32_t num) = 0;
};

}   /* namespace scx */

#endif  /* defined(ARM_SCX_DEBUG_REGISTERS_IF_H) */
