/*
 * SCX: scx_debug_target_if.h - SystemC eXport support: Debug target interface,
 *                              including registers, memory spaces and parameters.
 *
 * Copyright 2016 ARM Limited.
 * All rights reserved.
 *
 * WARNING! This file is not part of a stable public API and
 *          is subject to change without notice.
 */

#ifndef ARM_SCX_DEBUG_TARGET_IF_H
#define ARM_SCX_DEBUG_TARGET_IF_H

/**
 * \file        scx_debug_target_if.h
 * \brief       SystemC eXport support: Debug target interface, including
 *              registers, memory spaces and parameters.
 * \date        Copyright 2016 ARM Limited. All rights reserved.
 *
 * SystemC eXport support debug target interface, including registers, memory
 * spaces and parameters.
 */

/* Includes */
#include <eslapi/CAInterface.h>

#include <string>

/* Forwards */
namespace eslapi {
class CADI;
}

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Forwards */
class scx_debug_registers_if;
class scx_debug_memory_if;
class scx_debug_params_if;
class scx_debug_control_if;
class scx_debug_disass_if;

/* Typedefs */

/*!
 * \brief       Debug resource access values.
 */
enum scx_debug_access_type {
    SCX_DEBUG_ACCESS_READ = 0x1,
    SCX_DEBUG_ACCESS_WRITE = 0x2,
    SCX_DEBUG_ACCESS_READ_WRITE = SCX_DEBUG_ACCESS_READ | SCX_DEBUG_ACCESS_WRITE,
};

/*!
 * \brief       Debug parameter access values.
 */
enum scx_debug_param_value_type
{
    SCX_DEBUG_PARAM_INVALID,
    SCX_DEBUG_PARAM_STRING,
    SCX_DEBUG_PARAM_BOOL,
    SCX_DEBUG_PARAM_INT
};

/* Datatypes */

/**
 * \brief   Debug target interface, including registers, memory spaces and
 *          parameters.
 *
 * This interface specifies operations provided by debug implementation to
 * models, as well as operations models use to populate debug views of their
 * registers, memory spaces, and parameters.
 */
class scx_debug_target_if {

    /* Construction */
    protected:

        /*!
         * \brief   Destructor.
         *
         * This version of the function does not allow destroying instance
         * through this interface.
         */
        virtual
        ~scx_debug_target_if() {
        }

    /* Accessors - Debug target base */
    public:

        /*!
         * \brief   Return this debug target name.
         */
        virtual std::string
        get_target_name() const = 0;

        /*!
         * \brief   Return this debug instance name.
         */
        virtual std::string
        get_instance_name() const = 0;

        /*!
         * \brief   Return this debug component type.
         */
        virtual std::string
        get_component_type() const = 0;

        /*!
         * \brief   Return this debug target version.
         */
        virtual std::string
        get_target_version() const = 0;

        /*!
         * \brief   Return a pointer to underlying CADI instance.
         * \see     eslapi::CADI
         */
        virtual eslapi::CADI *
        get_cadi() const = 0;

        /*!
         * \brief   Register new eslapi::CAInterface with this debug target.
         * \param   if_name name of interface
         * \param   if_rev highest interface revision implemented by \a if_ptr
         * \param   if_ptr pointer to new eslapi::CAInterface to register.
         * \note    There is no need to un-register interfaces at destruction.
         */
        virtual void
        register_interface(eslapi::if_name_t if_name,
                           eslapi::if_rev_t if_rev,
                           eslapi::CAInterface * if_ptr) = 0;

        /*!
         * \brief   Helper function to register new eslapi::CAInterface with
         *          this debug target.
         */
        template <class IF>
        void register_pointer(IF * if_ptr) {
            register_interface(IF::IFNAME(), IF::IFREVISION(), if_ptr);
        }

    /* Debug target - Registers */
    public:

        /*!
         * \brief   Set new scx_debug_registers_if for use by this debug target.
         * \param   registers_if pointer to scx_debug_registers_if to use.
         */
        virtual void
        set_registers_if(scx_debug_registers_if * registers_if) = 0;

        /*!
         * \brief   Add a register group to this debug target.
         * \param   name group name
         * \param   description group description
         * \return  group ID.
         */
        virtual uint32_t
        add_group(const std::string & name,
                  const std::string & description = "") = 0;

        /*!
         * \brief   Return current register group ID.
         *
         * The current group is the group added last.
         *
         * \return  current register group ID, or 0 when this debug target has
         *          no register.
         */
        virtual uint32_t
        get_curr_group_id() const = 0;

        /*!
         * \brief   Add a register to this debug target.
         *
         * This function adds register to current group.
         * This function adds a default group when no current group exists.
         *
         * \param   name register name
         * \param   description register description
         * \param   bits_wide bitwidth of register
         * \param   attr register access attributes.
         * \return  register number.
         */
        virtual uint32_t
        add_register(const std::string & name,
                     const std::string & description,
                     uint32_t bits_wide,
                     scx_debug_access_type attr = SCX_DEBUG_ACCESS_READ_WRITE) = 0;

        /*!
         * \brief   Return current register number.
         *
         * The current register is the register added last.
         *
         * \return  current register number, or 0 when this debug target has no
         *          register.
         */
        virtual uint32_t
        get_curr_register_number() const = 0;

        /*!
         * \brief   Specify given register is of compound type.
         *
         * A compound register is a register that contains subregisters or
         * bitfields, as specified by add_bitfield().
         *
         * \param   num register number.
         * \see     add_bitfield().
         */
        virtual void
        register_is_compound(uint32_t num) = 0;

        /*!
         * \brief   Add a bitfield to given \a parent register.
         *
         * A model \em may expose registers in such a way that they form a
         * Directed Acyclic Graph (DAG),
         * however it \em must \em not form a cyclic graph.
         *
         * \param   parent parent register number
         * \param   name register name
         * \param   description register description
         * \param   offset least significant bit offset relative to bit \c 0 in
         *          \a parent register
         * \param   bits_wide bitwidth of register
         * \return  register number.
         */
        virtual uint32_t
        add_bitfield(uint32_t parent,
                     const std::string & name,
                     const std::string & description,
                     uint32_t offset,
                     uint32_t bits_wide) = 0;

        /*!
         * \brief   Specify symbolic values for given register.
         *
         * Symbolic values are for use by debuggers.
         *
         * \param   num register number
         * \param   num_symbols number of symbolic values
         * \param   symbols array of symbolic values as strings.
         */
        virtual void
        register_has_symbols(uint32_t num,
                             uint32_t num_symbols,
                             const char * * symbols) = 0;

        /*!
         * \brief   Specify given register is program counter.
         *
         * Program counter register can be used for disassembly display by
         * debuggers.
         *
         * \param   num register number.
         */
        virtual void
        register_is_pc(uint32_t num) = 0;

        /*!
         * \brief   Specify given register is stack pointer.
         *
         * Stack pointer register can be used for disassembly display by
         * debuggers.
         *
         * \param   num register number.
         */
        virtual void
        register_is_sp(uint32_t num) = 0;

        /*!
         * \brief   Specify given register is link register.
         *
         * Link register can be used for disassembly display by
         * debuggers.
         *
         * \param   num register number.
         */
        virtual void
        register_is_lr(uint32_t num) = 0;

        /*!
         * \brief   Specify given register is status register.
         *
         * Status register can be used for disassembly display by
         * debuggers.
         *
         * \param   num register number.
         */
        virtual void
        register_is_status(uint32_t num) = 0;

        /*!
         * \brief   Add specified register to specified group.
         *
         * This function is used to add existing register to another group than
         * the current one.
         *
         * \param   reg register number
         * \param   group register group number.
         */
        virtual void
        register_is_in_group(uint32_t reg, uint32_t group) = 0;

        /*!
         * \brief   Perform checks related to register breakpoint when given
         *          register is accessed.
         *
         * When a breakpoint is hit, this function will:
         * - notifies debugers of breakpoints which have been hit
         * - causes simulation to stop.
         *
         * \param   num register number
         * \param   read \c true for read access, \false for write access.
         * \return  \c true when simulation will stop as result of breakpoint
         *          being hit, \c false otherwise.
         */
        virtual bool
        register_bpt_check(uint32_t num, bool read) = 0;

        /*!
         * \brief   Return whether given register has breakpoint(s).
         * \param   num register number
         */
        virtual bool
        has_register_bpt(uint32_t num) = 0;

    /* Debug target - Memory spaces */
    public:

        /*!
         * \brief   Set new scx_debug_memory_if for use by this debug target.
         * \param   memory_if pointer to scx_debug_memory_if to use.
         */
        virtual void
        set_memory_if(scx_debug_memory_if * memory_if) = 0;

        /*!
         * \brief   Add memory space to this debug target.
         * \param   name memory space name
         * \param   description memory space description
         * \param   bits_per_mau Minimum Addressable Unit
         *          (for example, 8 for bytes)
         * \param   max maximum address of this memory space
         * \param   min minimum address of this memory space
         * \param   progmem \c true for program memory
         *          (only one memory space can be program memory)
         * \param   virtmem \c true for virtual memory, \c false for physical
         *          memory
         * \param   endianness \c 0 for variable endianness (architecturally defined),
         *          \c 1 for little-endian, \c 2 for big-endian.
         * \return  memory space ID.
         */
        virtual uint32_t
        add_space(const std::string & name,
                  const std::string & description = "",
                  uint32_t bits_per_mau = 8,
                  const uint64_t & max = uint64_t(~0),
                  const uint64_t & min = 0,
                  bool progmem = false,
                  bool virtmem = false,
                  unsigned char endianness = 0) = 0;

        /*!
         * \brief   Add memory block to this debug target.
         *
         * This function adds memory block to current space.
         * This function adds a default space when no current space exists.
         *
         * \param   name memory block name
         * \param   description memory block description
         * \param   start start address of this memory block
         * \param   end end address of this memory block
         * \param   attr memory block access attributes.
         * \return  memory block ID.
         */
        virtual uint32_t
        add_block(const std::string & name,
                  const std::string & description = "",
                  const uint64_t & start = 0,
                  const uint64_t & end = 0,
                  scx_debug_access_type attr = SCX_DEBUG_ACCESS_READ_WRITE) = 0;

        /*!
         * \brief   Perform checks related to memory breakpoint when given
         *          address is accessed.
         *
         * When memory address is hit, this function will:
         * - notifies debuggers of breakpoints which have been hit
         * - causes simulation to stop.
         *
         * \param   space memory space identifier
         * \param   addr memory address
         * \param   read \c true for read access, \false for write access.
         * \return  \c true when simulation will stop as result of breakpoint
         *          being hit, \c false otherwise.
         */
        virtual bool
        memory_bpt_check(uint32_t space, const uint64_t & addr, bool read) = 0;

        /*!
         * \brief   Return whether given memory address has breakpoint(s).
         * \param   space memory space identifier
         * \param   addr memory address
         */
        virtual bool
        has_memory_bpt(uint32_t space, const uint64_t & addr) = 0;

    /* Debug target - Parameters */
    public:

        /*!
         * \brief   Set new scx_debug_params_if for use by this debug target.
         * \param   params_if pointer to scx_debug_params_if to use.
         */
        virtual void
        set_params_if(scx_debug_params_if * params_if) = 0;

        /*!
         * \brief   Add parameter to this debug target.
         * \param   name parameter name
         * \param   description parameter description
         * \param   type parameter type
         * \param   runtime \c true when parameter can be changed at run-time,
         *          \c false otherwise
         * \param   min minimum admissible value
         * \param   max maximum admissible value
         * \param   value default value for SCX_DEBUG_PARAM_BOOL or
         *          SCX_DEBUG_PARAM_INT parameter
         * \param   str default string for SCX_DEBUG_PARAM_STRING parameter
         * \return  parameter ID.
         */
        virtual uint32_t
        add_param(const std::string & name,
                  const std::string & description,
                  scx_debug_param_value_type type = SCX_DEBUG_PARAM_INVALID,
                  bool runtime = false,
                  int64_t min = 0,
                  int64_t max = 0,
                  int64_t value = 0,
                  const char * str = "") = 0;

    /* Debug target - Control */
    public:

        /*!
         * \brief   Set new scx_debug_control_if for use by this debug target.
         * \param   control_if pointer to scx_debug_control_if to use.
         */
        virtual void
        set_control_if(scx_debug_control_if * control_if) = 0;

        /*!
         * \brief   Perform checks related to program breakpoint when given
         *          address is accessed.
         *
         * When program address is hit, this function will:
         * - notifies debuggers of breakpoints which have been hit
         * - causes simulation to stop.
         *
         * \param   space program space identifier
         * \param   addr program address
         * \return  \c true when simulation will stop as result of breakpoint
         *          being hit, \c false otherwise.
         */
        virtual bool
        program_bpt_check(uint32_t space, const uint64_t & addr) = 0;

        /*!
         * \brief   Return whether given program address has breakpoint(s).
         * \param   space program space identifier
         * \param   addr program address
         */
        virtual bool
        has_program_bpt(uint32_t space, const uint64_t & addr) = 0;

        /*!
         * \brief   Called by a model when it has reached a stopped state.
         *
         * stop_acknowledge is called by the model after it handles a stop
         * request as signified by scx_debug_control_if::debug_stop().
         *
         * \param   success @c true when model at a debug-able point, @c false otherwise.
         * \see     scx_debug_control_if::debug_stop()
         */
        virtual void
        stop_acknowledge(bool success) {
        }

        /*!
         * \brief   Called by a model when it has executed specified number of steps.
         *
         * step_acknowledge is called by the model after it handles a step
         * request as signified by scx_debug_control_if::debug_step().
         *
         * \param   success @c true when model at a debug-able point, @c false otherwise.
         * \see     scx_debug_control_if::debug_step()
         */
        virtual void
        step_acknowledge(bool success) {
        }

        /*!
         * \brief   Called by a model when it finishes its reset sequence.
         *
         * Model issues this notification after a few simulated cycles, to let the
         * CADI implementation know that the reset sequence has finished, and the model
         * is assumed to be at a debuggable state.
         *
         * \note Intended usage: To enable application/data load APIs, which could be
         * called before the complete system construction, or at a point where model
         * does not know its memory map/TCMs etc, or before PC/memory is writable.
         * Implementation of such APIs should wait for model to send this notification,
         * before issuing debug writes.
         *
         * \note This call is not expected while the model handles
         * scx_debug_control_if::debug_reset(), which is a controlled/blocking call.
         *
         * \note This call is expected while the simulation is running, since the model
         * needs clock ticks to execute the reset sequence. When this call is received,
         * the CADI/SCX framework is allowed to modify the model's state via debug writes,
         * and leave the SCXSimulationEngine::isControlTargetDebuggable() related state as
         * CpuState::SCX_TARGET_IS_RUNNING.
         *
         */
        virtual void
        reset_done() {
        }

    /* Debug target - Disassembly */
    public:

        /*!
         * \brief   Set new scx_debug_disass_if for use by this debug target.
         * \param   disass_if pointer to scx_debug_disass_if to use.
         */
        virtual void
        set_disass_if(scx_debug_disass_if * disass_if) = 0;

        /*!
         * \brief   Add mode to this debug target.
         * \param   mode mode number, \c 0 for current or don't care
         * \param   name mode name.
         */
        virtual void
        add_mode(uint32_t mode, const std::string & name) = 0;
};

}   /* namespace scx */

#endif  /* defined(ARM_SCX_DEBUG_TARGET_IF_H) */
