/*
 * SCX: scx_sched.h - SystemC eXport support: scheduling API.
 *
 * Copyright 2011-2015 ARM Limited.
 * All rights reserved.
 */

#ifndef ARM_SCX_SCHED_H
#define ARM_SCX_SCHED_H

/*!
 * \file    scx_sched.h
 * \brief   Scheduling API.
 * \date    Copyright 2011-2015 ARM Limited. All rights reserved.
 *
 * SystemC eXport support scheduling API.
 */

/* Includes */

#include <sg/SGSchedulerInterfaceForComponents.h>

/* Namespaces */

/*!
 * \brief   SystemC eXport support namespace.
 */
namespace scx {

/* Forwards */
class scx_simcontrol_if;

/* Functions */

/*!
 * \brief   Add a future synchronisation point in time.
 * \param   sync_time specifies the time of the future synchronisation point
 *          relative to the start of the current quantum. Measured in seconds.
 *
 * SystemC components call this function to provide a hint to the scheduler
 * implementation when a potentially useful system synchronisation point will
 * occur in the future.
 * The scheduler uses this information to determine the quantum size of threads
 * as they get scheduled.
 * Threads that already ran their quantum are unaffected, all other threads
 * (including the current thread) will run until the synchronisation point
 * specified by \a sync_time.
 *
 * Calling scx_sync() again will add another synchronisation point.
 *
 * Synchronisation points automatically vanish when the simulation time
 * reaches them.
 *
 * \warning ARM deprecates this function. Use IEEE 1666 SystemC 2011
 *          \c sc_core::sc_prim_channel:async_request_update() instead.
 */
extern void
scx_sync(double sync_time);

/*!
 * \brief   Sets the minimum synchronization latency for this scheduler.
 * \param   t specifies the minimum synchronization latency.
 *          Measured in seconds.
 *
 * The minimum synchronization latency helps to ensure that sufficient simulated
 * time has passed between two synchronization points for synchronization to be
 * efficient.
 * A small latency increases accuracy but decreases simulation speed.
 * A large latency dicreases accuracy but increases simulation speed.
 * The scheduler uses this information to compute the next synchronization point
 * as returned by sg::SchedulerInterfaceForComponents::getNextSyncPoint().
 *
 * \see     scx_get_min_sync_latency()
 */
extern void
scx_set_min_sync_latency(double t);
extern void
scx_set_min_sync_latency(sg::ticks_t t);

/*!
 * \brief   Returns the minimum synchronization latency for this scheduler.
 * \return  minimum synchronization latency. Measured in seconds.
 *
 * \see     scx_set_min_sync_latency()
 */
extern double
scx_get_min_sync_latency();
extern sg::ticks_t
scx_get_min_sync_latency(sg::Tag<sg::ticks_t> *);

/*!
 * \brief   Sets the maximum number of seconds to simulate.
 * \param   t number of seconds to simulate. Default to unlimited.
 */
extern void
scx_simlimit(double t);

/*!
 * \brief   Returns a pointer to a new instance of the default implementation
 *          of the scheduler mapping provided with Fast Models.
 * \param   simcontrol pointer to an already existing simulation controller.
 *          When this is \c NULL, this function returns \c NULL.
 */
extern sg::SchedulerInterfaceForComponents *
scx_create_default_scheduler_mapping(scx_simcontrol_if * simcontrol);

/*!
 * \brief   Returns a pointer to the current implementation of the scheduler
 *          mapping interface.
 */
extern sg::SchedulerInterfaceForComponents *
scx_get_curr_scheduler_mapping();

}   /* namespace scx */

#endif  /* defined(ARM_SCX_SCHED_H) */
