/*!
 * \file    ActivationFlag.h
 * \brief   Support for efficient read-access to a flag which changes state infrequently
 * \date    Copyright 2009 ARM Limited. All rights reserved.
 */

#ifndef ACTIVATIONFLAG_H
#define ACTIVATIONFLAG_H


/*! \brief Clients implement this to be notified whenever a flag changes state. */
class ActivationFlagObserver
{
public:
    /*! \brief Receive notification when the flag state changes.
     *
     *  \param new_state The new state of the flag being observer.
     */
    virtual void flagStateChanged(bool new_state) = 0;

    /*! \brief Receive notification when the flag is destroyed.
     *
     * The base-flag object is still valid at this point, so it is
     * legal to read its final boolean state.
     *
     * The observer should not attempt to unregister itself as part of
     * this callback.
     */
    virtual void flagDestroyed() = 0;

    virtual ~ActivationFlagObserver() {}
};

/*! \brief Base-class for objects having an observable boolean value.
 *
 * This provides a boolean-like object which can be observed, but doesn't
 * provide a mechanism for controlling when the boolean value changes.
 */
class ActivationValue
{
protected:
    /*! Create a flag with an initial boolean state. */
    ActivationValue(bool initial_value = false);

    /*! Destroy a flag and notify all its observers. */
    ~ActivationValue();

    void changeValue(bool new_value);

public:
    /*! \brief Returns the current state of the activation flag. */
    bool isActive() const
    {
        return value;
    }

    /*! \brief Allow the flag object to be used in a boolean context. */
    operator bool() const
    {
        return value;
    }

    inline bool operator == (bool o) const { return o == isActive(); }
    inline bool operator != (bool o) const { return o != isActive(); }


    /*! \brief Register to receive change notifications.
     *
     * \param new_observer The object to receive notifications.
     */
    void addObserver(ActivationFlagObserver *new_observer) const;

    /*! \brief Stop an observer fom receiving notifications.
     *
     * This call may be slow if the flag has a large number of observers!!
     *
     *  \param existing_observer An object previously registered using addObserver()
     */
    void removeObserver(ActivationFlagObserver *existing_observer) const;

private:
    //! Block use of copy construction
    ActivationValue(const ActivationValue&);

    //! Block use of copy assignment
    ActivationValue& operator=(const ActivationValue&);

    bool   value;     //!< Current state of the flag.

    /*! \brief Internal type to manage the list of observers. */
    class ObserverList;
    mutable ObserverList*   observers; //!< List of observers of the flag.
};


/*! \brief An observable, modifiable boolean flag.
 *
 *  ActivationFlag is intended to be used in situations where a flag changes
 *  state infrequently, but needs to be read in critical-path code.  To
 *  support this, the flag's current state can be broadcast to a number of
 *  ActivationFlagMonitor objects.
 *
 *  Typically, an owner of an ActivationFlag will use it as follows:
 *  \code
 *  class Owner {
 *    ActivationFlag activeness;
 *  public:
 *    // A const pointer to the flag can be monitored but not modified.
 *    ActivationValue const * getActiveness() {
 *      return &activeness;
 *    }
 *  };
 *  \endcode
 *
 *  The ActivationFlag can be treated as a boolean object:
 *  \code
 *    if (activeness) ...
 *    activeness = false;
 *  \endcode
 *
 *  When a client needs to test the activation flag very frequently, it can
 *  use an ActivationFlagMonitor object to hold a copy of the flag's state.
 *  This copy is kept up-to-date automatically, and can be read efficiently.
 *  See ActivationFlagMonitor for details.
 */
class ActivationFlag : public ActivationValue
{
public:
    /*! \brief Construct an ActivationFlag in the given state.
     *
     *  The default constructor initialises an ActivationFlag in the "false"
     *  state.  The optional argument explicitly sets the initial state.
     */
    ActivationFlag(bool active = false) : ActivationValue(active) {}

    /*! \brief Copy construction. (copy just the value) */
    ActivationFlag(ActivationFlag const& other) : ActivationValue(other.isActive()) {}

    /*! \brief Construct a flag with the same boolean state as another flag. */
    ActivationFlag(ActivationValue const& other) : ActivationValue(other.isActive()) {}

    /*! \brief Destroy a flag and remove all of its observers. */
    ~ActivationFlag() {}

    /*! \brief Allow the owner of the flag to change its state. */
    void setActive(bool active = true);

    /*! \brief Allow the owner of the flag to change its state using assignment. */
    ActivationFlag& operator = (bool active);

    /*! \brief Copy assignment (copy just the value). */
    ActivationFlag& operator = (ActivationFlag const&);

    /*! \brief Allow one flag's boolean state to be assigned from another value. */
    ActivationFlag& operator = (ActivationValue const&);
};

/*! \brief A Monitor can track the state of any number of other
 *  ActivationValue objects (ActivationFlag or other ActivationMonitors)
 *  and can efficiently test whether any of them is true.
 *
 *  The monitor is set up as follows:
 *  \code
 *    ActivationFlagMonitor is_active;
 *    is_active.monitor(owner->getActiveness()); // one-off setup
 *    ..
 *  \endcode
 *  It can be tested as it if is a normal boolean - this is a very efficient
 *  inline test:
 *  \code
 *    if (is_active) ...
 *  \endcode
 *
 *  A client can monitor multiple flags, and quickly test whether
 *  whether one or many of them have been set:
 *  \code
 *    ActivationFlagMonitor are_active;
 *    are_active->monitor(owner1->getActiveness());
 *    are_active->monitor(owner2->getActiveness());
 *    ..
 *    if (are_active) ...  // true whenever owner1 or owner2 is active.
 *  \endcode
 *
 * Normally the monitor watches for any flags being true, but it can
 * also watch for flags being false.
 * \code
 *    is_inactive->monitor(owner->getActiveness(), false);
 *    if (is_inactive) // true if owner is not active.
 * \endcode
 */
class ActivationFlagMonitor : public ActivationValue
{
    class MonitorEntry;
    typedef MonitorEntry * FlagHandle_t;

    /*! Track how many flags are currently active. */
    unsigned count_active;

    /*! Track which flags we are monitoring. */
    MonitorEntry *monitor_of;

    void clearFlagList();
    void updateState();

public:
    /*! \brief Construct a monitor that is optionally watching a single flag. */
    ActivationFlagMonitor(ActivationValue const * flag = 0, bool active_sense = true);

    /*! \brief Tidy-up on exit. */
    ~ActivationFlagMonitor();

    /*! \brief Construct a monitor that has the same initial state as another moniotr.
     *
     *  The new monitor will be watching the same flag(s) as the existing monitor, but
     *  will have no observers.
     */
    ActivationFlagMonitor(ActivationFlagMonitor const &other);

    /*! \brief Copy-assignment: set this monitor to watch the same flags as another monitor.
     *
     * This monitor will set to watch the same set of ActivationFlag objects
     * as the other monitor, but it will keep its current set of observers.
     * If this assignment causes the monitor's boolean state to change,
     * all observers will be notified.
     */
    ActivationFlagMonitor & operator= (ActivationFlagMonitor const &other);

    /*! \brief Stop monitoring any flags. */
    void unmonitor();

    /*! \brief Start monitoring a specific flag. */
    FlagHandle_t monitor(ActivationValue const * flag, bool active_sense = true);

    /*! \brief Stop monitoring a specific flag. */
    void unmonitor(ActivationValue const * flag);

    /*! \brief Stop monitoring a specific flag. */
    void unmonitor(FlagHandle_t handle);
};
#endif
