/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   AlignedMemory.h
 *  \brief  Utility functions for aligned memory
 *  \date   Copyright ARM Limited 2014-2016 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 */

#ifndef INCLUDED_SG_ALIGNEDMEMORY_H
#define INCLUDED_SG_ALIGNEDMEMORY_H

#include <cstddef>
#include <cstdlib>
#include <malloc.h>


namespace sg {

/*! Allocate memory with specific alignement
 *
 *  \param  granule Required alignment, this must be a power of 2 and a multiple of sizeof(void*)
 *  \param  size    Size of requested allocation, this must be a multiple of granule
 *  \return allocated memory if successful or null pointer
 */
inline void* malloc_aligned(size_t granule,size_t size)
{
// TODO use C11 aligned_alloc() when it makes it's way into a supported C++ standard
// aligned_alloc was introduced to glibc in v2.16 but still need to support RHEL6 which is on v2.12
#if defined(_WIN32)
    return _aligned_malloc(size,granule);
#elif defined(__unix)
    void *ptr;
    if (posix_memalign(&ptr, granule, size) != 0)
    {
        ptr = 0;
    }
    return ptr;
#else
#error "Unsupported OS"
#endif

}

/*! Free memory allocated with malloc_aligned
 *
 * \param    ptr     a pointer allocated with malloc_aligned
 */
inline void free_aligned(void* ptr)
{
#if defined(_WIN32)
    // TODO use free() when C11 aligned_alloc() makes it's way into a supported C++ standard
    _aligned_free(ptr);
#elif defined(__unix)
    free(ptr);
#else
#error "Unsupported OS"
#endif
}

}
#endif

/* eof Memory.h */
