/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   Attributes.h
 *  \brief  Abstraction for useful compiler hints
 *  \date   Copyright ARM Limited 2013 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 */

#ifndef INCLUDED_ATTRIBUTES_H
#define INCLUDED_ATTRIBUTES_H

/** C++ 17 specifics */
#if  __cplusplus >= 201703L
    #define SG_FALLTHROUGH [[fallthrough]]
    #define SG_MAYBEUNUSED [[maybe_unused]]
    #define SG_NODISCARD   [[nodiscard]]
#endif

/**
 * Test for GCC being the compiler and being at least the given version.
 */
#define SG_IS_GCC_AND_AT_LEAST(major_ver, minor_ver) \
    (defined(__GNUC__) && (__GNUC__ > major_ver || (__GNUC__ >= major_ver && __GNUC_MINOR__ >= minor_ver)))

/* Clang aims to remain compatible with gcc, and defines __GNUC__. Therefore we
 * need to check if the compiler is clang before checking if it's GCC.
 */
#if defined(__clang__)
    #define SG_FORCE_INLINE __inline

    /* SG_ALWAYS_INLINE: A compiler hint prefix to always inline this function
       NOTE: gcc 5.3.0 and later require the extra inline keyword to avoid a warning
             when used on static functions  */
    #define SG_ALWAYS_INLINE __attribute__((always_inline)) inline

    #define SG_PURE         __attribute__((pure))
    #define SG_CONST        __attribute__((const))
    #define SG_DEPRECATED   __attribute__((deprecated))
    #define SG_UNUSED       __attribute__((unused))
    #define SG_NORETURN     __attribute__((noreturn))
    #define SG_WARN_UNUSED_RESULT __attribute__((warn_unused_result))
    #define SG_FORMATSTR(type, fmt_idx, var_idx) __attribute__((format(type, fmt_idx, var_idx)))

    /* Clang introduced C++11 override support in Clang 3.0, and has
     * a handy feature intrinsic to check for it...
     * Note that both final and override keywords use the same feature name.
     */
    #define SG_HAS_CXX11_FINAL    (__has_feature(cxx_override_control))
    #define SG_HAS_CXX11_OVERRIDE (__has_feature(cxx_override_control))

    #define SG_PACKED_STRUCT(declaration) struct __attribute__((packed)) declaration

    #define SG_FUNCTION_SIGNATURE __PRETTY_FUNCTION__

    /* Wrapper to use allow the use of C++11 noexcept(expr) in function specifiers.
     * C++11 by default has noexcept(true) for destructors, and occasionally it's
     * useful to override this. C++03 doesn't have the behaviour so we can ignore
     * the specifier there. Clang 3.0(?) and later support it in C++11 mode.
     */
    #if __has_feature(cxx_noexcept)
    #define SG_NOEXCEPT(expr) noexcept(expr)
    #else
    #define SG_NOEXCEPT(expr)
    #endif

    #if  __cplusplus < 201703L
        #if __has_cpp_attribute(clang::fallthrough)
        #define SG_FALLTHROUGH [[clang::fallthrough]]
        #else
        #define SG_FALLTHROUGH
        #endif

        #if __has_cpp_attribute(unused)
        #define SG_MAYBEUNUSED __attribute__((unused))
        #else
        #define SG_MAYBEUNUSED
        #endif

        #if __has_cpp_attribute(clang::warn_unused_result)
        #define SG_NODISCARD [[clang::warn_unused_result]]
        #else
        #define SG_NODISCARD
        #endif
    #endif

    #define SG_HAS_CXX11_CONSTEXPR  (__has_feature(cxx_constexpr))
    #define SG_HAS_CXX14_CONSTEXPR  (__has_feature(cxx_relaxed_constexpr))

    // Static branch prediction
    // only use when profiling has revealed a significant benefit
    #define SG_EXPECT(expr)     __builtin_expect(expr, 1)
    #define SG_UNEXPECT(expr)   __builtin_expect(expr, 0)

#elif defined(__GNUC__)
    /* SG_FORCE_INLINE: A compiler hint prefix to inline this function */
    #define SG_FORCE_INLINE __inline

    /* SG_ALWAYS_INLINE: A compiler hint prefix to always inline this function
       NOTE: gcc 5.3.0 and later require the extra inline keyword to avoid a warning
             when used on static functions  */
    #define SG_ALWAYS_INLINE __attribute__((always_inline)) inline

    /* SG_PURE: Function has no side effects except return values */
    #define SG_PURE         __attribute__((pure))

    /* SG_CONST: Function is pure *and* doesn't depend on anything but the function arguments
     * and doesn't depend on dereferenced argument pointers.
     */
    #define SG_CONST        __attribute__((const))

    /* SG_DEPRECATED: Warn if a variable or type is used anywhere in the program */
    #define SG_DEPRECATED   __attribute__((deprecated))

    /* SG_UNUSED: Tell the compiler a variable is intentionally unused */
    #define SG_UNUSED       __attribute__((unused))

    /* SG_NORETURN: Hint to compiler that the function never returns.
     * Useful to hide spurious uninitialised value warnings in callers.
     */
    #define SG_NORETURN     __attribute__((noreturn))

    /**
     * SG_WARN_UNUSED_RESULT: warn if a function's return value is not used.
     */
    #define SG_WARN_UNUSED_RESULT __attribute__((warn_unused_result))

    /* SG_FORMATSTR: Used to indicate we're implementing a printf/scanf-like function */
    /* NB if using with class functions, the function has an implicit 'this' as the first arg */
    #define SG_FORMATSTR(type, fmt_idx, var_idx) __attribute__((format(type, fmt_idx, var_idx)))

    /* GCC 4.7 and later need -std=c++11 or -std=gnu++11 to fully
     * support override, but have __final in non-C++11 mode.
     * The standard way to check for C++11 support is via __cplusplus,
     * but some g++s don't set it correctly, so we check for
     * __GXX_EXPERIMENTAL_CXX0X__ too.
     */
    #if defined(__GXX_EXPERIMENTAL_CXX0X__) || __cplusplus >= 201103L
    #define SG_HAS_CXX11_FINAL    SG_IS_GCC_AND_AT_LEAST(4, 7)
    #define SG_HAS_CXX11_OVERRIDE SG_IS_GCC_AND_AT_LEAST(4, 7)
    #else
    #define SG_HAS_GCC___FINAL    SG_IS_GCC_AND_AT_LEAST(4, 7)
    #endif

    /* SG_PACKED_STRUCT(name): Tell the compiler to tightly pack a struct
     */
    #define SG_PACKED_STRUCT(name) struct __attribute__((packed)) name

    /* Macro for the current function */
    #define SG_FUNCTION_SIGNATURE __PRETTY_FUNCTION__

    /* Wrapper to use allow the use of C++11 noexcept(expr) in function specifiers.
     * C++11 by default has noexcept(true) for destructors, and occasionally it's
     * useful to override this. C++03 doesn't have the behaviour so we can ignore
     * the specifier there. Supported by GCC 4.6 and later in C++11 mode.
     */
    #if SG_IS_GCC_AND_AT_LEAST(4, 7) && (defined(__GXX_EXPERIMENTAL_CXX0X__) || __cplusplus >= 201103L)
    #define SG_NOEXCEPT(expr) noexcept(expr)
    #else
    #define SG_NOEXCEPT(expr)
    #endif

    #if  __cplusplus < 201703L
        #if SG_IS_GCC_AND_AT_LEAST(7, 1)
        #define SG_FALLTHROUGH [[gnu::fallthrough]]
        #else
        #define SG_FALLTHROUGH
        #endif
        
        #if SG_IS_GCC_AND_AT_LEAST(4, 8)
        #define SG_MAYBEUNUSED [[gnu::unused]]
        #define SG_NODISCARD   [[gnu::warn_unused_result]]
        #else
        #define SG_MAYBEUNUSED
        #define SG_NODISCARD
        #endif
    #endif 

    #define SG_HAS_CXX11_CONSTEXPR  SG_IS_GCC_AND_AT_LEAST(4, 6)
    #define SG_HAS_CXX14_CONSTEXPR  SG_IS_GCC_AND_AT_LEAST(5, 0) && __cplusplus >= 201402L

    // Static branch prediction
    // only use when profiling has revealed a significant benefit
    #define SG_EXPECT(expr)     __builtin_expect(expr, 1)
    #define SG_UNEXPECT(expr)   __builtin_expect(expr, 0)

#elif defined(_MSC_VER)
    #define SG_FORCE_INLINE __forceinline
    #define SG_ALWAYS_INLINE SG_FORCE_INLINE
    #define SG_PURE
    #define SG_CONST
    #define SG_DEPRECATED
    #define SG_UNUSED
    #define SG_NORETURN
    #define SG_WARN_UNUSED_RESULT
    #define SG_FORMATSTR(type, fmt_idx, var_idx)

    /* Visual Studio 2005 and later have override support.
     * Visual Studio 2005 added final support, but called it 'sealed'
     * Visual Studio 2012 added final support with the proper C++11 name
     */
    #define SG_HAS_CXX11_FINAL    (_MSC_VER >= 1700)
    #define SG_HAS_CXX11_OVERRIDE (_MSC_VER >= 1400)
    #define SG_HAS_MSVC_SEALED    (_MSC_VER >= 1400)

    // We can't just wrap the pragma around the name of the struct as MSVC++
    // 2010 doesn't like this Unfortunately wrapping the macro around the
    // entire struct causes syntax highlighting errors but is the only way to
    // make it work.
    #define SG_PACKED_STRUCT(declaration) __pragma(pack(push, 1)) struct declaration __pragma(pack(pop))

    #define SG_FUNCTION_SIGNATURE __FUNCSIG__

    /* As of Visual Studio 2013, MSVC doesn't support noexcept:
     * See http://msdn.microsoft.com/enu-us/library/hh567368.aspx
     * Visual Studio 2012 and later have C++11 support which have
     * dtors with noexcept(true) by default, and we don't (yet)
     * have a way to override this. Earlier Visual Studios use
     * C++03 so don't require any declaration
     */
    #if (_MSC_VER >= 1700)
    // SG_NOEXCEPT is not defined for Visual Studio 2012+ so that using it
    // causes an error. (Unfortunately, there is no clean way to generate a
    // user-specified error message on *using* a macro)
    #else
    #define SG_NOEXCEPT(expr)
    #endif

    #if  __cplusplus < 201703L
    #define SG_FALLTHROUGH
    #define SG_MAYBEUNUSED
    #define SG_NODISCARD
    #endif

    #define SG_HAS_CXX11_CONSTEXPR  (_MSC_VER >= 1900)
    #define SG_HAS_CXX14_CONSTEXPR  (_MSC_VER >= 1910)

    // Static branch prediction not supported
    #define SG_EXPECT(expr)     (expr)
    #define SG_UNEXPECT(expr)   (expr)

#else
    #error "Unsupported compiler for attributes"
#endif


/* Annotate a virtual method saying it *must* override a virtual
 * method in a parent class. This avoids the case where the method
 * is accidentally made non-virtual in the parent class, or the
 * method accidentally has the wrong signature and therefore isn't
 * overriding one in the base class.
 *
 * Usage: virtual void foo_method() SG_OVERRIDE;
 */
#if SG_HAS_CXX11_OVERRIDE
    #define SG_OVERRIDE override
#else
    #define SG_OVERRIDE /* NO SG_OVERRIDE SUPPORT */
#endif


/* 1. Annotate a virtual method to say no subclass can override it.
 * 2. Annotate a class to say it may not be subclassed
 *
 * Usage: class FinalClass SG_FINAL : public Parent {};
 *        virtual void final_method() SG_FINAL;
 */
#if SG_HAS_CXX11_FINAL
    #define SG_FINAL final
#elif SG_HAS_GCC___FINAL
    #define SG_FINAL __final
#elif SG_HAS_MSVC_SEALED
    #define SG_FINAL sealed
#else
    #define SG_FINAL /* NO SG_FINAL SUPPORT */
#endif

/* A wrapper function to perform explicit valgrind defined value checks
 * to all the values being read from a component's memory or register
 * interfaces, or sent across ports.
 *
 * This allows us to enforce an interface boundary across which
 * uninitialized values will not pass.
 *
 * If uninitialised values propagate into the memory subsystem or
 * generated code, they become extremely diffcult to debug due to the
 * indirection involved.
 *
 * Enable at build time with the Scons variable 'valgrind_checks'
 */
#ifdef SG_VALGRIND_INTERFACE_CHECKS
#include <memcheck.h>
template<typename T> static inline T const& sg_check_is_defined(T const& t)
{
    VALGRIND_CHECK_VALUE_IS_DEFINED(t);
    return t;
}
#else
#define sg_check_is_defined(t) t
#endif //SG_VALGRIND_INTERFACE_CHECKS

/* Mark a function as constexpr if supported by the compiler.
 *
 * Your code cannot use the resulting function in compile-time expressions
 * (e.g. template arguments), since constexpr is not supported on older
 * compilers. Think of it as:
 * - A hint to the compiler
 * - A likely candidate for conversion to a real 'constexpr' once all compilers
 *   support it.
 *
 * Usage:   SG_CONSTEXPR_FUNC void foo();
 */
#if SG_HAS_CXX11_CONSTEXPR
    #define SG_CONSTEXPR_FUNC constexpr
#else
    #define SG_CONSTEXPR_FUNC /* NO CONSTEXPR SUPPORT */
#endif

/* Mark a function as constexpr with C++14 features (aka "extended constexpr"
 * or "relaxed constexpr") if supported by the compiler.
 *
 * Your code cannot use the resulting function in compile-time expressions
 * (e.g. template arguments), since C++14 constexpr is not supported on older
 * compilers. Think of it as:
 * - A hint to the compiler
 * - A likely candidate for conversion to a real 'constexpr' once all compilers
 *   support it.
 *
 * Usage:   SG_CONSTEXPR_CXX14_FUNC void foo();
 */
#if SG_HAS_CXX14_CONSTEXPR
    #define SG_CONSTEXPR_CXX14_FUNC constexpr
#else
    #define SG_CONSTEXPR_CXX14_FUNC /* NO C++14 CONSTEXPR SUPPORT */
#endif

/* Mark a variable as constexpr if supported by the compiler, or const otherwise.
 *
 * Your code cannot necessarily use the resulting variable in compile-time
 * expressions (e.g. template arguments), since constexpr is not supported on
 * older compilers. Think of it as:
 * - A hint to the compiler
 * - A likely candidate for conversion to a real 'constexpr' once all compilers
 *   support it.
 *
 * Usage:   SG_CONSTEXPR_VAR void foo();
 */
#if SG_HAS_CXX11_CONSTEXPR
    #define SG_CONSTEXPR_VAR constexpr
#else
    #define SG_CONSTEXPR_VAR const /* NO CONSTEXPR SUPPORT */
#endif

#endif
/* eof Attributes.h */
