/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   Concurrency.h
 *  \brief  Low level constructs for memory concurrency
 *  \date   Copyright ARM Limited 2011 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 *  Lock Free algorithms require specific memory constructs that are not
 *  provided by C++.
 */

#ifndef INCLUDED_CONCURRENCY_H
#define INCLUDED_CONCURRENCY_H

#include "stdint.h"
#include "sg/Macros.h"


    
#if defined(_MSC_VER) && !defined (__CLANG_UNDER_WINDOWS__)
    
    // ----------------------------------------------------------------------
    //                        Microsoft Definitions
    // ----------------------------------------------------------------------

    // It appears that using intrin.h and some combination of pragmas break
    // xlocale.
    #include <intrin.h>
    #pragma intrinsic(_ReadWriteBarrier)
    #define sg_rwbarrier() _ReadWriteBarrier()
    
namespace sg {
    // ------------------------------
    //  atomic_postincrement() atomic_postdecrement()
    // ------------------------------
    
    /// increment a value and return the original value. equivalent to an atomic *ptr++
    int64_t atomic_postincrement(volatile int64_t* valp);

    /// decrement a value and return the original value. equivalent to an atomic *ptr--
    int64_t atomic_postdecrement(volatile int64_t* valp);
}

#elif defined(__GNUC__) || defined (__CLANG_UNDER_WINDOWS__)
    
    // ----------------------------------------------------------------------
    //                        GCC Definitions
    // ----------------------------------------------------------------------

    #define sg_rwbarrier() __asm__ __volatile__ ("" : : : "memory")

namespace sg {
    // ------------------------------
    //  atomic_postincrement() atomic_postdecrement()
    // ------------------------------
    #if __GNUC__ == 4 && __GNUC_MINOR__ <= 7
        // increment a value and return the original value. equivalent to an atomic *ptr++
        inline int64_t atomic_postincrement(volatile int64_t* valp) { return __sync_fetch_and_add(valp, 1); }
        
        // decrement a value and return the original value. equivalent to an atomic *ptr--
        inline int64_t atomic_postdecrement(volatile int64_t* valp) { return __sync_fetch_and_sub(valp, 1); }
    #else
        // increment a value and return the original value. equivalent to an atomic *ptr++
        inline int64_t atomic_postincrement(volatile int64_t* valp) { return __atomic_fetch_add(valp, 1, __ATOMIC_SEQ_CST); }
             
        // decrement a value and return the original value. equivalent to an atomic *ptr--
        inline int64_t atomic_postdecrement(volatile int64_t* valp) { return __atomic_fetch_sub(valp, 1, __ATOMIC_SEQ_CST); }
    #endif
}
    
#else
#error "Unsupported Compiler"
#endif
    
namespace sg {
    // ----------------------------------------------------------------------
    //                         Common definitions
    // ----------------------------------------------------------------------

    template <typename T>
    inline T load_acquire (T const volatile* src) { T v = *src;  sg_rwbarrier(); return v; }
    template <typename T>
    inline T load_acquire (T const* src) { T v =  *const_cast<T const volatile*>(src); sg_rwbarrier(); return v; }

    template <typename T>
    inline void store_release(T volatile* dst, T src) { sg_rwbarrier(); *dst = src; }
    template <typename T>
    inline void store_release(T* dst, T src) { sg_rwbarrier(); *const_cast<T volatile*>(dst) = src; }

#undef sg_rwbarrier
    
}

#endif

/* eof Concurrency.h */
