/*!
 * \file    ConnectorBase.h
 * \brief   Base class for handling abstract connector behaviour.
 * \date    Copyright 2006,2017 ARM Limited. All rights reserved.
 * \author  Andrew Bolt and Graeme Barnes.
 *
 * Define a policy-neutral, non-templated base class for connectors.
 * The templatisation occurs in the subclass and supporting classes
 * defined in ConcreteConnectorBase.h.  Connection policy is defined
 * in PeerConnector.h and DirectedConnector.h
 */

#ifndef CONNECTORBASE_H
#define CONNECTORBASE_H

#include "sg/FastDelegate.h"    // Could use separate out FastDelegateMemento.h
#include <list>
#include <vector>
#include <string>
#include <utility>
#include <cassert>
#include <iosfwd>
#include <atomic>
#include <memory>
#include <iostream>

#include "sg/SGNode.h"
#include "sg/IndexedDelegateMemento.h"
#include <stdint.h>

namespace sg {

class Connectable;
class ConnectorBase;

// Forward declarations for pseudo- dynamic multiple dispatch
class Port;
class PeerConnectorBase;
class DirectedConnectorBase;
class SimpleConnectorBase;


class InvokeBase;

/// A CallID is a (mostly) opaque token that represents a call that the
/// connector is about to make, or is making.  CallID bindings can receive
/// this as their first argument.  It is guaranteed that no connector will
/// use the default constructed ConnectorCallID.
class ConnectorCallID {
public:
    using Token = void*;
    
    ConnectorCallID(Token token_ = nullptr, uint64_t value_ = 0)
        : token(token_), value(value_) {}

    void increment()                                   { ++ value; }
    bool operator == (ConnectorCallID const& o_) const { return token == o_.token && value == o_.value; }
    bool operator != (ConnectorCallID const& o_) const { return !((*this) == o_); }
    bool isValid() const                               { return token != nullptr && value != 0; }
    bool isSpecialCallID() const                       { return value == SPECIAL; }
    bool isSpecialCallIDWithToken(Token token_) const  { return value == SPECIAL && token == token_; }

    static ConnectorCallID makeSpecial(Token token_ = nullptr) { return ConnectorCallID(token_, SPECIAL); }


    Token     getToken() const { return token; }
    uint64_t  getValue() const { return value; }

    /// Return the CallID that is prior to this one.
    ConnectorCallID previous() const { return ConnectorCallID(getToken(), getValue() - 1); }
    
private:
    Token                 token;
    static uint64_t const SPECIAL = ~0ull;
    uint64_t              value;
};

std::ostream& operator <<(std::ostream& os_, ConnectorCallID const& t_);


/*! Allow Connectables to have an 'Interceptor' examine and interfere with the
 *  edge connection process.
 *
 *  This can be used by the Interceptor to add wrap function calls on the
 *  connectors, or an Interceptor might recognise the port itself and replace
 *  the connection with a specific adaptor that bridges the two ends of the
 *  call.
 *
 *  It does this by the return value of installEdge() which is called:
 *
 *      - for peer and directed connectors, the proposed edge is checked for being
 *        able to be connected and the result passed into the installEdge() call
 *
 *      - for ports, before any behaviours are attempted to be bound. Ports can
 *        always be bound together, but they might not have any common
 *        behaviours to bind!
 *
 *  Thus the Interceptor can install adaptors to bridge between different types
 *  of ports, or a connector that bridges between two different connectors.
 *
 *  The Interceptor can generically ask for an 'adaptor' behaviour to be bound
 *  on the connector (via bindAdaptor() and unbindAdaptor()).  The Interceptor
 *  then has to emulate the call through the edge using a CallList that the
 *  adaptor behaviour will execute.
 *
 *  The Interceptor must track the edges in the graph as necessary.  It uses a
 *  data member in each connector of type Interceptor::InterceptorNode called
 *  interceptor_node.
 *
 *  A port contains connectors, and the hierarchical containment relationship is
 *  tracked by 'EdgeLabels'.  An EdgeLabel is an object created by the
 *  implementation of the Interceptor to contain a pointer to the Interceptor
 *  and any information necessary to describe that edge.  It is passed to the
 *  connectToVia() call of the port and then a new EdgeLabel is created by
 *  EdgeLabel::createEdgeLabel() in the port and then passed to the connectors
 *  in the port.  This lets the Interceptor track which port they are being
 *  bound from.  The lifetime of the EdgeLabel are tracked by shared_ptr.
 *
 *  The first level EdgeLabel is created by asking the _specific implementation_
 *  of the Interceptor to make one.  Thus, in the Interceptor interface there
 *  are no methods to create an initial EdgeLabel.  It is up to the Interceptor
 *  implementation to define a suitable methodology for creating one.
 *
 */
class Interceptor
{
public:
    virtual ~Interceptor() {}
    
    using CallID = ConnectorCallID;
    class InterceptorNode;
    
    ///
    /// Configure how to print the Interceptor
    ///
    class print_config_t
    {
    public:
    };


    ///
    /// Print a representation of the Interceptor.
    ///
    virtual void print(std::ostream& os_, print_config_t const& = print_config_t()) const = 0;
        
    /// The CallList for an adaptor to execute.
    ///
    /// An adaptor is a behaviour bound on a Connector and so will receive the
    /// normal messages destined for bound behaviours on that Connector.
    ///
    /// The adaptor's job is to then call a bunch of other objects with the same
    /// connector type and it does this by the CallList.
    ///
    /// The CallList is the base class of an object that knows how to deliver a
    /// sequence of objects for the adaptor to call.  It is created by the
    /// Interceptor and its lifetime is managed by shared_ptr<>.
    ///
    /// The CallList is given to the adaptor as const and is immutable except
    /// for the valid flag which is atomically changed to invalid when
    /// necessary.
    ///
    /// However, whilst processing the CallList then some mutable state is
    /// necessary and we use a queue abstraction to do this:
    ///
    ///     // CallList::queue holds the mutable state for iterating through
    ///     // the call_list.
    ///     CallList::queue  q(interceptor_node);
    ///
    ///     if (q.hasCalls())
    ///         while (InvokeBase* invoke_base = q.pop_and_return_front())
    ///         {
    ///              // use invoke_base to invoke the call.
    ///         }
    ///
    /// NOTE the use of hasCalls() is to avoid some extra tests in
    /// pop_and_return_front() and can't be missed.
    ///
    /// <201708100708>
    ///
    /// NOTE that for PeerConnectors then the queue::pop_and_return_front()
    /// (which calls CallList::getNextInvokeBase()) has to not call the graph
    /// component for which interceptor_node is the in-charge node or that
    /// component would be called twice.  This is because the CallList is shared
    /// between all nodes and includes a list of all in-charge nodes, including
    /// the one that is executing the CallList.  Thus it has to recognise
    /// through queue::originator_InvokeBase which node is executing it and
    /// hence not return it.
    ///
    /// In addition, then the Interceptor has to cope with self-calls used to
    /// process split-edges (see <201708110958/>) as follows...
    ///
    /// Any normal invocation of a behaviour will have a CallID that is unique to
    /// that call and is distinguishable from 'special' CallIDs.
    ///
    /// When we use an adaptor to transfer a call across an edge then if there
    /// is an adaptor bound on the other end of the edge then it will also
    /// receive that call and so it has to recognise the self-call so that it
    /// doesn't transfer it back along the edge it came from.
    ///
    /// The edges for which both ends have bound adaptors are:
    ///
    ///    - all edges between PeerConnector graphs
    ///    - edges between split nodes in DirectedConnector graphs
    ///
    /// For PeerConnectors, then any adaptor receiving a forwarded call should
    /// filter it out as the original in_charge adaptor will be responsible for
    /// forwarding to all graph components.  It does this by using the singular
    /// special CallID produced by makeSpecial().  This is all done by the
    /// process_call_list helper and the Interceptor need do nothing.  See
    /// <201708100704/>.
    ///
    /// For DirectedGraphs, then only the self call for that specific edge has
    /// to be filtered out as it can then flow to an adaptor that was bound to
    /// forward it along another edge (that may or may not be between split
    /// nodes).  For example, if we have nodes:
    ///
    ///         A <--split--> B --intercepted--> C
    ///
    /// The node A and B are connected by a split-edge.  B and C are connected
    /// by an intercepted edge and so there is an adaptor bound on B that is to
    /// transfer the call across the edge B->C.  The adaptor bound on B though
    /// also will notice that there is a split edge between A and B and so would
    /// normally attempt to transfer the call across B to A.  However, if this
    /// call was originally made from A then this self-call has to be removed.
    ///
    /// The way that the Interceptor's implementation has to do this is by
    /// recognising the self-call by the CallID.  When a call is transferred
    /// across an edge, then the CallID in the CallList::queue is marked as
    /// 'Special' and with the origin InvokeBase as the 'token' in the CallID.
    /// When the CallList is processing getNextInvokeBase() then it has to
    /// remove any call that would transfer it this origin InvokeBase.
    ///
    /// </201708100708>
    class CallList {
    public:
        virtual ~CallList();

        CallList();
        
    public:
        // Interface used from the adaptor.  It can only access const members.
        
        /// Is the CallList valid?
        bool isValid() const                               { return valid.load(); }

        ///
        /// An abstraction onto iterating through the CallList.
        ///
        class queue {
        public:
            // Interface used when invoking the CallList.

            inline queue(InterceptorNode&, CallID const&);

            bool hasCalls() const { return call_list != nullptr; }

            InvokeBase* pop_and_return_front()
            {
                return call_list->getNextInvokeBase(*this);
            }

        public:
            // Interface used by CallList implementation


            /// NOTE that the shared_ptr ensures that the call list is not
            /// destroyed whilst we are executing from it.
            std::shared_ptr<CallList const>   call_list;

            /// The data element provided by the implementation of the
            /// Interceptor in InterceptorNode::d_for_CallList.
            ///
            /// <201709060916>
            ///
            /// This is copied from the interceptor_node when the queue is
            /// created so that if the node is recoloured during executing the
            /// queue then the InterceptorNode::d_for_CallList can be changed
            /// without interfering with the ongoing call.
            ///
            /// </201709060916>
            ///
            /// As the same CallList might be used by multiple nodes, depending
            /// on what the implementation of the Interceptor is trying to
            /// achieve, then this data element can act as a fast way of
            /// identifying the specific node on which the CallList was
            /// installed.
            ///
            uint64_t                          d_for_CallList;

            /// The InvokeBase of the InterceptorNode that bound the queue.
            /// This is used by the implementation to remove self-calls to the
            /// same graph component.
            InvokeBase* originator_InvokeBase()
            {
                return interceptor_node.getInvokeBase();
            }

            /// The InterceptorNode that this queue is being executed for.  This
            /// is useful to provide access to the InterceptorNode::d/p fields
            /// that are for use by the implementation.
            ///
            /// See <201709060916/> for InterceptorNode::d_for_CallList
            InterceptorNode&                  interceptor_node;
            
            /// This is the CallID of the call we are trying to process.
            /// See <201708100708/>.
            CallID                            call_id;
            
            /// For use by the derived implementation.  The very first call is
            /// identified by data[0] == 0.
            ///
            /// The obtuse way of providing some storage here for the
            /// CallList implementation to use is because:
            ///    - CallList is immutable (barring the valid flag)
            ///    - the CallList implementation will need mutable data for
            ///      iterating through the list
            ///    - this must be fast so we have to avoid new/delete for
            ///      the mutable state.
            ///
            /// Thus we provide a small amount of storage on the stack (queue is
            /// allocated on the stack) for use by the implementation of
            /// CallList.
            uintptr_t                         data[4];

        private:
            queue(queue const&)                    = delete;
            void  operator = (queue const&)        = delete;
            void* operator new(size_t)             = delete;
            void  operator delete(void*, size_t)   = delete;
            void* operator new[](size_t)           = delete;
            void  operator delete[](void*, size_t) = delete;
        };
        
    private:
        friend class queue;
        
        /// Interface used by queue.
        ///
        /// NOTE with care <201708100708/> which places responsibility on the
        /// implementation of the CallList for removing some edges from the call
        /// list.
        virtual InvokeBase* getNextInvokeBase(queue&) const = 0;

    protected:
        mutable std::atomic<bool> valid;
    };
    
    /// This is the response of the Interceptor to the installEdge() call to the
    /// connector to says what to do.
    enum ConnectionAction_t
    {
        /// Run the logic that normally connects the two connectors together.
        PERFORM_NORMAL_CONNECTION_REAL_EDGE,
        
        /// Run the logic that normally connects the two connectors together,
        /// except this is a special flag for the Port connectors to tell it
        /// that the connectors in the port should be bound, but this represents
        /// a 'fake' edge -- that is one that is not represented in the
        /// connection graph of the Connectors because an adaptor is going to do
        /// the job.  No other connectors will receive this.
        PERFORM_NORMAL_CONNECTION_FAKE_EDGE,
        
        /// Pretend as though the connection has already occurred and was
        /// successfully.  Normally the Interceptor has installed an adaptor or
        /// called bind_adaptor() a forwarder to it.
        CONNECTION_FORMED_OK,
        
        /// Deny the connection being formed.
        CONNECTION_DENIED
    };
    
    ///
    /// A object that is passed down as the connector is bound by connectToVia().
    ///
    /// This is intended to encapsulate a pointer to an Interceptor and also an
    /// object that the Interceptor can use to keep track of what connectors are
    /// in which ports, etc.
    ///
    /// They are a way of the Interceptor tracking which connectors exist in a
    /// port.
    ///
    /// These interfaces do not specify a way of creating an initial EdgeLabel,
    /// only a way of creating new EdgeLabel from one that is used to connect
    /// two ports together and that new EdgeLabel will be given to the
    /// Connectors in the Port.
    ///
    /// As the interpretation of the EdgeLabels is specific to the
    /// implementation of the Interceptor and what it is trying to achieve then
    /// it is up to the Interceptor implementation to provide a mechanism for
    /// creating the initial EdgeLabels.
    ///
    class EdgeLabel {
    public:
        virtual ~EdgeLabel();

        Interceptor* getInterceptor() const { return interceptor; }

        // Take the current EdgeLabel and use it to create a new one based on
        // the information in the current one.  The EdgeLabel created is then
        // passed to connectToVia() for each Connector in the Port.
        virtual std::shared_ptr<EdgeLabel> createEdgeLabel(
            Port& from_, Port& to_
            ) = 0;

        // Does the edge represent an edge between a node that has been split in
        // two?
        //
        // <201708110958>
        //
        // A split-node is where we have two nodes that represent the same node.
        //
        // The nodes connected to them are partitioned by the user into a 'left'
        // and 'right' set.  The edge between the split nodes are required to
        // forward calls between the left and right hand set by using adaptors.
        // This edge is known as the split-edge.
        //
        // A node can only be split into two and only a single edge should will
        // exist between them.
        //
        // As an example, suppose the original (directed) graph is:
        //
        //  +----------------------------------+
        //  |                                  |
        //  |      2 >-+         +-< 3         |
        //  |           \       /              |
        //  |            +- X -+               |
        //  |           /       \              |
        //  |      4 <-+         +-> 5         |
        //  |                                  |
        //  +----------------------------------+
        //
        // Thus:
        //     - a call on 2 would call behaviours bound on 2,X,4 and 5
        //     - a call on 3 would call behaviours bound on 3,X,4 and 5
        //     - a call on X would call behaviours bound on X,4 and 5
        //     - a call on 4 would call behaviours bound on 4
        //     - a call on 5 would call behaviours bound on 5
        //
        // (NOTE that there is no implication in the above about the order in
        // which they will be called.  In practice, there will be a difference
        // from the non-intercepted graph.)
        //
        // Suppose, for some reason, the user wants to partition the nodes 2 and
        // 4 into one set and the nodes 3 and 5 into a different set and observe
        // any transfer of calls between the sets.
        //
        // The user can do this by splitting node X into two:
        //
        //  +----------------------------------+
        //  |                                  |
        //  |  2 >-+                   +-< 3   |
        //  |       \                 /        |
        //  |        +- 0 <- - -> 1 -+         |
        //  |       /                 \        |
        //  |  4 <-+                   +-> 5   |
        //  |                                  |
        //  +----------------------------------+
        //
        // Where the node X has been split into node 0 and 1 the connecting edge
        // is called a 'split-edge' and the Interceptor can observe calls
        // transferred across that edge.
        //
        // The observed calls are the same except that any behaviour bound to
        // 'X' in the first diagram is replaced by any behaviours bound to
        // either 0 or 1:
        //     - a call on 2 would call behaviours bound on 2,0,1,4 and 5
        //     - a call on 3 would call behaviours bound on 3,0,1,4 and 5
        //     - a call on 0 would call behaviours bound on 0,1,4 and 5
        //     - a call on 1 would call behaviours bound on 0,1,4 and 5
        //     - a call on 4 would call behaviours bound on 4
        //     - a call on 5 would call behaviours bound on 5
        //
        // </201708110958>
        bool isSplitEdge() const;
        
    protected:
        EdgeLabel(Interceptor* interceptor_, bool split_edge_);
        
    private:
        EdgeLabel(EdgeLabel const&) = delete;
        void operator = (EdgeLabel const&) = delete;

        Interceptor* interceptor;
        bool         split_edge;
    };

    using EdgeLabelPtr = std::shared_ptr<EdgeLabel>;
    
    /// This is the information stored in a connector node from the Interceptor
    /// using the information from the EdgeLabel.
    ///
    /// It's destructor is responsible for removing all edges from the graph
    /// that connect to this node.
    ///
    /// It may not know the Interceptor until it starts participating in a graph
    /// that has been exposed to the Interceptor.  However, it is up to the
    /// Interceptor's implementation about when and if the InterceptorNode
    /// becomes bound to the Interceptor.
    ///
    /// The InterceptorNode resides in InvokeBase.
    class InterceptorNode
    {
    public:

        InterceptorNode(InvokeBase*, Connectable*);
        ~InterceptorNode();

        /// \name RegisterEdges Register an edge
        ///
        /// These are called as nodes are connected together.  They only tell the
        /// Interceptor about direct edges between groups and so the Interceptor
        /// must track any indirect connections that it needs by itself.
        ///
        /// The EdgeLabelPtr might be nullptr.
        ///
        /// The return value indicates what the rest of the connection code
        /// should do.
        ///
        /// If EdgeLabelPtr is non-null then this is an edge that has been
        /// labelled to be of interest to the Interceptor.  The Interceptor may
        /// in response, allow the normal edge connection process, deny it,
        /// claim that it has performed the connect, etc.
        ///
        /// If it claims that it has created the edge, then we normally refer to
        /// this edge as a _fake_ edge as the Interceptor has usually bound an
        /// adaptor (bindAdaptor) and arranged for the adaptor to forward the
        /// call across the logical edge -- but the edge itself does not exist
        /// in the Connector's graph.
        ///
        /// Even if the EdgeLabel is nullptr then if either end of the edge has
        /// been connected to the Interceptor then the Interceptor will still be
        /// called.  It is then up to the Interceptor on how to interpret a
        /// nullptr EdgeLabel.
        ///
        /// is_connectable_ is whether the two Connectors would be connectable
        /// in the absence of the presence of the Interceptor.
        ///
        //@{
        ConnectionAction_t installEdge(EdgeLabelPtr const&, Port& to_);
        ConnectionAction_t installEdge(EdgeLabelPtr const&, PeerConnectorBase& to_,     bool is_connectable_);
        ConnectionAction_t installEdge(EdgeLabelPtr const&, DirectedConnectorBase& to_, bool is_connectable_);
        //@}

        /// Disconnect all edges from this node.
        void disconnect();
        
        /// \name UnRegisterEdges Un-register an edge, returns true if was a fake edge.
        ///
        /// See the discussion of installEdge() on the meaning of a 'fake' edge.
        //@{
        bool uninstallEdge(Port& to_);
        bool uninstallEdge(PeerConnectorBase& to_);
        bool uninstallEdge(DirectedConnectorBase& to_);
        //@}

        Interceptor* getInterceptor() const { return interceptor; }
        Connectable* getConnectable() const { return connectable; }
        InvokeBase*  getInvokeBase()  const { return node;        }

        ///
        /// DirectedConnectors have a concept of 'local only'.
        ///
        /// This is called when DirectedConnectorBase::setLocalOnly() is called.
        /// It is up to the Interceptor to query the initial state of the
        /// 'LocalOnly' property of the DirectedConnector when initially it
        /// discovers the DirectedConnector.  From then on this function will
        /// keep the Interceptor informed of changes in this property.
        ///
        void setDirectedConnectorSetLocalOnly(bool local_);

        /// The data part for use by the implementation of the Interceptor.
        uint64_t     d;
        
        /// The data part for use by the implementation of the Interceptor.
        void*        p;

        /// A data part for use by the implementation that get copied into the
        /// d_for_CallList data member of a CallList::queue object when an
        /// adaptor sets up to execute a CallList.
        ///
        /// This is setup by the return result of rebuildCallList().
        uint64_t     d_for_CallList;

    private:
        friend class CallList::queue;
        // Interface used to execute the calls on an adaptor
        //
        // See the macros ARM_ADAPTOR_DIRECTED_CALL and
        // ARM_ADAPTOR_PEER_CALL.

        /// Checks to see if the CallList is currently valid and if not then
        /// rebuilds a new CallList, if it is 'in charge' then returns the
        /// CallList to call.  Otherwise returns nullptr.
        ///
        /// A node is 'in charge' if it is required to execute the CallList.
        /// This allows a single CallList to be shared between multiple nodes
        /// (and so the invalidating/rebuild behaviour can be shared) but only
        /// a subset of the nodes be responsible for executing it.
        ///
        /// NOTE that the lifetime of the CallList is managed by a shared_ptr<>.
        std::shared_ptr<CallList const> getCallList()
        {
            if (call_list == nullptr // first time round!
                || ! call_list->isValid())
                slow_getCallList();
            
            return in_charge ? call_list : nullptr;
        }

    private:
        InterceptorNode(InterceptorNode const&) = delete;
        void operator =(InterceptorNode const&) = delete;

        Interceptor*                     interceptor { nullptr };
        Connectable*                     connectable { nullptr };
        InvokeBase*                      node        { nullptr };

        /// This node is responsible for executing the CallList and therefore
        /// getCallList() will return non-null.  When the call list is rebuilt
        /// then we acquire the knowledge if we are in_charge or not.
        ///
        /// NOTE this piece of data is considered to belong to the adaptor.  It
        /// is not altered directly.  When the adaptor tries to determine if it
        /// is in_charge and should execute a CallList then it recognises that
        /// it needs to Interceptor::rebuildCallList() which returns if it is
        /// in_charge and the new CallList.
        bool                             in_charge { false };

        /// This is the call list for this node.  The node is tested for
        /// validity before being used.  Usually, all nodes that have been
        /// coloured by the Interceptor will have a non-nullptr, even those
        /// nodes that are not in charge.  However, the exact semantics are up
        /// to the Interceptor.
        ///
        /// NOTE that the interface is such that, when executing, it need only
        /// ever use the const interface.
        ///
        /// NOTE that this piece of state is considered to belong to the
        /// adaptor.  The pointer is not altered directly.  The object it points
        /// to (the const CallList) is not altered except for being made
        /// asynchronously invalid.
        ///
        /// See the notes for in_charge.
        std::shared_ptr<CallList const>  call_list;

        friend class Interceptor;

        /// Connect up the InterceptorNode to the Interceptor.  Called by
        /// the Interceptor::initInterceptorNode().
        void init(Interceptor* interceptor_, uint64_t d_);

        /// We have missed the local cached version of call_list so go and ask
        /// the interceptor what it should be.
        void slow_getCallList();
    };

    
    ///
    /// This is the return structure for the rebuildCallList() call.
    ///
    struct rebuildCallListResult_t {
        std::shared_ptr<CallList const>   call_list;
        bool                              in_charge;
        uint64_t                          d_for_CallList;

        rebuildCallListResult_t(
            std::shared_ptr<CallList const> const&  call_list_,
            bool                                    in_charge_,
            uint64_t                                d_for_CallList_
            ) : call_list(call_list_), in_charge(in_charge_), d_for_CallList(d_for_CallList_)
        {
        }            
    };
protected:
    friend class InterceptorNode;

    /// \name RegisterEdges Register an edge (used by InterceptorNode)
    ///
    /// These are called as nodes are connected together.  They only tell the
    /// Interceptor about direct edges between groups and so the Interceptor
    /// must track any indirect connections that it needs by itself.
    ///
    /// It can do this by following the connection graph of the Connectors and
    /// installing itself in the InterceptorNode parts of the Connectors that it
    /// discovers.  From then on, any change to the topology of the Connectors
    /// will be reported to the Interceptor.
    ///
    //@{
    virtual ConnectionAction_t installEdge(EdgeLabelPtr const&,
                                           Port& from_,
                                           Port& to_) = 0;
    virtual ConnectionAction_t installEdge(EdgeLabelPtr const&,
                                           PeerConnectorBase& from_,
                                           PeerConnectorBase& to_,
                                           bool is_connectable_) = 0;
    virtual ConnectionAction_t installEdge(EdgeLabelPtr const&,
                                           DirectedConnectorBase& from_,
                                           DirectedConnectorBase& to_,
                                           bool is_connectable_) = 0;
    //@}

    /// \name DisconnectNode Remove the node entirely from its graph.
    //@{
    virtual void disconnect(InvokeBase*) = 0;
    //@}

    /// \name RegisterEdges Un-register an edge (used by InterceptorNode)
    //@{
    virtual bool uninstallEdge(Port& from_, Port& to_) = 0;
    virtual bool uninstallEdge(PeerConnectorBase& from_, PeerConnectorBase& to_) = 0;
    virtual bool uninstallEdge(DirectedConnectorBase& from_, DirectedConnectorBase& to_) = 0;
    //@}

    /// The DirectedConnectorBase::setLocalOnly(local_) has been called, which
    /// calls InterceptorNode::setDirectedConnectorSetLocalOnly() which calls
    /// this.
    ///
    /// When a DirectedConnector is initially found then the Interceptor should
    /// query its LocalOnly property, once it has installed itself in the
    /// InterceptorNode then any _changes_ to the LocalOnly property will be
    /// notified to the Interceptor via this call.
    virtual void setDirectedConnectorSetLocalOnly(InterceptorNode&, bool local_) = 0;
    
    /// The specified node is being destroyed.  At this moment, the node has
    /// been partially destroyed so do not access any fields in it!
    virtual void nodeBeingDestroyed(InterceptorNode&) = 0;
        
    /// The CallList in the InterceptorNode, either does not exist or has
    /// declared itself to no longer be valid.  Rebuild a new CallList and
    /// return it.
    ///
    /// The result's in_charge data element is true if the node is expected to
    /// be in charge (and so execute the CallList).  If it is false then the
    /// node is not expected to execute anything (it is not 'in charge').
    ///
    /// The result's d_for_CallList is used to set node_.d_for_CallList.
    virtual rebuildCallListResult_t rebuildCallList(InterceptorNode& node_) = 0;

private:
    friend class ConnectorBase; // for disconnectFrom

protected:

    // These allow the derived implementation call bindAdaptor() and
    // unbindAdaptor() on the Connectable.
    void bindAdaptor(Connectable&);
    void unbindAdaptor(Connectable&);

    // These allow the derived implementation to initialise the InterceptorNode
    void initInterceptorNode(InterceptorNode&, uint64_t d_);
};

static inline std::ostream& operator << (std::ostream& os_, Interceptor const& t_)
{
    t_.print(os_);
    return os_;
}

/*! A Protocol ID
 *
 *  Protocol IDs are strings and are namespaced.
 *
 *  ARM reserves the namespace "ARM".
 *
 *  In LISA Protocols the property:
 *       protocol_namespace = "...";
 *  says that the protocol's name is in the specified namespace,
 *  and thus has a semantic meaning.
 *
 *  Protocol IDs in C++ versions of the protocol must call:
 *       setProtocolID("namespace.protocol_name");
 *
 *  Namespaces may be hierarchical and the namespace names are C-identifiers
 *  separated by '::'.
 *
 *  Examples:
 *      ARM::sg.Signal  -- the ARM::sg namespace, the Signal protocol in it.
 *      ARM::pv.PVBus   -- the ARM::pv namespace, the PVBus protocol in it.
 *
 *  There may be a correspondance between the namespace and the C++ namespaces
 *  but such a correspondance is not guaranteed.
 *
 *  In general, only the Ports themselves with have a ProtocolID, however, for
 *  specific purposes then a Connector might have a ProtocolID.
 *
 *  If no namespace is declared, either in the LISA file or in the C++ version
 *  then the protocol has no 'official' name.
 *
 */
class ProtocolID
{
public:

    ProtocolID() {}
    explicit ProtocolID(std::string const& id_) : id(id_) {}

    bool isValid() const                                          { return ! id.empty();   }
                                                              
    bool operator == (ProtocolID o_) const                        { return id == o_.id;    }
    bool operator != (ProtocolID o_) const                        { return !(*this == o_); }
                                                              
    // Enable sorting.                                            
    bool operator < (ProtocolID o_) const                         { return id < o_.id; }

    std::string asString() const                                  { return id; }
    
private:
    std::string id;
};

std::ostream& operator <<(std::ostream& os_, ProtocolID const& t_);
    
/*! A generic connectable/disconnectable thing. */
class Connectable : public Node
{
 protected:
    Node *parent;
    std::string name;
    
    Connectable() : parent(NULL), name("(unused connector)") { }

public:
    /*! Attach this connector to a remote connector, checking that it has
     *  exactly the same type.
     */
    virtual bool connectTo(Connectable *target) = 0;
    
    virtual bool connectToVia(Connectable *target, Interceptor::EdgeLabelPtr const&) = 0;

    // check whether it would be possible to connect two ports
    virtual bool canConnectTo(const Connectable* target) const = 0;

    /*! Detach this connector from a remote connector, checking that it has
     *  exactly the same type.
     */
    virtual bool disconnectFrom(Connectable *target) = 0;

    /*! Disconnect this object from everything it is connected to.
     */
    virtual void disconnect() = 0;

    void setName(std::string const &n) { name = n; }
    virtual std::string getName() const { return name; }
    virtual std::string getClassName() const;
    virtual const Node * getParent() const { return parent; }
    virtual void setParent(Node *p) { parent = p; }

    virtual void dump() = 0;

    // Fale dynamic dispatch from Node
    virtual Connectable* asConnectable() { return this; }
    virtual const Connectable* asConnectable() const { return this; }

    // Fake dynamic multiple dispatch
    virtual Port *asPort() { return NULL; }
    virtual PeerConnectorBase *asPeerConnector() { return NULL; }
    virtual DirectedConnectorBase *asDirectedConnector() { return NULL; }
    virtual SimpleConnectorBase* asSimpleConnector() { return NULL; }
    virtual const Port *asPort() const { return NULL; }
    virtual const PeerConnectorBase *asPeerConnector() const { return NULL; }
    virtual const DirectedConnectorBase *asDirectedConnector() const{ return NULL; }
    virtual const SimpleConnectorBase* asSimpleConnector() const { return NULL; }

    ProtocolID        getProtocolID() const;
    void              setProtocolID(ProtocolID protocol_id_);

    /// Set the Protocol ID by string.  Returns true if string was valid.
    bool              setProtocolID(std::string const&);
    
    /// Set the Protocol ID by string.
    ///
    /// In debug builds will assert fail if the string is unparseable.
    void              setProtocolIDWithAssert(std::string const&);

    /// Make a new Connectable of the same kind (but it doesn't have any of the
    /// bindings as the existing one).
    virtual Connectable* makeSameKind() = 0;
    
private:
    ProtocolID protocol_id;

    friend class Interceptor;

    /// This is a call from the Interceptor to ask for the adaptor to be bound.
    virtual void bindAdaptor() = 0;

    /// This is a call from the Interceptor to ask for the adaptor to be unbound.
    virtual void unbindAdaptor() = 0;
};

#ifdef _MSC_VER
#pragma warning(disable:4250) // implementing a virtual function in one of two classes derived from a common virtual base.
#endif

/*! A base class for manipulating connectors without knowing about their
 *  connection policy or their templated invocation argument list.  At
 *  this level, all bindings are handled via generic DelegateMemento
 *  lists.
 *
 *  Connectors provide two lists of bindings: local bindings, and default
 *  bindings.  The local bindings can be exported to other connectors in
 *  a graph, according to a connection policy.  This may then generate a
 *  set of active bindings to be called when this connector is invoked.
 *  If the active bindings list is empty, then the connector's default
 *  bindings are called.
 *
 *  All concrete connector types inherit from ConnectorBase via two paths:
 *  one path manages the propagation rules (peer-to-peer or directed)
 *  on abstract delegate types; the other path provides support for the
 *  formal arguments and return types.
 *
 *  The latter is provided by the ConcreteConnectorBase templated class,
 *  which provides an invoke() method that will call all of the delegates
 *  in the active_binding_list, or, if that list is empty, the
 *  default_bindings list.  The updateInvocableBehaviour() method is
 *  implemented in ConcreteConnectorBase, to precompute a delegate to
 *  handle the invoke method efficiently in the case where there is only
 *  one behaviour in the list.
 *
 *  The propagation rules are handled by implementing the virtual methods
 *  propagateNewBinding(), propagateRemovedBinding(), connectTo() and
 *  disconnectFrom().  These are called by
 *  the connectTo()/disconnectFrom() methods
 */
 class ConnectorBase : public Connectable
{
public:

    using CallID = ConnectorCallID;
    
protected:
    /*! Provide a local name for the type representing 'something that
     *  can be bound, copied, compared and invoked'.
     */
    typedef fastdelegate::DelegateMemento Binding;

    /*! A collection of bound delegates. */
    struct BindingList
    {
        Interceptor*                interceptor { nullptr };
        
        std::list<Binding>          call_id_bindings;
        std::list<Binding>          normal_bindings;

        using iterator = std::list<Binding>::iterator;

        /// Is the BindingList empty of any bindings.
        bool empty() const { return call_id_bindings.empty() && normal_bindings.empty();          }

        /// Clear the whole container.
        void clear();

        // Copy the bindings from one list onto this list.
        void push_back(BindingList const& o_);

        bool isThereOnlyASingleNormalBinding() const
        {
            return call_id_bindings.empty() && normal_bindings.size() == 1;
        }

        Binding& getSingleNormalBinding()
        {
            assert(isThereOnlyASingleNormalBinding());
            return *normal_bindings.begin();
        }

        // remove first normal binding matching
        void removeBinding(Binding const& old_binding);

        // remove first CallID binding matching
        void removeCallIDBinding(Binding const& old_binding);
    };

    /*! A collection of pointers to delegates for vector behaviours. */
    typedef std::list<IndexedDelegateMemento *> IndexedBindingList;

 protected:
    /*! The list of active bindings to call when invoked.  If this list is
     *  not defined, or is empty, then the default bindings list will be
     *  invoked instead.
     *
     *  NB The list is a cache that's lazily updated on demand, including
     *  from some const methods.
     */
    mutable BindingList *active_behaviour_list;

    /*! Check whether the active behaviour list gets modified in the
     *  middle of an invocation.
     */
    bool active_list_locked;

    /*! The set of bindings attached to this connector. */
    BindingList local_bindings;

    /*! The set of vector behaviours attached to this connector. */
    IndexedBindingList local_indexed_bindings;

    /*! Default local behaviour bindings, to be invoked if there are no
     *  bindings in the peer group.
     */
    BindingList default_bindings;

    ConnectorBase();

    /*! CallID passed to the bindings that want the CallID for the first parameter. */
    CallID call_id;

    /*! CallID that will be generated next.
     *
     *  This is used, in part, to implement invokeWithCallID:
     *
     *      - call_id is always the version that is given to any CallID bindings.
     *
     *      - next_call_id is always the next CallID to use when we consume call_id.
     *
     *      - invokeWithCallID knows whether there is a binding
     *
     */
    CallID next_call_id;

    /*! Get and Consume the current CallID */
    CallID getAndConsumeCallID() { CallID ret = call_id; call_id = next_call_id; next_call_id.increment(); return ret; }

    /*! If call_id might have been overridden by invokeWithCallID() then restore
     *  the correct version. */
    void restoreCallID() { call_id = next_call_id.previous(); }

 protected:
    /*! Allow the connection policy to take care of propagating a
     *  newly-added local binding.
     */
    virtual void propagateNewBinding(const Binding &new_binding) = 0;

    /*! Allow the connection policy to take care of retracting a
     *  removed local binding.
     */
    virtual void propagateRemovedBinding(const Binding &old_binding) = 0;

    /*! Allow the connection policy to take care of propagating a
     *  newly-added local binding.
     */
    virtual void propagateNewCallIDBinding(const Binding &new_binding) = 0;

    /*! Allow the connection policy to take care of retracting a
     *  removed local binding.
     */
    virtual void propagateRemovedCallIDBinding(const Binding &old_binding) = 0;

    /*! Add a binding to the local bindings list, and pass it on to the
     *  connection propagation policy.  The concrete subclasses provide
     *  public versions of this method that take care of checking that the
     *  parameter lists match.
     */
    void addBinding(const Binding &new_binding);

    /*! Remove an existing binding. */
    void removeBinding(const Binding &old_binding);

    /*! Add a binding for a vector behaviour.  The templated subclass
     *  is responsible for capturing the vector behaviour as an
     *  IndexedDelegateMemento object so we can manipulate it.
     *  Note: We take ownership of the IndexedDelegateMemento object.
     */
    void addIndexedBinding(IndexedDelegateMemento *indexed_binding);

    /*! Delete a vector behaviour that matches the given binding. */
    void deleteIndexedBinding(IndexedDelegateMemento const &pattern);

    /*! Add a binding for a CallID behaviour.  The templated subclass
     *  is responsible for capturing the CallID behaviour as an
     *  CallIDDelegateMemento object so we can manipulate it.
     *  Note: We take ownership of the CallIDDelegateMemento object.
     */
    void addCallIDBinding(const Binding &new_binding);
    
    /*! Remove an existing CallID binding. */
    void removeCallIDBinding(const Binding &old_binding);

    /*! Add a binding to the local defaults list, to be invoked if the
     *  active binding list is empty. */
    void addDefaultBinding(const Binding &new_binding);

    /*! Remove a default binding. */
    void removeDefaultBinding(const Binding &old_binding);

public:
    /*! Test if this connector can be connected to the specified target
     * connector.
     */
    bool isConnectable(const ConnectorBase* target) const;

 protected:
    /*! Make sure invocable_behaviour is the correct delegate for the
     *  current active_behaviour_list / default_behaviours.
     */
    virtual void updateInvocableBehaviour() = 0;

    virtual BindingList& getActiveBehaviourList() = 0;


    // For connectors used to implement ports, returns the enclosing port
    Port* parentPort();

 public:
    /*! Allow subclasses to destroy themselves cleanly. */
    virtual ~ConnectorBase();

    /*! Are there no active or default behaviours on this connector? */
    bool empty() const;

    /*! Inverted alias for empty() which is suitable for LISA code. LISA users
     * do not know what a connector is and that a connector can be 'empty'.
     * Instead they can use if (read.implemented()) ... to check whether
     * behavior read is implemented.
     */
    bool implemented() const { return !empty(); }

    /*! Copy local bindings into another list - used to propagate our
     *  bindings in accordance with a connection policy.
     */
    void copyBindingsToEndOf(BindingList &target);

    /*! Get the Token used in CallIDs invoked via this connector.
     */
    CallID::Token getToken() const { return next_call_id.getToken(); }
};

/// InvokeBase is a unique base class of the connector.
///
/// This is so that the list of calls to make, as returned by getCallList(), is
/// a list of InvokeBase* and can be static_cast<>() to the true connector class
/// and hence can call invokeWithCallID() on the connector class.
///
/// Whilst it does inherit from ConnectorBase, it is unsuitable as it is a
/// virtual base and it is not possible to cast from a virtual base to the
/// derived type.
class InvokeBase
{
public:
    InvokeBase(Connectable*);

    /*! Get the InterceptorNode, this is expected to be called only by the
     *  Interceptor */
    Interceptor::InterceptorNode& getInterceptorNode() { return interceptor_node; }
private:
    /*! Argument to the guard called by the adaptor.  This is filled out by the
     *  Interceptor when the adaptor is bound to the port. */
    Interceptor::InterceptorNode interceptor_node;
};

inline Interceptor::CallList::queue::queue(
    Interceptor::InterceptorNode& interceptor_node_,
    Interceptor::CallID const&    call_id_
    )
    : // NOTE that we _copy_ call_list and d_for_CallList_ from the
      // interceptor_node_ as the fields might change if the node gets
      // recoloured whilst we are executing the queue.
      call_list(interceptor_node_.getCallList()),
      d_for_CallList(interceptor_node_.d_for_CallList),
      interceptor_node(interceptor_node_),
      call_id(call_id_)
{
    // Q: Should we have an initial virtual call to CallList to
    //    setup the queue?
    // A: If we did then we would need a second virtual function call
    //    to getNextInvokeBase() and we must try and reduce the cost
    //    as much as we can.   In reality, the implementation of
    //    getNextInvokeBase() is quite likely to need a counter and
    //    data[0] is as good as anything.  There seems little benefit
    //    to trying to do anything more complicated.
    data[0] = 0;
}

} // namespace sg

/*! Reduce compilation times and object sizes by allowing the compiler
 *  to not instantiate connector classes in every compilation unit.
 *  Any header-file that declares a protocol class can use the macros
 *  in DirectedConnector.h to mention each behaviour signature used.
 *  By default, these macros will tell the compiler to treat each
 *  template as being instantiated externally.
 *  One compilation unit MUST predefine SG_INSTANTIATE_CONNECTORS prior
 *  to including any SG connector header files.  This will force the
 *  compiler to emit all required template definitions.
 */
#ifdef SG_INSTANTIATE_CONNECTORS
    /*! Force connector templates to be instantiated in this compilation unit. */
    #define SG_MENTION_TEMPLATE_CLASS template class
#else
    /*! Force connector templates to be marked as external to this compilation unit. */
    #define SG_MENTION_TEMPLATE_CLASS extern template class
#endif

namespace details_process_call_list {

    using Interceptor     = sg::Interceptor;
    using InterceptorNode = Interceptor::InterceptorNode;
    using CallID          = Interceptor::CallID;
    
    // The code for executing the CallList is almost identical, varying only on
    // if the calls return a value or not.  This is done by using INVOKE_PREFIX
    // to capture the return value if one exists.
    //
    // NOTE the use of 'special' CallIDs, see  <201708100704/>.
    //
#define ARM_EXECUTE_CALL_LIST( INVOKE_PREFIX )                          \
                                                                        \
    do                                                                  \
    {                                                                   \
        if (IsPeerConnector                                             \
            && call_id_.isSpecialCallID())                              \
            /* Self-call. */                                            \
            break;                                                      \
                                                                        \
        Interceptor::CallList::queue q(interceptor_node_, call_id_);    \
        if (q.hasCalls())                                               \
        {                                                               \
            auto const call_id_to_forward                               \
                = IsPeerConnector                                       \
                ? CallID::makeSpecial()                                 \
                : CallID::makeSpecial(interceptor_node_.getInvokeBase()) \
                ;                                                       \
                                                                        \
            while (sg::InvokeBase* base = q.pop_and_return_front())     \
            {                                                           \
                INVOKE_PREFIX static_cast<ConnectorType*>(base)         \
                    ->invokeWithCallID(                                 \
                        call_id_to_forward, std::forward<Args_>(args_)... \
                        );                                              \
            }                                                           \
        }                                                               \
    } while (0)

    // <201708100704>
    //
    // IsPeerConnector is true if this is a PeerConnector.
    //
    // If it is, then we have to distinguish calls from that were propagated
    // from outside of the current graph component (via fake edges) from those
    // that originated within the graph component.
    //
    // For example, if we have nodes A, B, C, X, Y, Z connected as follows:
    //
    //       A <---> B <---> C <- -> X <---> Y <---> Z
    //
    // where C <- -> X is the fake edge between the two graph components (ABC
    // and XYZ).  The other edges are real edges.
    //
    // The intended effect is that a call on any of the nodes will call all of
    // the behaviours bound to all of the nodes.
    //
    // If an interceptor has created a fake edge between C and X then it has
    // bound adaptors on both C and X.
    //
    // Say that the behaviour is invoked on one of ABC, then the adaptor on C
    // will be called.  It will then have to forward it to the graph XYZ
    // (typically by invoking the behaviour on X).  The behaviours on the graph
    // XYZ will then be invoked, this will include the adaptor bound to X.  That
    // adaptor must notice that it should not propagate the call back along the
    // edge X to C, otherwise the calls will continue indefinitely.  Similarly,
    // if the call originated in the XYZ graph component then the adaptor on C
    // has to similarly not forward back to X.
    //
    // When an adaptor forwards a call across a fake edge then it uses
    // CallID::makeSpecial().  When an adaptor receives a call with that special
    // CallID then it knows that it is not responsible for forwarding it across
    // the edge.  Thus the first in_charge node that a call reaches is
    // responsible for calling all other graph components in the complete graph.
    // The Interceptor has to create a CallList that obeys this property.
    //
    // </201708100704>
    //
    // For a directed connector, then each edge that crosses a fake edge that
    // isn't a split-edge, acts near-enough independently and so there is no
    // need for this complication.  For split-edges, then it is complicated and
    // needs the Interceptor to co-operate, see <201708100708/>.  See also
    // <201708110958/>.
    //
    template<typename ConnectorType, typename RetType, bool IsPeerConnector>
    struct process_call_list {
        template<typename... Args_>
        static RetType exec(
            InterceptorNode&              interceptor_node_,
            CallID                        call_id_,
            Args_&&...                    args_
            )
        {
            RetType ret {};

            ARM_EXECUTE_CALL_LIST( ret = );
            
            return ret;
        }
    };
    
    template<typename ConnectorType, bool IsPeerConnector>
    struct process_call_list<ConnectorType, void, IsPeerConnector> {
        template<typename... Args_>
        static void exec(
            InterceptorNode&              interceptor_node_,
            CallID                        call_id_,
            Args_&&...                    args_
            )
        {
            // We need some actual arguments to this that effectively do nothing
            ARM_EXECUTE_CALL_LIST( ; );
        }
    };

#undef ARM_EXECUTE_CALL_LIST
}

/*! Define the code fragment for use by the adaptors executing the CallList.
 *
 * We define multiple versions to use the implementation specific behaviours of
 * dealing with passing no arguments for the ... argument.
 *
 * In order to avoid an extraneous ',' then on g++ you need to token paste
 * __VA_ARGS__ onto the comma.  In Windows, then it silently just removes the
 * ','.
 *
 */
#ifdef __GNUC__
#define ARM_ADAPTOR_DIRECTED_CALL(RetType, call_id_, ...)  \
    return ::details_process_call_list::process_call_list<ConnectorType, RetType, /*IsPeerConnector*/false>::exec(getInterceptorNode(), call_id_, ## __VA_ARGS__)
#define ARM_ADAPTOR_PEER_CALL(RetType, call_id_, ...)      \
    return ::details_process_call_list::process_call_list<ConnectorType, RetType, /*IsPeerConnector*/true>::exec(getInterceptorNode(), call_id_, ## __VA_ARGS__)
#elif _MSC_VER
#define ARM_ADAPTOR_DIRECTED_CALL(RetType, call_id_, ...)  \
    return ::details_process_call_list::process_call_list<ConnectorType, RetType, /*IsPeerConnector*/false>::exec(getInterceptorNode(), call_id_, ## __VA_ARGS__)
#define ARM_ADAPTOR_PEER_CALL(RetType, call_id_, ...)      \
    return ::details_process_call_list::process_call_list<ConnectorType, RetType, /*IsPeerConnector*/true>::exec(getInterceptorNode(), call_id_, ## __VA_ARGS__)
#else
#error "Unknown compiler"
#endif

#endif
