/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   DynLib.h
 *  \brief  DSO Management Interface
 *  \date   Copyright ARM Limited 2009 All Rights Reserved.
 *          Copyright ARM Limited 2011 All Rights Reserved.
 *          Copyright ARM Limited 2013 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 */

#ifndef INCLUDED_FSDYNLIB_H
#define INCLUDED_FSDYNLIB_H

#include <string>
#include <sstream>

#include "eslapi/eslapi_stdint.h"

#include "sg/SystemInclude.h"

namespace sg {

#ifdef _WIN32
typedef HMODULE DynHandle;
#else
typedef void * DynHandle;
#endif

/*! Get the fully qualified directory of the host executable
 * \param   loc string to accept the location
 * \return  0 on success or a value to be translated by sg::TranslateError
 */
int GetExecutableDir(std::string & loc);

/*! Get the fully qualified directory containing the object containing a symbol
 * \param    sym    address of the symbol
 * \param    loc    string to accept the location
 * \return   0 on sucess or a value to be translated by sg::TranslateError
 *           DynError may also contain information.
 */
int GetSymbolDir(void* sym, std::string & loc);

/*! Get the most recent error associated with dynamic library loading
 * \return  The human readable text of the most recent error or a null string
 *          if there was no error.
 */
std::string DynError();

/*! Open a dymanic library
 * \param   name    string of the library filename
 * \return  Handle of the library, or NULL upon failure
 */
DynHandle DynOpen (const std::string & name);

/*! Open a dynamic library relative to object containing sym or standard search
 *  path
 * \param   sym     symbol to load relative to the object containing
 * \param   name    full name of object to load including suffix
 */
DynHandle DynOpenRelativeOrSearched(void* sym, const std::string & name,std::ostringstream & err);

/*! Close a dymanic library
 * \param   handle of the library to be closed
 * \return  Success of the operation
 */
bool DynClose(DynHandle handle);

/*! Look up symbol in a library
 * \param   handle of the library
 * \param   symbol string to look up
 * \return  Address of the desired symbol
 */
void* DynSym(DynHandle handle, const std::string & symbol);

/*! Look up symbol in all libraries, including the executable, in default search order
 * \param   symbol string to look up
 * \return  Address of the desired symbol
 */
void* DynSymDefault(const std::string & symbol);

/*! Look up address in all libraries, including the executable, in default search order
 * \param   address to lookup
 * \param   symbol name if found
 * \param   offset of address from actual symbol address if found
 * \return  returns true if symbol name found
 */
bool DynAddrDefault(const void* address, std::string& symbol, intptr_t* offset);

}

#endif

/* eof DynLib.h */
