/*!
 *  \file   FsFastVectorOfPOD.h
 *  \brief  A dynamic vector container that includes some statically pre-allocated elements
 *  \date   Copyright ARM Limited 2013 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 *  The use case for this class is where the majority of the time the number
 *  of elements is small, but could be unbounded.
 *
 *  Template arguments select pre-allocated size and enable/disable profiling
 */

#ifndef INCLUDED_SG_FAST_VECTOR_OF_POD_H
#define INCLUDED_SG_FAST_VECTOR_OF_POD_H


#include <cstdlib>
#include <stdexcept>
#include <vector>

#include "sg/SizeAndConstructionProfiler.h"


namespace sg {


//! A dynamic vector container that includes some statically pre-allocated element
// \param T the POD type of the vector elements
// \param SIZE the size of the static allocation (this has performance not functional implications)
// \param PROFILE when true enable the built in profiler
template <typename T, size_t SIZE=4, bool PROFILE=false>
class FastVectorOfPOD
{
private:
    typedef FastVectorOfPOD<T,SIZE,PROFILE>  This;

    size_t          elements;
    T               array[SIZE];
    std::vector<T>  dynamic;

    //! Return a reference to the element at position index
    T& get(size_t index)
    {
        return index < SIZE ?  array[index] : dynamic[index - SIZE];
    }

    //! Return a const reference to the element at position index
    const T& get(size_t index) const
    {
        return index < SIZE ?  array[index] : dynamic[index - SIZE];
    }

public:
    FastVectorOfPOD() : elements(0)
    {
        if (PROFILE) fast_vector_profiler.construct();
    }

    ~FastVectorOfPOD()
    {
        if (PROFILE) fast_vector_profiler.destruct();
    }

    //! Return the total number of elements
    size_t size() const { return elements; }

    //! Return true if there are no elements
    bool empty() const { return size() == 0; }

    //! At a new element at the end
    void push_back(const T& val)
    {
        if (elements < SIZE)
        {
            array[elements] = val;
        }
        else
        {
            dynamic.push_back(val);
        }

        ++elements;

        if (PROFILE) fast_vector_profiler.sample(elements);
    }

    //! Remove the last element
    void pop_back()
    {
        if (--elements >= SIZE)
        {
            dynamic.pop_back();
        }

        if (PROFILE) fast_vector_profiler.sample(elements);
    }

    //! Clear
    void clear()
    {
        if (elements > SIZE)
        {
            dynamic.clear();
        }

        elements = 0;
    }

    //! Return a reference to the first element
          T& front()       { return get(0); }
    const T& front() const { return get(0); }

    //! Return reference to the element at position index
          T& operator[](size_t index)       { return get(index); }
    const T& operator[](size_t index) const { return get(index); }

    //! Return a reference to the element at position index
    T& at(size_t index)
    {
        if (index >= elements)
            throw std::out_of_range("FastVectorOfPOD::_M_range_check");

        return get(index);
    }
    const T& at(size_t index) const
    {
        if (index >= elements)
            throw std::out_of_range("FastVectorOfPOD::_M_range_check");

        return get(index);
    }

    //! Return a reference to the last element
          T& back()       { return get(elements - 1); }
    const T& back() const { return get(elements - 1); }


    //! Transfer elements from that to back of this
    void splice_back(This& that)
    {
        for(unsigned i=0; i<that.size(); ++i)
        {
            push_back(that[i]);
        }

        that.clear();
    }

    //! Return the index of the first element that equals 'val', starting at 'first'
    size_t find(size_t first, T const & val) const
    {
        for(; first < size() ; ++first)
        {
            if (get(first) == val)
                return first;
        }
        return size();
    }
};

} // namespace sg

#endif
