/*!
 * \file    MTDomainGate.h
 * \brief   A gate between two MTDomains.
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 *
 * This defines a private interface for the use of by ARM.  The interface is
 * subject to change.
 *
 */
#ifndef ARM_INCLUDE_sg_MTDomainGate_h
#define ARM_INCLUDE_sg_MTDomainGate_h

#include <cstddef>
#include <string>
#include <vector>

#include "sg/MTDomainID.h"

namespace sg
{
    ///
    /// This is called when transitioning between two MTDomains.
    ///
    /// It is meant to be used RAII-style.
    class MTDomainGate
    {
    public:
        MTDomainGate(MTDomainID incoming_mtdomain_);
        ~MTDomainGate();

        static MTDomainID identifyCurrentDomainAs(std::string const& name_);
        
    private:
        MTDomainID outgoing;
    };

    class ComponentBase;
    class ComponentFactory;
    class Params;
    class SimulationContext;

    struct MTComponentWrapperResult {
        /// The MTDomainID of this component.
        MTDomainID             mtdomain_id;
        
        /// This is the instantiated component without the wrappers.  You should
        /// only talk to this components via MTDomainGates.
        ComponentBase*         raw_component;

        /// This is the component that most users should use to access the
        /// component.  It proxies the calls and transitions between domains.
        ComponentBase*         wrapper_component;
    };


    ///
    /// The parsed tool configuration information passed to MTComponentWrapper
    /// to tell it how to wrap the component.
    ///
    class MTWrapperConfig
    {
    public:
        //
        // This is the 'mtdomains_tool_config' property that we are using to
        // configure this component.  This is the mtdomains_tool_config property
        // of the component in which the component is being instantiated.
        //
        // These strings are usually only expected to be used when reporting
        // errors in the config items.
        //
        std::string mtdomains_tool_config_raw;
        std::string mtdomains_tool_config_post_interpolation;

        //
        // The mtdomains_tool_config property is substituted from parameters (of
        // the component whose mtdomains_tool_config we are parsing) and then
        // parsed as:
        //
        //       item0(option0, option1), ...
        //
        // Options may contain '(' and ')' but only if they balance.
        // Options may contain ',' but only within balanced brackets.
        //
        class item_t {
        public:

            struct string_and_pos_t {
                std::string                str;

                // This is the position in
                // mtdomains_tool_config_post_interpolation where str was sliced
                // from.
                std::string::size_type     begin;
            };

            /// The item's name.
            string_and_pos_t               item_name;

            /// The list of comma separated strings that made up the item's
            /// options.
            ///
            /// There is at least one option, even if that is empty.  All the
            /// arguments have preceding and trailing whitespace removed.
            std::vector<string_and_pos_t>  options;
        };
        std::vector<item_t> items;

        //
        // There was an error whilst assembling this information.
        //
        bool error { false };
    };
    
    /// Instantiate and wrap a component with an MT-safe wrapper.
    MTComponentWrapperResult MTComponentWrapper(
        ComponentFactory*                      factory_,
        MTWrapperConfig const&                 wrapper_config_,
        std::string const&                     mtdomain_name_,
        std::string const&                     instance_name_,
        SimulationContext*                     simulation_context_,
        Params const&                          params_
        );
}

#endif
