/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   Macros.h
 *  \brief  Preprocessor Macros
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  Some preprocessor 'tricks' to make common operations more terse or more
 *  portable.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 */

#ifndef INCLUDED_MACROS_H
#define INCLUDED_MACROS_H


/*! Convert the parameter into a string literal. */
#define SG_STRINGIZE(s) #s
/*! Expands the parameter and then stringize it stringized. */
#define SG_XSTRINGIZE(s) SG_STRINGIZE(s)

/*! Ensure the parameter of this macro is a string literal at compile time.
 *
 * \param s     A string literal
 */
#define SG_ENSURE_STRING_LITERAL(s) static_assert((s "") != nullptr, "must be string literal")
// The static assert is mostly to allow it to be used in non-expression contexts.

/*! Portable drop-in replacement for x##y in macros which works also with __LINE__.
 * Use:
 *  SG_CONCAT_IDENTIFIER(myvar_, __LINE__)
 * rather than
 *  myvar_##__LINE__
 * which results in
 *  myvar___LINE__
 */
#define SG_CONCAT_IDENTIFIER(x,y) SG_CONCAT_IDENTIFIER_internal(x,y)
#define SG_CONCAT_IDENTIFIER_internal(x,y) x##y

/* The SG_DISABLE_DIAG macros allow temporary disabling of diagnostics for
 * small sections of a source file, rather than using CFLAGS;
 * You will notice that the SG_[compiler_]DISABLE_DIAG_END macros do not expect
 * parameters, as they basically act as pragma push/pop.
 *
 * Diagnostics for all supported compilers
 * ---------------------------------------
 *   Use `SG_DISABLE_DIAG("-Wmy-warning")` to disable a warning for the entire
 *   source file.
 *   Use `SG_DISABLE_DIAG_BEGIN("-Wmy-warning")` to begin a section for which a
 *   warning should be disabled, and `SG_DISABLE_DIAG_END()` to end the section.
 *
 * Diagnostics for specific compilers
 * ----------------------------------
 *   These are basically the same as diagnostics for all targeted compilers,
 *   but will only be issued if the source file is compiled with that compiler.
 *   The commands are:
 *    SG_GCC_DISABLE_DIAG("-Wmy-warning")
 *    SG_GCC_DISABLE_DIAG_BEGIN("-Wmy-warning")
 *    SG_GCC_DISABLE_DIAG_END()
 *    SG_CLANG_DISABLE_DIAG("-Wmy-warning")
 *    SG_CLANG_DISABLE_DIAG_BEGIN("-Wmy-warning")
 *    SG_CLANG_DISABLE_DIAG_END()
 * WARNING!
 *    Gcc is broken for the moment, I've tested from 4.6 (where "support" for
 *    pragma bits was introduced), through 4.7, 4.8 and 4.9; On none of these
 *    versions will the pragma stack work correctly (i.e. at all), regardless
 *    of whether added manually, or using `_Pragma`. This bug is allegedly
 *    fixed in gcc >= 5, so for versions older than that, both
 *    `SG_DISABLE_DIAG_BEGIN` and `SG_DISABLE_DIAG_END` are defined, but will
 *    not do anything. `SG_DISABLE_DIAG` however, will work as expected.
 */
#if defined(__clang__)
  #define _SG_DIAG_PRAGMA(x) SG_STRINGIZE(clang diagnostic ignored x)
  #define SG_DISABLE_DIAG(x) _Pragma(_SG_DIAG_PRAGMA(x))
  #define SG_DISABLE_DIAG_BEGIN(x) \
        _Pragma("clang diagnostic push") \
        SG_DISABLE_DIAG(x)
  #define SG_DISABLE_DIAG_END(x) _Pragma("clang diagnostic pop")
  #define SG_CLANG_DISABLE_DIAG(x) SG_DISABLE_DIAG(x)
  #define SG_CLANG_DISABLE_DIAG_BEGIN(x) SG_DISABLE_DIAG_BEGIN(x)
  #define SG_CLANG_DISABLE_DIAG_END() SG_DISABLE_DIAG_END()
  #define SG_GCC_DISABLE_DIAG(x)
  #define SG_GCC_DISABLE_DIAG_BEGIN(x)
  #define SG_GCC_DISABLE_DIAG_END()
#elif defined(__GNUC__) && (__GNUC__ >= 5)
  #define _SG_DIAG_PRAGMA(x) SG_STRINGIZE(GCC diagnostic ignored x)
  #define SG_DISABLE_DIAG(x) _Pragma(_SG_DIAG_PRAGMA(x))
  #define SG_DISABLE_DIAG_BEGIN(x) \
        _Pragma("GCC diagnostic push") \
        SG_DISABLE_DIAG(x)
  #define SG_DISABLE_DIAG_END(x) _Pragma("GCC diagnostic pop")
  #define SG_CLANG_DISABLE_DIAG(x)
  #define SG_CLANG_DISABLE_DIAG_BEGIN(x)
  #define SG_CLANG_DISABLE_DIAG_END()
  #define SG_GCC_DISABLE_DIAG(x) SG_DISABLE_DIAG(x)
  #define SG_GCC_DISABLE_DIAG_BEGIN(x) SG_DISABLE_DIAG_BEGIN(x)
  #define SG_GCC_DISABLE_DIAG_END() SG_DISABLE_DIAG_END()
#elif defined(__GNUC__)
  #define _SG_DIAG_PRAGMA(x) SG_STRINGIZE(GCC diagnostic ignored x)
  #define SG_DISABLE_DIAG(x) _Pragma(_SG_DIAG_PRAGMA(x))
  #define SG_DISABLE_DIAG_BEGIN(x)
  #define SG_DISABLE_DIAG_END()
  #define SG_CLANG_DISABLE_DIAG(x)
  #define SG_CLANG_DISABLE_DIAG_BEGIN(x)
  #define SG_CLANG_DISABLE_DIAG_END()
  #define SG_GCC_DISABLE_DIAG(x) SG_DISABLE_DIAG(x)
  #define SG_GCC_DISABLE_DIAG_BEGIN(x) SG_DISABLE_DIAG_BEGIN(x)
  #define SG_GCC_DISABLE_DIAG_END() SG_DISABLE_DIAG_END()
#else
  #define SG_DISABLE_DIAG(x)
  #define SG_DISABLE_DIAG_BEGIN(x)
  #define SG_DISABLE_DIAG_END()
  #define SG_CLANG_DISABLE_DIAG(x)
  #define SG_CLANG_DISABLE_DIAG_BEGIN(x)
  #define SG_CLANG_DISABLE_DIAG_END()
  #define SG_GCC_DISABLE_DIAG(x)
  #define SG_GCC_DISABLE_DIAG_BEGIN(x)
  #define SG_GCC_DISABLE_DIAG_END()

#endif

#if defined(_MSC_VER)

  #define SG_ALIGNOF(T) __alignof(T)
  // NOTE that MSVC can only accept integer literals for 'N'.
  #define SG_ALIGNAS(N) __declspec(align(N))

#elif defined(__GNUC__)

  #if defined(__clang__)
      #define SG_ALIGNAS(N) alignas(N)
  #elif __GNUC__ == 4 && __GNUC_MINOR__ <= 7
      // doesn't work until 4.8 and higher
      #define SG_ALIGNAS(N) __attribute__((aligned(N)))
  #else
      #define SG_ALIGNAS(N) alignas(N)
  #endif
  #define SG_ALIGNOF(T) alignof(T)

#else

  #error "Unsupported compiler"
    
#endif

/*! On builds with asserts enabled, equivalent to assert().
 *  On builds with asserts disabled, pass a hint to the compiler that the given
 *  expression is always true.
 *  The expression will be evaluated, so passing an expression with side
 *  effects is strongly discouraged since it will prevent it from being
 *  optimised away. */
#if defined(NDEBUG)
  #if defined(__GNUC__)
    #define SG_ASSUME(exp) do { if (!(exp)) __builtin_unreachable(); } while (0)
  #elif defined(_MSC_VER)
    #define SG_ASSUME(exp) do { (void)(exp); __assume(exp); } while (0)
  #endif
#else
  #define SG_ASSUME(exp) assert(exp)
#endif

/*! Mark this point in the code as unreachable.  */
#if defined(_MSC_VER)
    namespace
    {
        __declspec(noreturn) void sg_does_not_return() {}
    }
    #define SG_UNREACHABLE do { SG_ASSUME(0); sg_does_not_return(); } while (0)
#else
    #define SG_UNREACHABLE SG_ASSUME(0)
#endif

#endif

/* eof Macros.h */
