/*!
 * \file    PeerConnector.h
 * \brief   Base class for handling peer connector behaviour.
 * \date    Copyright 2006 ARM Limited. All rights reserved.
 * \author  Andrew Bolt and Graeme Barnes.
 *
 * Define a non-templated base class for peer-to-peer connectors.
 */

// THIS FILE IS GENERATED FROM PeerConnector.hxx

#ifndef PEERCONNECTOR_H
#define PEERCONNECTOR_H

#include "sg/ConnectorBase.h"

namespace sg {


/*! Manage a single peer connector, in a graph of connectors.  The connection
 *  graph must be acyclic.
 *
 *  All connected connectors share a pooled BehaviourPeerGroup: when a new
 *  behaviour is bound to any of the connectors in the group, all other
 *  connectors see it.
 *
 *  When a connection is made between two graphs, the BehaviourPeerGroups and
 *  pooled behaviours are merged.
 *
 *  When a connection is broken it creates two separate graphs; a new
 *  BehaviourPeerGroup is created, and everything on one side is moved
 *  from the old group into the new one.
 */
class PeerConnectorBase : public virtual ConnectorBase, public InvokeBase
{
protected:
    typedef std::list<PeerConnectorBase *> PeerList;

    /*! Construct an empty connector, with no connections. */
    PeerConnectorBase();

    /*! Handle collection of peers. */
    class BehaviourPeerGroup;

private:
    /*! Who we're directly connected to. */
    PeerList direct_peers;
    
    /*! Container for all connectors within a graph. */
    BehaviourPeerGroup *group;

    void leaveGroup();
    void connectToGroup(BehaviourPeerGroup* group);
    void constructGroup();
    void destroyGroup();
    void removePeer(PeerConnectorBase*);
    void addPeer(PeerConnectorBase*);
    void disconnectPeers();
    bool connectTo(Connectable* target, Interceptor::EdgeLabelPtr const&);

protected:

    virtual BindingList& getActiveBehaviourList() override;

protected: // Implement base-class behaviour.

    /*! Add a new binding to the group's list. */
    virtual void propagateNewBinding(const Binding &new_binding) override;

    /*! Remove a binding from the group's list. */
    virtual void propagateRemovedBinding(const Binding &old_binding) override;

    /*! Add a new binding to the group's list. */
    virtual void propagateNewCallIDBinding(const Binding &new_binding) override;

    /*! Remove a binding from the group's list. */
    virtual void propagateRemovedCallIDBinding(const Binding &old_binding) override;
public:
    /*! Print a debug log of the connector's contents. */
    virtual void dump() override;

    /*! Disconnected from our peer group cleanly. */
    virtual void disconnect() override;

    /*! Make sure we're disconnected when we die. */
    ~PeerConnectorBase();
    
    
    /*! Allow BehaviourPeerGroup to find out who we're connected to. */
    const PeerList &getPeers();

    virtual PeerConnectorBase *asPeerConnector() override;
    virtual const PeerConnectorBase *asPeerConnector() const override;

    virtual bool connectTo(Connectable *target) override;
    virtual bool connectToVia(Connectable *target, Interceptor::EdgeLabelPtr const& edge_label) override;
    virtual bool disconnectFrom(Connectable *target) override;
    virtual bool canConnectTo(const Connectable* target) const override;
};

// Predeclare a class that can be specified on the function signature.
// This will be specialised for each length of arg list, to derive from
// the corresponding PeerConnectorN class.
template <typename Signature>
class PeerConnector;

} // namespace sg


#include "sg/ConcreteConnectorBase.h"

namespace sg {

//N=0
template< class RetType=void >
class PeerConnector0 : public ConcreteConnectorBase0< RetType >, public PeerConnectorBase
{
    typedef PeerConnector0 < RetType> ConnectorType;

public:

    ~PeerConnector0() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector0();                               }

    RetType adaptor(CallID call_id_)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_);
    }
};

// Allow the use of pretty signatures
template<typename R>
class PeerConnector< R (  ) >
    : public PeerConnector0 < R >
{
 public:
    typedef PeerConnector0 < R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=1
template< class Param1, class RetType=void >
class PeerConnector1 : public ConcreteConnectorBase1< Param1, RetType >, public PeerConnectorBase
{
    typedef PeerConnector1 < Param1, RetType> ConnectorType;

public:

    ~PeerConnector1() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector1();                               }

    RetType adaptor(CallID call_id_, Param1 p1)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1>
class PeerConnector< R ( Param1 ) >
    : public PeerConnector1 < Param1, R >
{
 public:
    typedef PeerConnector1 < Param1, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=2
template< class Param1, class Param2, class RetType=void >
class PeerConnector2 : public ConcreteConnectorBase2< Param1, Param2, RetType >, public PeerConnectorBase
{
    typedef PeerConnector2 < Param1, Param2, RetType> ConnectorType;

public:

    ~PeerConnector2() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector2();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2>
class PeerConnector< R ( Param1, Param2 ) >
    : public PeerConnector2 < Param1, Param2, R >
{
 public:
    typedef PeerConnector2 < Param1, Param2, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=3
template< class Param1, class Param2, class Param3, class RetType=void >
class PeerConnector3 : public ConcreteConnectorBase3< Param1, Param2, Param3, RetType >, public PeerConnectorBase
{
    typedef PeerConnector3 < Param1, Param2, Param3, RetType> ConnectorType;

public:

    ~PeerConnector3() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector3();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3>
class PeerConnector< R ( Param1, Param2, Param3 ) >
    : public PeerConnector3 < Param1, Param2, Param3, R >
{
 public:
    typedef PeerConnector3 < Param1, Param2, Param3, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=4
template< class Param1, class Param2, class Param3, class Param4, class RetType=void >
class PeerConnector4 : public ConcreteConnectorBase4< Param1, Param2, Param3, Param4, RetType >, public PeerConnectorBase
{
    typedef PeerConnector4 < Param1, Param2, Param3, Param4, RetType> ConnectorType;

public:

    ~PeerConnector4() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector4();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4>
class PeerConnector< R ( Param1, Param2, Param3, Param4 ) >
    : public PeerConnector4 < Param1, Param2, Param3, Param4, R >
{
 public:
    typedef PeerConnector4 < Param1, Param2, Param3, Param4, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=5
template< class Param1, class Param2, class Param3, class Param4, class Param5, class RetType=void >
class PeerConnector5 : public ConcreteConnectorBase5< Param1, Param2, Param3, Param4, Param5, RetType >, public PeerConnectorBase
{
    typedef PeerConnector5 < Param1, Param2, Param3, Param4, Param5, RetType> ConnectorType;

public:

    ~PeerConnector5() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector5();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5 ) >
    : public PeerConnector5 < Param1, Param2, Param3, Param4, Param5, R >
{
 public:
    typedef PeerConnector5 < Param1, Param2, Param3, Param4, Param5, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=6
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class RetType=void >
class PeerConnector6 : public ConcreteConnectorBase6< Param1, Param2, Param3, Param4, Param5, Param6, RetType >, public PeerConnectorBase
{
    typedef PeerConnector6 < Param1, Param2, Param3, Param4, Param5, Param6, RetType> ConnectorType;

public:

    ~PeerConnector6() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector6();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6 ) >
    : public PeerConnector6 < Param1, Param2, Param3, Param4, Param5, Param6, R >
{
 public:
    typedef PeerConnector6 < Param1, Param2, Param3, Param4, Param5, Param6, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=7
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class RetType=void >
class PeerConnector7 : public ConcreteConnectorBase7< Param1, Param2, Param3, Param4, Param5, Param6, Param7, RetType >, public PeerConnectorBase
{
    typedef PeerConnector7 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, RetType> ConnectorType;

public:

    ~PeerConnector7() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector7();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7 ) >
    : public PeerConnector7 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, R >
{
 public:
    typedef PeerConnector7 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=8
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class RetType=void >
class PeerConnector8 : public ConcreteConnectorBase8< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, RetType >, public PeerConnectorBase
{
    typedef PeerConnector8 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, RetType> ConnectorType;

public:

    ~PeerConnector8() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector8();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8 ) >
    : public PeerConnector8 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, R >
{
 public:
    typedef PeerConnector8 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=9
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class RetType=void >
class PeerConnector9 : public ConcreteConnectorBase9< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, RetType >, public PeerConnectorBase
{
    typedef PeerConnector9 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, RetType> ConnectorType;

public:

    ~PeerConnector9() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector9();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9 ) >
    : public PeerConnector9 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, R >
{
 public:
    typedef PeerConnector9 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=10
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class RetType=void >
class PeerConnector10 : public ConcreteConnectorBase10< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, RetType >, public PeerConnectorBase
{
    typedef PeerConnector10 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, RetType> ConnectorType;

public:

    ~PeerConnector10() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector10();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10 ) >
    : public PeerConnector10 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, R >
{
 public:
    typedef PeerConnector10 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=11
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class RetType=void >
class PeerConnector11 : public ConcreteConnectorBase11< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, RetType >, public PeerConnectorBase
{
    typedef PeerConnector11 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, RetType> ConnectorType;

public:

    ~PeerConnector11() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector11();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10, Param11 p11)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11 ) >
    : public PeerConnector11 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, R >
{
 public:
    typedef PeerConnector11 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=12
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class RetType=void >
class PeerConnector12 : public ConcreteConnectorBase12< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, RetType >, public PeerConnectorBase
{
    typedef PeerConnector12 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, RetType> ConnectorType;

public:

    ~PeerConnector12() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector12();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10, Param11 p11, Param12 p12)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12 ) >
    : public PeerConnector12 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, R >
{
 public:
    typedef PeerConnector12 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=13
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13, class RetType=void >
class PeerConnector13 : public ConcreteConnectorBase13< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, RetType >, public PeerConnectorBase
{
    typedef PeerConnector13 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, RetType> ConnectorType;

public:

    ~PeerConnector13() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector13();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10, Param11 p11, Param12 p12, Param13 p13)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13 ) >
    : public PeerConnector13 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, R >
{
 public:
    typedef PeerConnector13 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=14
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13, class Param14, class RetType=void >
class PeerConnector14 : public ConcreteConnectorBase14< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, RetType >, public PeerConnectorBase
{
    typedef PeerConnector14 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, RetType> ConnectorType;

public:

    ~PeerConnector14() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector14();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10, Param11 p11, Param12 p12, Param13 p13, Param14 p14)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13, class Param14>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14 ) >
    : public PeerConnector14 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, R >
{
 public:
    typedef PeerConnector14 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};

//N=15
template< class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13, class Param14, class Param15, class RetType=void >
class PeerConnector15 : public ConcreteConnectorBase15< Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, Param15, RetType >, public PeerConnectorBase
{
    typedef PeerConnector15 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, Param15, RetType> ConnectorType;

public:

    ~PeerConnector15() { disconnect(); }

private:
    virtual void bindAdaptor() override         { this->bind_call_id_behaviour(this, &ConnectorType::adaptor);   }
    virtual void unbindAdaptor() override       { this->unbind_call_id_behaviour(this, &ConnectorType::adaptor); }
    virtual Connectable* makeSameKind() override { return new PeerConnector15();                               }

    RetType adaptor(CallID call_id_, Param1 p1, Param2 p2, Param3 p3, Param4 p4, Param5 p5, Param6 p6, Param7 p7, Param8 p8, Param9 p9, Param10 p10, Param11 p11, Param12 p12, Param13 p13, Param14 p14, Param15 p15)
    {
        ARM_ADAPTOR_PEER_CALL(RetType, call_id_, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15);
    }
};

// Allow the use of pretty signatures
template<typename R, class Param1, class Param2, class Param3, class Param4, class Param5, class Param6, class Param7, class Param8, class Param9, class Param10, class Param11, class Param12, class Param13, class Param14, class Param15>
class PeerConnector< R ( Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, Param15 ) >
    : public PeerConnector15 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, Param15, R >
{
 public:
    typedef PeerConnector15 < Param1, Param2, Param3, Param4, Param5, Param6, Param7, Param8, Param9, Param10, Param11, Param12, Param13, Param14, Param15, R > BaseType;
    
    virtual Connectable* makeSameKind() override { return new PeerConnector(); }
};


} // namespace sg

#endif
