/**
 *  \file   PluginRegisterPublishingInterface.h
 *  \brief  Interface to allow plugins to publish registers to a core
 *  \date   Copyright ARM Limited 2012 All Rights Reserved.
 */
#ifndef FRAMEWORK_INCLUDE_SG_PLUGINREGISTERPUBLISHINGINTERFACE_H_E06AEF67
#define FRAMEWORK_INCLUDE_SG_PLUGINREGISTERPUBLISHINGINTERFACE_H_E06AEF67 1

#include <eslapi/CAInterface.h>
#include <stdlib.h>

namespace sg
{
    /**
     * Call back interface for access checks on a register.
     */
    class PluginRegisterAccessControl
    {
    public:
        /**
         * Result of an access check.  Not all results may be valid for every type of register
         * interface.
         */
        enum CheckResult
        {
            CHECK_OK,                   ///< Access will not fault
            CHECK_NOP,                  ///< Access should behave as a NOP.
            CHECK_UNDEF,                ///< Access is UNDEFINED.
            CHECK_TRAP_TO_EL1,          ///< Access should be trapped to EL1
            CHECK_TRAP_TO_EL2,          ///< Access should be trapped to EL2
            CHECK_TRAP_TO_EL3,          ///< Access should be trapped to EL3
            CHECK_TRAP_TO_DEBUG,        ///< Access should be trapped to debug state
            CHECK_RESULT_MAX
        };

        /**
         * Security state of the core.
         */
        enum SecurityState
        {
            SS_SECURE = 0,
            SS_NON_SECURE = 1
        };

        /**
         * Protection level of the core.
         */
        enum ProtectionLevel
        {
            PL_EL0 = 0,                 ///< EL0 -- user mode
            PL_EL1 = 1,                 ///< EL1 -- kernel mode
            PL_EL2 = 2,                 ///< EL0 -- hypervisor mode
            PL_EL3 = 3                  ///< EL0 -- secure monitor mode
        };

        /**
         * Handle a access check for a read of a register.
         *
         * This function may use any mechanism it wishes to determine a if access should proceed,
         * or how it will fail.  The core's current state may be inspected using SCADI if
         * necessary.
         */
        virtual CheckResult registerCheckRead(SecurityState security_state, ProtectionLevel protection_level, void *user_data) = 0;

        /**
         * Handle a access check for a write to a register.
         *
         * This function may use any mechanism it wishes to determine a if access should proceed,
         * or how it will fail.  The core's current state may be inspected using SCADI if
         * necessary.
         */
        virtual CheckResult registerCheckWrite(SecurityState security_state, ProtectionLevel protection_level, void *user_data) = 0;

        virtual ~PluginRegisterAccessControl() { }
    }; // class PluginRegisterAccessControl

    /**
     * Call back interface for handling accesses to a register.
     *
     * By the time the read or write function on this interface is called all permission and trap
     * checking should have been taken care of by the access checks implemented by
     * PluginRegisterAccessControl.  The only reason an access may fail at this stage is for a
     * reason associated with the value itself.
     */
    class PluginRegisterHandler
    {
    public:
        enum AccessResult
        {
            ACCESS_OK,
            ACCESS_ABORT
        };

        /**
         * Handle a read of a register.
         *
         * The read must normally succeed, but may abort for data dependent reasons.
         *
         * \param[in]   data        Pointer to a location to store the value read.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual AccessResult registerRead(void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a write to a register.
         *
         * The write must normally succeed, but may abort for data dependent reasons.
         *
         * \param[in]   data        Pointer to the value accompanying the write.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual AccessResult registerWrite(const void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a debug read of a register.
         *
         * A debug read of a register is made by an external entity, such as a CADI debugger.
         * Debug reads may wish to avoid some of the side effects that a normal read would have.
         *
         * \param[in]   data        Pointer to a location to store the value read.  The location
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual AccessResult registerReadDebug(void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a debug write to a register.
         *
         * A debug write to a register is made by an external entity, such as a CADI debugger.
         * Debug reads may wish to avoid some of the side effects that a normal write would have.
         *
         * \param[in]   data        Pointer to the value accompanying the write.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual AccessResult registerWriteDebug(const void *data, unsigned width_bits, void *user_data) = 0;

        virtual ~PluginRegisterHandler() { }
    }; // class PluginRegisterHandler

    /**
     * Call back interface for handling interception of a register.
     *
     * For read callbacks, the read data will have been initialized with the value that would have
     * been returned if the register was not being intercepted.  The callback may inspect and
     * modify this value before it is returned as the actual result of the read.
     *
     * For write callbacks, the write data is the value initially supplied with the access.  The
     * plugin may inspect and modify this value before the real register access uses it.
     */
    class PluginRegisterInterceptHandler
    {
    public:
        /**
         * Handle a read of a register.
         *
         * The read must normally succeed, but may abort for data dependent reasons.
         *
         * \param[in]   data        Pointer to a location to store the value read.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual void registerRead(void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a write to a register.
         *
         * The write must normally succeed, but may abort for data dependent reasons.
         *
         * \param[in]   data        Pointer to the value accompanying the write.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual void registerWrite(void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a debug read of a register.
         *
         * A debug read of a register is made by an external entity, such as a CADI debugger.
         * Debug reads may wish to avoid some of the side effects that a normal read would have.
         *
         * \param[in]   data        Pointer to a location to store the value read.  The location
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual void registerReadDebug(void *data, unsigned width_bits, void *user_data) = 0;

        /**
         * Handle a debug write to a register.
         *
         * A debug write to a register is made by an external entity, such as a CADI debugger.
         * Debug reads may wish to avoid some of the side effects that a normal write would have.
         *
         * \param[in]   data        Pointer to the value accompanying the write.  The value will
         *                          be a uint32_t, uint64_t etc depending on \p width_bits.
         * \param[in]   width_bits  The width of this access.  Will usually be the same as the \p
         *                          width_bits specified when the handler was added.
         * \param[in]   user_data   The \p user_data provided when the register was added.
         */
        virtual void registerWriteDebug(void *data, unsigned width_bits, void *user_data) = 0;

        virtual ~PluginRegisterInterceptHandler() { }
    }; // class PluginRegisterInterceptHandler

    /**
     * Interface for an object that allows a plugin to add or modify registers.
     */
    class PluginRegisterPublishingInterface
        : public eslapi::CAInterface
    {
    public:
        typedef uint32_t RegisterNumber;

        /**
         * Add a single register with the given attributes.
         *
         * \param[in]   register_number  Encoded register number.
         * \param[in]   width_bits       Width of the register in bits.
         * \param[in]   name             Name of the register.  The string will be copied.
         * \param[in]   description      User visible textual description of the register.
         * \param[in]   access_controls  An array of access controls for the register
         * \param[in]   num_access_controls  The number of controls in \p access_controls
         * \param[in]   register_handler Handler for the register itself.
         * \param[in]   user_data        Context for the access controls and handler.
         *
         * The access controls and register handler remain owned by the plugin but must remain live
         * until the plugin is unloaded, when they may be cleaned up.  The list of access controls
         * will be copied and may be disposed of after the call.
         */
        virtual bool addRegister(
                RegisterNumber      register_number,
                uint32_t            width_bits,
                const char *        name,
                const char *        description,
                PluginRegisterAccessControl *access_controls[],
                size_t              num_access_controls,
                PluginRegisterHandler *register_handler,
                void *              user_data
        ) = 0;

        /**
         * Intercept data accesses to an existing register.  The specified register must already
         * exist and the plugin can only change the data values read or written from/to the
         * register.  It is not possible to change the access controls on the register.
         *
         * \param[in]   register_number  Encoded register number.
         * \param[in]   width_bits       Width of the register in bits.
         * \param[in]   register_handler Handler for the register itself.
         * \param[in]   user_data        Context for the handler.
         *
         * The handler remain owned by the plugin but must remain live until the plugin is
         * unloaded, when they may be cleaned up.
         */
        virtual bool interceptRegister(
                RegisterNumber      register_number,
                uint32_t            width_bits,
                PluginRegisterInterceptHandler *register_handler,
                void *              user_data
        ) = 0;

        static eslapi::if_name_t IFNAME()       { return "sg.PluginRegisterPublishingInterface"; }
        static eslapi::if_rev_t  IFREVISION()   { return 1; }
    }; // PluginRegisterPublishingInterface

} // namespace sg

#endif // FRAMEWORK_INCLUDE_SG_PLUGINREGISTERPUBLISHINGINTERFACE_H_E06AEF67
