/*!
 * \file    Range.h
 * \brief   A range type.
 * \date    Copyright 2010-2012 ARM Limited. All rights reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 */

#ifndef SG_RANGE_H
#define SG_RANGE_H

#include <iostream>
#include <algorithm>
#include <cassert>
#include <limits>

namespace sg
{
    template< typename T > struct range_t;
    template< typename T > inline range_t<T> range_begin_end_incl( T begin_, T end_incl_ );
    template< typename T > inline range_t<T> range_empty();

    ///
    /// Subtracting two T may give a non-T, this allows a user to implement this
    /// functionality.
    ///
    template< typename T > struct range_t_traits
    {
        typedef T size_type;

        ///
        /// \name Members for use in creating an empty and full ranges.
        ///
        /// If asked to create an empty range then we need to create a T for
        /// such purposes, these functions construct such a T.
        ///
        //@{
        static T begin_for_empty()      { return 1; }
        static T end_incl_for_empty()   { return 0; }
        static T begin_for_full_range() { return (std::numeric_limits<T>::min)(); }
        static T end_incl_for_full_range() { return (std::numeric_limits<T>::max)(); }
        //@}

        // This is called within the size() method of the range (u_).  If
        // the range represents the full range then we cannot represent the
        // full size in the return result.  Prefer to assert fail!
        template<typename U>
        static void assert_size_not_full(U const& u_)
        {
            (void)u_;
            assert( ! u_.full() );
        }
    };

    ///
    /// Partial specialisation for pointers, the subtraction of two pointers yields a ptrdiff_t
    ///
    template< typename T > struct range_t_traits<T*>
    {
        typedef std::ptrdiff_t size_type;
        static T* begin_for_empty()    { return (T*)1; }
        static T* end_incl_for_empty() { return (T*)0; }

//        static T begin_for_full_range();        // NO! See comment for assert_size_not_full
//        static T end_incl_for_full_range();     // NO! See comment for assert_size_not_full

        // This is called within the size() method of the range (u_).  If
        // the range represents the full range then we cannot represent the
        // full size in the return result.
        //
        // However, we don't expect people to construct full ranges of host
        // memory pointers.  The use for such a thing is limited so we have
        // prevented the ability to detect such a thing.  Generally one would
        // expect a finite range or the empty range.
        template<typename U>
        static void assert_size_not_full(U const& u_)
        {
            (void)u_; // do nothing
        }
    };

    ///
    /// Represent a range, can be constructed with either a semi-open or closed
    /// range using range_begin_end/range_begin_end_incl.
    ///
    template< typename T >
    struct range_t
    {
        typedef T                                      value_type;
        typedef range_t_traits<T>                      traits;
        typedef typename traits::size_type             size_type;

        T  begin;
        T  end_incl;

        // NOTE that this is _not_ a copy constructor -- C++'s rules say that a
        // copy constructor cannot be a template.
        template< typename U >
        range_t( range_t<U> const& o_ )
            : begin(o_.begin), end_incl(o_.end_incl)
        {
            // Assert that o_ is representable in this type.
            assert( begin == o_.begin );
            assert( end_incl == o_.end_incl );
        }

        // NOTE no constructor with two arguments -- so we are forced at call
        // site to name what we mean.

        range_t() : begin(traits::begin_for_empty()),
                    end_incl(traits::end_incl_for_empty())
        {

        }

        // Compiler-generated copy-constructor and assignment operator are OK

        /// Align the range start and end with a specified power-of-two granule.
        ///
        /// A post-condition is that an empty range stays empty, a range can
        /// only get bigger and that the begin and end points are
        /// granule-aligned.  In addition, the expansion of the range is minimal
        /// -- that is the difference between the new begin and the old is <
        /// size_ and similarly the new end and the old is < size_.
        ///
        void expand_to_cover_power_of_two_granule( size_type size_ )
        {
            // is size_ a power of two?
            assert( size_ != 0 && (size_ & (T(0) - size_)) == size_ );
            // TODO make this work if T is a pointer!  We need some additional
            // traits to allow us to do it!
            if (empty())
                return;
            begin    &= T(~T(size_ - 1));
            end_incl |= T(size_ - 1);
        }

        // Align start and end such that start and end are aligned to the granule
        // and start and end both always lie inside the original range. This can make
        // the range empty.
        void reduce_to_contained_power_of_two_granule( size_type size_ )
        {
            // is size_ a power of two?
            assert( size_ != 0 && (size_ & (T(0) - size_)) == size_ );
            // empty range stays empty
            if (empty())
                return;
            // handle corner cases at the end and the beginning of the space
            if ((begin > (T(0) - size_)) || (end_incl < (size_ - 1)))
            {
                *this = make_empty();
                return;
            }
            begin = (begin + (size_ - 1)) & T(~T(size_ - 1));
            end_incl = (end_incl - (size_ - 1)) | T(size_ - 1);
            if (end_incl < begin)
                *this = make_empty();
        }

        /// Is the range empty or aligned to a power of two size.
        bool is_empty_or_aligned_to_power_of_two_size( size_type size_ ) const
        {
            // is size_ a power of two?
            assert( size_ != 0 && (size_ & (size_type(0) - size_)) == size_ );

            size_type const mask = size_ - 1;
            
            return empty()
                || ( (begin & mask) == 0
                     && (end_incl & mask) == mask );
        }

        /// Is the range empty or the range size is a power of two and it is
        /// aligned to that size.
        bool is_empty_or_size_a_power_of_two_and_aligned_to_size() const
        {
            if (empty())
                return true;

            // We can't call size() if we are full(), but the following code
            // should work independent of if we are full or not.

            size_type const mask = end_incl - begin;
            size_type const s = mask + 1;  // size
            bool const size_is_power_of_two
                =   s == 0 // overflowed representation and so assume is a power of two
                || (s & (size_type(0) - s)) == s
                ;
            
            return size_is_power_of_two
                && (begin & mask) == 0
                && (end_incl & mask) == mask;
        }

        /// Does this range intersect the argument, are there a non-zero number
        /// of elements in the intersection?
        ///
        /// This answers the question does (*this `intersection` o_) != null_set.
        ///
        /// As such, if either of the ranges are the null_set then the result is
        /// false.
        bool intersects( range_t const& o_ ) const
        {
            bool const ret = end_incl >= o_.begin && begin <= o_.end_incl
                && ! empty() && ! o_.empty()
                ;
            return ret;
        }

        /// Compute the intersection of two ranges.
        range_t intersection_with( range_t const& o_ ) const
        {
            if (empty() || o_.empty())
                return range_empty<T>();
            range_t<T> r;
            r.begin    = (std::max)( begin,    o_.begin    );
            r.end_incl = (std::min)( end_incl, o_.end_incl );

            if (r.begin > r.end_incl) {
                r = range_empty<T>();
            }

            assert( intersects( o_ ) == ! r.empty() );

            return r;
        }

        /// Compute the part of the range that is below the intersection of
        /// *this and o_.
        ///
        /// If this or the argument is empty then result is UNDEFINED.
        ///
        /// If this and the argument do not intersect then the result is UNDEFINED.
        ///
        /// Returns the empty range if there is no range below.  The second of
        /// the pair returns is UNDEFINED.
        ///
        /// Returns a non-empty range and true if *this is 'below' o_.
        ///
        /// Returns a non-empty range and false if *this is 'above' o_.
        ///
        /// In diagrams (where we don't indicate the upper bound as it is irrelevant).
        ///
        ///      |<---....    this
        ///      |<----...    o_
        ///                   result empty
        ///
        ///    |<---.....     this
        ///         |<---.    o_
        ///    |<->|          result, true
        ///
        ///         |<---.    this
        ///    |<---.....     o_
        ///    |<->|          result, false
        ///
        std::pair<range_t,bool> range_below_intersection( range_t const& o_ ) const
        {
            assert( intersects(o_) );

            if (begin == o_.begin)
                return std::make_pair(make_empty(), /*UNDEFINED*/true);
            
            return
                begin < o_.begin
                ? std::make_pair( from_begin_end(begin,    o_.begin), true  )
                : std::make_pair( from_begin_end(o_.begin, begin   ), false );
        }
        
        /// Compute the part of the range that is above the intersection of
        /// *this and o_.
        ///
        /// If this or the argument is empty then result is UNDEFINED.
        ///
        /// If this and the argument do not intersect then the result is UNDEFINED.
        ///
        /// Returns the empty range if there is no range above.  The second of
        /// the pair returns is UNDEFINED.
        ///
        /// Returns a non-empty range and true if *this is 'above' o_.
        ///
        /// Returns a non-empty range and false if *this is 'below' o_.
        ///
        /// In diagrams (where we don't indicate the lower bound as it is irrelevant).
        ///
        ///      ...--->|     this
        ///      ...--->|     o_
        ///                   result empty
        ///
        ///         ...--->|  this
        ///    ...--->|       o_
        ///            |<->|  result, true
        ///
        ///    ...--->|       this
        ///         ...--->|  o_
        ///            |<->|  result, false
        ///
        std::pair<range_t,bool> range_above_intersection( range_t const& o_ ) const
        {
            assert( intersects(o_) );
            
            if (end_incl == o_.end_incl)
                return std::make_pair(make_empty(), /*UNDEFINED*/true);
            
            return
                end_incl < o_.end_incl
                ? std::make_pair( from_begin_end_incl(end_incl + 1,    o_.end_incl), false )
                : std::make_pair( from_begin_end_incl(o_.end_incl + 1, end_incl   ), true  );
        }


        /// True if the range is empty.
        bool empty() const
        {
            // Although there are many ways of representing an empty range,
            // it is an error to construct anything other than the canonical representation.
            // This simplifies the comparison operator, and reduces the number of corner-cases
            // to write tests for.
            assert(!(end_incl < begin) || (begin == traits::begin_for_empty() && end_incl == traits::end_incl_for_empty()));
            return end_incl < begin;
        }

        /// True if the range is the full range
        bool full() const
        {
            // NOTE that for types it might not make sense for a full range
            // and the traits will not have the begin_for_full_range(), etc.
            // in and this will fail to compile.
            return begin == traits::begin_for_full_range()
                && end_incl == traits::end_incl_for_full_range();
        }

        /// How many elements are in the range?  NOTE that if this covers the
        /// whole range then this may return a bogus value as size_type may not
        /// be able to express the whole range.  You can use full() to determine
        /// if it is the full range.
        size_type size() const
        {
            traits::assert_size_not_full(*this);
            return empty() ? 0 :  end_incl - begin + 1;
        }

        /// Does the range contain the point specified in the argument.
        bool contains( T t_ ) const
        {
            bool const ret = begin <= t_ && t_ <= end_incl;
            assert( intersects(range_begin_end_incl(t_,t_)) == ret );
            return ret;
        }

        /// Get offset of a value from start of range.
        /// Precondition: value is not less than start of range.
        size_type offset_of( T t_ ) const
        {
            assert(t_ >= begin);
            return t_ - begin;
        }

        /// Does the range contain (or is the same as) the argument.
        ///
        /// NOTE that the empty set is a subset of every set (including itself).
        /// Therefore every set is_a_superset_or_same_as() of the empty set.
        ///
        /// NOTE that strictly 'superset' means that 'A' is a superset of 'A'.
        /// However, some people disagree, so we disambiguate by using
        /// 'or_same_as'.
        ///
        bool is_a_superset_or_same_as( range_t t_ ) const
        {
            bool const ret
                = (begin <= t_.begin && t_.end_incl <= end_incl)
                || t_.empty()                // any set is a superset of empty()
                || (empty() && t_.empty())   // the empty() set is a superset of itself
                ;
            return ret;
        }

        /// This range is below the specified point.  If the range is empty then the result is undefined.
        bool range_is_below_point( T t_ ) const
        {
            assert( ! empty() );
            return end_incl < t_;
        }

        /// This range is above the specified point.  If the range is empty then the result is undefined.
        /// If this returns true then it means that t_ is not a member of the range.
        bool range_is_above_point( T t_ ) const
        {
            assert( ! empty() );
            bool const ret = t_ < begin;
            assert( (  ret && ! contains(t_)) ||
                    (! ret && (contains(t_) || range_is_below_point(t_)) ) );
            return ret;
        }

        bool operator == ( range_t const& o_ ) const
        {
            return begin == o_.begin && end_incl == o_.end_incl && empty() == o_.empty();
        }

        bool operator != ( range_t const& o_ ) const
        {
            return !(*this == o_);
        }

        /// Defining an ordering over the range type, so that values can be
        /// stored in an ordered container.
        /// The empty range compares less than everything apart from itself.
        /// Otherwise, ranges are ordered by start address, then by end address.
        bool operator < ( range_t const & r) const
        {
            if (empty()) return !r.empty();
            return begin < r.begin
               || (begin == r.begin && end_incl < r.end_incl);
        }


        /// \name Models sequence concept
        //@{
        T    front()     const { assert( ! empty() ); return begin; }
        void pop_front()       {
            assert( ! empty() );
            if (begin != end_incl) {
                ++begin;
                assert( !empty() );
            } else {
                *this = range_empty<T>();
                assert( empty() );
            }
        }
        // empty() is defined above
        //@}

        /// \name Named constructors
        //@{
        static range_t from_begin_end(T begin_, T end_) {

            // NOTE that case where end = MIN_T, therefore end - 1 is MAX_T
            // So [begin, MIN_T) (which should be the empty set) becomes [begin, MAX_T)
            // which is a rather large set!  So we have to filter that out first.

            if (end_ <= begin_)
                return make_empty();

            range_t<T> r;
            r.begin = begin_;
            r.end_incl = T(end_ - 1);
            return r;
        }

        static range_t from_begin_end_incl(T begin_, T end_incl_) {

            if (end_incl_ < begin_)
                return make_empty();

            range_t<T> r;
            r.begin = begin_;
            r.end_incl = end_incl_;
            return r;
        }

        static range_t from_begin_size(T begin_, typename traits::size_type size_ )
        {
            if (size_ == 0)
                return range_empty<T>();

            range_t<T> r = range_begin_end_incl( begin_, begin_ + (size_ - 1) );
            assert( r.size() == size_ );
            return r;
        }

        static range_t make_empty() {
            range_t<T> r;
            r.begin = traits::begin_for_empty();
            r.end_incl = traits::end_incl_for_empty();
            assert(r.empty());
            return r;
        }

        static range_t make_full_range() {
            range_t<T> r;
            r.begin = traits::begin_for_full_range();
            r.end_incl = traits::end_incl_for_full_range();
            return r;
        }
        //@}

    };

    ///
    /// Construct a range [begin, end) -- begin is part of the range, but
    /// end is not.
    ///
    template< typename T >
    inline range_t<T> range_begin_end( T begin_, T end_ )
    {
        return range_t<T>::from_begin_end(begin_, end_);
    }

    ///
    /// Construct a range [begin, begin + size) -- begin is part of the range, but
    /// begin + size is not.
    ///
    template< typename T >
    inline range_t<T> range_begin_size( T begin_, typename range_t_traits<T>::size_type size_ )
    {
        return range_t<T>::from_begin_size(begin_, size_);
    }

    ///
    /// Construct a range [begin, end] -- both begin and end are part of the
    /// range.
    ///
    template< typename T>
    inline range_t<T> range_begin_end_incl( T begin_, T end_incl_ )
    {
        return range_t<T>::from_begin_end_incl(begin_, end_incl_);
    }

    ///
    /// Construct an empty range.  This is the unique representation of the empy set.
    ///
    template< typename T >
    inline range_t<T> range_empty()
    {
        return range_t<T>::make_empty();
    }

    ///
    /// Construct a full range.
    ///
    template< typename T >
    inline range_t<T> range_full_range()
    {
        return range_t<T>::make_full_range();
    }

    ///
    /// Construct an sg::range_t from another equivalent representation.
    ///
    template< typename T >
    inline range_t<typename T::value_type> range_from_range( T t_ )
    {
        range_t<typename T::value_type> r; r.begin = t_.begin; r.end_incl = t_.end_incl;
        return r;
    }

    ///
    /// Construct the minimal superset that covers both ranges.
    ///
    /// This constructs a range that contains all members of the two ranges and
    /// also covers the range between the two ranges.
    ///
    /// i.e. the new range has the lower bound of the minimum of the begin of
    /// the two ranges, and an upper bound of the maximum of the end_incl of the
    /// two ranges.
    ///
    template<typename T>
    inline range_t<T> superset_of(range_t<T> const& a_, range_t<T> const& b_)
    {
        if (a_.empty())
            return b_;

        if (b_.empty())
            return a_;

        return range_begin_end_incl<T>(
            std::min( a_.begin,    b_.begin    ),
            std::max( a_.end_incl, b_.end_incl )
            );
    }

    template<typename T>
    inline std::ostream& operator <<( std::ostream& os_, range_t<T> t_ )
    {
        return os_ << "[" << t_.begin << ", " << t_.end_incl << "]";
    }


    ///
    /// Given that a_ and b_ represent objects with non-empty ranges then
    /// break into three regions represented by calling:
    ///
    ///        collection_.below_region( T a_or_b_, below_range, is_a_ )             // called only if non-empty
    ///        collection_.intersection_region( T a_, T b_, intersection_range )     // called only if non-empty
    ///        collection_.above_region( T a_or_b_, above_range, is_a_ )             // called only if non-empty
    ///
    /// If the ranges do not intersect then the regions are classified as
    /// above and below.
    ///
    /// NOTE that in general the *_region functions should not alter the T
    /// arguments because they would be subsequently passed to the other
    /// *_region functions.
    ///
    /// The calls are called in the specified order.
    ///
    /// collection also has a method range(T const&) that returns a range_t<>
    /// for the specified object.
    ///
    template<typename T0, typename T1, typename Collection>
    void break_into_ranges(
        T0&            a_,
        T1&            b_,
        Collection&    collection_
        )
    {
        auto const a_range = collection_.range(a_);
        auto const b_range = collection_.range(b_);

        assert( ! a_range.empty() );
        assert( ! b_range.empty() );
            
        auto const intersection_range = a_range.intersection_with(b_range);

        if (intersection_range.empty())
        {
            // Special case.
            //
            // NOTE that the extra brackets are for a parse-bug in g++ 4.8.3
            // that interprets the < as a template bracket, despite there
            // being no context at all to suggest that it should be a template.
            bool const a_is_below_b = (a_range.begin) < (b_range.begin);
            if (a_is_below_b)
            {
                collection_.below_region(a_, a_range, /*is_a_*/true);
                collection_.above_region(b_, b_range, /*is_a_*/false);
            }
            else
            {
                collection_.below_region(b_, b_range, /*is_a_*/false);
                collection_.above_region(a_, a_range, /*is_a_*/true);
            }
            return;
        }

        auto const p_below     = a_range.range_below_intersection(b_range);
        auto const below_range = p_below.first;
        auto const below_is_a  = p_below.second;
            
        auto const p_above     = a_range.range_above_intersection(b_range);
        auto const above_range = p_above.first;
        auto const above_is_a  = p_above.second;

        if ( ! below_range.empty() )
        {
            if (below_is_a)
                collection_.below_region(a_, below_range, /*is_a_*/true);
            else
                collection_.below_region(b_, below_range, /*is_a_*/false);
        }

        collection_.intersection_region(a_, b_, intersection_range);

        if ( ! above_range.empty() )
        {
            if (above_is_a)
                collection_.above_region(a_, above_range, /*is_a_*/true);
            else
                collection_.above_region(b_, above_range, /*is_a_*/false);
        }
    }

    /// Break a simple range_t<T> into a below, intersection and above.
    template<typename T>
    void break_into_simple_ranges(
        range_t<T> const&    a_,
        range_t<T> const&    b_,

        range_t<T>&          out_below_,
        range_t<T>&          out_intersection_,
        range_t<T>&          out_above_
        )
    {
        typedef range_t<T> R;
        
        struct collection {
            R&          out_below;
            R&          out_intersection;
            R&          out_above;

            collection(
                R&          out_below2_,
                R&          out_intersection2_,
                R&          out_above2_
                ) : out_below        ( out_below2_        ),
                    out_intersection ( out_intersection2_ ),
                    out_above        ( out_above2_        ) {}

            R range(R const& t_) const { return t_; }
            
            void below_region(R const&, R const& r_, bool)            { out_below = r_; }
            void intersection_region(R const&, R const&, R const& r_) { out_intersection = r_; }
            void above_region(R const&, R const& r_, bool)            { out_above = r_; }
        } c( out_below_, out_intersection_, out_above_ );

        out_below_        = R();
        out_intersection_ = R();
        out_above_        = R();
        
        break_into_ranges( a_, b_, c );
    }
    
}

#endif
