 /*!
 * \file    SGComponent.h
 * \brief   Component/Port infrastructure using Connectors
 * \date    Copyright 2005 ARM Limited. All rights reserved.
 * \author  Andrew Bolt and Graeme Barnes.
 *
 */

#ifndef ARM_SG_COMPONENT_H
#define ARM_SG_COMPONENT_H

#include <map>

#include "sg/Attributes.h"
#include "sg/SGPort.h"
#include "sg/SGComponentBase.h"
#include "sg/SGParams.h"
#include "sg/SGControlProtocol.h"
#include "sg/SGSimulationContext.h"


namespace sg {

// Components are collections of named ports, and of other components
class Component : public ComponentBase
{
    class stream_t;
    friend class stream_t;

    // Must be before msg_fatal_error, etc.
    stream_t* s_fatal_error;
    stream_t* s_error;
    stream_t* s_warning;
    stream_t* s_info;
    stream_t* s_debug;
    
    std::string name;

    std::string hierarchicalName;

    Node *parent;

    // at some point this might change to point to a sub-section
    // of a global parameter set
    Params params; // init/runtime parameters

protected:
    /*! Global simulation context */
    SimulationContext* simulationContext;

    typedef std::list<ComponentBase*> ComponentList;

    // nested components - should these be named?
    ComponentList subcomponents;

public:
    /*! Provide a port that can be used to control this component's life-span.
     *  This port will be published as '$control'.
     */
    SGControlProtocol control_port;

    /*! Provide a port that can be used to control all child components.
     *  This port will be connected to the $control port on all children.
     */
    SGControlProtocol child_control_port;

    /*! Provide message functions for all components.
     */
    void message(const std::string & msg, sg::message::MessageType msgType);
    void message(sg::message::MessageType msgType, const char *fmt, ...) SG_FORMATSTR(printf, 3, 4);

    /*! The MSG_FATAL_ERROR stream.
     *  You must use std::endl or a flush() to push the data to the actual
     *  message handler.
     *
     *  Streaming a message to this will cause the simulation to stop and exit
     *  at the next available opportunity.
     */
    std::ostream&     msg_fatal_error;

    /*! The MSG_ERROR stream.
     *  You must use std::endl or a flush() to push the data to the actual
     *  message handler.
     *
     *  Streaming a message to this will cause the simulation to have a non-zero
     *  exit code.
     */
    std::ostream&     msg_error;

    /*! The MSG_WARNING stream.
     *  You must use std::endl or a flush() to push the data to the actual
     *  message handler.
     */
    std::ostream&     msg_warning;

    /*! The MSG_INFO stream.
     *  You must use std::endl or a flush() to push the data to the actual
     *  message handler.
     */
    std::ostream&     msg_info;

    /*! The MSG_DEBUG stream.
     *  You must use std::endl or a flush() to push the data to the actual
     *  message handler.
     */
    std::ostream&     msg_debug;

private:

    typedef std::map<std::string, Port*>  PortMap;
    typedef std::pair<std::string, Port*> PortEntry;

    // public ports.
    PortMap       ports;

    // locally registered CADI interfaces.
    CADIMap localCADI;

protected:
    Component(std::string const & name, SimulationContext* simulationContext = NULL);

    /*! Publish a port so that external components can connect FROM it.
     *
     * \param name textual name for lookup
     * \param port port to add
     * \return true if port was added, false if there is a port with that name
     */
    bool publishMasterPort(const std::string& name, Port& port);

    /*! Publish a port so that external components can connect TO it.
     *
     * \param name textual name for lookup
     * \param port port to add
     * \return true if port was added, false if there is a port with that name
     */
    bool publishSlavePort(const std::string& name, Port& port);

    /*! Publish a port so that external components can connect to or from it.
     *
     * \param name textual name for lookup
     * \param port port to add
     * \return true if port was added, false if there is a port with that name
     */
    bool publishPeerPort(const std::string& name, Port& port);

    /*! Backwards compatibility. */
    bool addPort(const std::string& name_, Port& port) {
        return publishPeerPort(name_, port);
    }

    /*! Add an array of ports the list of named ports
     *
     * The ports will be named "name[0]", "name[1]" ... "name[size-1]"
     * Template type T is any subclass of Port.
     */
    template <typename T>
    bool addPortArray(const std::string& name_, T* array, unsigned int size) {
        return publishPeerPortArray(name_, array, size);
    }

    template <typename T>
    bool publishPeerPortArray(const std::string& name_, T* array, unsigned int size) {
        for (unsigned int i=0; i<size; ++i) {
            publishPeerPortIndex(name_, array[i], i);
        }
        return true;
    }

    template <typename T>
    bool publishMasterPortArray(const std::string& name_, T* array, unsigned int size) {
        for (unsigned int i=0; i<size; ++i) {
            publishMasterPortIndex(name_, array[i], i);
        }
        return true;
    }

    template <typename T>
    bool publishSlavePortArray(const std::string& name_, T* array, unsigned int size) {
        for (unsigned int i=0; i<size; ++i) {
            publishSlavePortIndex(name_, array[i], i);
        }
        return true;
    }

    /*! Publish an port for external bidirectional connection.
     *  The port will have a name of "name[n]" where n is the value of the
     *  index parameter.
     */
    bool publishPeerPortIndex(const std::string& name, Port& port, unsigned int index);

    /*! Publish an port for external components to connect FROM.
     *  The port will have a name of "name[n]" where n is the value of the
     *  index parameter.
     */
    bool publishMasterPortIndex(const std::string& name, Port& port, unsigned int index);

    /*! Publish an port for external components to connect TO.
     *  The port will have a name of "name[n]" where n is the value of the
     *  index parameter.
     */
    bool publishSlavePortIndex(const std::string& name, Port& port, unsigned int index);

    /*! Remove the port from the component. Note this will remove
     *  all registrations of the port
     *  \return true if one or more ports were removed
     */
    bool unpublishPort(Port& port);

    /*! Backwards compatibility. */
    bool removePort(Port& port) { return unpublishPort(port); }

    /*! Remove a port by name.
     *  Faster than the above
     *  \return the found port or NULL if no port was found
     */
    Port* unpublishPort(const std::string& name);

    /*! Remove an array of ports by name
     *
     */
    void unpublishPortArray(const std::string& name, unsigned int size);

public:
    /*! Default implementation of populateCADIMap that populates the
     *  map with the mxdi interface of this component and those of all
     *  subcomponents.
     */
    virtual void populateCADIMap(CADIMap &map, std::string const &prefix);

    /*! Add a named mxdi interface. The default mxdi interface should have
     *  a name "".
     */
    void addCADI(const std::string& name, eslapi::CAInterface* cadi);

    eslapi::CAInterface* removeCADI(const std::string& name);

    /*! Return a hierarchical name of subcomponent: "hierarchicalName:subcompName".
     *  This is used to pass the hierarchicalName into a subcomponent's constructor. Usually used
     *  in instantiation stage.
     *  E.g. If we use a factory to instantiate a subcomponent:
     *      add(factory->instantiate(subcompHierName("name")));
     *
     *  E.g. If we new a subcomponent:
     *      Component* comp = new Subcomp(subcompHierName("name"));
     *      add(comp);    // Or: comp->setParent(this);
     *
     *  Note: Some sub-components' name already contains dot '.' (such as "cpu0.l1cache"),
     *        so a colon is used as the separator instead. Otherwise we can not extract
     *        the subcomponent name.
     */
    inline const std::string subcompHierName(const std::string& subcompName)
    {
        return hierarchicalName + ":" + subcompName;
    }

public:
    virtual ~Component();

    /*! Get an abstract port by name
     *
     * \param name textual name of port added with addPort().
     * \return reference to port if found or reference to Component::NoPort.
     */
    Port& findPort(std::string const & name);

    /*! Get a port from a port array by name and index.
     *
     *  \param name textual name of port added with addPortArray().
     *  \param index index into port array, between 0 and size-1.
     *  \return reference to port if found, or reference to Component::NoPort.
     */
    Port& findPort(std::string const & name, int index);


    /*! Add a component to the list of 'sub' components
     *
     */
    void add(ComponentBase *c);

    void remove(ComponentBase *c);

    // apply parameter settings recursively
    void apply_params();

    virtual std::string getName() const { return name; }

    virtual std::string getHierarchicalName() const { return hierarchicalName; }

    virtual std::string getClassName() const;

    virtual const Node * getParent() const { return parent; }
    virtual void setParent(Node *p) { parent = p; }
};

class GroupComponent : public Component {
  public:
    GroupComponent(std::string const & name_) : Component(name_) { }
};

} // namespace sg

#endif // ARM_SG_COMPONENT_H
