/*!
 * \file    SGComponentBaseForwarder.h
 * \brief   Abstract Component base-class forwarding class
 * \date    Copyright 2017 ARM Limited. All rights reserved.
 *
 * This defines a private interface for the use of by ARM.  The interface is
 * subject to change.
 *
 */


#ifndef ARM_INCLUDE_sg_SGComponentBaseForwarder_h
#define ARM_INCLUDE_sg_SGComponentBaseForwarder_h

#include <string>
#include "sg/SGPort.h"
#include "sg/SGComponentBase.h"

namespace sg
{
    ///
    /// Forward from ComponentBase to a ComponentBase.
    ///
    class ComponentBaseForwarder : public sg::ComponentBase
    {
    public:
        ComponentBaseForwarder(sg::ComponentBase* object_ = nullptr);
        ~ComponentBaseForwarder();

        void setComponentBase(sg::ComponentBase* object_);

        //
        // From ComponentBase
        //
        virtual Port& findPort(std::string const & name) override;
        virtual Port& findPort(std::string const & name, int index) override;

        //
        // From Node
        //
        virtual std::string getName() const override;
        virtual std::string getClassName() const override;
        virtual const Node * getParent() const override;
        virtual void setParent(Node * p) override;
        virtual std::string getHierarchicalName() const override;

        // ComponentBase's are never connectables.
        virtual Connectable* asConnectable() override;
        virtual const Connectable* asConnectable() const override;

        //
        // Utilities for connecting ports to this class
        //

        // This is the kind of the port in the wrapped component.
        enum kind_in_wrapped_component_t {
            KIND_IN_WRAPPED_COMPONENT_Master,
            KIND_IN_WRAPPED_COMPONENT_Slave,
            KIND_IN_WRAPPED_COMPONENT_Peer
        };

        // Given that a port exists with the name port_name_ of the specified
        // kind_ then connect it to port_.
        void connectPort(
            kind_in_wrapped_component_t   kind_,
            std::string const&            port_name_,
            sg::Port&                     port_
            );

        // Given that a port array exists with the base name port_base_name_ of
        // the specified kind_ then connect it to port_array_.
        template<typename TypeOfPort, unsigned NumberOfPorts>
        void connectPort(
            kind_in_wrapped_component_t   kind_,
            std::string const&            port_base_name_,
            TypeOfPort                    (&port_array_)[NumberOfPorts]
            )
        {
            Port* port_pointers[NumberOfPorts];
            for (unsigned i = 0; i != NumberOfPorts; ++ i)
                port_pointers[i] = &port_array_[i]; // also checks that TypeOfPort derives from sg::Port

            connectPortArray( kind_, port_base_name_, port_pointers, NumberOfPorts );
        }
        
    private:
        sg::ComponentBase*       object;

        void connectPortArray(
            kind_in_wrapped_component_t   kind_,
            std::string const&            port_base_name_,
            sg::Port* const*              array_of_port_pointers_,
            unsigned                      number_of_ports_
            );

        ComponentBaseForwarder(ComponentBaseForwarder const&) = delete;
        void operator =(ComponentBaseForwarder const&) = delete;
    };
}

#endif
