 /*!
 * \file    SGControlProtocol.h
 * \brief   Internal port protocol used to connect hierarchy
 *          of SystemGenerator components.
 * \date    Copyright 2006 ARM Limited. All rights reserved.
 *          Copyright 2013 ARM Limited. All rights reserved.
 *
 * This file defines the port protocol used on the system '$control'
 * ports implemented by all components.
 */

#ifndef ARM_SG_CONTROL_PROTOCOL_H
#define ARM_SG_CONTROL_PROTOCOL_H

#include <map>
#include <string>

#include "sg/SGPort.h"
#include "sg/SGParams.h"
#include "cp/SerializationInterface.h"

class SimulationEngine;

namespace eslapi {
    class CAInterface;
} // namespace eslapi

namespace sg {

namespace message
{
    enum MessageTypeEnum
    {

        // Actual problem from which simulation cannot recover
        MSG_FATAL_ERROR = -2,

        // Actual problem from which simulation might be able to recover
        MSG_ERROR = -1,

        // Potential problem but simulation can carry on
        MSG_WARNING = 0,

        // Informative messages
        MSG_INFO = 1,
        MSG_DEBUG = 2
    };

    typedef int MessageType;
} // namespace message

typedef std::map< std::string, eslapi::CAInterface* > CADIMap;

// protocol class definition
class SGControlProtocol : public Port
{
public:
    // connector instances
    SlaveConnector< void (const Params&) > configure;
    SlaveConnector< void (void) > init;
    SlaveConnector< void (void) > interconnect;
    SlaveConnector< void (int) > reset;
    SlaveConnector< void (const std::string&) > loadApplicationFile;
    SlaveConnector< void (const char *name, cp::SerializationInterface*) > serializeState;
    SlaveConnector< void (void) > disconnect;
    SlaveConnector< void (void) > terminate;
    SlaveConnector< void  (CADIMap &, std::string const &prefix) > populateCADIMap;
    SlaveConnector< void (void) > step;
    SlaveConnector< std::string (const std::string&) > getProperty;
    SlaveConnector< std::string (const std::string&) > bypass;

    MasterConnector< std::string (const std::string&) > callSimEngine;

    // event-driven simulation engine
    SlaveConnector< void (void) > run;
    SlaveConnector< void (void) > stop; // asynchronous?
    PeerConnector< void (void) > idle;
    MasterConnector< void (void) > quit;
    SlaveConnector< void (SimulationEngine*) > setSimulationEngine;
    MasterConnector< void (const std::string&, message::MessageType) > message;
    SlaveConnector< void (const std::string&, const std::string&, const std::string&) > setParameterAttribute;


    static char const* PROTOCOL_ID() { return "ARM::Builtin.ControlProtocol"; } // SPECIFIC TO SGControlProtocol
    
    // constructor
    SGControlProtocol()
    {
        setProtocolIDWithAssert(PROTOCOL_ID());
        
        define_behaviour(configure);
        define_behaviour(init);
        define_behaviour(interconnect);
        define_behaviour(reset);
        define_behaviour(loadApplicationFile);
        define_behaviour(serializeState);
        define_behaviour(disconnect);
        define_behaviour(terminate);
        define_behaviour(populateCADIMap);
        define_behaviour(step);
        define_behaviour(getProperty);
        define_behaviour(bypass);

        define_behaviour(callSimEngine);

        // event-driven simulation engine
        define_behaviour(run);
        define_behaviour(stop);
        define_behaviour(idle);
        define_behaviour(quit);
        define_behaviour(setSimulationEngine);
        define_behaviour(message);
        define_behaviour(setParameterAttribute);
    }
};

} // namespace sg

#endif /* ifndef ARM_SG_CONTROL_PROTOCOL_H */
