///
/// \file    SGDeferredTraceDefaultAllocator.h
/// \brief   An allocator suitable as the default allocator for deferred trace.
/// \date    Copyright 2013 ARM Limited. All rights reserved.
///

#ifndef ARM_INCLUDE_SGDEFERREDTRACEDEFAULTALLOCATOR_H
#define ARM_INCLUDE_SGDEFERREDTRACEDEFAULTALLOCATOR_H

#include <cstddef>
#include <cassert>
#include <cstdint>

#include <iostream>

//#define HERE_STASH if (0) ; std::cerr << __FILE__ << ":" << __LINE__ << ": "

///
/// Models the allocator concept for the deferred trace sources.
///

namespace trace_source_details
{
    struct storage_t
    {
        // storage_len is not a particularly special number.  Choose it so
        // that the total size is something like 64 KiB on the basis that
        // the heap allocators might be more efficient for a power of two?
        static std::size_t const storage_len = 4096;
        // 64 * 1024 - sizeof(storage_t*) - sizeof(std::size_t);
        storage_t() : next(0), valid_data_size(0)
        {
            assert( end() - begin() == 0       );
            assert( free_size() == storage_len );
        }

        storage_t*   next;
        std::size_t  valid_data_size;  // the amount of storage[] storing valid data.

        uint8_t      storage[storage_len];

        uint8_t*           begin()       { return storage;          }
        uint8_t*           end()         { return storage + size(); }

        uint8_t const*     begin() const { return storage;          }
        uint8_t const*     end()   const { return storage + size(); }

        std::size_t        size()  const { return valid_data_size;  }

        std::size_t        free_size() const
        {
            assert( valid_data_size <= sizeof(storage) );
            return std::size_t(sizeof(storage) - valid_data_size);
        }

        uint8_t*     allocate_from_this_store(std::size_t sz_);

        storage_t*   end_of_chain()
        {
            storage_t* p = this;
            while (p->next != 0)
                p = p->next;
            return p;
        }

        void clear_this()
        {
            valid_data_size = 0;
        }

        void checks() const;
    };

    class stash_t
    {
        // We lazy allocate head and end_of_chain so we only pay the cost if
        // we use them.

        storage_t*         head;
        storage_t*         end_of_chain;

        bool check_end_of_chain() const;
        inline void checks() const;
        uint8_t* priv_slow_path_allocate(std::size_t sz_);

    public:
        uint8_t* priv_allocate(std::size_t sz_);

        bool has_stash() const { return head != 0; }

        struct traverser {
            traverser( storage_t* storage_ = 0 ) : storage(storage_), next_chunk(0) {}

            storage_t*    storage;

            // The amount of storage that has already been consumed in the
            // node 'storage'.
            std::size_t   next_chunk;

            uint8_t* get_chunk();
            void     advance(std::size_t sz_);

            std::size_t get_chunk_size() const;
        };

        traverser begin() { return traverser(head); }

        stash_t() : head(0), end_of_chain(0) {}
        ~stash_t();

        /// Unconditionally clear the current stash without
        /// executing anything.
        void clear();
    };

    storage_t* attach_more_storage( storage_t* end_node_ = 0 );



    ///
    /// This allocator has a chain of chunks of memory that it uses to
    /// allocate memory.
    ///
    class default_allocator
    {
    public:

#ifdef _MSC_VER
#pragma warning( push )
#pragma warning(disable: 4355)
#endif
        default_allocator() : stash(), count_of_nested_exec_and_clear(0), dbg_count_of_nested_execs(0) {}
#ifdef _MSC_VER
#pragma warning( pop )
#endif

        ~default_allocator();

        // ----------------------------------------------------------------------
        //                   Allocator concept methods
        // ----------------------------------------------------------------------
        //
        // All allocators for deferred trace have to implement the following
        // interface...

        // A call that allocates space for a 'virtual thunk' -- the storage
        // where the information about how to call the deferred function is
        // stored.  This is called first by the deferred function system and
        // once per function.
        uint8_t* allocate_virtual_thunk(std::size_t sz_)
        {
            uint8_t* ret = stash.priv_allocate(sz_);
//            HERE_STASH << "allocate_virtual_thunk(" << sz_ << ") -> " << (void*)ret << "\n";
            return ret;
        }

        // A call that allocates space for storing all the arguments to the
        // deferred function.  This is represented as an 'arg_holder_t<>' that
        // doesn't necessarily store the args directly as their original types
        // but as 'stash_types' as mapped by the traits supplied to the deferred
        // function system.  For example, storing all the built-in types (int,
        // double, etc.) then their stashed type and their actual type will be
        // the same, but a std::string object will be stored as a more
        // complicated representation which will use dynamic data (as we will
        // see...)
        //
        // This is called second by the deferred function system and is called
        // once per function.
        uint8_t* allocate_arg_holder(std::size_t sz_)
        {
            uint8_t* ret = stash.priv_allocate(sz_);
//            HERE_STASH << "allocate_arg_holder(" << sz_ << ") -> " << (void*)ret << "\n";
            return ret;
        }

        // So if we have a more complicated type to stash, say a std::string or
        // char const* then we have to copy the data out of that string and so
        // we need to allocate some dynamic data and also have in the 'stash
        // type' of std::string a representation that can point to that dynamic
        // data.   So we require the allocator to supply a 'pointer' type.
        //
        // We could just typedef uint8_t* pointer, but here we wrap it in a
        // struct to show that it doesn't need to have many properties to work.
        //
        // This is called zero or more times after allocate_arg_holder()
        // depending on what the stashed types are.
        struct pointer {
            uint8_t*  p;

            // Public interface of pointer concept
            pointer() {}
            operator uint8_t*()             { return p; }
            operator uint8_t const*() const { return p; }
        private:
            friend class default_allocator;

            void init(uint8_t* p_) { p = p_; }

            pointer(pointer const&); // disallow
            void operator = (pointer const&); // disallow
        };

        // And we need a representation of a 'size'.  The default allocator has
        // a maximum size of a little less than 6 KiB so we use uint16_t just to
        // save a few bytes.  This is expected to be convertible to std::size_t.
        //
        // NOTE that most of the other functions do not use this by std::size_t
        // this is to avoid having to do lots of casts, though perhaps not such
        // a great decision?
        typedef uint16_t size_type;

        // The function a stash type asks the allocator for dynamic data is this
        // one.
        //
        // NOTE that we update a pointer in situ, so that we don't require a
        // pointer to have a copy constructor or assignment operator.  This
        // would allow more 'interesting' pointer types to be used.
        void allocate_dynamic_data(std::size_t sz_, pointer& o_)
        {
            uint8_t* ret = stash.priv_allocate(sz_);
            o_.init(ret);

//            HERE_STASH << "allocate_dynamic_data(" << sz_ << ") -> " << (void*)ret << "\n";
        }

        // Once a stash type has allocated some memory then it may need to copy
        // in the representation of the data that it is looking for.
        //
        // NOTE not inline to void pulling in <algorithm>
        void copy( uint8_t const* begin_, uint8_t const* end_, pointer& destination_ );

        // A stash type might want to compare two bytes streams are the same.
        //
        // NOTE not inline to void pulling in <algorithm>
        static bool is_equal( pointer const& begin_, pointer const& other_begin_, std::size_t sz_ );

        // At some point you are going to want to execute those stashed thunks.
        //
        // From the functions called by exec() do _not_ call clear() or
        // exec_and_clear().
        void exec();

        // Clear up the storage associated with this.
        //
        // Do not call if current we are executing functions via exec() or
        // exec_and_clear().
        void clear();

        // Or perhaps do both at once!
        //
        // NOTE that if any more deferred functions are added during execution
        // of a deferred function then they _will_ be executed by this call
        // completing.
        //
        // NOTE that if a deferred function calls exec_and_clear() during its
        // call then those functions that have _already_ been executed will
        // execute not execute again.
        //
        // Deferred functions will execute in the order in which they were
        // added; even in the presence of adding more functions and calling
        // exec_and_clear() from within the deferred function.
        //
        // Do NOT call exec() or clear() within a deferred function call.
        //
        void exec_and_clear();

        // Are there no functions stashed?
        bool empty() const;

        // ----------------------------------------------------------------------
        //    Additional methods specific to trace source deferred allocators
        // ----------------------------------------------------------------------

        // We use this so that we can chose whether to defer or not defer trace
        // sources globally.   This provides us with a testing methodology.
        bool is_deferring() const
        {
            return true;
        }

    public:

        // ----------------------------------------------------------------------
        //                        Implementation
        // ----------------------------------------------------------------------

        //
        // This allocator stores the memory as a set nodes, carving chunks out
        // from the end as necessary.  It maintains a free pool of nodes so that
        // once it has built up a working set of node then it probably doesn't
        // need to call new or delete again.
        //
        // The trace sources call the allocate functions in the following order:
        //      * allocate_virtual_thunk() (once)
        //      * allocate_arg_holder() (once)
        //      * allocate_dynamic_data() (zero or more times).
        //
        // We store them next to each other in memory and we don't bother
        // aligning anything.
        //


        static std::size_t const storage_len = storage_t::storage_len;
    private:

        stash_t               stash;
        unsigned              count_of_nested_exec_and_clear;
        unsigned              dbg_count_of_nested_execs;
        stash_t::traverser    current_exec_and_clear_position;

        void priv_exec(stash_t::traverser&);
        void priv_clear();

    };


    // ----------------------------------------------------------------------
    //                  Allocator used for testing purposes
    // ----------------------------------------------------------------------
    class test_default_allocator : public default_allocator
    {
    public:
        test_default_allocator() : defer_trace(true) {}

        bool is_deferring() const             { return defer_trace;         }
        void set_deferring(bool defer_trace_) { defer_trace = defer_trace_; }

    private:
        bool              defer_trace;
    };

    // ----------------------------------------------------------------------
    //                                Inlines
    // ----------------------------------------------------------------------

}

#endif
