/*!
 * \file    SGLCLib.h
 * \brief   License checking related functionality.
 * \date    Copyright 2011-2017 ARM Limited. All rights reserved.
 *
 */

#ifndef SGLCLIB_H
#define SGLCLIB_H

#include <map>
#include <string>
#include <eslapi/CAInterface.h>


#ifndef UNOBFUSCATE_LICENSE_CHECKING
// obfuscate the LCLib symbols by renaming the functions/classes to arbitrary names
// classes
#define LCLibLicenseCheckingModule             dWJXfW5n
#define LCLibLicenseCheckingModuleRegistry     AypUi241
#define LCLibContext S15061_0025
// functions
#define LCLibDecryptString S15061_0027
#define LCLibGetFieldValue S15061_0028
#define LCLibGenerateMessage S15061_0029
#define LCLibGetMessageMap S15061_0030
#define LCLibEncryptString S15061_0031
#define LCLibGetHexChecksum S15061_0032
#define LCLibRegisterLicenseCheckingModule     U87AsLMu
#define LCLibUnregisterLicenseCheckingModule   C8FyBl4H
#define LCLibRegisterCallback S15061_0035
#define LCLibUnregisterCallback S15061_0036
#define LCLibDoCommand S15061_0037
#endif

namespace sg
{
    // --- License checking module interface -----------------------------------------------------

    // Interface of a license checking module.
    class LCLibLicenseCheckingModule
    {
    public:
        // constructor and destructor
        LCLibLicenseCheckingModule() {}
        virtual ~LCLibLicenseCheckingModule() {}

        // Execute one license checking command.
        // 'input' contains the command and its parameters. It must be encoded using
        // EncryptString(GenerateMessage(map<key,value>)).
        // 'output' is overwritten and contains the return values. It must be decrypted
        // using s=DecryptString() and then the fields can be accessed with GetMessageValue(s, "key").
        // The input message always contains a 'command' field. The following commands are (and must be)
        // supported (external command set):
        // - command=CheckOut: check out a license
        // - command=CheckIn: check a license back in which has been checked out before
        // The CheckOut command has the following parameters:
        // - feature: Name of the feature to check out
        // - version: Version of the feature to check out
        // The CheckIn command has the following parameters:
        // - feature: Name of the feature to check in
        virtual void LCLibDoCommand(const std::string &input, std::string &output) = 0;

        // register/unregister callback object (reserved for future extensions requiring callbacks)
        virtual void LCLibRegisterCallback(LCLibLicenseCheckingModule *callbackObject) = 0;
        virtual void LCLibUnregisterCallback(LCLibLicenseCheckingModule *callbackObject) = 0;
    };

    // Interface of the license checking module registry.
    // This is accessible using getGlobalInterface()/scx::scx_get_global_interface() using
    // sg::obtainComponentInterfacePointer<sg::LicenseCheckingModuleRegistry>(globalInterface, "lmRegistry").
    class LCLibLicenseCheckingModuleRegistry: public eslapi::CAInterface
    {
    public:
        // constructor and destructor
        LCLibLicenseCheckingModuleRegistry() {}
        virtual ~LCLibLicenseCheckingModuleRegistry() {}

        // CAInterface functionality
        static eslapi::if_name_t IFNAME() {
            // "sg.LCLibLicenseCheckingModuleRegistry" obfuscated in binary
            static char buf[] = "nb)G>Gd]Gd^`in`>c`^fdibHj_pg`M`bdnomt";
            // decode the first time we come here
            if (buf[0] == 'n')
                for (char *p = buf; *p; p++)
                    *p += 5;
            return buf;
        }
        static eslapi::if_rev_t IFREVISION() { return 0; }
        virtual eslapi::CAInterface *ObtainInterface(eslapi::if_name_t ifName, eslapi::if_rev_t minRev, eslapi::if_rev_t *actualRev) = 0;

        // register/unregister a LicenseCheckingModule
        // (the moduleId is the same as the 'Id' field of the 'MapFeature RouteTo=LicenseBackend Id=foo' statement)
        virtual void LCLibRegisterLicenseCheckingModule(const char *moduleId, LCLibLicenseCheckingModule *lcm) = 0;
        virtual void LCLibUnregisterLicenseCheckingModule(const char *moduleId) = 0;
    };

    // Opaque encryption/decryption context
    class LCLibContext
    {
    public:
        void* context;
        virtual ~LCLibContext() {}
    };

    // --- Encryption, decryption and checksum functions -----------------------------------------------------

    // Decrypt string and return the plain text.
    std::string LCLibDecryptString(const std::string& encryptedData, const std::string& key1, unsigned int key2, LCLibContext **contextPtrPtr = 0);

    // Return the value of one specific 'field' in the decrypted input string.
    // If the field is not present the defaultValue is returned.
    std::string LCLibGetFieldValue(const std::string &plainText, const std::string &field, const std::string &defaultValue = std::string());

    // Convert a map from field name to field value into a single string in a reversible way.
    // This is the counterpart of GetFieldValue().
    std::string LCLibGenerateMessage(const std::map<std::string,std::string> &fields);

    // Return a map<string,string> which maps the field names onto their values.
    std::map<std::string,std::string> LCLibGetMessageMap(const std::string &plainText);

    // Encrypt plain text and return encrypted data.
    // The returned data is a binary string potentially containing all bytes from 0 to 255.
    std::string LCLibEncryptString(const std::string& plainText, const std::string& key1, unsigned int key2, LCLibContext **contextPtrPtr = 0);

    // Return a cryptographically safe checksum of string s which can be and data or plain text.
    // The returned string is the checksum in hex format.
    std::string LCLibGetHexChecksum(const std::string& s);

} // namespace sg

#endif
