/*!
 * \file    SGSchedulerInterfaceForComponents_v1.h
 * \brief   Abstract scheduler interface used by modelling components to create threads and timers etc.
 * \date    Copyright 2015 ARM Limited. All rights reserved.
 */

#ifndef SGSCHEDULERINTERFACEFORCOMPONENTSV1_H
#define SGSCHEDULERINTERFACEFORCOMPONENTSV1_H

#include "SGSchedulerInterfaceForComponents_v0.h"

namespace sg
{
namespace v1_scheduler
{

/*! Base class for all scheduler objects and interfaces. */
using v0_scheduler::SchedulerObject;

/*! FrequencyObserver: abstract base class for objects which want to receive notification messages from
 * FrequencySource objects when the clock frequency changed.
 */
using v0_scheduler::FrequencyObserver;

/*! Frequency source: Provides a clock frequency.
 * The clock frequency can be queried (getFrequency()) and the object can also notify
 * other objects about clock frequency changes (register/unregisterFrequencyObserver()).
 */
using v0_scheduler::FrequencySource;

/*! TimerCallback:
 * Callback base class for timers created through SchedulerInterfaceForComponents::createTimer().
 */
using v0_scheduler::TimerCallback;

/*! Interface of timers. Instances of objects implementing this
 *  interface are created using the createTimer() method on SchedulerInterfaceForComponents.
 *
 * \warning setFrequency() function is deprecated and is not used.
 *
 * \warning values passed to set() represent ticks relative to simulated time resolution.
 *
 * \warning values returned by remaining() represent ticks relative to simulated time resolution.
 */
using v0_scheduler::Timer;

/*! Interface of threads (runnable side): These objects are created and implemented by the modelling components and
 * a pointer to such an interface is passed to SchedulerInterfaceForComponents::createThread().
 * The scheduler uses this interface to run the thread and control the thread execution.
 *
 * \warning breakQuantum() function is deprecated.
 *          Use SchedulerInterfaceForComponents::addSynchronizationPoint() instead.
 */
using v0_scheduler::SchedulerRunnable;

/*! Modified version of SchedulerRunnable interface class: getName() is renamed here to getRunnableName().
 * The sole reason for the existence of this class is to disambiguate this from other classes which might
 * have a getName() method with a different prototype or different semantics which then defeats multiple
 * inheritance.
 * The rest of the interface and semantics of this class is identical to class SchedulerRunnable.
 */
using v0_scheduler::SchedulerRunnableWithGetRunnableName;

/*! see v0_scheduler::SchedulerThread
 */
class SchedulerThread: public v0_scheduler::SchedulerThread
{
public:
    /*! returns the SchedulerRunnable of this thread.
    * the runnable might be used to access ThreadProperties
    */
    virtual SchedulerRunnable *getRunnable() const { return 0; }
};

/*! ThreadSignal:
 * A non schedulable but waitable event on which many threads may wait.
 * When the event is signalled all waiting threads are scheduled to run.
 */
using v0_scheduler::ThreadSignal;

/*! SchedulerCallback:
 * Callback base class for callbacks registered with for SchedulerInterfaceForComponents::addCallback/removeCallback().
 */
using v0_scheduler::SchedulerCallback;

/*! Tag:
 * This class is use to select v1 versus v0 versions of get...() methods of the SchedulerInterfaceForComponents.
 * The v1 version is selected by passing a null pointer to the extra Tag<> argument.
 * \see getGlobalQuantum(), getMinSyncLatency(), getNextSyncPoint(), getCurrentSimulatedTime().
 */
template<class T>
class Tag
{
private:
    /*! Constructor: Hidden as this class is not intended to be instantiated.
     */
    Tag() {};
};

/*! SchedulerInterfaceForComponents:
 * This interface is used by modelling components to create threads, timers etc.
 * This interface is usually not used by the owner of the scheduler. Every simulation framework must implement this
 * interface and can put any kind of scheduler behind this interface.
 * Modelling components usually use this to get access to this interface:
 * \code
 * sg::obtainComponentInterfacePointer<sg::SchedulerInterfaceForComponents>(simulationContext->getGlobalInterface(),
 *                                                                          "scheduler")
 * \endcode
 * \see sg/SGComponentRegistry.h
 */
class SchedulerInterfaceForComponents: public v0_scheduler::SchedulerInterfaceForComponents
{
private:

    // Used in default implementation over v0 version of this interface
    class GlobalFrequencySource: public FrequencySource
    {

    /* Construction */
    public:
        explicit GlobalFrequencySource(double f): m_frequency(f) {}
        ~GlobalFrequencySource() {}

    /* Accessors */
    public:
        const char *getName() const { return "sg.GlobalFrequencySource"; }
        void registerFrequencyObserver(FrequencyObserver *) {}
        void unregisterFrequencyObserver(FrequencyObserver *) {}
        double getFrequency() { return (m_frequency); }

    /* Implementation */
    protected:
        void notifyFrequencyObservers() {}

    private:
        double m_frequency;
    };

public:
    // CAInterface related stuff
    static eslapi::if_name_t IFNAME() { return "sg.SchedulerInterfaceForComponents"; }
    static eslapi::if_rev_t IFREVISION() { return 1; }

    // Main interface, used by modelling components
    //
    // Note: Modelling components which want to use these scheduler objects should generally use the generic
    // scheduler related modelling components instead of using the functionality of this module directly:
    // - Use ClockTimer and ClockTimer64 instead of using createTimer() directly.
    // - Use AsyncSignal instead of using addCallback() and removeCallback() directly.
    // - Use MasterClock and ClockDivider to create clock trees instead of using FrequencySource and FrequencyObserver
    //   directly.

    /*! Wait: Block current thread and run other threads for some time.
     * This function should only be called from within a SchedulerRunnable::threadProc() context.
     * Calling this from a context outside of a threadProc() is invalid.
     * It will block the currently running thread for the specified amount of time and let
     * other threads run.
     * This function blocks while the other threads etc are running and returns
     * when the calling thread should continue to run. This is the coroutine switching point.
     * Typically a thread calls wait(n) in its loop each time it completed n ticks of work.
     * n is called a quantum.
     * ticks may also be 0: This means that the current thread is potentially blocked and other callbacks which
     * are pending (if any) are run, but no time is advanced for this thread and this thread is most likely
     * scheduled again immediately.
     * \param ticks Time to wait for. This is relative to simulated time resolution.
     */
    virtual void wait(ticks_t ticks)
    {
        static GlobalFrequencySource f(1.0 / getSimulatedTimeResolution());

        wait(ticks, & f);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::wait;
    virtual void wait(ticks_t ticks, FrequencySource *timebase)
    {
        if (timebase && (timebase->getFrequency() > 0))
        {
            wait(ticks_t(double(ticks) / timebase->getFrequency() / getSimulatedTimeResolution() + 0.5));
        }
        else
        {
            wait(ticks_t(0));
        }
    }

    /*! Sets the value of the global quantum.
     * The global quantum is the maximum time a thread can run ahead of simulation time.
     * All threads must synchronize on timing points that are multiples of the global quantum.
     * \param ticks global quantum value measured in ticks relative to simulated time resolution.
     */
    virtual void setGlobalQuantum(ticks_t ticks)
    {
        static GlobalFrequencySource f(1.0 / getSimulatedTimeResolution());

        setGlobalQuantum(ticks, & f);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::setGlobalQuantum;
    virtual void setGlobalQuantum(ticks_t ticks, FrequencySource * timebase)
    {
        if (timebase && (timebase->getFrequency() > 0))
        {
            setGlobalQuantum(ticks_t(double(ticks) / timebase->getFrequency() / getSimulatedTimeResolution() + 0.5));
        }
        else
        {
            setGlobalQuantum(ticks_t(0));
        }
    }

    /*! Returns the value of the global quantum in ticks relative to simulated time resolution.
     */
    virtual ticks_t getGlobalQuantum(Tag<ticks_t> *)
    {
        return ticks_t(getGlobalQuantum() / getSimulatedTimeResolution() + 0.5);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::getGlobalQuantum;
    virtual double getGlobalQuantum()
    {
        return double(getGlobalQuantum(0)) * getSimulatedTimeResolution();
    }

    /*! Sets the minimum synchronization latency for this scheduler.
     * The minimum synchronization latency helps to ensure that sufficient simulated time has passed between two
     * synchronization points for synchronization to be efficient.
     * A small latency increases accuracy but decreases simulation speed.
     * A large latency decreases accuracy but increases simulation speed.
     * The scheduler uses this information to compute the next synchronization point as returned by getNextSyncPoint().
     * \param ticks specifies the minimum synchronization latency measured in ticks relative to simulated time resolution.
     * \see getMinSyncLatency()
     */
    virtual void setMinSyncLatency(ticks_t ticks)
    {
        static GlobalFrequencySource f(1.0 / getSimulatedTimeResolution());

        setMinSyncLatency(ticks, & f);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::setMinSyncLatency;
    virtual void setMinSyncLatency(ticks_t ticks, FrequencySource * timebase)
    {
        if (timebase && (timebase->getFrequency() > 0))
        {
            setMinSyncLatency(ticks_t(double(ticks) / timebase->getFrequency() / getSimulatedTimeResolution() + 0.5));
        }
        else
        {
            setMinSyncLatency(ticks_t(0));
        }
    }

    /*! Returns the minimum synchronization latency for this scheduler.
     * \return minimum synchronization latency. Measured in ticks relative to simulated time resolution.
     * \see setMinSyncLatency()
     */
    virtual ticks_t getMinSyncLatency(Tag<ticks_t> *)
    {
        return ticks_t(getMinSyncLatency() / getSimulatedTimeResolution() + 0.5);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::getMinSyncLatency;
    virtual double getMinSyncLatency()
    {
        return double(getMinSyncLatency(0)) * getSimulatedTimeResolution();
    }

    /*! Add future global synchronisation point in time.
     * Modelling components can call this function to provide a hint to the scheduler implementation
     * when a potentially useful global synchronisation point will occur in the future.
     * The scheduler uses this information to determine the quantum sizes of threads as
     * they get scheduled.
     * Calling this function again will add another global synchronisation point.
     * Synchronisation points automatically vanish when the time is reaching them.
     * \param ticks Time relative to the start of the current quantum (for threads) or relative
     *        to the current scheduler time (now) for all other callbacks.
     *        Measured in ticks relative to simulated time resolution.
     *
     * \warning addSynchronisationPoint() function is deprecated.
     *          Use sc_core::sc_prim_channel::async_request_update() instead.
     */
    virtual void addSynchronisationPoint(ticks_t ticks)
    {
        static GlobalFrequencySource f(1.0 / getSimulatedTimeResolution());

        addSynchronisationPoint(ticks, & f);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::addSynchronisationPoint;
    virtual void addSynchronisationPoint(ticks_t ticks, FrequencySource *timebase)
    {
        if (timebase && (timebase->getFrequency() > 0))
        {
            addSynchronisationPoint(ticks_t(double(ticks) / timebase->getFrequency() / getSimulatedTimeResolution() + 0.5));
        }
        else
        {
            addSynchronisationPoint(ticks_t(0));
        }
    }

    /*! Return the next future global synchronisation point in time.
     * Modelling components can call this function to get a hint for when a potentially useful global synchronisation
     * point will occur in the future.
     * The quantum keeper uses this information to determine when core threads need to synchronize.
     * \return next global synchronisation point in ticks, relative to simulated time resolution.
     */
    virtual ticks_t getNextSyncPoint(Tag<ticks_t> *)
    {
        return ticks_t(getNextSyncPoint() / getSimulatedTimeResolution() + 0.5);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::getNextSyncPoint;
    virtual double getNextSyncPoint()
    {
        return double(getNextSyncPoint(0)) * getSimulatedTimeResolution();
    }

    /*! Return the next synchronisation range, that is [ minimum synchronization latency, next synchronisation point ].
     * Modelling components can call this function to get a hint for when a potentially useful global synchronisation
     * point will occur in the future, as well as to update their view of the minimum synchronization latency.
     * The core quantum keeper uses this information to determine when core threads need to and can synchronize.
     * \param min_sync_latency minimum synchronization latency. Measured in ticks, relative to simulated time resolution.
     * \param next_sync_pt next global synchronisation point. Measured in ticks, relative to simulated time resolution.
     */
    using v0_scheduler::SchedulerInterfaceForComponents::getNextSyncRange;
    virtual void getNextSyncRange(ticks_t & min_sync_latency, ticks_t & next_sync_pt)
    {
        min_sync_latency = getMinSyncLatency(0);
        next_sync_pt = getNextSyncPoint(0);
    }

    /*! Get current elapsed simulated time in seconds since the scheduler was created.
     * The granularity of this clock is determined by the timers and their
     * expiry intervals and by the intervals of the wait() statements.
     * It always accurately reflects the time on the last timer callback invocation or
     * the last return from SchedulerThread::wait(), whichever occurred last.
     * The return values are guaranteed to monotonically increase over (real or simulated) time.
     * Used by: ClockDivider and MasterClock (ClockSignalProtocol::currentTicks())
     * \returns Current simulated time in ticks, relative to simulated time resolution.
     */
    virtual ticks_t getCurrentSimulatedTime(Tag<ticks_t> *)
    {
        return ticks_t(getCurrentSimulatedTime() / getSimulatedTimeResolution() + 0.5);
    }
    using v0_scheduler::SchedulerInterfaceForComponents::getCurrentSimulatedTime;
    virtual double getCurrentSimulatedTime()
    {
        return double(getCurrentSimulatedTime(0)) * getSimulatedTimeResolution();
    }

    /*! Get simulated time resolution in seconds.
     * \returns Simulated time resolution in seconds.
     */
    virtual double getSimulatedTimeResolution()
    {
        return 1e-12;
    };

    /*! Set simulated time resolution in seconds.
     * \param resolution Simulated time resolution in seconds.
     *
     * \note Setting simulated time resolution after simulation has started is not possible.
     *
     * \note Setting simulated time resolution after timers have been set is not possible.
     */
    virtual void setSimulatedTimeResolution(double resolution) { (void)resolution; }

    /*! Return current SchedulerThread object or NULL if not in any threadProc() call.
     */
    virtual SchedulerThread * currentThread()
    {
        return 0;
    }
    
    /*! Returns a v1_scheduler::SchedulerThread
    * see v0_scheduler::SchedulerInterfaceForComponents::createThread
    */
    virtual SchedulerThread *createThread(const char* name, SchedulerRunnable *runnable) = 0;
};

} // namespace v1_scheduler

} // namespace sg

#endif  /* SGSCHEDULERINTERFACEFORCOMPONENTSV1_H */
