/*!
 * \file    SGTime.h
 * \brief   Class definition for simulated time representation.
 * \date    Copyright 2016 ARM Limited. All rights reserved.
 */

#ifndef SGTIME_H
#define SGTIME_H

/* Includes */

#include "SGTicks.h"
#include "SGSimulationContext.h"
#include "SGSchedulerInterfaceForComponents.h"

#include <cassert>
#include <string>

/* Namespaces */

namespace sg {

/* Datatypes */

/*!
 * \brief   Class Time is used to represent simulated time, as well as time intervals, including delays and time-outs.
 *
 * An object of class Time is constructed from a ticks_t value and a FrequencySource instance.
 * Time is represented internally as ticks_t, relative to simulated time resolution.
 *
 * Objects of this class are intended to be used by Component objects to manipulate simulated time values.
 *
 * \see     FrequencySource, SimulationContext
 */
class Time {

    /* Construction */
    public:
        Time();
        Time(ticks_t, double simulation_ticks_per_tick);
        Time(ticks_t, FrequencySource *, SimulationContext *);
        Time(ticks_t, FrequencySource *, SchedulerInterfaceForComponents *);
        Time(ticks_t, double, SimulationContext *);
        Time(ticks_t, double, SchedulerInterfaceForComponents *);
        static Time fromSeconds(double, SimulationContext *);
        static Time fromSeconds(double, SchedulerInterfaceForComponents *);
        static Time fromTicks(ticks_t);

    /* Copying */
        Time(const Time &);
        Time & operator=(const Time &);

    /* Conversions */
        ticks_t ticks() const;
        ticks_t toTicks(FrequencySource *, SimulationContext *) const;
        ticks_t toTicks(FrequencySource *, SchedulerInterfaceForComponents *) const;
        ticks_t toTicks(double, SimulationContext *) const;
        ticks_t toTicks(double, SchedulerInterfaceForComponents *) const;
        double toSeconds(SimulationContext *) const;
        double toSeconds(SchedulerInterfaceForComponents *) const;
        std::string toString(SimulationContext *) const;
        std::string toString(SchedulerInterfaceForComponents *) const;

    /* Operators */
        bool operator==(const Time &) const;
        bool operator!=(const Time &) const;
        bool operator<(const Time &) const;
        bool operator<=(const Time &) const;
        bool operator>(const Time &) const;
        bool operator>=(const Time &) const;
        Time & operator+=(const Time &);
        Time & operator-=(const Time &);
        Time & operator*=(double);
        Time & operator/=(double);
        
    /* Implementation  */
    private:
        ticks_t m_value;
};

/* Functions - Inlines */

/*!
 * \brief   Default constructor.
 */
inline
Time::Time():
    m_value(0) {
}

/*!
 * \brief   Construct a Time instance from a local ticks_t value and a simulation
 *          ticks multiplier
 */
inline
Time::Time(ticks_t local_ticks, double simulation_ticks_per_local_tick)
    : m_value(ticks_t(double(local_ticks) * simulation_ticks_per_local_tick + 0.5))
{}

/*!
 * \brief   Creates an instance of Time from the given ticks_t value.
 */
inline Time
Time::fromTicks(ticks_t v) {
    Time t;

    t.m_value = v;
    return (t);
}

/*!
 * \brief   Copy constructor.
 */
inline
Time::Time(const Time & t):
    m_value(t.m_value) {
}

/*!
 * \brief   Copy assignment.
 */
inline Time &
Time::operator=(const Time & t) {
    m_value = t.m_value;
    return * this;
}

/*!
 * \brief   This function returns the underlying representation of time.
 */
inline ticks_t
Time::ticks() const {
    return m_value;
}

/*!
 * \brief   Comparison operator: equality.
 */
inline bool
Time::operator==(const Time & t) const {
    return (m_value == t.m_value);
}

/*!
 * \brief   Comparison operator: unequality.
 */
inline bool
Time::operator!=(const Time & t) const {
    return (m_value != t.m_value);
}

/*!
 * \brief   Comparison operator: less than.
 */
inline bool
Time::operator<(const Time & t) const {
    return (m_value < t.m_value);
}

/*!
 * \brief   Comparison operator: less then or equal.
 */
inline bool
Time::operator<=(const Time & t) const {
    return (m_value <= t.m_value);
}

/*!
 * \brief   Comparison operator: greater than.
 */
inline bool
Time::operator>(const Time & t) const {
    return (m_value > t.m_value);
}

/*!
 * \brief   Comparison operator: greater than or equal.
 */
inline bool
Time::operator>=(const Time & t) const {
    return (m_value >= t.m_value);
}

/*!
 * \brief   Arithmetic operator: addition.
 */
inline Time &
Time::operator+=(const Time & t) {
    m_value += t.m_value;
    return * this;
}

/*!
 * \brief   Arithmetic operator: substraction.
 */
inline Time &
Time::operator-=(const Time & t) {
    m_value -= t.m_value;
    return * this;
}

/*!
 * \brief   Arithmetic operator: multiplication with a \c double value in seconds.
 */
inline Time &
Time::operator*=(double v) {
    assert(v >= 0);
    m_value = ticks_t(double(m_value) * v + 0.5);
    return * this;
}

/*!
 * \brief   Arithmetic operator: division with a \c double value in seconds.
 */
inline Time &
Time::operator/=(double v) {
    assert(v > 0);
    m_value = ticks_t(double(m_value) / v + 0.5);
    return * this;
}

/* Functions - Globals */

/*!
 * \brief   Arithmetic operator: addition.
 */
inline const Time
operator+(const Time & t1, const Time & t2) {
    Time tmp(t1);

    return tmp += t2;
}

/*!
 * \brief   Arithmetic operator: substraction.
 */
inline const Time
operator-(const Time & t1, const Time & t2) {
    Time tmp(t1);

    return tmp -= t2;
}

/*!
 * \brief   Arithmetic operator: multiplication with a \c double value in seconds.
 */
inline const Time
operator*(const Time & t, double v) {
    Time tmp(t);

    assert(v >= 0);
    return tmp *= v;
}

/*!
 * \brief   Arithmetic operator: multiplication with a \c double value in seconds.
 */
inline const Time
operator*(double v, const Time & t) {
    Time tmp(t);

    assert(v >= 0);
    return tmp *= v;
}

/*!
 * \brief   Arithmetic operator: division with a \c double value in seconds.
 */
inline const Time
operator/(const Time & t, double v) {
    Time tmp(t);

    assert(v > 0);
    return tmp /= v;
}

/*!
 * \brief   Constant ZERO_TIME represents a time value of zero.
 */
extern const Time ZERO_TIME;

/*!
 * \brief   The function getSimulatedTimeResolution() returns the simulated time resolution in seconds.
 *
 * \see     SchedulerInterfaceForComponents
 */
double getSimulatedTimeResolution(SimulationContext *);

/*! SchedulerHelperForComponents:
 * This class contains helper functions related to SchedulerInterfaceForComponents
 */
class TimeHelper{
public:
    /*! Get the unit of simulation time used by the scheduler as string.
    * \returns Simulated time unit as string. "tic" if unknown.
    */
    static std::string getSimulatedTimeResolutionStr(SchedulerInterfaceForComponents * sch);
};

}   /* namespace sg */

#endif  /* SGTIME_H */
