/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   SingleBitOps.h
 *  \brief  Operations on single bit values
 *  \date   Copyright ARM Limited 2009-2014 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 *  There are two forms of operation: ones which operate in place on a reference
 *  and ones which which return a modified value.
 *
 *  Both types are provided with different naming conventions. Those which
 *  operate in place are suffixed "_in".
 */

#ifndef INCLUDED_SG_SINGLEBITOPS_H
#define INCLUDED_SG_SINGLEBITOPS_H

namespace sg {
namespace bitops {

/*! Select and return a bit
 *  \param  data    value to operate on
 *  \param  pos     bit position
 *
 *  \return bit value in place
 */
template <typename D>
inline D bit(D data, unsigned pos)
{
    return (data & (static_cast<D>(1) << pos));
}

/*! Select and return as a boolean the state of a bit
 *  \param  data    value to operate on
 *  \param  pos     bit position
 *
 *  \return bit value as a boolean
 */
template <typename D>
inline bool flag(D data, unsigned pos)
{
    return (bit(data,pos) != 0);
}

/*! Select and return a bit
 * \param   data    value to operate on
 * \param   pos     bit position
 *
 * \return bit value as a 'D' in bit 0
 */
template<typename D>
inline D slice(D data, unsigned pos)
{
    return static_cast<D>((data >> pos) & 1);
}

/*! Assign a bit value in place
  * \param value    true for 1 false for 0
  * \param pos      position to place in
  * \param data     value to be modified
  */
template<typename D>
inline void put_bit_in(bool value, unsigned pos, D& data)
{
    D const mask = static_cast<D>(1) << pos;
    data = ((data & ~mask) | (value ? mask : D(0)));
}

/*! Assign a bit value in place based on a boolean
  * \param value    true for 1 false for 0
  * \param pos      position to place in
  * \param data     original value to operate on
  *
  * \return the modified value
  */
template<typename D>
inline D put_bit(bool value, unsigned pos, D data)
{
    D const mask = static_cast<D>(1) << pos;
    return ((data & ~mask) | (value ? mask : 0));
}

/*! Set a bit value in place based on a position
  * \param data     original value to operate on
  * \param pos      position to set
  */
template <typename D>
inline void set_bit_in(D& data, unsigned pos)
{
    data |= (static_cast<D>(1) << pos);
}

/*! Set a bit value
  * \param data     original value to operate on
  * \param pos      position to set
  *
  * \return the modified value
  */
template <typename D>
inline D set_bit(D data, unsigned pos)
{
    return (data | (static_cast<D>(1) << pos));
}

/*! Clear a bit value in place based on a position
  * \param data     original value to operate on
  * \param pos      position to set
  */
template <typename D>
inline void clear_bit_in(D& data, unsigned pos)
{
    data &= ~(static_cast<D>(1) << pos);
}

/*! Clear a bit value based on a position
  * \param data     original value to operate on
  * \param pos      position to set
  *
  * \return the modified value
  */
template <typename D>
inline D clear_bit(D data, unsigned pos)
{
    return (data & ~(static_cast<D>(1) << pos));
}

}
}

#endif

/* eof SingleBitOps.h */
