/* -*- c++ -*- */
/* vim: syntax=cpp
 */
/*!
 *  \file   StructureOverlay.h
 *  \brief  Overlays a structure on a byte array
 *  \date   Copyright ARM Limited 2014 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 */

#ifndef INCLUDED_STRUCTUREOVERLAY_H
#define INCLUDED_STRUCTUREOVERLAY_H

#include <sg/IncludeMeFirst.h>
#include <inttypes.h>

#include <cassert>

namespace sg {

/*! A template which overlays a structure on a byte array and provides methods
 * to read and write that array. The first template parameter *must* be a packed
 * structure the second parameter is the extent of the underlying data in bytes
 * and may be larger than the structure.
 *
 * Bits are allowed to be tagged as write-to-clear (wtc) and also as read-only
 * (filter).
 *
 * Accesses of different widths can be allowed/forbidden.
 */
template <typename T,size_t EXTENT,size_t ARRAY=1>
class StructureOverlay
{
private:
    void write8Internal(pv::bus_addr_t addr, uint8_t data)
    {
        assert(addr < EXTENT);

        if(byte_is_wtc[addr])
            backing.storage[addr] =  backing.storage[addr] & ~(filter.storage[addr] & data);
        else
            backing.storage[addr] = (backing.storage[addr] & ~(filter.storage[addr])) | (data & filter.storage[addr]);
    }

    uint8_t read8Internal(pv::bus_addr_t addr) const
    {
        assert(addr < EXTENT);

        return backing.storage[addr];
    }

    bool    access8Allowed;
    bool    access16Allowed;
    bool    access32Allowed;
    bool    access64Allowed;

public:
    StructureOverlay() {
        static_assert(sizeof(T)*ARRAY <= EXTENT, "sizeof(T)*ARRAY <= EXTENT");

        for(size_t i = 0 ; i < EXTENT ; ++i)
        {
            backing.storage[i] = 0;
            filter.storage[i] = 0;
            byte_is_wtc[i] = false;
        }

        access8Allowed  = true;
        access16Allowed = true;
        access32Allowed = true;
        access64Allowed = true;
    }
    ~StructureOverlay() {
    }

    /*! Backing storage and masks, bits set in the filter are writeable */
    union { T overlay; T array[ARRAY]; uint8_t storage[EXTENT]; } backing, filter;
    /*! Tags for if the byte in question is Write To Clear true if it is, otherwise false */
    bool byte_is_wtc[EXTENT];


    void setAccess8(bool access)
    {
        access8Allowed = access;
    }

    void setAccess16(bool access)
    {
        access16Allowed = access;
    }

    void setAccess32(bool access)
    {
        access32Allowed = access;
    }

    void setAccess64(bool access)
    {
        access64Allowed = access;
    }

    /// 'addr' is offset in structure
    bool write8(pv::bus_addr_t addr, uint8_t data)
    {
        if(access8Allowed) {
            write8Internal(addr,data);
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool write16(pv::bus_addr_t addr, uint16_t data)
    {
        if(access16Allowed) {
            write8Internal(addr++,uint8_t(data & 0xFF));
            write8Internal(addr,uint8_t((data >> 8) & 0xFF));
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool write32(pv::bus_addr_t addr, uint32_t data)
    {
        if(access32Allowed) {
            write8Internal(addr++,uint8_t(data & 0xFF));
            write8Internal(addr++,uint8_t((data >> 8) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 16) & 0xFF));
            write8Internal(addr,uint8_t((data >> 24) & 0xFF));
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool write64(pv::bus_addr_t addr, uint64_t data)
    {
        if(access64Allowed) {
            write8Internal(addr++,uint8_t(data & 0xFF));
            write8Internal(addr++,uint8_t((data >> 8) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 16) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 24) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 32) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 40) & 0xFF));
            write8Internal(addr++,uint8_t((data >> 48) & 0xFF));
            write8Internal(addr,uint8_t((data >> 56) & 0xFF));
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool read8(pv::bus_addr_t addr, uint8_t* __restrict rv) const
    {
        assert(rv != NULL);

        if(access8Allowed) {
            *rv = read8Internal(addr);
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool read16(pv::bus_addr_t addr,uint16_t* __restrict rv) const
    {
        assert(rv != NULL);

        if(access16Allowed) {
            *rv = read8Internal(addr++);
            *rv |= read8Internal(addr) << 8;
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool read32(pv::bus_addr_t addr,uint32_t* __restrict rv) const
    {
        assert(rv != NULL);

        if(access32Allowed) {
            *rv = read8Internal(addr++);
            *rv |= uint32_t(read8Internal(addr++)) << 8;
            *rv |= uint32_t(read8Internal(addr++)) << 16;
            *rv |= uint32_t(read8Internal(addr)) << 24;
            return true;
        } else {
            return false;
        }
    }

    /// 'addr' is offset in structure
    bool read64(pv::bus_addr_t addr,uint64_t* __restrict rv) const
    {
        assert(rv != NULL);

        if(access64Allowed) {
            *rv = read8Internal(addr++);
            *rv |= uint64_t(read8Internal(addr++)) << 8;
            *rv |= uint64_t(read8Internal(addr++)) << 16;
            *rv |= uint64_t(read8Internal(addr++)) << 24;
            *rv |= uint64_t(read8Internal(addr++)) << 32;
            *rv |= uint64_t(read8Internal(addr++)) << 40;
            *rv |= uint64_t(read8Internal(addr++)) << 48;
            *rv |= uint64_t(read8Internal(addr)) << 56;
            return true;
        } else {
            return false;
        }
    }
};

}

#endif

/* eof StructureOverlay.h */
