/**
 *  \file   SysReg32RegisterNumber.h
 *  \brief  Encodings for the sysreg32 plugin register interface
 *  \date   Copyright ARM Limited 2012 All Rights Reserved.
 */
#ifndef FRAMEWORK_INCLUDE_SG_SYSREG32REGISTERNUMBER_H_6C0327BD
#define FRAMEWORK_INCLUDE_SG_SYSREG32REGISTERNUMBER_H_6C0327BD 1

#include <sg/PluginRegisterPublishingInterface.h>

namespace sg
{
    /**
     * Helper structure to encode / decode register numbers for the sysreg32 interface.
     *
     * Most fields are as used by the AArch32 MCR/MRC or MCRR/MRRC instructions.  \c bank
     * represents the system register bank, as selected by SCR_EL3.NS.
     */
    struct SysReg32RegisterNumber
    {
        unsigned coproc, crn, opc1, crm, opc2, bank;

        /**
         * Register encodings as used by MCR/MRC instructions.
         */
        SysReg32RegisterNumber(
            unsigned coproc_,
            unsigned crn_,
            unsigned opc1_,
            unsigned crm_,
            unsigned opc2_,
            unsigned bank_
        )
            : coproc(coproc_)
            , crn(crn_)
            , opc1(opc1_)
            , crm(crm_)
            , opc2(opc2_)
            , bank(bank_)
        {
        }

        /**
         * Register encodings as used by MCRR/MRRC instructions.
         */
        SysReg32RegisterNumber(
            unsigned coproc_,
            unsigned crm_,
            unsigned opc1_,
            unsigned bank_
        )
            : coproc(coproc_)
            , crn(0)
            , opc1(opc1_)
            , crm(crm_)
            , opc2(0)
            , bank(bank_)
        {
        }

        /**
         * Decode a register number in to split out fields.
         */
        static SysReg32RegisterNumber decode(sg::PluginRegisterPublishingInterface::RegisterNumber register_number)
        {
            return SysReg32RegisterNumber(
                    field(COPROC_BIT, 4, register_number),  // coproc
                    field(   CRN_BIT, 4, register_number),  // crn
                    field(  OPC1_BIT, 4, register_number),  // opc1
                    field(   CRM_BIT, 4, register_number),  // crm
                    field(  OPC2_BIT, 3, register_number),  // opc2
                    field(  BANK_BIT, 1, register_number)   // bank
            );
        }

        /**
         * Encode the split out fields into a register number.
         */
        sg::PluginRegisterPublishingInterface::RegisterNumber encode() const
        {
            return ((coproc & 0xf) << COPROC_BIT)
                 | ((crn    & 0xf) <<    CRN_BIT)
                 | ((opc1   & 0xf) <<   OPC1_BIT)
                 | ((crm    & 0xf) <<    CRM_BIT)
                 | ((opc2   & 0x7) <<   OPC2_BIT)
                 | ((bank   & 0x1) <<   BANK_BIT)
                 ;
        }

    private:
        enum {
            COPROC_BIT =  0,
               CRN_BIT =  4,
              OPC1_BIT =  8,
               CRM_BIT = 12,
              OPC2_BIT = 16,
              BANK_BIT = 20
        };

        static unsigned field(unsigned lsb, unsigned width, unsigned value)
        {
            return (value >> lsb) & ((1 << width) -1);
        }
    }; // class SysReg32RegisterNumber

} // namespace sg

#endif // FRAMEWORK_INCLUDE_SG_SYSREG32REGISTERNUMBER_H_6C0327BD
