/*!
 *  \file   VectorOfPtrs.h
 *  \brief  A vector of pointers container with a re-entrant safe iteration mechanism
 *  \date   Copyright ARM Limited 2013 All Rights Reserved.
 *
 *  This file is not part of a stable public API and it may
 *  change without notice.
 *
 *  Typical use case is maintaining a list of observers
 *
 */

#ifndef INCLUDED_SG_VECTOR_OF_PTRS_H
#define INCLUDED_SG_VECTOR_OF_PTRS_H

#include <cassert>
#include <cstdlib>
#include <vector>


#include "sg/FastVectorOfPOD.h"


namespace sg {


template <typename T, size_t SIZE=4, bool PROFILE=false>
class VectorOfPtrs : protected sg::FastVectorOfPOD<T*,SIZE,PROFILE>
{
private:
    typedef VectorOfPtrs<T,SIZE,PROFILE>          This;
    typedef sg::FastVectorOfPOD<T*,SIZE,PROFILE>  Base;
    typedef T*                                    Ptr;

    static const size_t INDEX_BEFORE_FRONT = size_t(-1);

    unsigned  order_lock_count;

    bool is_index_valid(size_t index) const
    {
        assert(INDEX_BEFORE_FRONT >= Base::size());

        return index < Base::size();
    }

    static bool is_ptr_null(const Ptr& ptr)
    {
        return ptr == 0;
    }

    bool is_element_null(size_t index) const
    {
        assert(is_index_valid(index));
        return is_ptr_null(Base::operator[](index));
    }

    bool is_back_null() const
    {
        return is_ptr_null(back());
    }

    void set_element_null(size_t index)
    {
        assert(is_index_valid(index));

        Base::operator[](index) = 0;

        assert(is_element_null(index));
    }

    //! Compact vector by removing null elements
    void compact()
    {
        assert(!is_order_locked());

        for(size_t i=0; i<Base::size(); ++i)
        {
            if (is_element_null(i))
            {
                assert(!is_back_null());

                Base::operator[](i) = back();
                Base::pop_back();

                assert(!empty() && !is_element_null(i));

                while(is_back_null())
                {
                    Base::pop_back();
                }

                assert(!empty());
            }
        }
    }

    //! Move index to first non-null element or to an invalid value if
    //  there are no more non-null elements
    template <int DIRECTION>
    void skip_null(size_t& index) const
    {
        if (is_order_locked())
        {
            // There may be null elements
            while(is_index_valid(index) && is_element_null(index))
            {
                index += DIRECTION;
            }
        }
#ifndef NDEBUG
        else if (is_index_valid(index))
        {
            assert(!is_element_null(index));
        }
#endif
    }

    //! Get the index of the last element in the vector
    size_t back_index() const
    {
        return Base::size() - 1;
    }

public:
    VectorOfPtrs() : order_lock_count(0) {}

    ~VectorOfPtrs()
    {
        // Re-entrancy makes this assert not possible e.g. an iteration can be in progress
        // but blocked when the simulation is stopped and destructed
        //assert(!is_order_locked());
    }

    // Access to the permitted base class methods
    void        push_back(const Ptr& value)  { Base::push_back(value); assert(!is_back_null()); }
    bool        empty() const                { return Base::empty();   }
    Ptr&        back()                       { return Base::back();    }
    const Ptr&  back() const                 { return Base::back();    }

    Ptr& front()
    {
        size_t index = 0;
        skip_null<+1>(index);
        return Base::operator[](index);
    }

    const Ptr& front() const
    {
        return const_cast<Ptr&>(front());
    }

    void clear()
    {
        Base::clear();
    }

    //! Remove last element and reduce size() by one
    //  Ensures that the last element is never null
    void pop_back()
    {
        Base::pop_back();

        if (is_order_locked())
        {
            while(!empty() && is_back_null())
            {
                Base::pop_back();
            }
        }

        assert(empty() || !is_back_null());
    }

    //! \return true if the order of the elements in the vector is being maintained
    bool is_order_locked() const { return order_lock_count != 0; }

    //! Ensure remove() and remove_assert_unique() maintain order
    //  Top and tail code that requires order to be maintained with lock_order()
    //  and unlock_order()
    void lock_order()
    {
        ++order_lock_count;
    }

    //! Allow remove() and remove_assert_unique() to reorder the vector
    void unlock_order()
    {
        assert(is_order_locked());

        --order_lock_count;

        if (!is_order_locked())
        {
            compact();
        }
    }

    //! Erase a single element as efficiently as possible, O(1)
    //  If the order_lock is on then the order of elements is maintained
    void erase(size_t index)
    {
        assert(is_index_valid(index));

        if (!is_order_locked())
        {
            // Order unimportant, swap with last element and reduce size() by one
            Base::operator[](index) = back();
            Base::pop_back();
        }
        else if (index == (Base::size() - 1))
        {
            pop_back();
        }
        else
        {
            // Temporarily null the element until order lock is removed
            set_element_null(index);
        }
    }

    //! Remove all elements equal to value
    //  WARNING cost is O(n)
    void remove(const Ptr& value)
    {
        size_t index = 0;

        while(is_index_valid(index))
        {
            if (Base::operator[](index) == value)
            {
                erase(index);
            }
            else
            {
                ++index;
            }
        }
    }

    //! Remove the first element found that matches value
    //  WARNING cost is O(n)
    void remove_once(const Ptr& value)
    {
        size_t index = 0;

        while(is_index_valid(index))
        {
            if (Base::operator[](index) == value)
            {
                erase(index);
                break;
            }
            else
            {
                ++index;
            }
        }
    }

    //! Push a unique value onto the back
    //  Functionality is equivalent to std::vector<>::push_back() but with an extra condition
    //  that the value is not already in the vector.
    void push_back_assert_unique(const Ptr& value)
    {
#ifndef NDEBUG
        for(size_t i=0; i<Base::size(); ++i)
        {
            assert(Base::at(i) != value);
        }
#endif
        push_back(value);
        assert(!is_back_null());
    }

    //! Remove a unique element from a vector by value
    //  WARNING cost is O(n)
    void remove_assert_unique(const Ptr& value)
    {
        bool found_ptr_in_vector = remove_if_found_assert_unique(value);
        (void) found_ptr_in_vector;
        assert(found_ptr_in_vector);
    }


    //! Remove a unique element from a vector by value
    // \return true if found, false and no assert if not found
    //  WARNING cost is O(n)
    bool remove_if_found_assert_unique(const Ptr& value)
    {
        for(size_t i=0; i<Base::size(); ++i)
        {
            if (Base::operator[](i) == value)
            {
                erase(i);

#ifndef NDEBUG
                while(++i < Base::size())
                {
                    assert(Base::at(i) != value);
                }
#endif
                return true;
            }
        }

        return false;
    }

    //! Return the total number of elements
    //  WARNING cost is O(n) when the element order is locked and O(1) when unlocked
    //  it is guessed that the element order is mostly unlocked
    size_t size() const
    {
        if (!is_order_locked())
            return Base::size();

        size_t total = Base::size();

        for(size_t i=0; i<Base::size(); ++i)
        {
            if (is_element_null(i))
            {
                --total;
            }
        }

        return total;
    }

    //! Transfer elements from that to back of this
    void splice_back(This& that)
    {
        Queue q(that);

        while(!q.empty())
        {
            push_back(q.front());
            q.pop();
        }

        that.clear();
    }

    using Base::find;


    //! A class with queue like semantics used to (re-entrant) safely iterate over a
    //  VectorOfPtrs. It is safe to call clear(), remove() or remove_assert_unique()
    //  on any element during an iteration using DirectionQueue
    template <int DIR>
    class DirectionQueue
    {
    private:
        This&    vector;
        size_t   index;

    public:
        DirectionQueue(This& vector_)
            : vector(vector_)
        {
            if (DIR == +1)
            {
                index = 0;
                vector.skip_null<DIR>(index);
            }
            else
            {
                index = vector_.back_index();
                assert(!vector.is_index_valid(index) || !vector.is_element_null(index)); // If vector is not empty, its last element must not be null.
            }
            vector.lock_order();
        }

        ~DirectionQueue()
        {
            vector.unlock_order();
        }

        //! Returns true if iteration over all elements of the VectorOfPtrs<T> is complete
        bool empty() const
        {
            assert(vector.is_order_locked());

            return !vector.is_index_valid(index);
        }

        //! Returns value of current iteration element
        Ptr front()
        {
            assert(vector.is_order_locked());
            assert(!empty());

            return vector[index];
        }

        //! Removes the current iteration element and moves front() to the next one
        void pop()
        {
            assert(vector.is_order_locked());

            if (!empty())
            {
                index += DIR;
                vector.skip_null<DIR>(index);
            }
        }
    };

    //! For forward iteration from first index to last
    typedef DirectionQueue<+1>  Queue;

    //! For reverse iteration from last index to first
    typedef DirectionQueue<-1>  ReverseQueue;
};


} // namespace sg

#endif
