//
// (C) COPYRIGHT 2017 ARM Limited. ALL RIGHTS RESERVED.
//

#ifndef ARM_MODELS_SCAXI4_H
#define ARM_MODELS_SCAXI4_H

/*
 * SystemC adapters from SLAXI4 AXI4Master/Slave.
 */

#include <ARM/TLM/arm_axi4.h>

#include "SLAXI4.h"

namespace ARM
{
namespace Models
{
namespace SCAXI4
{

/* Process lines of an exception's what() message into SC_REPORT calls. */
inline void exception_to_sc_report(const char* msg_type, std::exception& except)
{
    const char* message = except.what();
    const char* nl = std::strchr(message, '\n');

    while (nl)
    {
        if (nl != message)
        {
            std::string line = std::string(message, nl - message);
            SC_REPORT_INFO(msg_type, line.c_str());
        }
        message = nl + 1;
        nl = std::strchr(message, '\n');
    }
    /* If the message is empty or ends in \n, give a generic fatal message. */
    if (*message == '\0')
        SC_REPORT_FATAL(msg_type, "Exiting");
    else
        SC_REPORT_FATAL(msg_type, message);
}

/* Catch std::exception and convert to SystemC SC_REPORT_FATAL. */
#define ARM_CATCH_TO_SC_REPORT_FATAL(MODEL) \
    catch (std::exception& except) \
    { \
        static const char* msg_type = "/ARM/Models/" MODEL; \
        exception_to_sc_report(msg_type, except); \
        throw; /* This makes function return checks happy. */ \
    }

class ErrorHandler : public SignalLevel::ErrorHandler
{
public:
    void error(const std::string& message, const std::string& context = "")
    {
        if (!context.empty())
        {
            std::string context_message = message + ": " + context;
            SC_REPORT_FATAL("/ARM/Models", context_message.c_str());
        } else
        {
            SC_REPORT_FATAL("/ARM/Models", message.c_str());
        }
    }
};

template <typename ProtocolType>
class TLMTieOffInterface :
    public tlm::tlm_fw_transport_if<ProtocolType>,
    public tlm::tlm_bw_transport_if<ProtocolType>
{
public:
    typedef typename ProtocolType::tlm_payload_type Payload;
    typedef typename ProtocolType::tlm_phase_type Phase;

    static void tie_off_error(const char* function)
    {
        SC_REPORT_FATAL("TLMTieOffInterface", ("Function: " + std::string(function)
            + " called on tied-off TLM socket").c_str());
    }

    static void zero_time_exception(const std::string& function)
    {
        SC_REPORT_WARNING("TLMTieOffInterface", (function +
            "with time != 0 is not supported by native ports").c_str());
    }

    tlm::tlm_sync_enum nb_transport_fw(Payload& trans, Phase& phase,
        sc_core::sc_time& delay)
    { tie_off_error("nb_transport_fw"); return tlm::TLM_ACCEPTED; }

    tlm::tlm_sync_enum nb_transport_bw(Payload& trans, Phase& phase,
        sc_core::sc_time& delay)
    { tie_off_error("nb_transport_bw"); return tlm::TLM_ACCEPTED; }

    void b_transport(Payload& trans, sc_core::sc_time& delay)
    { tie_off_error("b_transport"); }

    bool get_direct_mem_ptr(Payload& trans, tlm::tlm_dmi& dmi_data)
    { tie_off_error("get_direct_mem_ptr"); return false; }

    void invalidate_direct_mem_ptr(sc_dt::uint64 start_range, sc_dt::uint64 end_range)
    { tie_off_error("invalidate_direct_mem_ptr"); }

    unsigned int transport_dbg(Payload& trans)
    { tie_off_error("transport_dbg"); return 0; }
};

/* Adapter from a TLM master parent to an SLAXI4 master child. */
class AXI4Master :
    public SignalLevel::AXI4Master::AXI4FwInterface,
    public TLMTieOffInterface<AXI4::ProtocolType>
{
public:
    typedef TLM::BaseMasterSocket<AXI4::ProtocolType> SCPortType;

protected:
    typedef TLMTieOffInterface<AXI4::ProtocolType> TieOffInterface;

    typedef AXI4::ProtocolType::tlm_payload_type Payload;
    typedef AXI4::ProtocolType::tlm_phase_type Phase;

    ErrorHandler error_handler;

    TLM::BaseMasterSocket<AXI4::ProtocolType>* parent;
    SignalLevel::AXI4Master child;

    /* TLM bw. */
    tlm::tlm_sync_enum nb_transport_bw(Payload& trans, Phase& phase,
        sc_core::sc_time& delay) try
    {
        if (delay != sc_core::SC_ZERO_TIME)
            TieOffInterface::zero_time_exception("nb_transport_bw");

        if (child.bw(trans, phase))
            return tlm::TLM_UPDATED;
        else
            return tlm::TLM_ACCEPTED;
    }
    ARM_CATCH_TO_SC_REPORT_FATAL("AXI4Master(nb_transport_bw)")

    /* AXI4Master::AXI4FwInterface. */
    bool fw(Payload& payload, Phase& phase) try
    {
        sc_core::sc_time delay = sc_core::SC_ZERO_TIME;

        tlm::tlm_sync_enum result = (*parent)->nb_transport_fw(payload, phase, delay);

        if (delay != sc_core::SC_ZERO_TIME)
            TieOffInterface::zero_time_exception("nb_transport_fw");

        return result != tlm::TLM_ACCEPTED;
    }
    ARM_CATCH_TO_SC_REPORT_FATAL("AXI4Master(fw)")

public:
    AXI4Master(const std::string& name, TLM::Protocol protocol) :
        parent(NULL),
        child(name, protocol)
    { child.set_fw_interface(*this); }

    /* Forwarded calls to SignalLevel::AXI4Master. */
    void set_signal_interface(SignalLevel::AXI4MasterInterface& signal_interface)
    { child.set_signal_interface(signal_interface); }
    void posedge() { child.posedge(); }
    void negedge() { child.negedge(); }
    
    /* Bind onto a SystemC master socket. */
    void bind_to_systemc_parent(TLM::BaseMasterSocket<AXI4::ProtocolType>& parent_)
    {
        parent = &parent_;
        parent->bind(*this);
    }
};

/* Adapter from a TLM slave parent to an SLAXI4 slave child. */
class AXI4Slave :
    public SignalLevel::AXI4Slave::AXI4BwInterface,
    public TLMTieOffInterface<AXI4::ProtocolType>
{
public:
    typedef TLM::BaseSlaveSocket<AXI4::ProtocolType> SCPortType;

protected:
    typedef TLMTieOffInterface<AXI4::ProtocolType> TieOffInterface;

    typedef AXI4::ProtocolType::tlm_payload_type Payload;
    typedef AXI4::ProtocolType::tlm_phase_type Phase;

    ErrorHandler error_handler;

    TLM::BaseSlaveSocket<AXI4::ProtocolType>* parent;
    SignalLevel::AXI4Slave child;

    /* TLM fw. */
    tlm::tlm_sync_enum nb_transport_fw(Payload& trans, Phase& phase,
        sc_core::sc_time& delay) try
    {
        if (delay != sc_core::SC_ZERO_TIME)
            TieOffInterface::zero_time_exception("nb_transport_fw");

        if (child.fw(trans, phase))
            return tlm::TLM_UPDATED;
        else
            return tlm::TLM_ACCEPTED;
    }
    ARM_CATCH_TO_SC_REPORT_FATAL("AXI4Slave(nb_transport_fw)")

    bool bw(Payload& payload, Phase& phase) try
    {
        sc_core::sc_time delay = sc_core::SC_ZERO_TIME;

        tlm::tlm_sync_enum result = (*parent)->nb_transport_bw(payload, phase, delay);

        if (delay != sc_core::SC_ZERO_TIME)
            TieOffInterface::zero_time_exception("nb_transport_bw");

        return result != tlm::TLM_ACCEPTED;
    }
    ARM_CATCH_TO_SC_REPORT_FATAL("AXI4Slave(bw)")

public:
    AXI4Slave(const std::string& name, TLM::Protocol protocol) :
        parent(NULL),
        child(name, protocol)
    { child.set_bw_interface(*this); }

    /* Forwarded calls to SignalLevel::AXI4Slave. */
    void set_signal_interface(SignalLevel::AXI4SlaveInterface& signal_interface)
    {
      child.set_signal_interface(signal_interface);
    }
    void posedge() { child.posedge(); }
    void negedge() { child.negedge(); }
    
    /* Bind onto a SystemC slave socket. */
    void bind_to_systemc_parent(TLM::BaseSlaveSocket<AXI4::ProtocolType>& parent_)
    {
        parent = &parent_;
        parent->bind(*this);
    }
};

}
}
}

#endif // ARM_MODELS_SCAXI4_H
