//
// (C) COPYRIGHT 2017 ARM Limited. ALL RIGHTS RESERVED.
//

#include <cassert>
#include <sstream>
#include <algorithm>

#include "SLAXI4.h"

#define MAX_DATA_WIDTH_IN_BYTES 128
#define MAX_DATA_STROBE_WIDTH_IN_BYTES MAX_DATA_WIDTH_IN_BYTES/8

namespace ARM
{
namespace Models
{
namespace SignalLevel
{

using namespace ARM::AXI4;
using namespace AXI4Signal;

namespace
{

std::ostream& operator <<(std::ostream& stream, Phase phase)
{
    switch (phase)
    {
    case PHASE_UNINITIALIZED: stream << "PHASE_UNINITIALIZED"; break;
    case AW_VALID: stream << "AW_VALID"; break;
    case AW_READY: stream << "AW_READY"; break;
    case W_VALID: stream << "W_VALID"; break;
    case W_VALID_LAST: stream << "W_VALID_LAST"; break;
    case W_READY: stream << "W_READY"; break;
    case B_VALID: stream << "B_VALID"; break;
    case B_READY: stream << "B_READY"; break;
    case AR_VALID: stream << "AR_VALID"; break;
    case AR_READY: stream << "AR_READY"; break;
    case R_VALID: stream << "R_VALID"; break;
    case R_VALID_LAST: stream << "R_VALID_LAST"; break;
    case R_READY: stream << "R_READY"; break;
    case AC_VALID: stream << "AC_VALID"; break;
    case AC_READY: stream << "AC_READY"; break;
    case CR_VALID: stream << "CR_VALID"; break;
    case CR_READY: stream << "CR_READY"; break;
    case CD_VALID: stream << "CD_VALID"; break;
    case CD_VALID_LAST: stream << "CD_VALID_LAST"; break;
    case CD_READY: stream << "CD_READY"; break;
    case WACK: stream << "WACK"; break;
    case RACK: stream << "RACK"; break;
    default: stream << "Phase(" << static_cast<unsigned int>(phase) << ')'; break;
    }

    return stream;
}

std::ostream& operator <<(std::ostream& stream, Size size)
{
    switch (size)
    {
    case SIZE_1: stream << "SIZE_1"; break;
    case SIZE_2: stream << "SIZE_2"; break;
    case SIZE_4: stream << "SIZE_4"; break;
    case SIZE_8: stream << "SIZE_8"; break;
    case SIZE_16: stream << "SIZE_16"; break;
    case SIZE_32: stream << "SIZE_32"; break;
    case SIZE_64: stream << "SIZE_64"; break;
    case SIZE_128: stream << "SIZE_128"; break;
    default: stream << "Size(" << static_cast<unsigned int>(size) << ')'; break;
    }

    return stream;
}

/* Render some value into a string with operator << */
template <typename Type>
std::string to_string(const Type& value)
{
    std::ostringstream ret;

    ret << value;

    return ret.str();
}

/*
 * Render some value into a string with operator << after first setting the stream flags
 * to the given flags.
 */
template <typename Type>
std::string to_string(const Type& value, std::ios_base::fmtflags flags)
{
    std::ostringstream ret;

    ret.flags(flags);
    ret << value;

    return ret.str();
}

/* ceil(a / b). */
inline uint64_t ceil_div(uint64_t a, uint64_t b) { return (a + (b - 1)) / b; }

/* ceil(log_2 a). Returns 0 on arguments of 0. */
inline unsigned int ceil_log2(uint64_t a)
{
    if (a == 0)
        return 0;

    unsigned int ret = 0;

    while (a != 0)
    {
        ret++;
        a >>= 1;
    }

    return ret - 1;
}

/*
 * Data width in bits translated to Size enum.
 * If round_up is true then the function will assert if data_width is
 * not a power of 2 of a multiple of 8. If not, the Size returned will be for the input
 * rounded up to the nearest power of 2 bytes.
 */
Size data_width_to_size(unsigned int data_width, bool round_up = false)
{
    assert(round_up || (((data_width - 1) & data_width) == 0 && (data_width % 8 == 0)));

    unsigned int log_width = ceil_log2(ceil_div(data_width, 8));
    assert(log_width <= SIZE_128);
    return Size(log_width);
}

/* Convert Size value to the number of bytes carried per beat. */
inline unsigned int size_to_bytes(Size size) { return 1 << size; }

/* Size enum translated to data width in bits. */
inline unsigned int size_to_data_width(Size size) { return size_to_bytes(size) * 8; }

/* A default error handler using its own exception class. */
class DefaultErrorHandler : public ErrorHandler
{
protected:
    class Exception : public std::exception
    {
    const std::string message;

    public:
        Exception(const std::string& message_) throw() : message(message_) { }
        ~Exception() throw () {}

        const char* what() const throw() { return message.c_str(); }
    };

public:
    void error(const std::string& message, const std::string& context = "")
    {
        if (context.empty())
            throw Exception(message);
        else
            throw Exception(context + ": " + message);
    }
};

DefaultErrorHandler default_error_handler;

}

AXI4TransactionMapper::AXI4TransactionMapper(unsigned int max_id_, ErrorHandler* error_handler_) :
    error_handler((error_handler_ ? error_handler_ : &default_error_handler)),
    current_id(1),
    max_id(max_id_)
{}

unsigned int AXI4TransactionMapper::add_payload(AXI4::Payload* payload)
{
    while (id_to_payload.find(current_id) != id_to_payload.end())
    {
        current_id++;

        if (current_id > max_id)
        {
            if (id_to_payload.size() >= max_id)
            {
                error_handler->error("More than " + to_string(max_id) +
                    "outstanding transactions held by the signal level transaction mapper");
            }

            current_id = 1;
        }
    }

    id_to_payload[current_id] = payload;
    payload_to_id[payload] = current_id;

    return current_id;
}

void AXI4TransactionMapper::remove_payload(AXI4::Payload* payload)
{
    unsigned int id = payload_to_id[payload];

    payload_to_id.erase(payload);
    id_to_payload.erase(id);
}

AXI4::Payload* AXI4TransactionMapper::lookup_payload(unsigned int id)
{
    if (id_to_payload.find(id) != id_to_payload.end())
        return id_to_payload[id];

    return NULL;
}

const char* const AXI4NameMapping::uppercase_names[] =
{
    "AWID", "AWADDR", "AWLEN", "AWSIZE", "AWBURST", "AWLOCK", "AWCACHE", "AWPROT",
        "AWQOS", "AWREGION", "AWUSER", "AWVALID", "AWREADY",
    "WID", "WDATA", "WSTRB", "WLAST", "WUSER", "WVALID", "WREADY",
    "BID", "BRESP", "BUSER", "BVALID", "BREADY",
    "ARID", "ARADDR", "ARLEN", "ARSIZE", "ARBURST", "ARLOCK", "ARCACHE", "ARPROT",
        "ARQOS", "ARREGION", "ARUSER", "ARVALID", "ARREADY",
    "RID", "RDATA", "RRESP", "RLAST", "RUSER", "RVALID", "RREADY",
    "ARSNOOP", "ARDOMAIN", "ARBAR",
    "AWSNOOP", "AWDOMAIN", "AWBAR", "AWUNIQUE",
    "ACVALID", "ACREADY", "ACADDR", "ACSNOOP", "ACPROT",
    "CRVALID", "CRREADY", "CRRESP",
    "CDVALID", "CDREADY", "CDDATA", "CDLAST",
    "RACK",
    "WACK"
};

const char* const AXI4NameMapping::lowercase_names[] =
{
    "awid", "awaddr", "awlen", "awsize", "awburst", "awlock", "awcache", "awprot",
        "awqos", "awregion", "awuser", "awvalid", "awready",
    "wid", "wdata", "wstrb", "wlast", "wuser", "wvalid", "wready",
    "bid", "bresp", "buser", "bvalid", "bready",
    "arid", "araddr", "arlen", "arsize", "arburst", "arlock", "arcache", "arprot",
        "arqos", "arregion", "aruser", "arvalid", "arready",
    "rid", "rdata", "rresp", "rlast", "ruser", "rvalid", "rready",
    "arsnoop", "ardomain", "arbar",
    "awsnoop", "awdomain", "awbar", "awunique",
    "acvalid", "acready", "acaddr", "acsnoop", "acprot",
    "crvalid", "crready", "crresp",
    "cdvalid", "cdready", "cddata", "cdlast",
    "rack",
    "wack"
};

AXI4InfixNameMapping::AXI4InfixNameMapping(const std::string& prefix_,
    const std::string& suffix_,
    bool is_uppercase_) :
    prefix(prefix_),
    suffix(suffix_),
    is_uppercase(is_uppercase_)
{}

std::string AXI4InfixNameMapping::map(AXI4Signal::AXI4SignalId id)
{
    return prefix + (is_uppercase ? uppercase_names : lowercase_names)[id] + suffix;
}

std::ostream& operator <<(std::ostream& stream, AXI4SignalId id)
{
    stream << AXI4NameMapping::uppercase_names[id];
    return stream;
}

AXI4NameMapping AXI4NameMapping::default_mapping;

AXI4PortParams::AXI4PortParams(ErrorHandler* error_handler_) :
    error_handler((error_handler_ ? error_handler_ : &default_error_handler)),
    /*
     * Break a cyclic dependency between cache_line_width and data_width and between address
     * widths.
     */
    dvm_addr_width(0),
    cache_line_width(512)
{
    set_protocol(TLM::PROTOCOL_AXI4);
    set_addr_width(64);
    set_data_width(128);
    set_id_width(8);
    set_user_width(64);
    set_qos_width(0);
    set_region_width(0);
}

void AXI4PortParams::set_error_handler(ErrorHandler* error_handler_)
{ error_handler = (error_handler_ ? error_handler_ : &default_error_handler); }

/*
 * It is only critical that this width is correct to the nearest byte so as to prevent buffer
 * over-run.
 */
unsigned int AXI4PortParams::expected_signal_width(AXI4SignalId id)
{
    switch (id)
    {
    case AWVALID:
    case AWREADY:
    case WVALID:
    case WREADY:
    case WLAST:
    case BVALID:
    case BREADY:
    case ARVALID:
    case ARREADY:
    case RVALID:
    case RREADY:
    case RLAST:
    case AWUNIQUE:
    case ACVALID:
    case ACREADY:
    case CRVALID:
    case CRREADY:
    case CDVALID:
    case CDREADY:
    case CDLAST:
    case SIGNAL_RACK:
    case SIGNAL_WACK:
        return 1;
    case AWID:
        return wid_width;
    case WID:
        return wid_width;
    case BID:
        return wid_width;
    case ARID:
        return rid_width;
    case RID:
        return rid_width;
    case AWADDR:
        return awaddr_width;
    case ARADDR:
        return araddr_width;
    case AWLEN:
    case ARLEN:
        return 8;
    case AWSIZE:
    case ARSIZE:
        return 3;
    case AWBURST:
    case ARBURST:
        return 2;
    case AWLOCK:
    case ARLOCK:
        /* 2 for AXI3. */
        return 1;
    case AWCACHE:
    case ARCACHE:
        return 4;
    case AWPROT:
    case ARPROT:
    case ACPROT:
        return 3;
    case AWQOS:
    case ARQOS:
        return qos_width;
    case AWREGION:
    case ARREGION:
        return region_width;
    case AWUSER:
        return awuser_width;
    case WUSER:
        return wuser_width;
    case BUSER:
        return buser_width;
    case ARUSER:
        return aruser_width;
    case RUSER:
        return ruser_width;
    case WDATA:
    case RDATA:
    case CDDATA:
        return data_width;
    case WSTRB:
        return data_width_in_bytes;
    case BRESP:
        return 2;
    case RRESP:
        return (protocol == TLM::PROTOCOL_ACE ? 4 : 2);
    case ARSNOOP:
    case ACSNOOP:
        return 4;
    case AWSNOOP:
        return 3;
    case ARDOMAIN:
    case AWDOMAIN:
        return 2;
    case ARBAR:
    case AWBAR:
        return 2;
    case ACADDR:
        return acaddr_width;
    case CRRESP:
        return 5;
    default:
        error_handler->error("AXI4PortParams::expected_signal_width: Unrecognised AXI4SignalId: "
            + to_string(id));
        return 0;
    }
}

uint8_t AXI4PortParams::signal_default_value(AXI4Signal::AXI4SignalId id)
{
    switch (id)
    {
    case WLAST:
    case RLAST:
    case CDLAST:
        return 1;
    default:
        return 0;
    }
}

uint64_t AXI4PortParams::crop_address(uint64_t addr, unsigned int addr_width)
{
    return (addr_width == 64 ? addr : ((1ULL << addr_width) - 1) & addr);
}

void AXI4PortParams::set_addr_width(unsigned int addr_width_)
{
    set_awaddr_width(addr_width_);
    set_araddr_width(addr_width_);
    set_acaddr_width(addr_width_);
}

void AXI4PortParams::set_awaddr_width(unsigned int awaddr_width_)
{
    if (awaddr_width_ > 64)
    {
        error_handler->error(
            "AXI4PortParams::set_awaddr_width: address width must be <= 64, given: " + to_string(awaddr_width_));
    }
    awaddr_width = awaddr_width_;
}

void AXI4PortParams::set_araddr_width(unsigned int araddr_width_)
{
    if (araddr_width_ > 64)
    {
        error_handler->error(
            "AXI4PortParams::set_araddr_width: address width must be <= 64, given: " + to_string(araddr_width_));
    }
    araddr_width = std::max(araddr_width_, dvm_addr_width);
}

void AXI4PortParams::set_acaddr_width(unsigned int acaddr_width_)
{
    if (acaddr_width_ > 64)
    {
        error_handler->error(
            "AXI4PortParams::set_acaddr_width: address width must be <= 64, given: " + to_string(acaddr_width_));
    }
    acaddr_width = std::max(acaddr_width_, dvm_addr_width);
}

void AXI4PortParams::set_dvm_addr_width(unsigned int dvm_addr_width_)
{
    dvm_addr_width = dvm_addr_width_;
    /* Widen {AR,AC}ADDR width if necessary. */
    set_araddr_width(araddr_width);
    set_acaddr_width(acaddr_width);
}

void AXI4PortParams::set_protocol(TLM::Protocol protocol_)
{
    protocol = protocol_;
}

void AXI4PortParams::set_data_width(unsigned int data_width_)
{
    if (((data_width_ - 1) & data_width_) != 0)
    {
        error_handler->error(
            "AXI4PortParams::set_data_width: data width must be a non-0 power of two, given: "
            + to_string(data_width_));
    }
    data_width = data_width_;
    data_size = data_width_to_size(data_width);
    data_width_in_bytes = (data_width + 7) / 8;
    data_strobe_width_in_bytes = (data_width_in_bytes + 7) / 8;

    /* Update cache line beat count. */
    set_cache_line_width(cache_line_width);
}

void AXI4PortParams::set_user_width(unsigned int user_width_)
{
    set_awuser_width(user_width_);
    set_wuser_width(user_width_);
    set_buser_width(user_width_);
    set_aruser_width(user_width_);
    set_ruser_width(user_width_);
}

void AXI4PortParams::set_awuser_width(unsigned int awuser_width_)
{
    awuser_width = awuser_width_;
}

void AXI4PortParams::set_wuser_width(unsigned int wuser_width_)
{
    wuser_width = wuser_width_;
}

void AXI4PortParams::set_buser_width(unsigned int buser_width_)
{
    buser_width = buser_width_;
}

void AXI4PortParams::set_aruser_width(unsigned int aruser_width_)
{
    aruser_width = aruser_width_;
}

void AXI4PortParams::set_ruser_width(unsigned int ruser_width_)
{
    ruser_width = ruser_width_;
}

void AXI4PortParams::set_wid_width(unsigned int wid_width_)
{
    wid_width = wid_width_;
}

void AXI4PortParams::set_rid_width(unsigned int rid_width_)
{
    rid_width = rid_width_;
}

void AXI4PortParams::set_id_width(unsigned int id_width_)
{
    set_rid_width(id_width_);
    set_wid_width(id_width_);
}

void AXI4PortParams::set_qos_width(unsigned int qos_width_)
{
    qos_width = qos_width_;
}

void AXI4PortParams::set_region_width(unsigned int region_width_)
{
    region_width = region_width_;
}

void AXI4PortParams::set_cache_line_width(unsigned int cache_line_width_)
{
    if (((cache_line_width_ - 1) & cache_line_width_) != 0)
    {
        error_handler->error(
            "AXI4PortParams::set_cache_line_width: cache line width must be a non-0 power of two, given: "
            + to_string(cache_line_width_));
    }

    cache_line_width = cache_line_width_;
    cache_line_beat_count = cache_line_width / data_width;

    if ((cache_line_beat_count * data_width) != cache_line_width)
    {
        error_handler->error(
            "AXI4PortParams::set_cache_line_width: cache line width: " + to_string(cache_line_width_)
            + " must be a multiple of data width: " + to_string(data_width));
    }
}

void AXI4SlaveInterface::reset()
{
    set_awvalid(false);
    set_wvalid(false);
    set_bready(false);
    set_arvalid(false);
    set_rready(false);
    set_acready(false);
    set_crvalid(false);
    set_cdvalid(false);

    /* FIXME, drive all other signals including ones which need payloads. */
}

AXI4PortSignalAccesses::AXI4PortSignalAccesses(ErrorHandler* error_handler_) :
    error_handler((error_handler_ ? error_handler_ : &default_error_handler)),
    params(error_handler)
{
    for (unsigned int signal = 0; signal < NUM_AXI_SIGNALS; signal++)
        signals[signal] = NULL;
}

void AXI4PortSignalAccesses::set_error_handler(ErrorHandler* error_handler_)
{
    error_handler = (error_handler_ ? error_handler_ : &default_error_handler);
    params.set_error_handler(error_handler);
}

AXI4PortSignalAccesses::~AXI4PortSignalAccesses()
{
    std::list<SignalAccess*>::iterator it = tie_offs.begin();

    while (it != tie_offs.end())
    {
        delete *it;
        ++it;
    }
}

void AXI4PortSignalAccesses::bind_to_signals(SignalAccesses& new_signals, AXI4NameMapping& mapping,
    std::ostream* debug_stream)
{
    for (unsigned int signal = 0; signal < NUM_AXI_SIGNALS; signal++)
    {
        if (!signals[signal])
        {
            std::string name = mapping.map(AXI4SignalId(signal));
            SignalAccess* new_signal = new_signals.find_signal(name);
            if (debug_stream)
            {
                *debug_stream << "For signal " << AXI4SignalId(signal) << ", trying name: " << name
                    << (new_signal ? " matched" : " no match") << '\n';
            }
            signals[signal] = new_signal;
        }
    }
}

void AXI4PortSignalAccesses::print_binding_report(std::ostream& stream)
{
    for (unsigned int i = 0; i < NUM_AXI_SIGNALS; i++)
    {
        AXI4SignalId signal_id = AXI4SignalId(i);
        SignalAccess* signal = signals[signal_id];

        if (!signal)
        {
            stream << "Unbound signal: " << signal_id << '\n';
        } else
        {
            unsigned int expected_width = params.expected_signal_width(signal_id);
            unsigned int actual_width = signal->get_width();

            if (expected_width != actual_width)
            {
                stream << "Width mismatch in signal: " << signal_id <<
                    ", expected: " << expected_width << " got: " << actual_width << '\n';
            }
        }
    }
}

namespace
{

/* A tie off wide enough to satisfy up to 'width' bits of 0s. */
class TieOffSignalAccess : public SignalAccess
{
protected:
    const uint8_t byte;

public:
    TieOffSignalAccess(unsigned int width_, uint8_t byte_) :
        SignalAccess(width_),
        byte(byte_)
    {}

    void set_ptr(uint8_t* value) {}
    void get_ptr(uint8_t* value) const { std::memset(value, byte, bits_to_bytes(get_width())); }
};

}

void AXI4PortSignalAccesses::tie_off_signals(std::ostream* debug_stream)
{
    for (unsigned int i = 0; i < NUM_AXI_SIGNALS; i++)
    {
        AXI4SignalId signal_id = AXI4SignalId(i);
        SignalAccess* signal = signals[signal_id];

        if (!signal)
        {
            SignalAccess* tie_off = new TieOffSignalAccess(
                params.expected_signal_width(signal_id),
                params.signal_default_value(signal_id));
            signals[signal_id] = tie_off;
            tie_offs.push_back(tie_off);

            if (debug_stream)
            {
                *debug_stream << "Tying off signal " << signal_id <<
                    " with value: " << int(params.signal_default_value(signal_id)) << '\n';
            }
        }
    }
}

void AXI4PortSignalAccesses::check_size(AXI4::Size size)
{
    if (size_to_data_width(size) > params.get_data_width())
    {
        unsigned int data_width = params.get_data_width();

        error_handler->error("Data SIZE: " + to_string(size) + " (" +
            to_string(size_to_bytes(size)) + " bytes, " + to_string(size_to_data_width(size)) +
            " bits) is too wide for data channel of width: " + to_string(data_width) + " bits");
    }
}

void AXI4PortSignalAccesses::check_value_fits_in_width(uint64_t value, unsigned int width,
    const std::string& name)
{
    if (value != (width >= 64 ? value : (value & ((1ULL << width) - 1))))
    {
        error_handler->error("Payload " + name + ": 0x" + to_string(value, std::ios_base::hex) +
            " too wide for width: " + to_string(width) + " bits");
    }
}

AXI4SignalAccessSlaveAdapter::AXI4SignalAccessSlaveAdapter(ErrorHandler* error_handler_) :
    AXI4PortSignalAccesses(error_handler_),
    transaction_mapper(NULL)
{}

void AXI4SignalAccessSlaveAdapter::set_transaction_mapper(AXI4TransactionMapper& transaction_mapper_)
{ transaction_mapper = &transaction_mapper_; }

void AXI4SignalAccessSlaveAdapter::set_error_handler(ErrorHandler* error_handler_)
{ AXI4PortSignalAccesses::set_error_handler(error_handler_); }

void AXI4SignalAccessSlaveAdapter::set_aw(Payload& payload)
{
    uint64_t addr = params.crop_address(payload.get_address(), params.get_awaddr_width());

    check_value_fits_in_width(payload.id, params.get_wid_width(), "AWID");
    check_size(payload.get_size());

    uint64_t user = (transaction_mapper ? transaction_mapper->add_payload(&payload) : payload.user);

    check_value_fits_in_width(user, params.get_awuser_width(), "AWUSER");
    signals[AWUSER]->set(user);

    signals[AWID]->set(payload.id);
    signals[AWADDR]->set(addr);
    signals[AWLEN]->set(payload.get_len());
    signals[AWSIZE]->set(payload.get_size());
    signals[AWBURST]->set(payload.get_burst());
    signals[AWLOCK]->set(payload.lock);
    signals[AWCACHE]->set(payload.cache);
    signals[AWPROT]->set(payload.prot);
    signals[AWQOS]->set(payload.qos);
    signals[AWREGION]->set(payload.region);
    signals[AWSNOOP]->set(payload.snoop);
    signals[AWDOMAIN]->set(payload.domain);
    signals[AWBAR]->set(payload.bar);
    signals[AWUNIQUE]->set(payload.unique);
}

void AXI4SignalAccessSlaveAdapter::set_awvalid(bool valid) { signals[AWVALID]->set(valid); }
bool AXI4SignalAccessSlaveAdapter::get_awready() { return signals[AWREADY]->get8(); }

void AXI4SignalAccessSlaveAdapter::set_w(Payload& payload, unsigned int beat_index)
{
    check_size(payload.get_size());

    Size data_size = params.get_data_size();
    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];
    uint8_t strobe[MAX_DATA_STROBE_WIDTH_IN_BYTES];
    /* Zero all bytes to ensure data and strobe don't care byte lanes always show the same value. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);
    std::memset(strobe, 0, MAX_DATA_STROBE_WIDTH_IN_BYTES);

    payload.write_out_beat_raw(data_size, beat_index, data);
    payload.write_out_beat_raw_strobe(data_size, beat_index, strobe);

    signals[WID]->set(payload.id);
    signals[WDATA]->set_ptr(data);
    signals[WSTRB]->set_ptr(strobe);
    signals[WLAST]->set(beat_index == payload.get_len());
    signals[WUSER]->set(payload.user);
}

void AXI4SignalAccessSlaveAdapter::set_wvalid(bool valid) { signals[WVALID]->set(valid); }
bool AXI4SignalAccessSlaveAdapter::get_wready() { return signals[WREADY]->get8(); }

void AXI4SignalAccessSlaveAdapter::get_b(Payload& payload)
{
    if (transaction_mapper)
        transaction_mapper->remove_payload(&payload);

    payload.set_resp(signals[BRESP]->get8());
}

uint32_t AXI4SignalAccessSlaveAdapter::get_bid() { return signals[BID]->get32(); }

bool AXI4SignalAccessSlaveAdapter::get_bvalid() { return signals[BVALID]->get8(); }
void AXI4SignalAccessSlaveAdapter::set_bready(bool ready) { signals[BREADY]->set(ready); }

void AXI4SignalAccessSlaveAdapter::set_ar(Payload& payload)
{
    uint64_t addr = params.crop_address(payload.get_address(), params.get_araddr_width());

    check_value_fits_in_width(payload.id, params.get_rid_width(), "ARID");
    check_size(payload.get_size());

    uint64_t user = (transaction_mapper ? transaction_mapper->add_payload(&payload) : payload.user);

    check_value_fits_in_width(user, params.get_aruser_width(), "ARUSER");
    signals[ARUSER]->set(user);

    signals[ARID]->set(payload.id);
    signals[ARADDR]->set(addr);
    signals[ARLEN]->set(payload.get_len());
    signals[ARSIZE]->set(payload.get_size());
    signals[ARBURST]->set(payload.get_burst());
    signals[ARLOCK]->set(payload.lock);
    signals[ARCACHE]->set(payload.cache);
    signals[ARPROT]->set(payload.prot);
    signals[ARQOS]->set(payload.qos);
    signals[ARREGION]->set(payload.region);
    signals[ARSNOOP]->set(payload.snoop);
    signals[ARDOMAIN]->set(payload.domain);
    signals[ARBAR]->set(payload.bar);
}

void AXI4SignalAccessSlaveAdapter::set_arvalid(bool valid) { signals[ARVALID]->set(valid); }
bool AXI4SignalAccessSlaveAdapter::get_arready() { return signals[ARREADY]->get8(); }

void AXI4SignalAccessSlaveAdapter::get_r(Payload& payload)
{
    check_size(payload.get_size());

    if (transaction_mapper && get_rlast())
        transaction_mapper->remove_payload(&payload);

    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];
    /* Zero all bytes to ensure data always show the same value even for don't cares. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);
    signals[RDATA]->get_ptr(data);

    Resp resp = signals[RRESP]->get8();
    payload.read_in_beat_raw(params.get_data_size(), data, resp);
}

uint32_t AXI4SignalAccessSlaveAdapter::get_rid() { return signals[RID]->get32(); }
bool AXI4SignalAccessSlaveAdapter::get_rlast() { return signals[RLAST]->get8(); }
uint64_t AXI4SignalAccessSlaveAdapter::get_ruser() { return signals[RUSER]->get64(); }
bool AXI4SignalAccessSlaveAdapter::get_rvalid() { return signals[RVALID]->get8(); }
void AXI4SignalAccessSlaveAdapter::set_rready(bool ready) { signals[RREADY]->set(ready); }

AXI4::Payload* AXI4SignalAccessSlaveAdapter::get_ac()
{
    AXI4::Payload* payload = Payload::new_payload(COMMAND_SNOOP,
        /* addr_width pruning only happens in slaves. */
        signals[ACADDR]->get64(), params.get_data_size(),
        params.get_cache_line_beat_count() - 1, AXI4::BURST_INCR);

    payload->snoop = signals[ACSNOOP]->get8();
    payload->prot = signals[ACPROT]->get8();

    return payload;
}

bool AXI4SignalAccessSlaveAdapter::get_acvalid() { return signals[ACVALID]->get8(); }
void AXI4SignalAccessSlaveAdapter::set_acready(bool valid) { signals[ACREADY]->set(valid); }

void AXI4SignalAccessSlaveAdapter::set_cr(Payload& payload)
{
    signals[CRRESP]->set(payload.get_resp());
}

void AXI4SignalAccessSlaveAdapter::set_crvalid(bool valid) { signals[CRVALID]->set(valid); }
bool AXI4SignalAccessSlaveAdapter::get_crready() { return signals[CRREADY]->get8(); }

void AXI4SignalAccessSlaveAdapter::set_cd(Payload& payload, unsigned int beat_index)
{
    Size data_size = params.get_data_size();
    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];

    /* Zero all bytes to ensure data don't care byte lanes always show the same value. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);

    payload.read_out_beat_raw(data_size, beat_index, data);

    signals[CDDATA]->set_ptr(data);
    signals[CDLAST]->set(beat_index == payload.get_len());
}


void AXI4SignalAccessSlaveAdapter::set_cdvalid(bool valid) { signals[CDVALID]->set(valid); }
bool AXI4SignalAccessSlaveAdapter::get_cdready() { return signals[CDREADY]->get8(); }
void AXI4SignalAccessSlaveAdapter::set_wack(bool wack) { signals[SIGNAL_WACK]->set(wack); }
void AXI4SignalAccessSlaveAdapter::set_rack(bool rack) { signals[SIGNAL_RACK]->set(rack); }

AXI4Slave::AXI4Slave(const std::string& name_, TLM::Protocol protocol_) :
    name(name_),
    protocol(protocol_),
    error_handler(&default_error_handler),
    bw_interface(NULL),
    signal_interface(NULL),
    aw_state(CHANNEL_STATE_CLEAR),
    w_state(CHANNEL_STATE_CLEAR),
    b_state(CHANNEL_STATE_CLEAR),
    ar_state(CHANNEL_STATE_CLEAR),
    r_state(CHANNEL_STATE_CLEAR),
    ac_state(CHANNEL_STATE_CLEAR),
    cr_state(CHANNEL_STATE_CLEAR),
    cd_state(CHANNEL_STATE_CLEAR),
    wack_state(CHANNEL_STATE_CLEAR),
    rack_state(CHANNEL_STATE_CLEAR),
    aw(NULL),
    w(NULL),
    b(NULL),
    ar(NULL),
    r(NULL),
    ac(NULL),
    cr(NULL),
    cd(NULL),
    rlast(false),
    w_beat_count(0),
    cd_beat_count(0),
    input_event(NULL),
    signal_input_event_at_clock_edges(true),
    handshake_event(NULL),
    allow_same_cycle_reads(false)
{}

void AXI4Slave::set_signal_interface(AXI4SlaveInterface& signal_interface_)
{
    signal_interface = &signal_interface_;
    signal_interface->set_error_handler(error_handler);
    signal_interface->reset();
}

void AXI4Slave::set_bw_interface(AXI4BwInterface& bw_interface_)
{ bw_interface = &bw_interface_; }

void AXI4Slave::set_error_handler(ErrorHandler* error_handler_)
{
    error_handler = (error_handler_ ? error_handler_ : &default_error_handler);
    if (signal_interface)
        signal_interface->set_error_handler(error_handler);
}

void AXI4Slave::set_input_event(InputEvent& input_event_) { input_event = &input_event_; }

void AXI4Slave::enable_clock_event(bool enable) { signal_input_event_at_clock_edges = enable; }

void AXI4Slave::enable_same_cycle_reads(bool enable) { allow_same_cycle_reads = enable; }

void AXI4Slave::set_handshake_event(InputEvent& handshake_event_) { handshake_event = &handshake_event_; }

bool AXI4Slave::fw(Payload& payload, Phase& phase)
{
    try {
        if (input_event)
            input_event->event();

        /*
         * Forward data moves from transactions onto the signals of the signal-level model.
         * Each forward clause is composed of:
         * 1) Check the relevant channel is clear.
         * 2) Setting the 'on the wires' payload members aw, w, etc. including ref-ing
         *      the payload (as it is retained until posedge).
         * 3) Setting data signals using signal_interface->set... calls.
         * 4) For W: increment write_beat_count
         * 5) Push the payload onto all the queues defining the ordering of responses.
         * 6) Put valid on the signal-level valid and step the channel state to REQ
         * 6a) Signal a valid to ready event if a signal level agent needs to combinatorially
         *      update ready from valid.
         * 7) Checking for an immediate signal-level ready and modifying phase to ...READY
         *      and step the channel state to ACK.
         *
         * Forward ready acknowledgements are composed of:
         * 1) Check the relevant channel is in REQ.
         * 2) Setting ready signals using signal_interface->set... calls.
         * 3) Stepping the channels state on to ACK
         *
         * Queues are popped and 'on the wires' members are NULLed and unref-ed at posedge.
         */
        switch (phase)
        {
        case AW_VALID:
            {
                if (aw_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected AWVALID before accepting a previous AW");

                aw = &payload;
                aw->ref();

                signal_interface->set_aw(*aw);

                b_queue[aw->id].push_back(aw);
                aw->ref();

                signal_interface->set_awvalid(true);
                aw_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_awready();

                if (ready)
                {
                    aw_state = CHANNEL_STATE_ACK;
                    phase = AW_READY;
                }

                return ready;
            }
        case W_VALID:
        case W_VALID_LAST:
            {
                if (w_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected WVALID before accepting a previous W");

                w = &payload;
                w->ref();

                signal_interface->set_w(*w, w_beat_count);
                w_beat_count++;

                Phase expected_phase = (w_beat_count == w->get_beat_count() ? W_VALID_LAST : W_VALID);

                if (phase != expected_phase)
                {
                    error_handler->error("Unexpected W phase: " + to_string(phase) +
                        " after receiving: " + to_string(w_beat_count) + " of: " +
                        to_string(w->get_beat_count()) + " beats, expected: " + to_string(phase));
                }

                signal_interface->set_wvalid(true);
                w_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_wready();

                if (ready)
                {
                    w_state = CHANNEL_STATE_ACK;
                    phase = W_READY;
                }

                return ready;
            }
        case B_READY:
            {
                if (b_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected BREADY without BVALID");

                signal_interface->set_bready(true);
                b_state = CHANNEL_STATE_ACK;
                return false;
            }
        case AR_VALID:
            {
                if (ar_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected ARVALID before accepting a previous AR");

                ar = &payload;
                ar->ref();

                signal_interface->set_ar(*ar);

                r_queue[ar->id].push_back(ar);
                ar->ref();

                signal_interface->set_arvalid(true);
                ar_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                /*
                 * Try to perform this cycle's R straightaway if same cycle reads are allowed.
                 * This should work irrespective of the ordering of this call and negedge as
                 * r_state will be appropriately updated.
                 */
                if (allow_same_cycle_reads)
                {
                    negedge_r_bw();

                    if (handshake_event)
                        handshake_event->event();
                }

                bool ready = signal_interface->get_arready();

                if (ready)
                {
                    ar_state = CHANNEL_STATE_ACK;
                    phase = AR_READY;
                }

                return ready;
            }
        case R_READY:
            {
                if (r_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected RREADY without RVALID");

                signal_interface->set_rready(true);
                r_state = CHANNEL_STATE_ACK;
                return false;
            }
        case CR_VALID:
            {
                if (cr_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected CRVALID before accepting a previous CR");

                cr = &payload;
                cr->ref();

                signal_interface->set_cr(*cr);

                signal_interface->set_crvalid(true);
                cr_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_crready();

                if (ready)
                {
                    cr_state = CHANNEL_STATE_ACK;
                    phase = CR_READY;
                    return true;
                }

                return ready;
            }
        case CD_VALID:
        case CD_VALID_LAST:
            {
                if (cd_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected CDVALID{,_LAST} before accepting a previous CD");

                cd = &payload;
                cd->ref();

                signal_interface->set_cd(*cd, cd_beat_count);
                cd_beat_count++;

                Phase expected_phase = (cd_beat_count == cd->get_beat_count() ? CD_VALID_LAST : CD_VALID);

                if (phase != expected_phase)
                {
                    error_handler->error("Unexpected CD phase: " + to_string(phase) +
                        " after receiving: " + to_string(cd_beat_count) + " of: " +
                        to_string(cd->get_beat_count()) + " beats, expected: " + to_string(phase));
                }

                signal_interface->set_cdvalid(true);
                cd_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_cdready();

                if (ready)
                {
                    cd_state = CHANNEL_STATE_ACK;
                    phase = CD_READY;
                    return true;
                }

                return ready;
            }
        case AC_READY:
            {
                if (ac_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected ACREADY without ACVALID");

                signal_interface->set_acready(true);
                ac_state = CHANNEL_STATE_ACK;
                return false;
            }
        case WACK:
            {
                wack = &payload;
                wack->ref();

                signal_interface->set_wack(true);
                wack_state = CHANNEL_STATE_REQ;
                return false;
            }
        case RACK:
            {
                rack = &payload;
                rack->ref();

                signal_interface->set_rack(true);
                rack_state = CHANNEL_STATE_REQ;
                return false;
            }
        default:
            error_handler->error("Uncaught phase: " + to_string(phase));
            return false;
        }
    } catch (std::exception& e)
    {
        error_handler->error(e.what(), "In fw on signal level port: " + name);
        return false;
    }
}

void AXI4Slave::posedge()
{
    if (input_event && signal_input_event_at_clock_edges)
        input_event->event();

    if (aw_state == CHANNEL_STATE_ACK)
    {
        assert(aw);
        aw_state = CHANNEL_STATE_CLEAR;
        aw->unref();
        aw = NULL;
    }

    if (w_state == CHANNEL_STATE_ACK)
    {
        assert(w);
        w_state = CHANNEL_STATE_CLEAR;

        /* Reset the beat count at the end of a burst. */
        if (w_beat_count == w->get_beat_count())
            w_beat_count = 0;

        w->unref();
        w = NULL;
    }

    if (b_state == CHANNEL_STATE_ACK)
    {
        assert(b);
        b_state = CHANNEL_STATE_CLEAR;

        TransactionQueuesById::iterator found_queue = b_queue.find(b->id);

        if (found_queue == b_queue.end() || found_queue->second.empty())
        {
            error_handler->error("Unexpected B with ID: " + to_string(b->id)
                + ", no AW matches it");
        }

        if (b != found_queue->second.front())
        {
            error_handler->error("Unexpected B with payload uid: " + to_string(b->uid) +
                " expecting payload with uid: " + to_string(found_queue->second.front()->uid) +
                ", bad ordering");
        }

        /* Pop from b_queue. */
        found_queue->second.pop_front();
        b->unref();

        b->unref();
        b = NULL;
    }

    if (ar_state == CHANNEL_STATE_ACK)
    {
        assert(ar);
        ar_state = CHANNEL_STATE_CLEAR;
        ar->unref();
        ar = NULL;
    }

    if (r_state == CHANNEL_STATE_ACK)
    {
        assert(r);
        r_state = CHANNEL_STATE_CLEAR;

        TransactionQueuesById::iterator found_queue = r_queue.find(r->id);

        if (found_queue == r_queue.end() || found_queue->second.empty())
        {
            error_handler->error("Unexpected R with ID: " + to_string(r->id)
                + ", no AR matches it");
        }

        if (r != found_queue->second.front())
        {
            error_handler->error("Unexpected R with payload uid: " + to_string(r->uid) +
                " expecting payload with uid: " + to_string(found_queue->second.front()->uid) +
                ", bad ordering");
        }

        /* Consider popping the head of r_queue if we're at the end of a burst. */
        if (rlast)
        {
            found_queue->second.pop_front();
            r->unref();
        }

        r->unref();
        r = NULL;
    }

    if (ac_state == CHANNEL_STATE_ACK)
    {
        assert(ac);
        ac_state = CHANNEL_STATE_CLEAR;
        ac->unref();
        ac = NULL;
    }

    if (cr_state == CHANNEL_STATE_ACK)
    {
        assert(cr);
        cr_state = CHANNEL_STATE_CLEAR;

        cr->unref();
        cr = NULL;
    }

    if (cd_state == CHANNEL_STATE_ACK)
    {
        assert(cd);
        cd_state = CHANNEL_STATE_CLEAR;

        /* Reset the beat count at the end of a burst. */
        if (cd_beat_count == cd->get_beat_count())
            cd_beat_count = 0;

        cd->unref();
        cd = NULL;
    }

    if (wack_state == CHANNEL_STATE_REQ)
    {
        assert(wack);
        wack_state = CHANNEL_STATE_CLEAR;
        wack->unref();
        wack = NULL;
    }

    if (rack_state == CHANNEL_STATE_REQ)
    {
        assert(rack);
        rack_state = CHANNEL_STATE_CLEAR;
        rack->unref();
        rack = NULL;
    }
}

void AXI4Slave::negedge_r_bw()
{
    if (r_state == CHANNEL_STATE_CLEAR && signal_interface->get_rvalid())
    {
        uint32_t rid = signal_interface->get_rid();

        TransactionQueuesById::iterator found_queue = r_queue.find(rid);

        if (found_queue == r_queue.end() || found_queue->second.empty())
        {
            error_handler->error("Unexpected R with ID: " + to_string(rid)
                + ", no AR matches it");
        }

        r = found_queue->second.front();
        r->ref();

        signal_interface->get_r(*r);
        rlast = signal_interface->get_rlast();

        Phase phase = (rlast ? R_VALID_LAST : R_VALID);
        bool ready = bw_interface->bw(*r, phase);
        r_state = CHANNEL_STATE_REQ;

        if (ready)
        {
            assert(phase == R_READY);
            signal_interface->set_rready(true);
            r_state = CHANNEL_STATE_ACK;
        }
    }
}

void AXI4Slave::negedge()
{
    if (input_event && signal_input_event_at_clock_edges)
        input_event->event();

    /*
     * First reset handshake signalling from the previous cycle. If fw has already set any
     * VALID/READY signal, then the .._state will *not* be CLEAR.
     */
    if (aw_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_awvalid(false);

    if (w_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_wvalid(false);

    if (b_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_bready(false);

    if (ar_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_arvalid(false);

    if (r_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_rready(false);

    if (ac_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_acready(false);

    if (cr_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_crvalid(false);

    if (cd_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_cdvalid(false);

    if (wack_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_wack(false);

    if (rack_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_rack(false);

    /* Propagate signal changes before sampling other ready/valids. */
    if (allow_same_cycle_reads && handshake_event)
        handshake_event->event();

    /*
     * Handle shovelling signal-level communications into transactions;
     *
     * Checking ..._state here looks like an ordering problem between fw@negedge and this negedge
     * but actually fw would not have set ..._state to REQ unless it was not responded to straight
     * away, by checking get_...ready, and get_...ready() must not be able to change around negedge
     * between fw and this function.
     *
     * Backwards data moves from signals of the signals-level model to transactions.
     * Each backward clause is composed of:
     * 1) Check the relevant channel is clear and the signal-level valid signal is high.
     * 2) Setting the 'on the wires' payload members b, r, etc. including ref-ing
     *      the payload (as it is retained until posedge). The payload is found from the
     *      relevant response queue.
     * 3) Setting transaction members on the payload from signals using signal_interface->get...
     *      calls.
     * 6) Send the req as a VALID phase using bw.
     * 7) Checking for an immediate transaction ready and setting the ready signal
     *      and step the channel state to ACK.
     *
     * Backward ready acknowledgements are composed of:
     * 1) Check the relevant channel is in REQ and the signal-level ready signal is high.
     * 2) Stepping the channels state on to ACK.
     * 3) Send the ack back as a READY phase using bw.
     *
     * Queues are popped and 'on the wires' members are NULLed and unref-ed at posedge.
     */

    if (aw_state == CHANNEL_STATE_REQ && signal_interface->get_awready())
    {
        Phase phase = AW_READY;
        bw_interface->bw(*aw, phase);
        aw_state = CHANNEL_STATE_ACK;
    }

    if (w_state == CHANNEL_STATE_REQ && signal_interface->get_wready())
    {
        Phase phase = W_READY;
        bw_interface->bw(*w, phase);
        w_state = CHANNEL_STATE_ACK;
    }

    if (b_state == CHANNEL_STATE_CLEAR && signal_interface->get_bvalid())
    {
        uint32_t bid = signal_interface->get_bid();

        TransactionQueuesById::iterator found_queue = b_queue.find(bid);

        if (found_queue == b_queue.end() || found_queue->second.empty())
        {
            error_handler->error("Unexpected B with ID: " + to_string(bid)
                + ", no AW matches it");
        }

        b = found_queue->second.front();
        b->ref();
        signal_interface->get_b(*b);

        Phase phase = B_VALID;
        bool ready = bw_interface->bw(*b, phase);
        b_state = CHANNEL_STATE_REQ;

        if (ready)
        {
            assert(phase == B_READY);
            signal_interface->set_bready(true);
            b_state = CHANNEL_STATE_ACK;
        }
    }

    if (ar_state == CHANNEL_STATE_REQ && signal_interface->get_arready())
    {
        Phase phase = AR_READY;
        bw_interface->bw(*ar, phase);
        ar_state = CHANNEL_STATE_ACK;
    }
    
    negedge_r_bw();

    if (ac_state == CHANNEL_STATE_CLEAR && signal_interface->get_acvalid())
    {
        ac = signal_interface->get_ac();
        /* First ref from get_ac counts as the ac->ref. */

        Phase phase = AC_VALID;
        bool ready = bw_interface->bw(*ac, phase);

        ac_state = CHANNEL_STATE_REQ;

        if (ready)
        {
            assert(phase == AC_READY);
            signal_interface->set_acready(true);
            ac_state = CHANNEL_STATE_ACK;
        }
    }

    if (cr_state == CHANNEL_STATE_REQ && signal_interface->get_crready())
    {
        Phase phase = CR_READY;
        bw_interface->bw(*cr, phase);
        cr_state = CHANNEL_STATE_ACK;
    }

    if (cd_state == CHANNEL_STATE_REQ && signal_interface->get_cdready())
    {
        Phase phase = CD_READY;
        bw_interface->bw(*cd, phase);
        cd_state = CHANNEL_STATE_ACK;
    }
}

void AXI4MasterInterface::reset()
{
    set_awready(false);
    set_wready(false);
    set_bvalid(false);
    set_arready(false);
    set_rvalid(false);
    set_acvalid(false);
    set_crready(false);
    set_cdready(false);

    /* FIXME, drive all other signals including ones which need payloads. */
}

AXI4SignalAccessMasterAdapter::AXI4SignalAccessMasterAdapter(ErrorHandler* error_handler_) :
    AXI4PortSignalAccesses(error_handler_),
    transaction_mapper(NULL)
{}

void AXI4SignalAccessMasterAdapter::set_transaction_mapper(AXI4TransactionMapper& transaction_mapper_)
{ transaction_mapper = &transaction_mapper_; }

void AXI4SignalAccessMasterAdapter::set_error_handler(ErrorHandler* error_handler_)
{ AXI4PortSignalAccesses::set_error_handler(error_handler_); }

AXI4::Payload* AXI4SignalAccessMasterAdapter::get_aw()
{
    AXI4::Payload* payload = NULL;
    uint64_t user = signals[AWUSER]->get64();

    if (transaction_mapper)
    {
        AXI4::Payload* parent_payload = transaction_mapper->lookup_payload(user);
        if (parent_payload)
        {
            payload = parent_payload->descend(COMMAND_WRITE,
                /* addr_width pruning only happens in slaves. */
                signals[AWADDR]->get64(), signals[AWSIZE]->get8(),
                signals[AWLEN]->get8(), signals[AWBURST]->get8());
        }
    }

    if (!payload)
    {
        payload = Payload::new_payload(COMMAND_WRITE,
            /* addr_width pruning only happens in slaves. */
            signals[AWADDR]->get64(), signals[AWSIZE]->get8(),
            signals[AWLEN]->get8(), signals[AWBURST]->get8());
        payload->user = user;
    }

    payload->id = signals[AWID]->get32();
    payload->lock = signals[AWLOCK]->get8();
    payload->cache = signals[AWCACHE]->get8();
    payload->prot = signals[AWPROT]->get8();
    payload->qos = signals[AWQOS]->get8();
    payload->region = signals[AWREGION]->get8();
    payload->snoop = signals[AWSNOOP]->get8();
    payload->domain = signals[AWDOMAIN]->get8();
    payload->bar = signals[AWBAR]->get8();
    payload->unique = signals[AWUNIQUE]->get8();

    return payload;
}

bool AXI4SignalAccessMasterAdapter::get_awvalid() { return signals[AWVALID]->get8(); }
void AXI4SignalAccessMasterAdapter::set_awready(bool valid) { signals[AWREADY]->set(valid); }

void AXI4SignalAccessMasterAdapter::get_w(AXI4::Payload& payload)
{
    check_size(payload.get_size());

    Size data_size = params.get_data_size();
    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];
    uint8_t strobe[MAX_DATA_STROBE_WIDTH_IN_BYTES];
    /* Zero all bytes to ensure data and strobe don't care byte lanes always show the same value. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);
    std::memset(strobe, 0, MAX_DATA_STROBE_WIDTH_IN_BYTES);

    signals[WDATA]->get_ptr(data);
    signals[WSTRB]->get_ptr(strobe);

    payload.write_in_beat_raw(data_size, data, strobe);
}

uint32_t AXI4SignalAccessMasterAdapter::get_wid() { return signals[WID]->get32(); }
bool AXI4SignalAccessMasterAdapter::get_wlast() { return signals[WLAST]->get8(); }
uint64_t AXI4SignalAccessMasterAdapter::get_wuser() { return signals[WUSER]->get64(); }

bool AXI4SignalAccessMasterAdapter::get_wvalid() { return signals[WVALID]->get8(); }
void AXI4SignalAccessMasterAdapter::set_wready(bool ready) { signals[WREADY]->set(ready); }

void AXI4SignalAccessMasterAdapter::set_bid(uint32_t id) { signals[BID]->set(id); }
void AXI4SignalAccessMasterAdapter::set_bresp(AXI4::Resp resp) { signals[BRESP]->set(resp); }
void AXI4SignalAccessMasterAdapter::set_bvalid(bool valid) { signals[BVALID]->set(valid); }
bool AXI4SignalAccessMasterAdapter::get_bready() { return signals[BREADY]->get8(); }

AXI4::Payload *AXI4SignalAccessMasterAdapter::get_ar()
{
    AXI4::Payload* payload = NULL;
    uint64_t user = signals[ARUSER]->get64();

    if (transaction_mapper)
    {
        AXI4::Payload* parent_payload = transaction_mapper->lookup_payload(user);
        if (parent_payload)
        {
            payload = parent_payload->descend(COMMAND_READ,
                /* addr_width pruning only happens in slaves. */
                signals[ARADDR]->get64(), signals[ARSIZE]->get8(),
                signals[ARLEN]->get8(), signals[ARBURST]->get8());
        }
    }

    if (!payload)
    {
        payload = Payload::new_payload(COMMAND_READ,
            /* addr_width pruning only happens in slaves. */
            signals[ARADDR]->get64(), signals[ARSIZE]->get8(),
            signals[ARLEN]->get8(), signals[ARBURST]->get8());
        payload->user = user;
    }

    payload->id = signals[ARID]->get32();
    payload->lock = signals[ARLOCK]->get8();
    payload->cache = signals[ARCACHE]->get8();
    payload->prot = signals[ARPROT]->get8();
    payload->qos = signals[ARQOS]->get8();
    payload->region = signals[ARREGION]->get8();
    payload->snoop = signals[ARSNOOP]->get8();
    payload->domain = signals[ARDOMAIN]->get8();
    payload->bar = signals[ARBAR]->get8();

    return payload;
}

bool AXI4SignalAccessMasterAdapter::get_arvalid() { return signals[ARVALID]->get8(); }
void AXI4SignalAccessMasterAdapter::set_arready(bool ready) { signals[ARREADY]->set(ready); }

void AXI4SignalAccessMasterAdapter::set_r(AXI4::Payload& payload, unsigned int beat_index)
{
    check_size(payload.get_size());

    Size data_size = params.get_data_size();
    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];
    /* Zero all bytes to ensure data don't care byte lanes always show the same value. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);

    payload.read_out_beat_raw(data_size, beat_index, data);

    signals[RID]->set(payload.id);
    signals[RDATA]->set_ptr(data);
    signals[RRESP]->set(payload.read_out_beat_resp(beat_index));
    signals[RLAST]->set(beat_index == payload.get_len());
    signals[RUSER]->set(payload.user);
}

void AXI4SignalAccessMasterAdapter::set_rvalid(bool valid) { signals[RVALID]->set(valid); }
bool AXI4SignalAccessMasterAdapter::get_rready() { return signals[RREADY]->get8(); }

void AXI4SignalAccessMasterAdapter::set_ac(Payload& payload)
{
    uint64_t addr = params.crop_address(payload.get_address(), params.get_acaddr_width());

    signals[ACADDR]->set(addr);
    signals[ACSNOOP]->set(payload.snoop);
    signals[ACPROT]->set(payload.prot);
}

void AXI4SignalAccessMasterAdapter::set_acvalid(bool valid) { signals[ACVALID]->set(valid); }
bool AXI4SignalAccessMasterAdapter::get_acready() { return signals[ACREADY]->get8(); }

void AXI4SignalAccessMasterAdapter::get_cr(AXI4::Payload& payload)
{
    payload.set_resp(signals[CRRESP]->get8());
}

bool AXI4SignalAccessMasterAdapter::get_crvalid() { return signals[CRVALID]->get8(); }
void AXI4SignalAccessMasterAdapter::set_crready(bool ready) { signals[CRREADY]->set(ready); }

void AXI4SignalAccessMasterAdapter::get_cd(AXI4::Payload& payload)
{
    uint8_t data[MAX_DATA_WIDTH_IN_BYTES];
    /* Zero all bytes to ensure data don't care byte lanes always show the same value. */
    std::memset(data, 0, MAX_DATA_WIDTH_IN_BYTES);
    signals[CDDATA]->get_ptr(data);

    payload.read_in_beat_raw(params.get_data_size(), data);
}

bool AXI4SignalAccessMasterAdapter::get_cdlast() { return signals[CDLAST]->get8(); }
bool AXI4SignalAccessMasterAdapter::get_cdvalid() { return signals[CDVALID]->get8(); }
void AXI4SignalAccessMasterAdapter::set_cdready(bool ready) { signals[CDREADY]->set(ready); }
bool AXI4SignalAccessMasterAdapter::get_wack() { return signals[SIGNAL_WACK]->get8(); }
bool AXI4SignalAccessMasterAdapter::get_rack() { return signals[SIGNAL_RACK]->get8(); }

AXI4Master::AXI4Master(const std::string& name_, TLM::Protocol protocol_) :
    name(name_),
    protocol(protocol_),
    error_handler(&default_error_handler),
    fw_interface(NULL),
    signal_interface(NULL),
    aw_state(CHANNEL_STATE_CLEAR),
    w_state(CHANNEL_STATE_CLEAR),
    b_state(CHANNEL_STATE_CLEAR),
    ar_state(CHANNEL_STATE_CLEAR),
    r_state(CHANNEL_STATE_CLEAR),
    ac_state(CHANNEL_STATE_CLEAR),
    cr_state(CHANNEL_STATE_CLEAR),
    cd_state(CHANNEL_STATE_CLEAR),
    wack_state(CHANNEL_STATE_CLEAR),
    rack_state(CHANNEL_STATE_CLEAR),
    aw(NULL),
    w(NULL),
    b(NULL),
    ar(NULL),
    r(NULL),
    ac(NULL),
    cr(NULL),
    cd(NULL),
    wack(NULL),
    rack(NULL),
    wlast(false),
    input_event(NULL),
    signal_input_event_at_clock_edges(true),
    handshake_event(NULL),
    allow_same_cycle_reads(false)
{}

void AXI4Master::set_signal_interface(AXI4MasterInterface& signal_interface_)
{
    signal_interface = &signal_interface_;
    signal_interface->set_error_handler(error_handler);
    signal_interface->reset();
}

void AXI4Master::set_fw_interface(AXI4FwInterface& fw_interface_)
{ fw_interface = &fw_interface_; }

void AXI4Master::set_input_event(InputEvent& input_event_) { input_event = &input_event_; }

void AXI4Master::set_error_handler(ErrorHandler* error_handler_)
{
    error_handler = (error_handler_ ? error_handler_ : &default_error_handler);
    if (signal_interface)
        signal_interface->set_error_handler(error_handler);
}

void AXI4Master::enable_clock_event(bool enable) { signal_input_event_at_clock_edges = enable; }

void AXI4Master::enable_same_cycle_reads(bool enable) { allow_same_cycle_reads = enable; }

void AXI4Master::set_handshake_event(InputEvent& handshake_event_) { handshake_event = &handshake_event_; }

bool AXI4Master::bw(Payload& payload, Phase& phase)
{
    try {
        if (input_event)
            input_event->event();

        /*
         * Backward data moves from transactions onto the signals of the signal-level model.
         * Each backward clause is composed of:
         * 1) Check the relevant channel is clear.
         * 2) Setting the 'on this wires' payload members aw, w, etc. including ref-ing
         *      the payload (as it is retained until posedge).
         * 3) Setting data signals using signal_interface->set... calls.
         * 4) For R: increment read_beat_counts[]
         * 5) Push the payload onto all the queues defining the ordering of responses.
         * 6) Put valid on the signal-level valid and step the channel state to REQ
         * 6a) Signal a valid to ready event if a signal level agent needs to combinatorially
         *      update ready from valid.
         * 7) Checking for an immediate signal-level ready and modifying phase to ...READY
         *      and step the channel state to ACK.
         *
         * Backward ready acknowledgements are composed of:
         * 1) Check the relevant channel is in REQ.
         * 2) Setting ready signals using signal_interface->set... calls.
         * 3) Stepping the channels state on to ACK
         *
         * Queues are popped and 'on the wires' members are NULLed and unref-ed at posedge.
         */
        switch (phase)
        {
        case AW_READY:
            {
                if (aw_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected AWREADY without AWVALID");

                signal_interface->set_awready(true);
                aw_state = CHANNEL_STATE_ACK;
                return false;
            }
        case W_READY:
            {
                if (w_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected WREADY without WVALID");

                signal_interface->set_wready(true);
                w_state = CHANNEL_STATE_ACK;
                return false;
            }
        case B_VALID:
            {
                if (b_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected BVALID before accepting a previous B");

                b = &payload;
                b->ref();

                signal_interface->set_bid(payload.id);
                signal_interface->set_bresp(payload.get_resp());

                signal_interface->set_bvalid(true);
                b_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_bready();

                if (ready)
                {
                    b_state = CHANNEL_STATE_ACK;
                    phase = B_READY;
                }

                return ready;
            }
            break;
        case AR_READY:
            {
                if (ar_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected ARREADY without ARVALID");

                signal_interface->set_arready(true);
                ar_state = CHANNEL_STATE_ACK;
                return false;
            }
        case R_VALID:
        case R_VALID_LAST:
            {
                if (r_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected RVALID{,_LAST} before accepting a previous R");

                r = &payload;
                r->ref();

                BeatCountMap::iterator beat_count = read_beat_counts.find(payload.uid);

                if (beat_count == read_beat_counts.end())
                    error_handler->error("Unexpected R, no AR matches it");

                signal_interface->set_r(payload, beat_count->second);
                beat_count->second++;

                Phase expected_phase = (beat_count->second == r->get_beat_count() ? R_VALID_LAST : R_VALID);

                if (phase != expected_phase)
                {
                    error_handler->error("Unexpected R phase: " + to_string(phase) +
                        " after receiving: " + to_string(beat_count->second) + " of: " +
                        to_string(r->get_beat_count()) + " beats, expected: " + to_string(phase));
                }

                signal_interface->set_rvalid(true);
                r_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_rready();

                if (ready)
                {
                    r_state = CHANNEL_STATE_ACK;
                    phase = R_READY;
                }

                return ready;
            }
        case AC_VALID:
            {
                if (ac_state != CHANNEL_STATE_CLEAR)
                    error_handler->error("Unexpected ACVALID before accepting a previous AC");

                ac = &payload;
                ac->ref();

                signal_interface->set_ac(payload);

                cr_queue.push_back(ac);
                ac->ref();
                /*
                 * CR responses cause cr_queue elements to be promoted (or not)
                 * to the cd_queue.
                 */

                signal_interface->set_acvalid(true);
                ac_state = CHANNEL_STATE_REQ;

                if (handshake_event)
                    handshake_event->event();

                bool ready = signal_interface->get_acready();

                if (ready)
                {
                    ac_state = CHANNEL_STATE_ACK;
                    phase = AC_READY;
                }

                return ready;
            }
        case CR_READY:
            {
                if (cr_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected CRREADY without CRVALID");

                signal_interface->set_crready(true);
                cr_state = CHANNEL_STATE_ACK;
                return false;
            }
        case CD_READY:
            {
                if (cd_state != CHANNEL_STATE_REQ)
                    error_handler->error("Unexpected CDREADY without CDVALID");

                signal_interface->set_cdready(true);
                cd_state = CHANNEL_STATE_ACK;
                return false;
            }
        default:
            error_handler->error("Uncaught phase: " + to_string(phase));
            return false;
        }
    } catch (std::exception& e)
    {
        error_handler->error(e.what(), "In bw on signal level port: " + name);
        return false;
    }
}

void AXI4Master::posedge()
{
    if (input_event && signal_input_event_at_clock_edges)
        input_event->event();

    if (aw_state == CHANNEL_STATE_ACK)
    {
        assert(aw);
        aw_state = CHANNEL_STATE_CLEAR;
        aw->unref();
        aw = NULL;
    }

    if (w_state == CHANNEL_STATE_ACK)
    {
        assert(w);
        w_state = CHANNEL_STATE_CLEAR;

        if (w_queue.empty())
            error_handler->error("Unexpected W, no AW matches it");

        if (w != w_queue.front())
        {
            error_handler->error("Unexpected W with payload uid: " + to_string(w->uid) +
                " expecting payload with uid: " + to_string(w_queue.front()->uid) +
                ", bad ordering");
        }

        /* Pop w_queue at the end of a burst. */
        if (wlast)
        {
            w_queue.pop_front();
            w->unref();
        }

        w->unref();
        w = NULL;
    }

    if (b_state == CHANNEL_STATE_ACK)
    {
        assert(b);
        b_state = CHANNEL_STATE_CLEAR;
        b->unref();
        b = NULL;
    }

    if (ar_state == CHANNEL_STATE_ACK)
    {
        assert(ar);
        ar_state = CHANNEL_STATE_CLEAR;
        ar->unref();
        ar = NULL;
    }

    if (r_state == CHANNEL_STATE_ACK)
    {
        assert(r);
        r_state = CHANNEL_STATE_CLEAR;

        /* Reset the beat count at the end of a burst. */
        if (read_beat_counts[r->uid] == r->get_beat_count())
            read_beat_counts.erase(r->uid);

        r->unref();
        r = NULL;
    }

    if (ac_state == CHANNEL_STATE_ACK)
    {
        assert(ac);
        ac_state = CHANNEL_STATE_CLEAR;
        ac->unref();
        ac = NULL;
    }

    if (cr_state == CHANNEL_STATE_ACK)
    {
        assert(cr);
        cr_state = CHANNEL_STATE_CLEAR;

        if (cr_queue.empty())
            error_handler->error("Unexpected CR, no AC matches it");

        if (cr != cr_queue.front())
        {
            error_handler->error("Unexpected CR with payload uid: " + to_string(cr->uid) +
                " expecting payload with uid: " + to_string(cr_queue.front()->uid) +
                ", bad ordering");
        }

        /* Pop from cr_queue. */
        cr_queue.pop_front();
        cr->unref();

        cr->unref();
        cr = NULL;
    }

    if (cd_state == CHANNEL_STATE_ACK)
    {
        assert(cd);
        cd_state = CHANNEL_STATE_CLEAR;

        if (cd_queue.empty())
            error_handler->error("Unexpected CD, no AC matches it");

        if (cd != cd_queue.front())
        {
            error_handler->error("Unexpected CD with payload uid: " + to_string(cd->uid) +
                " expecting payload with uid: " + to_string(cd_queue.front()->uid) +
                ", bad ordering");
        }

        /* Consider popping the head of cd_queue if we're at the end of a burst. */
        if (cdlast)
        {
            cd_queue.pop_front();
            cd->unref();
        }

        cd->unref();
        cd = NULL;
    }

    if (wack_state == CHANNEL_STATE_REQ)
    {
        assert(wack);
        wack_state = CHANNEL_STATE_CLEAR;

        if (wack_queue.empty())
            error_handler->error("Unexpected WACK, no AW matches it");

        if (wack != wack_queue.front())
        {
            error_handler->error("Unexpected WACK with payload uid: " + to_string(wack->uid) +
                " expecting payload with uid: " + to_string(wack_queue.front()->uid) +
                ", bad ordering");
        }

        wack_queue.pop_front();
        wack->unref();

        wack->unref();
        wack = NULL;
    }

    if (rack_state == CHANNEL_STATE_REQ)
    {
        assert(rack);
        rack_state = CHANNEL_STATE_CLEAR;

        if (rack_queue.empty())
            error_handler->error("Unexpected RACK, no AR matches it");

        if (rack != rack_queue.front())
        {
            error_handler->error("Unexpected RACK with payload uid: " + to_string(rack->uid) +
                " expecting payload with uid: " + to_string(wack_queue.front()->uid) +
                ", bad ordering");
        }

        rack_queue.pop_front();
        rack->unref();

        rack->unref();
        rack = NULL;
    }
}

void AXI4Master::negedge()
{
    if (input_event && signal_input_event_at_clock_edges)
        input_event->event();

    /*
     * First reset handshake signalling from the previous cycle. If bw has already set any
     * VALID/READY signal, then the .._state will *not* be CLEAR.
     */
    if (aw_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_awready(false);

    if (w_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_wready(false);

    if (b_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_bvalid(false);

    if (ar_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_arready(false);

    if (r_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_rvalid(false);

    if (ac_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_acvalid(false);

    if (cr_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_crready(false);

    if (cd_state == CHANNEL_STATE_CLEAR)
        signal_interface->set_cdready(false);

    /* Propagate signal changes before sampling other ready/valids. */
    if (allow_same_cycle_reads && handshake_event)
        handshake_event->event();

    /*
     * Handle shovelling signal-level communications into transactions;
     *
     * Checking ..._state here looks like an ordering problem between bw@negedge and this negedge
     * but actually bw would not have set ..._state to REQ unless it was not responded to straight
     * away, by checking get_...ready, and get_...ready() must not be able to change around negedge
     * between bw and this function.
     *
     * Backwards data moves from signals of the signals-level model to transactions.
     * Each backward clause is composed of:
     * 1) Check the relevant channel is clear and the signal-level valid signal is high.
     * 2) Setting the 'on the wires' payload members aw, ar, etc. including ref-ing
     *      the payload (as it is retained until posedge). The payload is found from the
     *      relevant response queue.
     * 3) Setting transaction members on the payload from signals using signal_interface->get...
     *      calls.
     * 6) Send the req as a VALID phase using bw.
     * 7) Checking for an immediate transaction ready and setting the ready signal
     *      and step the channel state to ACK.
     *
     * Backward ready acknowledgements are composed of:
     * 1) Check the relevant channel is in REQ and the signal-level ready signal is high.
     * 2) Stepping the channels state on to ACK.
     * 3) Send the ack back as a READY phase using bw.
     *
     * Queues are popped and 'on the wires' members are NULLed and unref-ed at posedge.
     */

    if (aw_state == CHANNEL_STATE_CLEAR && signal_interface->get_awvalid())
    {
        aw = signal_interface->get_aw();
        /* payload.ref(); get_aw creates a payload with refcount of 1. */

        Phase phase = AW_VALID;
        bool ready = fw_interface->fw(*aw, phase);
        aw_state = CHANNEL_STATE_REQ;

        w_queue.push_back(aw);
        aw->ref();

        /* Keep an ordered list of all AWs for WACK. */
        if (protocol == TLM::PROTOCOL_ACE)
        {
            wack_queue.push_back(aw);
            aw->ref();
        }

        if (ready)
        {
            assert(phase == AW_READY);
            signal_interface->set_awready(true);
            aw_state = CHANNEL_STATE_ACK;
        }
    }

    if (w_state == CHANNEL_STATE_CLEAR && signal_interface->get_wvalid())
    {
        if (w_queue.empty())
        {
            /* Stall early Ws where the transaction's identity isn't known. */
        } else
        {
            if (w_queue.empty())
                error_handler->error("Unexpected WVALID received. No previous AW matches it");

            w = w_queue.front();
            w->ref();

            signal_interface->get_w(*w);
            wlast = signal_interface->get_wlast();

            Phase phase = (wlast ? W_VALID_LAST : W_VALID);
            bool ready = fw_interface->fw(*w, phase);
            w_state = CHANNEL_STATE_REQ;

            if (ready)
            {
                assert(phase == W_READY);
                signal_interface->set_wready(true);
                w_state = CHANNEL_STATE_ACK;
            }
        }
    }

    if (b_state == CHANNEL_STATE_REQ && signal_interface->get_bready())
    {
        Phase phase = B_READY;
        fw_interface->fw(*b, phase);
        b_state = CHANNEL_STATE_ACK;
    }

    if (ar_state == CHANNEL_STATE_CLEAR && signal_interface->get_arvalid())
    {
        ar = signal_interface->get_ar();
        /* payload.ref(); get_ar creates a payload with refcount of 1. */

        read_beat_counts[ar->uid] = 0;

        Phase phase = AR_VALID;
        bool ready = fw_interface->fw(*ar, phase);
        ar_state = CHANNEL_STATE_REQ;

        /* Keep an ordered list of all ARs for RACK. */
        if (protocol == TLM::PROTOCOL_ACE)
        {
            rack_queue.push_back(ar);
            ar->ref();
        }

        if (ready)
        {
            assert(phase == AR_READY);
            signal_interface->set_arready(true);
            ar_state = CHANNEL_STATE_ACK;
        }
    }

    if (r_state == CHANNEL_STATE_REQ && signal_interface->get_rready())
    {
        Phase phase = R_READY;
        fw_interface->fw(*r, phase);
        r_state = CHANNEL_STATE_ACK;
    }

    if (ac_state == CHANNEL_STATE_REQ && signal_interface->get_acready())
    {
        Phase phase = AC_READY;
        fw_interface->fw(*ac, phase);
        ac_state = CHANNEL_STATE_ACK;
    }

    if (cr_state == CHANNEL_STATE_CLEAR && signal_interface->get_crvalid())
    {
        if (cr_queue.empty())
            error_handler->error("Unexpected CRVALID received. No previous AC matches it");

        cr = cr_queue.front();
        cr->ref();

        signal_interface->get_cr(*cr);

        /* If there is data to transfer, move the transaction into the cd_queue. */
        if (cr->get_resp().has(AXI4::RESP_DATA_TRANSFER))
        {
            cd_queue.push_back(cr);
            cr->ref();
        }

        Phase phase = CR_VALID;
        bool ready = fw_interface->fw(*cr, phase);
        cr_state = CHANNEL_STATE_REQ;

        if (ready)
        {
            assert(phase == CR_READY);
            signal_interface->set_crready(true);
            cr_state = CHANNEL_STATE_ACK;
        }
    }

    /*
     * It is important that CR and CD are sequenced this way so that cd_queue pushes from above can
     * be noticed in the following clause.
     */

    if (cd_state == CHANNEL_STATE_CLEAR && signal_interface->get_cdvalid())
    {
        if (cd_queue.empty())
        {
            /* Stall early CDs where the transaction's identity isn't known. */
        } else
        {
            cd = cd_queue.front();
            cd->ref();

            signal_interface->get_cd(*cd);
            cdlast = signal_interface->get_cdlast();

            Phase phase = (cdlast ? CD_VALID_LAST : CD_VALID);
            bool ready = fw_interface->fw(*cd, phase);
            cd_state = CHANNEL_STATE_REQ;

            if (ready)
            {
                assert(phase == CD_READY);
                signal_interface->set_cdready(true);
                cd_state = CHANNEL_STATE_ACK;
            }
        }
    }

    if (protocol == TLM::PROTOCOL_ACE)
    {
        if (signal_interface->get_wack())
        {
            if (wack_state != CHANNEL_STATE_CLEAR)
                error_handler->error("Unexpected WACK before accepting a previous WACK");

            if (wack_queue.empty())
                error_handler->error("Unexpected WACK received. No previous AW matches it");

            wack = wack_queue.front();
            wack->ref();

            Phase phase = WACK;
            fw_interface->fw(*wack, phase);
            wack_state = CHANNEL_STATE_REQ;
        }

        if (signal_interface->get_rack())
        {
            if (rack_state != CHANNEL_STATE_CLEAR)
                error_handler->error("Unexpected RACK before accepting a previous RACK");

            if (rack_queue.empty())
                error_handler->error("Unexpected RACK received. No previous AR matches it");

            rack = rack_queue.front();
            rack->ref();

            Phase phase = RACK;
            fw_interface->fw(*rack, phase);
            rack_state = CHANNEL_STATE_REQ;
        }
    }
}

}
}
}
