//
// (C) COPYRIGHT 2017 ARM Limited. ALL RIGHTS RESERVED.
//

#ifndef ARM_MODELS_SLAXI4_H
#define ARM_MODELS_SLAXI4_H

/*
 * Types which adapt from signal-level models' implementations of AXI to ARM performance modelling
 * AMBA AXI4 TLM interfaces.
 */

#include <deque>
#include <iostream>
#include <list>
#include <map>
#include <string>

#include <ARM/TLM/arm_axi4_phase.h>
#include <ARM/TLM/arm_axi4_payload.h>
#include <ARM/TLM/arm_tlm_protocol.h>

#include "SLSignalAccess.h"
#include "SLEvents.h"

namespace ARM
{
namespace Models
{
namespace SignalLevel
{

/* Error handling base class. */
class ErrorHandler
{
public:
    /*
     * Report an error and end the simulation. This can be implemented with an exception throw but
     * it is not necessary to guarantee that anything in the caller's world is in a state to be
     * inspected. If given, context is a description of the context in which message is meant.
     */
    virtual void error(const std::string& message, const std::string& context = "") = 0;
};

/*
 * Id <-> Payload mapping to allow bridging signal-level modules to keep the AXI4::Payload parent
 * relationships as transactions flow between slave and master ports.  The id here is a short
 * integer used to identify in-flight payloads. It is neither the AXI ID or the Payload AXI4::uid.
 * The id can be passed in the USER field of an AR/AW channel communication to traverse a
 * module and allow the payload to be recovered on the other side.
 */
class AXI4TransactionMapper
{
protected:
    ErrorHandler* error_handler;

    /* Counter of which id to offer for newly added payloads. */
    unsigned int current_id;

    /* Maximum id value which the mapper will assign. */
    unsigned int const max_id;

    /* Mappings in both directions. */
    std::map<unsigned int, AXI4::Payload*> id_to_payload;
    std::map<AXI4::Payload*, unsigned int> payload_to_id;

public:
    explicit AXI4TransactionMapper(unsigned int max_id_ = 0xFFFF,
        ErrorHandler* error_handler_ = NULL);

    /*
     * Add an association. This object does not ref/unref payloads as it's assumed that
     * payloads will always be live *somewhere* if a mapping is kept. Returns the new
     * id value associated with the payload (which will be unique for payloads managed
     * by this mapper).
     */
    unsigned int add_payload(AXI4::Payload* payload);

    /* Remove an association. */
    void remove_payload(AXI4::Payload* payload);

    /* Lookup an association by id. Returns NULL if no payload can be found. */
    AXI4::Payload* lookup_payload(unsigned int id);
};

/* Open a namespace for convenience so signals names can be short in the implementation. */
namespace AXI4Signal
{

/* All the distinct signals which can be part of an AXI/ACE port. */
enum AXI4SignalId
{
    /* AW channel. */
    AWID, AWADDR, AWLEN, AWSIZE, AWBURST, AWLOCK, AWCACHE, AWPROT, AWQOS, AWREGION,
        AWUSER, AWVALID, AWREADY,

    /* W channel. */
    WID, WDATA, WSTRB, WLAST, WUSER, WVALID, WREADY,

    /* B channel. */
    BID, BRESP, BUSER, BVALID, BREADY,

    /* AR channel. */
    ARID, ARADDR, ARLEN, ARSIZE, ARBURST, ARLOCK, ARCACHE, ARPROT, ARQOS, ARREGION,
        ARUSER, ARVALID, ARREADY,

    /* R channel. */
    RID, RDATA, RRESP, RLAST, RUSER, RVALID, RREADY,

    /* ACE channels. */

    /* ACE AR additions. */
    ARSNOOP, ARDOMAIN, ARBAR,

    /* ACE AW additions. */
    AWSNOOP, AWDOMAIN, AWBAR, AWUNIQUE,

    /* ACE AC channel. */
    ACVALID, ACREADY, ACADDR, ACSNOOP, ACPROT,

    /* ACE CR channel. */
    CRVALID, CRREADY, CRRESP,

    /* ACE CD channel. */
    CDVALID, CDREADY, CDDATA, CDLAST,

    /*
     * ACK RACK channel. Prefixed with SIGNAL_ so that this enum can exist in the same context as
     * AXI4::Phase
     */
    SIGNAL_RACK,

    /* ACK WACK channel. */
    SIGNAL_WACK,

    NUM_AXI_SIGNALS
};

}

/*
 * Functor to map signal ids into a name to search for when discovering AXI
 * port signal connections. This class is a base for more complicated mapping
 * schemes where the member function 'map' is replaced.
 * The default mapping assumes that signals have the same name (and case) as
 * the element of enum AXI4SignalId.
 */
class AXI4NameMapping
{
public:
    virtual ~AXI4NameMapping() {}

    /* Default names to match against: "AWADDR", "AWID" ... */
    static const char* const uppercase_names[];
    /* The same but in lower case. */
    static const char* const lowercase_names[];

    /* Mapping function to overload. */
    virtual std::string map(AXI4Signal::AXI4SignalId id)
    { return uppercase_names[id]; }

    /* Default mapping (as above) to use as a default function argument. */
    static AXI4NameMapping default_mapping;
};

/*
 * Map signal names such as AWADDR to prefixAWADDRsuffix or prefixawaddrsuffix
 * or given choice of prefix/suffix and choice of stem name case.
 */
class AXI4InfixNameMapping : public AXI4NameMapping
{
protected:
    const std::string prefix;
    const std::string suffix;
    const bool is_uppercase;

public:
    AXI4InfixNameMapping(const std::string& prefix_,
        const std::string& suffix_,
        bool is_uppercase_);

    std::string map(AXI4Signal::AXI4SignalId id);
};

/*
 * Parameters that affect the behaviour and signal widths of AXI ports.
 */
class AXI4PortParams
{
protected:
    ErrorHandler* error_handler;

    /*
     * Protocol to use on the port. This affects the treatment of WACK and RACK, whether AC, CR and
     * CD channels need be connected, and the widths of some extra signals.
     */
    TLM::Protocol protocol;

    /* Width of AW, AR and AC addresses. */
    unsigned int awaddr_width;
    unsigned int araddr_width;
    unsigned int acaddr_width;

    /* Minimum width of ARADDR and ACADDR due to DVM command packing. */
    unsigned int dvm_addr_width;

    /*
     * W, R and CD channel data widths in bits, bytes, as Size, and the number of bytes needed to
     * hold the strobe.
     */
    unsigned int data_width;
    AXI4::Size data_size;
    unsigned int data_width_in_bytes;
    unsigned int data_strobe_width_in_bytes;

    /* Width of USER signals. */
    unsigned int awuser_width;
    unsigned int wuser_width;
    unsigned int buser_width;
    unsigned int aruser_width;
    unsigned int ruser_width;

    /* Width of read/write ID signals. */
    unsigned int wid_width;
    unsigned int rid_width;

    /* Width of AWQOS, ARQOS signals. */
    unsigned int qos_width;

    /* Width of AWREGION, ARREGION signals. */
    unsigned int region_width;

    /* Width of a cache line in bits and multiples of data_width. */
    unsigned int cache_line_width;
    unsigned int cache_line_beat_count;

public:
    explicit AXI4PortParams(ErrorHandler* error_handler_ = NULL);

    /* Set the handler used to report errors. NULL means use the default handler. */
    void set_error_handler(ErrorHandler* error_handler_ = NULL);

    /* Set the protocol to use of the port. */
    void set_protocol(TLM::Protocol protocol_);

    /*
     * Set the width (in bits) of AR, AW, AC addresses. Where dvm_addr_width is
     * wider than addr_width_, dvm_addr_width will be used for araddr_width and
     * acaddr_width. Defaults to 64.
     */
    void set_addr_width(unsigned int addr_width_);

    /* Set the minimum width of AR and AC address when using DVM. Defaults to 0. */
    void set_dvm_addr_width(unsigned int dvm_addr_width_);

    /* Set individual address withs of AW, AR and AC. Defaults to 64. */
    void set_awaddr_width(unsigned int awaddr_width_);
    void set_araddr_width(unsigned int araddr_width_);
    void set_acaddr_width(unsigned int acaddr_width_);

    /* Set the width (in bits) of AR, AW, AC addresses. Defaults to 128. */
    void set_data_width(unsigned int data_width_);

    /* Set the width (in bits) of all ...USER signals. Defaults to 64. */
    void set_user_width(unsigned int user_width_);

    /* Set individual channel USER widths. */
    void set_awuser_width(unsigned int awuser_width_);
    void set_wuser_width(unsigned int wuser_width_);
    void set_buser_width(unsigned int buser_width_);
    void set_aruser_width(unsigned int aruser_width_);
    void set_ruser_width(unsigned int ruser_width_);

    /* Set the width (in bits) of all ...ID signals. Defaults to 8. */
    void set_id_width(unsigned int id_width_);

    /* Set individual channel ID widths. */
    void set_awid_width(unsigned int awid_width_);
    void set_wid_width(unsigned int wid_width_);
    void set_bid_width(unsigned int bid_width_);
    void set_arid_width(unsigned int arid_width_);
    void set_rid_width(unsigned int rid_width_);

    /* Set the width (in bits) of AWQOS and ARQOS signals. Defaults to 0. */
    void set_qos_width(unsigned int qos_width_);

    /* Set the width (in bits) of AWREGION and ARREGION signals. Defaults to 0. */
    void set_region_width(unsigned int region_width_);

    /*
     * Set the width of a cache line. This must be a multiple of data_width.
     * Defaults to 512.
     */
    void set_cache_line_width(unsigned int cache_line_width_);

    /* Simple getters. */
    TLM::Protocol get_protocol() const { return protocol; }
    unsigned int get_awaddr_width() const { return awaddr_width; }
    unsigned int get_araddr_width() const { return araddr_width; }
    unsigned int get_acaddr_width() const { return acaddr_width; }
    unsigned int get_dvm_addr_width() const { return dvm_addr_width; }
    unsigned int get_data_width() const { return data_width; }
    AXI4::Size get_data_size() const { return data_size; }
    unsigned int get_data_width_in_bytes() const { return data_width_in_bytes; }
    unsigned int get_data_strobe_width_in_bytes() const { return data_strobe_width_in_bytes; }
    unsigned int get_awuser_width() const { return awuser_width; }
    unsigned int get_wuser_width() const { return wuser_width; }
    unsigned int get_buser_width() const { return buser_width; }
    unsigned int get_aruser_width() const { return aruser_width; }
    unsigned int get_ruser_width() const { return ruser_width; }
    unsigned int get_wid_width() const { return wid_width; }
    unsigned int get_rid_width() const { return rid_width; }
    unsigned int get_qos_width() const { return qos_width; }
    unsigned int get_region_width() const { return region_width; }
    unsigned int get_cache_line_width() const { return cache_line_width; }
    unsigned int get_cache_line_beat_count() const { return cache_line_beat_count; }

    /*
     * Returns the expected width (in bits) of the given signal. A width of 0 indicates that the
     * signal should not be present for this protocol/set of parameters.
     */
    unsigned int expected_signal_width(AXI4Signal::AXI4SignalId id);

    /*
     * Safe default value for tied off signal's bytes. A tie off should present this value
     * as every byte of the signal.
     */
    uint8_t signal_default_value(AXI4Signal::AXI4SignalId id);

    /* Crop an address to the given width. */
    uint64_t crop_address(uint64_t addr, unsigned int addr_width);
};

/*
 * Abstract base class for adaption from AXI4::Payload to signal-level slave port implementations.
 * set... functions source transaction-level data into signal-level signals.
 * get... functions extract signal-level values into transaction-level data.
 */
class AXI4SlaveInterface
{
public:
    virtual ~AXI4SlaveInterface() {}

    /* Reset all driven signals. */
    void reset();

    /* AW channel */
    virtual void set_aw(AXI4::Payload& payload) = 0;
    virtual void set_awvalid(bool valid) = 0;
    virtual bool get_awready() = 0;

    /* W channel */
    virtual void set_w(AXI4::Payload& payload, unsigned int beat_index) = 0;
    virtual void set_wvalid(bool valid) = 0;
    virtual bool get_wready() = 0;

    /* B channel */
    virtual void get_b(AXI4::Payload& payload) = 0;
    virtual uint32_t get_bid() = 0;
    virtual bool get_bvalid() = 0;
    virtual void set_bready(bool ready) = 0;

    /* AR channel */
    virtual void set_ar(AXI4::Payload& payload) = 0;
    virtual void set_arvalid(bool valid) = 0;
    virtual bool get_arready() = 0;

    /* R channel */
    virtual void get_r(AXI4::Payload& payload) = 0;
    virtual uint32_t get_rid() = 0;
    virtual bool get_rlast() = 0;
    virtual uint64_t get_ruser() = 0;
    virtual bool get_rvalid() = 0;
    virtual void set_rready(bool ready) = 0;

    /* AC channel */
    /* get_ac must construct a new payload. */
    virtual AXI4::Payload* get_ac() = 0;
    virtual bool get_acvalid() = 0;
    virtual void set_acready(bool valid) = 0;

    /* CR channel */
    virtual void set_cr(AXI4::Payload& payload) = 0;
    virtual void set_crvalid(bool valid) = 0;
    virtual bool get_crready() = 0;

    /* CD channel */
    virtual void set_cd(AXI4::Payload& payload, unsigned int beat_index) = 0;
    virtual void set_cdvalid(bool valid) = 0;
    virtual bool get_cdready() = 0;

    /* WACK/RACK channels */
    virtual void set_wack(bool wack) = 0;
    virtual void set_rack(bool rack) = 0;

    /* Set the handler used to report errors. NULL means use the default handler. */
    virtual void set_error_handler(ErrorHandler* error_handler_ = NULL) = 0;
};

/*
 * A collection of SignalAccesses onto an AXI port.
 */
class AXI4PortSignalAccesses
{
protected:
    ErrorHandler* error_handler;

    /* Discovered signals of the underlying model. */
    SignalAccess* signals[AXI4Signal::NUM_AXI_SIGNALS];

    /* SignalAccess used to tie off unconnected signals. */
    std::list<SignalAccess*> tie_offs;

public:
    /* Configuration of the port's behaviour and signal binding expectations. */
    AXI4PortParams params;

protected:
    /* Check that an AxSIZE value is valid for the data channels of this port. */
    void check_size(AXI4::Size size);

    /*
     * Check the given value can fit into the given width (in bits). name is the
     * name of the field to use in any exception message.
     */
    void check_value_fits_in_width(uint64_t value, unsigned int width, const std::string& name);

public:
    explicit AXI4PortSignalAccesses(ErrorHandler* error_handler_ = NULL);
    ~AXI4PortSignalAccesses();

    /* Set the handler used to report errors. NULL means use the default handler. */
    void set_error_handler(ErrorHandler* error_handler_ = NULL);

    /*
     * Try bindings underlying signal to this interface. This function can be called multiple times
     * to bind to more than one SignalAccesses. All the discoverable signals are attempted using
     * 'mapping' to map into signal names before comparing with new_signals.
     * If debug_stream points to an ostream, debug messages are printed to that stream while
     * binding.
     */
    void bind_to_signals(SignalAccesses& new_signals,
        AXI4NameMapping& mapping = AXI4NameMapping::default_mapping,
        std::ostream* debug_stream = NULL);

    /*
     * Print a report on how completely the adapter is bound. That is, which signals remain unbound.
     */
    void print_binding_report(std::ostream& stream = std::cout);

    /*
     * Tie off all remaining signals to read as 0 and safely ignore writes.
     * If debug_stream points to an ostream, debug messages are printed to that stream while
     * binding.
     */
    void tie_off_signals(std::ostream* debug_stream = NULL);
};

/*
 * A SlaveInterface implementation which binds to SignalAccess signals in its implementation of
 * signals underlying an AXI4/ACE slave interface.
 */
class AXI4SignalAccessSlaveAdapter : public AXI4PortSignalAccesses,
    public AXI4SlaveInterface
{
protected:
    AXI4TransactionMapper* transaction_mapper;

protected:
    /* AXI4SlaveInterface interface called by the transactional port. */
    void set_aw(AXI4::Payload& payload);
    void set_awvalid(bool valid);
    bool get_awready();
    void set_w(AXI4::Payload& payload, unsigned int beat_index);
    void set_wvalid(bool valid);
    bool get_wready();
    void get_b(AXI4::Payload& payload);
    AXI4::Resp get_bresp();
    uint32_t get_bid();
    bool get_bvalid();
    void set_bready(bool ready);
    void set_ar(AXI4::Payload& payload);
    void set_arvalid(bool valid);
    bool get_arready();
    void get_r(AXI4::Payload& payload);
    uint32_t get_rid();
    bool get_rlast();
    uint64_t get_ruser();
    bool get_rvalid();
    void set_rready(bool ready);
    AXI4::Payload* get_ac();
    bool get_acvalid();
    void set_acready(bool valid);
    void set_cr(AXI4::Payload& payload);
    void set_crvalid(bool valid);
    bool get_crready();
    void set_cd(AXI4::Payload& payload, unsigned int beat_index);
    void set_cdvalid(bool valid);
    bool get_cdready();
    void set_wack(bool wack);
    void set_rack(bool rack);

public:
    explicit AXI4SignalAccessSlaveAdapter(ErrorHandler* error_handler_ = NULL);

    /* Set the transaction mapper for linking parent payloads to generated. */
    void set_transaction_mapper(AXI4TransactionMapper& transaction_mapper_);

    /* Set the handler used to report errors. NULL means use the default handler. */
    void set_error_handler(ErrorHandler* error_handler_ = NULL);
};

/*
 * State of individual, clocked, READY/VALID-handshaking channels.
 * Channels state in state CLEAR and progress using the state machine:
 *
 *     ,-.                     ,-.
 *     | |    @negedge         | |   @negedge
 *     | v    VALID & !READY   | v   READY
 *    CLEAR -----------------> REQ -----------> ACK
 *     ^ |                                      ^ |
 *     | |    @negedge                          | |
 *     | |    VALID & READY                     | |
 *     | `--------------------------------------' |
 *     |                                          |
 *     |      @posedge                            |
 *     `------------------------------------------'
 *
 * For transitions from CLEAR, READY is signalled by an immediate phase update.
 * The transition from REQ to ACK is a signalled, possibly delayed, READY from
 * the receiving end.
 */
enum AXI4ChannelState
{
    /* No communication this cycle (yet). */
    CHANNEL_STATE_CLEAR,
    /* VALID signalled but not immediately acknowledged. */
    CHANNEL_STATE_REQ,
    /* VALID & READY but not yet reset to CLEAR by posedge of the clock. */
    CHANNEL_STATE_ACK
};

/*
 * AXI4 slave using AXI4SlaveInterface for its fw implementation.
 */
class AXI4Slave
{
public:
    class AXI4BwInterface
    {
    public:
        virtual bool bw(AXI4::Payload& payload, AXI4::Phase& phase) = 0;
    };

protected:
    const std::string name;
    const TLM::Protocol protocol;

    ErrorHandler* error_handler;

    /* Offered backwards interface for TLM upstream communication. */
    AXI4BwInterface* bw_interface;

    /* Offered interface onto signals. */
    AXI4SlaveInterface* signal_interface;

    /*
     * State of handshaking on each channel from CLEAR (!VALID & !READY) to REQ (VALID & !READY) to
     * ACK (VALID & READY). Transtions from CLEAR happen in negedge in response to communications
     * (either from TLM or by sampling signals). Transitions back to CLEAR (from ACK) happen in
     * posedge but signals are not reset until the start of the following cycle's negedge.
     */
    AXI4ChannelState aw_state;
    AXI4ChannelState w_state;
    AXI4ChannelState b_state;
    AXI4ChannelState ar_state;
    AXI4ChannelState r_state;
    AXI4ChannelState ac_state;
    AXI4ChannelState cr_state;
    AXI4ChannelState cd_state;
    /* WACK and RACK only transition from CLEAR to REQ and back. */
    AXI4ChannelState wack_state;
    AXI4ChannelState rack_state;

    typedef std::deque<AXI4::Payload*> TransactionQueue;
    typedef std::map<unsigned int, TransactionQueue> TransactionQueuesById;

    /*
     * Payloads 'on the wires' between CHANNEL_STATE_REQ through CHANNEL_STATE_ACK to
     * CHANNEL_STATE_CLEAR at posedge.
     */
    AXI4::Payload* aw;
    AXI4::Payload* w;
    AXI4::Payload* b;
    AXI4::Payload* ar;
    AXI4::Payload* r;
    AXI4::Payload* ac;
    AXI4::Payload* cr;
    AXI4::Payload* cd;
    AXI4::Payload* wack;
    AXI4::Payload* rack;

    /*
     * Transaction tracking to allow returning transactions from the signal-level model to be mapped
     * back into transaction objects by using AXI ordering.
     */
    TransactionQueuesById b_queue;
    TransactionQueuesById r_queue;

    /* RLAST value from negedge (where RVALID was high) to posedge. */
    bool rlast;

    /* Number of beat seen on W. This rolls back to 0 after the last beat of W. */
    unsigned int w_beat_count;

    /* Number of beat seen on CD. This rolls back to 0 after the last beat of CD. */
    unsigned int cd_beat_count;

    /* Event at start of fw before signal assignments. */
    InputEvent *input_event;
    /* Also signal an input event on clock edges? */
    bool signal_input_event_at_clock_edges;
    /* Event after setting VALID, before sampling READY to allow for combinatorial updates. */
    InputEvent *handshake_event;
    /*
     * Insert extra evaluation and immediate checking to allow combinatorial paths from from
     * ARVALID to RVALID and RREADY to ARREADY to allow same cycle read responses from AR.
     */
    bool allow_same_cycle_reads;
      
    /* The operation to match an R to an AR in negedge and send bw. */
    void negedge_r_bw();

public:
    AXI4Slave(const std::string& name_, TLM::Protocol protocol_);

    /* Set the interface to use for the underlying set_aw... calls. */
    void set_signal_interface(AXI4SlaveInterface& signal_interface_);

    /* Set the interface to use for bw TLM calls. */
    void set_bw_interface(AXI4BwInterface& bw_interface_);

    /* Set the handler used to report errors. NULL means use the default handler. */
    virtual void set_error_handler(ErrorHandler* error_handler_ = NULL);

    /* Set the event to call at the start of fw. */
    void set_input_event(InputEvent& input_event_);
    /* Enable/disable calling input_event on clock edges. */
    void enable_clock_event(bool enable = true);
    /*
     * Enable/disable read responses in the same cycle as a read request. Enabling this will incur
     * more hanshake events to update the model between all signal sets and gets which might
     * conceivably be combinatorial linked.
     */
    void enable_same_cycle_reads(bool enable = true);
    /* Set the event to call between a received VALID in fw and sampling READY. */
    void set_handshake_event(InputEvent& handshake_event_);

    /* TLM fw call from environment. */
    bool fw(AXI4::Payload& payload, AXI4::Phase& phase);
    

    /* Clock activity from environment. */
    void posedge();
    void negedge();

};

/*
 * Abstract base class for adaption from AXI4::Payload to signal-level master port implementations.
 * set... functions source transaction-level data into signal-level signals.
 * get... functions extract signal-level values into transaction-level data.
 */
class AXI4MasterInterface
{
public:
    virtual ~AXI4MasterInterface() {}

    /* Reset all driven signals. */
    void reset();

    /* AW channel */
    /* get_aw must construct a new payload. */
    virtual AXI4::Payload* get_aw() = 0;
    virtual bool get_awvalid() = 0;
    virtual void set_awready(bool valid) = 0;

    /* W channel */
    virtual void get_w(AXI4::Payload& payload) = 0;
    virtual uint32_t get_wid() = 0;
    virtual bool get_wlast() = 0;
    virtual uint64_t get_wuser() = 0;
    virtual bool get_wvalid() = 0;
    virtual void set_wready(bool ready) = 0;

    /* B channel */
    virtual void set_bid(uint32_t id) = 0;
    virtual void set_bresp(AXI4::Resp resp) = 0;
    virtual void set_bvalid(bool valid) = 0;
    virtual bool get_bready() = 0;

    /* AR channel */
    /* get_ar must construct a new payload. */
    virtual AXI4::Payload* get_ar() = 0;
    virtual bool get_arvalid() = 0;
    virtual void set_arready(bool ready) = 0;

    /* R channel */
    virtual void set_r(AXI4::Payload& payload, unsigned int beat_index) = 0;
    virtual void set_rvalid(bool valid) = 0;
    virtual bool get_rready() = 0;

    /* AC channel */
    virtual void set_ac(AXI4::Payload& payload) = 0;
    virtual void set_acvalid(bool valid) = 0;
    virtual bool get_acready() = 0;

    /* CR channel */
    virtual void get_cr(AXI4::Payload& payload) = 0;
    virtual bool get_crvalid() = 0;
    virtual void set_crready(bool ready) = 0;

    /* CD channel */
    virtual void get_cd(AXI4::Payload& payload) = 0;
    virtual bool get_cdlast() = 0;
    virtual bool get_cdvalid() = 0;
    virtual void set_cdready(bool ready) = 0;

    /* WACK/RACK channels */
    virtual bool get_wack() = 0;
    virtual bool get_rack() = 0;

    /* Set the handler used to report errors. NULL means use the default handler. */
    virtual void set_error_handler(ErrorHandler* error_handler_ = NULL) = 0;
};

/*
 * A MasterInterface implementation which binds to SignalAccess signals in its implementation of
 * signals underlying an AXI4/ACE slave interface.
 */
class AXI4SignalAccessMasterAdapter : public AXI4PortSignalAccesses,
    public AXI4MasterInterface
{
protected:
    AXI4TransactionMapper* transaction_mapper;

protected:
    /* AXI4MasterInterface interface called by the transactional port. */
    AXI4::Payload* get_aw();
    bool get_awvalid();
    void set_awready(bool valid);
    void get_w(AXI4::Payload& payload);
    uint32_t get_wid();
    bool get_wlast();
    uint64_t get_wuser();
    bool get_wvalid();
    void set_wready(bool ready);
    void set_bid(uint32_t id);
    void set_bresp(AXI4::Resp resp);
    void set_bvalid(bool valid);
    bool get_bready();
    AXI4::Payload* get_ar();
    bool get_arvalid();
    void set_arready(bool ready);
    void set_r(AXI4::Payload& payload, unsigned int beat_index);
    void set_rvalid(bool valid);
    bool get_rready();
    void set_ac(AXI4::Payload& payload);
    void set_acvalid(bool valid);
    bool get_acready();
    void get_cr(AXI4::Payload& payload);
    bool get_crvalid();
    void set_crready(bool ready);
    void get_cd(AXI4::Payload& payload);
    bool get_cdlast();
    bool get_cdvalid();
    void set_cdready(bool ready);
    bool get_wack();
    bool get_rack();

public:
    explicit AXI4SignalAccessMasterAdapter(ErrorHandler* error_handler_ = NULL);

    /* Set the transaction mapper for linking parent payloads to generated. */
    void set_transaction_mapper(AXI4TransactionMapper& transaction_mapper_);

    /* Set the handler used to report errors. NULL means use the default handler. */
    void set_error_handler(ErrorHandler* error_handler_ = NULL);
};

/*
 * AXI4 master using AXI4MasterInterface for its bw inplementation.
 */
class AXI4Master
{
public:
    class AXI4FwInterface
    {
    public:
        virtual bool fw(AXI4::Payload& payload, AXI4::Phase& phase) = 0;
    };

protected:
    const std::string name;
    const TLM::Protocol protocol;

    ErrorHandler* error_handler;

    /* Offered forwards interface for TLM downstream communication. */
    AXI4FwInterface* fw_interface;

    /* Offered interface onto signals. */
    AXI4MasterInterface* signal_interface;

    /*
     * State of handshaking on each channel from CLEAR (!VALID & !READY) to REQ (VALID & !READY) to
     * ACK (VALID & READY). Transtions from CLEAR happen in negedge in response to communications
     * (either from TLM or by sampling signals). Transitions back to CLEAR (from ACK) happen in
     * posedge but signals are not reset until the start of the following cycle's negedge.
     */
    AXI4ChannelState aw_state;
    AXI4ChannelState w_state;
    AXI4ChannelState b_state;
    AXI4ChannelState ar_state;
    AXI4ChannelState r_state;
    AXI4ChannelState ac_state;
    AXI4ChannelState cr_state;
    AXI4ChannelState cd_state;
    /* WACK and RACK only transition from CLEAR to REQ and back. */
    AXI4ChannelState wack_state;
    AXI4ChannelState rack_state;

    typedef std::deque<AXI4::Payload*> TransactionQueue;
    typedef std::map<unsigned int, TransactionQueue> TransactionQueuesById;

    /*
     * Payloads 'on the wires' between CHANNEL_STATE_REQ through CHANNEL_STATE_ACK to
     * CHANNEL_STATE_CLEAR at posedge.
     */
    AXI4::Payload* aw;
    AXI4::Payload* w;
    AXI4::Payload* b;
    AXI4::Payload* ar;
    AXI4::Payload* r;
    AXI4::Payload* ac;
    AXI4::Payload* cr;
    AXI4::Payload* cd;
    AXI4::Payload* wack;
    AXI4::Payload* rack;

    /*
     * Transaction tracking to allow returning transactions from the signal-level model to be mapped
     * back into transaction objects by using AXI ordering.
     */
    TransactionQueue w_queue;
    TransactionQueue cr_queue;
    TransactionQueue cd_queue;
    TransactionQueue wack_queue;
    TransactionQueue rack_queue;

    /* WLAST value from negedge (where WVALID was high) to posedge. */
    bool wlast;

    /* CDLAST value from negedge (where CDVALID was high) to posedge. */
    bool cdlast;

    /* Event at start of bw before signal assignments. */
    InputEvent *input_event;
    /* Also signal an input event on clock edges? */
    bool signal_input_event_at_clock_edges;
    /* Event after setting VALID, before sampling READY to allow for combinatorial updates. */
    InputEvent *handshake_event;
    /*
     * Insert extra evaluation and immediate checking to allow combinatorial paths from from
     * ARVALID to RVALID and RREADY to ARREADY to allow same cycle read responses from AR.
     */
    bool allow_same_cycle_reads;

    /*
     * Beat counts for each read payload in flight.
     * beat_counts could as easily have been a map<rid<deque<count>> structure.
     */
    typedef std::map<uint64_t, unsigned int> BeatCountMap;
    BeatCountMap read_beat_counts;

public:
    AXI4Master(const std::string& name_, TLM::Protocol protocol_);

    /* Set the interface to use for the underlying set_aw... calls. */
    void set_signal_interface(AXI4MasterInterface& signal_interface_);

    /* Set the interface to use for fw TLM calls. */
    void set_fw_interface(AXI4FwInterface& fw_interface_);

    /* Set the handler used to report errors. NULL means use the default handler. */
    virtual void set_error_handler(ErrorHandler* error_handler_ = NULL);

    /* Set the event to call at the start of bw. */
    void set_input_event(InputEvent& input_event_);
    /* Enable/disable calling input_event on clock edges. */
    void enable_clock_event(bool enable = true);
    /*
     * Enable/disable read responses in the same cycle as a read request. Enabling this will incur
     * more hanshake events to update the model between all signal sets and gets which might
     * conceivably be combinatorial linked.
     */
    void enable_same_cycle_reads(bool enable = true);
    /* Set the event to call between a received VALID in bw and sampling READY. */
    void set_handshake_event(InputEvent& handshake_event_);

    /* TLM bw call from environment. */
    bool bw(AXI4::Payload& payload, AXI4::Phase& phase);
    
    /* Clock activity from environment. */
    void posedge();
    void negedge();
};

}
}
}

#endif // ARM_MODELS_SLAXI4_H
