//
// (C) COPYRIGHT 2017 ARM Limited. ALL RIGHTS RESERVED.
//

#ifndef ARM_MODELS_SLSIGNALACCESS_H
#define ARM_MODELS_SLSIGNALACCESS_H

#include <stdint.h>
#include <map>
#include <string>
#include <cstring>

/*
 * Classes to allow access to signals in signal-level models through a common SignalAccess
 * interface.  SLSignalAccess.h is intended to be the only part of the SL/SignalLevel infrastructure
 * which needs to be built into a model (or a model's wrapper).
 */

namespace ARM
{
namespace Models
{
namespace SignalLevel
{

/*
 * Abstract base class for get/set access onto a signal from a signal-level model.  The model must
 * override {get,set}_ptr to copy out/in data and get_width(). All other {set,get}... member
 * functions are syntactic sugar on top of {get,set}_ptr.
 */
class SignalAccess
{
protected:
    /* Signal width in bits. */
    const unsigned int width;

    /* Space to use when shuffling between ints and get_ptr/set_ptr. */
    uint8_t* temp_data;

public:
    SignalAccess(unsigned int width_) : width(width_)
    { temp_data = new uint8_t[bits_to_bytes(width)]; }

    virtual ~SignalAccess() { delete[] temp_data; }

    /* Core functions to override. */

    /*
     * Get data from a signal. 'value' must point to a pre-allocated piece of memory large enough to
     * hold the whole width of the signal.  Data will be returned in byte-wise little-endian fashion
     * with the 8 least-significant bits of the signal's value being returned in value[0] and next 8
     * least-significant bits in value[1] and so on. Any remaining bits above the most-significant
     * bit packed into the last byte will be set to 0.
     */
    virtual void get_ptr(uint8_t* value) const = 0;

    /*
     * Set data on a signal. 'value' must point to a pre-initialised piece of memory at least as
     * large as the signal's data. Data is set in byte-wise little-endian fashion with value[0]
     * setting the least-significant 8 bits of the signal and value[1] setting the next
     * least-significant 8 bits and so on.  Any remaining bits above the most-significant bit packed
     * into the last byte of value must be set to 0.
     */
    virtual void set_ptr(uint8_t* value) = 0;

    /* Other functions. */

    /*
     * Get the width (in bits) of the signal. As get and set only work in multiples of a byte, in
     * some circumstances it may be acceptable to return a width which is only correct to the
     * byte above the actual bit width.  bits_to_bytes(get_width()) can be used to find the size
     * required for the buffers used with get_ptr and set_ptr.
     */
    unsigned int get_width() const { return width; }

    /*
     * Returns the rounded up number of bytes needs to store the given number of bits.
     * Returns 0 for 0 bits.
     */
    static unsigned int bits_to_bytes(unsigned int bits) { return (bits + (8 - 1)) / 8; }

    /* Helper functions calling {set,get}_ptr. */

    /*
     * Convenience functions mapping {get,set}_ptr onto integer types.
     * This assumes little endian data.
     */
    void set64(uint64_t value)
    {
        if (width <= 64)
        {
            set_ptr(reinterpret_cast<uint8_t*>(&value));
        } else
        {
            std::memset(temp_data, 0, bits_to_bytes(width));
            std::memcpy(temp_data, &value, sizeof(uint64_t));
            set_ptr(temp_data);
        }
    }

    void set(bool value) { set64(value); }
    void set(uint8_t value) { set64(value); }
    void set(uint16_t value) { set64(value); }
    void set(uint32_t value) { set64(value); }
    void set(uint64_t value) { set64(value); }
    void set(int8_t value) { set64(value); }
    void set(int16_t value) { set64(value); }
    void set(int32_t value) { set64(value); }
    void set(int64_t value) { set64(value); }

    uint8_t get8() { return get64(); }
    uint16_t get16() { return get64(); }
    uint32_t get32() { return get64(); }

    uint64_t get64()
    {
        uint64_t value = 0;

        if (width <= 64)
        {
            get_ptr(reinterpret_cast<uint8_t*>(&value));
        } else
        {
            get_ptr(temp_data);
            std::memcpy(&value, temp_data, sizeof(uint64_t));
        }

        return value;
    }
};

/* A simple persistent signal that can be read/written. */
class SimpleSignal : public SignalAccess
{
protected:
    /* When read_only is set, the signal's value can not be changed. */
    bool read_only;

    /* width as bytes. Allocated size of data. */
    unsigned int bytes;

    /* Value of the signal. */
    uint8_t* data;

public:
    SimpleSignal(unsigned int width_, bool read_only_ = false) :
        SignalAccess(width_),
        read_only(false),
        bytes((width_ + (8 - 1)) / 8),
        data(new uint8_t[bytes])
    {
        set(0);
        set_read_only(read_only_);
    }

    /* Set/reset the read_only flag. */
    void set_read_only(bool read_only_ = true) { read_only = read_only_; }

    void set_ptr(uint8_t* value) { if (!read_only) std::memcpy(data, value, bytes); }
    void get_ptr(uint8_t* value) const { std::memcpy(value, data, bytes); }
};

/* Base class for keepers of named SignalAccesses. */
class SignalAccesses
{
public:
    virtual ~SignalAccesses() {}

    /* Find a SignalAccess by name. Returns NULL on failure. */
    virtual SignalAccess* find_signal(const std::string& name) = 0;
};

/* A collection of SignalAccesses held in a map from signal name to SignalAccesses. */
class SignalAccessMap : public SignalAccesses
{
protected:
    typedef std::map<std::string, SignalAccess*> SignalMap;
    SignalMap signals;

public:
    /* Find a signal by name. Returns NULL on failure. */
    SignalAccess* find_signal(const std::string& name)
    {
        SignalMap::iterator found = signals.find(name);
        return (found == signals.end() ? NULL : found->second);
    }

    /* Add a signal to the set. */
    void add_signal(const std::string& name, SignalAccess* signal)
    { signals[name] = signal; }
};

}
}
}

#endif // ARM_MODELS_SLSIGNALACCESS_H
