//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// This file forms part of the ARM standard for performance model
// memory transactions for SystemC TLM 2.0 and should not be
// modified.
//
// This standard is currently under review and is subject to change.
//-------------------------------------------------------------------

#ifndef ARM_TLM_HELPERS_H
#define ARM_TLM_HELPERS_H

#include <stdint.h>

namespace ARM
{
namespace TLM
{

/**
 * Wrapper for enumeration values to allow their size (as object data members)
 * to be controlled and to provide a base for attaching functions to enumeration
 * values. Enum-typed values can be extracted from an EnumWrapper with an
 * implicit cast.
 */
template <typename Enum, typename Int = uint32_t>
class EnumWrapper
{
protected:
    /**
     * The wrapped enumeration value held as a raw integer to limit its
     * allocated size.
     */
    Int value;

public:
    /**
     * Default construction with initialized value of 0. It is assumed that 0
     * will be a sensible default value in type Enum
     */
    EnumWrapper() : value(0) {}

    /** Construction from an Enum value. */
    EnumWrapper(Enum value_) : value(value_) {}

    /** Construction from a raw integer. */
    EnumWrapper(Int value_) : value(value_) {}

    /** Implicit cast to the wrapped enumeration type. */
    operator Enum () const { return static_cast<Enum>(value); }

    /** Assignment from an Enum value. */
    EnumWrapper& operator= (Enum value_) { value = value_; return *this; }
};

/**
 * BitEnumWrapper wraps a value of an enumeration type which can be
 * bitwise-composed from elements of another enumeration type. Elements of the
 * BitEnum type can be bitwise or-ed (with operator|) into a BitEnumWrapper and
 * a composed Enum value can be extracted from the wrapper using an implicit
 * cast.
 *
 * For example:
 *
 * enum Flags { NA_NB = 0, NA_B = 1, A_NB = 2, A_B = 3 };
 *
 * Flags is effectively a 2 bit record of elements A and B.
 * BitEnumWrapper can express this with:
 *
 * enum FlagsBits { A = 2, B = 1 };
 * typedef BitEnumWrapper<Flags, FlagsBits> WrappedFlags.
 *
 * The value A_B of Flags could then be constructed using:
 *
 * WrappedFlags() | A | B
 */
template <typename Enum, typename BitEnum, typename Int = uint32_t>
class BitEnumWrapper
{
protected:
    /**
     * The wrapped enumeration value held as a raw integer to limit its
     * allocated size.
     */
    Int value;

public:
    /**
     * Default construction with initialized value of 0. It is assumed that 0
     * will be a sensible default value in type Enum
     */
    BitEnumWrapper() : value(0) {}

    /** Construction from an Enum value. */
    BitEnumWrapper(Enum value_) : value(value_) {}

    /** Construction from a raw integer value. */
    BitEnumWrapper(Int value_) : value(value_) {}

    /** Implicit cast to the wrapped enumeration type. */
    operator Enum () const { return static_cast<Enum>(value); }

    /** Assignment from an Enum value. */
    BitEnumWrapper& operator= (Enum value_) { value = value_; return *this; }

    /** Bitor in a bitfield from the BitOr type. */
    BitEnumWrapper& operator|= (BitEnum rhs) { value |= rhs; return *this; }

    /**
     * Make a new wrapper from the bitwise or-ed combination of this wrapper and
     * a BitEnum.
     */
    BitEnumWrapper operator| (BitEnum rhs) const
    {
        return BitEnumWrapper<Enum, BitEnum, Int>(
            static_cast<Enum>(value | rhs));
    }

    /**
     * Does this wrapper's value have all the bits set that are set in the
     * given BitEnum. Useful for testing BitEnum flag bit values.
     */
    bool has(BitEnum rhs) { return value & rhs; }

    /** Make a new wrapper with only the bits covered by the given BitEnum. */
    BitEnumWrapper mask(BitEnum rhs) { return value & rhs; }
};

}
}

#endif // ARM_TLM_HELPERS_H
