/*
 * Arm SystemC Cycle Models: scx_schedule_channel.h - Arm SystemC Cycle Models schedule channel class and related definitions.
 *
 * Copyright 2019 Arm Limited (or its affiliates). All rights reserved.
 */

#ifndef SCX_SCHEDULE_CHANNEL_H
#define SCX_SCHEDULE_CHANNEL_H

/**
 * @file scx_schedule_channel.h
 *
 * Arm SystemC Cycle Models schedule channel class and related definitions.
 */

#include <systemc>
#ifndef __carbon_capi_h_
#include "carbon/carbon_capi.h"
#endif

/**
* @brief Arm SystemC Cycle Models schedule channel class.
* @param COMP component model.
*
* This schedule channel class includes member functions related to scheduling the compiled model:
* - request_schedule()
* - schedule()
* - update_outputs()
* scx_schedule_channel ensures to call carbonSchedule() once per consecutive delta cycles.
* This is required when integrating with mixed-language simulators, as those might insert delta cycles at the mixed-language
* boundary, especially around clocks.
*/
template<class COMP>
#ifndef CM_SCX_OLD_SCHEDULE
class scx_schedule_channel: public sc_core::sc_module
#else
class scx_schedule_channel: public sc_core::sc_prim_channel
#endif
{
public:
#ifndef CM_SCX_OLD_SCHEDULE
    SC_HAS_PROCESS(scx_schedule_channel);
#endif

    /**
     * @brief Constructor.
     * @param name_ SystemC module name of the compiled model.
     */
#ifndef CM_SCX_OLD_SCHEDULE
    scx_schedule_channel(const sc_core::sc_module_name & name_):
        sc_core::sc_module(name_),
#else
    explicit scx_schedule_channel(const char * name_):
        sc_core::sc_prim_channel(name_),
#endif
        m_component(nullptr),
        m_object(nullptr),
#ifndef CM_SCX_OLD_SCHEDULE
        m_carbon_schedule_event("m_carbon_schedule_event"),
        m_deposit_count(0),
#else
        m_carbon_write_outputs_event("m_carbon_write_outputs_event"),
#endif
        m_prev_time_stamp(sc_core::sc_max_time())
    {
#ifndef CM_SCX_OLD_SCHEDULE
        SC_METHOD(schedule);
        sensitive << m_carbon_schedule_event;
#endif
    }

    /**
     * @brief Sets the component model for use by this schedule channel.
     */
    void set_component(COMP *comp_)
    {
        m_component = comp_;
    }

    /**
     * @brief Sets the compiled model for use by this schedule channel.
     */
    void set_object(CarbonObjectID *obj_)
    {
        m_object = obj_;
    }

    /**
     * @brief Requests a schedule of the model.
     *
     * request_schedule() is typically called when a model input signal changed.
     * The schedule will happen at the next delta cycle at the earliest,
     * unless additional calls to request_schedule() happened before.
     */
    void request_schedule()
    {
#ifndef CM_SCX_OLD_SCHEDULE
        m_deposit_count ++;
        m_carbon_schedule_event.notify(sc_core::SC_ZERO_TIME);
#else
        request_update();
#endif
    }
#ifdef CM_SCX_OLD_SCHEDULE

    /**
     * @brief Returns reference to event notified when output pins on the model can be written.
     *
     */
    const sc_core::sc_event & write_outputs_event() const
    {
        return m_carbon_write_outputs_event;
    }
protected:

    /**
     * @brief Implement @c sc_core::sc_prim_channel::update() function.
     *
     * This version of the function calls schedule().
     *
     */
    void update() override
    {
        schedule();
    }
#endif
private:
    COMP             *m_component;
    CarbonObjectID   *m_object;
#ifndef CM_SCX_OLD_SCHEDULE
    sc_core::sc_event m_carbon_schedule_event;
    sc_dt::uint64     m_deposit_count;
#else
    sc_core::sc_event m_carbon_write_outputs_event;
#endif
    sc_core::sc_time  m_prev_time_stamp;

    /**
     * @brief Performs a schedule of the model.
     *
     * schedule() calls carbonSchedule(), unless there is pending activity at current time and
     * additional calls to request_schedule() happened.
     *
     * @see schedule_event()
     */
    void schedule()
    {
#ifndef CM_SCX_OLD_SCHEDULE
        if (sc_core::sc_pending_activity_at_current_time() && (m_deposit_count > 0))
        {
            m_deposit_count = 0;
            m_carbon_schedule_event.notify(sc_core::SC_ZERO_TIME);
            return;
        }
#endif
        sc_assert(m_component != nullptr && m_object != nullptr);
#ifdef CM_SCX_TRACE
        std::cout << "CM_SCX_TRACE: " << m_component->name() << ": ["
            << sc_core::sc_time_stamp() << "][" << sc_core::sc_delta_count()
            << "] calling carbonSchedule()\n";
#endif
        carbonSchedule(m_object, sc_core::sc_time_stamp().value());
#ifdef CM_SCX_OLD_SCHEDULE
        m_carbon_write_outputs_event.notify(sc_core::SC_ZERO_TIME);
#endif

        /* Make sure scx_wrapper_feature::update() is only called once per simulation time.
         * See DSGCM-5206. */
        auto new_time_stamp = sc_core::sc_time_stamp();

        if (new_time_stamp != m_prev_time_stamp)
        {
            for (unsigned int i = 0; (i < m_component->mFeatures.size()); i ++)
                  if (m_component->mFeatures[i]->isEnabled())
                      m_component->mFeatures[i]->update();
            m_prev_time_stamp = new_time_stamp;
        }
#ifndef CM_SCX_OLD_SCHEDULE
        m_component->write_outputs();
        m_deposit_count = 0;
#endif
    }
};

#endif  /* SCX_SCHEDULE_CHANNEL_H */
