/**
 * @file        scx_signal_sizer.h
 * @brief       Class scx_signal_sizer definition.
 * @copyright   Copyright 2016-2017 ARM Limited. All rights reserved.
 *
 * This file is not part of a stable public API and may change without notice.
 */

#ifndef SCX_SIGNAL_SIZER_H
#define SCX_SIGNAL_SIZER_H

#include <systemc>

/**
 * @brief       SCX namespace.
 */
namespace scx {

/**
 * @brief       Channel intended to model behaviour of wire with down/up sizing.
 *
 * Class scx_signal_sizer is a channel intended to model the behavior of a single
 * piece of wire carrying a digital electronic signal, with additional down or
 * up sizing capability, based on input and output types.
 *
 * @note        No warnings are generated when bits are being squashed, when input
 *              and output are integer or logic types.
 *
 * @param       IN specifies input type for which predefined semantics for assignment
 *              and equality are adequate, that is C++ type or SystemC data type
 * @param       OUT specifies output type for which predefined semantics for assignment
 *              and equality are adequate, that is C++ type or SystemC data type
 * @param       WRITER_POLICY specifies whether one or many writers are allowed.
 *
 * @note        @a IN and @a OUT types are required to provide adequate semantices
 *              for conversion from one to the other.
 *
 * @see         @c sc_core::sc_signal
 */
template<typename IN, typename OUT, sc_core::sc_writer_policy POL = sc_core::SC_ONE_WRITER>
class scx_signal_sizer : public sc_core::sc_prim_channel {
public:

    /* Typedefs - convenience */
    typedef scx_signal_sizer<IN, OUT, POL> this_type;

    /**
     * Default constructor.
     */
    scx_signal_sizer():
        sc_core::sc_prim_channel{sc_core::sc_gen_unique_name("scx_signal_sizer")},
        m_cur_val_if{this},
        m_new_val_if{this},
        m_value_changed_event_p{nullptr},
        m_posedge_event_p{nullptr},
        m_negedge_event_p{nullptr},
        m_delta_count{~sc_dt::UINT64_ONE},
        m_input{nullptr} {
    }

    /**
     * Constructor.
     *
     * @param   name name of this channel instance
     */
    explicit
    scx_signal_sizer(const char * name):
        sc_core::sc_prim_channel{name},
        m_cur_val_if{this},
        m_new_val_if{this},
        m_value_changed_event_p{nullptr},
        m_posedge_event_p{nullptr},
        m_negedge_event_p{nullptr},
        m_delta_count{~sc_dt::UINT64_ONE},
        m_input{nullptr} {
    }

    /**
     * Constructor.
     *
     * @param   name name of this channel instance
     * @param   initial_value initial value of the signal
     */
    scx_signal_sizer(const char * name, const IN & initial_value):
        sc_core::sc_prim_channel{name},
        m_cur_val_if{this},
        m_new_val_if{this},
        m_value_changed_event_p{nullptr},
        m_posedge_event_p{nullptr},
        m_negedge_event_p{nullptr},
        m_delta_count{~sc_dt::UINT64_ONE},
        m_input{nullptr} {
        m_new_val_if.m_val = initial_value;
        m_cur_val_if.m_val = OUT(initial_value);
    }

    /**
     * Destructor.
     */
    virtual
    ~scx_signal_sizer() {
        delete m_value_changed_event_p;
        delete m_posedge_event_p;
        delete m_negedge_event_p;
    }

    /**
     * Returns string @c "scx_signal_sizer".
     */
    virtual const char*
    kind() const override {
        return "scx_signal_sizer";
    }

    /**
     * Modify value of signal sizer such that signal sizer appears to have new value (as returned by read())
     * in next delta cycle but not before then.
     *
     *  @param  value new value.
     *
     *  @see    operator=()
     */
    void
    write(const IN & value) {
        if (! check_write()) {
            return;
        }
        m_new_val_if.m_val = value;
        if (m_cur_val_if.read() != OUT(value)) {
            request_update();
        }
    }

    /**
     * Modify value of signal sizer such that signal sizer appears to have new value (as returned by read())
     * in next delta cycle but not before then.
     *
     *  @param  value new value.
     *
     *  @see    write()
     */
    this_type &
    operator =(const IN & value) {
        write(value);
        return (* this);
    }

    /**
     * Modify value of signal sizer such that signal sizer appears to have new value (as returned by read())
     * in next delta cycle but not before then.
     *
     *  @param  value new value.
     *
     *  @see    write()
     */
    this_type &
    operator =(this_type & a) {
        write(a.read());
        return (* this);
    }

    /**
     * Return reference to current value of signal sizer.
     */
    const OUT &
    read() const {
        return m_cur_val_if.read();
    }

    /**
     * Return reference to current value of signal sizer.
     *
     * @see     read()
     */
    operator const OUT & () const {
        return (read());
    }

    /**
     *  Print current value of signal sizer to stream passed as argument
     *
     *  @param    os output stream, default to @c std::cout.
     *
     *  @see    dump()
     */
    virtual void
    print(std::ostream &os = std::cout) const override {
        os << m_cur_val_if.read();
    }

    /**
     * Print diagnostic information to help user debug application to stream
     * passed as argument
     *
     * @param    os output stream, default to @c std::cout.
     *
     *  @see    print()
     */
    virtual void
    dump(std::ostream &os = std::cout) const override {
        os << "     name = " << name() << std::endl;
        os << "    value = " << m_cur_val_if.read() << std::endl;
        os << "new value = " << m_new_val_if.read() << std::endl;
    }

    /**
     * Return reference to signal sizer input interface.
     * This reference can be used for binding this signal sizer.
     */
    operator sc_core::sc_signal_inout_if<IN> & () {
        return m_new_val_if;
    }

    /**
     * Return reference to signal sizer output interface,
     * This reference can be used for binding this signal sizer.
     */
    operator sc_core::sc_signal_inout_if<OUT> & () {
        return m_cur_val_if;
    }

    /**
     * Return @c true when value of signal sizer was updated in preceding delta cycle and at current simulation time.
     */
    bool
    event() const {
        return ((m_delta_count == sc_core::sc_delta_count()) && (m_time_stamp == sc_core::sc_time_stamp()));
    }
private:

    /* Helpers */
    virtual void
    update() override {
        m_cur_val_if.m_val = OUT(m_new_val_if.read());
        m_delta_count = sc_core::sc_delta_count();
        m_time_stamp = sc_core::sc_time_stamp();
        if (m_value_changed_event_p) {
            m_value_changed_event_p->notify(sc_core::SC_ZERO_TIME);
        }
    }
    bool
    check_write() {
        if (POL == sc_core::SC_ONE_WRITER) {
            sc_process_handle writer = sc_core::sc_get_current_process_handle();

            if (! m_writer.valid()) {
                m_writer.swap(writer);
            } else if (writer.valid() && (m_writer != writer)) {
                report_more_than_one_writer(m_writer.get_process_object(), writer.get_process_object());
                return false;
            }
        }
        return true;
    }
    void
    check_port(sc_core::sc_port_base & port, bool is_input) {
        if ((POL == sc_core::SC_ONE_WRITER) && is_input) {
            if (m_input != nullptr) {
                report_more_than_one_writer(m_input, & port);
            }
            m_input = & port;
        }
    }
    void
    report_attempt_to_write() {
        std::stringstream msg;

        msg << "\n signal sizer "
               "`" << name() << "' "
              " (" << kind() << ")"
            << "\n incorrect template parameters specification?";
        SC_REPORT_ERROR("attempt to write the value of an scx_signal_sizer<IN, OUT> instance from its OUT interface",
                        msg.str().c_str());
    }
    void
    report_more_than_one_writer(sc_core::sc_object * first_writer, sc_core::sc_object * second_writer) {
        std::stringstream msg;

        msg << "\n signal sizer "
               "`" << name() << "' "
              " (" << kind() << ")";
        if (first_writer) {
            msg << "\n first writer "
                   "`" << first_writer->name() << "' "
                  " (" << first_writer->kind() << ")";
        }
        if (second_writer) {
            msg << "\n second writer "
                   "`" << second_writer->name() << "' "
                   "(" << second_writer->kind() << ")";
        }
        SC_REPORT_ERROR("scx_signal_sizer<IN, OUT> cannot have more than one writer", msg.str().c_str());
    }

    /* Derived sc_signal interface */
    template<class T, bool IS_IN>
    struct inout_if : public sc_core::sc_signal_inout_if<T> {
        explicit
        inout_if(scx_signal_sizer *outer):
            m_outer{outer} {
        }
        virtual void
        write(const T & val) override {
            if (IS_IN) {
                m_outer->write(IN(val));
            } else {
                m_outer->report_attempt_to_write();
            }
        }
        virtual const sc_core::sc_event &
        default_event() const override {
            return value_changed_event();
        }
        virtual const sc_core::sc_event &
        value_changed_event() const override {
            if (m_outer->m_value_changed_event_p == nullptr) {
                m_outer->m_value_changed_event_p = new sc_core::sc_event(sc_core::sc_gen_unique_name("change_event"));
            }
            return (* m_outer->m_value_changed_event_p);
        }
        virtual const T &
        read() const override {
            return m_val;
        }
        virtual const T &
        get_data_ref() const override {
            return m_val;
        }
        virtual bool
        event() const override {
            return m_outer->event();
        }
        virtual void
        register_port(sc_core::sc_port_base & port, const char * if_typename) {
            m_outer->check_port(port, IS_IN);
        }
        T                       m_val;
        scx_signal_sizer *      m_outer;
    };

    /* Specialized derived sc_signal interface */
    template<bool IS_IN>
    struct inout_if<bool, IS_IN> : public sc_core::sc_signal_inout_if<bool> {
        explicit
        inout_if(scx_signal_sizer *outer):
            m_outer{outer} {
        }
        virtual void write(const bool & val) {
            if (IS_IN) {
                m_outer->write(IN(val));
            } else {
                m_outer->report_attempt_to_write();
            }
        }
        virtual const sc_core::sc_event & default_event() const override {
            return value_changed_event();
        }
        virtual const sc_core::sc_event & posedge_event() const override {
            if (m_outer->m_posedge_event_p == nullptr) {
                m_outer->m_posedge_event_p = new sc_core::sc_event(sc_core::sc_gen_unique_name("posedge_event"));
            }
            return (* m_outer->m_posedge_event_p);
        }
        virtual const sc_core::sc_event & negedge_event() const override {
            if (m_outer->m_negedge_event_p == nullptr) {
                m_outer->m_negedge_event_p = new sc_core::sc_event(sc_core::sc_gen_unique_name("negedge_event"));
            }
            return (* m_outer->m_negedge_event_p);
        }
        virtual const sc_core::sc_event & value_changed_event() const override {
            if (m_outer->m_value_changed_event_p == nullptr) {
                m_outer->m_value_changed_event_p = new sc_core::sc_event(sc_core::sc_gen_unique_name("change_event"));
            }
            return (* m_outer->m_value_changed_event_p);
        }
        virtual const bool & read() const override {
            return m_val;
        }
        virtual const bool & get_data_ref() const override {
            return m_val;
        }
        virtual bool event() const override {
            return m_outer->event();
        }
        virtual bool posedge() const override {
            return event() && m_outer->read();
        }
        virtual bool negedge() const override {
            return event() && !m_outer->read();
        }
        virtual void
        register_port(sc_core::sc_port_base & port, const char * if_typename) {
            m_outer->check_port(port, IS_IN);
        }
        bool               m_val;
        scx_signal_sizer * m_outer;
    };

    /* Member variables */
    inout_if<OUT, false>       m_cur_val_if;
    inout_if<IN, true>         m_new_val_if;
    sc_core::sc_event *        m_value_changed_event_p;
    sc_core::sc_event *        m_posedge_event_p;
    sc_core::sc_event *        m_negedge_event_p;
    sc_dt::uint64              m_delta_count; /* Whether an update occured in preceding delta cycle */
    sc_core::sc_time           m_time_stamp;
    sc_core::sc_port_base *    m_input;
    sc_core::sc_process_handle m_writer;
};

} /* namespace scx */

#endif /* SCX_SIGNAL_SIZER_H */

