//
// (C) COPYRIGHT 2017 ARM Limited. ALL RIGHTS RESERVED
//

#ifndef SCX_WRAPPER_FEATURE_H
#define SCX_WRAPPER_FEATURE_H

#include "scx/scx_net_parameter.h"

// this is the virtual class to implement features within the SystemC wrapper
class scx_wrapper_feature
{
private:
  int mNumCPUs;                  // number of cpus or cores
  CarbonObjectID* mHandle;       // Handle to the Cycle Model
  scx::scx_debug_target_if& mTarget; // debug target
  std::string mName;             // Unique name to identify the feature
  std::string mParamName;        // Debug parameter name
  std::string mDescription;      // Debug parameter description
  std::string mModuleName;       // sc_module instance name
  bool mEnabled;                 // whether feature is enabled or not
  void* mICMModel;
  std::vector<scx_feature_parameter*> mParams; // hold a collection of parameters
  scx_feature_parameter* findParameter(std::string name)
  {
    for (unsigned int i = 0; i < mParams.size(); ++i)
      if (mParams[i]->name() == name)
	return mParams[i];
    return NULL;
  }

public:
  scx_wrapper_feature(int num_cpus, CarbonObjectID* cycle_model_handle,scx::scx_debug_target_if& target):
    mNumCPUs(num_cpus),mHandle(cycle_model_handle),mTarget(target)
  {
    mEnabled = false;
  }
  
  virtual ~scx_wrapper_feature()
  {
    // free params
    for (unsigned int i = 0; i < mParams.size(); ++i)
      delete mParams[i];
    mParams.clear();
  }
  
  // add a net parameter to this feature
  bool addNetParameter(std::string name,std::string net_path,std::string description,
		       int64_t maxValue=LLONG_MAX,int64_t defaultValue=0,
		       scx::scx_debug_param_value_type scx_type=scx::SCX_DEBUG_PARAM_INT,bool runtime=true)
  {
    // create the net parameter
    scx_net_parameter* netParam = new scx_net_parameter(name,net_path,handle());
    netParam->setDescription(description);
    mParams.push_back(netParam);
    // add parameter to debug target
    mTarget.add_param(netParam->name(),netParam->description(),scx_type,runtime,0,maxValue,defaultValue);
    netParam->setValue(scx::to_string(defaultValue));
    return true;
  }

  // add a generic string parameter to this feature
  bool addParameter(std::string name,std::string value,std::string description)
  {
    // create the parameter
    scx_feature_parameter* param = new scx_feature_parameter(name,value,description);
    mParams.push_back(param);
    // add parameter to debug target
    mTarget.add_param(param->name(),param->description(),scx::SCX_DEBUG_PARAM_STRING,true,0,0,0,value.c_str());
    return true;    
  }
  
  // add a generic integer parameter to this feature
  bool addParameter(std::string name,int value,std::string description="",int64_t min=0,int64_t max=INT_MAX)
  {
    // create the parameter
    std::string valstr = scx::to_string(value);
    scx_feature_parameter* param = new scx_feature_parameter(name,valstr,description);
    mParams.push_back(param);
    // add parameter to debug target
    mTarget.add_param(param->name(),param->description(),scx::SCX_DEBUG_PARAM_INT,true,min,max,value);
    return true;    
  }
  
  // set the value for a string parameter
  bool setParameter(std::string name, std::string value)
  {
    scx_feature_parameter* param = findParameter(name);
    if (param)
      return param->setValue(value);
    return false;
  }

  // set the value for a integer parameter
  bool setParameter(std::string name, int value)
  {
    scx_feature_parameter* param = findParameter(name);
    if (param)
    {
      return param->setValue(scx::to_string(value));
    }
    return false;
  }

  // get the value for a string parameter
  bool getParameter(std::string name,std::string& value)
  {
    scx_feature_parameter* param = findParameter(name);
    if (param)
      return param->getValue(value);
    return false;    
  }
  
  // get the value for an integer parameter
  bool getParameter(std::string name,int& value)
  {
    scx_feature_parameter* param = findParameter(name);
    if (param)
    {
      std::string valstr;
      param->getValue(valstr);
      value = atoi(valstr.c_str());
      return true;
    }
    return false;    
  }

  // add a parameter to the debug target to enable or disable this feature
  void addEnableParameter()
  {
    mTarget.add_param(paramName(),description(),scx::SCX_DEBUG_PARAM_BOOL,true,0,1,isEnabled());
  }

  const char* name() { return mName.c_str(); }
  void setName(const char* name) { mName = name; mParamName = mName + "_ENABLED"; }
  const char* paramName() { return mParamName.c_str(); }
  void setParamName(const char* name) { mParamName = name; }
  const char* description() { return mDescription.c_str(); }
  void setDescription(const char* desc) { mDescription = desc; }
  const char* moduleName() { return mModuleName.c_str(); }
  void setModuleName(const char* name) { mModuleName = name; }
  int numCPUs() { return mNumCPUs; }
  void setNumCPUs(int numCPUs) { mNumCPUs = numCPUs; }
  CarbonObjectID* handle() { return mHandle; }
  bool isEnabled() { return mEnabled; }
  
  virtual void enable() { mEnabled = true; }
  virtual void disable() { mEnabled = false; }
  virtual void update() {}                     // called in the update method of the CarbonScheduleChannel
  virtual void start_of_simulation() {}        // called at start_of_simulation
  virtual void end_of_simulation() {}          // called at end_of_simulation
  virtual void end_of_elaboration() {}         // called at end_of_elaboration
  virtual void before_end_of_elaboration() {}  // called at before_end_of_elaboration
  virtual void run(std::ostream& os)           // called by the user to do something
  {
    SCX_REPORT_WARNING(moduleName(),"Run not supported for the %s feature",name());
  }
};

#endif
