/// \file icm_api.h
/// \brief Global API Registry
//
/// \copyright 2017 ARM. All rights reserved.
//
/// The global API registery allows components (particularly IP models) to register APIs that they
/// support so that other code can make user of tose APIs.  The APIs are registered with an opaque
/// pointer to the component that is registering it, as well as the name of the API.  To make use
/// of the API, the consumer must have an understanding of the API associated with the name and
/// how to decode the API pointer into the appropriate functions

#ifndef ICM_API_H
#define ICM_API_H

#include "icm/api/icm_types.h"
#include "icm/api/icm_exports.h"

#ifdef __cplusplus
extern "C" {
#endif


///  Register an API
//
/// \param component [in] The component that is registering the API
/// \param name [in] The name associated with this particular API
/// \param api [in] The pointer to the actual api

ICM_API_CALL void icm_api_register(icm_model_t component, const char *name, void *api);

///  Get a named API for a given component
//
/// \param component [in] The component that is registering the API
/// \param name [in] The name associated with the API
//
/// \returns a pointer to the requested API, or NULL if no such API is regsitered for the component

ICM_API_CALL void *icm_api_get(icm_model_t component, const char *name);

/// Unregister an API
//
/// \param component [in] The component that is registering the API
/// \param name [in] The name associated with this particular API
//
/// Remove the API named 'name' and associated with component from the registry.  After being
/// unregistered, any further calls to icm_api_get() will return NULL.
//
/// Other components may already hold a pointer to the API.  It is up to the creator of any API
/// to determine the lifetime of that API and how to manage invalidation of API pointers

ICM_API_CALL void icm_api_unregister(icm_model_t component, const char *name);

/// Register an API
//
/// \param component [in] The component that is registering the API
//
/// Like icm_api_unregister but removes ALL apis associated with component

ICM_API_CALL void icm_api_unregister_all(icm_model_t component);

#ifdef __cplusplus
}
#endif

#endif
