/// \file icm_control.h
/// \copyright 2017 ARM. All rights reserved.
///
/// \detail ICM APIs for debugger run control
/// The control APIs provide functionality required to support debugger 's run control commands like
/// requests for run/halt a simulation model, program-counter breakpoints etc.
///
/// \terminology Here is the explanation for a few terms/concepts related to these APIs
///
/// (1) halted state: a model is considered in a halted-state, if clocking this model does not advance 
/// its internal state machine like executing any instructions, updating any registers etc. 
///
/// (2) debuggable point: a state where the model's other debug view APIs can be invoked/relied upon.
/// 
#ifndef icm_control_h
#define icm_control_h

#include "icm/api/icm_types.h"
#include "icm/api/icm_error.h"

/// API KEY
UNUSED_ATTR static const char *api_id_arm_control_1_0 = "arm.cycle_model.control/1.0";

#ifdef __cplusplus
extern "C" {
#endif

/// To move the model out of a halted state 
///
/// \param[in]   model
/// \param[in] err
/// \return 1 if model switched to running and the model will issue a callback like model_halted (or breakpoint hit) at some point
///         0 if model failed to switch state (some error, or, command ignored (halt_core)
typedef int (*icm_control_exit_halt_fn)(icm_model_t model, icm_error_t *err);

/// To move the model to halted state
///
/// \param[in]   model
/// \param[in] err
/// \return 1 if transitioned to 'halting', and the model will issue a callback like model_halted at some point
///         0 if model failed to switch state (some error, or, command ignored (halt_core)
typedef int (*icm_control_enter_halt_fn)(icm_model_t model, icm_error_t *err);

/// signature of a callback function, which model should use to indicate having reached a halted state
/// Ths should happen in response to a enter_halt command
///
/// \param[in] model handle to the icm model making this call
/// \param[in] client_context the client context which can process this callback. 
/// \param[in] success if set to 1, it indicates that the model had success reaching a debuggable state.
/// This context is conveyed during "setup the callback" phase
typedef void (*icm_control_callback_model_halted_fn)(icm_model_t model, void *client_context, int32_t success);

/// register callback regarding model hitting the halted state
//
/// \param[in]   model
/// \param[in]   client_context pointer to the parent class, which shall receive the callback
/// \param[in]   callback pointer to a function which can listen to model having reached halted state
/// \param[in] err
typedef void (*icm_control_set_callback_model_halted_fn)(icm_model_t model, void *client_context,
                                                            icm_control_callback_model_halted_fn callback,
                                                            icm_error_t *err);

/// request the model to do instruction steps
//
/// \param[in]   model
/// \param[in]   step_count number of instructions steps
/// \param[in] err
/// \return 1 if switched to stepping and the model will issue a callback like step_done at some point
///         0 if model failed to switch state (some error, or, command ignored (halt_core)
typedef int (*icm_control_request_instr_step_fn)(icm_model_t model, int32_t step_count, icm_error_t *err);


/// signature of a callback function, which model should use to indicate having finished instruction stepping
/// Ths should happen in response to a request_instr_step command.
/// Note: if an instruction step gets interrupted due to some reason (say breakpoints): model is not obligated 
/// to make this call
///
/// \param[in] model handle to the icm model making this call
/// \param[in] client_context the client context which can process this callback. 
/// \param[in] success if set to 1, it indicates that the model had success reaching a debuggable state.
/// This context is conveyed during "setup the callback" phase
typedef void (*icm_control_callback_instructions_stepped_fn)(icm_model_t model, void *client_context, int32_t success);

/// register callback regarding instruction step completion
//
/// \param[in]   model
/// \param[in] client_context the client context which can process the callback. 
/// \param[in]   callback pointer to a function which can listen to program breakpoint hit notification
/// \param[in] err
typedef void (*icm_control_set_callback_instructions_stepped_fn)(icm_model_t model, void *client_context,
                                                            icm_control_callback_instructions_stepped_fn callback,
                                                            icm_error_t *err);

/// Set a program breakpoint
/// \param[in] model
/// \param[in] address
/// \param[in] err
typedef void (*icm_control_set_program_breakpoint_fn)(icm_model_t model, uint64_t address, icm_error_t *err);

/// Clear a program breakpoint
/// \param[in] model
/// \param[in] address
/// \param[in] err
typedef void (*icm_control_clear_program_breakpoint_fn)(icm_model_t model, uint64_t address, icm_error_t *err);

/// Enable a program breakpoint
/// \param[in] model
/// \param[in] address
/// \param[in] err
typedef void (*icm_control_enable_program_breakpoint_fn)(icm_model_t model, uint64_t address, icm_error_t *err);

/// Disable a program breakpoint
/// \param[in] model
/// \param[in] address
/// \param[in] err
typedef void (*icm_control_disable_program_breakpoint_fn)(icm_model_t model, uint64_t address, icm_error_t *err);

/// signature of a function which can listen to program breakpoint hit notification
/// \param[in] model
/// \param[in] client_context the client context which can process the callback. 
/// \param[in] address
typedef void (*icm_control_program_breakpoint_callback_fn)(icm_model_t model, void *client_context, uint64_t address);

/// register callback for breapoints
//
/// \param[in]   model
/// \param[in]   client_context pointer to the parent class, which shall receive program breakpoint notification callbacks
/// \param[in]   callback pointer to a function which can listen to program breakpoint hit notification
/// \param[in] err
typedef void (*icm_control_set_program_breakpoint_callback_fn)(icm_model_t model, void *client_context,
                                                            icm_control_program_breakpoint_callback_fn callback,
                                                            icm_error_t *err);

/// Get count of instruction executed by the model since the last call.
//
/// \param[in]   model
/// \param[in] err
//\return number of instructions executed
typedef int64_t (*icm_control_get_instruction_count_fn)(icm_model_t model, icm_error_t *err);

/// check if model is at a debug point
//
/// \param[in]   model
/// \param[in] err
/// \return a status indicator: 1 suggests that model is at a debuggable point.
typedef int32_t (*icm_control_is_debuggable_fn)(icm_model_t model, icm_error_t *err);

/// set persistent executation state of a model to support asynchronous control (for HALT_CORE)
//
/// \param[in]   model
/// \param[in]   stay_halted if set to 1, model is expected to ignore any future
/// run/step commands, until this function is called with stay_halted=0
/// \param[in] err
typedef void (*icm_control_set_execution_state_fn)(icm_model_t model,
                                                   int stay_halted,
                                                   icm_error_t *err);

/// signature of a callback function, which the model should use to report having finished its
/// reset sequence.
///
/// \param[in] model handle to the icm model making this call
/// \param[in] client_context the client context which can process this callback.
/// This context is conveyed during "setup the callback" phase
typedef void (*icm_control_callback_model_reset_done_fn)(icm_model_t model, void *client_context);

/// register callback regarding the model having finished its reset sequence
//
/// \param[in]   model
/// \param[in]   client_context pointer to the parent class, which shall receive the callback
/// \param[in]   callback pointer to a function which can listen to model having reached reset_done state
/// \param[in] err
typedef void (*icm_control_set_callback_model_reset_done_fn)(icm_model_t model, void *client_context,
                                                            icm_control_callback_model_reset_done_fn callback,
                                                            icm_error_t *err);

/// \brief Collection of debugger run control APIs
struct icm_api_control_t
{
  icm_control_exit_halt_fn                 exit_halt;
  icm_control_enter_halt_fn                enter_halt;
  icm_control_set_callback_model_halted_fn set_cb_halted;

  icm_control_request_instr_step_fn                request_step;
  icm_control_set_callback_instructions_stepped_fn set_cb_stepped;


  icm_control_set_program_breakpoint_fn          set_program_breakpoint;
  icm_control_clear_program_breakpoint_fn        clear_program_breakpoint;
  icm_control_enable_program_breakpoint_fn       enable_program_breakpoint;
  icm_control_disable_program_breakpoint_fn      disable_program_breakpoint;
  icm_control_set_program_breakpoint_callback_fn set_cb_program_breakpoint_hit;

  icm_control_get_instruction_count_fn instr_count;

  icm_control_is_debuggable_fn at_debug_point;

  icm_control_set_execution_state_fn set_execution_state;

  icm_control_set_callback_model_reset_done_fn set_cb_reset_done;
};

#ifdef __cplusplus
}
#endif

#endif /// icm_control_h
