/// \file icm_disasm.h
/// \brief ICM APIs for disassembler
/// \copyright 2017 ARM. All rights reserved.
//
/// The Disassembler APIs provide functionality required to enable disassembly view
///

#ifndef icm_disasm_h
#define icm_disasm_h

#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_disasm_1_0 = "arm.cycle_model.disasm/1.0";

/// \brief opaque data type for icm disassembler objects
typedef void *icm_disasm_t;

/// \brief Get number of disassembly modes supported by the model
/// \param[in] model
/// \param[out] err
/// \return number of disassembly modes 
typedef uint32_t (*icm_disasm_mode_count_fn)(icm_model_t model, icm_error_t *err);

/// \brief Get information(name) about each supported mode
/// \param[in]    model
/// \param[in]    index        mode index is allowed 0..n-1 (n=supported modes)
/// \param[out]  mode_name     Pointer to a string holding mode name (e.g. arm, thumb, thumb2, arm64, current)
/// \param[out]   err       error
/// Note : memory management for *name will be taken care by the API implementation.
///         Caller is expected to immediately copy the returned value. The pointers maybe reused by the ICM
///         implementation during any following ICM API call.
typedef icm_disasm_t (*icm_disasm_mode_info_fn)(icm_model_t model, uint32_t index, const char **mode_name, icm_error_t *err);

/// \brief Get index for current mode of disassembly. 
///  The return value should be between 0 and n-1, where n = return value of icm_disasm_mode_count_fn.
///  This value will be interpreted the "index" used in icm_disasm_mode_info_fn API; it should
///  correspond to the disassembler object which can provide correct disassembly (ARM, Thumb, Thumb2, I64 etc) 
/// based on CPU's internal state like CPSR/CP15 registers etc
/// \param[in] model
/// \param[out] err
/// \return index of the disassembler/mode 
typedef uint32_t (*icm_disasm_get_current_mode_index_fn)(icm_model_t model, icm_error_t *err);


/// \brief Get disassembly and opcode for a given address/memory-space
///
//// \param[in]   handler - disassembly request handler
//// \param[in]   index - mode index that was used during info API call
//// \param[in]   space	memory space
//// \param[in]   addr	memory address
//// \param[out]  opcode     Pointer to be set to point to the opcode
//// \param[out]  disassembly     Pointer to be set to the disassembly of the instruction
//// \param[out]  next_addr Pointer must be set to the new value of addr, based on instruction size just decoded
//// \param[out]   err       error
/// Note : memory management for *opcode and *disassembly will be taken care by the API implementation.
///         Caller is expected to immediately copy the returned value. The pointers maybe reused by the ICM
///         implementation during any following ICM API call.

typedef void (*icm_disasm_disassemble_fn)(icm_disasm_t handler, uint32_t index,
                                          uint32_t space, uint64_t addr,
                                          const char **opcode,
                                          const char **disassembly,
                                          uint64_t *next_addr,
                                          icm_error_t *err);

/// \brief Collection of disassembly APIs
struct icm_api_disasm_t
{
  icm_disasm_mode_count_fn mode_count;
  icm_disasm_mode_info_fn mode_info;
  icm_disasm_get_current_mode_index_fn get_curr_mode;
  icm_disasm_disassemble_fn disassemble;
};

#ifdef __cplusplus
}
#endif

#endif
