/// \file icm_error.h
/// \brief ICM APIs for error support
/// \copyright 2017 ARM. All rights reserved.

#ifndef icm_error_h
#define icm_error_h

#include "icm/api/icm_types.h"
#include "icm/api/icm_exports.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_error_1_0 = "arm.cycle_model.error/1.0";

/// \brief opaque data type for icm error object
typedef void *icm_error_t;

/// An error code/status decoded from icm_error_t
///
/// The space of error codes is divided into related
/// blocks:
/// 0      to 999     General errors
/// 1000   to 1999    Register errors
/// 2000   to 2999    Memory errors
///
enum icm_status_t
{
  /// \brief no error
  ICM_OKAY = 0,
  /// \brief some error (generic error code), simulation can continue
  ICM_WARNING = 1,
  /// \brief a fatal error occurred, simulation should not be continued
  ICM_FATAL_ERROR = 2,
  /// \brief API call failed because the target is busy at the moment
  ICM_TARGET_BUSY = 3,
  /// \brief API call was serviced but partially ( need more details)
  ICM_INCOMPLETE_RESPONSE = 4,
  
  /// std::exceptions that are not being assigned an 
  /// ICM error code.
  ICM_STD_EXCEPTION = 5,

  /// Errors that we can't classify.  Gate keeper for some exception that 
  /// is not derived from std::exception.  Should never happen, but implementations
  /// are expected to handle it 
  ICM_UNKNOWN_INTERNAL_ERROR = 6,

  /// Provided index was out of range
  ICM_INDEX_OUT_OF_RANGE = 7,

  ICM_INTERNAL_ERROR = 9,

  /// Should be avoided - intened for use during development.
  ICM_NOT_IMPLEMENTED = 100,  
  ICM_NOT_REGISTERED =  101,

  /// register error messages:
  ICM_REGISTER_ERROR = 1000,
  ICM_REG_WRITE_FAILED_READ_ONLY_REG = 1000,

  /// memory error messages:
  ICM_MEMORY_ERROR = 2000,
  ICM_MEM_WRITE_FAILED_TO_READ_ONLY = 2000
};
/// \brief Decode error message
/// \param[in]   err       error
/// \param[out]    message      A pointer to  error messages, which will be populated by the model,
/// \return true to indicate that error was successfully decoded.
/// \details
/// The returned strings in message may be reused w/ the next api call.  Caller is expected to immediately copy
/// values
/// \return an int, which is an "icm_status_t" value, indicating nature of the error

typedef int64_t (*icm_error_decode_fn)(icm_error_t err, const char **message);


/// Delete error class err
///
/// \param [in] err handle to the error class
///
/// The caller manages the lifetime of an error, but it 
/// was created by the API library so we provide a delete
/// routine. 
typedef void (*icm_error_delete_fn)(icm_error_t err);

/// \brief Collection of error APIs
struct icm_api_error_t
{
  /// \brief decode error message
  icm_error_decode_fn decode;
  icm_error_delete_fn delete_error;
};

#ifdef __cplusplus
}
#endif

#endif
