/// \file icm_memory.h
/// \brief ICM APIs for memory view support
/// \copyright 2017 ARM. All rights reserved.
//
/// The memory APIs provide functionality required to support debug views for a
/// model's memory.
/// NOTE: These APIs do not cover the interfaces required by a model, if it needs to send out a request
///
/// \terminology Here is the explanation for a few terms used in these APIs
///
/// (1) memory object : a memory implementation, which can serve read/write
/// commands.
/// (2) Endianness - CADI based debuggers expect one of the three: 
/// (A) arch-defined (B) Little Endian (C) Big Endian
/// "architecture defined" endianness is default, and no distinction is made between BE8 or BE32
/// The attributes defined in this header allow for a little more flexibility, which maybe
/// used with future debugger APIs like IRIS
/// 
/// All the read/write calls are expected to be transfered as "nx 1-byte transfers" as if little-endian
/// It will be the responsibility of the final target to reorder the data it received, or translate address 
/// to match its own endianness, assuming that incoming data was in Little Endian.
/// IRIS based debuggers might leverage flexibility permitted for specifying endianness: e.g. BE8/BE32, 
/// as well as an option to NOT specify endianness. 

#ifndef icm_memory_h
#define icm_memory_h

#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_memory_1_0 = "arm.cycle_model.memory/1.0";

/// \brief opaque data type for icm memory object
typedef void *icm_memory_t;

/// Each memory object's info will contain a uint64_t mem_attr, where each bit
/// would represent different attributs of the memory object
enum icm_memory_attribute_t
{
  /// \brief bit#0 if set, implies the memory object is read-only
  ICM_MEM_IS_READ_ONLY = (1 << 0),
  /// \brief bit#1 if set, implies the memory object is write-only
  ICM_MEM_IS_WRITE_ONLY = (1 << 1),
  /// \brief bit#2 if set, implies memory is little -endian
  ICM_MEM_ENDIAN_LE = (1 << 2),
  /// \brief bit#3 if set, implies memory is big-endian - byte invariant (BE8)
  ICM_MEM_ENDIAN_BE8 = (1 << 3),
  /// \brief bit#4 if set, implies memory is big-endian - word invariant (BE32)
  ICM_MEM_ENDIAN_BE32 = (1 << 4),
  /// \brief bit#5 if set, implies memory is big-endian
  ICM_MEM_ENDIAN_VARIABLE = (1 << 5),
  /// \brief bit#6 if set, implies memory endianness is unspecified. Client should assume little endian
  ICM_MEM_ENDIAN_UNSPECIFIED = (1 << 6),
};

/// Get memory objects count (memory spaces)
/// \param[in] model
/// \param[out] err
/// \return number of memory objects in ICM
typedef uint32_t (*icm_memory_count_fn)(icm_model_t model, icm_error_t *err);

/// Get info about a memory by index
///
/// \param[in]   model
/// \param[in]   index	memory index
/// \param[out]  name     Pointer to be set to point to the name of the memory.
/// \param[out]  desc    Pointer to be set to the memory description.
/// \param[out] start    modified to have start of the memory object's address range
/// \param[out] end      modified to have end of the memory object's address range
/// \param[out] mem_attr    a bit-pattern, where each bit represent different
/// attributes of the memory ( \link  icm_memory_attribute_t  \endlink)
///
/// \note : memory management for *name/*desc will be taken care by the API implementation.
///         Caller is expected to immediately copy the returned values. The pointers maybe reused by the ICM
///         implementation during any following ICM API call.
/// \return handle to the memory object
typedef icm_memory_t (*icm_memory_info_fn)(icm_model_t model, uint32_t index, const char **name, const char **desc,
                                            uint64_t *start, uint64_t *end, uint64_t *mem_attr, icm_error_t *err);

/// Read memory from a given address for a given memory space
/// \param[in]    mem_handle  memory implementation which can handle the request
/// \param[in]    addr      memory address to access
/// \param[in]    num_bytes      Number of bytes to access
/// \param[out]  bytes_processed number of bytes processed (continuous bytes, starting from start address)
/// \param[out]   buf       data buffer to fill with read data.
/// \param[out]   err       error
/// Note: Caller is responsible for ensuring that buf is large enough to cover
/// the full width of the memory request
typedef void (*icm_memory_read_fn)(icm_memory_t mem_handle, const uint64_t addr, uint32_t num_bytes, 
				   uint32_t *bytes_processed, uint8_t *buf, icm_error_t *err);

/// Write  memory to a given address for a given memory space
/// \param[in]    mem_handle  memory implementation which can handle the request
/// \param[in]    addr      memory address to access
/// \param[in]    num_bytes Number of bytes to access
/// \param[out]  bytes_processed number of bytes processed (continuous bytes, starting from start address)
/// \param[out]   buf       data buffer w/ data to write.
/// \param[out]   err       error
/// Note: Caller is responsible for ensuring that buf is large enough to cover
/// the full width of the memory request
typedef void (*icm_memory_write_fn)(icm_memory_t mem_handle, const uint64_t addr, uint32_t num_bytes,
                                    uint32_t* bytes_processed, const uint8_t *buf, icm_error_t *err);

/// \brief Collection of memory APIs
struct icm_api_memory_t
{
  icm_memory_count_fn count;
  icm_memory_info_fn info;
  icm_memory_read_fn read;
  icm_memory_write_fn write;
};

#ifdef __cplusplus
}
#endif

#endif

