/// \file icm_memory_ports.h
/// \brief ICM APIs for supporting outgoing memory requests
/// \copyright 2017 ARM. All rights reserved.
///
/// Memory-port APIs enable models to send out a debug memory read/write request.
/// This 
///
/// \note Here is an explanation for a few terms/concepts related to these APIs
///
/// (1) outgoing ports - an IP model could have 0 or 1 more than 1 outgoing memory ports. E.g.
/// (a) a CPU could have a memory port to send out requests it can't handle from $/TCMs.
/// (b) an interconnect could have multiple ports, which may goto different connected IP models.
/// 
/// (2) memory request handlers - As part of the integration of the IP model, 
/// the integration layer may setup the model with various client-contexts and memory request 
/// handlers per outgoing port, where memory-request-handlers will be C-functions, which should be
/// called by the IP model, to send out the request on the respective port. 
///
/// (3) The client-context could be anything those C-functions may need to be able to service those requests. 
/// e.g. a TLM port object pointer, a pointer to the class providing actual memory implementation, etc.
///
/// (4) Servicing a memory request could mean sending request out to a SystemC component using TLM ports, or it could be
/// something else, based on how the memory port is expected to be serviced for that model in a sytstem.
/// 
/// (5) port names - mapping: A port-names exposed by the model should either match the names the consumer/integration 
/// layer is expecting (e.g. to create outgoing TLM ports) OR, the integration laye should have some mapping 
/// between ports it exposes at its own boundary vs. ports exposed by the model using these memory-port APIs.
///
#ifndef icm_icm_memory_ports_h
#define icm_icm_memory_ports_h

#include "icm/api/icm_types.h"
#include "icm/api/icm_error.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_memory_ports_1_0 = "arm.cycle_model.memory_ports/1.0";

/// number of outgoing memory ports of the model
/// \param[in]    model
/// \param[out]   err       error
typedef uint32_t (*icm_memory_ports_count_fn)(icm_model_t model, icm_error_t *err);

///  Get information about individual ports exposed by the model.
/// \param[in]    model
/// \param[in]    index    port index (0.. count-1, where count = value returned by icm_memory_ports_count_fn function)
/// \param[out]   name     pointer to a char * containing name of the port.
/// \param[out]   desc     pointer to a char * containing any description the model (free format at the moment)
/// \param[out]   err      error
/// \details
/// Note : memory management for *name/*desc will be taken care by the API implementation.
///        Caller is expected to immediately copy the returned values. The pointers maybe reused by the ICM
///        implementation during any following ICM API call.
typedef void (*icm_memory_port_info_fn)(icm_model_t model, uint32_t index, const char **name, const char **desc,
                                   icm_error_t *err);

/// signature of the functions used by the model to send out a read command
///
/// \param[in]   client_context pointer to the parent class, which shall receive the memory request
/// \param[in]   model
/// \param[in]   addr       start address
/// \param[in]   num_bytes  number of bytes to be processed
/// \param[out]  bytes_processed number of bytes processed (continuous bytes, starting from start address)
/// \param[inout]  buffer       data buffer to hold data to be read/written. Caller is responsible for ensuring it is
/// large enough to cover the entire memory request
/// \return      value 0 if the request is serviced successfully
//
/// \details Such a function should be implemented by a wrapper model, which wants to allow icm model to be able to send
/// out memory requests.

typedef int (*icm_memory_port_read_handler_fn)(icm_model_t model, void *client_context, uint64_t addr, 
						uint32_t num_bytes, uint32_t *bytes_processed, uint8_t *buffer);
/// signature of a function used by the model to send out a write command
typedef int (*icm_memory_port_write_handler_fn)(icm_model_t model, void *client_context, uint64_t addr, 
						uint32_t num_bytes, uint32_t *bytes_processed, const uint8_t *buffer);

/// set memory request handlers for a given port
/// \param[in] model
/// \param[in]  index    port index (0.. count-1, where count = value returned by icm_memory_ports_count_fn function)
/// \param[in]  client_context pointer to the parent class, which shall receive the memory request
/// \param[in]  callback pointer to a function which shall receive outgoing memory read/write commands
/// specified by the icm model during icm_memory_port_info_fn, the ICM model will not use this callback and report a warning.
///
/// \param[out] err
typedef void (*icm_memory_port_set_read_handler_fn)(icm_model_t model, uint32_t index, void *client_context, 
                                                    icm_memory_port_read_handler_fn callback, icm_error_t *err);
typedef void (*icm_memory_port_set_write_handler_fn)(icm_model_t model, uint32_t index, void *client_context, 
                                                     icm_memory_port_write_handler_fn callback, icm_error_t *err);
/// \brief Collection of memory port APIs
struct icm_api_memory_ports_t
{
  icm_memory_ports_count_fn count;
  icm_memory_port_info_fn info;
  icm_memory_port_set_read_handler_fn set_reader;
  icm_memory_port_set_write_handler_fn set_writer;
};

#ifdef __cplusplus
}
#endif

#endif
