/// \file icm_model_info.h
/// \brief ICM APIs for model information like 'raw model, number of cores
/// within a multicore/multi-cluster
/// \copyright 2017 ARM. All rights reserved.

#ifndef icm_model_info_h
#define icm_model_info_h

#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_modelinfo_1_0 =
    "arm.cycle_model.model_info/1.0";

/// Get handle to raw core model object.
//
/// \param[in]   model
/// \param[out] err
/// \details
/// \return raw model (like CarbonObject*, or UBD*, etc implemnetation)
typedef void *(*icm_get_raw_model_fn)(icm_model_t model, icm_error_t *err);

/// Get basic info about an icm model
//
/// \param[in]   model
/// \param[out]  name     Pointer to be set to point to the name of the model.
/// Name is non-hierarchical (ie 'cpu0', and *not* 'CortexR8.cpu0' )
/// \param[out]  desc     Pointer to be set to the model description, which has
/// no fixed format at this point.
/// \param[out] child_count    number of children icm models.
/// \param[out] err
/// \details
/// The returned strings in name and desc may be reused w/ the next api call.
/// Caller is expected to immediately copy
/// values

typedef void (*icm_model_info_fn)(icm_model_t model, const char **name,
                                  const char **desc, int32_t *child_count,
                                  icm_error_t *err);

/// Get info about a child  model by index.
//
/// \param[in]   model
/// \param[in]   index	index of child that a use wants to access.
/// \param[out] err
// \return handle to the child model
typedef icm_model_t (*icm_model_get_child_fn)(icm_model_t model, uint32_t index,
                                              icm_error_t *err);

/// Retrieve a model property's value, if implemented by the model; else return
/// null. It is upto the model to interpret
/// type of the model property's value (as integer, hex, or a string)
///
/// NOTE: if the model wants to return a numeric value, it must use C-style
/// prefixes to imply decimal, or hexa (0x..), binary (0b..)
/// or octal (0..), etc.
//
/// \param[in]   model
/// \param[in]   prop_name     Pointer to model property name. This has no
/// specified format at the moment
/// \param[out] err
/// \return Pointer to be set to the model property value, if model has a
/// property that matches prop_name; else null

typedef const char *(*icm_model_get_property_fn)(icm_model_t model,
                                                 const char *prop_name,
                                                 icm_error_t *err);

/// Set a model property's value, if model can't use the value being passed:
/// either because property is read-only, or the input value is unacceptable:
/// it can return -1; else returns 0 to imply success.
/// In case of failure in setting of the value, the model can convey the cause
/// of the failure using 'err' object
///
/// NOTE: if the model wants to set a numeric value, the value has to be conveyed as a 
/// C-string, but it should use C-style prefixes to imply decimal, or hexa (0x..), binary (0b..)
/// or octal (0..), etc.
//
/// \param[in]   model
/// \param[in]   prop_name     Pointer to model property name. This has no
/// specified format at the moment.
/// \param[in]   prop_value    Pointer to model property value. This has no
/// specified format at the moment.
/// \param[out] err    an optional description of the issue in case of the model
/// rejecting the value being set.
/// \return -1 in case of failure; 0 in case of success.

typedef int (*icm_model_set_property_fn)(icm_model_t model,
                                         const char *prop_name,
                                         const char *prop_value,
                                         icm_error_t *err);

/// \brief Collection of APis to get basic Model Info
struct icm_api_model_info_t
{
  icm_get_raw_model_fn raw_model;
  icm_model_info_fn info;
  icm_model_get_child_fn child;
  icm_model_get_property_fn get_property;
  icm_model_set_property_fn set_property;
};

#ifdef __cplusplus
}
#endif

#endif
