/// \file icm_registers.h
/// \brief ICM APIs for register support
/// \copyright 2017 ARM. All rights reserved.
//
/// The register APIs provide functionality required to support debug views for model's registers
///
/// \terminology Here is the explanation for a few terms used in these APIs
///
/// (1) register object : a group of registers, an individual registers, or a register bitfield
#ifndef icm_registers_h
#define icm_registers_h

#include "icm/api/icm_types.h"
#include "icm/api/icm_error.h"

#ifdef __cplusplus
extern "C" {
#endif

/// API KEY
UNUSED_ATTR static const char *api_id_arm_registers_1_0 = "arm.cycle_model.registers/1.0";

  /// \brief opaque data type for icm register object
typedef void *icm_reg_t;

/// Each register object's info will contain a uint64_t reg_type, where each bit would represent
/// different attributs of the register
enum icm_reg_attribute_t
{
  ICM_REG_NOATTRIBUTE = 0,
  /// \brief bit#0 if set, implies register object and its children (if any) are invisible.
  /// e.g. setting a register group invisibile for a register group would force its registers
  /// to be invisible. Similarly, if a register is set to "invisible", its bitfields will be invisible
  ICM_REG_INVISIBLE = (1 << 0),
  /// \brief bit#1 if set, implies register is Program counter
  ICM_REG_PC = (1 << 1),
  /// \brief bit#2 if set, implies register is a Stack pointer register
  ICM_REG_SP = (1 << 2),
  /// \brief bit#3 if set, implies register is a Link register
  ICM_REG_LR = (1 << 3),
  /// \brief bit#4 if set, implies register is a program control/status register
  ICM_REG_CPSR = (1 << 4),
  /// \brief bit#5 if set, implies register is defined as "PC_MEMSPACE" register by the model
  ICM_REG_PC_MEMSPACE = (1 << 5),
  /// \bried bits 6:63: Reserved

  /// \todo discuss whether we need any of the following attributes
  /// \todo ICM_REG_IS_READ_ONLY
  /// \todo ICM_REG_IS_WRITE_ONLY
  /// \todo ICM_REG_IS_WRITE_ONE_TO_CLEAR
  //
};

/// Get register objects count (all the groups, group.registers, group.reg.bitfields)
/// \param[in] model
/// \param[out] err
/// \return number of register objects in ICM
typedef uint32_t (*icm_reg_count_fn)(icm_model_t model, icm_error_t *err);

/// Get info about a register object by index
//
/// \param[in]   model
/// \param[in]   index	index of register object, 0 to count-1, where count comes from icm_reg_count_fn
/// \param[out]  name     Pointer to be set to point to the name of the register object
/// \param[out]  desc     Pointer to be set to the description.
/// \param[out]  width    modified to have width of register or bitfield (in terms of number of bits)
/// \param[out]  offset modified to have offset of the register or register-bitfield (in terms of number of bits)
/// \param[out] reg_attribute    a bit-pattern, where each bit represent different attributes of the register
///             ( \link  icm_reg_attribute_t  \endlink)
//
/// Note 1: Conventions about the Name:
///  Names of the register objects are alpha-numeric values typically allowed for register groups/registers/bitfields.
///  However, it uses "." as a delimiter, to distinguish between various register objects
///    - For a "group",   name can not contain any ".".                          : e.g. "Core"
///    - For a "register ",name must contain <group name> + "." + <register name> : e.g. "Core.CPSR"
///    - For a "bitfield",name must contain <group name> + "." + <register name> +"."+<bitfield name> : e.g.
///    "Core.CPSR.F"
///
/// Note 2: memory management for *name/*desc will be taken care by the API implementation.
///         Caller is expected to immediately copy the returned values. The pointers maybe reused by the ICM
///         implementation during any following ICM API call.
/// \return handle to the register object - if an error occurs the returned handle may be invalid ( == 0)
typedef icm_reg_t (*icm_reg_info_fn)(icm_model_t model, uint32_t index, const char **name, const char **desc,
                                     uint32_t *width, uint32_t *offset, uint64_t *reg_attribute, icm_error_t *err);

/// Read data to the specified register
/// \param[in]    reg 	    handle to the register object
/// \param[out]   buf       data buffer to fill with read data.
/// \param[out]   err       error
//
/// Note: Caller is responsible for ensuring that buf is large enough to cover the full width of the register
typedef void (*icm_reg_read_fn)(icm_reg_t reg, uint64_t *buf, icm_error_t *err);

///  Write data to the specified register
/// \param[in]    reg 	    handle to the register object
/// \param[in]    buf       data buffer w/ data to write.
/// \param[out]   err       error
///
/// Note: Caller is responsible for ensuring that buf is large enough to cover the full width of the register
typedef void (*icm_reg_write_fn)(icm_reg_t reg, const uint64_t *buf, icm_error_t *err);

/// \brief Collection of registers APIs
struct icm_api_registers_t
{
  icm_reg_count_fn count;
  icm_reg_info_fn info;
  icm_reg_read_fn read;
  icm_reg_write_fn write;
};

#ifdef __cplusplus
}
#endif

#endif

