/// \file        icm_scx_control.h
/// \brief       Class to simplify usage of ICM Control APIs
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_control_h
#define icm_scx_control_h

#include <iostream>
#include <scx/scx.h>

#include "icm/api/icm_api.h"
#include "icm/api/icm_control.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_control_callback.h"
#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{
/// \brief icm_scx_control implements usage of ICM Control APIs.
/// 1. Forwards various calls to the ICM models, and implements bsaic error checks
/// 2. Provide function to setup callbacks
class icm_scx_control
{
public:
  icm_api_control_t *api_control;
  /// \brief Constructor takes ICM Model handle as input, fetches handle to ICM Control and error API objects
  /// \param[in]  model - handle to the ICM model

  icm_scx_control(icm_model_t icm_in) : icm(icm_in)
  {
    api_control = static_cast<icm_api_control_t *>(icm_api_get(icm, api_id_arm_control_1_0));
    api_err = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  }
  ~icm_scx_control() {}

  /// \return \c true if switched to running and the model will issue a callback like model_halted (or breakpoint hit) at some point
  ///         \c false if model failed to switch to running state (either due to some error, or, because it ignored the command (halt_core)
  bool exit_halt()
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      bool status = (api_control->exit_halt(icm, &icm_err)!=0);
      if (!icm_err) return status;
      icm_scx_decode_error(api_err, icm_err);
      return false; // dont expect a response back
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "exit_halt failed:no ICM Control APIs");
    }
    return false;
  }

  /// \brief this function should be called when model wants to enter Halted state
  /// \return \c true indicates that model has transitioned to halting state, and will make a callback like
  /// model_halted at some point.
  ///  \c false indicates that model can stop immediately.
  bool enter_halt()
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      bool status = (api_control->enter_halt(icm, &icm_err) !=0);
      if (!icm_err) return status;
      icm_scx_decode_error(api_err, icm_err);
      return false;
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "enter_halt failed:no ICM Control APIs");
    }
    return false;
  }

  /// \brief set program breakpoint
  /// \param space[in] memory space - currently ignored everywhere.
  /// \param addr[in]  address to set breakpoint
  void set_program_breakpoint(uint32_t space, const uint64_t &addr)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      api_control->set_program_breakpoint(icm, addr, &icm_err);
      if (icm_err) icm_scx_decode_error(api_err, icm_err);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "set_program_breakpoint failed:no ICM Control APIs");
    }
  }

  /// \brief clear program breakpoint
  /// \param space[in] memory space - currently ignored everywhere.
  /// \param addr[in]  address from where to clear a breakpoint
  void clear_program_breakpoint(uint32_t space, const uint64_t &addr)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      api_control->clear_program_breakpoint(icm, addr, &icm_err);
      if (icm_err) icm_scx_decode_error(api_err, icm_err);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "clear_program_breakpoint failed:no ICM Control APIs");
    }
  }

  /// \brief send a request to do n-instruction steps
  /// \param count[in]  number of instruction steps
  /// \return \c true if switched to stepping; model will callback like step_done later
  ///         \c false if model failed to switch to stepping state (some error or command ignored (halt_core)
  bool request_step(uint32_t count)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      bool status = (api_control->request_step(icm, count, &icm_err) !=0);
      if (!icm_err) return status;
      icm_scx_decode_error(api_err, icm_err);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "request_step failed:no ICM Control APIs");
    }
    return false; //dont expect a response
  }

  /// \brief return current instruction count
  /// \param instruction_count[out]  current instruction count
  /// \return false if control api is not implemented
  bool instr_count(uint64_t &instruction_count)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      instruction_count = api_control->instr_count(icm, &icm_err);
      if (icm_err)
      {
        icm_scx_decode_error(api_err, icm_err);
        return false;
      }
      return true;
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "instr_count failed:no ICM Control APIs");
    }
    return false;
  }

  /// \brief return cycle count between two calls to this function
  /// \param cycle_count[out]  number of cycle steps
  /// \param system_cycles[in]  system level cycles?
  /// \return false if control api is not implemented
  bool cycle_count(uint64_t &cycle_count, bool system_cycles)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (system_cycles)
    {
      // SCX_REPORT_INFO_VERB(scx::SCX_DEBUG,"icm_scx_control::debug_get_cycle_count()", "system_cycles not
      // supported!");  return false;
    }
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "cycle_count failed:no ICM Control APIs");

    return false;
  }

  /// \brief set control related callbacks
  /// \param client_context[in]  A pointer to icm_scx_control_callback_if, which can receive various callbacks
  void set_control_callback(icm_scx_control_callback_if *client_context)
  {
    ICM_SCX_FUNCTION_LOGGING;
    if (api_control)
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "calling set_cb_halted");
      api_control->set_cb_halted(icm, client_context, c_icm_control_callback_model_halted, &icm_err);
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "calling set_cb_reset_done");
      api_control->set_cb_reset_done(icm, client_context, c_icm_control_callback_model_reset_done, &icm_err);
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "returned from  set_cb_halted");
      if (icm_err) icm_scx_decode_error(api_err, icm_err);

      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "calling set_cb_stepped");
      api_control->set_cb_stepped(icm, client_context, c_icm_control_callback_instructions_stepped, &icm_err);
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "returned from  set_cb_stepped");
      if (icm_err) icm_scx_decode_error(api_err, icm_err);
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "calling set_cb_program_breakpoint_hit");
      api_control->set_cb_program_breakpoint_hit(icm, client_context, c_icm_control_program_breakpoint_callback,
                                                 &icm_err);

      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "set_control_callback", "returned from  set_cb_program_breakpoint_hit");
      if (icm_err) icm_scx_decode_error(api_err, icm_err);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "set_control_callback failed:no ICM Control APIs");
    }
  }

  void set_execution_state(bool stay_halted_flag)
  {
    ICM_SCX_FUNCTION_LOGGING;
    int stay_halted = (stay_halted_flag?1:0);

    if (api_control)
    {
      api_control->set_execution_state(icm, stay_halted, &icm_err);
      if (icm_err)
      {
        icm_scx_decode_error(api_err, icm_err);
      }
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "set_execution_state failed:no ICM Control APIs");
    }
  }

  /// \brief report whether the mdoel is at a debuggable point or not.
  /// \return \c true by default, \c false only if the control API at_debug returns false.
  bool is_debuggable() {
    bool at_debug = true; /* default */
      ICM_SCX_FUNCTION_LOGGING;
      if (api_control) {
        at_debug = (api_control->at_debug_point(icm, &icm_err) != 0);
        if (icm_err) icm_scx_decode_error(api_err, icm_err);
        return at_debug;
      }
      else {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control", "is_debuggable failed:no ICM Control APIs");
      }
      return at_debug;
  }

private:
  icm_model_t icm;

  // error object/api: used in several functions
  icm_error_t icm_err;
  icm_api_error_t *api_err;
};
} // icm_scx

#endif
