/// \file        icm_scx_control_callback.h
/// \brief       Interface class and code to receive ICM Model's Control callbacks.
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_control_callback_h
#define icm_scx_control_callback_h

#include <scx/scx.h>

#include "icm/api/icm_api.h"
#include "icm/api/icm_control.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{
/// \brief icm_scx_control_callback_if is an interface class to receive various ICM control callbacks
class icm_scx_control_callback_if
{
public:
  /// \bried constructor
  icm_scx_control_callback_if() {}
  /// \brief destructor
  virtual ~icm_scx_control_callback_if() {}

  /// \brief This function is called when the ICM model reports entering halted state
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  success - 1 implies that model had no issues
  virtual void model_halted(icm_model_t model, int32_t success) {}
 /// \brief This function is called when the ICM model reports finishing reset sequence.
  /// \param[in]  model - handle to the ICM model
  virtual void model_reset_done(icm_model_t model) {}

  /// \brief This function is called when the ICM model reports finishing an earlier instruction step request
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  success - 1 implies that model had no issues

  virtual void instructions_stepped(icm_model_t model, int32_t success) {}

  /// \brief This function is called when the ICM model reports hitting a program breakpoint
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  address - breakpoint location that is hit
  virtual void program_breakpoint_hit(icm_model_t model, uint64_t address) {}
};

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Following code is a 'utility' code, meant to hookup ICM Model's C-callbacks to the icm_scx_control_callback_if's
// member functions
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/// \brief C-callback when ICM model halts. Convert it to icm_scx_control_callback_if::model_halted
inline void c_icm_control_callback_model_halted(icm_model_t model, void *client_context, int32_t success)
{
  ICM_SCX_FUNCTION_LOGGING;
  try
  {
    if (client_context)
    {
      ((icm_scx_control_callback_if *)client_context)->model_halted(model, success);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control_callback.h",
                           "c_icm_control_callback_model_halted received with a null client-context. Will be ignored.");
    }
  }
  catch (std::exception &ex)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding model_halted caused an exception (%s)", ex.what());
  }
  catch (...)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding model_halted caused an unknown exception");
  }
}

/// \brief callback when ICM model finishes its reset sequence: forward to model_reset_done.
inline void c_icm_control_callback_model_reset_done(icm_model_t model, void *client_context)
{
  ICM_SCX_FUNCTION_LOGGING;
  try
  {
    if (client_context)
    {
      ((icm_scx_control_callback_if *)client_context)->model_reset_done(model);
    }
    else
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_control_callback.h",
                           "c_icm_control_callback_model_reset_done received with a null client-context. Will be ignored.");
    }
  }
  catch (std::exception &ex)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding model_reset_done caused an exception (%s)", ex.what());
  }
  catch (...)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding model_reset_done caused an unknown exception");
  }
}

/// \brief C-callback when ICM model finishes request for instruction step. Convert it to
/// icm_scx_control_callback_if::instructions_stepped
inline void c_icm_control_callback_instructions_stepped(icm_model_t model, void *client_context, int32_t success)
{
  ICM_SCX_FUNCTION_LOGGING;
  try
  {
    if (client_context)
    {
      ((icm_scx_control_callback_if *)client_context)->instructions_stepped(model, success);
    }
    else
    {
      SCX_REPORT_INFO_VERB(
          scx::SCX_DEBUG, "icm_scx_control_callback.h",
          "c_icm_control_callback_instructions_stepped received with a null client-context. Will be ignored.");
    }
  }
  catch (std::exception &ex)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding instructions_stepped caused an exception (%s)",
                     ex.what());
  }
  catch (...)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding instructions_stepped caused an unknown exception");
  }
}

/// \brief C-callback when ICM model hits a program breakpoint. Convert it to
/// icm_scx_control_callback_if::program_breakpoint_hit
inline void c_icm_control_program_breakpoint_callback(icm_model_t model, void *client_context, uint64_t address)
{
  ICM_SCX_FUNCTION_LOGGING;
  try
  {
    if (client_context)
    {
      ((icm_scx_control_callback_if *)client_context)->program_breakpoint_hit(model, address);
    }
    else
    {
      SCX_REPORT_INFO_VERB(
          scx::SCX_DEBUG, "icm_scx_control_callback.h",
          "c_icm_control_program_breakpoint_callback received with a null client-context. Will be ignored.");
    }
  }
  catch (std::exception &ex)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding program_breakpoint_hit caused an exception (%s)",
                     ex.what());
  }
  catch (...)
  {
    SCX_REPORT_FATAL("icm_scx_control_callback.h", "forwarding program_breakpoint_hit caused an unknown exception");
  }
}

} // icm_scx
#endif
