/// \file        icm_scx_debug_impl.h
/// \brief       Class icm_scx_debug_impl definition, which implements various scx_debug_target_if interfaces
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_debug_impl_h
#define icm_scx_debug_impl_h

#include <scx/scx.h>

#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#include "carbon/carbon_shelltypes.h"
#include "icm/integration/icm_scx_model.h"

namespace icm_scx
{
/// \brief  Immplement scx_cadi_target related interfaces, and receive icm control callbacks
///  Class scx_cadi_target implements all the forward debug calls like register, memory, debug, run control,
/// disassembler calls, as part of various scx::debug_xxx interfaces.
/// It also implements ICM::control callback interface, to receive notification from the ICM model
/// regarding breakpoint hit, model entering halt state, as well as finishing a request for instruction
/// steps.

class icm_scx_debug_impl : public virtual scx::scx_debug_registers_if,
                           public virtual scx::scx_debug_params_if,
                           public virtual scx::scx_debug_memory_if,
                           public virtual scx::scx_debug_control_if,
                           public virtual scx::scx_debug_disass_if,
                           public virtual icm_scx_control_callback_if
{

public:
  /// \brief Constructs base class creates a scx_cadi_target tied to its interfaces, as well as
  /// pass icm_in to icm_wrapper, which uses ICM interfaces supported by the icm model,
  /// to provide 'cadi' support.
  /// \param[in]   icm_in handle to the ICM model
  /// \param[in] name for the model. This can be a null too, in which case, icm model supplies the name
  explicit icm_scx_debug_impl(icm_model_t icm_in, const char *name = nullptr);

  /// /brief destructor usual cleanup, destroys scx_cadi_target object
  ~icm_scx_debug_impl() { delete mCADI; }

  /// \brief get access to scx_debug_target_if
  scx::scx_debug_target_if *get_scx_debug_target_if() const { return mCADI; }

  /// \brief get the child targets
  std::vector<icm_scx_debug_impl*> get_children() { return children_targets; }

  // \brief get the name of the target
  std::string& get_name() { return m_name; }

  ////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // following section has scx_cadi_target related functions
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////
  /* Debug support - Registers */
  virtual bool debug_register_read(uint32_t num, uint8_t *datap) override { return icm_wrapper.regs.read(num, datap); }
  virtual bool debug_register_write(uint32_t num, const uint8_t *datap) override
  {
    return icm_wrapper.regs.write(num, datap);
  }
  virtual void debug_register_bpt_set(uint32_t num) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_register_bpt_set is not supported!");
  }
  virtual void debug_register_bpt_clear(uint32_t num) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_register_bpt_clear is not supported!");
  }
  /* Debug support - Parameters */
  virtual bool debug_param_get(const std::string &name, std::string &value) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_param_get is not implemented!");
    return false;
  }
  virtual bool debug_param_set(const std::string &name, const std::string &value) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_param_set is not implemented!");
    return false;
  }
  /* Debug support - Memory */
  virtual bool debug_memory_read(uint32_t space, const uint64_t &addr, uint32_t size, uint8_t *datap) override
  {
    return icm_wrapper.memories.read(space, addr, size, datap);
  }

  virtual bool debug_memory_write(uint32_t space, const uint64_t &addr, uint32_t size, const uint8_t *datap) override
  {
    return icm_wrapper.memories.write(space, addr, size, datap);
  }

  virtual void debug_memory_bpt_set(uint32_t space, const uint64_t &addr) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_memory_bpt_set is not supported");
  }

  virtual void debug_memory_bpt_clear(uint32_t space, const uint64_t &addr) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_memory_bpt_clear is not supported!");
  }

  /* Debug support - Control */
  virtual void debug_program_bpt_set(uint32_t space, const uint64_t &addr) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_program_bpt_set:PC space= %d", space);
    icm_wrapper.control.set_program_breakpoint(space, addr);
  }
  virtual void debug_program_bpt_clear(uint32_t space, const uint64_t &addr) override
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, m_name.c_str(), "debug_program_bpt_clear: PC space= %d", space);
    icm_wrapper.control.clear_program_breakpoint(space, addr);
  }

  virtual bool debug_run() override { return icm_wrapper.control.exit_halt(); }
  virtual bool debug_step(uint32_t count) override { return icm_wrapper.control.request_step(count); }

  virtual bool debug_stop() override { return icm_wrapper.control.enter_halt(); }

  virtual void debug_reset(uint32_t level) override
  {
    SCX_REPORT_WARNING(m_name.c_str(), "Reset from debugger is not supported");
  }

  virtual bool debug_get_instruction_count(uint64_t &instruction_count) override
  {
    return icm_wrapper.control.instr_count(instruction_count);
  }
  virtual bool debug_get_cycle_count(uint64_t &cycle_count, bool system_cycles) override
  {
    return icm_wrapper.control.cycle_count(cycle_count, system_cycles);
  }

  virtual uint32_t debug_get_pcmemspace() override
  {
    return icm_wrapper.regs.read_pc_memspace();
  }

  virtual void set_execution_state(bool stay_halted) override
  {
    icm_wrapper.control.set_execution_state(stay_halted);
  }
  virtual bool is_debuggable() override
  {
    return icm_wrapper.control.is_debuggable();
  }

  /* Debug support - Disassembly */
  virtual uint32_t debug_get_curr_mode() override
  {
      return icm_wrapper.disasm.get_current_mode_index();
  }
  virtual bool debug_disassemble(uint32_t space, uint64_t &addr, std::string &opcode, std::string &disassembly,
                                 uint32_t mode) override
  {
    return icm_wrapper.disasm.disassemble(space, addr, opcode, disassembly, mode);
  }

  ////////////////////////////////////////////////////////////////////////////////////////////////
  //
  // following section implements icm_scx_control_callback_if functions
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////
  /// \brief This function is called when the ICM model reports entering halted state
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  success - 1 implies that model had no issues
  virtual void model_halted(icm_model_t model, int32_t success) override;

  /// \brief This function is called when the ICM model reports finishing an earlier instruction step request
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  success - 1 implies that model had no issues
  virtual void instructions_stepped(icm_model_t model, int32_t success) override;

  /// \brief This function is called when the ICM model reports hitting a program breakpoint
  /// \param[in]  model - handle to the ICM model
  /// \param[in]  address - breakpoint location that is hit
  virtual void program_breakpoint_hit(icm_model_t model, uint64_t address) override;

  /// \brief This function is called when the ICM model reports finishing reset sequence.
  /// \param[in]  model - handle to the ICM model
  virtual void model_reset_done(icm_model_t model) override;
public:
  void set_mem_port_debug_handler(icm_scx_mem_port_if *actual_port_impl)
  {
    icm_wrapper.memory_ports.set_mem_port_debug_handler(actual_port_impl);
  }

  std::vector<std::string>* get_mem_port_names_list()
  {
    return icm_wrapper.memory_ports.get_mem_port_names_list();
  }

private:
  // \brief This function creates instances of a this class corresponding to icm model's children icm models
  void create_children_targets();

private:
  // \brief model name, either passed as a non-null constructor value, or read from icm model
  std::string m_name;

  // \brief "CADI" pointer
  scx::scx_cadi_target *mCADI;

  // \brief handle to the ICM model
  icm_model_t icm;

  /// \brief an object, which can process all the ICM model information like parsing registers, hooks up interface for
  /// control callbacks etc
  icm_scx_model icm_wrapper;

  /// \brief children target objects corresponding to ICM Model's children. Populated by create_children_targets()
  std::vector<icm_scx_debug_impl *> children_targets;
};

} // icm_scx

/// To keep this file trimmed, some code has been moved to a .hpp, which is inlined here
#include "icm/integration/icm_scx_debug_impl.hpp"

#endif
