/// \file        icm_scx_debug_impl.hpp
/// \brief       implements some functions from icm_scx_debug_impl.h
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_debug_impl_hpp
#define icm_scx_debug_impl_hpp

#include <iostream>
#include <limits>
#include <string>

/// \brief Constructs base class creates a scx_cadi_target tied to its interfaces, as well as
/// pass icm_in to icm_wrapper, which uses ICM interfaces supported by the icm model,
/// to provide 'cadi' support.
/// \param[in]   icm_in handle to the ICM model
/// \param[in] name for the model. This can be a null too, in which case, icm model supplies the name

namespace icm_scx
{

inline icm_scx_debug_impl::icm_scx_debug_impl(icm_model_t icm_in, const char *name)
    : mCADI(nullptr), icm(icm_in), icm_wrapper(icm_in)
{
  ICM_SCX_FUNCTION_LOGGING;
  /// \note caller can override the model-specified name by specifying a non-null parameter for name
  /// by default, model specified name fetched from model_info API is used
  m_name = (name != nullptr) ? name : icm_wrapper.model_info.get_instance_name();

  // CADI object
  mCADI = new scx::scx_cadi_target(icm_wrapper.model_info.get_target_name(),
                                    m_name/*instance*/,
                                    icm_wrapper.model_info.get_target_type());

  // pass CADI object to icm_wrapper, which can use this object to populate various debug views
  icm_wrapper.setup_debug_view(mCADI);

  mCADI->set_params_if(this);

  if (icm_wrapper.supports_registers){
    mCADI->set_control_if(this);
    mCADI->set_registers_if(this);
    // ICM callback handler
    icm_wrapper.control.set_control_callback(this);
    if (icm_wrapper.supports_disassembler) mCADI->set_disass_if(this);
  }
  if (icm_wrapper.supports_memory) mCADI->set_memory_if(this);

  // create Target handlers for ICM model's children
  create_children_targets();
}

inline void icm_scx_debug_impl::create_children_targets()
{
  ICM_SCX_FUNCTION_LOGGING;
  int num_children = icm_wrapper.model_info.get_child_count();
  if (num_children > 0)
  {
    icm_error_t icm_err;
    children_targets.resize(num_children);
    for (int i = 0; i < num_children; i++)
    {
      auto child = icm_wrapper.api_model_info->child(icm, i, &icm_err);
      if (icm_err) icm_scx_decode_error(icm_wrapper.api_error, icm_err);

      const char *child_model_name, *child_model_desc;
      int child_child_count = 0; /*a dont care at this level, but we need to pass a variable anyways*/
      icm_wrapper.api_model_info->info(child, &child_model_name, &child_model_desc, &child_child_count, &icm_err);

      // use the model-specified 'model name' to create child's target implementation
      // prefixed with m_name (parent component's name)
      std::string hier_name = m_name + "." + child_model_name;
      auto dbg = new icm_scx_debug_impl(child, hier_name.c_str());
      children_targets.push_back(dbg);
    }
  }
}

/// \brief This function is called when the ICM model reports entering halted state
/// \param[in]  model - handle to the ICM model
/// \param[in]  success - 1 implies that model reached debuggable point successfully.
inline void icm_scx_debug_impl::model_halted(icm_model_t model, int32_t success)
{
  ICM_SCX_FUNCTION_LOGGING;
  mCADI->stop_acknowledge(success);
}


/// \brief This function is called when the ICM model reports finishing reset sequence.
/// \param[in]  model - handle to the ICM model
inline void icm_scx_debug_impl::model_reset_done(icm_model_t model)
{
  ICM_SCX_FUNCTION_LOGGING;
   mCADI->reset_done();
}
/// \brief This function is called when the ICM model reports finishing an earlier instruction step request
/// \param[in]  model - handle to the ICM model
/// \param[in]  success - 1 implies that model reached debuggable point successfully.
inline void icm_scx_debug_impl::instructions_stepped(icm_model_t model, int32_t success)
{
  ICM_SCX_FUNCTION_LOGGING;
  mCADI->step_acknowledge(success);
}

/// \brief This function is called when the ICM model reports hitting a program breakpoint
/// \param[in]  model - handle to the ICM model
/// \param[in]  address - breakpoint location that is hit
inline void icm_scx_debug_impl::program_breakpoint_hit(icm_model_t model, uint64_t address)
{
  int space = debug_get_pcmemspace();
  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "program_breakpoint_hit", "@PC=0x%llx, space=%u", (unsigned long long)address,
                       space);
  bool really_hit = mCADI->program_bpt_check(space, address);
  if (!really_hit)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "program_breakpoint_hit",
                         "CADI::program_bpt_check returned false. Resuming by calling exit_halt ");
    icm_wrapper.control.exit_halt(); //return value ignored.
  }
}

} // icm_scx

#endif
