/// \file        icm_scx_disasm.h
/// \brief       Class to simplify usage of ICM disasm APIs and defines a local class
/// icm_scx_disasm_object to encapsulate disasm-objects received from ICM Model
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

/// \note Terms related to disassembler APIs
///
/// An ICM model can specify several modes like ARM, Thumb, Thumb2, Arm64 etc. as possible options for disassembly.
///
/// This CADI specific implementation class automatically adds a mode "Current" before registering these modes with
/// CADI,
/// as  mode#0.
//. When diwassembly view is viewed with "current" mode, disassemble API is called with mode=0. In this case,
/// disassemble API in this file calls "get_current_mode_index" API, which returns the actul mode based on PC state
/// (CPSR etc).
///
/// If the model does not report any modes, a dummy "Current" mode is added, which is to avoid crashing
/// SCX layer.
///

#ifndef icm_scx_disasm_h
#define icm_scx_disasm_h

#include <iostream>
#include <limits>
#include <map>
#include <string>
#include <vector>

#include "icm/api/icm_api.h"
#include "icm/api/icm_disasm.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{

/// \brief icm_scx_disasm_object is a container for disasm objects received from ICM
/// Model.
struct icm_scx_disasm_object
{
  icm_scx_disasm_object(const char *name_, uint32_t index_, icm_disasm_t handle_)
  {
    name = (name_ == nullptr) ? "<mode name missing>" : name_;
    mode_info_index = index_;
    handle = handle_;
  }
  ~icm_scx_disasm_object() = default;

  /// \brief disasm object's name
  std::string name;
  /// \brief Model specified mode index (during info api)
  uint32_t mode_info_index;
  /// \brief handle to the disasm object which implements disasm read/write functions
  icm_disasm_t handle;
};

/// \brief Class to process ICM disasm APIs
/// 1. gather all the disasm objects reported by the icm disasm api
/// 2. push all the disasm spaces into CADI
/// 3. Provide means to get access to "disasm handle" for a given "cadi  registration index"
class icm_scx_disasm
{
public:
  /// \brief consturctor fetches ICM Register and Error APIs handle
  /// \param icm_in[in] ICM model
  icm_scx_disasm(icm_model_t icm_in);
  ~icm_scx_disasm() { cadi_to_disasm_map.clear(); }

  /// \brief setup_debug_view registers all the disassembly modes with CADI
  /// \return true if there is any disassembly mode
  bool setup_debug_view(scx::scx_debug_target_if *parent_cadi);

  /// \brief Get disassembly and opcode for the instruction at a given address/memory-space using the specified mode
  ///
  //// \param[in]   space memory space
  //// \param[inout]   addr  memory address;  updated to points to the next address for decoding if call is successful
  //// \param[out]  opcode     storage for holding opcode  of the instruction
  //// \param[out]  disassembly     storage for holding disassembly of the instruction
  //// \param[out]  mode mode-index that corresponds to the mode value used during cadi registeration using add_mode
  bool disassemble(uint32_t space, uint64_t &addr, std::string &opcode, std::string &disassembly, uint32_t mode);

  /// \brief return cadi-index for the disasm-mode reported to be 'current' by icm model.
  uint32_t get_current_mode_index();

private:
  /// \brief extract all the disasms from the ICM Model via ICM disasm API
  /// and push those into a local storage.
  void process_icm_disasm_info();

  /// \brief add a disasm space entry, as read from ICM model into a vector
  /// \param name_[in] disasm object name
  /// \param index[in] mode index for which this mode was received from the model
  /// \param handl[in] disasm handle
  bool add_disasm_info(const char *name_, uint32_t index_, icm_disasm_t handl);

  /// \brief find a disasm object in the current hierarchy
  /// \param name_[in] name of the disasm group or disasm or bitfield to search
  /// \return disasm object, if found, else null
  icm_scx_disasm_object *find_node(const char *name_);

  /// \brief register a mode with CADI as well as add to local bookkeeping.
  /// \param cadi_mode[in] mode to use while registring with cadi
  /// \param name[in] mode name
  /// \param disasm_obj[in] disassembly handler object
  void register_mode(uint32_t cadi_mode, const char *name, icm_scx_disasm_object *disasm_obj);

  /// \brief retrieve icm_disasm_t handle for a given CADI index
  /// \param cadi_index[in]  disasm index requested at CADI layer
  /// \return disasm object, if found within the disasm objects map, else nullptr
  bool add_to_registered_disasms(icm_scx_disasm_object *obj)
  {
    cadi_to_disasm_map.push_back(obj);
    return true;
  }
  /// \brief retrieve icm_disasm_t handle for a given CADI index
  /// \param cadi_index[in]  disasm index requested at CADI layer
  /// \return disasm object, if found within the disasm objects map, else nullptr
  icm_scx_disasm_object *get_disasm_handle(uint32_t cadi_index);

private:
  icm_model_t icm;
  icm_api_disasm_t *api_disasm;

  icm_api_error_t *api_error;
  scx::scx_debug_target_if *mCADI;
  icm_error_t icm_err;

  /// \brief disasm-handlers: in the order of their "CADI registration index"
  std::vector<icm_scx_disasm_object *> cadi_to_disasm_map;

  /// \brief collection of disasm objects
  std::vector<icm_scx_disasm_object *> disasms_from_model_info;
};

} // icm_scx
#include "icm_scx_disasm.hpp"

#endif
