/// \file        icm_scx_disasm.hpp
/// \brief       implementation of some of the functions in icm_scx_disasm.h
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_disasm_hpp
#define icm_scx_disasm_hpp

#include <algorithm>
#include <string>

namespace icm_scx
{

/// \constructor retrieves and holds all the disassembler objects received from ICM model
/// \param icm_in[in] ICM model
inline icm_scx_disasm::icm_scx_disasm(icm_model_t icm_in)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm = icm_in;
  api_disasm = static_cast<icm_api_disasm_t *>(icm_api_get(icm, api_id_arm_disasm_1_0));
  if (api_disasm == nullptr)
  {
    return;
  }
  api_error = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  process_icm_disasm_info();
}

/// \brief extract all the disassebly modes and corresponding disassembler handlers
/// from the ICM Model and add to this class' storage
inline void icm_scx_disasm::process_icm_disasm_info()
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t num_modes = api_disasm->mode_count(icm, &icm_err);
  if ((icm_err) || (num_modes == 0))
  {
    icm_scx_decode_error(api_error, icm_err);
    return;
  }

  const char *mode_name = nullptr;
  for (uint32_t i = 0; i < num_modes; ++i)
  {
    auto handle = api_disasm->mode_info(icm, i, &mode_name, &icm_err);
    if (icm_err)
    {
      icm_scx_decode_error(api_error, icm_err);
    }
    else
    {
      if (add_disasm_info(mode_name, i, handle) == false)
      {
        if (mode_name != nullptr)
          SCX_REPORT_WARNING("process_icm_disasm_info", "Error adding disassembler object(%s)", mode_name);
      }
    }
  }
}

/// \brief add a disassembler mode entry, as read from ICM model
/// \param mode_name[in] disassembler object name
/// \param index[in] mode index (as par 'get info' api)
/// \param handle[in] disassembler handle which can handle this particular mode
inline bool icm_scx_disasm::add_disasm_info(const char *mode_name, uint32_t index, icm_disasm_t handle)
{
  ICM_SCX_FUNCTION_LOGGING;
  if (mode_name == nullptr)
  {
    SCX_REPORT_WARNING("icm_scx_disasm:add_disasm_info", "null name - not allowed\n");
    return false;
  }

  auto obj = find_node(mode_name);
  if (obj != nullptr)
  {
    SCX_REPORT_WARNING("icm_scx_disasm:add_disasm_info", "mode: %s already reported by the model\n", mode_name);
    return false;
  }
  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "add_disasm_info", "mode: %s reported by the model\n", mode_name);
  obj = new icm_scx_disasm_object(mode_name, index, handle);
  disasms_from_model_info.push_back(obj);
  return true;
}

/// \brief find a disassembler object in the modes reported by the model
/// \param name_[in] name of the disassembler mode to search
/// \return disassembler object, if found, else null
inline icm_scx_disasm_object *icm_scx_disasm::find_node(const char *name_)
{
  for (auto &c : disasms_from_model_info)
  {
    if (c->name.compare(name_) == 0)
    {
      return c;
    }
  }
  return nullptr;
}

/// \brief register a mode with CADI as well as add to list of registered modes.
inline void icm_scx_disasm::register_mode(uint32_t mode, const char *name, icm_scx_disasm_object *disasm_obj)
{
  ICM_SCX_FUNCTION_LOGGING;
  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "register_mode", "mode:%u: %s registered\n", mode, name);

  mCADI->add_mode(mode, name);
  add_to_registered_disasms(disasm_obj);
}

/// \brief setup_debug_view registers all the disassembly modes with CADI
/// note: this function addes mode-0 as "current": a requirement for SCX add_mode API
/// \param parent_cadi[in] cadi target handle
/// \return true if there is any disassembly mode

inline bool icm_scx_disasm::setup_debug_view(scx::scx_debug_target_if *parent_cadi)
{
  ICM_SCX_FUNCTION_LOGGING;
  mCADI = parent_cadi;

  // start by adding a default
  register_mode(0, "Current", nullptr);

  if (disasms_from_model_info.size() == 0)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "setup_debug_view", "No Disassembler mode specified.");
    return false;
  }

  uint32_t cadi_mode_index = 1;
  /// \note add each disasm mode into CADI
  for (auto disasm : disasms_from_model_info)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "setup_debug_view", "Added disassembler:%d=%s", cadi_mode_index,
                         disasm->name.c_str());
    register_mode(cadi_mode_index++, disasm->name.c_str(), disasm);
  }
  return true;
}

/// \brief retrieve icm_disasm_t handle for a given CADI index
/// \param cadi_index[in]  disasm index requested at CADI layer
/// \return disasm object, if found within the disasm objects map, else nullptr
inline icm_scx_disasm_object *icm_scx_disasm::get_disasm_handle(uint32_t cadi_index)
{
  ICM_SCX_FUNCTION_LOGGING;

  if (api_disasm == nullptr)
  {
    return nullptr;
  }

  if (cadi_index == 0)
  {
    cadi_index = get_current_mode_index();
  }
  // if it's still 0: we can't handle default case
  if (cadi_index == 0)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "get_disasm_handle", "cadi_index=%d. no disassembler found", cadi_index);
    return nullptr;
  }
  uint32_t mode_info_index = cadi_index;
  if (mode_info_index < cadi_to_disasm_map.size())
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "get_disasm_handle",
                         "cadi_index=%d. Disassembler[mode_info_index=%d] will be used", cadi_index, mode_info_index);
    return cadi_to_disasm_map[mode_info_index];
  }
  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "get_disasm_handle", "cadi_index=%d. out of bound!", cadi_index);
  return nullptr;
}

/// \brief return cadi index for the disasm-mode reported to be 'current' by icm model.
inline uint32_t icm_scx_disasm::get_current_mode_index()
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t cadi_index = 0; // some arbit default

  if (api_disasm == nullptr)
  {
    return cadi_index;
  }

  uint32_t model_current_index; // to receive index of the diassembler which is "current"

  model_current_index = api_disasm->get_curr_mode(icm, &icm_err);
  if (icm_err)
  {
    icm_scx_decode_error(api_error, icm_err);
    return cadi_index; // return default
  }

  // cadi indexes are always model_index + 1, as 0th is a custom entry "Current"
  cadi_index = model_current_index + 1;

  return cadi_index;
}

/// \brief Get disassembly and opcode for the instruction at a given address/memory-space using the specified mode
///
//// \param[in]   space memory space
//// \param[inout]   addr  memory address;  updated to points to the next address for decoding if call is successful
//// \param[out]  opcode     storage for holding opcode  of the instruction
//// \param[out]  disassembly     storage for holding disassembly of the instruction
//// \param[out]  mode mode-index that corresponds to the mode value used during cadi registeration using add_mode
inline bool icm_scx_disasm::disassemble(uint32_t space, uint64_t &addr, std::string &opcode, std::string &disassembly,
                                        uint32_t mode)
{
  ICM_SCX_FUNCTION_LOGGING;
  if (api_disasm == nullptr)
  {
    return false;
  }
  auto obj = get_disasm_handle(mode);
  if (obj == nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "disassemble", "called for a null disasm object, Mode index =  %d", mode);
    return false;
  }
  const char *opcode_ptr = nullptr;
  const char *disassembly_ptr = nullptr;
  uint64_t next_addr = 0;

  (*api_disasm->disassemble)(obj->handle, obj->mode_info_index, space, addr, &opcode_ptr, &disassembly_ptr, &next_addr,
                             &icm_err);
  if (icm_err && (icm_scx_decode_error(api_error, icm_err) == ICM_FATAL_ERROR)) return false;

  if (opcode_ptr != nullptr)
  {
    opcode = opcode_ptr;
  }
  else
  {
    SCX_REPORT_WARNING("debug_disassemble", "Opcode field is empty!");
    opcode = "\?\?";
  }
  if (disassembly_ptr != nullptr)
  {
    disassembly = disassembly_ptr;
  }
  else
  {
    SCX_REPORT_WARNING("debug_disassemble", "Disassembly field is empty!");
    disassembly = "\?\?";
  }
  addr = next_addr;
  return true;
}
} // icm_scx

#endif
