/// \file        icm_scx_error.h
/// \brief       A utility function to decode icm_error object, inlined/included in each icm client class
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_error_h
#define icm_scx_error_h

#include <scx/scx.h>

#include "icm/api/icm_error.h"
#include "icm/api/icm_types.h"

#include <cstring>
#include <string>

namespace icm_scx
{
/// \brief logger for reporting entry/exit of a function
class icm_scx_function_entry_logger
{
  const int ICM_SCX_DEBUG = scx::SCX_DEBUG + 100;
                          /* ICM messages are duplicate of CADI in several cases, and useful
                             almost exclusively for ICM_SCX layer developers.
                             Therefore, reduced severity:setenv FM_SCX_VERBOSITY_LEVEL 600*/

public:
  icm_scx_function_entry_logger(const char *filename, const char *func)
  {
    fname = filename;
    func_name = func;
    SCX_REPORT_INFO_VERB(ICM_SCX_DEBUG, fname, "Entering: %s ", func_name);
  }
  ~icm_scx_function_entry_logger() { SCX_REPORT_INFO_VERB(ICM_SCX_DEBUG, fname, "Exiting: %s ", func_name); }

private:
  const char *fname;
  const char *func_name;
};

#ifndef NDEBUG
#define ICM_SCX_FUNCTION_LOGGING                                                                                       \
  icm_scx_function_entry_logger tmp_icm_scx_function_entry_logging(__FILE__, __FUNCTION__);
#else
#define ICM_SCX_FUNCTION_LOGGING
#endif

/// \brief this is a common error handling code inlined within each ICM client class
/// \return value decoded error code. Expected to interpreted based on icm_status_t enums,defined in icm_error.h
inline int icm_scx_decode_error(icm_api_error_t *api_error, icm_error_t err)
{
  static bool skip_decode = (getenv("ICM_SKIP_DECODE") != nullptr);
  if (skip_decode) return 0;
  if (api_error && err)
  {
    const char *msg;
    int err_code = (*api_error->decode)(err, &msg);
    if (msg) SCX_REPORT_WARNING("icm_wrapper", "%s (error code:%d)", msg, err_code);
    return err_code;
  }
  return 0;
}

} // icm_scx

#endif
