/// \file        icm_scx_memory.h
/// \brief       Class to simplify usage of ICM memory APIs and defines a local class
/// icm_scx_mem_object to encapsulate memory-objects received from ICM Model
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.
#ifndef icm_scx_memory_h
#define icm_scx_memory_h

#include <iostream>
#include <limits>
#include <map>
#include <string>
#include <vector>

#include "icm/api/icm_api.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_memory.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{
static const uint32_t ICM_MEM_BITS_PER_MAU = 8;
static const bool ICM_MEM_WHETHER_PROGRAMMABLE = true;
static const bool ICM_MEM_WHETHER_VIRTUAL_MEM = true;

enum ICM_CADI_SPECIFIC_ENDIAN_FLAGS
{
  ICM_CADI_ENDIANNESS_ARCH_DEFINED = 0,
  ICM_CADI_ENDIANNESS_LE = 1,
  ICM_CADI_ENDIANNESS_BE = 2
};

/// \brief a container for memory objects received from ICM Model.
struct icm_scx_mem_object
{
  icm_scx_mem_object(uint32_t mem_space_, const char *name_, const char *desc_ = "<no description>",
                     icm_memory_t handle = nullptr, uint64_t min_addr_ = 0,
                     uint64_t max_addr_ = std::numeric_limits<uint32_t>::max(), uint64_t attr = 0)
  {
    mem_space = mem_space_;
    name = (name_ == nullptr) ? " " : name_;
    desc = (desc_ == nullptr) ? " " : desc_;
    mem_handle = handle;
    min_addr = min_addr_;
    max_addr = max_addr_;
    attributes = attr;
  }
  ~icm_scx_mem_object() = default;

  ///\brief mem-space specified by the model
  uint32_t mem_space;
  /// \brief memory object's name
  std::string name;
  /// \brief an optional description
  std::string desc;
  /// \brief  start address of mem space
  uint64_t min_addr;
  /// \brief  end address of mem space
  uint64_t max_addr;
  /// \brief attributes based on icm_memory_attribute_t defined in icm_memory.h
  uint64_t attributes;
  /// \brief handle to the memory object which implements memory read/write functions
  icm_memory_t mem_handle;
};

/// \brief Class to process ICM memory APIs
/// 1. gather all the memory objects reported by the icm memory api
/// 2. push all the memory spaces into CADI
/// 3. Provide means to get access to "memory handle" for a given memory space (identified by "CADI registration index")
class icm_scx_memory
{
public:
  /// \brief consturctor fetches ICM Register and Error APIs handle
  /// \param icm_in[in] ICM model
  icm_scx_memory(icm_model_t icm_in);
  ~icm_scx_memory() { map_cadi_to_mem.clear(); }

  /// \brief add all the mem spaces to CADI
  /// \return true if there is any memory
  bool setup_debug_view(scx::scx_debug_target_if *parent_cadi);

  /// \brief memory read function
  bool read(uint32_t space, const uint64_t &addr, uint32_t size, uint8_t *datap);

  /// \brief memory write function
  bool write(uint32_t space, const uint64_t &addr, uint32_t size, const uint8_t *datap);

private:
  /// \brief extract all the memorys from the ICM Model via ICM memory API
  /// and push those into a local storage (memories).
  void process_icm_mem_info();

  /// \brief add a memory space entry, as read from ICM model into a vector
  /// \param mem_space[in] memory space index
  /// \param name_[in] memory object name
  /// \param desc_[in] memory object description
  /// \param handl[in] memory handle
  /// \param min_addr_[in] memory width
  /// \param max_addr_[in] memory offset
  /// \param attr[in] attributes based on icm_mem_attribute_t defined in icm_memorys.h
  bool add_memory_info(uint32_t mem_space, const char *name_, const char *desc_, icm_memory_t handl, uint64_t min_addr_,
                       uint64_t max_addr_, uint64_t attr);

  /// \brief find a memory object in the current hierarchy
  /// \param name_[in] name of the memory object to search
  /// \return memory object, if found, else null
  icm_scx_mem_object *find_node(const char *name_);

  /// \brief store icm_memory_t handle for a given CADI index
  /// \param cadi_index[in]  memory index requested at CADI layer
  /// \return \c false when already present, \c true when added
  bool update_map_cadi_to_mem(uint32_t cadi_index, icm_scx_mem_object *mem_object)
  {
    if (map_cadi_to_mem.find(cadi_index) != map_cadi_to_mem.end())
    {
      return false;
    }

    map_cadi_to_mem[cadi_index] = mem_object;
    return true;
  }
  /// \brief retrieve icm_memory_t handle for a given CADI index
  /// \param cadi_index[in]  memory index requested at CADI layer
  /// \return memory object, if found within the memory objects map, else nullptr
  icm_scx_mem_object *get_mem_handle(uint32_t cadi_index)
  {
    auto ret = map_cadi_to_mem.find(cadi_index);
    if (ret != map_cadi_to_mem.end())
    {
      return ret->second;
    }
    return nullptr;
  }

private:
  icm_model_t icm;
  icm_api_memory_t *api_mem;

  icm_api_error_t *api_error;
  scx::scx_debug_target_if *mCADI;
  icm_error_t icm_err;

  /// \brief a mapping of memory-handlers and their "CADI registration index"
  std::map<uint32_t, icm_scx_mem_object *> map_cadi_to_mem;

  /// \brief collection of memory objects
  std::vector<icm_scx_mem_object *> memories;
};

} // icm_scx
#include "icm_scx_memory.hpp"

#endif
