/// \file        icm_scx_memory.hpp
/// \brief       implementation of some of the functions in icm_scx_memory.h
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_memory_hpp
#define icm_scx_memory_hpp

#include <cinttypes>

namespace icm_scx
{

/// \constructor creates the mem_objects, which holds all the memory objects received from ICM model, gets handle for
/// ICM Memory and error APIs
/// \param icm_in[in] ICM model
inline icm_scx_memory::icm_scx_memory(icm_model_t icm_in)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm = icm_in;
  api_mem = static_cast<icm_api_memory_t *>(icm_api_get(icm, api_id_arm_memory_1_0));
  if (api_mem == nullptr)
  {
    return;
  }
  api_error = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  process_icm_mem_info();
}

/// \brief extract all the memorys from the ICM Model via ICM memory API
/// and push those into mem_objects.
inline void icm_scx_memory::process_icm_mem_info()
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t num_mems = api_mem->count(icm, &icm_err);
  if ((icm_err) || (num_mems == 0))
  {
    icm_scx_decode_error(api_error, icm_err);
    return;
  }

  const char *name = nullptr;
  const char *description = nullptr;
  uint64_t min_addr;
  uint64_t max_addr;
  uint64_t mem_attr;
  for (uint32_t i = 0; i < num_mems; ++i)
  {
    icm_memory_t mem_handle = api_mem->info(icm, i, &name, &description, &min_addr, &max_addr, &mem_attr, &icm_err);
    if (icm_err)
    {
      icm_scx_decode_error(api_error, icm_err);
    }
    else
    {
      if (add_memory_info(i, name, description, mem_handle, min_addr, max_addr, mem_attr) == false)
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "process_icm_mem_info", "Error adding memory object(%s)", name);
      }
    }
  }
}

/// \brief add a memory space entry, as read from ICM model
/// \param name_[in] memory object name
/// \param desc_[in] memory object description
/// \param handl[in] memory handle
/// \param wid[in] memory width
/// \param off[in] memory offset
/// \param attr[in] attributes based on icm_mem_attribute_t defined in icm_memorys.h
inline bool icm_scx_memory::add_memory_info(uint32_t mem_space, const char *mem_name, const char *desc_,
                                            icm_memory_t handl, uint64_t min_addr_, uint64_t max_addr_, uint64_t attr)
{
  ICM_SCX_FUNCTION_LOGGING;
  if (mem_name == nullptr)
  {
    SCX_REPORT_WARNING("add_memory_info", "Memory Object with null name - not allowed\n");
    return false;
  }

  auto mem = find_node(mem_name);
  if (mem != nullptr)
  {
    SCX_REPORT_WARNING("add_memory_info", "Memory Object: %s already registered?\n", mem_name);
    return false;
  }
  mem = new icm_scx_mem_object(mem_space, mem_name, desc_, handl, min_addr_, max_addr_, attr);
  memories.push_back(mem);
  return true;
}

/// \brief find a memory object in the list of memories registered
/// \param name_[in] name of the memory object to search
/// \return memory object, if found, else null
inline icm_scx_mem_object *icm_scx_memory::find_node(const char *name_)
{
  for (auto &c : memories)
  {
    if (c->name.compare(name_) == 0)
    {
      return c;
    }
  }
  return nullptr;
}

/// \brief setup_debug_view consumes all the memorys read from the ICM model and pushes those into the CADI view
/// \param icm_in[in] ICM model
/// \return true if there is any memory

inline bool icm_scx_memory::setup_debug_view(scx::scx_debug_target_if *parent_cadi)
{
  ICM_SCX_FUNCTION_LOGGING;
  mCADI = parent_cadi;

  if (memories.size() == 0)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "setup_debug_view", "No memory views reported by the model.");
    return false;
  }

  /// \note for each memory...
  ///    add memory space
  ///     IF NOT HIDDEN,
  ///      add to CADI

  for (auto &memory : memories)
  {
    auto be_set = false;
    unsigned char endian = ICM_CADI_ENDIANNESS_ARCH_DEFINED;
    auto attr = memory->attributes;
    // only check if both BE and LE are set. If set, they will precedence over the possibility
    // of having also specified 'undefined'/ 'arch defined' etc values
    // not testing all the combinations: LE takes priority
    if (((attr & ICM_MEM_ENDIAN_BE8) != 0) || ((attr & ICM_MEM_ENDIAN_BE32) != 0))
    {
      be_set = true;
      endian = ICM_CADI_ENDIANNESS_BE;
    }
    if ((attr & ICM_MEM_ENDIAN_LE) != 0)
    {
      endian = ICM_CADI_ENDIANNESS_LE;
      if (be_set)
      {
        SCX_REPORT_WARNING("setup_debug_view", "Both LE and BE can't be set:%s. Ignoring BE", memory->name.c_str());
      }
    }

    auto is_read_only = ((attr & ICM_MEM_IS_READ_ONLY) != 0);
    auto is_write_only = ((attr & ICM_MEM_IS_WRITE_ONLY) != 0);

    scx::scx_debug_access_type debug_access_attr = scx::SCX_DEBUG_ACCESS_READ_WRITE;
    if (is_write_only)
    {
      debug_access_attr = scx::SCX_DEBUG_ACCESS_WRITE;
    }
    if (is_read_only)
    {
      debug_access_attr = scx::SCX_DEBUG_ACCESS_READ;
    }
    if (is_read_only && is_write_only)
    {
      SCX_REPORT_WARNING("setup_debug_view", "Both Read-only and write-only shouldn't be set :%s. "
                                             "Assuming tht memory is Read-write",
                         memory->name.c_str());
      debug_access_attr = scx::SCX_DEBUG_ACCESS_READ_WRITE;
    }

    uint32_t cadi_space_id_index =
        mCADI->add_space(memory->name.c_str(), memory->desc.c_str(), ICM_MEM_BITS_PER_MAU, memory->max_addr,
                         memory->min_addr, ICM_MEM_WHETHER_PROGRAMMABLE, ICM_MEM_WHETHER_VIRTUAL_MEM, endian);

    mCADI->add_block(memory->name.c_str(), memory->desc.c_str(), memory->min_addr, /*start address*/
                     memory->max_addr,                                             /* end address */
                     debug_access_attr);

    update_map_cadi_to_mem(cadi_space_id_index, memory);

    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "setup_debug_view", "Added Memory:%s", memory->name.c_str());
  }
  return true;
}

/// \brief ICM memory Read function
/// \param num[in] CADI memory index
/// \param datap[out] data pointer
/// \return true, if everything went okay. else false
inline bool icm_scx_memory::read(uint32_t cadi_space, const uint64_t &addr, uint32_t num_bytes, uint8_t *datap)
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t bytes_processed = num_bytes;
  auto mem = get_mem_handle(cadi_space);
  if (mem == nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "read", "called for a null mem object, CADI mem space index=%d", cadi_space);
    return false;
  }
  (*api_mem->read)(mem->mem_handle, addr, num_bytes, &bytes_processed, datap, &icm_err);
  if (icm_err && (icm_scx_decode_error(api_error, icm_err) == ICM_FATAL_ERROR)) return false;
  if (bytes_processed != num_bytes)
  {
    SCX_REPORT_WARNING("memory_read", "Address=0x%" PRIx64 " : Number of bytes requested to read (%d) does not match numbers of bytes "
                               "reportedly read (%d)",
                       addr, num_bytes, bytes_processed);
  }

  return true;
}

/// \brief ICM memory Read function
/// \param num[in] CADI memory index
/// \param datap[in] data pointer
/// \return true, if everything went okay. else false
inline bool icm_scx_memory::write(uint32_t cadi_space, const uint64_t &addr, uint32_t num_bytes, const uint8_t *datap)
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t bytes_processed = num_bytes;
  auto mem = get_mem_handle(cadi_space);
  if (mem == nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "write", "called for a null mem object, CADI mem space index=%d", cadi_space);
    return false;
  }
  (*api_mem->write)(mem->mem_handle, addr, num_bytes, &bytes_processed, datap, &icm_err);
  if (icm_err && (icm_scx_decode_error(api_error, icm_err) == ICM_FATAL_ERROR)) return false;
  if (bytes_processed != num_bytes)
  {
    SCX_REPORT_WARNING("memory_write", "Address=0x%" PRIx64 " : Number of bytes requested for write (%d) does not match numbers of bytes "
                                "reportedly written(%d)",
                       addr, num_bytes, bytes_processed);
  }
  return true;
}

} // icm_scx

#endif
