/// \file        icm_scx_memory_ports.h
/// \brief       Class to simplify usage of ICM memory ports related APIs.
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.
#ifndef icm_scx_memory_ports_h
#define icm_scx_memory_ports_h

#include <iostream>
#include <string>
#include <vector>

#include "icm/api/icm_api.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_memory_ports.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"
#include "icm/integration/icm_scx_memory_ports_typedefs.h"

namespace icm_scx
{
/// \brief Class to process ICM memory port APIs
/// 1. gather all the memory objects reported by the icm memory api
/// 2. forward calls like set_reader/set_writer to ICM model, to convey callbacks for read/write
/// 3. provide means to access all the ports retrieved from the model in the order of reporting.
/// 4. keep list of all the "callback_contexts" reported for callback: These are a boundle of 'port index' and 'actual port r/w handler'
///    'callback_contexts' are maintained as an array and are overwritten every time "set_mem_port_debug_handler" conveys a new handler
class icm_scx_memory_ports
{
public:
  /// \brief consturctor fetches ICM Register and Error APIs handle
  /// \param icm_in[in] ICM model
  icm_scx_memory_ports(icm_model_t icm_in);
  ~icm_scx_memory_ports() { remove_existing_callback_contexts(); }

  std::vector<std::string>* get_mem_port_names_list() { return &memory_port_names_ordered_list; }
  void set_mem_port_debug_handler(icm_scx_mem_port_if *actual_port_impl);

private:
  /// \brief extract all the memory ports from the ICM Model
  /// and store locally for later reference.
  void process_icm_mem_ports_info();

  /// \brief add a memory port entry, as read from ICM model into a vector
  /// \return false in case of error like a duplicate entry
  bool add_port(uint32_t index, const char *name_, const char *desc_);

  /// \brief find a memory port in the list of ports reported by ICM model
  /// \param name_[in] name of the memory port to search
  /// \return true if the port is found
  bool find_port(const char *name_);

  /// \brief set read handler in the ICM model
  bool set_debug_mem_reader(uint32_t index, icm_memory_port_read_handler_fn func, void *context);
  /// \brief set write handler in the ICM model
  bool set_debug_mem_writer(uint32_t index, icm_memory_port_write_handler_fn func, void *context);

  // clear any contexts saved for port-bindings: either to override port-handlers or for final cleanup
  void remove_existing_callback_contexts();

private:
  icm_model_t icm;
  icm_api_memory_ports_t *api_mem_ports;

  icm_api_error_t *api_error;
  scx::scx_debug_target_if *mCADI;
  icm_error_t icm_err;

  /// \brief collection of memory objects
  std::vector<std::string> memory_port_names_ordered_list;

  // simple list of pointers to hold contexts of port handlers
  std::vector<icm_scx_mem_port_context_wrapper *> callback_contexts;
};

} // icm_scx
#include "icm_scx_memory_ports.hpp"

#endif
