/// \file        icm_scx_memory_ports.hpp
/// \brief       implementation of some of the functions in icm_scx_memory_ports.h
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_memory_ports_hpp
#define icm_scx_memory_ports_hpp

#include "icm/integration/icm_scx_memory_ports_callback.h"

namespace icm_scx
{

/// \constructor creates the mem_objects, which holds all the memory objects received from ICM model, gets handle for
/// ICM Memory and error APIs
/// \param icm_in[in] ICM model
inline icm_scx_memory_ports::icm_scx_memory_ports(icm_model_t icm_in)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm = icm_in;
  api_mem_ports = static_cast<icm_api_memory_ports_t *>(icm_api_get(icm, api_id_arm_memory_ports_1_0));
  api_error = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  if (api_mem_ports == nullptr)
  {
    return;
  }
  process_icm_mem_ports_info();
}

/// \brief extract all the memory ports from the ICM Model via ICM memory API
/// and push those into mem_objects.
inline void icm_scx_memory_ports::process_icm_mem_ports_info()
{
  ICM_SCX_FUNCTION_LOGGING;
  uint32_t num_ports = api_mem_ports->count(icm, &icm_err);
  if ((icm_err) || (num_ports == 0))
  {
    if(num_ports == 0){
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "process_icm_mem_ports_info", "No ports exposed.");
    }
    icm_scx_decode_error(api_error, icm_err);
    return;
  }

  const char *name = nullptr;
  const char *description = nullptr;
  for (uint32_t i = 0; i < num_ports; ++i)
  {
    api_mem_ports->info(icm, i, &name, &description, &icm_err);

    if (icm_err)
    {
      icm_scx_decode_error(api_error, icm_err);
    }
    else
    {
      if (add_port(i, name, description) == false)
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "process_icm_mem_ports_info", "Error adding memory port#%d", i);
      }
    }
  }
}

/// \brief add a memory space entry, as read from ICM model
/// \param name_[in] memory object name
/// \param desc_[in] memory object description
/// \param handl[in] memory handle
/// \param wid[in] memory width
/// \param off[in] memory offset
/// \param attr[in] attributes based on icm_mem_attribute_t defined in icm_memorys.h
inline bool icm_scx_memory_ports::add_port(uint32_t index, const char *mem_name, const char *desc_)
{
  ICM_SCX_FUNCTION_LOGGING;
  if (mem_name == nullptr)
  {
    SCX_REPORT_WARNING("add_port", "Memory Port with null name - not allowed\n");
    return false;
  }

  if (find_port(mem_name))
  {
    SCX_REPORT_WARNING("add_port", "%s already registered.\n", mem_name);
    return false;
  }
  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "add_port", "Added %d: %s\n", index, mem_name);
  memory_port_names_ordered_list.push_back(mem_name);
  return true;
}

/// \brief find a memory object in the list of memory ports registered
/// \param name_[in] name of the memory object to search
/// \return memory object, if found, else null
inline bool icm_scx_memory_ports::find_port(const char *name_)
{
  for (const auto &c : memory_port_names_ordered_list)
  {
    if (c.compare(name_) == 0)
    {
      return true;
    }
  }
  return false;
}

inline bool icm_scx_memory_ports::set_debug_mem_reader(uint32_t index, icm_memory_port_read_handler_fn func,
                                                       void *context)
{
  ICM_SCX_FUNCTION_LOGGING;
  api_mem_ports->set_reader(icm, index, context, func, &icm_err);
  if (icm_err)
  {
    icm_scx_decode_error(api_error, icm_err);
    return false;
  }
  return true;
}
inline bool icm_scx_memory_ports::set_debug_mem_writer(uint32_t index, icm_memory_port_write_handler_fn func,
                                                       void *context)
{
  ICM_SCX_FUNCTION_LOGGING;
  api_mem_ports->set_writer(icm, index, context, func, &icm_err);
  if (icm_err)
  {
    icm_scx_decode_error(api_error, icm_err);
    return false;
  }
  return true;
}

inline void icm_scx_memory_ports::set_mem_port_debug_handler(icm_scx_mem_port_if *actual_port_impl)
{
  ICM_SCX_FUNCTION_LOGGING;
  remove_existing_callback_contexts();
  for (uint32_t i = 0; i < memory_port_names_ordered_list.size(); i++)
  {
    auto cc = new icm_scx_mem_port_context_wrapper(i, actual_port_impl);
    callback_contexts.push_back(cc);
    set_debug_mem_reader(i, &c_mem_port_debug_read, cc);
    set_debug_mem_writer(i, &c_mem_port_debug_write, cc);
  }
}
inline void icm_scx_memory_ports::remove_existing_callback_contexts()
{
  ICM_SCX_FUNCTION_LOGGING;
  for (auto cc : callback_contexts)
  {
    delete cc;
  }
  callback_contexts.clear();
}
} // icm_scx

#endif
