/// \file        icm_scx_memory_to_tlm.h
/// \brief       utility class to handle memory port requests and forward to tlm simple sockets.
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
/// usage: refer icm_scx_memory_to_tlm_setup() at end of this file.
//
/// Expectation
/// 1. This class' instance can be used to recieve read/write calls from an ICM model
/// 2. once a call is received, based on "index" in read/write request, the request gets forwarded to
/// a tlm_initiator_socket already registered for that index.

#ifndef icm_scx_memory_to_tlm_h
#define icm_scx_memory_to_tlm_h

#include <iostream>
#include <scx/scx.h>
#include <tlm>

#include <typeinfo>

#include "icm/integration/icm_scx_memory_ports_typedefs.h"
#include <vector>

namespace icm_scx
{
/// \brief base class for any 'socket' that implements '->transport_dbg(tlm_gp)'
class base_forwarder
{
public:
  base_forwarder() = default;
  virtual int transport_dbg(tlm::tlm_generic_payload &gp) { return 0; }
  virtual ~base_forwarder() = default;
};

/// \brief contained which will hold actual 'socket' which can be of any type.
/// code assumes that this socket implements transport_dbg(tlm gp)
template <typename T> class type_forwarder : public virtual base_forwarder
{

public:
  type_forwarder(T *handler) : base_forwarder() { m_handler = handler; }
  virtual ~type_forwarder() = default;
  virtual int transport_dbg(tlm::tlm_generic_payload &gp) override { return (*m_handler)->transport_dbg(gp); }
private:
  T *m_handler;
};

/// \brief an inline utility function to encapsulate any-user-specified-socket
template <typename T> base_forwarder *make_base_forwarder(T &socket)
{
  auto b = new type_forwarder<T>(&socket);
  return b;
}

/// \brief class that handles all the ports' requests.
///  it allows a user to
/// (a) add_socket : to supply a socket for a given port-index
/// (b) implements icm_scx_mem_port_if functions, which receive 'port index' and 'debug memory request'
/// (c) once a debug read/write is received, it is sent to 'process' function, which forwards the calls to n'th socket
class icm_scx_memory_to_tlm : public virtual icm_scx::icm_scx_mem_port_if
{
  std::string m_name;
  std::vector<base_forwarder *> m_sockets;
  tlm::tlm_generic_payload gp;

 public:
  explicit icm_scx_memory_to_tlm(const char *name = nullptr)
  {
    m_name = (name != nullptr) ? name : "icm_scx_memory_to_tlm";
  }
  ~icm_scx_memory_to_tlm() { m_sockets.clear(); }
  void add_socket(uint32_t index, base_forwarder *tgt)
  {
    if (m_sockets.size() <= index)
    {
      m_sockets.resize(index + 1);
    }
    m_sockets[index] = tgt;
  }

  /// \brief icm_scx_mem_port_if function for read
  virtual int mem_port_debug_read(uint32_t index, uint64_t addr, uint32_t num_bytes, uint32_t *bytes_processed,
                                  uint8_t *buffer) override
  {
    return process(true /*read*/, index, addr, num_bytes, bytes_processed, reinterpret_cast<unsigned char *>(buffer));
  }
  /// \brief icm_scx_mem_port_if function for write
  virtual int mem_port_debug_write(uint32_t index, uint64_t addr, uint32_t num_bytes, uint32_t *bytes_processed,
                                   const uint8_t *buffer) override
  {
    return process(false /*write*/, index, addr, num_bytes, bytes_processed,
                   reinterpret_cast<unsigned char *>(const_cast<uint8_t *>(buffer)));
  }

  /// \brief common debug r/w handler
  /// (1) compose a tlm gp
  /// (2) forward to the socket based on index
  /// \param [in] is_read read vs. write
  /// \param [in] index - socket index, to which the request should be forwarded
  /// \param [in] addr - memory address to access
  /// \param [in] num_bytes  - number of bytes to access
  /// \param [out] bytes_processed  - updated to report number of bytes actually updated
  /// \param [inout] buffer - read/write data buffer
  /// \return 0 if success, otherwise failed
  int process(bool is_read, uint32_t index, uint64_t addr, uint32_t num_bytes, uint32_t *bytes_processed,
              unsigned char *buffer)
  {
    if ( (index >= m_sockets.size()) || (m_sockets[index] == nullptr))
    {
      SCX_REPORT_WARNING(m_name.c_str(), "debug memory access attempted to  forward to a port with invalid index: %d. Socket is null", index);
      return 0;
    }
    if (is_read)
      gp.set_command(tlm::TLM_READ_COMMAND);
    else
      gp.set_command(tlm::TLM_WRITE_COMMAND);
    gp.set_address(addr);
    gp.set_data_length(num_bytes);
    gp.set_streaming_width(num_bytes);
    gp.set_data_ptr(buffer);
    gp.set_dmi_allowed(false);
    gp.set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    int bytes = (*m_sockets[index]).transport_dbg(gp);
    if(bytes_processed)*bytes_processed = bytes;
    return (bytes-num_bytes); // bytes not processed. 0=>successfully processed entire request
  }

  /// \brief To map ICM reported port names to generated-port-names
  /// and register ports in the order of ICM port-reporting.
  /// \param [in] parent - name of the tlm wrapper (for error reporting)
  /// \param [in] icm_reported_ports - ports reported by ICM model
  /// \param [in] generated_port_names - port names as generated in the tlm wrapper
  /// \param [in] tlm_outgoing_ports - port/socket handles for generated_port_names
  /// \return true if no errors. false if there are issues in mapping
  bool setup_debug_mem_port_mapping(const char *parent, std::vector<std::string> *icm_reported_ports,
                                    std::vector<std::string> &generated_port_names,
                                    std::vector<icm_scx::base_forwarder *> &tlm_outgoing_ports)
  {
    // The order of generated port names may not be same as ICM model.
    // So we need the following logic.
    //  for each of icm_reported_ports
    // 1 locate 'index' in the "generated port names", which matches icm port name.
    // 2 if not found , report a warning.
    // 3 if found,
    // 3.1 add_socket the initiator_socket at the same index against this ICM model into forwarder
    // 3.2 set 'generated port names's name entry at the index as empty - to rule out duplicity.

    uint32_t generated_port_count = generated_port_names.size();
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, parent, "ICM model generated_port_count =%d", generated_port_count);
    if (icm_reported_ports == nullptr)
    {
      SCX_REPORT_WARNING(parent, "ICM model ports list is null!");
      if (generated_port_count != 0)
      {
        SCX_REPORT_INFO(parent, "ICM model expects 0 ports, whereas  wrapper exposed %d ports!", generated_port_count);
      }
      return false;
    }
    uint32_t icm_port_count = (*icm_reported_ports).size();
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, parent, "ICM model icm_port_count =%d", icm_port_count);
    if (icm_port_count > generated_port_count)
    {
      SCX_REPORT_WARNING(parent, "ICM model expects %d ports, whereas only %d were found in this wrapper",
                         icm_port_count, generated_port_count);
    }
    // go through each ICM port
    for (uint32_t icm_port_index = 0; icm_port_index < icm_port_count; icm_port_index++)
    {
      std::string icm_port_name = (*icm_reported_ports)[icm_port_index];
      auto result1 = std::find(std::begin(generated_port_names), std::end(generated_port_names), icm_port_name);
      size_t gen_port_index = std::distance(std::begin(generated_port_names), result1);
      if (gen_port_index >= generated_port_count)
      {
        add_socket(icm_port_index, nullptr);
        SCX_REPORT_WARNING(parent, "ICM model exposed a port:%s. No match found in this wrapper",
                           icm_port_name.c_str());
      }
      else
      {
        add_socket(icm_port_index, tlm_outgoing_ports[gen_port_index]);
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG,parent, "ICM model port %s mapped to port generated at index %d ",
                         generated_port_names[gen_port_index].c_str(), (int)gen_port_index);
        generated_port_names[gen_port_index] = ""; // hollow it to avoid more matches.
      }
    }
    return true;
  }
};

} // icm_scx
#endif

