/// \file        icm_scx_model.h
/// \brief       Class icm_scx_model definition, to encapsulate ICM Model API usage support
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_model_h
#define icm_scx_model_h

#include "icm/integration/icm_scx_control.h"
#include "icm/integration/icm_scx_control_callback.h"
#include "icm/integration/icm_scx_disasm.h"
#include "icm/integration/icm_scx_error.h"
#include "icm/integration/icm_scx_memory.h"
#include "icm/integration/icm_scx_memory_ports.h"
#include "icm/integration/icm_scx_model_info.h"
#include "icm/integration/icm_scx_register.h"

namespace icm_scx
{
/// note top level scx->icm implementation layer would access lots of
/// variables/functions from icm_scx_model. So instead of creating a tons of
/// get/set functions, giving that class permission to access those variables
class icm_scx_debug_impl;

/// \brief class to contain all the ICM API clients in one place.
/// at present, register and modelinfo APIs have been implemented in separate files
class icm_scx_model
{
public:
  icm_scx_model() = delete;
  /// \brief consructor creates various supporting class objects like regs, model_info, control, etc
  /// \param icm_[in] Handle to ICM model
  icm_scx_model(icm_model_t icm_)
      : icm(icm_), model_info(icm_), control(icm_), regs(icm_), memories(icm_), disasm(icm_), memory_ports(icm_)
  {
    ICM_SCX_FUNCTION_LOGGING;
    api_model_info = static_cast<icm_api_model_info_t *>(icm_api_get(icm, api_id_arm_modelinfo_1_0));
    api_error = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  }

  /// \brief destructor presently does nothing.
  virtual ~icm_scx_model() {}

  /// \brief setup_debug_view receives a "CADI" object and uses it to create various debug views via its supporting ICM
  /// client objects (like regs, memories, disasm)
  void setup_debug_view(scx::scx_debug_target_if *_mCADI)
  {
    ICM_SCX_FUNCTION_LOGGING;
    mCADI = _mCADI;
    supports_registers = regs.setup_debug_view(mCADI, model_info.get_target_type());
    supports_memory = memories.setup_debug_view(mCADI);
    supports_disassembler = disasm.setup_debug_view(mCADI);
  }

private:
  icm_model_t icm;
  scx::scx_debug_target_if *mCADI;
  icm_error_t icm_err;

protected:
  friend class icm_scx_debug_impl;

  /// \brief handler for ICM model_info API
  icm_scx_model_info model_info;
  /// \brief api_model_info is
  icm_api_model_info_t *api_model_info;
  /// \brief handler for ICM control API
  icm_scx_control control;
  /// \brief handler for ICM Model registers/APIs
  icm_scx_register regs;
  /// \brief handler for ICM Model memory/APIs
  icm_scx_memory memories;
  /// \brief handle for ICM disassembler APIs
  icm_scx_disasm disasm;
  /// \brief handler for ICM Model memory port APIs
  icm_scx_memory_ports memory_ports;
  /// \brief error api handle
  icm_api_error_t *api_error;

  bool supports_registers;
  bool supports_memory;
  bool supports_disassembler;
};
} // icm_scx
#endif
