/// \file        icm_scx_model_info.h
/// \brief       Class to simplify usage of ICM model_info APIs
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_model_info_h
#define icm_scx_model_info_h

#include <string>

#include "icm/api/icm_api.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_model_info.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{
/// \brief icm_scx_model_info implements usage of ICM model_info APIs.
/// to read and retain model name, descriptin, and number of children
class icm_scx_model_info
{
public:
  /// \brief constructore setups some default for model_name/description etc, fetches handles for ICM model_info and
  /// error APIs
  /// \param icm_in[in] handle to ICM Model
  icm_scx_model_info(icm_model_t icm_in)
      : icm(icm_in), model_name("undefined"), model_desc("undefined"), target_name(""), target_type("")
  {
    ICM_SCX_FUNCTION_LOGGING;
    api_model_info = static_cast<icm_api_model_info_t *>(icm_api_get(icm, api_id_arm_modelinfo_1_0));
    api_err = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));

    child_count = 0;
    const char *p_model_name;
    const char *p_model_desc;

    if (api_model_info)
    {
      api_model_info->info(icm, &p_model_name, &p_model_desc, &child_count, &icm_err);

      if (icm_err)
        icm_scx_decode_error(api_err, icm_err); /* decode_error is a common utility function to process icm_error*/
      if (p_model_name == nullptr)
      {
        SCX_REPORT_WARNING("icm_scx_model_info", "icm_api_model_info_t returned null for model-name.");
        model_name = "undefined";
      }
      else
      {
        model_name = p_model_name;
      }

      if (p_model_desc != nullptr) model_desc = p_model_desc;

      const char *value = get_model_property("componentName");
      if (value != nullptr) target_name = value;

      value = get_model_property("componentType");
      if (value != nullptr) target_type = value;

      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_model_info", "%s:componentType=%s,componentName=%s", model_name.c_str(),
                           target_name.c_str(), target_type.c_str());
    }
  }
  ~icm_scx_model_info() {}
  std::string get_instance_name() const { return model_name; }
  std::string get_target_name() const { return target_name; }
  std::string get_target_type() const { return target_type; }
  std::string get_model_description() const { return model_desc; }
  int get_child_count() const { return child_count; }

  const char *get_model_property(const char *prop_name)
  {
    const char *value = nullptr;
    if (api_model_info)
    {
      value = api_model_info->get_property(icm, prop_name, &icm_err);
      if (icm_err) icm_scx_decode_error(api_err, icm_err);
    }
    return value;
  }

private:
  icm_model_t icm;
  icm_error_t icm_err;
  icm_api_error_t *api_err;

  icm_api_model_info_t *api_model_info;
  std::string model_name;
  std::string model_desc;
  std::string target_name;
  std::string target_type;
  int child_count;
};
} // icm_scx

#endif
