/// \file        icm_scx_register.h
/// \brief       Class to simplify usage of ICM register APIs (icm_scx_register) and defines a local class
/// icm_scx_reg_object to encapsulate register-objects received from ICM Model
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.
#ifndef icm_scx_register_h
#define icm_scx_register_h

#include <iostream>
#include <map>
#include <string>
#include <vector>

#include "icm/api/icm_api.h"
#include "icm/api/icm_error.h"
#include "icm/api/icm_registers.h"
#include "icm/api/icm_types.h"

#include "icm/integration/icm_scx_error.h"

namespace icm_scx
{
/// \brief icm_scx_reg_object is a container for register objects (a group, or register, or a bitfield) received from
/// ICM
/// Model. Each register object holds a vector of 'children', which are its sub-'nodes' e.g. a group("Core")->children
/// contain all registers of a group ("Core0") such as "R0","R1","CPSR". a group("Core")->children["CPSR"]->children
/// contains bitfields related to Core.CPSR registers
class icm_scx_reg_object
{
public:
  /// \brief constructor for icm_scx_reg_object
  /// \param name_[in] register object name
  /// \param desc_[in] register object description
  /// \param handl[in] register handle
  /// \param wid[in] register width
  /// \param off[in] register offset
  /// \param attr[in] attributes based on icm_reg_attribute_t defined in icm_registers.h
  icm_scx_reg_object(const char *name_, const char *desc_ = "not defined", icm_reg_t handl = nullptr, uint32_t wid = 0,
                     uint32_t off = 0, uint64_t attr = 0);
  ~icm_scx_reg_object();

protected:
  /// \note icm_scx_register will need access to various functions of this class to add/access register objects
  friend class icm_scx_register;

  /// \brief add a new register object in the children/hierarchy after splitting name based on "." delimiter
  /// \note: arguments similar to constructor
  /// \return true if successful, false in case of failure to add
  bool add_reg_object(const char *name_, const char *desc_, icm_reg_t handle, uint32_t wid, uint32_t offset,
                      uint64_t attr);

private:
  /// \brief add a new register "group" object in the children/hierarchy
  /// \note: arguments similar to add_reg_object
  /// \return handle to the object just added
  icm_scx_reg_object *add_new_group(const char *name_, const char *desc_, icm_reg_t handl, uint32_t wid, uint32_t off,
                                    uint64_t attr);

  /// \note: similar to add_reg_object
  icm_scx_reg_object *add_new_register(const char *g_name, const char *r_name, const char *desc_, icm_reg_t handl,
                                       uint32_t wid, uint32_t off, uint64_t attr);

  /// \note: similar to add_reg_object
  icm_scx_reg_object *add_new_bitfield(const char *g_name, const char *r_name, const char *bf_name, const char *desc_,
                                       icm_reg_t handl, uint32_t wid, uint32_t off, uint64_t attr);

  /// \brief find a register object in the current hierarchy
  /// \param name_[in] name of the register group or register or bitfield to search
  /// \return register object, if found, else null
  icm_scx_reg_object *find_node(const char *name_);

  /// \brief split a register object name based on a delimiter
  /// \param full_name[in] - full register object name like group.register.bitfield
  /// \param del[in] - delimiter
  /// \param group_name[out] - register group name
  /// \param register_name[out] - register name
  /// \param bitfield_name[out] - register bitfield name
  /// \return number of splits made based on the delimiter (e.g. group =>1, group.reg => 2, group.reg.bitfield = 3)
  int split_reg_object_name(const char *full_name, const char *delimiter, std::string &group_name,
                            std::string &register_name, std::string &bitfield_name);

  /// \brief register object's name (can be group, register, bitfield)
  std::string name;
  /// \brief an optional description
  std::string desc;
  /// \brief  width, relevant only for a register or a bitfield
  int32_t width;
  /// \brief starting offset assuming little endian, relevant only for a register or a bitfield
  int32_t offset;

  /// \brief attributes based on icm_reg_attribute_t defined in icm_registers.h
  int64_t attributes;

  /// \brief handle to the register object which provide ICM register APIs
  icm_reg_t reg_handle;

  /// \brief collection of register objects at this level. Each child will have its own children.
  /// e.g. top level icm_scx_reg_object will be groups
  /// each group's children icm_scx_reg_object will be registers
  /// and each registers's children icm_scx_reg_object will be bitfields (if any).

  std::vector<icm_scx_reg_object *> children;
};

/// \brief Class to process ICM register APIs
/// 1. gather all the register objects reported by the icm register api
/// 2. push all the groups/regs/bitfields into CADI
/// 3. Provide means to get access to "register handle" for a given "cadi index"
class icm_scx_register
{
public:
  /// \brief consturctor fetches ICM Register and Error APIs handle
  /// \param icm_in[in] ICM model
  icm_scx_register(icm_model_t icm_in);
  ~icm_scx_register() { map_cadi_to_reg.clear(); }

  /// \brief add all the groups/registers/bitfields to CADI
  /// \return true if there is any register
  bool setup_debug_view(scx::scx_debug_target_if *parent_cadi, const std::string& target_type);

  /// \brief register read function
  bool read(uint32_t num, uint8_t *datap);

  /// \brief register write function
  bool write(uint32_t num, const uint8_t *datap);

/// \brief read pc_memspace register
  uint32_t read_pc_memspace();

private:
  /// \brief retrieve icm_reg_t handle for a given CADI index
  /// \param cadi_index[in]  register index requested at CADI layer
  /// \return register object, if found within the register objects map, else nullptr
  icm_reg_t get_reg_handle(int cadi_index)
  {
    auto ret = map_cadi_to_reg.find(cadi_index);
    if (ret != map_cadi_to_reg.end())
    {
      return ret->second;
    }
    return nullptr;
  }

  /// \brief retrieve icm_reg_t handle for a given CADI index
  /// \param cadi_index[in]  register index requested at CADI layer
  /// \return register object, if found within the register objects map, else nullptr
  bool update_map_cadi_to_reg(int cadi_index, icm_reg_t reg_handle)
  {
    if (map_cadi_to_reg.find(cadi_index) != map_cadi_to_reg.end())
    {
      return false;
    }

    map_cadi_to_reg[cadi_index] = reg_handle;
    return true;
  }

  /// \brief extract all the registers from the ICM Model via ICM register API
  /// and push those into reg_objects.
  void process_icm_reg_info();

private:
  icm_model_t icm;
  icm_api_registers_t *api_regs;

  icm_api_error_t *api_error;
  scx::scx_debug_target_if *mCADI;
  icm_error_t icm_err;

  /// \brief a mapping of register-handlers and their "CADI registration index"
  std::map<int, icm_reg_t> map_cadi_to_reg;

  /// \brief reg_objects holds all the register objects managed by this class
  icm_scx_reg_object reg_objects;

  int num_registers;

  icm_reg_t m_pc_memspace_reg_handle;
};

} // icm_scx
#include "icm_scx_register.hpp"

#endif
