/// \file        icm_scx_register.hpp.hpp
/// \brief       implementation of some of the functions in icm_scx_register.h
///
/// \copyright   Copyright 2017 ARM Limited. All rights reserved.
///
/// \note This file is not part of a stable public API and may change without notice.
/// Please refer to your legal agreement with ARM Limited to
/// understand what you are licensed to do with this code.

#ifndef icm_scx_register_hpp
#define icm_scx_register_hpp

namespace icm_scx
{
/// \constructor creates the reg_objects, which holds all the register objects received from ICM model, gets handle for
/// ICM Registrer and error APIs
/// \param icm_in[in] ICM model
inline icm_scx_register::icm_scx_register(icm_model_t icm_in) : reg_objects("register")
{
  ICM_SCX_FUNCTION_LOGGING;
  icm = icm_in;
  api_regs = static_cast<icm_api_registers_t *>(icm_api_get(icm, api_id_arm_registers_1_0));
  if (api_regs == nullptr)
  {
    return;
  }
  api_error = static_cast<icm_api_error_t *>(icm_api_get(icm, api_id_arm_error_1_0));
  process_icm_reg_info();
}

/// \brief setup_debug_view consumes all the registers read from the ICM model and pushes those into the CADI view
/// \param icm_in[in] ICM model
/// \return true if there is any register
inline bool icm_scx_register::setup_debug_view(scx::scx_debug_target_if *parent_cadi, const std::string& target_type)
{
  ICM_SCX_FUNCTION_LOGGING;
  mCADI = parent_cadi;

  m_pc_memspace_reg_handle = nullptr;
  std::string pc_memspace_reg_name;

  if (num_registers == 0)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "setup_debug_view", "No registers reported by the model.");
    return false;
  }
  /// \note for each group...
  ///        for each register in this group...
  ///            for each register bitfield in this group...
  ///                IF NOT HIDDEN,
  ///                add to CADI and process children

  for (auto &grp : reg_objects.children)
  {
    if (grp->attributes & ICM_REG_INVISIBLE)
    {
      SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view", "Hidden reg-group skipped: %s",
                           grp->name.c_str());
      continue;
    }
    // added_group is added to handle scenarios where a group may be emptry or have only hidden registers.
    // This flag gets reset for each group.
    // Once we identify at least one register that is added to CADI, we set this flag, besides adding the group to CADI.

    bool added_group = false;
    for (auto &reg : grp->children)
    {
      // test whether a register is marked as pc_memspace - do this before checking for Hidden - since, a hidden
      // register is not at all processed.
      if (reg->attributes & ICM_REG_PC_MEMSPACE)
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view:: PC_MEMSPACE found.", "name=%s",
                             reg->name.c_str());

        if (m_pc_memspace_reg_handle != nullptr)
        {
          SCX_REPORT_WARNING("icm_scx_register:setup_debug_view",
                             "More than one register tagged as ICM_REG_PC_MEMSPACE. Newer one will be used: %s",
                             reg->name.c_str());
        }
        m_pc_memspace_reg_handle = reg->reg_handle;
        pc_memspace_reg_name = reg->name;
        // skip this register, since FMRuntime/SCX adds this register on its own.
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view",
                             "Register %s will not be added to CADI (ICM_REG_PC_MEMSPACE).", reg->name.c_str());
        continue;
      }
      if (reg->attributes & ICM_REG_INVISIBLE)
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view", "Hidden register skipped: %s",
                             reg->name.c_str());
        continue;
      }

      if (!added_group)
      {
        // int grp_id =
        mCADI->add_group(grp->name, grp->desc);
        added_group = true;
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view", "Adding Group %s to cadi.",
                             grp->name.c_str());
      }
      int reg_id = mCADI->add_register(reg->name, reg->desc, reg->width, scx::SCX_DEBUG_ACCESS_READ_WRITE);
      if (false == update_map_cadi_to_reg(reg_id, reg->reg_handle))
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view",
                             "Error Adding %s to cadi. Potentially a duplicate!", reg->name.c_str());
        continue;
      }

      if (reg->attributes & ICM_REG_PC) mCADI->register_is_pc(reg_id);
      if (reg->attributes & ICM_REG_SP) mCADI->register_is_sp(reg_id);
      if (reg->attributes & ICM_REG_LR) mCADI->register_is_lr(reg_id);
      if (reg->attributes & ICM_REG_CPSR) mCADI->register_is_status(reg_id);
      if (reg->children.size() > 0)
      {
        for (auto &bf : reg->children)
        {
          mCADI->register_is_compound(reg_id);
          if (bf->attributes & ICM_REG_INVISIBLE)
          {
            SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view",
                                 "Hidden register bitfield skipped: %s", reg->name.c_str());
            continue;
          }
          // int reg_bf_id =
          mCADI->add_bitfield(reg_id, bf->name, bf->desc, bf->offset, bf->width);
        }
      }
    }
  }
  if (m_pc_memspace_reg_handle == nullptr)
  {
    if(target_type == "Core") {
       SCX_REPORT_WARNING("icm_scx_register:setup_debug_view",
                          "No PC_MEMSPACE register declared for a model of type=Core. PC_MEMSPACE will always be 0.");
    }
  }
  else
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view", "ICM_REG_PC_MEMSPACE found:%s",
                         pc_memspace_reg_name.c_str());
  }

  return true;
}

/// \brief extract all the registers from the ICM Model via ICM register API
/// and push those into reg_objects.
inline void icm_scx_register::process_icm_reg_info()
{
  ICM_SCX_FUNCTION_LOGGING;
  num_registers = api_regs->count(icm, &icm_err);
  if ((icm_err) || (num_registers == 0))
  {
    icm_scx_decode_error(api_error, icm_err);
    return;
  }

  const char *name = nullptr;
  const char *description = nullptr;
  uint32_t width;
  uint32_t offset;
  uint64_t reg_attr;
  for (int i = 0; i < num_registers; ++i)
  {
    icm_reg_t reg_handle = api_regs->info(icm, i, &name, &description, &width, &offset, &reg_attr, &icm_err);
    if (icm_err)
    {
      icm_scx_decode_error(api_error, icm_err);
    }
    else
    {
      if (reg_objects.add_reg_object(name, description, reg_handle, width, offset, reg_attr) == false)
      {
        SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register::setup_debug_view", "Error adding reg object(%s)", name);
      }
    }
  }
}

/// \brief ICM register Read function
/// \param num[in] CADI register index
/// \param datap[out] data pointer
/// \return true, if everything went okay. else false
inline bool icm_scx_register::read(uint32_t num, uint8_t *datap)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm_reg_t reg = get_reg_handle(num);
  if (reg == nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register::read", "called for a null reg object, index=%d", num);
    return false;
  }
  (*api_regs->read)(reg, (uint64_t *)datap, &icm_err);
  if (icm_err && (icm_scx_decode_error(api_error, icm_err) == ICM_FATAL_ERROR)) return false;
  return true;
}

/// \brief ICM register Read function
/// \param num[in] CADI register index
/// \param datap[in] data pointer
/// \return true, if everything went okay. else false
inline bool icm_scx_register::write(uint32_t num, const uint8_t *datap)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm_reg_t reg = get_reg_handle(num);
  if (reg == nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register::write", "called for a null reg object, index=%d", num);
    return false;
  }
  (*api_regs->write)(reg, (uint64_t *)datap, &icm_err);
  if (icm_err && (icm_scx_decode_error(api_error, icm_err) == ICM_FATAL_ERROR)) return false;
  return true;
}

/// icm_scx_registers_t implementation

/// \brief constructor for icm_scx_reg_object
/// \param name_[in] register object name
/// \param desc_[in] register object description
/// \param handl[in] register handle
/// \param wid[in] register width
/// \param off[in] register offset
/// \param attr[in] attributes based on icm_reg_attribute_t defined in icm_registers.h
inline icm_scx_reg_object::icm_scx_reg_object(const char *name_, const char *desc_, icm_reg_t handl, uint32_t wid,
                                              uint32_t off, uint64_t attr)
{
  name = (name_ != nullptr) ? name_ : "<empty>";
  desc = (desc_ != nullptr) ? desc_ : "";
  width = wid;
  offset = off;
  reg_handle = handl;
  attributes = attr;
}

/// \brief destructor cleanup all the children register objects
inline icm_scx_reg_object::~icm_scx_reg_object() { children.clear(); }

/// \brief add a new register object in the children/hierarchy after splitting name based on "." delimiter
/// \note: arguments similar to constructor
/// \return true if successful, false in case of failure to add
inline bool icm_scx_reg_object::add_reg_object(const char *name_, const char *desc_, icm_reg_t handle, uint32_t wid,
                                               uint32_t offset, uint64_t attr)
{
  std::string grp, reg, bitfield;
  int split = split_reg_object_name(name_, ".", grp, reg, bitfield);
  if (split == 1) return (add_new_group(grp.c_str(), desc_, handle, wid, offset, attr) != nullptr);
  if (split == 2) return (nullptr != add_new_register(grp.c_str(), reg.c_str(), desc_, handle, wid, offset, attr));
  if (split >= 3)
    return (nullptr != add_new_bitfield(grp.c_str(), reg.c_str(), bitfield.c_str(), desc_, handle, wid, offset, attr));

  SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:add_reg_object",
                       "%s :Too many fields : based on \".\" delimiter. Expected format "
                       "Group.RegisterName.Bitfield",
                       name_);
  return false;
}

/// \brief add a new register "group" object in the children/hierarchy
/// \note: arguments similar to add_reg_object
/// \return handle to the object just added
inline icm_scx_reg_object *icm_scx_reg_object::add_new_group(const char *name_, const char *desc_, icm_reg_t handl,
                                                             uint32_t wid, uint32_t off, uint64_t attr)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm_scx_reg_object *g = find_node(name_);
  if (g != nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:add_new_group",
                         "Register Object: %s already found in the list\n", name_);
    return g;
  }
  g = new icm_scx_reg_object(name_, desc_, handl, wid, off, attr);
  children.push_back(g);
  return g;
}

/// \note: similar to add_reg_object
inline icm_scx_reg_object *icm_scx_reg_object::add_new_register(const char *g_name, const char *r_name,
                                                                const char *desc_, icm_reg_t handl, uint32_t wid,
                                                                uint32_t off, uint64_t attr)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm_scx_reg_object *group = find_node(g_name);
  if (group == nullptr)
  {
    // all except invisible bit will get copied to group.
    group = add_new_group(g_name, "<missing group description>", nullptr, wid, off, attr & ~ICM_REG_INVISIBLE);
  }
  icm_scx_reg_object *r = group->find_node(r_name);
  if (r != nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:add_new_register",
                         "Register Object: %s already found in the list\n", name.c_str());
    return r;
  }
  r = new icm_scx_reg_object(r_name, desc_, handl, wid, off, attr);
  group->children.push_back(r);
  return r;
}

/// \note: similar to add_reg_object
inline icm_scx_reg_object *icm_scx_reg_object::add_new_bitfield(const char *g_name, const char *r_name,
                                                                const char *bf_name, const char *desc_, icm_reg_t handl,
                                                                uint32_t wid, uint32_t off, uint64_t attr)
{
  ICM_SCX_FUNCTION_LOGGING;
  icm_scx_reg_object *reg = add_new_register(g_name, r_name, "<missing register description>", nullptr, wid, off,
                                             attr & ~ICM_REG_INVISIBLE); // all except invisible bit get copied
  icm_scx_reg_object *bf_handle = reg->find_node(bf_name);
  if (bf_handle != nullptr)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:add_new_bitfield",
                         "Register Object: %s already found in the list\n", name.c_str());
    return bf_handle;
  }
  bf_handle = new icm_scx_reg_object(bf_name, desc_, handl, wid, off, attr);
  reg->children.push_back(bf_handle);
  return bf_handle;
}

/// \brief find a register object in the current hierarchy
/// \param name_[in] name of the register group or register or bitfield to search
/// \return register object, if found, else null
inline icm_scx_reg_object *icm_scx_reg_object::find_node(const char *name_)
{
  for (auto &c : children)
  {
    if (c->name.compare(name_) == 0)
    {
      return c;
    }
  }
  return nullptr;
}

/// \brief split a group.register.bitfield based on delimiter
inline int icm_scx_reg_object::split_reg_object_name(const char *full_name, const char *del, std::string &g,
                                                     std::string &r, std::string &b)
{
  if (!full_name) return 0; /* bad case: null input*/
  std::string s = full_name;
  std::string delimiter = del;

  auto first_pos = s.find(delimiter);

  if (first_pos == 0) return 0; /* bad case: name starts with a delimiter*/
  g = s.substr(0, first_pos);

  if (first_pos == std::string::npos) return 1; /* zero delimiter? everything in the input is register group name*/

  auto second_pos = s.find(delimiter, first_pos + 1);
  r = s.substr(first_pos + 1, second_pos - (first_pos + 1));

  if (r.length() == 0) return 1; /*"register" part is zero length?  it's as good as specifying only group.*/

  if (second_pos == std::string::npos)
    return 2; /* 1 delimiter? so it was group.reg. This came after checking that "register" part is non-zero length */

  auto third_pos = s.find(delimiter, second_pos + 1);
  b = s.substr(second_pos + 1, third_pos - (second_pos + 1));
  if (b.length() == 0) return 2; /*"bitfield" part is zero length?  it's as good as specifying only "group.register".*/

  if (third_pos != std::string::npos)
  {
    SCX_REPORT_INFO_VERB(scx::SCX_DEBUG, "icm_scx_register:setup_debug_view",
                         ":Register Object:%s: Ignored name after bitfield. ", full_name);
    return 3;
  }

  return 0;
}

/// \brief read pc_memspace register
/// \return 0 if the m_pc_memspace_reg_handle is null, else read the actual register and return value
inline uint32_t icm_scx_register::read_pc_memspace()
{
  if (m_pc_memspace_reg_handle == nullptr || api_regs == nullptr) return 0;
  uint64_t value = 0;
  (*api_regs->read)(m_pc_memspace_reg_handle, (uint64_t *)&value, nullptr);
  return (uint32_t)value; // pc_memspace is expected to be 32bit only.
}

} // icm_scx

#endif
