//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// Please refer to your legal agreement with ARM Limited to
// understand what you are licensed to do with this example code.
//-------------------------------------------------------------------

#include <iomanip>
#include <sstream>
#include <iostream>

#include "BasicUART.h"

using namespace tlm;
using namespace sc_core;

namespace ARM
{
namespace Models 
{

BasicUART::BasicUART(sc_module_name p_scName, std::ostream &stream_, std::string prefix_, int width_ ) :
    pending_printing(false), prefix(prefix_), stream(stream_),
    clock("clock_uart"),
    tSkt("tSkt_uart", *this, (&BasicUART::nb_transport_fw),  ARM::TLM::PROTOCOL_ACE, width_),
    directTskt("directTskt"),
    r_current(NULL),
    b_current(NULL),
    r_ready_pending(false),
    b_ready_pending(false),
    now(0)
{
    // clock pos evt is used for evaluate phase
    SC_METHOD(clock_posedge);
    sensitive << clock.pos();
    dont_initialize();

    // clock neg evt is use for propagate phase
    SC_METHOD(clock_negedge);
    sensitive << clock.neg();
    dont_initialize();

    directTskt.register_transport_dbg(this, &BasicUART::transport_dbg);
    
}

bool BasicUART::add_character(char chr)
{
    if (chr < 0x20 || chr >= 0x7f)
    {
        std::ostringstream escape;
        uint32_t cc = chr;
        cc &= 0xff;
        escape << '<' << std::hex << std::setw(2) << std::setfill('0') << cc << '>';

        line += escape.str();
    } else {
        line += chr;
    }

    pending_printing = true;

    /* Newline or EOT, print the accumulate line */
    if (chr == 0x0a || chr == 0x04)
    {
        flush();
        clear_line();
    }

    /* Received EOT? */
    return chr == 0x04;
}

void BasicUART::flush()
{
    if (pending_printing)
    {
        stream << prefix << line << '\n';
        stream.flush();

        pending_printing = false;
    }
}

void BasicUART::clear_line()
{
    line.clear();
    pending_printing = false;
}

BasicUART::~BasicUART()
{
    flush();
}

tlm_sync_enum BasicUART::nb_transport_fw(PayloadT& payload, PhaseT& phase)
{
    tlm_sync_enum ret_val;
    switch (phase)
    {
        case ARM::AXI4::AW_VALID:
            w_queue.push_back(TimedPayloadT(&payload, now));
            payload.ref();
            phase   = ARM::AXI4::AW_READY;
            ret_val = TLM_UPDATED;
            break;

        case ARM::AXI4::W_VALID_LAST:
        {
            wd_queue.push_back(TimedPayloadT(&payload, now)); // no delay on write response
            payload.ref();
            phase = ARM::AXI4::W_READY;
            ret_val = TLM_UPDATED;
        }
        //fall thru
        case ARM::AXI4::W_VALID:

            phase   = ARM::AXI4::W_READY;
            ret_val = TLM_UPDATED;
            break;

       case ARM::AXI4::B_READY:
            b_ready_pending  = false;
            ret_val = TLM_ACCEPTED;
            break;

        case ARM::AXI4::AR_VALID:
        {
            for (unsigned i = 0; i < payload.get_beat_count(); i++) {
                r_queue.push_back(std::pair<PayloadT*,uint64_t>(&payload, now));
                payload.ref();
            }
 
            uint8_t* temp_data = new uint8_t[payload.get_data_length()];
            memset(temp_data, 0, payload.get_data_length());
            payload.read_in(temp_data);

            phase   = ARM::AXI4::AR_READY;
            ret_val = TLM_UPDATED;
            break;
        }

        case ARM::AXI4::R_READY:
            r_ready_pending = false;
            ret_val = TLM_ACCEPTED;
            break;
        case ARM::AXI4::RACK:
        case ARM::AXI4::WACK:
            ret_val = TLM_ACCEPTED;
            break;
        default:
            ret_val = TLM_ACCEPTED;
            assert(!"uncaught command");
    }
    return ret_val;
}

void BasicUART::clock_posedge()
{
    #ifdef UBD_ENABLED
    static int sc_stop_reported_once = false;
    if(sc_stop_reported_once)
        return;
    #endif
    now++; // advance time
   
    // things stay in the queue until their timestamp passes
    if (!r_ready_pending && !r_queue.empty())
    {
        if (now >= r_queue.front().second)
        {
            r_current = r_queue.front().first;
            r_queue.pop_front();
        }
    }        

    if (!b_ready_pending && !w_queue.empty() && !wd_queue.empty())
    {
        if (now >= wd_queue.front().second)
        {
            assert(w_queue.front().first == wd_queue.front().first);
            b_current = w_queue.front().first;
            PayloadT *wd_current = wd_queue.front().first;
            wd_current->unref();
            if (b_current->get_data_length() == 1)
            {
                unsigned char uart_char;
                b_current->read_out(&uart_char);
                if (add_character (uart_char))
                {
#ifdef UBD_ENABLED
                    if(!sc_stop_reported_once){
                        SC_REPORT_WARNING("UART","Received end-of-simulation character");
                        sc_stop_reported_once=true;
                    }
                    /*UBD can not afford to see sc_stop being called, as calling it would result in CADI connection being lost*/
                    /* sc_core::sc_stop(); */
                    sleep(1);
		    return;
#else
		   sc_core::sc_stop();
#endif
                }
            }
            w_queue.pop_front();
            wd_queue.pop_front();
        }
    }
}

void BasicUART::clock_negedge()
{
    PhaseT        phase;
    tlm_sync_enum ret_val;

    if (r_current)
    {
        // check to see if the queue has more beats
        if (r_queue.size())
        {
            // If there are more beats in the queue and r_current is not the same as the the one at the
            // head of the queue then it must be the last from this transaction. 
            phase = (r_current != r_queue.front().first) ? ARM::AXI4::R_VALID_LAST:ARM::AXI4::R_VALID;
        }
        else 
        {
            // If the queue is empty then r_current is the last beat from this transaction since we pushed
            // all of them at once in nb_transport_fw
            phase = ARM::AXI4::R_VALID_LAST;
        }	
        r_current->set_resp(ARM::AXI4::RESP_OKAY);
        r_ready_pending = true;
        
        ret_val = tSkt.nb_transport_bw(*r_current, phase);
        
        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::R_READY);
            r_ready_pending = false;
        }
        r_current->unref();
        r_current = NULL;
    }


    if (b_current)
    {
        phase = ARM::AXI4::B_VALID;
        b_current->set_resp(ARM::AXI4::RESP_OKAY);
        
        b_ready_pending = true;
        ret_val = tSkt.nb_transport_bw(*b_current, phase);

        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::B_READY);
            b_ready_pending = false;
        }
        b_current->unref();
        b_current = NULL;
    }
}

unsigned int BasicUART::transport_dbg(tlm::tlm_generic_payload &trans)
{
    uint8_t *data_ptr = trans.get_data_ptr();

    // You need to be able to populate a read-only memory from the ELF loader
    // so immediate-writes do write

    if(trans.get_command() == tlm::TLM_WRITE_COMMAND)
    {
        // Do Nothing
    }
    else
    {
        memset(data_ptr, 0, trans.get_data_length());
    } 
    trans.set_response_status( tlm::TLM_OK_RESPONSE );

    return trans.get_data_length();
}


}
}
