//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// Please refer to your legal agreement with ARM Limited to
// understand what you are licensed to do with this example code.
//-------------------------------------------------------------------

#ifdef _WIN32
#include "elfdefinitions.h"
#define ELFMAG "\177ELF"
#else
#include <elf.h>
#endif
#include <cassert>
#include <zlib.h>
#include <cstring>
#include <cstdio>
#include <cstdlib>
#include <stdexcept>
#include "RAMBlock.h"
#include "ElfLoader.h"

namespace ARM
{
namespace Models
{

static void ElfLoaderLoadData (const char* filedata, RAMBlock* ram_block, f_dbg_write_t *fPTR)
{
    int elf_class = filedata[EI_CLASS];
    if (elf_class == ELFCLASS64)
    {
        Elf64_Ehdr* hdr = (Elf64_Ehdr*)filedata;
        Elf64_Phdr* phdrarr = (Elf64_Phdr*)&filedata[hdr->e_phoff];
        int i;
        for (i = 0; i < hdr->e_phnum; i++)
        {
            Elf64_Phdr* phdr = &phdrarr[i];
            if (phdr->p_type != PT_LOAD)
                continue;

            const char* data = &filedata[phdr->p_offset];
            //  Embedded tools like debugger will ignore any p_paddr values. If the p_paddr fields 
            //  is not ignored, wide spread common understanding is: if p_paddr = 0, ignore the value 
            //  and use p_vaddr instead. If p_paddr !=0 use the address to load the content for.
            uint64_t address;
            if (phdr->p_paddr != 0)
                address = phdr->p_paddr;
            else
                address = phdr->p_vaddr;
            uint64_t size = phdr->p_filesz;
            uint64_t zeros = phdr->p_memsz - size;
            if(ram_block)ram_block->write(address, (unsigned char*)data, size);
            if(fPTR)(*fPTR)(address, (unsigned char*)data, size);
            if (zeros)
            {
                void* zero_data = calloc(zeros, 1);
                if(ram_block)ram_block->write(address + size, (unsigned char*)zero_data, zeros);
                if(fPTR)(*fPTR)(address + size, (unsigned char*)zero_data, zeros);
                free(zero_data);
            }
        }
    }
    else if(elf_class == ELFCLASS32)
    {
        Elf32_Ehdr* hdr = (Elf32_Ehdr*)filedata;
        Elf32_Phdr* phdrarr = (Elf32_Phdr*)&filedata[hdr->e_phoff];
        int i;
        for (i = 0; i < hdr->e_phnum; i++)
        {
            Elf32_Phdr* phdr = &phdrarr[i];
            if (phdr->p_type != PT_LOAD)
                continue;

            const char* data = &filedata[phdr->p_offset];
            //  Embedded tools like debugger will ignore any p_paddr values. If the p_paddr fields 
            //  is not ignored, wide spread common understanding is: if p_paddr = 0, ignore the value 
            //  and use p_vaddr instead. If p_paddr !=0 use the address to load the content for.
            uint64_t address;
            if (phdr->p_paddr != 0)
                address = phdr->p_paddr;
            else
                address = phdr->p_vaddr;
            uint64_t size = phdr->p_filesz;
            uint64_t zeros = phdr->p_memsz - size;
            if(ram_block)ram_block->write(address, (unsigned char*)data, size);
            if(fPTR)(*fPTR)(address, (unsigned char*)data, size);
            if (zeros)
            {
                void* zero_data = calloc(zeros, 1);
                if(ram_block)ram_block->write(address + size, (unsigned char*)zero_data, zeros);
                if(fPTR)(*fPTR)(address + size, (unsigned char*)zero_data, zeros);
                free(zero_data);
            }
        }
    }
    else
        assert(0);
}

bool ElfLoaderLoad (const char* filename, RAMBlock* ram_block, f_dbg_write_t *fPTR)
{
    // Check whether filename starts with ~/,
    // and if so replace with $HOME/
    std::string modified_filename = filename;
    if (modified_filename.compare(0 /* start pos */, 2 /* length */, "~/") == 0)
    {
       modified_filename.replace(0 /* start pos */, 1 /* length */, getenv("HOME"));
       filename = modified_filename.c_str();
    }

    gzFile file = gzopen (filename, "rb");
    if(!file)
        return false;
    printf ("Loading: %s\n",filename);

    char* file_data = NULL;
    unsigned malloc_size = 0;

    malloc_size = 1024 * 1024 * 16;
    file_data = (char*) malloc(malloc_size);
    if(!file_data)
        throw std::bad_alloc();

    unsigned data_size = 0;
    while  (!gzeof(file))
        {
            if (data_size >= malloc_size)
            {
                malloc_size *= 2;
                char* new_file_data = (char*)realloc((void*)file_data, malloc_size);
                if(!new_file_data) {
                    free(file_data);
                    throw std::bad_alloc();
                } else {
                    file_data = new_file_data;
                }
            }
            data_size += gzread(file, &(file_data)[data_size], (unsigned)(malloc_size - data_size));
        }

    if (strncmp(ELFMAG, file_data, 4))
        {
        }
    else{
        ElfLoaderLoadData (file_data, ram_block,fPTR);
        }
    gzclose(file);
    free(file_data);
    return true;
}


bool UBD_ElfLoaderLoad (const char* filename, f_dbg_write_t *fptr){
    return ElfLoaderLoad(filename,nullptr,fptr);
}

};
};
