//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// Please refer to your legal agreement with ARM Limited to
// understand what you are licensed to do with this example code.
//-------------------------------------------------------------------

#include "SimpleBus.h"

using namespace std;
using namespace tlm;
using namespace sc_core;

namespace ARM
{
namespace Models
{

//Constructor
SimpleBus::SimpleBus(sc_core::sc_module_name p_scName, uint32_t numMasters_, uint32_t numTargets_, uint32_t busWidthBits_) :
  SimpleBusBase(p_scName,numMasters_,numTargets_,busWidthBits_),
  tSkt("tskt_bus", *this, (&SimpleBus::nb_transport_fw), ARM::TLM::PROTOCOL_ACE, busWidthBits_)
{
    assert(numMasters == 1);
    iSkt = new MasterSocketT*[numTargets];

    char port_name[15];
    for(uint32_t target=0; target < numTargets; target++)
    {
        sprintf(port_name, "iSkt_bus%d", target);
        iSkt[target] = new MasterSocketT(port_name,  *this, (&SimpleBus::nb_transport_bw), ARM::TLM::PROTOCOL_ACE, busWidthBits_);
    }
    prev_time = 0;
    SC_METHOD(negedge);
    sensitive << clock.neg();
    dont_initialize();
}
  
//Destructor
SimpleBus::~SimpleBus()
{
  for(uint32_t target=0; target < numTargets; target++)
  {
    delete iSkt[target];
  }
  delete [] iSkt;

  // check for any payloads left in queues
  for (int i = 0; i < AXI4_NUM_CHANNELS; ++i)
    if (bw_queues[i].size())
    {
      char msg[1024];
      sprintf(msg,"%lu payload(s) left in back pressure queue #%d at time %llu",bw_queues[i].size(),i,sc_time_stamp().value());
      SC_REPORT_WARNING("SimpleBus",msg);
    }
}

// TLM-2 forward non-blocking transport method
tlm_sync_enum SimpleBus::nb_transport_fw(PayloadT& payload, PhaseT& phase)
{
    uint32_t portId = memoryMap.decodeAddress(payload.get_base_address());

    if(portId == TargetMemoryMap::INVALID_PORT)
    {
        std::cerr << "ERROR: Access to unmapped address (0x" << std::hex << payload.get_base_address() << "). Transaction dropped." << std::endl;
        assert(0);
        return TLM_COMPLETED;
    }

    return iSkt[portId]->nb_transport_fw(payload, phase);
}

// have we seen this channel before during the same time  
bool SimpleBus::seen_channel(int channel)
{
  // look for previous phase on the same channel
  for (unsigned i = 0; i < prev_phases.size(); ++i)
  {
    // are channels the same?
    if (AXI4_CHANNEL_MASK(prev_phases[i]) == channel)
      return true;
  }
  return false;
}

// save the payload for sending later
void SimpleBus::save_payload(PayloadT& payload, PhaseT& phase,int channel,int channel_index)
{
  report_payload("save_payload",payload,phase);
  
  // save the payload, phase and time in a queue and wait till the next negedge to send it back
  bw_queues[channel_index].push_back(std::make_tuple(&payload,phase,sc_time_stamp().value()));
  payload.ref();
      
  // send the READY phase back to the caller
  phase = AXI4_CHANNEL_READY(channel);
}

// send the payload back
void SimpleBus::send_payload(int channel_index)
{
  if (bw_queues[channel_index].size())  // is there something in the queue to send?
  {
    auto front = bw_queues[channel_index].front();
    if (sc_time_stamp().value() > std::get<2>(front))  // has sim time passed since put in queue?
    {
      // send the payload and phase back to the initiator
      PayloadT* payload = std::get<0>(front);
      PhaseT phase = std::get<1>(front);
      bw_queues[channel_index].pop_front();
      report_payload("send_payload",*payload,phase);
      tSkt.nb_transport_bw(*payload,phase);
      payload->unref();
    }
  }
}

// report payload information for debug purposes
void SimpleBus::report_payload(const char* prefix,PayloadT& payload, PhaseT& phase)
{
  char msg[1024];
  sprintf(msg,"%s(time=%llu phase=%x payload=%p addr=%lx)",prefix,sc_time_stamp().value(),phase,&payload,payload.get_base_address());
  SC_REPORT_INFO_VERB("SimpleBus",msg,SC_DEBUG);
}
 
tlm_sync_enum SimpleBus::nb_transport_bw(PayloadT& payload, PhaseT& phase)
{
  int channel = AXI4_CHANNEL_MASK(phase);
  int channel_index = AXI4_CHANNEL_INDEX(channel);
  // Don't call nb_transport_bw twice on the same channel at the same time
  // This can cause an assertion back at the master.
  if (prev_time == sc_time_stamp().value())
  {
    if (seen_channel(channel))
    {
      save_payload(payload,phase,channel,channel_index);
      return TLM_UPDATED;
    }
  }
  else
  {
    prev_time = sc_time_stamp().value();
    prev_phases.clear();
    // if there is something already in the queue for this channel save this payload
    if (bw_queues[channel_index].size())
    {
      save_payload(payload,phase,channel,channel_index);
      return TLM_UPDATED;      
    }
  }
 
  prev_phases.push_back(phase);
  return tSkt.nb_transport_bw(payload,phase);
}

void SimpleBus::negedge()
{
  // go through all the channels and send back the front payload if any
  for (int i = 0; i < AXI4_NUM_CHANNELS; ++i)
    send_payload(i);
}
  
}
}
