//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// Please refer to your legal agreement with ARM Limited to
// understand what you are licensed to do with this example code.
//-------------------------------------------------------------------

#include "SimpleFlashImp.h"

#include <iostream>

namespace ARM
{
namespace Models
{
namespace ModelSimpleMem
{

using namespace ::tlm;

// Constructor
SimpleFlashImp::SimpleFlashImp(sc_core::sc_module_name p_scName, const SimpleFlashConfig &config)
  : sc_module(p_scName)
  , clock("clock_imp")
  , tSkt("tSkt_imp", *this, (&SimpleFlashImp::nb_transport_fw), ARM::TLM::PROTOCOL_ACE, config.busWidthBits)
  , directTskt("directTskt")
  , ram_block(config.ram_block)
  , debugLevel(config.debugLevel)
  , r_current(NULL)
  , b_current(NULL)
  , r_ready_pending(false)
  , b_ready_pending(false)
  , m_name(p_scName)
  , m_delay_read(config.read_delay)
  , m_delay_cache(config.cache_delay)
  , m_max_outstanding(config.max_outstanding)
  , m_cache_line_count(config.cache_line_count)
  , m_cache_line_bytes(config.cache_line_bytes)
  , access_count(0)
  , cache_hits(0)
{
    // Check cache line size is a power of two.
    if (0 != (m_cache_line_bytes & (m_cache_line_bytes - 1)) && m_cache_line_count > 0)
        throw "SimpleFlashImp: cache line size must be a power of two.";

    cache_address_mask = ~(m_cache_line_bytes - 1);

    // Initialise cache.
    {
        CacheLine invalid_line;
        invalid_line.address = 0;
        invalid_line.last_used = 0;
        invalid_line.time_ready = ~(uint64_t)0;
        mini_cache.assign(m_cache_line_count, invalid_line);
    }

    if (!ram_block)
        ram_block = new RAMBlock;
    //assert(config.busWidthBits == 128);

    // clock pos evt is used for evaluate phase
    SC_METHOD(clock_posedge);
    sensitive << clock.pos();
    dont_initialize();

    // clock neg evt is use for propagate phase
    SC_METHOD(clock_negedge);
    sensitive << clock.neg();
    dont_initialize();

    directTskt.register_transport_dbg(this, &SimpleFlashImp::transport_dbg);

    now = 0;

    debug_level = 0;
    if (const char *debug_env = getenv("FLASH_DEBUG"))
    {
        debug_level = (uint8_t)strtol(debug_env, NULL, 10);
    }

    beats_in_flash = 0;
}

// Destructor
SimpleFlashImp::~SimpleFlashImp()
{
    if(debug_level > 0)
    {
        std::cerr << "Number of flash accesses: " << std::dec << access_count << std::endl
                  << "Number of flash cache hits: " << cache_hits << std::endl;
    }
}

unsigned SimpleFlashImp::get_time_ready(uint64_t address)
{
    access_count++;

    if (m_cache_line_count > 0)
    {
        // Check whether it matches in the cache.
        Cache::iterator line = find_cache_line(address);
        if (line != mini_cache.end())
        {
            // Matched; record and update last used time.
            cache_hits++;
            line->last_used = now;

            return std::max(now, line->time_ready) + m_delay_cache;
        }
    }

    // Use standard delay if not found in cache (or no cache present).
    uint64_t time_ready = now + m_delay_read;

    if (m_cache_line_count > 0)
    {
        // Need to insert into the cache here.
        // First find a victim.
        Cache::iterator replacement_line = mini_cache.begin();
        for (Cache::iterator i = replacement_line; i != mini_cache.end(); i++)
        {
            // If not valid, this is always a good line.
            if (!i->is_valid())
            {
                replacement_line = i;
                break;
            }

            // Otherwise check whether this is better than the current replacement line.
            if (i->last_used < replacement_line->last_used)
                replacement_line = i;
        }
        // Now insert into cache.
        replacement_line->address = address & cache_address_mask;
        replacement_line->last_used = replacement_line->time_ready = time_ready;
    }

    return time_ready;
}

SimpleFlashImp::Cache::iterator SimpleFlashImp::find_cache_line(uint64_t address)
{
    uint64_t cache_line_base = address & cache_address_mask;

    for (Cache::iterator i = mini_cache.begin(); i != mini_cache.end(); i++)
    {
        // Check valid.
        if (!i->is_valid())
            continue;

        if (i->address == cache_line_base)
            return i;
    }

    return mini_cache.end();
}

void SimpleFlashImp::queue_insert(TimedPayloadQueue &queue, const TimedPayloadT &payload, PayloadT& actual_axi_payload)
{
    if (queue.empty())
    {
        queue.push_back(payload);
        actual_axi_payload.ref();
        return;
    }

    for (TimedPayloadQueue::iterator i = queue.begin(); i != queue.end(); i++)
    {
        if (payload.second < i->second)
        {
            queue.insert(i, payload);
            actual_axi_payload.ref();
        }
    }

    // Got to end.
    queue.push_back(payload);
    actual_axi_payload.ref();
}

tlm_sync_enum SimpleFlashImp::check_read_buffer()
{
    // Check whether we have anything to do.
    if (read_buffer.empty())
        return TLM_ACCEPTED;

    // Check whether we have the ability to add to the r_queue.
    if (r_queue.size() >= m_max_outstanding)
        return TLM_ACCEPTED;

    // All good; get payload/beat to process.
    PayloadAndBeat payload_info = read_buffer.front();
    read_buffer.pop_front();
    PayloadT &payload = *(payload_info.first);
    
    unsigned beat_num = payload_info.second;

    uint64_t time_ready = get_time_ready(payload.get_address() + (payload.get_beat_data_length() * beat_num));
    // If we have zero delay and we're not already processing something this cycle.
    if (time_ready <= now && !r_current)
    {
        // Support for Zero Delay Flash Memory, Send the response back instantly
        bool last_beat = (payload.get_beat_count() == (beat_num + 1));

        PhaseT ret_phase = last_beat ? ARM::AXI4::R_VALID_LAST : ARM::AXI4::R_VALID;

        tlm_sync_enum ret_val = tSkt.nb_transport_bw(payload, ret_phase);

        // Check whether we process the beat.
        if (ret_val == TLM_UPDATED){
            payload.unref();
            return ret_val;
        }
    }

    if (debug_level > 0)
    {
        std::cerr << "Processing payload @ 0x" << std::hex << payload.get_address() << ", beat num " << std::dec
                  << beat_num << "; will be ready @ " << time_ready << " (" << (time_ready - now) << " cycles time)."
                  << std::endl;
    }

    payload.unref(); // corresponding to removal of payload from read_buffer.

    // Insert beat into the queue, and update payload's ref count.
    queue_insert(r_queue, TimedPayloadT(&payload, time_ready), payload);

    return TLM_ACCEPTED;
}

// TLM-2 forward non-blocking transport method
tlm_sync_enum SimpleFlashImp::nb_transport_fw(PayloadT &payload, PhaseT &phase)
{
    tlm_sync_enum ret_val;
    switch (phase)
    {
    case ARM::AXI4::AW_VALID:
        w_queue.push_back(TimedPayloadT(&payload, now)); // No delay on writes as we ignore them.
        payload.ref();
        phase = ARM::AXI4::AW_READY;
        ret_val = TLM_UPDATED;
        break;

    case ARM::AXI4::W_VALID_LAST:
        // A read-only memory follows the protocol just like a normal memory
        // So writes are acknowledged, they just don't do anything to the memory content

        wd_queue.push_back(TimedPayloadT(&payload, now)); // no delay on write response
        payload.ref();
    // fall thru

    case ARM::AXI4::W_VALID:
        phase = ARM::AXI4::W_READY;
        ret_val = TLM_UPDATED;
        break;

    case ARM::AXI4::B_READY:
        b_ready_pending = false;
        ret_val = TLM_ACCEPTED;
        break;

    case ARM::AXI4::AR_VALID:
    {
        uint64_t addr = payload.get_base_address();
        uint8_t *temp_data = new uint8_t[payload.get_data_length()];
        ram_block->read(addr, temp_data, payload.get_data_length());
        payload.read_in(temp_data);
        delete[] temp_data;

        // Put each beat onto buffer for processing.
        for (unsigned i = 0; i < payload.get_beat_count(); i++)
        {
            if (debug_level > 0)
                std::cerr << "Adding 0x" << std::hex << payload.get_address() << " beat " << std::dec << i
                          << " to read buffer @ " << now << std::endl;
            read_buffer.push_back(std::make_pair(&payload, i));
            payload.ref();
            beats_in_flash++;
            if (debug_level > 1)
                std::cerr << "Now " << std::dec << beats_in_flash << " beats being processed." << std::endl;
        }

        // Process buffer (if we can).
        check_read_buffer();

        // Indicate that we've processed the transaction.
        ret_val = TLM_UPDATED;
        phase = ARM::AXI4::AR_READY;

        break;
    }

    case ARM::AXI4::R_READY:
        r_ready_pending = false;
        ret_val = TLM_ACCEPTED;
        break;
    case ARM::AXI4::RACK:
    case ARM::AXI4::WACK:
        ret_val = TLM_ACCEPTED;
        break;
    default:
        ret_val = TLM_ACCEPTED;
        assert(!"uncaught command");
    }
    return ret_val;
}

/*-------------------- Clock Actions ----------------------*/
void SimpleFlashImp::clock_posedge()
{
    now++; // advance time

    // things stay in the queue until their timestamp passes
    if (!r_ready_pending && !r_queue.empty())
    {
        if (now >= r_queue.front().second)
        {
            r_current = r_queue.front().first;
            r_queue.pop_front();

            // Update cache last used.
            if (m_cache_line_count > 0)
            {
                Cache::iterator line = find_cache_line(r_current->get_address());
                if (line != mini_cache.end())
                    line->last_used = now;
            }
        }
    }

    // Process the read buffer, in case we have anything else that
    // we can now process.
    check_read_buffer();

    if (!b_ready_pending && !w_queue.empty() && !wd_queue.empty())
    {
        if (now >= wd_queue.front().second)
        {
            assert(w_queue.front().first == wd_queue.front().first);
            b_current = w_queue.front().first;
            PayloadT *wd_current = wd_queue.front().first;
            wd_current->unref();
            w_queue.pop_front();
            wd_queue.pop_front();
        }
    }
}

void SimpleFlashImp::clock_negedge()
{
    PhaseT phase;
    tlm_sync_enum ret_val;

    if (r_current)
    {
        phase = (r_current != r_queue.front().first) ? ARM::AXI4::R_VALID_LAST : ARM::AXI4::R_VALID;
        r_current->set_resp(ARM::AXI4::RESP_OKAY);
        r_ready_pending = true;

        ret_val = tSkt.nb_transport_bw(*r_current, phase);

        if (debug_level > 0)
            std::cerr << "Returned data for payload at 0x" << std::hex << r_current->get_address() << " @ " << std::dec
                      << now << std::endl;
        beats_in_flash--;
        if (debug_level > 1)
            std::cerr << "Now (" << std::dec << now << ") " << beats_in_flash << " beats being processed." << std::endl;

        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::R_READY);
            r_ready_pending = false;
        }
        r_current->unref();
        r_current = NULL;
    }

    if (b_current)
    {
        phase = ARM::AXI4::B_VALID;
        b_current->set_resp(ARM::AXI4::RESP_OKAY);

        b_ready_pending = true;
        ret_val = tSkt.nb_transport_bw(*b_current, phase);

        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::B_READY);
            b_ready_pending = false;
        }
        b_current->unref();
        b_current = NULL;
    }
}

unsigned int SimpleFlashImp::transport_dbg(tlm::tlm_generic_payload &trans)
{
    uint8_t *data_ptr = trans.get_data_ptr();
    uint64_t addr = trans.get_address();

    // You need to be able to populate a read-only memory from the ELF loader
    // so immediate-writes do write

    if (trans.get_command() == tlm::TLM_WRITE_COMMAND)
    {
        ram_block->write(addr, data_ptr, trans.get_data_length());
    }
    else
    {
        ram_block->read(addr, data_ptr, trans.get_data_length());
    }
    trans.set_response_status(tlm::TLM_OK_RESPONSE);

    return trans.get_data_length();
}
}
}
}
