//-------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited.
//
//            (C) COPYRIGHT 2015 ARM Limited.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited.
//
// Please refer to your legal agreement with ARM Limited to
// understand what you are licensed to do with this example code.
//-------------------------------------------------------------------

#include "SimpleMemImp.h"

#include <iomanip>

using namespace std;
using namespace tlm;
using namespace sc_core;

namespace ARM
{
namespace Models
{
namespace ModelSimpleMem
{

// Constructor
SimpleMemImp::SimpleMemImp(sc_core::sc_module_name p_scName, const SimpleMemConfig &config)
  : sc_module(p_scName)
  , clock("clock_imp")
  , tSkt("tSkt_imp", *this, (&SimpleMemImp::nb_transport_fw), ARM::TLM::PROTOCOL_ACE, config.busWidthBits)
  , directTskt("directTskt")
  , ram_block(config.ram_block)
  , debugLevel(config.debugLevel)
  , r_current(NULL)
  , b_current(NULL)
  , r_ready_pending(false)
  , b_ready_pending(false)
  , m_name(p_scName)
  , m_delay_read(config.delay)
  , m_delay_write(config.delay)
  , m_readonly(config.read_only)
  , m_memory_dump_file(config.memory_dump_file)
  , m_memory_dump_min(config.dump_min_address)
  , m_memory_dump_max(config.dump_max_address)
{
    if (!ram_block)
        ram_block = new RAMBlock;

    // Only 128-bit bus width is supported.
    //assert(config.busWidthBits == 128);

    // clock pos evt is used for evaluate phase
    SC_METHOD(clock_posedge);
    sensitive << clock.pos();
    dont_initialize();

    // clock neg evt is use for propagate phase
    SC_METHOD(clock_negedge);
    sensitive << clock.neg();
    dont_initialize();

    directTskt.register_transport_dbg(this, &SimpleMemImp::transport_dbg);

    now = 0;

    // If we have a memory dump file specified,
    // install a SIGINT handler if we haven't already.
    if (m_memory_dump_file)
    {
        mem_imps.push_back(this);

        install_signal_handler();
    }
}

//Destructor
SimpleMemImp::~SimpleMemImp()
{
}

// TLM-2 forward non-blocking transport method
tlm_sync_enum SimpleMemImp::nb_transport_fw(PayloadT& payload, PhaseT& phase)
{
    tlm_sync_enum ret_val;
    switch (phase)
    {
        case ARM::AXI4::AW_VALID:
            w_queue.push_back(TimedPayloadT(&payload, now + m_delay_write));
            payload.ref();
            phase   = ARM::AXI4::AW_READY;
            ret_val = TLM_UPDATED;
            break;

        case ARM::AXI4::W_VALID_LAST:
        {
            uint64_t addr = payload.get_base_address();
            uint8_t* temp_data = new uint8_t[payload.get_data_length()];
            
            // A read-only memory follows the protocol just like a normal memory
            // So writes are acknowledged, they just don't do anything to the memory content

            if (! m_readonly)
            {
                ram_block->read (addr, temp_data, payload.get_data_length());
                payload.write_out(temp_data);
                ram_block->write (addr, temp_data, payload.get_data_length());
            }
            
            delete [] temp_data;
            
            wd_queue.push_back(TimedPayloadT(&payload, now)); // no delay on write response
            payload.ref();
            phase = ARM::AXI4::W_READY;
            ret_val = TLM_UPDATED;
        }
        //fall thru
        
        case ARM::AXI4::W_VALID:

            phase   = ARM::AXI4::W_READY;
            ret_val = TLM_UPDATED;
            break;

        case ARM::AXI4::B_READY:
            b_ready_pending  = false;
            ret_val = TLM_ACCEPTED;
            break;

        case ARM::AXI4::AR_VALID:
        {
            uint64_t addr = payload.get_base_address();
            uint8_t* temp_data = new uint8_t[payload.get_data_length()];
            ram_block->read (addr, temp_data, payload.get_data_length());
            payload.read_in(temp_data);
            delete [] temp_data;

            if(m_delay_read == 0) // Support for Zero Delay Flash Memory, Send the response back instantly
            {
                // Sending the response for the first beat instanstly and storing the rest of the beats
                for (unsigned i = 1; i < payload.get_beat_count(); i++)
                {
                    r_queue.push_back(pair<PayloadT*,uint64_t>(&payload, now+m_delay_read));
                    payload.ref();
                }

                PhaseT ret_phase = (payload.get_beat_count() == 1) ? ARM::AXI4::R_VALID_LAST : ARM::AXI4::R_VALID;

                ret_val = tSkt.nb_transport_bw(payload, ret_phase);

                if(ret_val != TLM_UPDATED)
                {
                    r_queue.push_back(pair<PayloadT*,uint64_t>(&payload, now+m_delay_read));
                    payload.ref();
                }
            }
            else
            {
                for (unsigned i = 0; i < payload.get_beat_count(); i++)
                {
                    r_queue.push_back(pair<PayloadT*,uint64_t>(&payload, now+m_delay_read));
                    payload.ref();
                }
            }
            phase   = ARM::AXI4::AR_READY;
            ret_val = TLM_UPDATED;
            break;
        }

        case ARM::AXI4::R_READY:
            r_ready_pending = false;
            ret_val = TLM_ACCEPTED;
            break;
        case ARM::AXI4::RACK:
        case ARM::AXI4::WACK:
            ret_val = TLM_ACCEPTED;
            break;
        default:
            ret_val = TLM_ACCEPTED;
            assert(!"uncaught command");
    }
    return ret_val;
}

/*-------------------- Clock Actions ----------------------*/
void SimpleMemImp::clock_posedge()
{
    now++; // advance time

    // things stay in the queue until their timestamp passes
    if (!r_ready_pending && !r_queue.empty())
    {
        if (now >= r_queue.front().second)
        {
            r_current = r_queue.front().first;
            r_queue.pop_front();
        }
    }                                   
        
    if (!b_ready_pending && !w_queue.empty() && !wd_queue.empty())
    {
        if (now >= wd_queue.front().second)
        {
            assert(w_queue.front().first == wd_queue.front().first);
            b_current = w_queue.front().first;
            PayloadT *wd_current = wd_queue.front().first;
            wd_current->unref();

            w_queue.pop_front();
            wd_queue.pop_front();
        }
    }
}

void SimpleMemImp::clock_negedge()
{
    PhaseT        phase;
    tlm_sync_enum ret_val;

    if (r_current)
    {
        if (r_queue.empty())
          phase = ARM::AXI4::R_VALID_LAST;
        else phase = (r_current != r_queue.front().first) ? ARM::AXI4::R_VALID_LAST:ARM::AXI4::R_VALID;
        r_current->set_resp(ARM::AXI4::RESP_OKAY);
        r_ready_pending = true;
        
        ret_val = tSkt.nb_transport_bw(*r_current, phase);
        
        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::R_READY);
            r_ready_pending = false;
        }
        r_current->unref();
        r_current = NULL;
    }

    if (b_current)
    {
        phase = ARM::AXI4::B_VALID;
        b_current->set_resp(ARM::AXI4::RESP_OKAY);
        
        b_ready_pending = true;
        ret_val = tSkt.nb_transport_bw(*b_current, phase);

        if (ret_val == TLM_UPDATED)
        {
            assert(phase == ARM::AXI4::B_READY);
            b_ready_pending = false;
        }
        b_current->unref();
        b_current = NULL;
    }
}

unsigned int SimpleMemImp::transport_dbg(tlm::tlm_generic_payload &trans)
{
    uint8_t *data_ptr = trans.get_data_ptr();
    uint64_t addr = trans.get_address();

    // You need to be able to populate a read-only memory from the ELF loader
    // so immediate-writes do write

    if(trans.get_command() == tlm::TLM_WRITE_COMMAND)
    {
        ram_block->write (addr, data_ptr , trans.get_data_length());
    }
    else
    {
        ram_block->read (addr, data_ptr, trans.get_data_length());
    }

    trans.set_response_status( tlm::TLM_OK_RESPONSE );

    return trans.get_data_length();
}

//
// Memory dump / signal handling.
//

std::vector<SimpleMemImp *> SimpleMemImp::mem_imps;
#ifndef _WIN32 //RSM
bool SimpleMemImp::installed_sigint_handler = false;
struct sigaction SimpleMemImp::previous_signal_handler;
#endif

void SimpleMemImp::install_signal_handler()
{
#ifndef _WIN32  //RSM
    if (installed_sigint_handler)
        return;

    struct sigaction new_action;
    new_action.sa_handler = &SimpleMemImp::sigint_handler;
    sigemptyset(&new_action.sa_mask);

    // Don't prevent signal from being received while in handler.
    new_action.sa_flags = SA_NODEFER;

    sigaction(SIGINT, &new_action, &previous_signal_handler);

    installed_sigint_handler = true;
#endif
}

void SimpleMemImp::sigint_handler(int signum)
{
#ifndef _WIN32 //RSM
    // Install previous signal handler.
    sigaction(SIGINT, &previous_signal_handler, NULL);

    for (std::vector<SimpleMemImp *>::iterator i = mem_imps.begin(); i != mem_imps.end(); i++)
    {
        (*i)->dump_memory();
    }

    // Re-raise the signal.
    kill(0, signum);
#endif
}

void SimpleMemImp::dump_memory()
{
    std::cerr << "Writing memory dump to " << m_memory_dump_file << "...";

    std::ofstream fh(m_memory_dump_file);

    if (!fh.good())
    {
        std::cerr << std::endl
                  << "ERROR: failed to open '" << m_memory_dump_file << "'." << std::endl;
        return;
    }

    const unsigned bytes_per_line = 0x10;

    for (unsigned addr = (m_memory_dump_min & ~(bytes_per_line - 1)); addr <= m_memory_dump_max; addr += bytes_per_line)
    {
        uint8_t data[bytes_per_line];
        ram_block->read(addr, data, bytes_per_line);

        fh << std::hex << std::setw(8) << std::setfill('0') << addr;

        for (unsigned offset = 0; offset < bytes_per_line; offset += 2)
        {
            fh << ' ' << std::setfill('0') << std::setw(2) << (uint32_t)data[offset];
            fh << std::setfill('0') << std::setw(2) << (uint32_t)data[offset + 1];
        }
        fh << std::endl;
    }
    fh.close();

    std::cerr << " done.\n";
}
}
}
}
