///
/// \file    SGArgMeta.h
/// \brief   A set of meta-programming widgets
/// \date    Copyright 2013 ARM Limited. All rights reserved.
///
/// This is an automatically generated file please do not edit.
///
#ifndef ARM_INCLUDE_SGARGMETA_H
#define ARM_INCLUDE_SGARGMETA_H

#include "sg/MSCFixes.h"
#include <cstddef>
#include <cassert>

/// \page Deferred function execution
///
/// This file contains meta programming abstractions to allow construction of
/// deferred function execution.
///
/// If execution is to be deferred then storage is required for the
/// arguments and what to call.  This is provided by an Allocator which is a
/// template argument.
///
/// The data needed to be stored is classified as:
///    * storage encoding what to call
///    * storage encoding the static amount of storage necessary for the
///      arguments
///    * dynamic storage necessary for storing variable length data for the
///      arguments.
///
/// The allocator models the following concept:
/// <code>
///        class AllocatorConcept
///        {
///        public:
///            AllocatorConcept();
///
///            struct pointer {
///                pointer();
///
///                operator unsigned char*       ()       { return (unsigned char*)this + offset;       }
///                operator unsigned char const* () const { return (unsigned char const*)this + offset; }
///            private:
///                pointer(pointer const&); // disallow
///                void operator = (pointer const&); // disallow
///            };
///
///            unsigned char* allocate_arg_holder(std::size_t sz_);
///            unsigned char* allocate_virtual_thunk(std::size_t sz_);
///            void allocate_dynamic_data(std::size_t sz_, pointer& o_);
///
///            // Copy the set of bytes indicated by [begin_,end_) to the destination_.
///            void copy( void const* begin_, void const* end_, pointer& destination_ );
///
///            // Is the set of bytes pointed to by [begin_, begin_ + sz_) equal to
///            // [other_begin_, other_begin_ + sz_)?
///            static bool is_equal( pointer const& begin_, pointer const& other_begin_, std::size_t sz_ );
///        };
/// </code>
///
/// The 'thing to call' inherits from the abstract base class virtual_thunk_if
/// which allows the thing to be called to be manipulated independent of its
/// signature.
///
/// The static storage for the arguments are held in arg_holder_t<>.  The
/// 'stashed' form of an argument may be different to the actual argument type
/// and a set of trait classes have to be provided to produce this mapping.  The
/// trait classes model the concept:
///
/// <code>
///        template<typename T>
///        struct TraitsConcept
///        {
///            template<typename Allocator>
///            struct bind {
///
///                // Some type called 'stash_type' (could be a typedef to a built-in type, etc.)
///                // stash_type has to support copy construction, assignment and operator ==().
///                struct stash_type;
///
///                static T&          from_stash(stash_type& t_);
///                static void        stash_in(Allocator&, stash_type& o_, T& t_);
///                static std::size_t dynamic_data_size(stash_type const&);
///            };
///        };
/// </code>
///
/// The arg_holder_t actually holds the stash_type equivalents of the real
/// argument types.
///
/// The dynamic storage is carved out by the allocator and is treated as a
/// stream of bytes and it is up to stash_in() and from_stash() to manipulate
/// that store.
///

namespace sg_meta
{

    // ----------------------------------------------------------------------
    //                      Virtual thunk objects
    // ----------------------------------------------------------------------

    class virtual_thunk_if
    {
    public:
        virtual ~virtual_thunk_if() {}

        struct sizes_t {
            sizes_t( std::size_t virtual_thunk_size_,
                     std::size_t arg_holder_size_,
                     std::size_t arg_holder_dynamic_size_ )
                : virtual_thunk_size(virtual_thunk_size_),
                  arg_holder_size(arg_holder_size_),
                  arg_holder_dynamic_size(arg_holder_dynamic_size_)
                {
                }

            std::size_t  virtual_thunk_size;
            std::size_t  arg_holder_size;
            std::size_t  arg_holder_dynamic_size;

            bool operator == (sizes_t const& o_) const
            {
                return virtual_thunk_size      == o_.virtual_thunk_size
                    && arg_holder_size         == o_.arg_holder_size
                    && arg_holder_dynamic_size == o_.arg_holder_dynamic_size;
            }

            bool operator != (sizes_t const& o_) const
            {
                return ! (*this == o_);
            }
        };

        virtual std::size_t sizeof_this() const = 0;
        virtual sizes_t     sizeof_all(void*) const = 0;
        virtual sizes_t     exec(void*) const = 0;
        virtual sizes_t     destroy_arg_holder(void*) const = 0;

        // We need an abstract concept of asking if two thunks are the
        // same.  This is, in general, quite hard, instead what we do
        // is that if the set of thunks that we place on the allocator
        // are of two specific types -- comparable and incomparable, then
        // the comparable ones are required to return a non-null pointer
        // such that two thunks are 'equal' if they return the same pointer.
        virtual void const* object_pointer() const = 0;

        // Return true if both are this and other_ are comparable to each
        // other and are equal to each other.
        virtual bool is_equal(void const* other_, void const* this_arg_holder_, void const* other_arg_holder_) const = 0;
    };

    template<typename OP, typename ARG_HOLDER>
    class op_and_arg_holder_thunk_t : public virtual_thunk_if
    {
        OP            op;
    public:

        op_and_arg_holder_thunk_t(OP const& op_) : op(op_) {}

        virtual std::size_t sizeof_this() const
        {
            return sizeof(*this);
        }

        virtual sizes_t sizeof_all(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            return sizes_t(sizeof(*this),
                           sizeof(ARG_HOLDER),
                           arg_holder->dynamic_data_size());
        }

        virtual sizes_t exec(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            sizes_t const szs(sizeof(*this),
                              sizeof(ARG_HOLDER),
                              arg_holder->dynamic_data_size());
            arg_holder->apply(op);
            return szs;
        }

        virtual sizes_t destroy_arg_holder(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            sizes_t const szs(sizeof(*this),
                              sizeof(ARG_HOLDER),
                              arg_holder->dynamic_data_size());
            arg_holder->~ARG_HOLDER();
            return szs;
        }

        // General deferred objects are incomparable.
        virtual void const* object_pointer() const { return 0; }

        // No is_equal operator
        virtual bool is_equal(void const*, void const*, void const*) const { assert(0); return false; }
    };

    // Forward decl
    template<typename Allocator, template<typename> class Traits, typename SIG> struct arg_holder_t {};
    template<typename T> struct traits {};

    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R>
    struct traits< R (C::*)() >
    {
        static unsigned const number_of_parameters = 0;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef return_type signature();

        typedef R    (C::*mem_func_ptr_type)();

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature();

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()()
                {
                    (obj->*mem_func_ptr)();
                }
        };
    };

    template<typename D, typename R, typename C>
    inline
    typename traits<R (C::*)()>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)())
    {
        typedef typename traits<R (C::*)()>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds no arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits>
    struct arg_holder_t<Allocator, Traits, void(void)>
    {
        typedef void                                                  signature();
        typedef void                                                  stash_signature();

        template<typename OP> void apply(OP op_)
        {
            op_();
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)();
        }

        void init(Allocator&)
        {
            ; // nothing to do
        }

        std::size_t dynamic_data_size() const
        {
            return 0;
        }

        bool operator == (arg_holder_t const&) const
        {
            return true;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init()
            {
                ah.init(a);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0>
    struct traits< R (C::*)(T0) >
    {
        static unsigned const number_of_parameters = 1;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;

        typedef return_type signature(T0);

        typedef R    (C::*mem_func_ptr_type)(T0);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_)
                {
                    (obj->*mem_func_ptr)(t0_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0>
    inline
    typename traits<R (C::*)(T0)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0))
    {
        typedef typename traits<R (C::*)(T0)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 0 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0>
    struct arg_holder_t<Allocator, Traits, void(T0)>
    {
        typedef T0                                                    arg0_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;

        stash0_type                                                   t0;

        typedef void                                                  signature(T0);
        typedef void                                                  stash_signature(stash0_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_)
            {
                ah.init(a, t0_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1>
    struct traits< R (C::*)(T0, T1) >
    {
        static unsigned const number_of_parameters = 2;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;

        typedef return_type signature(T0, T1);

        typedef R    (C::*mem_func_ptr_type)(T0, T1);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1>
    inline
    typename traits<R (C::*)(T0, T1)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1))
    {
        typedef typename traits<R (C::*)(T0, T1)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 1 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1>
    struct arg_holder_t<Allocator, Traits, void(T0, T1)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;

        typedef void                                                  signature(T0, T1);
        typedef void                                                  stash_signature(stash0_type,stash1_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_)
            {
                ah.init(a, t0_,t1_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2>
    struct traits< R (C::*)(T0, T1, T2) >
    {
        static unsigned const number_of_parameters = 3;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;

        typedef return_type signature(T0, T1, T2);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2>
    inline
    typename traits<R (C::*)(T0, T1, T2)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2))
    {
        typedef typename traits<R (C::*)(T0, T1, T2)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 2 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;

        typedef void                                                  signature(T0, T1, T2);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_)
            {
                ah.init(a, t0_,t1_,t2_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3>
    struct traits< R (C::*)(T0, T1, T2, T3) >
    {
        static unsigned const number_of_parameters = 4;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;

        typedef return_type signature(T0, T1, T2, T3);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 3 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;

        typedef void                                                  signature(T0, T1, T2, T3);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_)
            {
                ah.init(a, t0_,t1_,t2_,t3_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4>
    struct traits< R (C::*)(T0, T1, T2, T3, T4) >
    {
        static unsigned const number_of_parameters = 5;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;

        typedef return_type signature(T0, T1, T2, T3, T4);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 4 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;

        typedef void                                                  signature(T0, T1, T2, T3, T4);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5) >
    {
        static unsigned const number_of_parameters = 6;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 5 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6) >
    {
        static unsigned const number_of_parameters = 7;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 6 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7) >
    {
        static unsigned const number_of_parameters = 8;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 7 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8) >
    {
        static unsigned const number_of_parameters = 9;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 8 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9) >
    {
        static unsigned const number_of_parameters = 10;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 9 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10) >
    {
        static unsigned const number_of_parameters = 11;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 10 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11) >
    {
        static unsigned const number_of_parameters = 12;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 11 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12) >
    {
        static unsigned const number_of_parameters = 13;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 12 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13) >
    {
        static unsigned const number_of_parameters = 14;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 13 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14) >
    {
        static unsigned const number_of_parameters = 15;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 14 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15) >
    {
        static unsigned const number_of_parameters = 16;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 15 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16) >
    {
        static unsigned const number_of_parameters = 17;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 16 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17) >
    {
        static unsigned const number_of_parameters = 18;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 17 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18) >
    {
        static unsigned const number_of_parameters = 19;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 18 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19) >
    {
        static unsigned const number_of_parameters = 20;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;
        typedef T19   arg19_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_, T19 t19_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_, t19_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 19 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;
        typedef T19                                                    arg19_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;
        typedef typename Traits<arg19_type>::template bind<Allocator>  stash19_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;
        typedef typename stash19_traits::stash_type                    stash19_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;
        stash19_type                                                   t19;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type,stash19_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_, arg19_type const& arg19_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
            stash19_traits::stash_in(a_, t19, arg19_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18)+stash19_traits::dynamic_data_size(t19);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18 && t19 == o_.t19;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_,arg19_type t19_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_,t19_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20) >
    {
        static unsigned const number_of_parameters = 21;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;
        typedef T19   arg19_type;
        typedef T20   arg20_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_, T19 t19_, T20 t20_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_, t19_, t20_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 20 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;
        typedef T19                                                    arg19_type;
        typedef T20                                                    arg20_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;
        typedef typename Traits<arg19_type>::template bind<Allocator>  stash19_traits;
        typedef typename Traits<arg20_type>::template bind<Allocator>  stash20_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;
        typedef typename stash19_traits::stash_type                    stash19_type;
        typedef typename stash20_traits::stash_type                    stash20_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;
        stash19_type                                                   t19;
        stash20_type                                                   t20;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type,stash19_type,stash20_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_, arg19_type const& arg19_, arg20_type const& arg20_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
            stash19_traits::stash_in(a_, t19, arg19_);
            stash20_traits::stash_in(a_, t20, arg20_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18)+stash19_traits::dynamic_data_size(t19)+stash20_traits::dynamic_data_size(t20);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18 && t19 == o_.t19 && t20 == o_.t20;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_,arg19_type t19_,arg20_type t20_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_,t19_,t20_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21) >
    {
        static unsigned const number_of_parameters = 22;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;
        typedef T19   arg19_type;
        typedef T20   arg20_type;
        typedef T21   arg21_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_, T19 t19_, T20 t20_, T21 t21_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_, t19_, t20_, t21_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 21 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;
        typedef T19                                                    arg19_type;
        typedef T20                                                    arg20_type;
        typedef T21                                                    arg21_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;
        typedef typename Traits<arg19_type>::template bind<Allocator>  stash19_traits;
        typedef typename Traits<arg20_type>::template bind<Allocator>  stash20_traits;
        typedef typename Traits<arg21_type>::template bind<Allocator>  stash21_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;
        typedef typename stash19_traits::stash_type                    stash19_type;
        typedef typename stash20_traits::stash_type                    stash20_type;
        typedef typename stash21_traits::stash_type                    stash21_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;
        stash19_type                                                   t19;
        stash20_type                                                   t20;
        stash21_type                                                   t21;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type,stash19_type,stash20_type,stash21_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_, arg19_type const& arg19_, arg20_type const& arg20_, arg21_type const& arg21_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
            stash19_traits::stash_in(a_, t19, arg19_);
            stash20_traits::stash_in(a_, t20, arg20_);
            stash21_traits::stash_in(a_, t21, arg21_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18)+stash19_traits::dynamic_data_size(t19)+stash20_traits::dynamic_data_size(t20)+stash21_traits::dynamic_data_size(t21);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18 && t19 == o_.t19 && t20 == o_.t20 && t21 == o_.t21;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_,arg19_type t19_,arg20_type t20_,arg21_type t21_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_,t19_,t20_,t21_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22) >
    {
        static unsigned const number_of_parameters = 23;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;
        typedef T19   arg19_type;
        typedef T20   arg20_type;
        typedef T21   arg21_type;
        typedef T22   arg22_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_, T19 t19_, T20 t20_, T21 t21_, T22 t22_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_, t19_, t20_, t21_, t22_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 22 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;
        typedef T19                                                    arg19_type;
        typedef T20                                                    arg20_type;
        typedef T21                                                    arg21_type;
        typedef T22                                                    arg22_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;
        typedef typename Traits<arg19_type>::template bind<Allocator>  stash19_traits;
        typedef typename Traits<arg20_type>::template bind<Allocator>  stash20_traits;
        typedef typename Traits<arg21_type>::template bind<Allocator>  stash21_traits;
        typedef typename Traits<arg22_type>::template bind<Allocator>  stash22_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;
        typedef typename stash19_traits::stash_type                    stash19_type;
        typedef typename stash20_traits::stash_type                    stash20_type;
        typedef typename stash21_traits::stash_type                    stash21_type;
        typedef typename stash22_traits::stash_type                    stash22_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;
        stash19_type                                                   t19;
        stash20_type                                                   t20;
        stash21_type                                                   t21;
        stash22_type                                                   t22;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type,stash19_type,stash20_type,stash21_type,stash22_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21),
                stash22_traits::from_stash(t22)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21),
                stash22_traits::from_stash(t22)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_, arg19_type const& arg19_, arg20_type const& arg20_, arg21_type const& arg21_, arg22_type const& arg22_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
            stash19_traits::stash_in(a_, t19, arg19_);
            stash20_traits::stash_in(a_, t20, arg20_);
            stash21_traits::stash_in(a_, t21, arg21_);
            stash22_traits::stash_in(a_, t22, arg22_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18)+stash19_traits::dynamic_data_size(t19)+stash20_traits::dynamic_data_size(t20)+stash21_traits::dynamic_data_size(t21)+stash22_traits::dynamic_data_size(t22);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18 && t19 == o_.t19 && t20 == o_.t20 && t21 == o_.t21 && t22 == o_.t22;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_,arg19_type t19_,arg20_type t20_,arg21_type t21_,arg22_type t22_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_,t19_,t20_,t21_,t22_);
            }
        };
    };


    // ----------------------------------------------------------------------
    //         mem_fun_op_t is a functor and also a traits type
    // ----------------------------------------------------------------------

    template<typename C, typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22, typename T23>
    struct traits< R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23) >
    {
        static unsigned const number_of_parameters = 24;
        
        typedef C    class_type;
        typedef R    return_type;

        typedef T0   arg0_type;
        typedef T1   arg1_type;
        typedef T2   arg2_type;
        typedef T3   arg3_type;
        typedef T4   arg4_type;
        typedef T5   arg5_type;
        typedef T6   arg6_type;
        typedef T7   arg7_type;
        typedef T8   arg8_type;
        typedef T9   arg9_type;
        typedef T10   arg10_type;
        typedef T11   arg11_type;
        typedef T12   arg12_type;
        typedef T13   arg13_type;
        typedef T14   arg14_type;
        typedef T15   arg15_type;
        typedef T16   arg16_type;
        typedef T17   arg17_type;
        typedef T18   arg18_type;
        typedef T19   arg19_type;
        typedef T20   arg20_type;
        typedef T21   arg21_type;
        typedef T22   arg22_type;
        typedef T23   arg23_type;

        typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23);

        typedef R    (C::*mem_func_ptr_type)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23);

        struct op_t
        {
            typedef C           class_type;
            typedef R           return_type;
            typedef return_type signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23);

            C*                 obj;
            mem_func_ptr_type  mem_func_ptr;

            op_t(C* obj_, mem_func_ptr_type mem_func_ptr_)
                : obj(obj_), mem_func_ptr(mem_func_ptr_)
                {
                }

            op_t()
                {
                }

            R operator()(T0 t0_, T1 t1_, T2 t2_, T3 t3_, T4 t4_, T5 t5_, T6 t6_, T7 t7_, T8 t8_, T9 t9_, T10 t10_, T11 t11_, T12 t12_, T13 t13_, T14 t14_, T15 t15_, T16 t16_, T17 t17_, T18 t18_, T19 t19_, T20 t20_, T21 t21_, T22 t22_, T23 t23_)
                {
                    (obj->*mem_func_ptr)(t0_, t1_, t2_, t3_, t4_, t5_, t6_, t7_, t8_, t9_, t10_, t11_, t12_, t13_, t14_, t15_, t16_, t17_, t18_, t19_, t20_, t21_, t22_, t23_);
                }
        };
    };

    template<typename D, typename R, typename C, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22, typename T23>
    inline
    typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23)>::op_t
    make_mem_fun_op(D* obj_, R (C::*mem_func_ptr_)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23))
    {
        typedef typename traits<R (C::*)(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23)>::op_t op_t;
        return op_t(obj_,mem_func_ptr_);
    }

    // ----------------------------------------------------------------------
    //               This arg_holder_t holds 23 arguments
    // ----------------------------------------------------------------------

    template<typename Allocator, template<typename> class Traits, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22, typename T23>
    struct arg_holder_t<Allocator, Traits, void(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23)>
    {
        typedef T0                                                    arg0_type;
        typedef T1                                                    arg1_type;
        typedef T2                                                    arg2_type;
        typedef T3                                                    arg3_type;
        typedef T4                                                    arg4_type;
        typedef T5                                                    arg5_type;
        typedef T6                                                    arg6_type;
        typedef T7                                                    arg7_type;
        typedef T8                                                    arg8_type;
        typedef T9                                                    arg9_type;
        typedef T10                                                    arg10_type;
        typedef T11                                                    arg11_type;
        typedef T12                                                    arg12_type;
        typedef T13                                                    arg13_type;
        typedef T14                                                    arg14_type;
        typedef T15                                                    arg15_type;
        typedef T16                                                    arg16_type;
        typedef T17                                                    arg17_type;
        typedef T18                                                    arg18_type;
        typedef T19                                                    arg19_type;
        typedef T20                                                    arg20_type;
        typedef T21                                                    arg21_type;
        typedef T22                                                    arg22_type;
        typedef T23                                                    arg23_type;

        typedef typename Traits<arg0_type>::template bind<Allocator>  stash0_traits;
        typedef typename Traits<arg1_type>::template bind<Allocator>  stash1_traits;
        typedef typename Traits<arg2_type>::template bind<Allocator>  stash2_traits;
        typedef typename Traits<arg3_type>::template bind<Allocator>  stash3_traits;
        typedef typename Traits<arg4_type>::template bind<Allocator>  stash4_traits;
        typedef typename Traits<arg5_type>::template bind<Allocator>  stash5_traits;
        typedef typename Traits<arg6_type>::template bind<Allocator>  stash6_traits;
        typedef typename Traits<arg7_type>::template bind<Allocator>  stash7_traits;
        typedef typename Traits<arg8_type>::template bind<Allocator>  stash8_traits;
        typedef typename Traits<arg9_type>::template bind<Allocator>  stash9_traits;
        typedef typename Traits<arg10_type>::template bind<Allocator>  stash10_traits;
        typedef typename Traits<arg11_type>::template bind<Allocator>  stash11_traits;
        typedef typename Traits<arg12_type>::template bind<Allocator>  stash12_traits;
        typedef typename Traits<arg13_type>::template bind<Allocator>  stash13_traits;
        typedef typename Traits<arg14_type>::template bind<Allocator>  stash14_traits;
        typedef typename Traits<arg15_type>::template bind<Allocator>  stash15_traits;
        typedef typename Traits<arg16_type>::template bind<Allocator>  stash16_traits;
        typedef typename Traits<arg17_type>::template bind<Allocator>  stash17_traits;
        typedef typename Traits<arg18_type>::template bind<Allocator>  stash18_traits;
        typedef typename Traits<arg19_type>::template bind<Allocator>  stash19_traits;
        typedef typename Traits<arg20_type>::template bind<Allocator>  stash20_traits;
        typedef typename Traits<arg21_type>::template bind<Allocator>  stash21_traits;
        typedef typename Traits<arg22_type>::template bind<Allocator>  stash22_traits;
        typedef typename Traits<arg23_type>::template bind<Allocator>  stash23_traits;

        typedef typename stash0_traits::stash_type                    stash0_type;
        typedef typename stash1_traits::stash_type                    stash1_type;
        typedef typename stash2_traits::stash_type                    stash2_type;
        typedef typename stash3_traits::stash_type                    stash3_type;
        typedef typename stash4_traits::stash_type                    stash4_type;
        typedef typename stash5_traits::stash_type                    stash5_type;
        typedef typename stash6_traits::stash_type                    stash6_type;
        typedef typename stash7_traits::stash_type                    stash7_type;
        typedef typename stash8_traits::stash_type                    stash8_type;
        typedef typename stash9_traits::stash_type                    stash9_type;
        typedef typename stash10_traits::stash_type                    stash10_type;
        typedef typename stash11_traits::stash_type                    stash11_type;
        typedef typename stash12_traits::stash_type                    stash12_type;
        typedef typename stash13_traits::stash_type                    stash13_type;
        typedef typename stash14_traits::stash_type                    stash14_type;
        typedef typename stash15_traits::stash_type                    stash15_type;
        typedef typename stash16_traits::stash_type                    stash16_type;
        typedef typename stash17_traits::stash_type                    stash17_type;
        typedef typename stash18_traits::stash_type                    stash18_type;
        typedef typename stash19_traits::stash_type                    stash19_type;
        typedef typename stash20_traits::stash_type                    stash20_type;
        typedef typename stash21_traits::stash_type                    stash21_type;
        typedef typename stash22_traits::stash_type                    stash22_type;
        typedef typename stash23_traits::stash_type                    stash23_type;

        stash0_type                                                   t0;
        stash1_type                                                   t1;
        stash2_type                                                   t2;
        stash3_type                                                   t3;
        stash4_type                                                   t4;
        stash5_type                                                   t5;
        stash6_type                                                   t6;
        stash7_type                                                   t7;
        stash8_type                                                   t8;
        stash9_type                                                   t9;
        stash10_type                                                   t10;
        stash11_type                                                   t11;
        stash12_type                                                   t12;
        stash13_type                                                   t13;
        stash14_type                                                   t14;
        stash15_type                                                   t15;
        stash16_type                                                   t16;
        stash17_type                                                   t17;
        stash18_type                                                   t18;
        stash19_type                                                   t19;
        stash20_type                                                   t20;
        stash21_type                                                   t21;
        stash22_type                                                   t22;
        stash23_type                                                   t23;

        typedef void                                                  signature(T0, T1, T2, T3, T4, T5, T6, T7, T8, T9, T10, T11, T12, T13, T14, T15, T16, T17, T18, T19, T20, T21, T22, T23);
        typedef void                                                  stash_signature(stash0_type,stash1_type,stash2_type,stash3_type,stash4_type,stash5_type,stash6_type,stash7_type,stash8_type,stash9_type,stash10_type,stash11_type,stash12_type,stash13_type,stash14_type,stash15_type,stash16_type,stash17_type,stash18_type,stash19_type,stash20_type,stash21_type,stash22_type,stash23_type);

        template<typename OP> void apply(OP op_)
        {
            op_(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21),
                stash22_traits::from_stash(t22),
                stash23_traits::from_stash(t23)
                );
        }

        template<typename OP> void apply(OP* op_)
        {
            (*op_)(
                stash0_traits::from_stash(t0),
                stash1_traits::from_stash(t1),
                stash2_traits::from_stash(t2),
                stash3_traits::from_stash(t3),
                stash4_traits::from_stash(t4),
                stash5_traits::from_stash(t5),
                stash6_traits::from_stash(t6),
                stash7_traits::from_stash(t7),
                stash8_traits::from_stash(t8),
                stash9_traits::from_stash(t9),
                stash10_traits::from_stash(t10),
                stash11_traits::from_stash(t11),
                stash12_traits::from_stash(t12),
                stash13_traits::from_stash(t13),
                stash14_traits::from_stash(t14),
                stash15_traits::from_stash(t15),
                stash16_traits::from_stash(t16),
                stash17_traits::from_stash(t17),
                stash18_traits::from_stash(t18),
                stash19_traits::from_stash(t19),
                stash20_traits::from_stash(t20),
                stash21_traits::from_stash(t21),
                stash22_traits::from_stash(t22),
                stash23_traits::from_stash(t23)
                );
        }

        void init(Allocator& a_, arg0_type const& arg0_, arg1_type const& arg1_, arg2_type const& arg2_, arg3_type const& arg3_, arg4_type const& arg4_, arg5_type const& arg5_, arg6_type const& arg6_, arg7_type const& arg7_, arg8_type const& arg8_, arg9_type const& arg9_, arg10_type const& arg10_, arg11_type const& arg11_, arg12_type const& arg12_, arg13_type const& arg13_, arg14_type const& arg14_, arg15_type const& arg15_, arg16_type const& arg16_, arg17_type const& arg17_, arg18_type const& arg18_, arg19_type const& arg19_, arg20_type const& arg20_, arg21_type const& arg21_, arg22_type const& arg22_, arg23_type const& arg23_)
        {
            stash0_traits::stash_in(a_, t0, arg0_);
            stash1_traits::stash_in(a_, t1, arg1_);
            stash2_traits::stash_in(a_, t2, arg2_);
            stash3_traits::stash_in(a_, t3, arg3_);
            stash4_traits::stash_in(a_, t4, arg4_);
            stash5_traits::stash_in(a_, t5, arg5_);
            stash6_traits::stash_in(a_, t6, arg6_);
            stash7_traits::stash_in(a_, t7, arg7_);
            stash8_traits::stash_in(a_, t8, arg8_);
            stash9_traits::stash_in(a_, t9, arg9_);
            stash10_traits::stash_in(a_, t10, arg10_);
            stash11_traits::stash_in(a_, t11, arg11_);
            stash12_traits::stash_in(a_, t12, arg12_);
            stash13_traits::stash_in(a_, t13, arg13_);
            stash14_traits::stash_in(a_, t14, arg14_);
            stash15_traits::stash_in(a_, t15, arg15_);
            stash16_traits::stash_in(a_, t16, arg16_);
            stash17_traits::stash_in(a_, t17, arg17_);
            stash18_traits::stash_in(a_, t18, arg18_);
            stash19_traits::stash_in(a_, t19, arg19_);
            stash20_traits::stash_in(a_, t20, arg20_);
            stash21_traits::stash_in(a_, t21, arg21_);
            stash22_traits::stash_in(a_, t22, arg22_);
            stash23_traits::stash_in(a_, t23, arg23_);
        }

        std::size_t dynamic_data_size() const
        {
            return stash0_traits::dynamic_data_size(t0)+stash1_traits::dynamic_data_size(t1)+stash2_traits::dynamic_data_size(t2)+stash3_traits::dynamic_data_size(t3)+stash4_traits::dynamic_data_size(t4)+stash5_traits::dynamic_data_size(t5)+stash6_traits::dynamic_data_size(t6)+stash7_traits::dynamic_data_size(t7)+stash8_traits::dynamic_data_size(t8)+stash9_traits::dynamic_data_size(t9)+stash10_traits::dynamic_data_size(t10)+stash11_traits::dynamic_data_size(t11)+stash12_traits::dynamic_data_size(t12)+stash13_traits::dynamic_data_size(t13)+stash14_traits::dynamic_data_size(t14)+stash15_traits::dynamic_data_size(t15)+stash16_traits::dynamic_data_size(t16)+stash17_traits::dynamic_data_size(t17)+stash18_traits::dynamic_data_size(t18)+stash19_traits::dynamic_data_size(t19)+stash20_traits::dynamic_data_size(t20)+stash21_traits::dynamic_data_size(t21)+stash22_traits::dynamic_data_size(t22)+stash23_traits::dynamic_data_size(t23);
        }

        bool operator == (arg_holder_t const& o_) const
        {
            return t0 == o_.t0 && t1 == o_.t1 && t2 == o_.t2 && t3 == o_.t3 && t4 == o_.t4 && t5 == o_.t5 && t6 == o_.t6 && t7 == o_.t7 && t8 == o_.t8 && t9 == o_.t9 && t10 == o_.t10 && t11 == o_.t11 && t12 == o_.t12 && t13 == o_.t13 && t14 == o_.t14 && t15 == o_.t15 && t16 == o_.t16 && t17 == o_.t17 && t18 == o_.t18 && t19 == o_.t19 && t20 == o_.t20 && t21 == o_.t21 && t22 == o_.t22 && t23 == o_.t23;
        }

        struct deferrer {
            deferrer( Allocator& a_, arg_holder_t& ah_ )
                : a(a_), ah(ah_) {}

            Allocator&    a;
            arg_holder_t& ah;

            void init(arg0_type t0_,arg1_type t1_,arg2_type t2_,arg3_type t3_,arg4_type t4_,arg5_type t5_,arg6_type t6_,arg7_type t7_,arg8_type t8_,arg9_type t9_,arg10_type t10_,arg11_type t11_,arg12_type t12_,arg13_type t13_,arg14_type t14_,arg15_type t15_,arg16_type t16_,arg17_type t17_,arg18_type t18_,arg19_type t19_,arg20_type t20_,arg21_type t21_,arg22_type t22_,arg23_type t23_)
            {
                ah.init(a, t0_,t1_,t2_,t3_,t4_,t5_,t6_,t7_,t8_,t9_,t10_,t11_,t12_,t13_,t14_,t15_,t16_,t17_,t18_,t19_,t20_,t21_,t22_,t23_);
            }
        };
    };


    // ----------------------------------------------------------------------
    // Allocate a deferred op_, return an object which can be used to 'init()'
    // the arguments to be deferred.
    // ----------------------------------------------------------------------
    template<template<typename> class Traits, typename Allocator, typename OP>
    typename arg_holder_t<Allocator, Traits, typename OP::signature>::deferrer
    defer(Allocator& a_, OP op_)
    {
        typedef arg_holder_t<Allocator,Traits, typename OP::signature> this_arg_holder_t;
        typedef typename this_arg_holder_t::deferrer                   deferrer_t;
        typedef op_and_arg_holder_thunk_t< OP, this_arg_holder_t >     th_t;

        new (a_.allocate_virtual_thunk(sizeof(th_t))) th_t(op_);

        this_arg_holder_t* ah = new (a_.allocate_arg_holder(sizeof(this_arg_holder_t))) this_arg_holder_t;

        return deferrer_t( a_, *ah );
    }

} // end namespace

#endif // ARM_INCLUDE_SGARGMETA_H

