///
/// \file    SGComponentTraceHelper.h
/// \brief   A set of macros and meta-programming widgets to make tracing easier
/// \date    Copyright 2011-2013 ARM Limited. All rights reserved.
///
/// This is an automatically generated File please do not edit.
///
#ifndef ARM_INCLUDE_SGCOMPONENTTRACEHELPER_H
#define ARM_INCLUDE_SGCOMPONENTTRACEHELPER_H

#include "sg/MSCFixes.h"
#include "sg/SGComponentTrace.h"
#include "sg/SGArgMeta.h"

#include <cstdlib>
#include <vector>
#include <cstring>

#include "sg/SGDeferredTraceDefaultAllocator.h"

///
/// \page Deferred trace sources
///
/// Each trace event can be an opportunity for the code to lose control to a
/// trace callback and it is unknown what that callback might do.  This might
/// include making transactions into your component.
///
/// Thus each trace event has to be at a point at which you are re-entrant safe.
/// It is hard to test such a thing.  An approach that you can do is to defer
/// all trace events until a point in your code where you are forced to be
/// re-entrant safe because you are going to make an external bus transaction or
/// signal, etc.  Just before you lose control, then you flush the deferred
/// trace records out.
///
/// This methodology ensures that instead of (somehow) testing 100s of points at
/// which you have to be re-entrant safe then you only has to test a few points.
///
/// Deferring trace records mean that you have to have a stash to store them in.
/// Ideally this would be fibre-local -- so that each fibre buffers the trace
/// records until such time as it wishes to flush them, and that it can do this
/// without interference from other fibres.  In the absence of fibre-local
/// storage then we would have to pass the allocator/stash around to every
/// function-call that is going to generate a trace call.  This is very ugly at
/// a code level and also incurs execution cost in stacking and unstacking these
/// parameters.
///
/// Instead, we note that when a trace source is registered, then it can also be
/// given the allocator.  As a fibre only gives up control when it wants to then
/// only a single fibre will be actually executing in a specific component at
/// any particular moment.  Thus we need not need fibre-local caches as the
/// trace stream is flushed when the fibre loses control and if a new fibre
/// takes control then it can re-use the stash.
///
/// If the fibre, for whatever reason, didn't want to emit the trace immediately
/// before losing control then all it has to do is swap out the current
/// allocator and restore it when it re-gains control.  This is very neat and
/// efficient.  At a code level, it means that calling a trace source looks the
/// same without having to pass the stash around.
///
/// The problem is that the Allocator method used is a template parameter, thus
/// the Allocator type becomes a template parameter to the trace source.
///

template<unsigned N> struct trace_fixed_length_string {
    static unsigned const     size = N;

    char const* s;
    operator char const*() const { return s; }

// see <201305141725/> for why this is commented
//    uint8_t const* get()          const { return (uint8_t const*)s; }
//    std::size_t    dynamic_size() const { return ::strlen(s) + 1;   }

    trace_fixed_length_string( char const* s_ = "" ) : s(s_) {}
    trace_fixed_length_string( std::size_t, uint8_t const* p_ ) : s((char const*)p_) {}
};

namespace trace_source_details
{
    // <201304261013>
    //
    // stash_type is how to stash a traced value away for future use.
    //
    // It uses an allocator to get storage for how to proceed.
    //
    // The stash_type is expected to allocate data on construction only -- it
    // is expected that it will disallow the copy constructor and assignment
    // operator for the moment.   Further use-cases might mean we have to be
    // a bit more sophisticated.
    //
    // The stash_type is expected to have an operator ==() defined.
    //
    // </201304261013>

    template< typename T > struct traits_base
    {
        typedef T                               type;
        static MTI::EventFieldType::Type const  type_spec = MTI::EventFieldType::MTI_UNSIGNED_INT;
        static unsigned const                   size      = sizeof(type);
        static unsigned const                   max_size  = 0; // default for fixed length

        typedef T                               arg_type;

        static type convert_for_trace(T in_)
        {
            return in_;
        }

        // see <201304261013/> for stash_type
        template<typename Allocator> struct bind {
            typedef T stash_type;

            static void stash_in(Allocator const&, stash_type& o_, T in_)
            {
                o_ = in_;
            }

            static type from_stash(stash_type const& in_)
            {
                return in_;
            }

            static std::size_t dynamic_data_size( stash_type const& )
            {
                return 0;
            }

            static typename Allocator::size_type const max_size = 0; // default for fixed length -- this says
                                                                     // the stash needs no dynamically allocated memory
        };
    };

    template< typename T > struct traits : public traits_base<T>
    {
    };

    template<> struct traits<bool> : public traits_base<bool>
    {
        static MTI::EventFieldType::Type const type_spec = MTI::EventFieldType::MTI_BOOL;
    };

    template<> struct traits<char> : public traits_base<char>
    {
        static MTI::EventFieldType::Type const type_spec = MTI::EventFieldType::MTI_STRING;
    };

    template<typename T> struct traits_signed : public traits_base<T>
    {
        static MTI::EventFieldType::Type const type_spec = MTI::EventFieldType::MTI_SIGNED_INT;
    };
    template<> struct traits<int8_t> : public traits_signed<int8_t>
    {
    };
    template<> struct traits<int16_t> : public traits_signed<int16_t>
    {
    };
    template<> struct traits<int32_t> : public traits_signed<int32_t>
    {
    };
    template<> struct traits<int64_t> : public traits_signed<int64_t>
    {
    };

    namespace details {
        //
        // This stash_type can work for any T which has methods:
        //      std::size_t    dynamic_size() const
        //      uint8_t const* get() const
        //      T(std::size_t sz_, uint8_t const* ptr_);
        // and the resultant can be represented solely by that
        // byte stream.   It is assumed that the lifetime of the pointers
        // are 'appropriate' and so can just be stored.
        //
        // see <201304261013/> for stash_type
        //
        // T is the traced type as defined in the DEFINE_SIMPLE_TRACE_SOURCE_CLASS
        // type is the MTI trace type from the type traits for T.
        //
        template<typename T, typename type, typename Allocator, std::size_t MAX_SIZE>
        struct stash_common {
            struct stash_type {

                stash_type() {}

                typename Allocator::size_type  sz;
                typename Allocator::pointer    p;

                bool operator == (stash_type const& o_) const
                {
                    return sz == o_.sz && Allocator::is_equal(p,o_.p,sz);
                }

            private:
                stash_type(stash_type const&); // disallow
                void operator = (stash_type const&); // disallow
            };

            static void stash_in(Allocator& allocator_, stash_type& out_, T const& in_)
            {
                typedef typename Allocator::size_type sz_t;
                out_.sz = sz_t(in_.dynamic_size());
                allocator_.allocate_dynamic_data(out_.sz, out_.p);
                allocator_.copy( in_.get(), in_.get() + out_.sz, out_.p );
            }

            static type from_stash(stash_type const& in_)
            {
                return type(static_cast<std::size_t>(in_.sz),
                            static_cast<uint8_t const*>(in_.p));
            }

            static std::size_t dynamic_data_size( stash_type const& t_ )
            {
                return t_.sz;
            }

            static std::size_t const max_size = MAX_SIZE; // the stash needs to allocate at most MAX_SIZE
                                                          // dynamically allocated memory
        };
    }

    template<unsigned N> struct traits< trace_fixed_length_string<N> >
      : public traits_base< trace_fixed_length_string<N> >
    {
        typedef char const*                     type;
        static MTI::EventFieldType::Type const  type_spec = MTI::EventFieldType::MTI_STRING;
        static unsigned const                   size = 0;
        static unsigned const                   max_size  = N;

        // see <201304261013/> for stash_type
        //
        // <201305141725>
        //
        // Unfortunately as trace_fixed_length_string<N> isn't a first class
        // object in the trace system then we convert to char const* too early for
        // this binding statement to be used, instead we fall through the cracks
        // and use traits<char const*>.   The bind method comes from traits_base
        // which we specialise for char const* and hope for the best.
        //
        // traits<char const*> should not be used for normal purposes, as such we
        // have removed most of its fields so that if you use it directly then it
        // will compile-fail.  However, we have to leave the 'bind' system there
        // to cope with this case.
        //
        // The below is the code that would be used if the binding could be
        // used.
        //
        // </201305141725>
//       template<typename Allocator>
//       struct bind
//           : details::stash_common<trace_fixed_length_string<N>, type, Allocator, N> {
//       };
    };

    // These traits are used by the above code. Unfortunately using this
    // directly will lead to random data corruption due to the construction of
    // MTI.
    // Prefer trace_fixed_length_string<N> instead.
    template<> struct traits<char const*>
    {
        typedef int use_trace_fixed_length_string_instead;


        // see <201305141725/> for why we have to do this.
        template<typename Allocator> struct bind {
            typedef typename Allocator::size_type  size_type;
            typedef typename Allocator::pointer    pointer;

            struct stash_type {

                stash_type() {}

                size_type  sz;
                pointer    p;

                bool operator == (stash_type const& o_) const
                {
                    return sz == o_.sz && Allocator::is_equal(p,o_.p,sz);
                }

            private:
                stash_type(stash_type const&); // disallow
                void operator = (stash_type const&); // disallow
            };

            static void stash_in(Allocator& allocator_, stash_type& out_, char const* in_)
            {
                out_.sz = size_type(::strlen(in_) + 1);
                allocator_.allocate_dynamic_data(out_.sz, out_.p);
                allocator_.copy( (uint8_t const*)in_, (uint8_t const*)in_ + out_.sz, out_.p );
            }

            static char const* from_stash(stash_type const& in_)
            {
                return (char const*)static_cast<uint8_t const*>(in_.p);
            }

            static std::size_t dynamic_data_size( stash_type const& t_ )
            {
                return t_.sz;
            }
        };
    };

    template<size_t max_size_> struct traits<sg::size_and_pointer<max_size_> >
      : public traits_base< sg::size_and_pointer<max_size_> >
    {
        typedef sg::size_and_pointer<max_size_> type;
        static MTI::EventFieldType::Type const  type_spec = MTI::EventFieldType::MTI_UNSIGNED_INT;
        static unsigned const                   size      = 0; // variable length
        static unsigned const                   max_size  = max_size_;

        // see <201304261013/> for stash_type
        template<typename Allocator>
        struct bind
            : details::stash_common<sg::size_and_pointer<max_size_>, type, Allocator, max_size_> {
        };
    };

    /// helper class used to encapsulate the description and any symbols that may be attached.
    class description
    {
        std::string const d;

        // This will leak memory!
        mutable std::vector< std::pair<uint32_t, char const*> >   symbols;

        static char* global_deleter(char*);
    public:
        description( char const*        description_ ) : d(description_ == 0 ? "" : description_) {}
        description( std::string const& description_ ) : d(description_) {}

        std::string get_description() const { return d; }

        description& operator () (uint32_t v, std::string const& s_);

        /// This is 'dangerous' in the sense that it works on a const object
        description const& dangerous_add(uint32_t v, std::string const& s_) const;

        description& operator () (uint32_t v, char const* s_);

        /// This is 'dangerous' in the sense that it works on a const object
        description const& dangerous_add(uint32_t v, char const* s_) const;

        char const* get_symbol_name_from_value( uint32_t v ) const;

        description& symbol( uint32_t v, std::string const& s_ ) { return (*this)(v,s_); }
        description& symbol( uint32_t v, char const*        s_ ) { return (*this)(v,s_); }

        size_t get_number_of_symbols() const { return symbols.size(); }

        std::pair<uint32_t, char const*> get_symbol(size_t i_) const;
    };

    struct base
    {
        virtual ~base();
        virtual void init() = 0;

      protected:

        std::vector<char*>   strduped_strings_to_destroy;

        char* dup(std::string const& s_);

        void add_field(
            sg::EventSourceBase*          event_source_base_,
            std::string const&            field_name_,
            description const&            description_,
            ::MTI::EventFieldType::Type   type_,
            ::MTI::EventFieldType::Size   size_,
            ::MTI::EventFieldType::Size   max_size_
            );
    };

    /// This should never be instantiated, it means that the FIELD_TUPLES (or arguments to describe_fields)
    /// are the wrong number compared to the number of arguments in the SIGNATURE_TUPLE.
    template<typename T>
    struct error_wrong_number_of_fields_in_description
    {
    };

    static inline ::trace_source_details::description const& enhance_trace_description(
        ::trace_source_details::description const& description_,
        ...
        )
    {
        return description_;
    }
}


/// Base template for all trace_source_t, should never instantiated directly.
template<typename T, typename Allocator = trace_source_details::default_allocator> struct trace_source_t {};


///
/// Simplify registering objects with a trace sink.
///
/// For example, if you have an sg::CADIBase or fs::CADIInterface then you
/// can do:
/// <code>
///     register_objects_with_component_trace( my_cadi_interface )
///         << my_trace_source0
///         << my_trace_source1
///         << my_trace_source2
///         ;
/// </code>
///
/// There is no requirement to use this, it is supplied as a convenience to make
/// your code simpler.
///
template<typename T, typename Allocator>
struct register_objects_with_component_trace_t
{
    T&          sink;
    Allocator*  allocator;

    register_objects_with_component_trace_t( T& sink_, Allocator* allocator_ )
        : sink(sink_),
          allocator(allocator_)
    {
    }
    // default assignment, copy constructor and destructor are OK.

    template<typename TRACE_SOURCE>
    register_objects_with_component_trace_t& operator <<(TRACE_SOURCE& trace_source_)
    {
        trace_source_details::base& b = trace_source_;
        b.init();
        sink.RegisterTraceSource( &trace_source_ );
        set_allocator( &trace_source_, &trace_source_ );
        return *this;
    }

private:
    // Strange form is so that we can cope with being given objects that are not
    // trace_source_t<>.  The first form captures all those that are have no
    // trace_source_t<> in the inheritance hierarchy, the second form captures
    // all trace_source_t<> and all that are derived from them (calling the derived
    // set_allocator() if it has overridden it).
    template<typename U>                void set_allocator( U*, ... ) {}
    template<typename U, typename SIG>  void set_allocator( U* u_, trace_source_t<SIG,Allocator>* )
    {
         u_->set_allocator(allocator);
    }
};

template<typename T>
inline register_objects_with_component_trace_t<T,trace_source_details::default_allocator> register_objects_with_component_trace(T& sink_)
{
    return register_objects_with_component_trace_t<T,trace_source_details::default_allocator>(sink_, 0);
}

template<typename T, typename Allocator>
inline register_objects_with_component_trace_t<T,Allocator> register_objects_with_component_trace(T& sink_, Allocator* allocator_)
{
    return register_objects_with_component_trace_t<T,Allocator>(sink_, allocator_);
}

template<typename T>
struct register_hidden_objects_with_component_trace_t
{
    T&          sink;

    register_hidden_objects_with_component_trace_t( T& sink_ )
        : sink(sink_)
    {
    }
    // default assignment, copy constructor and destructor are OK.

    template<typename TRACE_SOURCE>
    register_hidden_objects_with_component_trace_t& operator <<(TRACE_SOURCE& trace_source_)
    {
        trace_source_details::base& b = trace_source_;
        b.init();
        sink.RegisterTraceSourceHidden( &trace_source_ );
        return *this;
    }
};

template<typename T>
inline register_hidden_objects_with_component_trace_t<T> register_hidden_objects_with_component_trace(T& sink_)
{
    return register_hidden_objects_with_component_trace_t<T>(sink_);
}

///
/// \define DEFINE_SIMPLE_TRACE_SOURCE_CLASS
///
/// Define a class CNAME that is-a trace_source_t.
///
/// NAME and DESCRIPTION is the name and description of the event source.
///
/// SIGNATURE_TUPLE is something like (uint64_t) or (uint64_t, char), i.e.
/// a parenthesised expression of types.
///
/// FIELD_TUPLES is a parenthesised tuple of names and descriptions for each of
/// the fields, e.g. ("field0", "field0 description", "field1", "field1 description")
///
/// In addition, the description can specify an enumeration, an example is:
/// <code>
///           description( "Line transformation incurred" )
///           ('I', "invalid")
///           ('D', "unaffected and dirty")
///           ('C', "unaffected and clean")
///           ('X', "already clean")
///           ('x', "was dirty now clean")
/// </code>
///
/// MTI limits us to enumerations values that are a maximum of 32 bits.  The
/// corresponding member of the signature must be sufficient to store the
/// enumerated values.
///
/// An example to show how to capture an active change event:
/// <code>
///      DEFINE_SIMPLE_TRACE_SOURCE_CLASS(
///          gic_log_warnings_out_trace_base_t,
///          "gic_log_warnings_out",
///          "Warnings from the VGIC model",
///          (trace_fixed_length_string<200>)
///          ( "message",
///            "The message from the VGIC" );
///          );
///
///      struct gic_log_warnings_out_trace_t : public gic_log_warnings_out_trace_base_t
///      {
///          virtual void activeStateChanged( bool new_active_state_ )
///          {
///              ...
///          }
///      } gic_log_warnings_out_trace;
///
///      gic_log_warnings_out_trace( "example message" ); // example on how to signal a trace
///  <code>

#define DEFINE_SIMPLE_TRACE_SOURCE_CLASS( CNAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    struct CNAME : trace_source_t<void SIGNATURE_TUPLE> {               \
                                                                        \
        typedef trace_source_details::description description;          \
        virtual void init()                                             \
        {                                                               \
            if (this->GetName()[0] == '\0')                             \
                this->setName(NAME);                                    \
            if (this->GetDescription()[0] == '\0')                      \
                this->setDescription(DESCRIPTION);                      \
            this->describe_fields FIELD_TUPLES;                         \
        }                                                               \
    }

///
/// \define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_CLASS
///
/// Same as DEFINE_SIMPLE_TRACE_SOURCE_CLASS except that it hides the
/// fire() and operator() methods so that you are forced to use the
/// defer() function call.
///

#define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_CLASS( CNAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_CLASS_WITH_ALLOCATOR(\
        ::trace_source_details::default_allocator,           \
        CNAME,                                               \
        NAME,                                                \
        DESCRIPTION,                                         \
        SIGNATURE_TUPLE,                                     \
        FIELD_TUPLES                                         \
        )

#define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_CLASS_WITH_ALLOCATOR( ALLOCATOR, CNAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    struct CNAME : trace_source_t<void SIGNATURE_TUPLE, ALLOCATOR> {    \
                                                                        \
        typedef trace_source_details::description description;          \
        virtual void init()                                             \
        {                                                               \
            if (this->GetName()[0] == '\0')                             \
                this->setName(NAME);                                    \
            if (this->GetDescription()[0] == '\0')                      \
                this->setDescription(DESCRIPTION);                      \
            this->describe_fields FIELD_TUPLES;                         \
        }                                                               \
    private:                                                            \
        void operator() SIGNATURE_TUPLE; /*disallow*/                   \
        void fire SIGNATURE_TUPLE; /*disallow*/                         \
    }

///
/// \define DEFINE_SIMPLE_TRACE_SOURCE
///
/// Same as DEFINE_SIMPLE_TRACE_SOURCE_CLASS except the INAME is instance name,
/// and the class is INAME_t.
///

#define DEFINE_SIMPLE_TRACE_SOURCE( INAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    struct INAME##_t : trace_source_t<void SIGNATURE_TUPLE> {           \
                                                                        \
        typedef trace_source_details::description description;          \
        virtual void init()                                             \
        {                                                               \
            if (this->GetName()[0] == '\0')                             \
                this->setName(NAME);                                    \
            if (this->GetDescription()[0] == '\0')                      \
                this->setDescription(DESCRIPTION);                      \
            this->describe_fields FIELD_TUPLES;                         \
        }                                                               \
    } INAME

///
/// \define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE
///
/// Same as DEFINE_SIMPLE_TRACE_SOURCE except that it hides the
/// fire() and operator() methods so that you are forced to use the
/// defer() function call.
///

#define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE( INAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_WITH_ALLOCATOR(                 \
        ::trace_source_details::default_allocator,                      \
        INAME,                                                          \
        NAME,                                                           \
        DESCRIPTION,                                                    \
        SIGNATURE_TUPLE,                                                \
        FIELD_TUPLES                                                    \
        )

#define DEFINE_SIMPLE_DEFERRED_TRACE_SOURCE_WITH_ALLOCATOR( ALLOCATOR, INAME, NAME, DESCRIPTION, SIGNATURE_TUPLE, FIELD_TUPLES ) \
    struct INAME##_t : trace_source_t<void SIGNATURE_TUPLE, ALLOCATOR> { \
                                                                        \
        typedef trace_source_details::description description;          \
        virtual void init()                                             \
        {                                                               \
            if (this->GetName()[0] == '\0')                             \
                this->setName(NAME);                                    \
            if (this->GetDescription()[0] == '\0')                      \
                this->setDescription(DESCRIPTION);                      \
            this->describe_fields FIELD_TUPLES;                         \
        }                                                               \
    private:                                                            \
        void operator() SIGNATURE_TUPLE; /*disallow*/                   \
        void fire SIGNATURE_TUPLE; /*disallow*/                         \
    } INAME


namespace trace_source_details
{
    // A virtual thunk that defers a call to obj->fire()
    template<typename Object, typename ARG_HOLDER>
    class fire_virtual_thunk_t : public sg_meta::virtual_thunk_if
    {
        Object* obj;
    public:

        fire_virtual_thunk_t(Object* obj_) : obj(obj_) {}

        virtual std::size_t sizeof_this() const
        {
            return sizeof(*this);
        }

        virtual sizes_t sizeof_all(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            return sizes_t(sizeof(*this),
                           sizeof(ARG_HOLDER),
                           arg_holder->dynamic_data_size());
        }

        virtual sizes_t exec(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            sizes_t const szs(sizeof(*this),
                              sizeof(ARG_HOLDER),
                              arg_holder->dynamic_data_size());
            arg_holder->apply(sg_meta::make_mem_fun_op(obj, &Object::fire));
            return szs;
        }

        virtual sizes_t destroy_arg_holder(void* arg_holder_) const
        {
            ARG_HOLDER* const arg_holder = static_cast<ARG_HOLDER*>(arg_holder_);
            sizes_t const szs(sizeof(*this),
                              sizeof(ARG_HOLDER),
                              arg_holder->dynamic_data_size());
            arg_holder->~ARG_HOLDER();
            return szs;
        }

        virtual void const* object_pointer() const { return obj; }

        virtual bool is_equal(void const* other_, void const* lhs_arg_holder_, void const* rhs_arg_holder_) const
        {
            if (other_ != obj || obj == 0)
                return false;

            ARG_HOLDER const* const lhs = static_cast<ARG_HOLDER const*>(lhs_arg_holder_);
            ARG_HOLDER const* const rhs = static_cast<ARG_HOLDER const*>(rhs_arg_holder_);

            return *lhs == *rhs;
        }
    };

    // ----------------------------------------------------------------------
    // Allocate a deferred call to obj_->fire(), return an object which can be
    // used to 'init()' the deferred call.
    //
    // OP is interesting in that in order to find out the signature of fire()
    // then you have to pass it make_mem_fun_op(obj_, &object_type::fire)
    //
    // It doesn't seem easily possible in C++ to deduce the fire signature
    // directly.  If you wrap this in another call that attempts to do the
    // make_mem_fun_op() for you then you can't figure out what the return
    // type should be.  The alternative is that we add 'signature' to the
    // trace_source_t object which would make this cleaner but produce a
    // larger source file.
    // ----------------------------------------------------------------------
    template<typename Allocator, typename OP>
    typename ::sg_meta::arg_holder_t<Allocator, traits, typename OP::signature>::deferrer
    defer_fire(Allocator& a_, OP op_)
    {
        typedef ::sg_meta::arg_holder_t<Allocator,traits, typename OP::signature>   this_arg_holder_t;
        typedef typename this_arg_holder_t::deferrer                                deferrer_t;
        typedef fire_virtual_thunk_t< typename OP::class_type, this_arg_holder_t >  th_t;

        new (a_.allocate_virtual_thunk(sizeof(th_t))) th_t(op_.obj);

        this_arg_holder_t* ah = new (a_.allocate_arg_holder(sizeof(this_arg_holder_t))) this_arg_holder_t;

        return deferrer_t( a_, *ah );
    }
}


// Zero case is sufficient different to be specified separately
template<typename Allocator>
struct trace_source_t<void(), Allocator>
    : trace_source_details::base,
      sg::EventSource0
{
    typedef sg::EventSource0 base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 0;

    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields()
    {
    }

    void defer()
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init();
            else
                this->fire();
        }
    }

    void operator()()
    {
        this->fire();
    }

};



template< typename T0, typename Allocator >
struct trace_source_t<void(T0), Allocator>
    : trace_source_details::base,
      sg::EventSource1<typename trace_source_details::traits<T0>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;

    typedef typename tr_0_t::type                              _0_t;


    typedef sg::EventSource1<_0_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 1;
    typedef typename tr_0_t::arg_type    arg0_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
    }

    void defer(arg0_type t0_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_) );
        }
    }

    void operator()( arg0_type t0 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename Allocator >
struct trace_source_t<void(T0,T1), Allocator>
    : trace_source_details::base,
      sg::EventSource2<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;


    typedef sg::EventSource2<_0_t,_1_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 2;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename Allocator >
struct trace_source_t<void(T0,T1,T2), Allocator>
    : trace_source_details::base,
      sg::EventSource3<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;


    typedef sg::EventSource3<_0_t,_1_t,_2_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 3;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3), Allocator>
    : trace_source_details::base,
      sg::EventSource4<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;


    typedef sg::EventSource4<_0_t,_1_t,_2_t,_3_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 4;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4), Allocator>
    : trace_source_details::base,
      sg::EventSource5<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;


    typedef sg::EventSource5<_0_t,_1_t,_2_t,_3_t,_4_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 5;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5), Allocator>
    : trace_source_details::base,
      sg::EventSource6<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;


    typedef sg::EventSource6<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 6;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6), Allocator>
    : trace_source_details::base,
      sg::EventSource7<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;


    typedef sg::EventSource7<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 7;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7), Allocator>
    : trace_source_details::base,
      sg::EventSource8<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;


    typedef sg::EventSource8<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 8;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8), Allocator>
    : trace_source_details::base,
      sg::EventSource9<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;


    typedef sg::EventSource9<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 9;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9), Allocator>
    : trace_source_details::base,
      sg::EventSource10<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;


    typedef sg::EventSource10<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 10;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10), Allocator>
    : trace_source_details::base,
      sg::EventSource11<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;


    typedef sg::EventSource11<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 11;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11), Allocator>
    : trace_source_details::base,
      sg::EventSource12<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;


    typedef sg::EventSource12<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 12;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12), Allocator>
    : trace_source_details::base,
      sg::EventSource13<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;


    typedef sg::EventSource13<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 13;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13), Allocator>
    : trace_source_details::base,
      sg::EventSource14<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;


    typedef sg::EventSource14<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 14;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14), Allocator>
    : trace_source_details::base,
      sg::EventSource15<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;


    typedef sg::EventSource15<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 15;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15), Allocator>
    : trace_source_details::base,
      sg::EventSource16<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;


    typedef sg::EventSource16<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 16;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16), Allocator>
    : trace_source_details::base,
      sg::EventSource17<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;


    typedef sg::EventSource17<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 17;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17), Allocator>
    : trace_source_details::base,
      sg::EventSource18<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;


    typedef sg::EventSource18<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 18;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18), Allocator>
    : trace_source_details::base,
      sg::EventSource19<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;


    typedef sg::EventSource19<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 19;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19), Allocator>
    : trace_source_details::base,
      sg::EventSource20<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type,typename trace_source_details::traits<T19>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;
    typedef typename trace_source_details::traits<T19>            tr_19_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;
    typedef typename tr_19_t::type                              _19_t;


    typedef sg::EventSource20<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t,_19_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 20;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;
    typedef typename tr_19_t::arg_type    arg19_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_,
        std::string const&       field_name19_,
        description const&       field_description19_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
        add_field( this, field_name19_,
                   enhance_trace_description(field_description19_, (arg19_type*)0),
                   ::trace_source_details::traits<arg19_type>::type_spec,
                   ::trace_source_details::traits<arg19_type>::size,
                   ::trace_source_details::traits<arg19_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_, arg19_type t19_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18, arg19_type t19 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20), Allocator>
    : trace_source_details::base,
      sg::EventSource21<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type,typename trace_source_details::traits<T19>::type,typename trace_source_details::traits<T20>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;
    typedef typename trace_source_details::traits<T19>            tr_19_t;
    typedef typename trace_source_details::traits<T20>            tr_20_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;
    typedef typename tr_19_t::type                              _19_t;
    typedef typename tr_20_t::type                              _20_t;


    typedef sg::EventSource21<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t,_19_t,_20_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 21;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;
    typedef typename tr_19_t::arg_type    arg19_type;
    typedef typename tr_20_t::arg_type    arg20_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_,
        std::string const&       field_name19_,
        description const&       field_description19_,
        std::string const&       field_name20_,
        description const&       field_description20_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
        add_field( this, field_name19_,
                   enhance_trace_description(field_description19_, (arg19_type*)0),
                   ::trace_source_details::traits<arg19_type>::type_spec,
                   ::trace_source_details::traits<arg19_type>::size,
                   ::trace_source_details::traits<arg19_type>::max_size );
        add_field( this, field_name20_,
                   enhance_trace_description(field_description20_, (arg20_type*)0),
                   ::trace_source_details::traits<arg20_type>::type_spec,
                   ::trace_source_details::traits<arg20_type>::size,
                   ::trace_source_details::traits<arg20_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_, arg19_type t19_, arg20_type t20_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18, arg19_type t19, arg20_type t20 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21), Allocator>
    : trace_source_details::base,
      sg::EventSource22<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type,typename trace_source_details::traits<T19>::type,typename trace_source_details::traits<T20>::type,typename trace_source_details::traits<T21>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;
    typedef typename trace_source_details::traits<T19>            tr_19_t;
    typedef typename trace_source_details::traits<T20>            tr_20_t;
    typedef typename trace_source_details::traits<T21>            tr_21_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;
    typedef typename tr_19_t::type                              _19_t;
    typedef typename tr_20_t::type                              _20_t;
    typedef typename tr_21_t::type                              _21_t;


    typedef sg::EventSource22<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t,_19_t,_20_t,_21_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 22;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;
    typedef typename tr_19_t::arg_type    arg19_type;
    typedef typename tr_20_t::arg_type    arg20_type;
    typedef typename tr_21_t::arg_type    arg21_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_,
        std::string const&       field_name19_,
        description const&       field_description19_,
        std::string const&       field_name20_,
        description const&       field_description20_,
        std::string const&       field_name21_,
        description const&       field_description21_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
        add_field( this, field_name19_,
                   enhance_trace_description(field_description19_, (arg19_type*)0),
                   ::trace_source_details::traits<arg19_type>::type_spec,
                   ::trace_source_details::traits<arg19_type>::size,
                   ::trace_source_details::traits<arg19_type>::max_size );
        add_field( this, field_name20_,
                   enhance_trace_description(field_description20_, (arg20_type*)0),
                   ::trace_source_details::traits<arg20_type>::type_spec,
                   ::trace_source_details::traits<arg20_type>::size,
                   ::trace_source_details::traits<arg20_type>::max_size );
        add_field( this, field_name21_,
                   enhance_trace_description(field_description21_, (arg21_type*)0),
                   ::trace_source_details::traits<arg21_type>::type_spec,
                   ::trace_source_details::traits<arg21_type>::size,
                   ::trace_source_details::traits<arg21_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_, arg19_type t19_, arg20_type t20_, arg21_type t21_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18, arg19_type t19, arg20_type t20, arg21_type t21 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22), Allocator>
    : trace_source_details::base,
      sg::EventSource23<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type,typename trace_source_details::traits<T19>::type,typename trace_source_details::traits<T20>::type,typename trace_source_details::traits<T21>::type,typename trace_source_details::traits<T22>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;
    typedef typename trace_source_details::traits<T19>            tr_19_t;
    typedef typename trace_source_details::traits<T20>            tr_20_t;
    typedef typename trace_source_details::traits<T21>            tr_21_t;
    typedef typename trace_source_details::traits<T22>            tr_22_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;
    typedef typename tr_19_t::type                              _19_t;
    typedef typename tr_20_t::type                              _20_t;
    typedef typename tr_21_t::type                              _21_t;
    typedef typename tr_22_t::type                              _22_t;


    typedef sg::EventSource23<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t,_19_t,_20_t,_21_t,_22_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 23;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;
    typedef typename tr_19_t::arg_type    arg19_type;
    typedef typename tr_20_t::arg_type    arg20_type;
    typedef typename tr_21_t::arg_type    arg21_type;
    typedef typename tr_22_t::arg_type    arg22_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_,
        std::string const&       field_name19_,
        description const&       field_description19_,
        std::string const&       field_name20_,
        description const&       field_description20_,
        std::string const&       field_name21_,
        description const&       field_description21_,
        std::string const&       field_name22_,
        description const&       field_description22_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
        add_field( this, field_name19_,
                   enhance_trace_description(field_description19_, (arg19_type*)0),
                   ::trace_source_details::traits<arg19_type>::type_spec,
                   ::trace_source_details::traits<arg19_type>::size,
                   ::trace_source_details::traits<arg19_type>::max_size );
        add_field( this, field_name20_,
                   enhance_trace_description(field_description20_, (arg20_type*)0),
                   ::trace_source_details::traits<arg20_type>::type_spec,
                   ::trace_source_details::traits<arg20_type>::size,
                   ::trace_source_details::traits<arg20_type>::max_size );
        add_field( this, field_name21_,
                   enhance_trace_description(field_description21_, (arg21_type*)0),
                   ::trace_source_details::traits<arg21_type>::type_spec,
                   ::trace_source_details::traits<arg21_type>::size,
                   ::trace_source_details::traits<arg21_type>::max_size );
        add_field( this, field_name22_,
                   enhance_trace_description(field_description22_, (arg22_type*)0),
                   ::trace_source_details::traits<arg22_type>::type_spec,
                   ::trace_source_details::traits<arg22_type>::size,
                   ::trace_source_details::traits<arg22_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_, arg19_type t19_, arg20_type t20_, arg21_type t21_, arg22_type t22_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18, arg19_type t19, arg20_type t20, arg21_type t21, arg22_type t22 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22) );
    }

    trace_source_t() : allocator(0) {}

};


template< typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9, typename T10, typename T11, typename T12, typename T13, typename T14, typename T15, typename T16, typename T17, typename T18, typename T19, typename T20, typename T21, typename T22, typename T23, typename Allocator >
struct trace_source_t<void(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9,T10,T11,T12,T13,T14,T15,T16,T17,T18,T19,T20,T21,T22,T23), Allocator>
    : trace_source_details::base,
      sg::EventSource24<typename trace_source_details::traits<T0>::type,typename trace_source_details::traits<T1>::type,typename trace_source_details::traits<T2>::type,typename trace_source_details::traits<T3>::type,typename trace_source_details::traits<T4>::type,typename trace_source_details::traits<T5>::type,typename trace_source_details::traits<T6>::type,typename trace_source_details::traits<T7>::type,typename trace_source_details::traits<T8>::type,typename trace_source_details::traits<T9>::type,typename trace_source_details::traits<T10>::type,typename trace_source_details::traits<T11>::type,typename trace_source_details::traits<T12>::type,typename trace_source_details::traits<T13>::type,typename trace_source_details::traits<T14>::type,typename trace_source_details::traits<T15>::type,typename trace_source_details::traits<T16>::type,typename trace_source_details::traits<T17>::type,typename trace_source_details::traits<T18>::type,typename trace_source_details::traits<T19>::type,typename trace_source_details::traits<T20>::type,typename trace_source_details::traits<T21>::type,typename trace_source_details::traits<T22>::type,typename trace_source_details::traits<T23>::type>
{
    typedef typename trace_source_details::traits<T0>            tr_0_t;
    typedef typename trace_source_details::traits<T1>            tr_1_t;
    typedef typename trace_source_details::traits<T2>            tr_2_t;
    typedef typename trace_source_details::traits<T3>            tr_3_t;
    typedef typename trace_source_details::traits<T4>            tr_4_t;
    typedef typename trace_source_details::traits<T5>            tr_5_t;
    typedef typename trace_source_details::traits<T6>            tr_6_t;
    typedef typename trace_source_details::traits<T7>            tr_7_t;
    typedef typename trace_source_details::traits<T8>            tr_8_t;
    typedef typename trace_source_details::traits<T9>            tr_9_t;
    typedef typename trace_source_details::traits<T10>            tr_10_t;
    typedef typename trace_source_details::traits<T11>            tr_11_t;
    typedef typename trace_source_details::traits<T12>            tr_12_t;
    typedef typename trace_source_details::traits<T13>            tr_13_t;
    typedef typename trace_source_details::traits<T14>            tr_14_t;
    typedef typename trace_source_details::traits<T15>            tr_15_t;
    typedef typename trace_source_details::traits<T16>            tr_16_t;
    typedef typename trace_source_details::traits<T17>            tr_17_t;
    typedef typename trace_source_details::traits<T18>            tr_18_t;
    typedef typename trace_source_details::traits<T19>            tr_19_t;
    typedef typename trace_source_details::traits<T20>            tr_20_t;
    typedef typename trace_source_details::traits<T21>            tr_21_t;
    typedef typename trace_source_details::traits<T22>            tr_22_t;
    typedef typename trace_source_details::traits<T23>            tr_23_t;

    typedef typename tr_0_t::type                              _0_t;
    typedef typename tr_1_t::type                              _1_t;
    typedef typename tr_2_t::type                              _2_t;
    typedef typename tr_3_t::type                              _3_t;
    typedef typename tr_4_t::type                              _4_t;
    typedef typename tr_5_t::type                              _5_t;
    typedef typename tr_6_t::type                              _6_t;
    typedef typename tr_7_t::type                              _7_t;
    typedef typename tr_8_t::type                              _8_t;
    typedef typename tr_9_t::type                              _9_t;
    typedef typename tr_10_t::type                              _10_t;
    typedef typename tr_11_t::type                              _11_t;
    typedef typename tr_12_t::type                              _12_t;
    typedef typename tr_13_t::type                              _13_t;
    typedef typename tr_14_t::type                              _14_t;
    typedef typename tr_15_t::type                              _15_t;
    typedef typename tr_16_t::type                              _16_t;
    typedef typename tr_17_t::type                              _17_t;
    typedef typename tr_18_t::type                              _18_t;
    typedef typename tr_19_t::type                              _19_t;
    typedef typename tr_20_t::type                              _20_t;
    typedef typename tr_21_t::type                              _21_t;
    typedef typename tr_22_t::type                              _22_t;
    typedef typename tr_23_t::type                              _23_t;


    typedef sg::EventSource24<_0_t,_1_t,_2_t,_3_t,_4_t,_5_t,_6_t,_7_t,_8_t,_9_t,_10_t,_11_t,_12_t,_13_t,_14_t,_15_t,_16_t,_17_t,_18_t,_19_t,_20_t,_21_t,_22_t,_23_t> base_type;
    typedef trace_source_details::description description;

    static unsigned const number_of_args = 24;
    typedef typename tr_0_t::arg_type    arg0_type;
    typedef typename tr_1_t::arg_type    arg1_type;
    typedef typename tr_2_t::arg_type    arg2_type;
    typedef typename tr_3_t::arg_type    arg3_type;
    typedef typename tr_4_t::arg_type    arg4_type;
    typedef typename tr_5_t::arg_type    arg5_type;
    typedef typename tr_6_t::arg_type    arg6_type;
    typedef typename tr_7_t::arg_type    arg7_type;
    typedef typename tr_8_t::arg_type    arg8_type;
    typedef typename tr_9_t::arg_type    arg9_type;
    typedef typename tr_10_t::arg_type    arg10_type;
    typedef typename tr_11_t::arg_type    arg11_type;
    typedef typename tr_12_t::arg_type    arg12_type;
    typedef typename tr_13_t::arg_type    arg13_type;
    typedef typename tr_14_t::arg_type    arg14_type;
    typedef typename tr_15_t::arg_type    arg15_type;
    typedef typename tr_16_t::arg_type    arg16_type;
    typedef typename tr_17_t::arg_type    arg17_type;
    typedef typename tr_18_t::arg_type    arg18_type;
    typedef typename tr_19_t::arg_type    arg19_type;
    typedef typename tr_20_t::arg_type    arg20_type;
    typedef typename tr_21_t::arg_type    arg21_type;
    typedef typename tr_22_t::arg_type    arg22_type;
    typedef typename tr_23_t::arg_type    arg23_type;


    Allocator*        allocator;

    void              set_allocator( Allocator* allocator_ ) { allocator = allocator_; }
    Allocator*        get_allocator() const                  { return allocator;       }

    void describe_fields( ... )
    {
        trace_source_details::error_wrong_number_of_fields_in_description<T0>::ok;
    }

    void describe_fields(
        std::string const&       field_name0_,
        description const&       field_description0_,
        std::string const&       field_name1_,
        description const&       field_description1_,
        std::string const&       field_name2_,
        description const&       field_description2_,
        std::string const&       field_name3_,
        description const&       field_description3_,
        std::string const&       field_name4_,
        description const&       field_description4_,
        std::string const&       field_name5_,
        description const&       field_description5_,
        std::string const&       field_name6_,
        description const&       field_description6_,
        std::string const&       field_name7_,
        description const&       field_description7_,
        std::string const&       field_name8_,
        description const&       field_description8_,
        std::string const&       field_name9_,
        description const&       field_description9_,
        std::string const&       field_name10_,
        description const&       field_description10_,
        std::string const&       field_name11_,
        description const&       field_description11_,
        std::string const&       field_name12_,
        description const&       field_description12_,
        std::string const&       field_name13_,
        description const&       field_description13_,
        std::string const&       field_name14_,
        description const&       field_description14_,
        std::string const&       field_name15_,
        description const&       field_description15_,
        std::string const&       field_name16_,
        description const&       field_description16_,
        std::string const&       field_name17_,
        description const&       field_description17_,
        std::string const&       field_name18_,
        description const&       field_description18_,
        std::string const&       field_name19_,
        description const&       field_description19_,
        std::string const&       field_name20_,
        description const&       field_description20_,
        std::string const&       field_name21_,
        description const&       field_description21_,
        std::string const&       field_name22_,
        description const&       field_description22_,
        std::string const&       field_name23_,
        description const&       field_description23_
        )
    {

        add_field( this, field_name0_,
                   enhance_trace_description(field_description0_, (arg0_type*)0),
                   ::trace_source_details::traits<arg0_type>::type_spec,
                   ::trace_source_details::traits<arg0_type>::size,
                   ::trace_source_details::traits<arg0_type>::max_size );
        add_field( this, field_name1_,
                   enhance_trace_description(field_description1_, (arg1_type*)0),
                   ::trace_source_details::traits<arg1_type>::type_spec,
                   ::trace_source_details::traits<arg1_type>::size,
                   ::trace_source_details::traits<arg1_type>::max_size );
        add_field( this, field_name2_,
                   enhance_trace_description(field_description2_, (arg2_type*)0),
                   ::trace_source_details::traits<arg2_type>::type_spec,
                   ::trace_source_details::traits<arg2_type>::size,
                   ::trace_source_details::traits<arg2_type>::max_size );
        add_field( this, field_name3_,
                   enhance_trace_description(field_description3_, (arg3_type*)0),
                   ::trace_source_details::traits<arg3_type>::type_spec,
                   ::trace_source_details::traits<arg3_type>::size,
                   ::trace_source_details::traits<arg3_type>::max_size );
        add_field( this, field_name4_,
                   enhance_trace_description(field_description4_, (arg4_type*)0),
                   ::trace_source_details::traits<arg4_type>::type_spec,
                   ::trace_source_details::traits<arg4_type>::size,
                   ::trace_source_details::traits<arg4_type>::max_size );
        add_field( this, field_name5_,
                   enhance_trace_description(field_description5_, (arg5_type*)0),
                   ::trace_source_details::traits<arg5_type>::type_spec,
                   ::trace_source_details::traits<arg5_type>::size,
                   ::trace_source_details::traits<arg5_type>::max_size );
        add_field( this, field_name6_,
                   enhance_trace_description(field_description6_, (arg6_type*)0),
                   ::trace_source_details::traits<arg6_type>::type_spec,
                   ::trace_source_details::traits<arg6_type>::size,
                   ::trace_source_details::traits<arg6_type>::max_size );
        add_field( this, field_name7_,
                   enhance_trace_description(field_description7_, (arg7_type*)0),
                   ::trace_source_details::traits<arg7_type>::type_spec,
                   ::trace_source_details::traits<arg7_type>::size,
                   ::trace_source_details::traits<arg7_type>::max_size );
        add_field( this, field_name8_,
                   enhance_trace_description(field_description8_, (arg8_type*)0),
                   ::trace_source_details::traits<arg8_type>::type_spec,
                   ::trace_source_details::traits<arg8_type>::size,
                   ::trace_source_details::traits<arg8_type>::max_size );
        add_field( this, field_name9_,
                   enhance_trace_description(field_description9_, (arg9_type*)0),
                   ::trace_source_details::traits<arg9_type>::type_spec,
                   ::trace_source_details::traits<arg9_type>::size,
                   ::trace_source_details::traits<arg9_type>::max_size );
        add_field( this, field_name10_,
                   enhance_trace_description(field_description10_, (arg10_type*)0),
                   ::trace_source_details::traits<arg10_type>::type_spec,
                   ::trace_source_details::traits<arg10_type>::size,
                   ::trace_source_details::traits<arg10_type>::max_size );
        add_field( this, field_name11_,
                   enhance_trace_description(field_description11_, (arg11_type*)0),
                   ::trace_source_details::traits<arg11_type>::type_spec,
                   ::trace_source_details::traits<arg11_type>::size,
                   ::trace_source_details::traits<arg11_type>::max_size );
        add_field( this, field_name12_,
                   enhance_trace_description(field_description12_, (arg12_type*)0),
                   ::trace_source_details::traits<arg12_type>::type_spec,
                   ::trace_source_details::traits<arg12_type>::size,
                   ::trace_source_details::traits<arg12_type>::max_size );
        add_field( this, field_name13_,
                   enhance_trace_description(field_description13_, (arg13_type*)0),
                   ::trace_source_details::traits<arg13_type>::type_spec,
                   ::trace_source_details::traits<arg13_type>::size,
                   ::trace_source_details::traits<arg13_type>::max_size );
        add_field( this, field_name14_,
                   enhance_trace_description(field_description14_, (arg14_type*)0),
                   ::trace_source_details::traits<arg14_type>::type_spec,
                   ::trace_source_details::traits<arg14_type>::size,
                   ::trace_source_details::traits<arg14_type>::max_size );
        add_field( this, field_name15_,
                   enhance_trace_description(field_description15_, (arg15_type*)0),
                   ::trace_source_details::traits<arg15_type>::type_spec,
                   ::trace_source_details::traits<arg15_type>::size,
                   ::trace_source_details::traits<arg15_type>::max_size );
        add_field( this, field_name16_,
                   enhance_trace_description(field_description16_, (arg16_type*)0),
                   ::trace_source_details::traits<arg16_type>::type_spec,
                   ::trace_source_details::traits<arg16_type>::size,
                   ::trace_source_details::traits<arg16_type>::max_size );
        add_field( this, field_name17_,
                   enhance_trace_description(field_description17_, (arg17_type*)0),
                   ::trace_source_details::traits<arg17_type>::type_spec,
                   ::trace_source_details::traits<arg17_type>::size,
                   ::trace_source_details::traits<arg17_type>::max_size );
        add_field( this, field_name18_,
                   enhance_trace_description(field_description18_, (arg18_type*)0),
                   ::trace_source_details::traits<arg18_type>::type_spec,
                   ::trace_source_details::traits<arg18_type>::size,
                   ::trace_source_details::traits<arg18_type>::max_size );
        add_field( this, field_name19_,
                   enhance_trace_description(field_description19_, (arg19_type*)0),
                   ::trace_source_details::traits<arg19_type>::type_spec,
                   ::trace_source_details::traits<arg19_type>::size,
                   ::trace_source_details::traits<arg19_type>::max_size );
        add_field( this, field_name20_,
                   enhance_trace_description(field_description20_, (arg20_type*)0),
                   ::trace_source_details::traits<arg20_type>::type_spec,
                   ::trace_source_details::traits<arg20_type>::size,
                   ::trace_source_details::traits<arg20_type>::max_size );
        add_field( this, field_name21_,
                   enhance_trace_description(field_description21_, (arg21_type*)0),
                   ::trace_source_details::traits<arg21_type>::type_spec,
                   ::trace_source_details::traits<arg21_type>::size,
                   ::trace_source_details::traits<arg21_type>::max_size );
        add_field( this, field_name22_,
                   enhance_trace_description(field_description22_, (arg22_type*)0),
                   ::trace_source_details::traits<arg22_type>::type_spec,
                   ::trace_source_details::traits<arg22_type>::size,
                   ::trace_source_details::traits<arg22_type>::max_size );
        add_field( this, field_name23_,
                   enhance_trace_description(field_description23_, (arg23_type*)0),
                   ::trace_source_details::traits<arg23_type>::type_spec,
                   ::trace_source_details::traits<arg23_type>::size,
                   ::trace_source_details::traits<arg23_type>::max_size );
    }

    void defer(arg0_type t0_, arg1_type t1_, arg2_type t2_, arg3_type t3_, arg4_type t4_, arg5_type t5_, arg6_type t6_, arg7_type t7_, arg8_type t8_, arg9_type t9_, arg10_type t10_, arg11_type t11_, arg12_type t12_, arg13_type t13_, arg14_type t14_, arg15_type t15_, arg16_type t16_, arg17_type t17_, arg18_type t18_, arg19_type t19_, arg20_type t20_, arg21_type t21_, arg22_type t22_, arg23_type t23_)
    {
        if (this->isActive())
        {
            if (allocator->is_deferring())
                ::trace_source_details::defer_fire( *get_allocator(), sg_meta::make_mem_fun_op(this, &trace_source_t::fire) )
                    .init(::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22_), ::trace_source_details::traits<arg23_type>::convert_for_trace(t23_));
            else
                this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0_), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1_), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2_), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3_), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4_), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5_), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6_), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7_), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8_), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9_), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10_), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11_), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12_), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13_), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14_), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15_), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16_), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17_), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18_), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19_), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20_), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21_), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22_), ::trace_source_details::traits<arg23_type>::convert_for_trace(t23_) );
        }
    }

    void operator()( arg0_type t0, arg1_type t1, arg2_type t2, arg3_type t3, arg4_type t4, arg5_type t5, arg6_type t6, arg7_type t7, arg8_type t8, arg9_type t9, arg10_type t10, arg11_type t11, arg12_type t12, arg13_type t13, arg14_type t14, arg15_type t15, arg16_type t16, arg17_type t17, arg18_type t18, arg19_type t19, arg20_type t20, arg21_type t21, arg22_type t22, arg23_type t23 )
    {
        this->fire( ::trace_source_details::traits<arg0_type>::convert_for_trace(t0), ::trace_source_details::traits<arg1_type>::convert_for_trace(t1), ::trace_source_details::traits<arg2_type>::convert_for_trace(t2), ::trace_source_details::traits<arg3_type>::convert_for_trace(t3), ::trace_source_details::traits<arg4_type>::convert_for_trace(t4), ::trace_source_details::traits<arg5_type>::convert_for_trace(t5), ::trace_source_details::traits<arg6_type>::convert_for_trace(t6), ::trace_source_details::traits<arg7_type>::convert_for_trace(t7), ::trace_source_details::traits<arg8_type>::convert_for_trace(t8), ::trace_source_details::traits<arg9_type>::convert_for_trace(t9), ::trace_source_details::traits<arg10_type>::convert_for_trace(t10), ::trace_source_details::traits<arg11_type>::convert_for_trace(t11), ::trace_source_details::traits<arg12_type>::convert_for_trace(t12), ::trace_source_details::traits<arg13_type>::convert_for_trace(t13), ::trace_source_details::traits<arg14_type>::convert_for_trace(t14), ::trace_source_details::traits<arg15_type>::convert_for_trace(t15), ::trace_source_details::traits<arg16_type>::convert_for_trace(t16), ::trace_source_details::traits<arg17_type>::convert_for_trace(t17), ::trace_source_details::traits<arg18_type>::convert_for_trace(t18), ::trace_source_details::traits<arg19_type>::convert_for_trace(t19), ::trace_source_details::traits<arg20_type>::convert_for_trace(t20), ::trace_source_details::traits<arg21_type>::convert_for_trace(t21), ::trace_source_details::traits<arg22_type>::convert_for_trace(t22), ::trace_source_details::traits<arg23_type>::convert_for_trace(t23) );
    }

    trace_source_t() : allocator(0) {}

};


#endif // SGCOMPONENTTRACEHELPER_H

